/*
 * RuntimeOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;

#if DATA
using System.Data;
#endif

using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Resources;

#if REMOTING
using System.Runtime.Remoting;
#endif

using System.Security;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Security.Policy;

#if !NATIVE
using System.Security.Principal;
#endif

using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Containers.Private;
using Eagle._Containers.Public;
using Eagle._Encodings;
using Eagle._Interfaces.Private;
using Eagle._Interfaces.Public;

#if !CONSOLE
using ConsoleColor = Eagle._Components.Public.ConsoleColor;
#endif

namespace Eagle._Components.Private
{
    [ObjectId("52155f4f-322b-4389-aacd-166fe334d164")]
    internal static class RuntimeOps
    {
        #region Private Data (MUST BE DONE PRIOR TO TOUCHING GlobalState)
        #region Synchronization
        //
        // BUGFIX: This is used by the IsTracingEnabled method, which is used
        //         by the DebugTrace method, which is used during the
        //         initialization of the GlobalState class; therefore, it must
        //         done before anything that touches the GlobalState class.
        //
        private static readonly object syncRoot = new object();
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Tracing
        //
        // HACK: This is part of a hack that solves a chicken-and-egg problem
        //       with the diagnostic tracing method used by this library.  We
        //       allow tracing to be disabled via an environment variable
        //       and/or the shell command line.  Unfortunately, by the time we
        //       disable tracing, many messages will have typically already
        //       been written to the trace listeners.  To prevent this noise
        //       (that the user wishes to suppress), we internalize the check
        //       (i.e. we do it from inside the core trace method itself) and
        //       initialize this variable [once] with the result of checking
        //       the environment variable.
        //
        private static bool? isTracingEnabled = null;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Constants
        #region Object Disposal
        private static readonly bool DefaultDispose = true;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Type Conversion
        private static readonly string systemSimpleName =
            AssemblyOps.GetFullName(typeof(object)); /* mscorlib */

        private static readonly string[] systemNamespaces = {
            /* System */
            typeof(object).Namespace
        };

        ///////////////////////////////////////////////////////////////////////

        private static readonly string assemblySimpleName =
            GlobalState.GetAssemblyFullName(); /* Eagle */

        private static readonly string[] assemblyNamespaces = {
            /* Eagle._Attributes */
            // typeof(AssemblyDateTimeAttribute).Namespace,

            /* Eagle._Components.Public */
            typeof(Engine).Namespace,

            /* Eagle._Containers.Public */
            typeof(ArgumentList).Namespace,

            /* Eagle._Encodings */
            // typeof(_Encodings.OneByteEncoding).Namespace,

            /* Eagle._Interfaces.Public */
            // typeof(IClientData).Namespace,
        };
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Interactive Loop
        internal static readonly string InteractiveCommandPrefix =
            Characters.NumberSign.ToString();
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Result Formatting
        public static readonly string emptyFormat = String.Empty;

        private const string codeOnlyFormat = "{0}{1}";
        private const string resultOnlyFormat = "{0}{2}";
        private const string codeAndResultFormat = "{0}{1}: {2}";
        private const string codeAndErrorLineFormat = "{0}{1}, line {3}";
        private const string codeResultAndErrorLineFormat = "{0}{1}, line {3}: {2}";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Property Value Defaults
        internal static readonly bool DefaultThrowOnFeatureNotSupported = true;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Binding Flags
        internal const BindingFlags CreateInstanceBindingFlags =
            BindingFlags.CreateInstance;

        internal const BindingFlags DelegateBindingFlags = BindingFlags.Instance |
            BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Miscellaneous Formatting
        private const string CertificateSubjectFormat = " - {0}";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Script Variable Handling
        private static readonly string DefaultVariableValue = null;
        private static readonly string DefaultGetVariableValue = String.Empty;
        private static readonly string DefaultSetVariableValue = null;
        private static readonly string DefaultUnsetVariableValue = null;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Script Update Checking
        //
        // NOTE: These procedure names are all RESERVED; however, they may
        //       legally be redefined to do nothing.
        //
        private const string CheckForUpdateScript = "checkForUpdate";
        private const string FetchUpdateScript = "fetchUpdate";
        private const string RunUpdateAndExitScript = "runUpdateAndExit";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Wait Handling
        private static readonly int WaitDivisor = 2;
        private static readonly int WaitMaximumSleepTime = 50; /* in milliseconds */

        private static readonly int WaitSlopDivisor = 40;
        private static readonly int WaitSlopMinimumTime = 25000;  /* in microseconds */
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Encoding Constants
        //
        // WARNING: Do not change this as it must be a pass-through one-byte
        //          per character encoding.
        //
        private static readonly Encoding RawEncoding = OneByteEncoding.OneByte;
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Hash Algorithm Constants
        //
        // NOTE: *WARNING* Change this value with great care because it may
        //       break custom script, file, and stream policies that rely on
        //       the hash result.
        //
        private static readonly string DefaultHashAlgorithm = "SHA1";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Resource Handling
        private const string InvalidInterpreterResourceManager =
            "invalid interpreter resource manager";

        private const string InvalidPluginResourceManager =
            "invalid plugin resource manager";
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Public Constants
        public static readonly string HostNoFeatureError =
            "interpreter host lacks support for the \"{0}\" feature";
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Private Data
        #region Native Stack Checking
#if NATIVE && WINDOWS
        private static LocalDataStoreSlot stackPtrSlot; /* ThreadSpecificData */
        private static LocalDataStoreSlot stackSizeSlot; /* ThreadSpecificData */
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Process Support Methods
        public static int GetCurrentProcessId()
        {
            try
            {
                Process process = Process.GetCurrentProcess();

                if (process != null)
                    return process.Id;
            }
            catch
            {
                // do nothing.
            }

            return 0;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Thread Support Methods
        public static bool IsStaThread()
        {
            Thread thread = Thread.CurrentThread;

            if (thread == null)
                return false;

            return (thread.GetApartmentState() == ApartmentState.STA);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Native Stack Checking Support Methods
        public static ReturnCode GetStackSize(
            ref UIntPtr used,
            ref UIntPtr allocated,
            ref UIntPtr extra,
            ref UIntPtr margin,
            ref UIntPtr maximum,
            ref UIntPtr reserve,
            ref UIntPtr commit,
            ref Result error
            ) /* ENTRY-POINT, THREAD-SAFE */
        {
#if NATIVE && WINDOWS
            lock (syncRoot)
            {
                if (stackSizeSlot != null)
                {
                    /* THREAD-SAFE, per-thread data */
                    NativeStack.StackSize stackSize = Thread.GetData(
                        stackSizeSlot) as NativeStack.StackSize; /* throw */

                    if (stackSize != null)
                    {
                        used = stackSize.used;
                        allocated = stackSize.allocated;
                        extra = stackSize.extra;
                        margin = stackSize.margin;
                        maximum = stackSize.maximum;
                        reserve = stackSize.reserve;
                        commit = stackSize.commit;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "thread stack size is invalid";
                        return ReturnCode.Error;
                    }
                }
                else
                {
                    error = "thread stack size slot is invalid";
                    return ReturnCode.Error;
                }
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        public static void InitializeStackChecking()
        {
            lock (syncRoot)
            {
                #region Native Stack Checking Thread Local Storage
                //
                // NOTE: These MUST to be done prior to evaluating any scripts
                //       or runtime stack checking will not work properly
                //       (which can potentially cause scripts that use deep
                //       recursion to cause a .NET exception to be thrown from
                //       the script engine itself because the script engine
                //       depends upon runtime stack checking working properly).
                //
                if (stackPtrSlot == null)
                    stackPtrSlot = Thread.AllocateDataSlot();

                if (stackSizeSlot == null)
                    stackSizeSlot = Thread.AllocateDataSlot();
                #endregion
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void FinalizeStackChecking()
        {
            lock (syncRoot)
            {
                //
                // NOTE: Dispose the cached stack pointer and size information
                //       for this thread.  It is "mostly harmless" to do this
                //       even if is still required by another interpreter in
                //       this thread because it will automatically re-created
                //       in that case.  The alternative is to never dispose of
                //       this data.
                //
                if (stackPtrSlot != null)
                {
                    object stackPtrData = Thread.GetData(stackPtrSlot); /* throw */

                    if (stackPtrData != null)
                    {
                        //
                        // NOTE: Remove our local reference to the data.
                        //
                        stackPtrData = null;

                        //
                        // NOTE: Clear out the data value for this thread.
                        //
                        Thread.SetData(stackPtrSlot, stackPtrData); /* throw */
                    }
                }

                ///////////////////////////////////////////////////////////////

                if (stackSizeSlot != null)
                {
                    object stackSizeData = Thread.GetData(stackSizeSlot); /* throw */

                    if (stackSizeData != null)
                    {
                        //
                        // NOTE: Remove our local reference to the data.
                        //
                        stackSizeData = null;

                        //
                        // NOTE: Clear out the data value for this thread.
                        //
                        Thread.SetData(stackSizeSlot, stackSizeData); /* throw */
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode CheckForStackSpace(
            Interpreter interpreter,
            ulong extraSpace
            ) /* THREAD-SAFE */
        {
            try
            {
                if (PlatformOps.IsWindowsOperatingSystem())
                {
                    lock (syncRoot)
                    {
                        //
                        // NOTE: Make sure we have our process-wide thread data slots.
                        //       We do not and cannot actually allocate or create them
                        //       here.  We are only reading these static variables and
                        //       we KNOW they are allocated during interpreter creation;
                        //       therefore, we do not need a lock around access to them.
                        //
                        if ((stackPtrSlot != null) && (stackSizeSlot != null))
                        {
                            //
                            // NOTE: Get the address of the local variable (so that we
                            //       know where in the stack we currently are.  The
                            //       commented-out code here sorta-kinda works; however,
                            //       checking the native stack pointer appears to be far
                            //       more reliable.
                            //
                            // GCHandle handle = GCHandle.Alloc(
                            //     reserved, GCHandleType.Pinned);
                            //
                            // UIntPtr innerStackPtr =
                            //     Conversion.ToUIntPtr(handle.AddrOfPinnedObject());
                            //
                            // handle.Free();
                            //
                            UIntPtr innerStackPtr = NativeStack.GetNativeStackPointer();

                            //
                            // NOTE: Attempt to get the intial address of the local
                            //       variable that we previously saved, if any.
                            //
                            /* THREAD-SAFE, per-thread data */
                            object objOuterThreadData = Thread.GetData(
                                stackPtrSlot); /* throw */

                            //
                            // NOTE: Now, if we got the value from the thread data slot,
                            //       it should be an IntPtr; otherwise, set the outer
                            //       value to zero (first time through) so that it will
                            //       be set to match the inner value (below).
                            //
                            UIntPtr outerStackPtr;

                            if (objOuterThreadData is UIntPtr)
                                outerStackPtr = (UIntPtr)objOuterThreadData;
                            else
                                outerStackPtr = UIntPtr.Zero;

                            //
                            // NOTE: If it was not previously saved, save it now.
                            //
                            if (outerStackPtr == UIntPtr.Zero)
                            {
                                //
                                // NOTE: First time through, set the outer value equal
                                //       to the inner (current) value.
                                //
                                outerStackPtr = innerStackPtr;

                                /* THREAD-SAFE, per-thread data */
                                Thread.SetData(stackPtrSlot, outerStackPtr); /* throw */
                            }

                            //
                            // NOTE: Make sure we got valid values for the stack bounds.
                            //
                            if ((outerStackPtr != UIntPtr.Zero) &&
                                (innerStackPtr != UIntPtr.Zero))
                            {
                                //
                                // NOTE: Get the stack size object for this thread.  If
                                //       it is invalid or has not been created yet, we
                                //       will create or reset it now.
                                //
                                /* THREAD-SAFE, per-thread data */
                                NativeStack.StackSize stackSize = Thread.GetData(
                                    stackSizeSlot) as NativeStack.StackSize; /* throw */

                                //
                                // NOTE: If it was not previously saved, save it now.
                                //
                                if (stackSize == null)
                                {
                                    stackSize = new NativeStack.StackSize();

                                    /* THREAD-SAFE, per-thread data */
                                    Thread.SetData(stackSizeSlot, stackSize); /* throw */
                                }

                                //
                                // NOTE: If necessary, copy the data about the current
                                //       executable (PE) file into the stack size for
                                //       future reference (e.g. reserve may be used
                                //       below).
                                //
                                if ((stackSize.reserve == UIntPtr.Zero) ||
                                    (stackSize.commit == UIntPtr.Zero))
                                {
                                    NativeStack.StackSize fileStackSize =
                                        FileOps.PeFileStackSize;

                                    if (fileStackSize != null)
                                    {
                                        if (stackSize.reserve == UIntPtr.Zero)
                                            stackSize.reserve = fileStackSize.reserve;

                                        if (stackSize.commit == UIntPtr.Zero)
                                            stackSize.commit = fileStackSize.commit;
                                    }
                                }

                                UIntPtr usedSpace;

                                //
                                // NOTE: Detect which way the stack is growing...
                                //
                                if (outerStackPtr.ToUInt64() > innerStackPtr.ToUInt64())
                                    usedSpace = new UIntPtr(
                                        outerStackPtr.ToUInt64() - innerStackPtr.ToUInt64());
                                else
                                    usedSpace = new UIntPtr(
                                        innerStackPtr.ToUInt64() - outerStackPtr.ToUInt64());

                                //
                                // NOTE: Update stack size object for this thread with the
                                //       amount of used space.
                                //
                                stackSize.used = usedSpace;

                                //
                                // NOTE: Update stack size object for this thread with the
                                //       amount of space currently allocated (because this
                                //       number grows automatically within the actual stack
                                //       limits, it is useless for the actual stack check
                                //       and is only used for informational purposes).
                                //
                                stackSize.allocated = NativeStack.GetNativeStackAllocated();

                                //
                                // NOTE: Update stack size object for this thread with the
                                //       requested amount of extra space.
                                //
                                stackSize.extra = new UIntPtr(extraSpace);

                                //
                                // NOTE: Calculate the approximate safety margin (overhead)
                                //       imposed by the .NET Framework runtime.  This is
                                //       merely an estimate and may need to be updated for
                                //       later versions of the .NET Framework.  Since this
                                //       number is currently constant for the lifetime of
                                //       the process, we calculate it once and then cache
                                //       it.
                                //
                                if (stackSize.margin == UIntPtr.Zero)
                                    stackSize.margin = NativeStack.GetNativeStackMargin();

                                UIntPtr stackMargin = stackSize.margin;

                                //
                                // NOTE: Get the current amount of stack reserved for this
                                //       thread from the Thread Environment Block (TEB).
                                //       Since it is highly unlikely that this number will
                                //       change during the lifetime of the thread, we cache
                                //       it.
                                //
                                /* THREAD-SAFE, per-thread data */
                                if (stackSize.maximum == UIntPtr.Zero)
                                    stackSize.maximum = NativeStack.GetNativeStackSize();

                                UIntPtr maximumSpace = stackSize.maximum;

                                //
                                // NOTE: Failing that, fallback on the stack reserve from the
                                //       PE file that started this process.
                                //
                                if (maximumSpace == UIntPtr.Zero)
                                    maximumSpace = stackSize.reserve;

                                //
                                // NOTE: Is the maximum stack size valid?
                                //
                                if (maximumSpace != UIntPtr.Zero)
                                {
                                    //
                                    // NOTE: Calculate the amount of space used with the safety
                                    //       margintaken into account.
                                    //
                                    UIntPtr usedSpaceWithMargin = new UIntPtr(
                                        usedSpace.ToUInt64() + extraSpace + stackMargin.ToUInt64());

                                    //
                                    // NOTE: Are we "out of stack space" taking our safety margin
                                    //       into account?  Also, it seems that some ThreadPool
                                    //       threads have a miserably low stack size (less than our
                                    //       safety margin); therefore, evaluating scripts on
                                    //       ThreadPool threads is not officially supported.
                                    //
                                    if (usedSpaceWithMargin.ToUInt64() <= maximumSpace.ToUInt64())
                                    {
                                        //
                                        // NOTE: Normal case, everything is verified to be ok.
                                        //
                                        return ReturnCode.Ok;
                                    }
                                    else
                                    {
                                        //
                                        // NOTE: We hit a "soft" stack-overflow error.  This error is
                                        //       guaranteed by the script engine to be non-fatal to
                                        //       the process, the application domain, and the script
                                        //       engine itself, and is fully recoverable.
                                        //
                                        DebugTrace(String.Format(
                                            "{0}: CheckForStackSpace: stack overflow, used space with " +
                                            "margin {1} is greater than maximum space {2}{3}",
                                            GlobalState.GetCurrentThreadId(), usedSpaceWithMargin,
                                            maximumSpace, Environment.NewLine), typeof(RuntimeOps).Name);

                                        return ReturnCode.Error;
                                    }
                                }
                                else
                                {
                                    //
                                    // NOTE: If we made it this far and do not have a
                                    //       valid maximum stack size, just return
                                    //       success.
                                    //
                                    return ReturnCode.Ok;
                                }
                            }
                            else
                            {
                                //
                                // NOTE: Runtime native stack checking is unavailable,
                                //       just assume everything is ok.
                                //
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Our process-wide data slot was not allocated or was
                            //       freed prematurely?  Just assume that runtime stack
                            //       checking was purposely disabled and that everything
                            //       is ok.
                            //
                            return ReturnCode.Ok;
                        }
                    }
                }
                else
                {
                    //
                    // NOTE: We are not running on Windows; therefore, we have no
                    //       native stack checking available.
                    //
                    return ReturnCode.Ok;
                }
            }
            catch (StackOverflowException)
            {
                //
                // NOTE: We hit a "hard" stack-overflow during the stack
                //       checking code?  This error should be non-fatal
                //       to the process, the application domain, and the
                //       script engine, and should also be fully recoverable;
                //       however, this is not guaranteed by the script engine
                //       as we are relying on the underlying .NET Framework
                //       stack unwinding semantics to function properly.
                //
                try
                {
                    //
                    // NOTE: We really want to report this condition to anybody
                    //       who might be listening; however, it is somewhat
                    //       dangerous to do so.  Therefore, wrap this block in
                    //       a try/catch just in case we re-trigger another stack
                    //       overflow.
                    //
                    DebugTrace(String.Format(
                        "{0}: CheckForStackSpace: caught stack overflow exception{1}",
                        GlobalState.GetCurrentThreadId(), Environment.NewLine),
                        typeof(RuntimeOps).Name);
                }
                catch (StackOverflowException)
                {
                    // do nothing.
                }

                return ReturnCode.Error;
            }
            catch (SecurityException)
            {
                //
                // NOTE: We may not be allowed to execute any native code;
                //       therefore, just assume that we always have enough
                //       stack space in that case.
                //
                return ReturnCode.Ok;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region AppDomain / Remoting Support Methods
        #region Dead Code
#if DEAD_CODE
        public static bool IsPrimaryAppDomain(
            Interpreter interpreter,
            AppDomain appDomain
            )
        {
            return (interpreter != null) && (appDomain != null) &&
                Object.ReferenceEquals(appDomain, interpreter.GetAppDomain());
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsPrimaryAppDomain(
            AppDomain appDomain
            )
        {
            return (appDomain != null) &&
                Object.ReferenceEquals(appDomain, AppDomain.CurrentDomain);
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        public static bool IsTransparentProxy(object proxy)
        {
#if REMOTING
            return RemotingServices.IsTransparentProxy(proxy);
#else
            return false;
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsDefaultAppDomain()
        {
            return IsDefaultAppDomain(AppDomain.CurrentDomain);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsDefaultAppDomain(
            AppDomain appDomain
            )
        {
            return ((appDomain != null) && appDomain.IsDefaultAppDomain());
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsSameAppDomain(
            Interpreter interpreter
            )
        {
            if (interpreter == null)
                return true;

            AppDomain appDomain = interpreter.GetAppDomain();

            if ((appDomain == null) ||
                (appDomain.Id == AppDomain.CurrentDomain.Id))
            {
                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

#if ISOLATED_PLUGINS
        public static bool IsPluginIsolated(
            IPlugin plugin
            )
        {
            return (plugin != null) &&
                FlagOps.HasFlags(plugin.Flags, PluginFlags.Isolated, true);
        }

        ///////////////////////////////////////////////////////////////////////

        private static OptionFlags GetEnumOptionFlags(
            TypeCode typeCode,
            bool strict
            )
        {
            switch (typeCode)
            {
                case TypeCode.Byte:
                case TypeCode.SByte:
                case TypeCode.Int16:
                case TypeCode.UInt16:
                case TypeCode.Int32:
                case TypeCode.UInt32:
                    return OptionFlags.MustBeInteger;
                case TypeCode.Int64:
                case TypeCode.UInt64:
                    return OptionFlags.MustBeWideInteger;
                default:
                    return strict ? OptionFlags.None :
                        OptionFlags.MustBeWideInteger;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // HACK: This is a horrible hack to workaround the issue with not being
        //       able to use plugin enumerated types from the "primary"
        //       application domain when the plugin has been loaded in isolated
        //       mode.
        //
        public static ReturnCode FixupOptions(
            IPlugin plugin,
            OptionDictionary options,
            bool strict,
            ref Result error
            )
        {
            if (plugin == null)
            {
                if (strict)
                {
                    error = "invalid plugin";
                    return ReturnCode.Error;
                }

                return ReturnCode.Ok;
            }

            if (options == null)
            {
                if (strict)
                {
                    error = "invalid options";
                    return ReturnCode.Error;
                }

                return ReturnCode.Ok;
            }

            if (!IsPluginIsolated(plugin))
                return ReturnCode.Ok;

            Assembly assembly = plugin.Assembly;

            foreach (KeyValuePair<string, IOption> pair in options)
            {
                IOption option = pair.Value;

                if (option == null)
                    continue;

                //
                // HACK: Skip options that do not have enumerated types.
                //       For now, these are the only options we really have to
                //       worry about because they are the only ones that can
                //       directly refer to user-defined types [of any kind].
                //
                if (!option.HasFlags(OptionFlags.MustBeEnum, true))
                    continue;

                //
                // NOTE: Grab the enumerated (?) type and figure out if it
                //       came from the plugin assembly.  If not, ignore it and
                //       continue.
                //
                Type type = option.Type;

                if ((type == null) || !type.IsEnum ||
                    !Object.ReferenceEquals(type.Assembly, assembly))
                {
                    continue;
                }

                //
                // NOTE: Get the current value of the option.
                //
                object oldValue = option.InnerValue;
                TypeCode typeCode = TypeCode.Empty;

                //
                // NOTE: Attempt to get the new value for the integral type for
                //       the enumeration value of this option, if any.  We must
                //       do this even if the original value is null because we
                //       must have the type code to properly reset the option
                //       flags.
                //
                object newValue = EnumOps.ConvertToTypeCodeValue(
                    type, (oldValue != null) ? oldValue : 0, ref typeCode,
                    ref error);

                if (newValue == null)
                    return ReturnCode.Error;

                //
                // NOTE: Get the option flags required for the integral type.
                //
                OptionFlags flags = GetEnumOptionFlags(typeCode, strict);

                if (flags == OptionFlags.None)
                {
                    error = String.Format(
                        "unsupported type code for enumerated type \"{0}\"",
                        type);

                    return ReturnCode.Error;
                }

                //
                // NOTE: Special handling for "flags" enumerations here.
                //
                if (type.IsDefined(typeof(FlagsAttribute), false))
                {
                    //
                    // HACK: Substitute our placeholder flags enumerated type.
                    //       It does not know about the textual values provided
                    //       by the actual enumerated type; however, at least
                    //       they can use the custom flags enumeration handling
                    //       (i.e. the "+" and "-" operators, etc).
                    //
                    option.Type = typeof(StubFlagsEnum);
                }
                else
                {
                    //
                    // NOTE: Remove the MustBeEnum flag for this option and add
                    //       the flag(s) needed for its integral type.
                    //
                    option.Flags &= ~OptionFlags.MustBeEnum;
                    option.Flags |= flags;

                    //
                    // NOTE: Clear the type for the option.  The type property
                    //       is only meaningful for enumeration-based options
                    //       and we are converting this option to use some kind
                    //       of integral type.
                    //
                    option.Type = null;
                }

                //
                // NOTE: If necessary, set the new [default] value for this
                //       option to the one we converted to an integral type
                //       value above.  If the old (original) value was null, we
                //       just discard the new value which will be zero anyhow.
                //
                option.Value = (oldValue != null) ? new Variant(newValue) : null;
            }

            return ReturnCode.Ok;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetOrCreatePluginAppDomain(
            Interpreter interpreter,
            string friendlyName,
            string packagePath,
#if CAS_POLICY
            Evidence evidence,
#endif
            IClientData clientData,
            bool isolated,
            bool useBasePath,
            ref AppDomain appDomain,
            ref Result error
            )
        {
            //
            // NOTE: Use an isolated application domain or the current one?
            //
            if (isolated)
            {
#if ISOLATED_PLUGINS
                //
                // BUGBUG: This feature does not currently work due to
                //         cross-domain marshalling issues.
                //
                return CreateAppDomain(
                    friendlyName, packagePath,
#if CAS_POLICY
                    evidence,
#endif
                    clientData, useBasePath, ref appDomain, ref error);
#else
                error = "not implemented";
#endif
            }
            else if (interpreter != null)
            {
                //
                // NOTE: Get the application domain configured for this
                //       interpreter.
                //
                appDomain = interpreter.GetAppDomain();

                if (appDomain != null)
                {
                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid application domain";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

#if ISOLATED_INTERPRETERS
        public static ReturnCode CreateTestAppDomain(
            string friendlyName,
            string packagePath,
            IClientData clientData, /* NOT USED */
            ref AppDomain appDomain,
            ref Result error
            )
        {
            return CreateAppDomain(
                friendlyName, packagePath,
#if CAS_POLICY
                null,
#endif
                clientData, true, ref appDomain, ref error);
        }
#endif

        ///////////////////////////////////////////////////////////////////////

#if APPDOMAINS || ISOLATED_INTERPRETERS || ISOLATED_PLUGINS
        public static AppDomainSetup CreateAppDomainSetup(
            string packagePath,
            bool useBasePath,
            ref Result error
            )
        {
            string basePath = useBasePath ?
                GlobalState.GetBasePath(GlobalState.GetAssembly()) : null;

            if (!useBasePath || !String.IsNullOrEmpty(basePath))
            {
                if ((packagePath == null) || !useBasePath ||
                    PathOps.IsUnderPath(packagePath, basePath))
                {
                    AppDomainSetup appDomainSetup = new AppDomainSetup();

                    //
                    // NOTE: Grab the path for this assembly.
                    //
                    string assemblyPath = GlobalState.GetAssemblyPath();

                    //
                    // NOTE: Use the base directory of the Eagle install as
                    //       the isolated application domain base directory.
                    //
                    appDomainSetup.ApplicationBase = useBasePath ?
                        basePath : (packagePath != null) ?
                            packagePath : assemblyPath;

                    //
                    // NOTE: If we are using the base path of this assembly,
                    //       then we need to modify the private binary path
                    //       so that it includes both the directory containing
                    //       this assembly and the directory containing the
                    //       package; otherwise, we can simply skip this step.
                    //
                    if (useBasePath)
                    {
                        //
                        // TODO: May need to adjust more options here.
                        //
                        string relativeAssemblyPath = (assemblyPath != null) ?
                            assemblyPath.Remove(
                                0, basePath.Length).Trim(PathOps.PathChars) : null;

                        string relativePackagePath = (packagePath != null) ?
                            packagePath.Remove(
                                0, basePath.Length).Trim(PathOps.PathChars) : null;

                        string privateBinPath = relativeAssemblyPath;

                        if (!String.IsNullOrEmpty(relativePackagePath))
                        {
                            if (!String.IsNullOrEmpty(privateBinPath))
                                privateBinPath += Characters.SemiColon;

                            privateBinPath += relativePackagePath;
                        }

                        appDomainSetup.PrivateBinPath = privateBinPath;
                    }

                    return appDomainSetup;
                }
                else
                {
                    error = "package path is not under base path";
                }
            }
            else
            {
                error = "invalid base path";
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode CreateAppDomain(
            string friendlyName,
            string packagePath,
#if CAS_POLICY
            Evidence evidence,
#endif
            IClientData clientData, /* NOT USED */
            bool useBasePath,
            ref AppDomain appDomain,
            ref Result error
            )
        {
            //
            // NOTE: *WARNING* Empty application domain names are allowed,
            //       please do not change this to "!String.IsNullOrEmpty".
            //
            if (friendlyName != null)
            {
                try
                {
                    AppDomainSetup appDomainSetup = CreateAppDomainSetup(
                        packagePath, useBasePath, ref error);

                    if (appDomainSetup != null)
                    {
                        appDomain = AppDomain.CreateDomain(
                            friendlyName,
#if CAS_POLICY
                            evidence,
#else
                            null,
#endif
                            appDomainSetup);

                        return ReturnCode.Ok;
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid friendly name";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode UnloadAppDomain(
            AppDomain appDomain,
            IClientData clientData, /* NOT USED */
            ref Result error
            )
        {
            if (appDomain != null)
            {
                try
                {
                    AppDomain.Unload(appDomain); /* throw */

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "invalid application domain";
            }

            //
            // NOTE: We do not really expect this method to fail;
            //       therefore, output a complaint about it.
            //
            DebugOps.Complain(ReturnCode.Error, error);

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Profiling Support Methods
        public static void StartProfiling(
            ref long startCount,    /* out */
            ref double microseconds /* in, out */
            )
        {
            microseconds = 0;

            startCount = PerformanceOps.GetCount();
        }

        ///////////////////////////////////////////////////////////////////////

        public static void StopProfiling(
            long startCount,        /* in */
            ref long stopCount,     /* out */
            ref double microseconds /* in, out */
            )
        {
            stopCount = PerformanceOps.GetCount();

            microseconds += PerformanceOps.GetMicroseconds(
                startCount, stopCount, 1);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Tracing Support Methods
        //
        // HACK: This is part of a hack that solves a chicken-and-egg problem
        //       with the diagnostic tracing method used by this library.  We
        //       allow tracing to be disabled via an environment variable
        //       and/or the shell command line.  Unfortunately, by the time we
        //       disable tracing, many messages will have typically already
        //       been written to the trace listeners.  To prevent this noise
        //       (that the user wishes to suppress), we internalize the check
        //       (i.e. we do it from inside the core trace method itself) and
        //       initialize this variable [once] with the result of checking
        //       the environment variable.
        //
        private static bool IsTracingEnabled()
        {
            bool result;

            //
            // NOTE: Grab the static data lock to prevent a race condition.
            //
            lock (syncRoot)
            {
                if (isTracingEnabled == null)
                {
                    //
                    // NOTE: We cannot use GlobalState.GetEnvironmentVariable
                    //       here because that method calls the DebugTrace
                    //       method (below), which calls us.
                    //
                    isTracingEnabled = (CommonOps.GetEnvironmentVariable(
                        EnvVars.NoTrace) == null);
                }

                result = (bool)isTracingEnabled;
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        [Conditional("DEBUG_WRITE")]
        public static void DebugWrite(
            Interpreter interpreter,
            string value,
            bool force
            )
        {
            DebugOps.Write(interpreter, String.Format("{0} ({1}): {2}",
                GlobalState.GetCurrentThreadId(), interpreter, value), force);
        }

        ///////////////////////////////////////////////////////////////////////

        [Conditional("DEBUG_TRACE")]
        public static void DebugTrace(
            string message,
            string category
            )
        {
            //
            // TODO: Redirect these writes to the active IHost, if any?
            //
            if (IsTracingEnabled())
                DebugOps.TraceWrite(message, category);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode WriteViaIExecute(
            Interpreter interpreter,
            string commandName, /* NOTE: Almost always null, for [puts]. */
            string channelId,   /* NOTE: Almost always null, for "stdout". */
            string value,
            ref Result result
            )
        {
            if (interpreter == null)
            {
                result = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (commandName == null)
                commandName = typeof(_Commands.Puts).Name.ToLowerInvariant();

            if (channelId == null)
                channelId = StandardChannel.Output;

            ReturnCode code;
            IExecute execute = null;

            code = interpreter.GetIExecuteViaResolvers(
                interpreter.GetResolveEngineFlags(true), commandName,
                null, true, true, ref execute, ref result);

            if (code != ReturnCode.Ok)
                return code;

            code = Engine.ExternalExecuteWithFrame(
                commandName, execute, interpreter, null,
                new ArgumentList(channelId, value), interpreter.EngineFlags,
                interpreter.SubstitutionFlags, interpreter.EngineEventFlags,
                interpreter.ExpressionFlags, ref result);

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Debugger Support Methods
        public static bool GetDebuggerBreakOnCancel(
            Interpreter interpreter,
            bool ignoreEnabled
            )
        {
#if DEBUGGER
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    IDebugger debugger = interpreter.Debugger;

                    if ((debugger != null) &&
                        (ignoreEnabled || debugger.Enabled))
                    {
                        return debugger.BreakOnCancel;
                    }
                }
            }
#endif

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetDebuggerCommand(
            Interpreter interpreter,
            bool ignoreEnabled
            )
        {
            string result = null;

#if DEBUGGER
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    IDebugger debugger = interpreter.Debugger;

                    if ((debugger != null) &&
                        (ignoreEnabled || debugger.Enabled))
                    {
                        //
                        // NOTE: Enforce "one-time" semantics.
                        //
                        result = debugger.Command;
                        debugger.Command = null;
                    }
                }
            }
#endif

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static HeaderFlags GetDebuggerHeaderFlags(
            HeaderFlags headerFlags,
            bool debug,
            bool show,
            bool empty,
            bool @default
            )
        {
            //
            // NOTE: If we are in debug mode and no header display flags have
            //       been explicitly set for the interpreter, initialize them
            //       to the default value.
            //
            if (@default && FlagOps.HasFlags(
                    headerFlags, HeaderFlags.Invalid, true))
            {
                //
                // NOTE: Remove the "these flags have not been setup before"
                //       indicator flag.
                //
                headerFlags &= ~HeaderFlags.Invalid;

                //
                // NOTE: Add the default header flags.
                //
                headerFlags |= HeaderFlags.Default;
            }

            //
            // NOTE: Only modify (set or unset) the active debugger flag if we
            //       have been told to do so; otherwise, the active debugger
            //       flag may have been manually changed and should be left
            //       alone.
            //
            if (show)
            {
                //
                // NOTE: Is there an active debugger?
                //
                if (debug)
                {
                    //
                    // NOTE: Set the active debugger flag.
                    //
                    headerFlags |= HeaderFlags.Debug;
                }
                else
                {
                    //
                    // NOTE: Unset the active debugger flag.
                    //
                    headerFlags &= ~HeaderFlags.Debug;
                }
            }

            //
            // NOTE: Show empty content?
            //
            if (empty)
                headerFlags |= HeaderFlags.EmptyContent;

            return headerFlags;
        }

        ///////////////////////////////////////////////////////////////////////

        public static EngineFlags GetDebuggerEngineFlags(
            bool debug
            )
        {
#if DEBUGGER
            if (debug)
                return Debugger.InteractiveEngineFlags;
#endif

            return EngineFlags.Interactive;
        }

        ///////////////////////////////////////////////////////////////////////

#if DEBUGGER
        public static bool CanHitBreakpoints(
            EngineFlags engineFlags,
            Interpreter interpreter
            )
        {
            return !Engine.HasNoBreakpoint(engineFlags) &&
                (interpreter != null) && interpreter.CanHitBreakpoints();
        }

        ///////////////////////////////////////////////////////////////////////

        public static IDebugger NewDebugger(
            bool isolated,
            string culture,
            CreateFlags createFlags,
            AppDomain appDomain,
            IHost host
            )
        {
            return new Debugger(isolated, culture, createFlags, appDomain, host);
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static bool SetDebuggerResult(
            Interpreter interpreter,
            ReturnCode code,
            Result result,
            bool ignoreEnabled
            )
        {
#if DEBUGGER
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    IDebugger debugger = interpreter.Debugger;

                    if ((debugger != null) &&
                        (ignoreEnabled || debugger.Enabled))
                    {
                        //
                        // NOTE: Enforce "copy" semantics.
                        //
                        debugger.Result = Result.Copy(code, result, true); /* COPY */

                        return true;
                    }
                }
            }
#endif

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Host Support Methods
        public static void Sleep(
            int milliseconds
            ) /* THREAD-SAFE */
        {
#if MONO_BUILD
#pragma warning disable 219
#endif
            ReturnCode code; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

            Result error = null;

            code = Sleep(milliseconds, ref error);

#if DEBUG && VERBOSE
            if (code != ReturnCode.Ok)
                DebugOps.Complain(code, error);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static void Sleep(
            Interpreter interpreter,
            int milliseconds
            ) /* THREAD-SAFE */
        {
            IEngineHost engineHost = null;

            if (interpreter != null)
            {
                lock (interpreter.syncRoot) /* TRANSACTIONAL */
                {                           /* FIELD */
                    //
                    // BUGFIX: Prevent a race condition between grabbing the
                    //         host and the interpreter being disposed.  This
                    //         is necessary because we are called in the
                    //         critical code path of the Wait and WaitVariable
                    //         methods.
                    //
                    if (!Interpreter.IsDisposed(interpreter))
                        engineHost = interpreter.Host;
                }
            }

#if MONO_BUILD
#pragma warning disable 219
#endif
            ReturnCode code; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

            Result error = null;

            code = HostSleep(engineHost, milliseconds, false, ref error);

#if DEBUG && VERBOSE
            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, error);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode Sleep(
            int milliseconds,
            ref Result error
            ) /* THREAD-SAFE */
        {
            Exception exception = null;

            return Sleep(milliseconds, ref exception, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode Sleep(
            int milliseconds,
            ref Exception exception,
            ref Result error
            ) /* THREAD-SAFE */
        {
            try
            {
                Thread.Sleep(milliseconds);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                exception = e;
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void WriteHostOrConsole(
            IInteractiveHost interactiveHost,
            string value
            )
        {
            if (interactiveHost != null)
                interactiveHost.Write(value);
#if CONSOLE
            else
                Console.Write(value);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static void WriteLineHostOrConsole(
            IInteractiveHost interactiveHost,
            string value
            )
        {
            if (interactiveHost != null)
                interactiveHost.WriteLine(value);
#if CONSOLE
            else
                Console.WriteLine(value);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static void WriteConsoleOrComplain(
            ReturnCode code,
            Result result
            )
        {
            WriteConsoleOrComplain(code, result, 0);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void WriteConsoleOrComplain(
            ReturnCode code,
            Result result,
            int errorLine
            )
        {
#if CONSOLE
            Console.WriteLine(FormatResult(code, result, errorLine));
#else
            DebugOps.Complain(code, result);
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsHostOpen(
            Interpreter interpreter,
            ref IInteractiveHost interactiveHost
            )
        {
            if (interpreter != null)
            {
                try
                {
                    interactiveHost = interpreter.Host; /* throw */

                    if (interactiveHost != null)
                        return interactiveHost.IsOpen(); /* throw */
                }
                catch (Exception e)
                {
                    DebugTrace(String.Format(
                        "{0}: {1}{2}",
                        GlobalState.GetCurrentThreadId(),
                        e, Environment.NewLine),
                        typeof(RuntimeOps).Name);
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetHostColors(
            IHost host,
            string name,
            bool foreground,
            bool background,
            bool strict,
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor,
            ref Result error
            )
        {
            ReturnCode code;

            //
            // NOTE: Is the interpreter host available (to make color decisions)?
            //
            if (host != null)
            {
                try
                {
                    //
                    // NOTE: If a "Null"-typed interpreter host is being used or
                    //       the host does not support colors, just skip this step.
                    //
                    if (!IsNullHostType(host.GetType()) &&
                        FlagOps.HasFlags(host.GetFlags(), HostFlags.AllColors, false))
                    {
                        code = host.GetColors(
                            null, name, foreground, background,
                            ref foregroundColor, ref backgroundColor,
                            ref error);
                    }
                    else if (strict)
                    {
                        error = String.Format(
                            HostNoFeatureError, HostFlags.AllColors);

                        code = ReturnCode.Error;
                    }
                    else
                    {
                        code = ReturnCode.Ok;
                    }
                }
                catch (Exception e)
                {
                    error = e;
                    code = ReturnCode.Error;
                }
            }
            else if (strict)
            {
                error = "interpreter host not available";
                code = ReturnCode.Error;
            }
            else
            {
                code = ReturnCode.Ok;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetHostScript(
            Interpreter interpreter,
            IEngineHost engineHost,
            string name,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            if (!HasNoHost(flags, ref result))
            {
                if (engineHost != null)
                {
                    try
                    {
                        HostFlags hostFlags = engineHost.GetFlags();

#if ISOLATED_PLUGINS
                        //
                        // HACK: If the current interpreter host is running
                        //       in an isolated application domain, use the
                        //       "backup" core host instead.
                        //
                        if (FlagOps.HasFlags(
                                hostFlags, HostFlags.Isolated, true))
                        {
                            IEngineHost coreEngineHost = interpreter.CoreHost;

                            if (coreEngineHost != null)
                            {
                                HostFlags coreHostFlags = coreEngineHost.GetFlags();

                                if (FlagOps.HasFlags(
                                        coreHostFlags, HostFlags.Script, true))
                                {
                                    return coreEngineHost.GetScript(
                                        name, ref flags, ref result); /* throw */
                                }
                            }
                        }
#endif

                        if (FlagOps.HasFlags(
                                hostFlags, HostFlags.Script, true))
                        {
                            return engineHost.GetScript(
                                name, ref flags, ref result); /* throw */
                        }
                        else
                        {
                            result = "interpreter host does not have script support";
                        }
                    }
                    catch (Exception e)
                    {
                        result = e;
                    }
                }
                else
                {
                    result = "interpreter host not available";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetHostStream(
            Interpreter interpreter,
            IEngineHost engineHost,
            string path,
            FileMode mode,
            FileAccess access,
            FileShare share,
            int bufferSize,
            FileOptions options,
            bool resolve,
            bool strict,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            )
        {
            if (engineHost != null)
            {
                try
                {
                    HostFlags hostFlags = engineHost.GetFlags();

#if ISOLATED_PLUGINS
                    //
                    // HACK: If the current interpreter host is running
                    //       in an isolated application domain, use the
                    //       "backup" core host instead.
                    //
                    if (FlagOps.HasFlags(
                            hostFlags, HostFlags.Isolated, true))
                    {
                        IEngineHost coreEngineHost = interpreter.CoreHost;

                        if (coreEngineHost != null)
                        {
                            HostFlags coreHostFlags = coreEngineHost.GetFlags();

                            if (FlagOps.HasFlags(
                                    coreHostFlags, HostFlags.Stream, true))
                            {
                                return coreEngineHost.GetStream(
                                    path, mode, access, share, bufferSize,
                                    options, resolve, ref fullPath, ref stream,
                                    ref error);
                            }
                        }
                    }
#endif

                    if (FlagOps.HasFlags(
                            engineHost.GetFlags(), HostFlags.Stream, true))
                    {
                        return engineHost.GetStream(
                            path, mode, access, share, bufferSize, options,
                            resolve, ref fullPath, ref stream, ref error);
                    }
                    else if (strict)
                    {
                        error = "interpreter host does not have stream support";
                    }
                    else
                    {
                        return NewFileStream(
                            path, mode, access, share, bufferSize, options,
                            resolve, ref fullPath, ref stream, ref error);
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = "interpreter host not available";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode HostSleep(
            IEngineHost engineHost,
            int milliseconds,
            bool strict,
            ref Result error
            )
        {
            if (engineHost != null)
            {
                try
                {
                    if (FlagOps.HasFlags(
                            engineHost.GetFlags(), HostFlags.Sleep, true))
                    {
                        if (engineHost.Sleep(milliseconds))
                            return ReturnCode.Ok;
                        else
                            error = "host sleep failed";
                    }
                    else if (strict)
                    {
                        error = String.Format(
                            HostNoFeatureError, HostFlags.Sleep);
                    }
                    else
                    {
                        return Sleep(milliseconds, ref error);
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else if (strict)
            {
                error = "interpreter host not available";
            }
            else
            {
                return Sleep(milliseconds, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool HasNoHost(
            ScriptFlags flags,
            ref Result error
            )
        {
            if (FlagOps.HasFlags(flags, ScriptFlags.NoHost, true))
            {
                error = "forbidden from getting script from host";

                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsNullHostType(
            Type type
            )
        {
            return (type != null) &&
                ((type == typeof(_Hosts.Null)) ||
                type.IsSubclassOf(typeof(_Hosts.Null)));
        }

        ///////////////////////////////////////////////////////////////////////

        public static IHost NewCustomHost(
            NewHostCallback callback,
            Interpreter interpreter,
            ResourceManager resourceManager,
            string profile,
            CreateFlags createFlags
            )
        {
            if (callback == null)
                return null;

            IHost host = callback(new HostData(
                null, null, null, ClientData.Empty,
                null, interpreter, resourceManager, profile,
                FlagOps.HasFlags(createFlags, CreateFlags.NoColor, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoTitle, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoIcon, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoProfile, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoCancel, true)));

            if (host != null)
            {
                //
                // NOTE: Dynamic fixup.  Since this host was created via the
                //       new host callback delegate, it will [probably] not
                //       have a valid type name; therefore, attempt to see if
                //       this host derives from the core host and then check
                //       the type name and fill it in now, if necessary.
                //
                _Hosts.Core coreHost = host as _Hosts.Core;

                if ((coreHost != null) && (coreHost.TypeName == null))
                    coreHost.TypeName = coreHost.GetType().Name;
            }

            return host;
        }

        ///////////////////////////////////////////////////////////////////////

#if CONSOLE
        public static IHost NewConsoleHost(
            Interpreter interpreter,
            ResourceManager resourceManager,
            string profile,
            CreateFlags createFlags
            )
        {
            return new _Hosts.Console(new HostData(
                null, null, null, ClientData.Empty,
                typeof(_Hosts.Console).Name,
                interpreter, resourceManager, profile,
                FlagOps.HasFlags(createFlags, CreateFlags.NoColor, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoTitle, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoIcon, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoProfile, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoCancel, true)));
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static IHost NewDiagnosticHost(
            Interpreter interpreter,
            ResourceManager resourceManager,
            string profile,
            CreateFlags createFlags
            )
        {
            return new _Hosts.Diagnostic(new HostData(
                null, null, null, ClientData.Empty,
                typeof(_Hosts.Diagnostic).Name,
                interpreter, resourceManager, profile,
                FlagOps.HasFlags(createFlags, CreateFlags.NoColor, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoTitle, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoIcon, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoProfile, true),
                FlagOps.HasFlags(createFlags, CreateFlags.NoCancel, true)));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Namespace Support Methods
        public static bool HasAbsoluteNamespace(
            string name
            )
        {
            return !String.IsNullOrEmpty(name) &&
                (String.Compare(name, 0, TclVars.GlobalNamespace, 0,
                    TclVars.GlobalNamespace.Length,
                    StringOps.SystemStringComparisonType) == 0);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsGlobalNamespace(
            string name
            )
        {
            return String.IsNullOrEmpty(name) ||
                (String.Compare(name, TclVars.GlobalNamespace,
                    StringOps.SystemStringComparisonType) == 0);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Regular Expression Support Methods
        public static string RegsubCountMatchCallback(
            Match match
            )
        {
            string result = null;

            IAnyPair<Interpreter, IClientData> pair =
                Interpreter.GetActivePair(typeof(RegsubClientData));

            if (pair != null)
            {
                Interpreter interpreter = pair.X;
                RegsubClientData regsubClientData = pair.Y as RegsubClientData;

                if ((interpreter != null) && (regsubClientData != null))
                {
                    //
                    // NOTE: Keep track of how many matches we have been given.
                    //
                    regsubClientData.Count++;

                    //
                    // NOTE: Get the additional parameters we need to perform
                    //       the callback from the clientData.
                    //
                    string pattern = regsubClientData.Pattern;
                    string input = regsubClientData.Input;
                    string replacement = regsubClientData.Replacement;
                    bool quote = regsubClientData.Quote;
                    bool args = regsubClientData.Args;
                    bool strict = regsubClientData.Strict;

                    //
                    // NOTE: Perform our custom replacements and return the
                    //       result.
                    //
                    result = TranslateRegSubSpec(pattern, input, replacement,
                        replacement, quote, args, strict, match);
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string RegsubEvaluateMatchCallback(
            Match match
            )
        {
            Result result = null;

            IAnyPair<Interpreter, IClientData> pair =
                Interpreter.GetActivePair(typeof(RegsubClientData));

            if (pair != null)
            {
                Interpreter interpreter = pair.X;
                RegsubClientData regsubClientData = pair.Y as RegsubClientData;

                if ((interpreter != null) && (regsubClientData != null))
                {
                    //
                    // NOTE: Keep track of how many matches we have been given.
                    //
                    regsubClientData.Count++;

                    //
                    // NOTE: Get the script to evaluate from the clientData.
                    //
                    string text = regsubClientData.Text;
                    IScriptLocation textLocation = regsubClientData.TextLocation;

                    if (!String.IsNullOrEmpty(text))
                    {
                        //
                        // NOTE: Get the additional parameters we need to
                        //       perform the script callback from the
                        //       clientData.
                        //
                        string pattern = regsubClientData.Pattern;
                        string input = regsubClientData.Input;
                        string replacement = regsubClientData.Replacement;
                        bool quote = regsubClientData.Quote;
                        bool args = regsubClientData.Args;
                        bool strict = regsubClientData.Strict;

                        //
                        // NOTE: Special processing to pass data to the script
                        //       to be evaluated.
                        //
                        StringList list = null;

                        //
                        // WARNING: Cannot cache list representation here, the
                        //          list is modified below.
                        //
                        ReturnCode code = Parser.SplitList(
                            interpreter, text, 0, Length.Invalid,
                            false, false, ref list, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            for (int index = 0; index < list.Count; index++)
                                list[index] = TranslateRegSubSpec(
                                    pattern, input, replacement,
                                    list[index], quote, args, strict, match);

                            text = list.ToString();

                            code = interpreter.EvaluateScript(text, textLocation, ref result);
                        }

                        if (code != ReturnCode.Ok)
                        {
                            Engine.AddErrorInformation(interpreter, result,
                                String.Format("{0}    (-regsub command)",
                                    Environment.NewLine));

                            //
                            // NOTE: This is our only way out of here.  This
                            //       exception will be caught by the command
                            //       handler for regsub and converted into a
                            //       script error.
                            //
                            throw new ScriptException(code, result);
                        }
                    }
                    else
                    {
                        //
                        // NOTE: This is allowed, translate a null or empty
                        //       script into an empty result.
                        //
                        result = String.Empty;
                    }
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string TranslateRegSubSpec(
            string pattern,     // the original pattern string.
            string input,       // the original input string.
            string replacement, // the original replacement string.
            string text,        // string containing the subSpecs to process.
            bool quote,         // use list element quoting?
            bool args,          // use of \P, \I, and \S permitted in subSpec.
            bool strict,        // use for strict conformance to the Tcl docs.
            Match match         // the current Regex match, if any.
            )
        {
            string result;

            if (!String.IsNullOrEmpty(text))
            {
                //
                // NOTE: We handle the following Tcl compatible regsub
                //       subSpecs:
                //
                //       "\&"          : Always replaced with a literal
                //                       ampersand.
                //
                //       "\\"          : Always replaced with a literal
                //                       backslash.
                //
                //       "&"           : Translated to "$&" OR replaced with
                //                       the portion of string that matched
                //                       exp if a valid match was supplied.
                //
                //       "\n" (0 to 9) : Translated to "$n", OR replaced with
                //                       the portion of string that matched the
                //                       Nth parenthesized subexpression of exp,
                //                       except for "\0" (which will be treated
                //                       just like "&") if a valid match was
                //                       supplied.
                //
                //       We also handle the following custom extensions
                //       typically used with "regsub -eval" (these are ONLY
                //       recognized if the "args" parameter is true; otherwise,
                //       they are ignored):
                //
                //       "\P"          : Always replaced with the original
                //                       pattern.  ONLY recognized if the
                //                       "args" parameter is true.
                //
                //       "\I"          : Always replaced with the original
                //                       input string.  ONLY recognized if
                //                       the "args" parameter is true.
                //
                //       "\S"          : Always replaced with the original
                //                       replacement (subSpec).  ONLY
                //                       recognized if the "args" parameter
                //                       is true.
                //
                //       Anything else will be passed through verbatim.
                //
                StringBuilder builder = StringOps.NewStringBuilder();

                for (int index = 0; index < text.Length; index++)
                {
                    char character = text[index];

                    switch (character)
                    {
                        case Characters.Ampersand:
                            {
                                //
                                // NOTE: If a match was supplied, replace this with
                                //       the entire matched value.
                                //
                                if (match != null)
                                {
                                    //
                                    // NOTE: Append the value of the entire match,
                                    //       quoting it if requested.
                                    //
                                    builder.Append(quote ?
                                        Parser.Quote(match.Value) :
                                        match.Value);
                                }
                                else
                                {
                                    //
                                    // NOTE: Translate an unescaped ampersand to the
                                    //       .NET Framework equivalent, which includes
                                    //       a dollar sign prefix and represents the
                                    //       entired matched expression.  An example
                                    //       is "&" to "$&".
                                    //
                                    builder.Append(
                                        Characters.DollarSign.ToString() +
                                        character.ToString());
                                }
                                break;
                            }
                        case Characters.Backslash:
                            {
                                //
                                // NOTE: Are there more characters remaining after
                                //       this one?
                                //
                                if ((index + 1) < text.Length)
                                {
                                    //
                                    // NOTE: Something has been escaped with a
                                    //       backslash, we need to look at the next
                                    //       character.
                                    //
                                    char nextCharacter = text[index + 1];

                                    switch (nextCharacter)
                                    {
                                        case Characters.Ampersand:
                                            {
                                                //
                                                // NOTE: We hit an escaped ampersand,
                                                //       insert the literal ampersand.
                                                //
                                                builder.Append(nextCharacter);
                                                break;
                                            }
                                        case Characters.Backslash:
                                            {
                                                //
                                                // NOTE: We hit an escaped backslash,
                                                //       insert the literal backslash.
                                                //
                                                builder.Append(nextCharacter);
                                                break;
                                            }
                                        case Characters.P:
                                            {
                                                //
                                                // NOTE: This feature is not supported
                                                //       in strict mode.
                                                //
                                                if (!args) goto default;

                                                //
                                                // NOTE: Append the original "exp"
                                                //       (pattern string) argument.
                                                //
                                                builder.Append(pattern);
                                                break;
                                            }
                                        case Characters.I:
                                            {
                                                //
                                                // NOTE: This feature is not supported
                                                //       in strict mode.
                                                //
                                                if (!args) goto default;

                                                //
                                                // NOTE: Append the original "string"
                                                //       (input string) argument.
                                                //
                                                builder.Append(input);
                                                break;
                                            }
                                        case Characters.S:
                                            {
                                                //
                                                // NOTE: This feature is not supported
                                                //       in strict mode.
                                                //
                                                if (!args) goto default;

                                                //
                                                // NOTE: Append the original "subSpec"
                                                //       (replacement string) argument.
                                                //
                                                builder.Append(replacement);
                                                break;
                                            }
                                        default:
                                            {
                                                //
                                                // NOTE: Is this a backslash followed by
                                                //       a digit?  If so, we need to
                                                //       append the applicable match
                                                //       group value, if any.
                                                //
                                                if (StringOps.CharIsAsciiDigit(nextCharacter))
                                                {
                                                    if (match != null)
                                                    {
                                                        int groupIndex = nextCharacter - Characters.Zero;

                                                        if (groupIndex < match.Groups.Count)
                                                        {
                                                            //
                                                            // NOTE: Append the value of
                                                            //       the match group,
                                                            //       quoting it if requested.
                                                            //
                                                            builder.Append(quote ?
                                                                Parser.Quote(match.Groups[groupIndex].Value) :
                                                                match.Groups[groupIndex].Value);
                                                        }
                                                    }
                                                    else
                                                    {
                                                        //
                                                        // NOTE: We hit a properly escaped subSpec,
                                                        //       insert the .NET Framework equivalent,
                                                        //       which includes a dollar sign prefix.
                                                        //       An example is "\1" to "$1".
                                                        //
                                                        builder.Append(
                                                            Characters.DollarSign.ToString() +
                                                            nextCharacter.ToString());
                                                    }
                                                }
                                                else
                                                {
                                                    if (strict)
                                                    {
                                                        //
                                                        // BUGFIX: No, we do not actually want to do that.
                                                        //         Even though this portion of the subSpec
                                                        //         pattern argument handling for [regsub]
                                                        //         is poorly specified in the Tcl
                                                        //         documentation, what we actually need to
                                                        //         do here is insert the literal backslash
                                                        //         followed by the literal character we
                                                        //         just encountered.
                                                        //
                                                        //         The exact rule is as follows:
                                                        //
                                                        //         "Any backslash in the subSpec pattern
                                                        //          argument NOT followed by an ampersand,
                                                        //          a single decimal digit, or another
                                                        //          backslash is treated as a literal
                                                        //          backslash."
                                                        //
                                                        //         As a consequence of the above rule, any
                                                        //         backslash followed by any character NOT
                                                        //         covered by the above rule will be
                                                        //         inserted into the output string literally.
                                                        //
                                                        builder.Append(character);
                                                        builder.Append(nextCharacter);
                                                    }
                                                    else
                                                    {
                                                        //
                                                        // NOTE: We hit an escaped character we do not
                                                        //       recognize, just insert the unescaped
                                                        //       equivalent.
                                                        //
                                                        builder.Append(nextCharacter);
                                                    }
                                                }
                                                break;
                                            }
                                    }

                                    //
                                    // NOTE: Now, skip beyond the escaped character
                                    //       we just handled.
                                    //
                                    index++;
                                }
                                else
                                {
                                    //
                                    // NOTE: We hit an isolated backslash at the
                                    //       end of the string, just insert a
                                    //       literal backslash.
                                    //
                                    builder.Append(character);
                                }
                                break;
                            }
                        default:
                            {
                                //
                                // NOTE: Anything we do not handle gets added
                                //       verbatim.
                                //
                                builder.Append(character);
                                break;
                            }
                    }
                }

                result = builder.ToString();
            }
            else
            {
                //
                // NOTE: Garbage in, garbage out.
                //
                result = text;
            }

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Hash Algorithm Support Methods
        public static byte[] HashScript(
            string algorithm,
            IScript script,
            Encoding encoding,
            ref Result error
            )
        {
            try
            {
                if (algorithm == null)
                    algorithm = DefaultHashAlgorithm;

                using (HashAlgorithm hashAlgorithm =
                        HashAlgorithm.Create(algorithm))
                {
                    ByteList bytes = new ByteList();

                    if (script != null)
                    {
                        string value = script.Text;

                        if (value != null)
                        {
                            if (encoding != null)
                                bytes.AddRange(encoding.GetBytes(value));
                            else
                                bytes.AddRange(RawEncoding.GetBytes(value));
                        }
                    }

                    hashAlgorithm.Initialize();

                    return hashAlgorithm.ComputeHash(bytes.ToArray());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static byte[] HashFile(
            string algorithm,
            string fileName,
            Encoding encoding,
            ref Result error
            )
        {
            try
            {
                if (!String.IsNullOrEmpty(fileName))
                {
                    if (!PathOps.IsRemoteUri(fileName))
                    {
                        if (File.Exists(fileName))
                        {
                            if (algorithm == null)
                                algorithm = DefaultHashAlgorithm;

                            using (HashAlgorithm hashAlgorithm =
                                    HashAlgorithm.Create(algorithm))
                            {
                                ByteList bytes = new ByteList();

                                if (encoding != null)
                                {
                                    bytes.AddRange(encoding.GetBytes(
                                        File.ReadAllText(
                                            fileName, encoding)));
                                }
                                else
                                {
                                    bytes.AddRange(File.ReadAllBytes(
                                        fileName));
                                }

                                hashAlgorithm.Initialize();

                                return hashAlgorithm.ComputeHash(
                                    bytes.ToArray());
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "couldn't read file \"{0}\": " +
                                "no such file or directory",
                                fileName);
                        }
                    }
                    else
                    {
                        error = "remote uri not supported";
                    }
                }
                else
                {
                    error = "invalid file name";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static byte[] HashString(
            string algorithm,
            string value,
            Encoding encoding,
            ref Result error
            )
        {
            try
            {
                if (algorithm == null)
                    algorithm = DefaultHashAlgorithm;

                using (HashAlgorithm hashAlgorithm =
                        HashAlgorithm.Create(algorithm))
                {
                    ByteList bytes = new ByteList();

                    if (value != null)
                    {
                        if (encoding != null)
                            bytes.AddRange(encoding.GetBytes(value));
                        else
                            bytes.AddRange(RawEncoding.GetBytes(value));
                    }

                    hashAlgorithm.Initialize();

                    return hashAlgorithm.ComputeHash(bytes.ToArray());
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return null;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Security Support Methods
#if !NATIVE
        private static ReturnCode IsAdministrator(
            ref bool administrator,
            ref Result error
            )
        {
            try
            {
                //
                // BUGBUG: This does not work properly on Mono due to their
                //         lack of support for checking the elevation status of
                //         the current process (i.e. it returns true even when
                //         running without elevation).
                //
                WindowsIdentity identity = WindowsIdentity.GetCurrent();

                administrator = (identity != null)
                    ? new WindowsPrincipal(identity).IsInRole(
                        WindowsBuiltInRole.Administrator) :
                    false;

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        public static bool IsAdministrator()
        {
#if NATIVE
            //
            // BUGBUG: This fails when running on Mono for Windows due to the
            //         bug that prevents native functions from being called by
            //         ordinal (e.g. "#680").
            //         https://bugzilla.novell.com/show_bug.cgi?id=636966
            //
            return SecurityOps.IsAdministrator();
#else
            //
            // BUGBUG: This does not work properly on Mono due to their lack of
            //         support for checking the elevation status of the current
            //         process (i.e. it returns true even when running without
            //         elevation).
            //
            bool administrator = false;
            Result error = null;

            return (IsAdministrator(ref administrator,
                ref error) == ReturnCode.Ok) && administrator;
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool ShouldCheckStrongNameVerified()
        {
            return (CommonOps.GetEnvironmentVariable(
                EnvVars.NoVerified) == null);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsStrongNameVerified(
            string fileName,
            bool force
            )
        {
#if NATIVE
            return SecurityOps.IsStrongNameVerified(fileName, force);
#else
            //
            // FIXME: Find some pure-managed way to do this?
            //
            return false;
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool ShouldCheckFileTrusted()
        {
            return (CommonOps.GetEnvironmentVariable(
                EnvVars.NoTrusted) == null);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsFileTrusted(
            string fileName
            )
        {
            return IsFileTrusted(fileName, false, false, true, false);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsFileTrusted(
            string fileName,
            bool userInterface,
            bool userPrompt,
            bool revocation,
            bool install
            )
        {
#if NATIVE
            return SecurityOps.IsFileTrusted(
                fileName, userInterface, userPrompt, revocation, install);
#else
            //
            // FIXME: Find some pure-managed way to do this?
            //
            // NOTE: We could use the AuthenticodeSignatureInformation
            //       class if we took a dependency on the .NET Framework
            //       3.5+; however, that would still be very unlikely to
            //       work on Mono.
            //
            return false;
#endif
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetCertificateSubject(
            Assembly assembly,
            bool trusted
            )
        {
            if (trusted && (assembly != null))
            {
                X509Certificate2 certificate =
                    AssemblyOps.GetCertificate2(assembly);

                string fileName = assembly.Location;

                if ((certificate != null) && IsFileTrusted(fileName))
                {
                    return String.Format(CertificateSubjectFormat,
                        certificate.GetNameInfo(X509NameType.SimpleName, false));
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static StringList CertificateToList(
            X509Certificate certificate,
            bool verbose
            )
        {
            StringList list = null;

            if (certificate != null)
            {
                list = new StringList();
                list.Add("subject", certificate.Subject);

                if (verbose)
                {
                    list.Add("issuer", certificate.Issuer);

                    list.Add("serialNumber",
                        certificate.GetSerialNumberString());

                    list.Add("hash",
                        certificate.GetCertHashString());
                    list.Add("effectiveDate",
                        certificate.GetEffectiveDateString());

                    list.Add("expirationDate",
                        certificate.GetExpirationDateString());

                    list.Add("algorithm",
                        certificate.GetKeyAlgorithm());

                    list.Add("algorithmParameters",
                        certificate.GetKeyAlgorithmParametersString());
                }
            }

            return list;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Shell Support Methods
        public static string BuildCommandLine(
            IEnumerable<string> args,
            bool quoteAll
            )
        {
            StringBuilder result = StringOps.NewStringBuilder();

            if (args != null)
            {
                char space = Characters.Space;
                char quote = Characters.QuotationMark;
                char slash = Characters.Backslash;
                char[] special = { space, quote, slash };

                foreach (string arg in args)
                {
                    if (arg != null)
                    {
                        if (result.Length > 0)
                            result.Append(space);

                        bool wrap = quoteAll ||
                            (arg.IndexOfAny(special) != Index.Invalid);

                        if (wrap)
                            result.Append(quote);

                        int length = arg.Length;

                        for (int index = 0; index < length; index++)
                        {
                            if (arg[index] == quote)
                            {
                                result.Append(slash);
                                result.Append(quote);
                            }
                            else if (arg[index] == slash)
                            {
                                int count = 0;

                                while ((index < length) &&
                                    (arg[index] == slash))
                                {
                                    count++; index++;
                                }

                                if (index < length)
                                {
                                    if (arg[index] == quote)
                                    {
                                        result.Append(slash, (count * 2) + 1);
                                        result.Append(quote);
                                    }
                                    else
                                    {
                                        result.Append(slash, count);
                                        result.Append(arg[index]);
                                    }
                                }
                                else
                                {
                                    result.Append(slash, count * 2);
                                    break;
                                }
                            }
                            else
                            {
                                result.Append(arg[index]);
                            }
                        }

                        if (wrap)
                            result.Append(quote);
                    }
                }
            }

            return result.ToString();
        }

        ///////////////////////////////////////////////////////////////////////

#if SHELL
        public static TextReader GetTextReaderForFile(
            Encoding encoding,
            string fileName,
            ref Result error
            )
        {
            try
            {
#if CONSOLE
                if (String.Equals(fileName,
                        CommandLineArgument.StandardInput,
                        StringOps.SystemNoCaseStringComparisonType) ||
                    String.Equals(fileName, StandardChannel.Input,
                        StringOps.SystemNoCaseStringComparisonType))
                {
                    //
                    // TODO: Allow the interpreter host (if available) to be
                    //       used here instead?
                    //
                    return Console.In;
                }
                else
#endif
                {
                    return (encoding != null) ?
                        new StreamReader(fileName, encoding) :
                        new StreamReader(fileName);
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ReadArgumentsFromFile(
            Interpreter interpreter,
            TextReader textReader,
            int count,
            ref IList<string> argv,
            ref Result error
            )
        {
            if (textReader == null)
            {
                error = "invalid stream";
                return ReturnCode.Error;
            }

            StringList newArgv = new StringList();

            while (true)
            {
                string line = textReader.ReadLine();

                if (line == null) // NOTE: End-of-file?
                    break;

                string trimLine = line.Trim();

                if (!String.IsNullOrEmpty(trimLine))
                {
                    if ((trimLine[0] != Characters.Comment) &&
                        (trimLine[0] != Characters.AltComment))
                    {
                        StringList list = null;

                        if (Parser.SplitList(interpreter, trimLine, 0,
                                Length.Invalid, false, true, ref list,
                                ref error) == ReturnCode.Ok)
                        {
                            newArgv.Add(list);
                        }
                        else
                        {
                            //
                            // NOTE: The line read from the file cannot be
                            //       parsed as a list, fail now.
                            //
                            return ReturnCode.Error;
                        }
                    }
                }
            }

            //
            // NOTE: *WARNING* This assumes that the arguments that need to be
            //       removed are at the start of the list provided by the
            //       caller; the ShellMainCore method can guarantee that will
            //       be the case and other callers should do so as well.
            //
            while ((argv != null) && (count-- > 0))
                GenericOps<string>.PopFirstArgument(ref argv);

            //
            // NOTE: If we used up all the arguments (i.e. there were only
            //       "count" arguments in the list), the original argument list
            //       (i.e. "argv") will now be null.  If that is the case, use
            //       a new list.
            //
            if (argv == null)
                argv = new StringList();

            //
            // NOTE: The count may already be zero at this point (if the above
            //       loop was actually fully executed); however, we must be 100%
            //       sure that it is zero beyond this point (for the loop below).
            //
            count = 0;

            //
            // NOTE: Insert each argument read from the file, in order, where the
            //       original argument(s) was/were removed.
            //
            foreach (string arg in newArgv)
                argv.Insert(count++, arg);

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: These methods are private because it currently seems unlikely
        //       that they will be useful to any external callers (i.e. those
        //       other than ShellMainCore).
        //
        public static void ShellMainCoreError(
            Interpreter interpreter,
            string savedArg,
            string arg,
            ReturnCode code,
            Result result,
            ref IInteractiveHost interactiveHost
            )
        {
            //
            // NOTE: This method overload is for non-API generated errors only.
            //       No error line info, script stack trace, or return code is
            //       needed.
            //
            IList<string> argv = null;

            ShellMainCoreError(interpreter, savedArg, arg, code, result,
                ref argv, ref interactiveHost);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void ShellMainCoreError(
            Interpreter interpreter,
            string savedArg,
            string arg,
            Result result,
            ref IList<string> argv,
            ref IInteractiveHost interactiveHost
            )
        {
            //
            // NOTE: This method overload is for non-API generated errors only.
            //       No error line info, script stack trace, or return code is
            //       needed.
            //
            ShellMainCoreError(interpreter, savedArg, arg, ReturnCode.Error,
                result, ref argv, ref interactiveHost);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void ShellMainCoreError(
            Interpreter interpreter,
            string savedArg,
            string arg,
            ReturnCode code,
            Result result,
            ref IList<string> argv,
            ref IInteractiveHost interactiveHost
            )
        {
            //
            // NOTE: This method overload is for [non-script] API generated
            //       errors only.  No error line info or script stack trace is
            //       required.
            //
            ShellMainCoreError(interpreter, savedArg, arg, code, result, 0,
                false, false, ref argv, ref interactiveHost);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void ShellMainCoreError(
            Interpreter interpreter,
            string savedArg,
            string arg,
            ReturnCode code,
            Result result,
            int errorLine,
            bool errorInfo,
            bool strict,
            ref IList<string> argv,
            ref IInteractiveHost interactiveHost
            )
        {
            DebugTrace(String.Format(
                "{0}: ShellMainCoreError: interpreter = {1}, " +
                "savedArg = {2}, arg = {3}, code = {4}, result = {5}, " +
                "errorLine = {6}, errorInfo = {7}, strict = {8}, " +
                "argv = {9}{10}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(interpreter),
                FormatOps.WrapOrNull(true, true, savedArg),
                FormatOps.WrapOrNull(true, true, arg), code,
                FormatOps.WrapOrNull(true, true, result), errorLine,
                errorInfo, strict, FormatOps.WrapOrNull(true, true, argv),
                Environment.NewLine), typeof(RuntimeOps).Name);

            if (interpreter == null)
            {
                if (strict)
                {
                    //
                    // NOTE: Interpreter is invalid; therefore, no interpreter
                    //       host is available.
                    //
                    WriteConsoleOrComplain(code, result, errorLine);
                }

                //
                // NOTE: Nothing else to do, return now.
                //
                return;
            }

            //
            // NOTE: Always grab the interpreter host fresh as it can change
            //       after any user code has been evaluated or executed.
            //
            interactiveHost = interpreter.Host;

            //
            // NOTE: See if quiet mode has been enabled for the interpreter.
            //       If so, skip doing any output because MSBuild may be
            //       watching us (i.e. it can cause the build to fail).
            //
            if (interpreter.ShouldBeQuiet())
                return;

            //
            // NOTE: Is the interpreter host unavailable now?
            //
            if (interactiveHost == null)
            {
                if (strict)
                {
                    //
                    // NOTE: No interpreter host is available.
                    //
                    WriteConsoleOrComplain(code, result, errorLine);
                }

                //
                // NOTE: Nothing else to do, return now.
                //
                return;
            }

            //
            // NOTE: Write the result to the interpreter host.  If the error
            //       line is zero, it will not actually be output.
            //
            interactiveHost.WriteResultLine(code, result, errorLine);

            //
            // NOTE: Do we want to report the script stack trace as well?
            //       First, see if debug mode has been enabled for the
            //       interpreter.
            //
            if (errorInfo && interpreter.Debug)
            {
                Result error = null;

                if (interpreter.CopyErrorInformation(
                        VariableFlags.None, ref error) == ReturnCode.Ok)
                {
                    interactiveHost.WriteResultLine(code, error.ErrorInfo);
                }
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Interactive Loop Support Methods
#if SHELL
        public static bool LooksLikeInteractiveCommand(
            string text
            )
        {
            string prefix = InteractiveCommandPrefix;

            return (!String.IsNullOrEmpty(prefix) &&
                !String.IsNullOrEmpty(text) &&
                text.Trim().StartsWith(prefix,
                    StringOps.SystemNoCaseStringComparisonType));
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode CheckForUpdate(
            Interpreter interpreter,
            int action,
            ReleaseType type,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            ref int errorLine,
            ref Result result
            )
        {
            ReturnCode code;

            //
            // NOTE: Evaluate the script we use to check for updates to the
            //       script engine.  If the proc has been redefined, this may
            //       not actually do anything.
            //
            errorLine = 0;

            code = Engine.EvaluateScript(
                interpreter, CheckForUpdateScript, engineFlags,
                substitutionFlags, eventFlags, expressionFlags,
                ref result, ref errorLine);

            //
            // NOTE: Evaluate the script we use to fetch an update to the
            //       script engine, if necessary.  If the proc has been
            //       redefined, this may not actually do anything.
            //
            if (code == ReturnCode.Ok)
            {
                //
                // NOTE: Attempt to parse the result of the check-for-update
                //       script as a list.
                //
                StringList list = null;

                code = Parser.SplitList(
                    interpreter, result, 0, Length.Invalid,
                    false, true, ref list, ref result);

                if (code == ReturnCode.Ok)
                {
                    //
                    // NOTE: If the specified action is greater than zero, we
                    //       need to actively interpret the result and then
                    //       fetch the update if necessary; otherwise, we do
                    //       nothing but simply reporting the result of the
                    //       check-for-update script.
                    //
                    if (action > 0)
                    {
                        //
                        // NOTE: If the result from the check-for-update script
                        //       is a list with more than one element, then a
                        //       download must be necessary.
                        //
                        if (list.Count > 1)
                        {
                            //
                            // NOTE: Parse the second element of the list as a
                            //       nested list containing [most of] the
                            //       arguments to pass to the fetch-an-update
                            //       script.
                            //
                            StringList list2 = null;

                            code = Parser.SplitList(
                                interpreter, list[1], 0, Length.Invalid,
                                false, true, ref list2, ref result);

                            if ((code == ReturnCode.Ok) && (list2.Count == 2))
                            {
                                string actionScript;

                                if (action > 1)
                                    actionScript = RunUpdateAndExitScript;
                                else
                                    actionScript = StringList.MakeList(
                                        FetchUpdateScript, list2[0], list2[1],
                                        type, PathOps.GetUnixPath(
                                            Path.GetTempPath()));

                                errorLine = 0;

                                code = Engine.EvaluateScript(
                                    interpreter, actionScript, engineFlags,
                                    substitutionFlags, eventFlags,
                                    expressionFlags, ref result,
                                    ref errorLine);

                                //
                                // NOTE: Return just the informational message
                                //       itself as the result.
                                //
                                result = String.Format(
                                    "{0}{1}{2}", list[0], Environment.NewLine,
                                    result);
                            }
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Return just the informational message itself
                        //       as the result or an error if none was provided.
                        //
                        result = (list.Count > 0) ? list[0] :
                            "invalid result from the check-for-update script";
                    }
                }
            }

            return code;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Command Support Methods
        public static void IncrementCommandCount(
            Interpreter interpreter
            )
        {
            if (interpreter != null)
                interpreter.IncrementCommandCount();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Command History Support Methods
#if HISTORY
        public static bool MatchHistoryData(
            IClientData clientData,
            IHistoryFilter historyFilter
            )
        {
            HistoryClientData historyClientData = clientData as HistoryClientData;

            if (historyClientData != null)
            {
                if (historyFilter != null)
                {
                    if ((historyFilter.StartLevel != Level.Invalid) &&
                        (historyClientData.Levels < historyFilter.StartLevel))
                    {
                        return false;
                    }

                    if ((historyFilter.StopLevel != Level.Invalid) &&
                        (historyClientData.Levels > historyFilter.StopLevel))
                    {
                        return false;
                    }

                    if ((historyFilter.HasFlags != HistoryFlags.None) &&
                        !FlagOps.HasFlags(historyClientData.Flags,
                            historyFilter.HasFlags, historyFilter.HasAll))
                    {
                        return false;
                    }

                    if ((historyFilter.NotHasFlags != HistoryFlags.None) &&
                        FlagOps.HasFlags(historyClientData.Flags,
                            historyFilter.NotHasFlags, historyFilter.NotHasAll))
                    {
                        return false;
                    }
                }

                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool MatchHistoryData(
            IHistoryData historyData,
            IHistoryFilter historyFilter
            )
        {
            int levels;
            HistoryFlags flags;

            if (historyData != null)
            {
                levels = historyData.Levels;
                flags = historyData.Flags;
            }
            else
            {
                levels = Level.Invalid;
                flags = HistoryFlags.None;
            }

            return MatchHistoryData(levels, flags, historyFilter);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool MatchHistoryData(
            int levels,
            HistoryFlags flags,
            IHistoryFilter historyFilter
            )
        {
            if (historyFilter != null)
            {
                if ((historyFilter.StartLevel != Level.Invalid) &&
                    (levels < historyFilter.StartLevel))
                {
                    return false;
                }

                if ((historyFilter.StopLevel != Level.Invalid) &&
                    (levels > historyFilter.StopLevel))
                {
                    return false;
                }

                if ((historyFilter.HasFlags != HistoryFlags.None) &&
                    !FlagOps.HasFlags(flags, historyFilter.HasFlags,
                        historyFilter.HasAll))
                {
                    return false;
                }

                if ((historyFilter.NotHasFlags != HistoryFlags.None) &&
                    FlagOps.HasFlags(flags, historyFilter.NotHasFlags,
                        historyFilter.NotHasAll))
                {
                    return false;
                }
            }

            return true;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Policy Support Methods
        public static IPolicy NewCorePolicy(
            ExecuteCallback callback,
            IClientData clientData,
            IPlugin plugin,
            ref Result error
            )
        {
            if (callback != null)
            {
                MethodInfo methodInfo = callback.Method;

                if (methodInfo != null)
                {
                    Type type = methodInfo.DeclaringType;

                    if (type != null)
                    {
                        _Policies.Core policy = new _Policies.Core(new PolicyData(
                            FormatOps.PolicyDelegateName(callback), null, null,
                            clientData, type.FullName, methodInfo.Name,
                            DelegateBindingFlags,
                            AttributeOps.GetMethodFlags(methodInfo),
                            PolicyFlags.None, plugin, 0));

                        policy.Callback = callback;
                        return policy;
                    }
                    else
                    {
                        error = "invalid policy callback method type";
                    }
                }
                else
                {
                    error = "invalid policy callback method";
                }
            }
            else
            {
                error = "invalid policy callback";
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsTrustedObject(
            Interpreter interpreter,
            string text,
            ObjectFlags flags,
            object @object,    /* NOT USED */
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (FlagOps.HasFlags(flags, ObjectFlags.Safe, true))
                    return true;

                error = String.Format(
                    "permission denied: safe interpreter cannot " +
                    "use object from \"{0}\"", text);

                return false;
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsTrustedType(
            Interpreter interpreter,
            string text,
            Type type,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (type != null)
                {
                    string name = type.FullName;

                    if (name != null)
                    {
                        ObjectDictionary trustedTypes =
                            new ObjectDictionary();

                        AddTrustedTypes(interpreter, trustedTypes);

                        if ((trustedTypes != null) &&
                            trustedTypes.ContainsKey(name))
                        {
                            return true;
                        }

                        error = String.Format(
                            "permission denied: safe interpreter cannot " +
                            "use type from \"{0}\"", text);
                    }
                    else
                    {
                        error = "invalid type name";
                    }
                }
                else
                {
                    error = "invalid type";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void AddTrustedDirectories(
            Interpreter interpreter,
            PathDictionary<object> directories
            )
        {
            if ((interpreter != null) && (directories != null))
            {
                lock (interpreter.SyncRoot)
                {
                    //
                    // BUGFIX: Cannot add null directory to the dictionary;
                    //         therefore, if the initialized path is invalid,
                    //         just skip it.
                    //
                    string initializedPath = interpreter.initializedPath;

                    if ((initializedPath != null) &&
                        !directories.ContainsKey(initializedPath))
                    {
                        directories.Add(initializedPath);
                    }

                    //
                    // NOTE: Add the paths trusted by the interpreter.
                    //
                    StringList trustedPaths = interpreter.TrustedPaths;

                    if (trustedPaths != null)
                    {
                        foreach (string trustedPath in trustedPaths)
                        {
                            if ((trustedPath != null) &&
                                !directories.ContainsKey(trustedPath))
                            {
                                directories.Add(trustedPath);
                            }
                        }
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void AddTrustedTypes(
            Interpreter interpreter,
            ObjectDictionary types
            )
        {
            if ((interpreter != null) && (types != null))
            {
                lock (interpreter.SyncRoot)
                {
                    //
                    // NOTE: Add the types trusted by the interpreter.
                    //
                    ObjectDictionary trustedTypes = interpreter.TrustedTypes;

                    if (trustedTypes != null)
                    {
                        foreach (KeyValuePair<string, object> pair in trustedTypes)
                        {
                            string trustedType = pair.Key;

                            if ((trustedType != null) &&
                                !types.ContainsKey(trustedType))
                            {
                                types.Add(trustedType, pair.Value);
                            }
                        }
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void AddTrustedUris(
            Interpreter interpreter, /* NOT USED */
            UriDictionary<object> uris
            )
        {
            if (uris != null)
            {
                Uri uri = GlobalState.GetAssemblyUri();

                //
                // NOTE: Add the URI for this assembly; however,
                //       make sure it is HTTPS.
                //
                if ((uri != null) &&
                    PathOps.IsHttpsUriScheme(uri) &&
                    !uris.ContainsKey(uri))
                {
                    uris.Add(uri, null);
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExtractPolicyContextAndScript( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IClientData clientData,
            ref IPolicyContext policyContext,
            ref IScript script,
            ref Result error
            )
        {
            if (clientData != null)
            {
                if (policyContext == null)
                    policyContext = clientData.Data as IPolicyContext;

                if (policyContext != null)
                {
                    script = policyContext.Script;

                    if (script != null)
                        return ReturnCode.Ok;
                    else
                        error = "invalid script";
                }
                else
                {
                    error = "policy clientData is not a policyContext object";
                }
            }
            else
            {
                error = "invalid policy clientData";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExtractPolicyContextAndFileName( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IClientData clientData,
            ref IPolicyContext policyContext,
            ref string fileName,
            ref Result error
            )
        {
            if (clientData != null)
            {
                if (policyContext == null)
                    policyContext = clientData.Data as IPolicyContext;

                if (policyContext != null)
                {
                    fileName = policyContext.FileName;

                    if (!String.IsNullOrEmpty(fileName))
                        return ReturnCode.Ok;
                    else
                        error = "invalid file name";
                }
                else
                {
                    error = "policy clientData is not a policyContext object";
                }
            }
            else
            {
                error = "invalid policy clientData";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExtractPolicyContextAndText( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IClientData clientData,
            ref IPolicyContext policyContext,
            ref string text,
            ref Result error
            )
        {
            if (clientData != null)
            {
                if (policyContext == null)
                    policyContext = clientData.Data as IPolicyContext;

                if (policyContext != null)
                {
                    text = policyContext.Text;
                    return ReturnCode.Ok;
                }
                else
                {
                    error = "policy clientData is not a policyContext object";
                }
            }
            else
            {
                error = "invalid policy clientData";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode LookupPluginCommandType( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IPlugin plugin,
            Type commandType,
            ref ICommand command
            )
        {
            Result error = null;

            return LookupPluginCommandType(
                interpreter, plugin, commandType, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode LookupPluginCommandType( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IPlugin plugin,
            Type commandType,
            ref ICommand command,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (plugin != null)
                {
                    if (commandType != null)
                    {
                        List<ICommand> commands = null;

                        if (interpreter.GetPluginCommands(plugin,
                                ref commands, ref error) == ReturnCode.Ok)
                        {
                            foreach (ICommand localCommand in commands)
                            {
                                IWrapper wrapper = localCommand as IWrapper;

                                if (wrapper != null)
                                {
                                    object @object = wrapper.Object;

                                    if (@object != null)
                                    {
                                        if (Object.ReferenceEquals(
                                                @object.GetType(), commandType))
                                        {
                                            command = localCommand;
                                            return ReturnCode.Ok;
                                        }
                                    }
                                }
                                else
                                {
                                    error = "invalid command wrapper";
                                    return ReturnCode.Error;
                                }
                            }
                        }
                    }
                    else
                    {
                        error = "invalid command type";
                    }
                }
                else
                {
                    error = "invalid plugin";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExtractPolicyContextAndCommand( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IClientData clientData,
            Type commandType,
            long commandToken,
            ref IPolicyContext policyContext,
            ref bool match,
            ref Result error
            )
        {
            ICommand command = null;

            return ExtractPolicyContextAndCommand(
                interpreter, clientData, commandType, commandToken,
                ref policyContext, ref command, ref match, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExtractPolicyContextAndCommand( /* POLICY HELPER METHOD */
            Interpreter interpreter,
            IClientData clientData,
            Type commandType,
            long commandToken,
            ref IPolicyContext policyContext,
            ref ICommand command,
            ref bool match,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (clientData != null)
                {
                    if (policyContext == null)
                        policyContext = clientData.Data as IPolicyContext;

                    if (policyContext != null)
                    {
                        IExecute execute = policyContext.Execute;

                        if (execute != null)
                        {
                            if (commandType != null)
                            {
#if ISOLATED_PLUGINS
                                IPlugin plugin = policyContext.Plugin;

                                if (IsPluginIsolated(plugin))
                                {
                                    command = null;

                                    if (((commandToken == 0) &&
                                        (LookupPluginCommandType(interpreter, plugin,
                                            commandType, ref command) == ReturnCode.Ok)) ||
                                        ((commandToken != 0) &&
                                        (interpreter.GetCommand(commandToken, true, true,
                                            ref command) == ReturnCode.Ok)))
                                    {
                                        match = Object.ReferenceEquals(execute, command);
                                    }
                                    else
                                    {
                                        match = false;
                                    }
                                }
                                else
#endif
                                {
                                    //
                                    // BUGBUG: This method call is a serious problem for
                                    //         isolated plugins.  The command type cannot be
                                    //         sent cleanly across the AppDomain boundry.
                                    //
                                    command = null;

                                    if (((commandToken == 0) &&
                                        (interpreter.GetCommand(commandType, true, true,
                                            ref command) == ReturnCode.Ok)) ||
                                        ((commandToken != 0) &&
                                        (interpreter.GetCommand(commandToken, true, true,
                                            ref command) == ReturnCode.Ok)))
                                    {
                                        match = Object.ReferenceEquals(execute, command);
                                    }
                                    else
                                    {
                                        match = false;
                                    }
                                }
                            }
                            else
                            {
                                //
                                // NOTE: Command type is null, skip type matching against
                                //       it (i.e. just extract it and return).
                                //
                                command = execute as ICommand;

                                match = (command != null);
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "policyContext does not contain an executable object";
                        }
                    }
                    else
                    {
                        error = "policy clientData is not a policyContext object";
                    }
                }
                else
                {
                    error = "invalid policy clientData";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Alias Support Methods
        public static ReturnCode GetAliasCommand(
            IAlias alias,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            //
            // NOTE: Gets the command associated with an alias.  This is the
            //       special command implementation (i.e. "_Commands.Alias")
            //       that also implements the IAlias interface so that it can
            //       maintain the linkage between the alias and the target.
            //
            if (alias != null)
            {
                //
                // NOTE: See if this alias also directly implements the
                //       ICommand interface.
                //
                command = alias as ICommand;

                if (command == null)
                {
                    //
                    // NOTE: See if this alias is merely a wrapper around
                    //       the real one.  In that case, unwrap it and
                    //       return the contained alias as an ICommand.
                    //
                    _Wrappers.Alias aliasWrapper = alias as _Wrappers.Alias;

                    if (aliasWrapper != null)
                        command = aliasWrapper.alias as ICommand;
                }

                if (command != null)
                    return ReturnCode.Ok;
                else
                    error = "alias is not a command";
            }
            else
            {
                error = "invalid alias";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetObjectAliasArguments(
            bool all,
            ref ArgumentList arguments,
            ref Result error
            )
        {
            arguments = new ArgumentList(new string[] {
                typeof(_Commands.Object).Name.ToLowerInvariant(),
                all ? "invokeall" : "invoke"
            });

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

#if NATIVE && TCL
        public static ReturnCode GetTclAliasArguments(
            ref ArgumentList arguments,
            ref Result error
            )
        {
            arguments = new ArgumentList(new string[] {
                typeof(_Commands.Tcl).Name.ToLowerInvariant(), "eval"
            });

            return ReturnCode.Ok;
        }
#endif

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: This method is static; however, in the future that may no
        //       longer be true.  In that case, it will need to move back to
        //       the Interpreter class.
        //
        public static IAlias NewAlias(
            string name,
            CommandFlags flags,
            IClientData clientData,
            string nameToken,
            Interpreter targetInterpreter,
            IExecute target,
            ArgumentList arguments,
            OptionDictionary options,
            int startIndex
            )
        {
            //
            // TODO: Modify this when full [namespace] support has been added.
            //
            string newName = StringOps.TrimLeadingColons(name);

            //
            // HACK: We do not necessarily know (and do not simply want to
            //       "guess") the plugin associated with the target of the
            //       command; therefore, we use a null value for the plugin
            //       argument here.
            //
            return new _Commands.Alias(
                new CommandData(newName, null, null, clientData,
                    typeof(_Commands.Alias).FullName, flags,
                    /* plugin */ null, 0),
                new AliasData(nameToken, targetInterpreter,
                    target, arguments, options, startIndex, 0));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Notification Support Methods
        public static IScriptEventArgs GetEventArgs(
            NotifyType type,
            NotifyFlags flags,
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            Result result,
            ScriptException exception
            )
        {
            return new ScriptEventArgs(
                GlobalState.NextId(interpreter), type, flags, interpreter,
                clientData, arguments, result, exception, null, null);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Event Manager Support Methods
        public static void Sleep(
            Interpreter interpreter,
            bool minimum
            ) /* THREAD-SAFE */
        {
            if (interpreter != null)
            {
                IEventManager eventManager = interpreter.EventManager;

                if (eventManager != null)
                    eventManager.Sleep(minimum);
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ProcessEvents(
            Interpreter interpreter,
            EventFlags eventFlags,
            EventPriority priority,
            int limit,
            bool stopOnError,
            bool errorOnEmpty,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                IEventManager eventManager = interpreter.EventManager;

                if (eventManager != null)
                {
                    return eventManager.ProcessEvents(
                        eventFlags, priority, limit, stopOnError, errorOnEmpty,
                        ref result);
                }
                else
                {
                    result = "event manager not available";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode DoOneEvent(
            Interpreter interpreter,
            EventFlags eventFlags,
            EventPriority priority,
            int limit, /* NOTE: Pass zero for ALL. */
            bool stopOnError,
            bool errorOnEmpty,
            bool userInterface,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                //
                // BUGFIX: Wrap in try/catch in case the interpreter is
                //         disposed (from WaitVariable).
                //
                try
                {
                    IEventManager eventManager = interpreter.EventManager;

                    if (eventManager != null)
                    {
                        return eventManager.DoOneEvent(
                            eventFlags, priority, limit, stopOnError,
                            errorOnEmpty, userInterface, ref result);
                    }
                    else
                    {
                        result = "event manager not available";
                    }
                }
                catch (Exception e)
                {
                    result = e;
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode Wait(
            Interpreter interpreter,
            long microseconds,
            bool strict,
            ref Result error
            ) /* THREAD-SAFE */
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                int waitCount;

                if ((waitCount = interpreter.EnterWait()) > 0)
                {
                    if (microseconds == 0)
                    {
#if NATIVE && WINDOWS && WINFORMS
                        //
                        // NOTE: If necessary, process all Windows messages
                        //       from the queue.
                        //
                        if (!strict)
                            code = WindowOps.ProcessEvents(interpreter, ref error);

                        if (code == ReturnCode.Ok)
#endif
                        {
                            //
                            // NOTE: Yield to other running threads.  This also
                            //       gives them an opportunity to cancel the
                            //       script in progress on this thread.
                            //
                            Sleep(0);
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Keep track of how many iterations through the
                        //       loop we take.
                        //
                        int iterations = 0;

                        //
                        // HACK: We need to account for our processing overhead;
                        //       therefore, use half of the requested delay.
                        //
                        int milliseconds = ConversionOps.ToInt(
                            PerformanceOps.GetMilliseconds(microseconds) /
                            WaitDivisor);

                        if (milliseconds < 0)
                            milliseconds = 0;

                        if (milliseconds > WaitMaximumSleepTime)
                            milliseconds = WaitMaximumSleepTime;

                        //
                        // NOTE: For more precise timing, use the high-resolution
                        //       CPU performance counter.
                        //
                        long startCount = PerformanceOps.GetCount();

                        //
                        // BUGFIX: Make sure the slop time does not exceed the
                        //         actual wait.
                        //
                        long slopMicroseconds = Math.Min(
                            microseconds / WaitSlopDivisor, WaitSlopMinimumTime);

                        //
                        // NOTE: Delay for approximately the specified number of
                        //       microseconds.
                        //
                        while (((code = Interpreter.Ready(interpreter,
                                false, false, false, ref error)) == ReturnCode.Ok) &&
                            !PerformanceOps.HasElapsed(startCount,
                                microseconds, slopMicroseconds))
                        {
#if NATIVE && WINDOWS && WINFORMS
                            if (!strict)
                            {
                                code = WindowOps.ProcessEvents(interpreter, ref error);

                                if (code != ReturnCode.Ok)
                                    break;
                            }
#endif

                            Sleep(interpreter, milliseconds); iterations++;
                        }

                        long stopCount = PerformanceOps.GetCount();

                        double elapsedMicroseconds = PerformanceOps.GetMicroseconds(
                            startCount, stopCount, 1);

                        DebugTrace(String.Format(
                            "{0}: Wait: code = {1}, iterations = {2}, microseconds = {3}, " +
                            "elapsedMicroseconds = {4}, sleepMilliseconds = {5}, " +
                            "slopMicroseconds = {6}, differenceMicroseconds = {7}, " +
                            "waitCount = {8}, error = {9}{10}", GlobalState.GetCurrentThreadId(),
                            code, iterations, microseconds, elapsedMicroseconds, milliseconds,
                            slopMicroseconds, elapsedMicroseconds - (double)microseconds,
                            waitCount, FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                            typeof(RuntimeOps).Name);
                    }

                    /* IGNORED */
                    interpreter.ExitWait();
                }
                else
                {
                    error = "wait subsystem locked";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                error = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Channel Support Methods
        public static string ChannelTypeToName(
            string name,
            ChannelType channelType
            )
        {
            if (FlagOps.HasFlags(channelType, ChannelType.Input, true))
                return (name != null) ? name : StandardChannel.Input;
            else if (FlagOps.HasFlags(channelType, ChannelType.Output, true))
                return (name != null) ? name : StandardChannel.Output;
            else if (FlagOps.HasFlags(channelType, ChannelType.Error, true))
                return (name != null) ? name : StandardChannel.Error;
            else
                return name;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsStandardChannelName( /* NOT USED */
            string name
            )
        {
            if (String.Compare(name, StandardChannel.Input,
                    StringOps.SystemStringComparisonType) == 0)
            {
                return true;
            }
            else if (String.Compare(name, StandardChannel.Output,
                    StringOps.SystemStringComparisonType) == 0)
            {
                return true;
            }
            else if (String.Compare(name, StandardChannel.Error,
                    StringOps.SystemStringComparisonType) == 0)
            {
                return true;
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Resource Support Methods
        public static string GetString(
            ResourceManager resourceManager,
            string name,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            if (resourceManager != null)
            {
                try
                {
                    return resourceManager.GetString(name, cultureInfo);
                }
                catch (Exception e)
                {
                    error = e;
                }
            }
            else
            {
                error = InvalidPluginResourceManager;
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetString(
            IPlugin plugin,
            ResourceManager resourceManager,
            string name,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            if ((plugin != null) &&
                !FlagOps.HasFlags(plugin.Flags, PluginFlags.NoResources, true))
            {
                ResourceManager pluginResourceManager = plugin.ResourceManager;

                if (pluginResourceManager != null)
                {
                    try
                    {
                        return pluginResourceManager.GetString(name, cultureInfo);
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                else
                {
                    error = InvalidPluginResourceManager;
                }
            }
            else
            {
                if (resourceManager != null)
                {
                    try
                    {
                        return resourceManager.GetString(name, cultureInfo);
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                else
                {
                    error = InvalidInterpreterResourceManager;
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ResourceManager NewResourceManager(
            AssemblyName assemblyName
            )
        {
            if (assemblyName != null)
            {
                try
                {
                    return NewResourceManager(assemblyName,
                        Assembly.Load(assemblyName));
                }
                catch (Exception e)
                {
                    DebugOps.Complain(ReturnCode.Error, e);
                }
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ResourceManager NewResourceManager(
            Assembly assembly
            )
        {
            if (assembly != null)
            {
                try
                {
                    return NewResourceManager(assembly.GetName(), assembly);
                }
                catch (Exception e)
                {
                    DebugOps.Complain(ReturnCode.Error, e);
                }

            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ResourceManager NewResourceManager(
            AssemblyName assemblyName,
            Assembly assembly
            )
        {
            if ((assemblyName != null) && (assembly != null))
            {
                try
                {
                    return new ResourceManager(assemblyName.Name, assembly);
                }
                catch (Exception e)
                {
                    DebugOps.Complain(ReturnCode.Error, e);
                }
            }

            return null;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Stream Support Methods
        public static ReturnCode NewFileStream(
            string path,
            FileMode mode,
            FileAccess access,
            bool resolve,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            )
        {
            return NewFileStream(
                path, mode, access, FileShare.Read, Channel.DefaultBufferSize,
                FileOptions.None, resolve, ref fullPath, ref stream, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode NewFileStream(
            string path,
            FileMode mode,
            FileAccess access,
            FileShare share,
            int bufferSize,
            FileOptions options,
            bool resolve,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            )
        {
            try
            {
                if (!PathOps.IsRemoteUri(path))
                {
                    string localFullPath = resolve ?
                        PathOps.ResolveFullPath(path) : path;

                    if (!String.IsNullOrEmpty(localFullPath))
                    {
                        fullPath = localFullPath;

                        stream = new FileStream(
                            localFullPath, mode, access, share,
                            bufferSize, options);

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = String.Format(
                            "could not resolve local path \"{0}\"",
                            path);
                    }
                }
                else
                {
                    error = String.Format(
                        "cannot open file stream for remote uri \"{0}\"",
                        path);
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Ensemble Support Methods
        public static ReturnCode SubCommandFromEnsemble(
            IEnsemble ensemble,         /* in */
            bool strict,                /* in */
            bool noCase,                /* in */
            ref string name,            /* in, out */
            ref Result error            /* out */
            )
        {
            ISubCommand subCommand = null;

            return SubCommandFromEnsemble(
                ensemble, strict, noCase, ref name,
                ref subCommand, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode SubCommandFromEnsemble(
            IEnsemble ensemble,         /* in */
            bool strict,                /* in */
            bool noCase,                /* in */
            ref string name,            /* in, out */
            ref ISubCommand subCommand, /* out */
            ref Result error            /* out */
            )
        {
            //
            // NOTE: *WARNING* Empty sub-command names are allowed,
            //       please do not change this to "!String.IsNullOrEmpty".
            //
            if (name != null)
            {
                if ((ensemble != null) &&
                    (ensemble.SubCommands != null) &&
                    (ensemble.SubCommands.Count > 0))
                {
                    bool exact = false;
                    List<KeyValuePair<string, ISubCommand>> subCommands =
                        new List<KeyValuePair<string, ISubCommand>>();

                    foreach (KeyValuePair<string, ISubCommand> pair in ensemble.SubCommands)
                    {
                        if (String.Compare(pair.Key, 0, name, 0, name.Length,
                                noCase ? StringOps.SystemNoCaseStringComparisonType :
                                StringOps.SystemStringComparisonType) == 0)
                        {
                            //
                            // NOTE: Was the key valid (this should always succeed).
                            //
                            if (pair.Key != null)
                            {
                                //
                                // NOTE: Did we match the whole string, regardless of case?
                                //
                                bool whole = (pair.Key.Length == name.Length);

                                //
                                // NOTE: It was a match; however, was it an exact match?
                                //
                                if (whole)
                                    //
                                    // NOTE: Preserve match, it may differ in case.
                                    //
                                    exact = true;

                                //
                                // NOTE: Was it an exact match or did we match at least one
                                //       character in a partial match?
                                //
                                if (whole || (name.Length > 0))
                                {
                                    //
                                    // NOTE: Store the exact or partial match in the result
                                    //       list.
                                    //
                                    subCommands.Add(pair);
                                }
                            }
                        }
                    }

                    //
                    // NOTE: If there was an exact match, just use it.
                    //
                    if (exact)
                    {
                        //
                        // NOTE: No need to change the caller's name variable,
                        //       it already contains the correct value; however,
                        //       we do need to provide the actual ISubCommand.
                        //
                        subCommand = subCommands[0].Value;

                        return ReturnCode.Ok;
                    }
                    else if (subCommands.Count == 1)
                    {
                        //
                        // NOTE: Normal case, exactly one sub-command matched.
                        //
                        name = subCommands[0].Key;
                        subCommand = subCommands[0].Value;

                        return ReturnCode.Ok;
                    }
                    else if (subCommands.Count > 1)
                    {
                        error = BadSubCommand(
                            "ambiguous", "option", name,
                            /* ensemble, */
                            subCommands, null, null);
                    }
                    else if (strict)
                    {
                        error = BadSubCommand(
                            null, "option", name,
                            ensemble, null, null);
                    }
                    else
                    {
                        //
                        // NOTE: Non-strict mode, leave the original sub-command
                        //       unchanged and let the caller deal with it.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = "invalid ensemble";
                }
            }
            else
            {
                error = "invalid sub-command name";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode TryExecuteSubCommandFromEnsemble(
            IEnsemble ensemble,         /* in */
            Interpreter interpreter,    /* in */
            IClientData clientData,     /* in */
            ArgumentList arguments,     /* in */
            bool strict,                /* in */
            bool noCase,                /* in */
            ref string name,            /* in, out */
            ref bool tried,             /* out */
            ref Result result           /* out */
            )
        {
            ISubCommand subCommand = null;

            return TryExecuteSubCommandFromEnsemble(
                ensemble, interpreter, clientData, arguments, strict,
                noCase, ref name, ref subCommand, ref tried, ref result);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode TryExecuteSubCommandFromEnsemble(
            IEnsemble ensemble,                  /* in */
            Interpreter interpreter,             /* in */
            IClientData clientData,              /* in */
            ArgumentList arguments,              /* in */
            bool strict,                         /* in */
            bool noCase,                         /* in */
            ref string name,                     /* in, out */
            ref ISubCommand subCommand,          /* out */
            ref bool tried,                      /* out */
            ref Result result                    /* out */
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                code = SubCommandFromEnsemble(
                    ensemble, strict, noCase, ref name, ref subCommand,
                    ref result);

                if (code == ReturnCode.Ok)
                {
                    if (subCommand != null)
                    {
                        //
                        // NOTE: Indicate that the sub-command has been
                        //       dispatched (i.e. there is no need for
                        //       the caller to handle this sub-command).
                        //       Even if the execution fails, we still
                        //       tried to execute it and the caller
                        //       should not try to handle it.
                        //
                        tried = true;

                        code = interpreter.Execute(
                            name, subCommand, clientData, arguments,
                            ref result);
                    }
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Error Message Support Methods
        public static Exception GetBaseException(
            Exception exception
            )
        {
            if (exception == null)
                return null;

            Exception baseException = exception.GetBaseException();

            if (baseException != null)
                return baseException;

            return exception;
        }

        ///////////////////////////////////////////////////////////////////////

        public static Result BadValue(
            string adjective,
            string type,
            string value,
            StringSortedList values,
            string prefix,
            string suffix
            )
        {
            if (values != null)
            {
                return String.Format(
                    "{0} {1} \"{2}\": must be {3}{4}",
                    !String.IsNullOrEmpty(adjective) ? adjective : "bad",
                    !String.IsNullOrEmpty(type) ? type : "value", value,
                    GenericOps<string>.ListToEnglish(
                        values, prefix, ", ",
                        !String.IsNullOrEmpty(suffix) ? null : "or "), suffix);
            }

            return String.Format(
                "{0} {1} \"{2}\"",
                !String.IsNullOrEmpty(adjective) ? adjective : "bad",
                !String.IsNullOrEmpty(type) ? type : "value", value); // FIXME: Fallback here?
        }

        ///////////////////////////////////////////////////////////////////////

        public static Result BadValue(
            string adjective,
            string type,
            string value,
            IEnumerable<string> values,
            string prefix,
            string suffix
            )
        {
            return BadValue(adjective, type, value,
                (values != null) ? new StringSortedList(values) : null,
                prefix, suffix);
        }

        ///////////////////////////////////////////////////////////////////////

        public static Result BadSubCommand(
            string adjective,
            string type,
            string subCommand,
            IEnsemble ensemble,
            string prefix,
            string suffix
            )
        {
            if ((ensemble != null) &&
                (ensemble.SubCommands != null))
            {
                return BadValue(
                    !String.IsNullOrEmpty(adjective) ? adjective : "bad",
                    !String.IsNullOrEmpty(type) ? type : "option", subCommand,
                    ensemble.SubCommands.Keys, prefix, suffix);
            }

            return String.Format(
                "{0} {1} \"{2}\"",
                !String.IsNullOrEmpty(adjective) ? adjective : "bad",
                !String.IsNullOrEmpty(type) ? type : "option", subCommand); // FIXME: Fallback here?
        }

        ///////////////////////////////////////////////////////////////////////

        public static Result BadSubCommand(
            string adjective,
            string type,
            string subCommand,
            IEnumerable<KeyValuePair<string, ISubCommand>> subCommands,
            string prefix,
            string suffix
            )
        {
            return BadValue(
                adjective, type, subCommand,
                (subCommands != null) ? new StringSortedList(subCommands) : null,
                prefix, suffix);
        }

        ///////////////////////////////////////////////////////////////////////

        public static Result WrongNumberOfArguments( /* NOT USED */
            IIdentifier identifier,
            int count,
            ArgumentList arguments,
            string suffix
            )
        {
            if ((count > 0) &&
                (arguments != null) &&
                (arguments.Count > 0))
            {
                return String.Format(
                    "wrong # args: should be \"{0}{1}{2}\"",
                    ArgumentList.GetRange(arguments, 0, Math.Min(count - 1, arguments.Count - 1)),
                    !String.IsNullOrEmpty(suffix) ? Characters.Space.ToString() : null,
                    suffix);
            }

            return "wrong # args"; // FIXME: Fallback here?
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Version Information Methods
        public static ReturnCode GetVersion(
            ref Result result
            )
        {
            //
            // NOTE: Return the engine version information.
            //
            Assembly assembly = GlobalState.GetAssembly();

            result = StringList.MakeList(
                ArrayEquals(
                    License.Hash, StringOps.HashString(
                        null, null, License.Summary + License.Text)) ?
                "Genuine" : "NonGenuine",
                GlobalState.GetPackageName(false),
                GlobalState.GetAssemblyVersion(),
                AttributeOps.GetAssemblyTag(assembly),
                AttributeOps.GetAssemblyRelease(assembly),
                AttributeOps.GetAssemblyText(assembly),
                AttributeOps.GetAssemblyConfiguration(assembly),
                "Tcl " + TclVars.PatchLevelValue,
                FormatOps.Iso8601DateTime(
                    AttributeOps.GetAssemblyDateTime(assembly), true),
                AttributeOps.GetAssemblySourceId(assembly),
                AttributeOps.GetAssemblySourceTimeStamp(assembly),
                CommonOps.Runtime.GetRuntimeNameAndVersion(),
                PlatformOps.OperatingSystem,
                PlatformOps.Processor);

            return ReturnCode.Ok;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Version Checking Methods
        public static int VersionCompare(
            Version version1,
            Version version2
            )
        {
            if ((version1 != null) && (version2 != null))
                return version1.CompareTo(version2);
            else if ((version1 == null) && (version2 == null))
                return 0;        // x (null) is equal to y (null)
            else if (version1 == null)
                return -1;     // x (null) is less than y (non-null)
            else
                return 1;  // x (non-null) is greater than y (null)
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool VersionSatisfies(
            Version version1,
            Version version2,
            bool exact
            )
        {
            if (exact)
                return (VersionCompare(version1, version2) == 0);
            else
                return (VersionCompare(version1, version2) >= 0);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Return / Exit Code Handling Methods
        public static bool IsSuccess(
            ReturnCode code,
            bool exceptions
            )
        {
            if (exceptions)
                return ((code != ReturnCode.Error) &&
                        ((code & ReturnCode.CustomError) == 0));
            else
                return ((code == ReturnCode.Ok) ||
                        ((code & ReturnCode.CustomOk) != 0));
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode CustomOkCode(uint value)
        {
            //
            // NOTE: These are always considered as "success codes".
            //
            return (ReturnCode.CustomOk | (ReturnCode)value);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode CustomErrorCode(uint value)
        {
            //
            // NOTE: These are always considered as "failure codes".
            //
            return (ReturnCode.CustomError | (ReturnCode)value);
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // HACK: Avoid "ExitCode" enumeration / property name collision.
        //
        public static ExitCode SuccessExitCode()
        {
            return ExitCode.Success;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // HACK: Avoid "ExitCode" enumeration / property name collision.
        //
        public static ExitCode FailureExitCode()
        {
            return ExitCode.Failure;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // HACK: Avoid "ExitCode" enumeration / property name collision.
        //
        public static ExitCode ExceptionExitCode()
        {
            return ExitCode.Exception;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode ExitCodeToReturnCode(
            ExitCode exitCode
            )
        {
            return (exitCode == SuccessExitCode()) ?
                ReturnCode.Ok : ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ExitCode ReturnCodeToExitCode(
            ReturnCode code,
            bool exceptions
            )
        {
            return IsSuccess(code, exceptions) ?
                SuccessExitCode() : FailureExitCode();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Result Methods
        #region Synchronized Result Methods
        public static Result CreateSynchronizedResult(
            string name
            )
        {
            Result synchronizedResult = String.Empty; /* FORCE VALID */

            synchronizedResult.ClientData =
                new ClientData(ThreadOps.CreateEvent(name));

            return synchronizedResult;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void CleanupSynchronizedResult(
            Result synchronizedResult
            )
        {
            if (synchronizedResult != null)
            {
                lock (synchronizedResult)
                {
                    //
                    // NOTE: Grab the client data (the event wait handle).
                    //
                    IClientData clientData = synchronizedResult.ClientData;

                    if (clientData != null)
                    {
                        EventWaitHandle @event =
                            clientData.Data as EventWaitHandle;

                        if (@event != null)
                            ThreadOps.CloseEvent(ref @event);
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool WaitSynchronizedResult(
            Result synchronizedResult,
            int milliseconds
            )
        {
            EventWaitHandle @event = null;

            if (synchronizedResult != null)
            {
                lock (synchronizedResult)
                {
                    //
                    // NOTE: Grab the client data (the event wait handle).
                    //
                    IClientData clientData = synchronizedResult.ClientData;

                    if (clientData != null)
                        @event = clientData.Data as EventWaitHandle;
                }
            }

            if (@event != null)
                return ThreadOps.WaitEvent(@event, milliseconds);

            return false;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void SetSynchronizedResult(
            Result synchronizedResult,
            ReturnCode code,
            Result result
            )
        {
            //
            // NOTE: Does the caller want to be notified about the result?
            //
            if (synchronizedResult != null)
            {
                lock (synchronizedResult)
                {
                    //
                    // NOTE: Grab the original client data (the event wait
                    //       handle).
                    //
                    IClientData clientData = synchronizedResult.ClientData;

                    //
                    // NOTE: Set the new client data (the result pair).
                    //
                    synchronizedResult.ClientData = new ClientData(
                        new AnyPair<ReturnCode, Result>(code, result));

                    //
                    // NOTE: If the original client data is valid and can be
                    //       cast to an event wait handle, signal it now.
                    //
                    if (clientData != null)
                    {
                        EventWaitHandle @event =
                            clientData.Data as EventWaitHandle;

                        if (@event != null)
                            /* IGNORED */
                            ThreadOps.SetEvent(@event);
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetSynchronizedResult(
            Result synchronizedResult,
            ref ReturnCode code,
            ref Result result,
            ref Result error
            )
        {
            if (synchronizedResult != null)
            {
                lock (synchronizedResult)
                {
                    IClientData clientData = synchronizedResult.ClientData;

                    if (clientData != null)
                    {
                        IAnyPair<ReturnCode, Result> pair =
                            clientData.Data as IAnyPair<ReturnCode, Result>;

                        if (pair != null)
                        {
                            result = pair.Y;
                            code = pair.X;

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "synchronized result clientData is not a pair";
                        }
                    }
                    else
                    {
                        error = "invalid synchronized result clientData";
                    }
                }
            }
            else
            {
                error = "invalid synchronized result";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Result Formatting Methods
        public static string FormatResult(
            ReturnCode code,
            Result result
            )
        {
            return FormatResult(code, result, 0);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string FormatResult(
            ReturnCode code,
            Result result,
            int errorLine
            )
        {
            return FormatResult(code, result, errorLine, false);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string FormatResult(
            ReturnCode code,
            Result result,
            int errorLine,
            bool exceptions
            )
        {
            return FormatResult(null, code, result, errorLine, exceptions);
        }

        ///////////////////////////////////////////////////////////////////////

        public static string FormatResult(
            string prefix,
            ReturnCode code,
            Result result,
            int errorLine,
            bool exceptions
            )
        {
            bool haveResult = !String.IsNullOrEmpty(result);
            bool haveErrorLine = (errorLine != 0);
            string format;

            if (IsSuccess(code, exceptions))
            {
                if (haveResult)
                    format = resultOnlyFormat;
                else
                    format = emptyFormat;
            }
            else
            {
                if (haveResult)
                {
                    if (haveErrorLine)
                        format = codeResultAndErrorLineFormat;
                    else
                        format = codeAndResultFormat;
                }
                else
                {
                    if (haveErrorLine)
                        format = codeAndErrorLineFormat;
                    else
                        format = codeOnlyFormat;
                }
            }

            return String.Format(format, prefix, code, result, errorLine);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Script Support Methods
        public static ReturnCode GetProcedureScriptLocation(
            Interpreter interpreter,
            IProcedure procedure,
            ref IScriptLocation location,
            ref Result error
            )
        {
            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (procedure == null)
            {
                error = "invalid procedure";
                return ReturnCode.Error;
            }

            if (FlagOps.HasFlags(
                    procedure.Flags, ProcedureFlags.ScriptLocation, true))
            {
                return GetScriptLocation(
                    interpreter, false, ref location, ref error);
            }

            location = procedure.Location;
            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetScriptLocation(
            ArgumentList arguments,
            int startIndex,
            ref IScriptLocation location,
            ref Result error
            )
        {
            if (arguments == null)
            {
                error = "invalid argument list";
                return ReturnCode.Error;
            }

            if ((startIndex < 0) || (startIndex >= arguments.Count))
            {
                error = "argument index out of range";
                return ReturnCode.Error;
            }

            Argument firstArgument = arguments[startIndex];
            Argument lastArgument = arguments[arguments.Count - 1];

            if ((firstArgument == null) && (lastArgument == null))
            {
                location = new ScriptLocation((IScriptLocation)null);
                return ReturnCode.Ok;
            }

            if (firstArgument != null)
            {
                location = new ScriptLocation(firstArgument.FileName,
                    firstArgument.StartLine, (lastArgument != null) ?
                        lastArgument.EndLine : firstArgument.EndLine,
                    firstArgument.ViaSource);
            }
            else
            {
                location = new ScriptLocation(lastArgument.FileName,
                    lastArgument.StartLine, lastArgument.EndLine,
                    lastArgument.ViaSource);
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // NOTE: FOR [info script] USE ONLY.
        //
        public static ReturnCode GetScriptLocation(
            Interpreter interpreter,
            bool viaSource,
            bool scrub,
            ref string fileName,
            ref Result error
            )
        {
            int currentLine = Parser.UnknownLine;

            return GetScriptLocation(
                interpreter, viaSource, scrub, ref fileName,
                ref currentLine, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetScriptLocation(
            Interpreter interpreter,
            bool viaSource,
            bool scrub,
            ref string fileName,
            ref int currentLine,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                ReturnCode code;
                IScriptLocation location = null;

                code = GetScriptLocation(
                    interpreter, viaSource, ref location, ref error);

                if (code == ReturnCode.Ok)
                {
                    string scriptFileName = (location != null) ?
                        location.FileName : null;

                    if (scrub && (scriptFileName != null))
                        fileName = PathOps.ScrubPath(
                            GlobalState.GetBasePath(), scriptFileName);
                    else
                        fileName = scriptFileName; /* NOTE: May be null. */

                    currentLine = (location != null) ?
                        location.StartLine : Parser.UnknownLine;

                    return ReturnCode.Ok;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode GetScriptLocation(
            Interpreter interpreter,
            bool viaSource,
            ref IScriptLocation location,
            ref Result error
            )
        {
            if (interpreter != null)
            {
#if !THREADING
                lock (interpreter.SyncRoot) /* TRANSACTIONAL */
#endif
                {
                    //
                    // NOTE: Grab whatever the caller previously manually
                    //       set the current script file name to, if any.
                    //
                    location = interpreter.ScriptLocation;

                    if (location == null)
                    {
                        StackList<IScriptLocation> locations =
                            interpreter.ScriptLocations;

                        if (locations != null)
                        {
                            int count = locations.Count;

                            if (count > 0)
                            {
                                for (int index = count - 1; index >= 0; index--)
                                {
                                    IScriptLocation thisLocation = locations[index];

                                    if (!viaSource || thisLocation.ViaSource)
                                    {
                                        //
                                        // NOTE: Grab the last (most recent) script
                                        //       location from the stack of active
                                        //       script locations that matches the
                                        //       via [source] flag set by the caller.
                                        //
                                        location = thisLocation;
                                        break;
                                    }
                                }
                            }
                        }
                    }

                    return ReturnCode.Ok;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static IPackage NewCorePackage(
            string name,
            string group,
            string description,
            IClientData clientData,
            PackageFlags flags,
            Version loaded,
            VersionStringDictionary ifNeeded
            )
        {
            return new _Packages.Core(new PackageData(name, group, description,
                clientData, flags, loaded, ifNeeded, 0));
        }

        ///////////////////////////////////////////////////////////////////////

        public static IProcedure NewCoreProcedure(
            string name,
            string group,
            string description,
            ProcedureFlags flags,
            ArgumentList arguments,
            string body,
            IScriptLocation location,
            IClientData clientData
            )
        {
            //
            // TODO: Modify this when full [namespace] support has been added.
            //
            string newName = StringOps.TrimLeadingColons(name);

            return new _Procedures.Core(new ProcedureData(newName, group,
                description, flags, arguments, body, location, clientData, 0));
        }

        ///////////////////////////////////////////////////////////////////////

        #region Library Support Methods
        public static ReturnCode GetFileScript(
            Interpreter interpreter,
            string directory,
            string name,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            //
            // NOTE: Check for the script on disk in the
            //       specified directory.
            //
            string fileName = PathOps.CombinePath(
                false, directory, name);

            if (!String.IsNullOrEmpty(fileName) &&
                File.Exists(fileName))
            {
                flags |= ScriptFlags.File;
                result = fileName;

                return ReturnCode.Ok;
            }
            else if (!FlagOps.HasFlags(flags, ScriptFlags.NoAutoPath, true))
            {
                //
                // NOTE: Check for the script on disk in the
                //       directories listed in the "auto_path".
                //
                StringList autoPathList = GlobalState.GetAutoPathList(
                    interpreter, false);

                foreach (string path in autoPathList)
                {
                    fileName = PathOps.CombinePath(
                        false, path, name);

                    if (!String.IsNullOrEmpty(fileName) &&
                        File.Exists(fileName))
                    {
                        flags |= ScriptFlags.File;
                        result = fileName;

                        return ReturnCode.Ok;
                    }
                }

                result = String.Format(
                    "cannot find a suitable \"{0}\" script in \"{1}\"",
                    name, autoPathList);
            }
            else
            {
                result = String.Format(
                    "cannot find a suitable \"{0}\" script in \"{1}\"",
                    name, fileName);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetLibraryScript(
            Interpreter interpreter,
            IEngineHost engineHost,
            string name,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                Result localResult = null;
                ResultList errors = new ResultList();

                //
                // NOTE: Where is the primary root directory where library
                //       script should be found (e.g. "<root>\lib\Eagle1.0")?
                //
                string directory = GlobalState.GetPackageNameRootPath();

                if (FlagOps.HasFlags(flags, ScriptFlags.PreferFileSystem, true))
                {
                    //
                    // NOTE: First, look for the script on the native file
                    //       system.
                    //
                    if (GetFileScript(interpreter, directory, name, ref flags,
                            ref localResult) == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Return the value (and flags) provided by
                        //       the file system, verbatim.
                        //
                        result = localResult;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        errors.Add(localResult);

                        //
                        // NOTE: Next, query the IHost.GetScript method for
                        //       the script.
                        //
                        if (GetHostScript(
                                interpreter, engineHost, name, ref flags,
                                ref localResult) == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Return the value (and flags) provided by
                            //       the host, verbatim.
                            //
                            result = localResult;

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            errors.Add(localResult);
                        }
                    }
                }
                else
                {
                    //
                    // NOTE: First, query the IHost.GetScript method for
                    //       the script.
                    //
                    if (GetHostScript(
                            interpreter, engineHost, name, ref flags,
                            ref localResult) == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Return the value (and flags) provided by
                        //       the host, verbatim.
                        //
                        result = localResult;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        errors.Add(localResult);

                        //
                        // NOTE: Look for the script on the native file
                        //       system.
                        //
                        if (GetFileScript(interpreter, directory, name,
                                ref flags, ref localResult) == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Return the value (and flags) provided by
                            //       the file system, verbatim.
                            //
                            result = localResult;

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            errors.Add(localResult);
                        }
                    }
                }

                result = errors;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetStartupScript(
            Interpreter interpreter,
            IEngineHost engineHost,
            string name,
            ref ScriptFlags flags,
            ref Result result,
            ref ResultList errors
            )
        {
            lock (syncRoot)
            {
                if (errors == null)
                    errors = new ResultList();

                Result localResult = null;

                if (GetHostScript(
                        interpreter, engineHost, name, ref flags,
                        ref localResult) == ReturnCode.Ok)
                {
                    result = localResult;

                    return ReturnCode.Ok;
                }
                else
                {
                    errors.Add(localResult);

                    if ((interpreter != null) && interpreter.GetVariableValue(
                        VariableFlags.GlobalOnly | VariableFlags.NoReady,
                        name, ref localResult) == ReturnCode.Ok)
                    {
                        if (GetHostScript(
                                interpreter, engineHost, localResult, ref flags,
                                ref localResult) == ReturnCode.Ok)
                        {
                            if (FlagOps.HasFlags(flags, ScriptFlags.File, true) &&
                                !PathOps.IsRemoteUri(localResult))
                            {
                                result = PathOps.ResolveFullPath(localResult);
                            }
                            else
                            {
                                result = localResult;
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            errors.Add(localResult);
                        }
                    }
                    else
                    {
                        errors.Add(localResult);
                    }
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Variable Support Methods
        public static object GetDefaultValue(
            BreakpointType breakpointType
            )
        {
            switch (breakpointType)
            {
                case BreakpointType.BeforeVariableGet:
                    return DefaultGetVariableValue;
                case BreakpointType.BeforeVariableSet:
                    return DefaultSetVariableValue;
                case BreakpointType.BeforeVariableUnset:
                    return DefaultUnsetVariableValue;
            }

            return DefaultVariableValue;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool IsWriteValueTrace(
            BreakpointType breakpointType
            )
        {
            return (breakpointType == BreakpointType.BeforeVariableSet) ||
                (breakpointType == BreakpointType.BeforeVariableReset) ||
                (breakpointType == BreakpointType.BeforeVariableUnset);
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool NeedValueForTrace(
            BreakpointType breakpointType,
            bool old
            )
        {
            return old ?
                ((breakpointType == BreakpointType.BeforeVariableReset) ||
                    (breakpointType == BreakpointType.BeforeVariableUnset)) :
                (breakpointType == BreakpointType.BeforeVariableSet);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void GatherTraceValues(
            string varIndex,
            object value,
            ObjectDictionary arrayValue,
            ref StringList values
            )
        {
            if ((varIndex == null) && (value == null) && (arrayValue == null))
                return;

            if (varIndex != null)
            {
                if (values == null)
                    values = new StringList();

                values.Add(varIndex);
            }

            if (value != null)
            {
                if (values == null)
                    values = new StringList();

                values.Add(GetStringFromObject(value));
            }

            if ((arrayValue != null) && (arrayValue.Count > 0))
            {
                if (values == null)
                    values = new StringList();

                values.Add(arrayValue.Keys);

                foreach (KeyValuePair<string, object> pair in arrayValue)
                {
                    if (pair.Value != null)
                        values.Add(GetStringFromObject(pair.Value));
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetLinkedVariableString(
            Interpreter interpreter,
            object value,
            ref Result result
            )
        {
            return MarshalOps.FixupReturnValue(
                interpreter, null, ObjectFlags.Default, null,
                value, false, false, false, ref result);
        }

        ///////////////////////////////////////////////////////////////////////

        public static void GetLinkedVariableMemberValue(
            Interpreter interpreter,
            object value,
            ref object memberValue
            )
        {
            ReturnCode code;
            MemberInfo memberInfo = null;
            Type type = null;
            object @object = null;
            Result error = null;

            code = GetLinkedVariableMemberAndValue(
                BreakpointType.None, null, null, null, value,
                ref memberInfo, ref type, ref @object,
                ref memberValue, ref error);

            if (code != ReturnCode.Ok)
                DebugOps.Complain(interpreter, code, error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode GetLinkedVariableMemberAndValue(
            BreakpointType breakpointType,
            IVariable variable,
            string name,
            string index,
            object value,
            ref MemberInfo memberInfo,
            ref Type type,
            ref object @object,
            ref object memberValue,
            ref Result error
            )
        {
            IAnyPair<MemberInfo, object> anyPair =
                value as IAnyPair<MemberInfo, object>;

            if (anyPair == null)
            {
                error = String.Format(
                    "can't {0} {1}: broken link",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index));

                return ReturnCode.Error;
            }

            memberInfo = anyPair.X;

            if (memberInfo == null)
            {
                error = String.Format(
                    "can't {0} {1}: missing member",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index));

                return ReturnCode.Error;
            }

            if (memberInfo is FieldInfo)
            {
                FieldInfo fieldInfo = (FieldInfo)memberInfo;

                type = fieldInfo.FieldType;

                if (type == null)
                {
                    error = String.Format(
                        "can't {0} {1}: missing field type",
                        FormatOps.Breakpoint(breakpointType),
                        FormatOps.ErrorVariableName(
                            variable, null, name, index));

                    return ReturnCode.Error;
                }

                @object = anyPair.Y;

                try
                {
                    memberValue = fieldInfo.GetValue(@object); /* throw */
                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    return ReturnCode.Error;
                }
            }
            else if (memberInfo is PropertyInfo)
            {
                PropertyInfo propertyInfo = (PropertyInfo)memberInfo;

                type = propertyInfo.PropertyType;

                if (type == null)
                {
                    error = String.Format(
                        "can't {0} {1}: missing property type",
                        FormatOps.Breakpoint(breakpointType),
                        FormatOps.ErrorVariableName(
                            variable, null, name, index));

                    return ReturnCode.Error;
                }

                @object = anyPair.Y;

                try
                {
                    //
                    // BUGBUG: Only non-indexed properties are currently
                    //         supported.
                    //
                    memberValue = propertyInfo.GetValue(@object, null); /* throw */
                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    return ReturnCode.Error;
                }
            }
            else
            {
                error = String.Format(
                    "can't {0} {1}: member must be field or property",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index));

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void GetLinkedVariableArrayValues(
            Interpreter interpreter,
            ObjectDictionary arrayValue,
            ref ObjectDictionary values
            )
        {
            if (arrayValue == null)
                return;

            ObjectDictionary localValues = new ObjectDictionary();

            foreach (KeyValuePair<string, object> pair in arrayValue)
            {
                object memberValue = null;

                GetLinkedVariableMemberValue(
                    interpreter, pair.Value, ref memberValue);

                if (memberValue == null)
                    continue;

                localValues.Add(pair.Key, memberValue);
            }

            if (localValues.Count > 0)
            {
                if (values == null)
                    values = new ObjectDictionary();

                values.Add(localValues);
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode SetLinkedVariableMemberValue(
            BreakpointType breakpointType,
            IVariable variable,
            string name,
            string index,
            MemberInfo memberInfo,
            object @object,
            object memberValue,
            ref Result error
            )
        {
            if (memberInfo == null)
            {
                error = String.Format(
                    "can't {0} {1}: missing member",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index));

                return ReturnCode.Error;
            }

            if (memberInfo is FieldInfo)
            {
                FieldInfo fieldInfo = (FieldInfo)memberInfo;

                try
                {
                    fieldInfo.SetValue(@object, memberValue); /* throw */
                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    return ReturnCode.Error;
                }
            }
            else if (memberInfo is PropertyInfo)
            {
                PropertyInfo propertyInfo = (PropertyInfo)memberInfo;

                try
                {
                    //
                    // BUGBUG: Only non-indexed properties are currently supported.
                    //
                    propertyInfo.SetValue(@object, memberValue, null); /* throw */
                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    return ReturnCode.Error;
                }
            }
            else
            {
                error = String.Format(
                    "can't {0} {1}: member must be field or property",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index));

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static void AddOldValuesToTraceInfo(
            ITraceInfo traceInfo,
            ObjectDictionary oldValues
            )
        {
            if ((traceInfo == null) || (oldValues == null))
                return;

            ObjectDictionary localOldValues = traceInfo.OldValues;

            if (localOldValues != null)
            {
                localOldValues.Add(oldValues);
            }
            else
            {
                localOldValues = new ObjectDictionary();
                localOldValues.Add(oldValues);

                traceInfo.OldValues = localOldValues;
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ITrace NewCoreTrace(
            TraceCallback callback,
            IClientData clientData,
            TraceFlags traceFlags,
            IPlugin plugin,
            ref Result error
            )
        {
            if (callback != null)
            {
                MethodInfo methodInfo = callback.Method;

                if (methodInfo != null)
                {
                    Type type = methodInfo.DeclaringType;

                    if (type != null)
                    {
                        _Traces.Core trace = new _Traces.Core(new TraceData(
                            FormatOps.TraceDelegateName(callback), null, null,
                            clientData, type.FullName, methodInfo.Name,
                            DelegateBindingFlags,
                            AttributeOps.GetMethodFlags(methodInfo),
                            traceFlags, plugin, 0));

                        trace.Callback = callback;
                        return trace;
                    }
                    else
                    {
                        error = "invalid trace callback method type";
                    }
                }
                else
                {
                    error = "invalid trace callback method";
                }
            }
            else
            {
                error = "invalid trace callback";
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ITraceInfo NewTraceInfo(
            Interpreter interpreter,
            ITrace trace,
            BreakpointType breakpointType,
            ICallFrame frame,
            IVariable variable,
            string name,
            string index,
            VariableFlags flags,
            object oldValue,
            object newValue,
            ObjectDictionary oldValues,
            ObjectDictionary newValues,
            StringList list,
            bool cancel,
            bool postProcess,
            ReturnCode returnCode
            )
        {
            //
            // HACK: This method is used to prevent creating a ton of redundant
            //       TraceInfo objects on the heap (i.e. whenever a variable is
            //       read, set, or unset).  Now, there is one TraceInfo object
            //       per-thread and it will be re-used as necessary.
            //
            ITraceInfo traceInfo;

            if (interpreter != null)
            {
                traceInfo = interpreter.TraceInfo;

                if (traceInfo != null)
                {
                    traceInfo = traceInfo.Update(
                       trace, breakpointType, frame, variable, name, index,
                       flags, oldValue, newValue, oldValues, newValues, list,
                       cancel, postProcess, returnCode);

                    if (traceInfo != null)
                        return traceInfo;
                }
            }

            traceInfo = new TraceInfo(
                trace, breakpointType, frame, variable, name, index,
                flags, oldValue, newValue, oldValues, newValues, list,
                cancel, postProcess, returnCode);

            if (interpreter != null)
                interpreter.TraceInfo = traceInfo;

            return traceInfo;
        }

        ///////////////////////////////////////////////////////////////////////

        //
        // WARNING: Do NOT call this method from "get" operation traces.  This
        //          method is ONLY for use by variable operations that cannot
        //          return a value (e.g. "set", "unset", "reset", "add").
        //
        public static ReturnCode FireTraces(
            IVariable variable,
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result error
            )
        {
            Result value = null;

            return FireTraces(variable, breakpointType, interpreter,
                traceInfo, ref value, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode FireTraces(
            IVariable variable,
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result value,
            ref Result error
            )
        {
            if (variable != null)
            {
                if (traceInfo != null)
                {
                    //
                    // NOTE: Save the original return code.  We will need it
                    //       later to figure out how to process the trace
                    //       callback results.
                    //
                    ReturnCode localCode = traceInfo.ReturnCode;
                    Result localResult = GetStringFromObject(traceInfo.OldValue);

                    //
                    // NOTE: Attempt to fire the traces for the variable, if any.
                    //
                    if (variable.FireTraces(
                            breakpointType, interpreter, traceInfo,
                            ref localResult) == ReturnCode.Ok)
                    {
                        //
                        // HACK: For "get" traces, we need a little bit more magic
                        //       here.
                        //
                        if (breakpointType == BreakpointType.BeforeVariableGet)
                        {
                            //
                            // NOTE: Did a trace callback cancel processing of a variable
                            //       operation that was previously regarded as unsuccessful?
                            //
                            if ((localCode != ReturnCode.Ok) && traceInfo.Cancel)
                            {
                                //
                                // NOTE: This was a failed "get" operation; however, it has
                                //       been canceled by a trace callback (presumably after
                                //       taking some more meaningful action) and is now
                                //       considered to be successful; therefore, place the
                                //       trace result into the OldValue property of the trace
                                //       object itself.  Also, this relies upon the old
                                //       value being an actual string, not a Result object.
                                //
                                traceInfo.OldValue =
                                    (localResult != null) ? localResult.ToString() : null;
                            }
                            else if ((localCode == ReturnCode.Ok) && traceInfo.Cancel)
                            {
                                //
                                // NOTE: This was a successful "get" operation; however, it
                                //       has now been canceled and the OldValue property of
                                //       the trace object will not be used.  Set the trace
                                //       result for the caller to grab.
                                //
                                value = localResult;
                            }
                        }

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        //
                        // NOTE: Give the caller the error from the trace callback.
                        //
                        error = localResult;
                    }
                }
                else
                {
                    error = "invalid trace";
                }
            }
            else
            {
                error = "invalid variable";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode SplitVariableName(
            Interpreter interpreter,
            VariableFlags flags,
            string name,
            ref string varName,
            ref string varIndex,
            ref Result error
            )
        {
            if (name != null)
            {
                if (name.Length > 0)
                {
                    if (FlagOps.HasFlags(flags, VariableFlags.NoSplit, true))
                    {
                        //
                        // HACK: Skip parsing, use the supplied name verbatim.
                        //
                        varName = name;
                        varIndex = null;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        string localVarName = null;
                        string localVarIndex = null;

                        if (Parser.SplitVariableName(name, ref localVarName,
                                ref localVarIndex, ref error) == ReturnCode.Ok)
                        {
                            if (localVarIndex != null)
                            {
                                if (!FlagOps.HasFlags(flags,
                                        VariableFlags.NoElement, true))
                                {
                                    varName = localVarName;
                                    varIndex = localVarIndex;

                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    error = "name refers to an element in an array";
                                }
                            }
                            else
                            {
                                //
                                // BUGFIX: Use the supplied name verbatim.
                                //
                                varName = name;
                                varIndex = null;

                                return ReturnCode.Ok;
                            }
                        }
                    }
                }
                else
                {
                    varName = String.Empty;
                    varIndex = null;

                    return ReturnCode.Ok;
                }
            }
            else
            {
                error = "invalid variable name";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode LinkVariable(
            Interpreter interpreter,
            ICallFrame localFrame,
            string localName,
            ICallFrame otherFrame,
            string otherName,
            ref Result error
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    if (localFrame != null)
                    {
                        localFrame = CallFrameOps.FollowLinks(localFrame);

                        if (otherFrame != null)
                        {
                            otherFrame = CallFrameOps.FollowLinks(otherFrame);

                            //
                            // NOTE: *WARNING* Empty variable names are allowed, please
                            //       do not change these to "!String.IsNullOrEmpty".
                            //
                            if (localName != null)
                            {
                                if (otherName != null)
                                {
                                    string localVarName = null;
                                    string localVarIndex = null;

                                    code = SplitVariableName(
                                        interpreter, VariableFlags.NoElement, localName,
                                        ref localVarName, ref localVarIndex, ref error);

                                    if (code == ReturnCode.Ok)
                                    {
                                        string otherVarName = null;
                                        string otherVarIndex = null;

                                        //
                                        // BUGFIX: Allow the other side of the link to
                                        //         be an array element.
                                        //
                                        code = SplitVariableName(
                                            interpreter, VariableFlags.None, otherName,
                                            ref otherVarName, ref otherVarIndex,
                                            ref error);

                                        if (code == ReturnCode.Ok)
                                        {
                                            if (!CallFrameOps.IsSame(
                                                    localFrame, otherFrame, localVarName,
                                                    otherVarName))
                                            {
                                                IVariable localVariable = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        localFrame, localVarName,
                                                        ref localVariable) == ReturnCode.Ok)
                                                {
                                                    //
                                                    // NOTE: If the local variable has been
                                                    //       flagged as undefined then go
                                                    //       ahead and allow them to use it
                                                    //       (it was not purged?).
                                                    //
                                                    if (!EntityOps.IsUndefined(localVariable))
                                                    {
                                                        //
                                                        // BUGFIX: If the local variable is a
                                                        //         link then go ahead and allow
                                                        //         them to use it.  We do this
                                                        //         for Tcl compatibility, which
                                                        //         allows for this "re-targeting"
                                                        //         of variable links to a
                                                        //         different variable.
                                                        //
                                                        if (!EntityOps.IsLink(localVariable))
                                                        {
                                                            error = String.Format(
                                                                "variable \"{0}\" already exists",
                                                                localVarName);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                }

                                                if (code == ReturnCode.Ok)
                                                {
                                                    IVariable otherVariable = null;

                                                    if (interpreter.GetVariableViaResolversWithSplit(
                                                            otherFrame, otherVarName,
                                                            ref otherVariable) == ReturnCode.Ok)
                                                    {
                                                        if (EntityOps.IsLink(otherVariable))
                                                            otherVariable = EntityOps.FollowLinks(otherVariable);

                                                        //
                                                        // NOTE: Make double sure now that we are
                                                        //       not trying to create a link to
                                                        //       ourselves.
                                                        //
                                                        if ((localVariable != null) &&
                                                            Object.ReferenceEquals(otherVariable, localVariable))
                                                        {
                                                            error = "can't upvar from variable to itself";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        otherVariable = new Variable(
                                                            otherFrame, otherVarName,
                                                            VariableFlags.None,
                                                            interpreter.GetTraces(null),
                                                            interpreter.VariableEvent);

                                                        if (interpreter.IsGlobalCallFrame(otherFrame))
                                                            EntityOps.SetGlobal(otherVariable, true);
                                                        else
                                                            EntityOps.SetLocal(otherVariable, true);

                                                        EntityOps.SetUndefined(otherVariable, true);

                                                        //
                                                        // TODO: Modify this when full [namespace]
                                                        //       support has been added.
                                                        //
                                                        string newOtherVarName =
                                                            StringOps.TrimLeadingColons(otherVarName);

                                                        otherFrame.Variables.Add(newOtherVarName, otherVariable);
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        if (localVariable != null)
                                                        {
                                                            localVariable.Reset();
                                                            localVariable.Link = otherVariable;
                                                            localVariable.LinkIndex = otherVarIndex;
                                                        }
                                                        else
                                                        {
                                                            localVariable = new Variable(
                                                                localFrame, localVarName, otherVariable,
                                                                otherVarIndex, interpreter.VariableEvent);
                                                        }

                                                        EntityOps.SetLink(localVariable, true);

                                                        //
                                                        // TODO: Modify this when full [namespace]
                                                        //       support has been added.
                                                        //
                                                        string newLocalVarName =
                                                            StringOps.TrimLeadingColons(localVarName);

                                                        if (localFrame.Variables.ContainsKey(newLocalVarName))
                                                            localFrame.Variables[newLocalVarName] = localVariable;
                                                        else
                                                            localFrame.Variables.Add(newLocalVarName, localVariable);

                                                        //
                                                        // BUGFIX: Mark the variable as "dirty"
                                                        //         AFTER the actual modifications
                                                        //         have been completed.
                                                        //
                                                        EntityOps.SetDirty(localVariable, true);
                                                    }
                                                }
                                            }
                                            else
                                            {
                                                error = "can't upvar from variable to itself";
                                                code = ReturnCode.Error;
                                            }
                                        }
                                    }
                                }
                                else
                                {
                                    error = "invalid \"other\" variable name";
                                    code = ReturnCode.Error;
                                }
                            }
                            else
                            {
                                error = "invalid \"local\" variable name";
                                code = ReturnCode.Error;
                            }
                        }
                        else
                        {
                            error = "invalid \"other\" call frame";
                            code = ReturnCode.Error;
                        }
                    }
                    else
                    {
                        error = "invalid \"local\" call frame";
                        code = ReturnCode.Error;
                    }
                }
            }
            else
            {
                error = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Data Support Methods
        public static bool ArrayEquals(
            byte[] array1,
            byte[] array2
            )
        {
            return GenericCompareOps<byte>.Equals(array1, array2);
        }

        ///////////////////////////////////////////////////////////////////////

#if DATA
        private static void GetDataReaderFieldNames(
            IDataReader reader,
            bool clear,
            ref StringList list
            )
        {
            if (reader == null)
                return;

            int fieldCount = reader.FieldCount;

            if (clear || (list == null))
                list = new StringList();

            for (int index = 0; index < fieldCount; index++)
                list.Add(reader.GetName(index));
        }

        ///////////////////////////////////////////////////////////////////////

        private static void GetDataReaderFieldValues(
            IDataReader reader,
            DateTimeBehavior dateTimeBehavior,
            string dateTimeFormat,
            bool clear,
            bool allowNull,
            bool pairs,
            bool names,
            ref StringList list
            )
        {
            if (reader == null)
                return;

            int fieldCount = reader.FieldCount;

            if (clear || (list == null))
                list = new StringList();

            for (int index = 0; index < fieldCount; index++)
            {
                object value = reader.GetValue(index);

                if (allowNull ||
                    ((value != null) && (value != DBNull.Value)))
                {
                    if (pairs)
                    {
                        StringList element = new StringList();

                        if (names)
                            element.Add(reader.GetName(index));

                        element.Add(MarshalOps.FixupDataValue(
                            value, dateTimeBehavior, dateTimeFormat,
                            null));

                        list.Add(element.ToString());
                    }
                    else
                    {
                        if (names)
                            list.Add(reader.GetName(index));

                        list.Add(MarshalOps.FixupDataValue(
                            value, dateTimeBehavior, dateTimeFormat,
                            null));
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode DataReaderToList(
            Interpreter interpreter,
            IDataReader reader,
            DateTimeBehavior dateTimeBehavior,
            string dateTimeFormat,
            int limit,
            bool allowNull,
            bool pairs,
            bool names,
            ref StringList list,
            ref Result error
            )
        {
            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (reader == null)
            {
                error = "invalid data reader";
                return ReturnCode.Error;
            }

            //
            // NOTE: Now, read the rows of data and populate the
            //       rows array.
            //
            while (reader.Read())
            {
                //
                // NOTE: Add the column values for this row to the list.
                //
                GetDataReaderFieldValues(
                    reader, dateTimeBehavior, dateTimeFormat, false,
                    allowNull, pairs, names, ref list);

                //
                // NOTE: Have we reached the limit for the number of rows we
                //       want to return, if any?
                //
                if ((limit != 0) && (--limit == 0))
                    break;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode DataReaderToArray(
            Interpreter interpreter,
            IDataReader reader,
            string varName,
            DateTimeBehavior dateTimeBehavior,
            string dateTimeFormat,
            int limit,
            bool allowNull,
            bool pairs,
            bool names,
            ref Result error
            )
        {
            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (reader == null)
            {
                error = "invalid data reader";
                return ReturnCode.Error;
            }

            ReturnCode code;

            //
            // NOTE: Make sure to reset the rows variable if it already exists.
            //
            bool badRequest = false;
            Result localError = null;

            if (((interpreter.DoesVariableExist(
                    VariableFlags.NoElement, null, varName, ref badRequest,
                    ref localError) != ReturnCode.Ok) && !badRequest) ||
                (!badRequest && interpreter.ResetVariable(
                    VariableFlags.NoElement, varName,
                    ref error) == ReturnCode.Ok))
            {
                //
                // NOTE: Get the column names.
                //
                StringList nameList = null;

                GetDataReaderFieldNames(reader, false, ref nameList);

                //
                // NOTE: Create the script array variable by setting the
                //       columns names.
                //
                code = interpreter.SetVariableValue2(
                    VariableFlags.None, varName, Vars.ResultSet.Names,
                    nameList.ToString(), null, ref error);

                if (code == ReturnCode.Ok)
                {
                    //
                    // NOTE: This is incremented inside the loop prior to be
                    //       used as an array element index; therefore, we will
                    //       never have a row zero that contains valid data.
                    //
                    int rowIndex = 0;

                    //
                    // NOTE: Now, read the rows of data and populate the
                    //       rows array.
                    //
                    while (reader.Read())
                    {
                        //
                        // NOTE: This list will contain the column names and
                        //       values for this row.  Including the column
                        //       names here as well as in the header row helps
                        //       us to detect when a given column value is null
                        //       (missing) for a row and retains semantic
                        //       compatibility with the upcoming TDBC standard.
                        //
                        StringList nameValueList = null;

                        GetDataReaderFieldValues(
                            reader, dateTimeBehavior, dateTimeFormat, false,
                            allowNull, pairs, names, ref nameValueList);

                        rowIndex++;

                        code = interpreter.SetVariableValue2(
                            VariableFlags.None, varName, rowIndex.ToString(),
                            nameValueList.ToString(), null, ref error);

                        if (code != ReturnCode.Ok)
                            break;

                        //
                        // NOTE: Have we reached the limit for the number of
                        //       rows we want to return, if any?
                        //
                        if ((limit != 0) && (--limit == 0))
                            break;
                    }

                    //
                    // NOTE: If we succeeded, set the overall row count now
                    //       that we know what it is.
                    //
                    if (code == ReturnCode.Ok)
                    {
                        code = interpreter.SetVariableValue2(
                            VariableFlags.None, varName,
                            Vars.ResultSet.Count, rowIndex.ToString(),
                            null, ref error);
                    }
                }
            }
            else
            {
                //
                // NOTE: The rows variable name is invalid (i.e. it refers
                //       to an array element) OR it already exists and we
                //       failed to reset it.
                //
                if (badRequest)
                    error = localError;

                code = ReturnCode.Error;
            }

            return code;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Package Support Methods
        private static ReturnCode FindHostPackages(
            Interpreter interpreter,
            StringList paths, /* NOT USED */
            PackageIndexCallback callback,
            bool resolve,
            bool refresh,
            ref PackageIndexDictionary packageIndexes, /* in, out */
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (paths != null) /* NOT USED */
                {
                    //
                    // NOTE: Initialize the package index collection if
                    //       necessary.
                    //
                    if (packageIndexes == null)
                        packageIndexes = new PackageIndexDictionary();

                    //
                    // NOTE: The file name we are interested in is the host
                    //       package index, named "pkgIndex".
                    //
                    string fileName = FileName.PackageIndex;

                    //
                    // NOTE: Initially mark the host package index as
                    //       "not found".  After the main search (below), if
                    //       the host package index is still marked "not found"
                    //       it will be purged.
                    //
                    if (MarkPackageIndex(
                            packageIndexes, fileName, PackageIndexFlags.Host,
                            PackageIndexFlags.Normal, PackageIndexFlags.Found,
                            true, false, false, ref error) != ReturnCode.Ok)
                    {
                        return ReturnCode.Error;
                    }

                    //
                    // NOTE: Have we seen this package index before?
                    //
                    bool exists = packageIndexes.ContainsKey(fileName);

                    //
                    // NOTE: If we are refreshing package indexes or we have
                    //       never seen this package index before, notify the
                    //       caller.
                    //
                    if (refresh || !exists)
                    {
                        Result result = null;

                        if ((callback != null) &&
                            (callback(interpreter, fileName, true, resolve,
                                refresh, ref result) != ReturnCode.Ok))
                        {
                            error = result;
                            return ReturnCode.Error;
                        }
                    }

                    //
                    // NOTE: If we have not seen this package index before add
                    //       it to the resulting collection now; otherwise,
                    //       mark it as "found" so that it will not be purged.
                    //
                    if (exists)
                        packageIndexes[fileName] |=
                            PackageIndexFlags.Host | PackageIndexFlags.Found;
                    else
                        packageIndexes.Add(fileName,
                            PackageIndexFlags.Host | PackageIndexFlags.Found);

                    //
                    // NOTE: Purge the host package index from the list if it
                    //       is still marked as "not found".
                    //
                    if (PurgePackageIndex(
                            packageIndexes, fileName, PackageIndexFlags.Host,
                            PackageIndexFlags.Normal | PackageIndexFlags.Found,
                            true, false, ref error) != ReturnCode.Ok)
                    {
                        return ReturnCode.Error;
                    }

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid paths";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode FindPackages(
            Interpreter interpreter,
            StringList paths,
            PackageIndexCallback callback,
            bool resolve,
            bool refresh,
            ref PackageIndexDictionary packageIndexes, /* in, out */
            ref Result error
            )
        {
            if (interpreter != null)
            {
                //
                // NOTE: Find all the package index files in the specified
                //       paths.
                //
                if (paths != null)
                {
                    //
                    // NOTE: Initialize the package index collection if
                    //       necessary.
                    //
                    if (packageIndexes == null)
                        packageIndexes = new PackageIndexDictionary();

                    //
                    // NOTE: Initially mark all package indexes as "not found".
                    //       After the main search loop (below), any remaining
                    //       package indexes that are still marked "not found"
                    //       will be purged.
                    //
                    if (MarkPackageIndexes(
                            packageIndexes, PackageIndexFlags.Normal,
                            PackageIndexFlags.Host, PackageIndexFlags.Found,
                            false, false, false, ref error) != ReturnCode.Ok)
                    {
                        return ReturnCode.Error;
                    }

                    //
                    // NOTE: Check each of the specified paths.
                    //
                    foreach (string path in paths)
                    {
                        //
                        // NOTE: Normalize the path prior to using it or adding
                        //       it to the dictionary.
                        //
                        string newPath = PathOps.ResolveFullPath(path);

                        //
                        // NOTE: Make sure the directory exists prior to
                        //       attempting to find any files in it; otherwise,
                        //       we just ignore it to reduce the burden on the
                        //       caller to validate that a given path actually
                        //       exists (which could be especially burdensome
                        //       if it is constructed dynamically from
                        //       environment variables, etc).
                        //
                        if (!String.IsNullOrEmpty(newPath) &&
                            Directory.Exists(newPath))
                        {
                            //
                            // NOTE: Find all package index files in the
                            //       specified directory.
                            //
                            StringList fileNames = new StringList(
                                Directory.GetFiles(newPath,
                                    FileName.PackageIndex,
                                    SearchOption.AllDirectories));

                            //
                            // NOTE: For each package index file, notify the
                            //       callback if it is new and/or add it to the
                            //       resulting collection.
                            //
                            foreach (string fileName in fileNames)
                            {
                                //
                                // NOTE: Have we seen this package index before?
                                //
                                bool exists = packageIndexes.ContainsKey(fileName);

                                //
                                // NOTE: If we are refreshing package indexes or
                                //       we have never seen this package index
                                //       before, notify the caller.
                                //
                                if (refresh || !exists)
                                {
                                    Result result = null;

                                    if ((callback != null) &&
                                        (callback(interpreter, fileName, false, resolve,
                                            refresh, ref result) != ReturnCode.Ok))
                                    {
                                        error = result;
                                        return ReturnCode.Error;
                                    }
                                }

                                //
                                // NOTE: If we have not seen this package index
                                //       before add it to the resulting
                                //       collection now; otherwise, mark it as
                                //       "found" so that it will not be purged.
                                //
                                if (exists)
                                    packageIndexes[fileName] |=
                                        PackageIndexFlags.Normal |
                                        PackageIndexFlags.Found;
                                else
                                    packageIndexes.Add(fileName,
                                        PackageIndexFlags.Normal |
                                        PackageIndexFlags.Found);
                            }
                        }
                    }

                    //
                    // NOTE: Purge any package indexes from the list that are
                    //       still marked as "not found".
                    //
                    if (PurgePackageIndexes(
                            packageIndexes, PackageIndexFlags.None,
                            PackageIndexFlags.Host | PackageIndexFlags.Found,
                            false, false, ref error) != ReturnCode.Ok)
                    {
                        return ReturnCode.Error;
                    }

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid paths";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode FindAllPackages(
            Interpreter interpreter,
            StringList paths,
            bool host,
            bool normal,
            bool resolve,
            bool refresh,
            ref PackageIndexDictionary packageIndexes,
            ref Result error
            )
        {
            if ((!host || (FindHostPackages(
                    interpreter, paths, SourcePackageIndexCallback, resolve,
                    refresh, ref packageIndexes, ref error) == ReturnCode.Ok)) &&
                (!normal || (FindPackages(
                    interpreter, paths, SourcePackageIndexCallback, resolve,
                    refresh, ref packageIndexes, ref error) == ReturnCode.Ok)))
            {
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode MarkPackageIndex(
            PackageIndexDictionary packageIndexes,
            string fileName,
            PackageIndexFlags hasFlags,
            PackageIndexFlags notHasFlags,
            PackageIndexFlags markFlags,
            bool hasAll,
            bool notHasAll,
            bool mark,
            ref Result error
            )
        {
            if (packageIndexes != null)
            {
                if (fileName != null)
                {
                    if (packageIndexes.ContainsKey(fileName))
                    {
                        PackageIndexFlags flags = packageIndexes[fileName];

                        if (FlagOps.HasFlags(flags, hasFlags, hasAll) &&
                            !FlagOps.HasFlags(flags, notHasFlags, notHasAll))
                        {
                            if (mark)
                                flags |= markFlags;
                            else
                                flags &= ~markFlags;

                            packageIndexes[fileName] = flags;
                        }
                    }

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid file name";
                }
            }
            else
            {
                error = "invalid package indexes";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode MarkPackageIndexes(
            PackageIndexDictionary packageIndexes,
            PackageIndexFlags hasFlags,
            PackageIndexFlags notHasFlags,
            PackageIndexFlags markFlags,
            bool hasAll,
            bool notHasAll,
            bool mark,
            ref Result error
            )
        {
            if (packageIndexes != null)
            {
                StringList list = new StringList(packageIndexes.Keys);

                for (int index = 0; index < list.Count; index++)
                {
                    PackageIndexFlags flags = packageIndexes[list[index]];

                    if (FlagOps.HasFlags(flags, hasFlags, hasAll) &&
                        !FlagOps.HasFlags(flags, notHasFlags, notHasAll))
                    {
                        if (mark)
                            flags |= markFlags;
                        else
                            flags &= ~markFlags;

                        packageIndexes[list[index]] = flags;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid package indexes";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode PurgePackageIndex(
            PackageIndexDictionary packageIndexes,
            string fileName,
            PackageIndexFlags hasFlags,
            PackageIndexFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            ref Result error
            )
        {
            if (packageIndexes != null)
            {
                if (fileName != null)
                {
                    if (packageIndexes.ContainsKey(fileName))
                    {
                        PackageIndexFlags flags = packageIndexes[fileName];

                        if (FlagOps.HasFlags(flags, hasFlags, hasAll) &&
                            !FlagOps.HasFlags(flags, notHasFlags, notHasAll))
                        {
                            packageIndexes.Remove(fileName);
                        }
                    }

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid file name";
                }
            }
            else
            {
                error = "invalid package indexes";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode PurgePackageIndexes(
            PackageIndexDictionary packageIndexes,
            PackageIndexFlags hasFlags,
            PackageIndexFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            ref Result error
            )
        {
            if (packageIndexes != null)
            {
                StringList list = new StringList(packageIndexes.Keys);

                for (int index = list.Count - 1; index >= 0; index--)
                {
                    PackageIndexFlags flags = packageIndexes[list[index]];

                    if (FlagOps.HasFlags(flags, hasFlags, hasAll) &&
                        !FlagOps.HasFlags(flags, notHasFlags, notHasAll))
                    {
                        packageIndexes.Remove(list[index]);
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid package indexes";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        private static ReturnCode SourcePackageIndexCallback(
            Interpreter interpreter,
            string fileName,
            bool host,
            bool resolve,
            bool refresh, /* IGNORED */
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                bool directory = false;

                try
                {
                    if (host)
                    {
                        //
                        // NOTE: It is important to note here that currently
                        //       there may only be a maximum of ONE package
                        //       index file provided by the host.
                        //
                        ScriptFlags scriptFlags =
                            ScriptFlags.PackageLibraryOptionalFile;

                        string text = interpreter.GetScript(
                            FileName.PackageIndex, ref scriptFlags);

                        if (!String.IsNullOrEmpty(text))
                        {
                            if (FlagOps.HasFlags(scriptFlags, ScriptFlags.File, true))
                            {
                                bool remoteUri = PathOps.IsRemoteUri(text);

                                if (remoteUri || File.Exists(text))
                                {
                                    string newText = text;

                                    if (resolve && !remoteUri)
                                    {
                                        //
                                        // NOTE: Attempt to resolve the file name to a
                                        //       fully qualified one.
                                        //
                                        newText = PathOps.ResolveFullPath(newText);

                                        //
                                        // NOTE: Failing that, fallback to the original
                                        //       file name which has already been
                                        //       "validated".
                                        //
                                        if (String.IsNullOrEmpty(newText))
                                            newText = text;
                                    }

                                    //
                                    // NOTE: The host for the interpreter seems
                                    //       to indicate we should be able to
                                    //       find the package index on the native
                                    //       file system?  Ok, fine.  Setup the
                                    //       directory variable properly.
                                    //
                                    code = interpreter.SetVariableValue(
                                        VariableFlags.None, TclVars.Directory,
                                        PathOps.GetDirectoryName(newText), null, ref result);

                                    if (code == ReturnCode.Ok)
                                    {
                                        directory = true;

                                        code = interpreter.EvaluateFile(newText, ref result);
                                    }
                                }
                                else
                                {
                                    result = String.Format(
                                        "the provided \"{0}\" script file \"{1}\" is not " +
                                        "a valid remote uri and does not exist locally",
                                        ScriptTypes.PackageIndex, text);

                                    code = ReturnCode.Error;
                                }
                            }
                            else
                            {
                                code = interpreter.SetVariableValue(
                                    VariableFlags.None, TclVars.Directory,
                                    null, null, ref result);

                                if (code == ReturnCode.Ok)
                                {
                                    directory = true;

                                    code = interpreter.EvaluateScript(text, ref result); /* EXEMPT */
                                }
                            }
                        }
                        else
                        {
                            //
                            // NOTE: This is optional; therefore, success.
                            //
                            code = ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        if (!String.IsNullOrEmpty(fileName))
                        {
                            bool remoteUri = PathOps.IsRemoteUri(fileName);

                            if (remoteUri || File.Exists(fileName))
                            {
                                string newFileName = fileName;

                                if (resolve && !remoteUri)
                                {
                                    //
                                    // NOTE: Attempt to resolve the file name to a
                                    //       fully qualified one.
                                    //
                                    newFileName = PathOps.ResolveFullPath(newFileName);

                                    //
                                    // NOTE: Failing that, fallback to the original
                                    //       file name which has already been
                                    //       "validated".
                                    //
                                    if (String.IsNullOrEmpty(newFileName))
                                        newFileName = fileName;
                                }

                                code = interpreter.SetVariableValue(
                                    VariableFlags.None, TclVars.Directory,
                                    PathOps.GetDirectoryName(newFileName), null, ref result);

                                if (code == ReturnCode.Ok)
                                {
                                    directory = true;

                                    code = interpreter.EvaluateFile(newFileName, ref result);
                                }
                            }
                            else
                            {
                                result = String.Format(
                                    "the provided \"{0}\" script file \"{1}\" is not " +
                                    "a valid remote uri and does not exist locally",
                                    ScriptTypes.PackageIndex, fileName);

                                code = ReturnCode.Error;
                            }
                        }
                        else
                        {
                            result = String.Format(
                                "the provided \"{0}\" script is invalid",
                                ScriptTypes.PackageIndex);

                            code = ReturnCode.Error;
                        }
                    }
                }
                catch (Exception e)
                {
                    result = String.Format(
                        "caught exception while sourcing package index: {0}",
                        e);

                    code = ReturnCode.Error;
                }
                finally
                {
                    //
                    // NOTE: If we set the directory variable, be sure to unset
                    //       it.
                    //
                    if (directory)
                    {
                        ReturnCode unsetCode;
                        Result unsetResult = null;

                        unsetCode = interpreter.UnsetVariable(
                            VariableFlags.None,
                            TclVars.Directory, ref unsetResult);

                        if (unsetCode != ReturnCode.Ok)
                            DebugOps.Complain(interpreter, unsetCode, unsetResult);
                    }
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Type Support Methods
        public static IClientData GetClientData(
            object @object
            )
        {
            IIdentifier identifier = @object as IIdentifier;

            if (identifier != null)
                return identifier.ClientData;
            else
                return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static string GetStringFromObject(
            object @object
            )
        {
            if (@object is string)
                return (string)@object;

            if (@object != null)
                return @object.ToString();

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode AddSystemObjectNamespaces(
            ref StringDictionary namespaces,
            ref Result error
            )
        {
            if (namespaces == null)
                namespaces = new StringDictionary();

            foreach (string systemNamespace in systemNamespaces)
                if (!namespaces.ContainsKey(systemNamespace))
                    namespaces.Add(systemNamespace, systemSimpleName);

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode AddAssemblyObjectNamespaces(
            ref StringDictionary namespaces,
            ref Result error
            )
        {
            if (namespaces == null)
                namespaces = new StringDictionary();

            foreach (string assemblyNamespace in assemblyNamespaces)
                if (!namespaces.ContainsKey(assemblyNamespace))
                    namespaces.Add(assemblyNamespace, assemblySimpleName);

            return ReturnCode.Ok;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Resolver Support Methods
        public static bool ShouldResolveHidden(
            EngineFlags flags,
            bool match
            )
        {
            return Engine.HasToExecute(flags) && !Engine.HasUseHidden(flags) &&
                (match ? Engine.HasMatchHidden(flags) : Engine.HasGetHidden(flags));
        }

        ///////////////////////////////////////////////////////////////////////

        public static IResolve NewCoreResolver(
            Interpreter interpreter
            )
        {
            return new _Resolvers.Core(new ResolveData(
                null, null, null, ClientData.Empty, interpreter, 0));
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Reflection Support Methods
        public static bool DoesClassTypeMatch(
            Type type,
            Type matchType,
            bool subClass
            )
        {
            bool result = false;

            if ((type != null) && (matchType != null))
            {
                //
                // NOTE: Make sure the type is a class.
                //
                if (type.IsClass)
                {
                    //
                    // NOTE: Are we matching against an interface type?
                    //
                    if (matchType.IsInterface)
                    {
                        //
                        // NOTE: Does the class implement the interface?
                        //
                        if (type.GetInterface(matchType.FullName) != null)
                            result = true;
                    }
                    else
                    {
                        //
                        // NOTE: Are the types equal; otherwise, [optionally]
                        //       is the type a sub-class of the type to match
                        //       against?
                        //
                        if (type.Equals(matchType))
                            result = true;
                        else if (subClass && type.IsSubclassOf(matchType))
                            result = true;
                    }
                }
            }
            else if ((type == null) && (matchType == null))
            {
                //
                // NOTE: If they are both null we consider that to be a match.
                //
                result = true;
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static PluginFlags GetSkipCheckPluginFlags()
        {
            PluginFlags result = PluginFlags.None;

            if (!ShouldCheckStrongNameVerified())
                result |= PluginFlags.SkipVerified;

            if (!ShouldCheckFileTrusted())
                result |= PluginFlags.SkipTrusted;

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static PluginFlags GetAssemblyPluginFlags(
            Assembly assembly,
            PluginFlags pluginFlags
            )
        {
            PluginFlags result = PluginFlags.None;

            if (assembly != null)
            {
                //
                // NOTE: Check if the plugin has a StrongName signature.
                //
                StrongName strongName = null;

                if ((AssemblyOps.GetStrongName(assembly,
                        ref strongName) == ReturnCode.Ok) &&
                    (strongName != null))
                {
                    result |= PluginFlags.StrongName;

                    //
                    // NOTE: Skip checking the StrongName signature?
                    //
                    if (!FlagOps.HasFlags(
                            pluginFlags, PluginFlags.SkipVerified, true))
                    {
                        //
                        // NOTE: See if the StrongName signature has really
                        //       been verified by the CLR itself [via the CLR
                        //       native API StrongNameSignatureVerificationEx].
                        //
                        if (IsStrongNameVerified(assembly.Location, true))
                        {
                            result |= PluginFlags.Verified;
                        }
                    }
                    else
                    {
                        result |= PluginFlags.SkipVerified;
                    }
                }

                //
                // NOTE: Check if the plugin has an Authenticode signature.
                //
                X509Certificate certificate = null;

                if ((AssemblyOps.GetCertificate(
                        assembly, ref certificate) == ReturnCode.Ok) &&
                    (certificate != null))
                {
                    result |= PluginFlags.Authenticode;

                    //
                    // NOTE: Skip checking the Authenticode signature?
                    //
                    if (!FlagOps.HasFlags(
                            pluginFlags, PluginFlags.SkipTrusted, true))
                    {
                        //
                        // NOTE: See if the Authenticode signature and
                        //       certificate are trusted by the operating
                        //       system [via the Win32 native API
                        //       WinVerifyTrust].
                        //
                        if (IsFileTrusted(assembly.Location))
                        {
                            result |= PluginFlags.Trusted;
                        }
                    }
                    else
                    {
                        result |= PluginFlags.SkipTrusted;
                    }
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode FindPrimaryPlugin(
            Assembly assembly,
            ref string typeName,
            ref Result error
            )
        {
            ReturnCode code;

            if (assembly != null)
            {
                TypeList types = null;
                ResultList errors = null;

                if (GetMatchingClassTypes(assembly, typeof(IPlugin),
                        typeof(IWrapper), true, ref types, ref errors))
                {
                    typeName = null;

                    foreach (Type type in types)
                    {
                        //
                        // NOTE: Is the plugin named "Default"?  If so, we need
                        //       to skip over it because it is used as the base
                        //       implementation for other plugins.
                        //
                        if (String.Compare(type.FullName,
                                typeof(_Plugins.Default).FullName,
                                StringOps.SystemStringComparisonType) != 0)
                        {
                            PluginFlags flags = AttributeOps.GetPluginFlags(type);

                            if (FlagOps.HasFlags(flags, PluginFlags.Primary, true))
                            {
                                typeName = type.FullName;
                                break;
                            }
                        }
                    }

                    if (typeName != null)
                    {
                        code = ReturnCode.Ok;
                    }
                    else
                    {
                        error = "no primary plugin found in assembly";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    errors.Insert(0,
                        "no plugins found in assembly");

                    error = errors;
                    code = ReturnCode.Error;
                }
            }
            else
            {
                error = "invalid assembly file name";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool GetMatchingClassTypes(
            Assembly assembly,
            Type matchType,    // must match this type
            Type nonMatchType, // must not match this type
            bool subClass,     // check sub-classes also (for not match)
            ref TypeList matchingTypes,
            ref ResultList errors
            )
        {
            bool result = false;

            if (errors == null)
                errors = new ResultList();

            try
            {
                if (assembly != null)
                {
                    if (matchingTypes == null)
                        matchingTypes = new TypeList();

                    foreach (Type type in assembly.GetTypes()) /* throw */
                    {
                        //
                        // NOTE: Make sure we are dealing with a valid type.
                        //
                        if ((type != null) &&
                            (type.IsClass || type.IsValueType))
                        {
                            //
                            // NOTE: Check the type against the matching criteria
                            //       supplied by the caller.  If it matches, add
                            //       it to the resulting list of types.
                            //
                            if (((matchType == null) ||
                                    DoesClassTypeMatch(type, matchType, subClass)) &&
                                ((nonMatchType == null) ||
                                    !DoesClassTypeMatch(type, nonMatchType, subClass)))
                            {
                                matchingTypes.Add(type);
                            }
                        }
                    }

                    result = true;
                }
            }
            catch (ReflectionTypeLoadException e)
            {
                errors.Add(e);

                //
                // NOTE: Add the loader exceptions, if they are available.
                //
                Exception[] exceptions = e.LoaderExceptions;

                if (exceptions != null)
                    foreach (Exception exception in exceptions)
                        errors.Add(exception);
            }
            catch (Exception e)
            {
                errors.Add(e);
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static bool GetMatchingDelegates(
            Assembly assembly,
            Type matchType,          // the delegate type we are expecting
            MethodFlags hasFlags,    // must match flag(s)
            MethodFlags notHasFlags, // must not match flag(s)
            bool hasAll,
            bool notHasAll,
            ref Dictionary<Delegate, MethodFlags> delegates,
            ref ResultList errors
            )
        {
            bool result = false;

            if (errors == null)
                errors = new ResultList();

            try
            {
                if (assembly != null)
                {
                    if (delegates == null)
                        delegates = new Dictionary<Delegate, MethodFlags>();

                    foreach (Type type in assembly.GetTypes()) /* throw */
                    {
                        //
                        // NOTE: Make sure we are dealing with a valid type.
                        //
                        if ((type != null) && (type.IsClass || type.IsValueType))
                        {
                            //
                            // NOTE: Check the type against the matching criteria
                            //       supplied by the caller.  If it matches, add
                            //       it to the resulting list of types.
                            //
                            MethodInfo[] methodInfo =
                                type.GetMethods(DelegateBindingFlags);

                            foreach (MethodInfo thisMethodInfo in methodInfo)
                            {
                                MethodFlags methodFlags =
                                    AttributeOps.GetMethodFlags(thisMethodInfo);

                                if (FlagOps.HasFlags(methodFlags, hasFlags, hasAll) &&
                                    !FlagOps.HasFlags(methodFlags, notHasFlags, notHasAll))
                                {
                                    Delegate @delegate =
                                        Delegate.CreateDelegate(
                                            matchType, null, thisMethodInfo, false);

                                    if (@delegate != null)
                                        delegates.Add(@delegate, methodFlags);
                                    else
                                        //
                                        // NOTE: This is not strictly an "error"; however,
                                        //       report it to the caller anyhow.
                                        //
                                        errors.Add(String.Format(
                                            "could not convert method \"{0}\" " +
                                            "to a delegate of type \"{1}\"",
                                            FormatOps.MethodFullName(
                                                thisMethodInfo.DeclaringType,
                                                thisMethodInfo.Name), matchType.FullName));
                                }
                            }
                        }
                    }

                    result = true;
                }
            }
            catch (ReflectionTypeLoadException e)
            {
                errors.Add(e);

                //
                // NOTE: Add the loader exceptions, if they are available.
                //
                Exception[] exceptions = e.LoaderExceptions;

                if (exceptions != null)
                    foreach (Exception exception in exceptions)
                        errors.Add(exception);
            }
            catch (Exception e)
            {
                errors.Add(e);
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode PopulatePluginCommands(
            IPlugin plugin,
            ref Result error
            )
        {
            if (plugin != null)
            {
                Assembly assembly = plugin.Assembly;

                if (assembly != null)
                {
                    CommandDataList commands = plugin.Commands;

                    if (commands != null)
                    {
                        TypeList types = null;
                        ResultList errors = null;

                        if (GetMatchingClassTypes(
                                assembly, typeof(ICommand),
                                typeof(IWrapper), true, ref types, ref errors))
                        {
                            foreach (Type type in types)
                                if (type != null)
                                    commands.Add(new CommandData(
                                        type.Name, null, null, null, type.FullName,
                                        AttributeOps.GetCommandFlags(type), plugin, 0));

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            errors.Insert(0,
                                "could not get matching command types");

                            error = errors;
                        }
                    }
                    else
                    {
                        error = "plugin has invalid command data list";
                    }
                }
                else
                {
                    error = "plugin has invalid assembly";
                }
            }
            else
            {
                error = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode PopulatePluginPolicies(
            IPlugin plugin,
            ref Result error
            )
        {
            if (plugin != null)
            {
                Assembly assembly = plugin.Assembly;

                if (assembly != null)
                {
                    PolicyDataList policies = plugin.Policies;

                    if (policies != null)
                    {
                        Dictionary<Delegate, MethodFlags> delegates = null;
                        ResultList errors = null;

                        if (GetMatchingDelegates(
                                assembly, typeof(ExecuteCallback),
                                MethodFlags.CommandPolicy | MethodFlags.ScriptPolicy |
                                MethodFlags.FilePolicy | MethodFlags.StreamPolicy,
                                MethodFlags.NoAdd, false, false, ref delegates, ref errors))
                        {
                            foreach (KeyValuePair<Delegate, MethodFlags> pair in delegates)
                            {
                                MethodInfo methodInfo = pair.Key.Method;

                                if (methodInfo != null)
                                {
                                    Type type = methodInfo.DeclaringType;

                                    if (type != null)
                                        policies.Add(new PolicyData(
                                            FormatOps.MethodFullName(type, methodInfo.Name),
                                            null, null, null, type.FullName, methodInfo.Name,
                                            DelegateBindingFlags, pair.Value, PolicyFlags.None,
                                            plugin, 0));
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            errors.Insert(0,
                                "could not get matching policy delegates");

                            error = errors;
                        }
                    }
                    else
                    {
                        error = "plugin has invalid policy data list";
                    }
                }
                else
                {
                    error = "plugin has invalid assembly";
                }
            }
            else
            {
                error = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode PrepareStaticPlugin(
            IPlugin plugin,
            ref Result error
            )
        {
            ReturnCode code = PopulatePluginCommands(plugin, ref error);

            if (code == ReturnCode.Ok)
                code = PopulatePluginPolicies(plugin, ref error);

            return code;
        }

        ///////////////////////////////////////////////////////////////////////

        public static void ThrowFeatureNotSupported( /* EXTERNAL USE ONLY */
            IPlugin plugin,
            string name
            )
        {
            Interpreter interpreter = Interpreter.GetActive();

            if (((interpreter == null) &&
                DefaultThrowOnFeatureNotSupported) ||
                ((interpreter != null) &&
                interpreter.ThrowOnFeatureNotSupported))
            {
                throw new ScriptException(String.Format(
                    "feature \"{0}\" not supported by plugin \"{1}\"",
                    name, plugin));
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Object Invocation Support Methods
        public static OptionDictionary GetCreateOptions()
        {
            return new OptionDictionary(
                new Option[] {
                new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                    Index.Invalid, "-objectname", null),
                new Option(null, OptionFlags.MustHaveTypeValue,
                    Index.Invalid, Index.Invalid, "-type", null),
                new Option(null, OptionFlags.MustHaveTypeListValue,
                    Index.Invalid, Index.Invalid, "-parametertypes", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-trace", null),
                new Option(typeof(ByRefArgumentFlags),
                    OptionFlags.MustHaveEnumValue, Index.Invalid,
                    Index.Invalid, "-argumentflags",
                    new Variant(MarshalOps.DefaultByRefArgumentFlags)),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nocreate", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nodispose", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noinvoke", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noargs", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-limit", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-index", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-alias", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                new Option(null, OptionFlags.MustHaveTclInterpreterValue,
                    Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noforcedelete", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-tostring", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-stricttype", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictmember", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictargs", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nocase", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-default", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-verbose", null),
                new Option(typeof(BindingFlags),
                    OptionFlags.MustHaveEnumValue, Index.Invalid,
                    Index.Invalid, "-flags",
                    new Variant(MarshalOps.DefaultBindingFlags)),
                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-objectflags",
                    new Variant(ObjectFlags.Default)),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, Option.EndOfOptions, null)
            });
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetInvokeOptions()
        {
            //
            // NOTE: The reason this is defined here is because it must be used
            //       anywhere that can create an alias that refers to this
            //       command (i.e. an "object alias").  Currently, this
            //       includes the [library] command, the [debug] command, and
            //       the custom event handling mechanism as well as the
            //       [object] command itself, specifically the "load",
            //       "create", and "invoke" sub-commands.
            //
            return new OptionDictionary(
                new Option[] {
                new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                    Index.Invalid, "-objectname", null),
                new Option(null, OptionFlags.MustHaveTypeValue |
                    OptionFlags.Unsafe, Index.Invalid, Index.Invalid,
                    "-type", null),
                new Option(null, OptionFlags.MustHaveTypeValue |
                    OptionFlags.Unsafe, Index.Invalid, Index.Invalid,
                    "-returntype", null),
                new Option(null, OptionFlags.MustHaveTypeListValue,
                    Index.Invalid, Index.Invalid, "-parametertypes", null),
                new Option(null, OptionFlags.Unsafe, Index.Invalid,
                    Index.Invalid, "-trace", null),
                new Option(typeof(ByRefArgumentFlags),
                    OptionFlags.MustHaveEnumValue, Index.Invalid,
                    Index.Invalid, "-argumentflags",
                    new Variant(MarshalOps.DefaultByRefArgumentFlags)),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-create", null),
                new Option(null, OptionFlags.Unsafe, Index.Invalid,
                    Index.Invalid, "-nodispose", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noinvoke", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noargs", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-limit", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-index", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-alias", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-aliasall", null),
                new Option(null, OptionFlags.Ignored, Index.Invalid,
                    Index.Invalid, "-invoke", null),
                new Option(null, OptionFlags.Ignored, Index.Invalid,
                    Index.Invalid, "-keepresults", null),
                new Option(null, OptionFlags.Ignored, Index.Invalid,
                    Index.Invalid, "-nocomplain", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-invokeall", null),
#if NATIVE && TCL
                new Option(null, OptionFlags.MustHaveTclInterpreterValue |
                    OptionFlags.Unsafe, Index.Invalid, Index.Invalid,
                    "-tcl", null),
#endif
                new Option(null, OptionFlags.Unsafe, Index.Invalid,
                    Index.Invalid, "-noforcedelete", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-tostring", null),
                new Option(null, OptionFlags.Unsafe, Index.Invalid,
                    Index.Invalid, "-verbose", null), /* FIXME: Unsafe? */
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nocase", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-default", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nonestedobject", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nonestedmember", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-stricttype", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictmember", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictargs", null),
                new Option(typeof(MemberTypes), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-membertypes",
                    new Variant(MarshalOps.DefaultMemberTypes)),
                new Option(typeof(BindingFlags), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-flags",
                    new Variant(MarshalOps.DefaultBindingFlags)),
                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-objectflags",
                    new Variant(ObjectFlags.Default)),
                new Option(null, OptionFlags.Unsafe, Index.Invalid,
                    Index.Invalid, "-identity", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, Option.EndOfOptions, null)
            });
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetInvokeAllOptions()
        {
            OptionDictionary options = GetInvokeOptions();

            options["-invoke"].Flags &= ~OptionFlags.Ignored;
            options["-invokeall"].Flags |= OptionFlags.Ignored;

            options["-keepresults"].Flags &= ~OptionFlags.Ignored;
            options["-nocomplain"].Flags &= ~OptionFlags.Ignored;

            return options;
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetDisposeOptions()
        {
            return new OptionDictionary(
                new Option[] {
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-synchronous", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nodispose", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nocomplain", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, Option.EndOfOptions, null)
            });
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetObjectOptions(
            ObjectOptionType type
            )
        {
            switch (type)
            {
                case ObjectOptionType.Create:
                    return GetCreateOptions();
                case ObjectOptionType.Invoke:
                    return GetInvokeOptions();
                case ObjectOptionType.InvokeAll:
                    return GetInvokeAllOptions();
                case ObjectOptionType.Dispose:
                    return GetDisposeOptions();
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetInvokeOptions(
            bool all
            )
        {
            return GetObjectOptions(all ?
                ObjectOptionType.InvokeAll : ObjectOptionType.Invoke);
        }

        ///////////////////////////////////////////////////////////////////////

        public static OptionDictionary GetCallOptions()
        {
            //
            // NOTE: These options are used by both the InvokeDelegate method
            //       (below) and the code for the [library call] command.
            //       Normally, this method would simply call into a static
            //       method exported from the _Commands.Library class; however,
            //       that class is only available when the library has been
            //       compiled with native code enabled; therefore, we define
            //       the actual options here and both the _Commands.Library
            //       class and the InvokeDelegate method can simply call us.
            //
            return new OptionDictionary(
                new Option[] {
                new Option(null, OptionFlags.MustHaveValue, Index.Invalid,
                    Index.Invalid, "-objectname", null),
                new Option(null, OptionFlags.MustHaveTypeValue, Index.Invalid,
                    Index.Invalid, "-type", null),
                new Option(typeof(ByRefArgumentFlags),
                    OptionFlags.MustHaveEnumValue, Index.Invalid,
                    Index.Invalid, "-argumentflags",
                    new Variant(MarshalOps.DefaultByRefArgumentFlags)),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-create", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nodispose", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noinvoke", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noargs", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-limit", null),
                new Option(null, OptionFlags.MustHaveIntegerValue,
                    Index.Invalid, Index.Invalid, "-index", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-alias", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-aliasall", null),
#if NATIVE && TCL
                new Option(null, OptionFlags.MustHaveTclInterpreterValue,
                    Index.Invalid, Index.Invalid, "-tcl", null),
#endif
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-noforcedelete", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-tostring", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-nocase", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-default", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictmember", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-strictargs", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-verbose", null),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, "-trace", null),
                new Option(typeof(BindingFlags), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-flags",
                    new Variant(MarshalOps.DefaultBindingFlags)),
                new Option(typeof(ObjectFlags), OptionFlags.MustHaveEnumValue,
                    Index.Invalid, Index.Invalid, "-objectflags",
                    new Variant(ObjectFlags.Default)),
                new Option(null, OptionFlags.None, Index.Invalid,
                    Index.Invalid, Option.EndOfOptions, null)
            });
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode InvokeDelegate(
            Interpreter interpreter,
            Delegate @delegate,
            ArgumentList arguments,
            ref Result result
            )
        {
            ///////////////////////////////////////////////////////////////////
            //                       ARGUMENT VALIDATION
            ///////////////////////////////////////////////////////////////////

            if (interpreter == null)
            {
                result = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (@delegate == null)
            {
                result = "invalid delegate";
                return ReturnCode.Error;
            }

            if (arguments == null)
            {
                result = "invalid arguments";
                return ReturnCode.Error;
            }

            ///////////////////////////////////////////////////////////////////
            //                        OPTION PROCESSING
            ///////////////////////////////////////////////////////////////////

            ReturnCode code = ReturnCode.Ok;
            OptionDictionary options = GetCallOptions();
            int argumentIndex = Index.Invalid;

            if (arguments.Count > 1)
                code = interpreter.GetOptions(
                    options, arguments, 0, 1, Index.Invalid, true,
                    ref argumentIndex, ref result);

            if (code != ReturnCode.Ok)
                return code;

            //
            // NOTE: Do we expect to create an object (e.g. System.Int32.Parse)?
            //
            bool create = false;

            if (options.IsPresent("-create"))
                create = true;

            bool dispose = true;

            if (options.IsPresent("-nodispose"))
                dispose = false;

            bool invoke = true;

            if (options.IsPresent("-noinvoke"))
                invoke = false;

            bool noArgs = false;

            if (options.IsPresent("-noargs"))
                noArgs = true;

            //
            // NOTE: Do we want to create a command alias for any newly created
            //       objects?
            //
            bool alias = false;

            if (options.IsPresent("-alias"))
                alias = true;

            bool aliasAll = false;

            if (options.IsPresent("-aliasall"))
                aliasAll = true;

            //
            // NOTE: Forcibly convert the object to a string and discard it?
            //
            bool toString = false;

            if (options.IsPresent("-tostring"))
                toString = true;

            bool noCase = false;

            if (options.IsPresent("-nocase"))
                noCase = true;

            bool @default = false;

            if (options.IsPresent("-default"))
                @default = true;

            bool verbose = false;

            if (options.IsPresent("-verbose"))
                verbose = true;

            bool strictMember = false;

            if (options.IsPresent("-strictmember"))
                strictMember = true;

            bool strictArgs = false;

            if (options.IsPresent("-strictargs"))
                strictArgs = true;

            bool trace = false;

            if (options.IsPresent("-trace"))
                trace = true;

            Variant value = null;
            ByRefArgumentFlags argumentFlags =
                MarshalOps.DefaultByRefArgumentFlags;

            if (options.IsPresent("-argumentflags", ref value))
                argumentFlags = (ByRefArgumentFlags)value.Value;

            Type type = null;

            if (options.IsPresent("-type", ref value))
                type = (Type)value.Value;

            BindingFlags bindingFlags = MarshalOps.DefaultBindingFlags;

            //
            // NOTE: Now check for and use the -nocase value.  It is also
            //       important to note here that a specifying the flags
            //       overrides this setting.  Therefore, if you want
            //       case-insensitive binding when specifying both the -nocase
            //       and -flags options, the IgnoreCase flag should also be
            //       included in your option value.
            //
            if (noCase)
                bindingFlags |= BindingFlags.IgnoreCase;

            if (options.IsPresent("-flags", ref value))
                bindingFlags = (BindingFlags)value.Value;

            int limit = invoke ? 1 : 0;

            if (options.IsPresent("-limit", ref value))
                limit = (int)value.Value;

            int index = Index.Invalid;

            if (options.IsPresent("-index", ref value))
                index = (int)value.Value;

            string interpName = null;

#if NATIVE && TCL
            if (options.IsPresent("-tcl", ref value))
                interpName = value.ToString();
#endif

            ObjectFlags objectFlags = ObjectFlags.Default;

            if (options.IsPresent("-objectflags", ref value))
                objectFlags = (ObjectFlags)value.Value;

            if (options.IsPresent("-noforcedelete"))
                objectFlags &= ~ObjectFlags.ForceDelete;

            string objectName = null;

            if (options.IsPresent("-objectname", ref value))
                objectName = value.ToString();

            ///////////////////////////////////////////////////////////////////
            //                    METHOD ARGUMENT BUILDING
            ///////////////////////////////////////////////////////////////////

            object[] args = null;
            int argumentCount = 0;

            if ((argumentIndex != Index.Invalid) &&
                (argumentIndex < arguments.Count))
            {
                //
                // NOTE: How many arguments were supplied?
                //
                argumentCount = (arguments.Count - argumentIndex);

                //
                // NOTE: Create and populate the array of arguments for the
                //       invocation.
                //
                args = new object[argumentCount];

                for (int index2 = argumentIndex; index2 < arguments.Count;
                        index2++)
                {
                    /* need String, not Argument */
                    args[index2 - argumentIndex] = arguments[index2].ToString();
                }
            }
            else if (invoke || !noArgs)
            {
                //
                // FIXME: When no arguments are specified, we actually need an
                //        array of zero arguments for the parameter to argument
                //        matching code to work correctly.
                //
                args = new object[0];
            }

            //
            // HACK: We want to use the existing marshalling code; therefore,
            //       we pre-bake some of the required arguments here (i.e. we
            //       KNOW what method we are going to call, however we want
            //       magical bi-directional type coercion, etc).
            //
            object delegateTarget = @delegate.Target;
            MethodInfo delegateMethodInfo = @delegate.Method;

            if (delegateMethodInfo == null)
            {
                result = "cannot invoke delegate, no method";
                return ReturnCode.Error;
            }

            string newObjectName = (delegateTarget != null) ?
                delegateTarget.GetType().FullName : null;

            string newMemberName = delegateMethodInfo.Name;
            MethodInfo[] methodInfo = new MethodInfo[] { delegateMethodInfo };

            if (methodInfo == null) // NOTE: Redundant [for now].
            {
                result = String.Format(
                    "delegate \"{0}\" has no methods matching \"{1}\"",
                    newObjectName, bindingFlags);

                return ReturnCode.Error;
            }

            ///////////////////////////////////////////////////////////////////
            //                    METHOD ARGUMENT CONVERSION
            ///////////////////////////////////////////////////////////////////

            IntList methodIndexList = null;
            ObjectArrayList argsList = null;
            ArgumentInfoList argumentInfoList = null;
            ResultList errors = null;

            //
            // NOTE: Attempt to convert the argument strings to something
            //       potentially more meaningful and find the corresponding
            //       method.
            //
            code = MarshalOps.FindMethodsAndFixupArguments(
                interpreter, interpreter.Binder, interpreter.CultureInfo,
                @delegate.GetType(), newObjectName, newMemberName,
                MemberTypes.Method, bindingFlags, methodInfo, null, args,
                limit, verbose, @default, ref methodIndexList, ref argsList,
                ref argumentInfoList, ref errors);

            if (code != ReturnCode.Ok)
            {
                result = errors;
                return code;
            }

            ///////////////////////////////////////////////////////////////////
            //                    METHOD OVERLOAD SELECTION
            ///////////////////////////////////////////////////////////////////

            if ((methodIndexList == null) || (methodIndexList.Count == 0) ||
                (argsList == null) || (argsList.Count == 0))
            {
                result = String.Format(
                    "method \"{0}\" of delegate \"{1}\" not found",
                    newMemberName, newObjectName);

                return ReturnCode.Error;
            }

            if ((index != Index.Invalid) &&
                ((index < 0) || (index >= methodIndexList.Count) ||
                (index >= argsList.Count)))
            {
                result = String.Format(
                    "method \"{0}\" of delegate \"{1}\" not found, " +
                    "invalid method index {2}, must be {3}",
                    newMemberName, newObjectName, index,
                    FormatOps.BetweenOrExact(0, methodIndexList.Count - 1));

                return ReturnCode.Error;
            }

            //
            // NOTE: Are we actually going to invoke the method or are we simply
            //       returning the list of matching method overloads?  For this
            //       method, the list of method overloads should always have
            //       exactly one result (i.e. it is somewhat redundant; however,
            //       it is designed to match the semantics of [object invoke]).
            //
            if (invoke)
            {
                if (strictMember && (methodIndexList.Count != 1))
                {
                    result = String.Format(
                        "matched {0} method overloads of \"{1}\" on delegate " +
                        "\"{2}\", need exactly 1", methodIndexList.Count,
                        newMemberName, newObjectName);

                    return ReturnCode.Error;
                }

                //
                // FIXME: Select the first method that matches.  More
                //        sophisticated logic may need to be added here later.
                //
                int methodIndex = (index != Index.Invalid) ?
                    methodIndexList[index] : methodIndexList[0];

                if (methodIndex == Index.Invalid)
                {
                    result = String.Format(
                        "method \"{0}\" of delegate \"{1}\" not found",
                        newMemberName, newObjectName);

                    return ReturnCode.Error;
                }

                ///////////////////////////////////////////////////////////////
                //               METHOD ARGUMENT ARRAY SELECTION
                ///////////////////////////////////////////////////////////////

                //
                // NOTE: Get the arguments we are going to use to perform
                //       the actual method call.
                //
                args = (index != Index.Invalid) ? argsList[index] : argsList[0];

                ///////////////////////////////////////////////////////////////
                //                       METHOD TRACING
                ///////////////////////////////////////////////////////////////

                if (trace)
                    DebugTrace(String.Format(
                        "{0}: InvokeDelegate: methodIndex = {1}, " +
                        "delegate = {2}, args = {3}{4}",
                        GlobalState.GetCurrentThreadId(), methodIndex,
                        FormatOps.WrapOrNull(@delegate),
                        FormatOps.WrapOrNull(new StringList(args)),
                        Environment.NewLine), typeof(RuntimeOps).Name);

                ///////////////////////////////////////////////////////////////
                //                      METHOD INVOCATION
                ///////////////////////////////////////////////////////////////

                object returnValue = null;

                code = Engine.ExecuteDelegate(
                    @delegate, args, ref returnValue, ref result);

                ///////////////////////////////////////////////////////////////
                //                   BYREF ARGUMENT HANDLING
                ///////////////////////////////////////////////////////////////

                if ((code == ReturnCode.Ok) && (argumentInfoList != null))
                {
                    code = MarshalOps.FixupByRefArguments(
                        interpreter, interpreter.Binder,
                        interpreter.CultureInfo, argumentInfoList, objectFlags,
                        GetInvokeOptions(aliasAll), interpName, args,
                        argumentFlags, strictArgs, create, dispose, alias,
                        aliasAll, toString, ref result);
                }

                ///////////////////////////////////////////////////////////////
                //                    RETURN VALUE HANDLING
                ///////////////////////////////////////////////////////////////

                if (code == ReturnCode.Ok)
                {
                    code = MarshalOps.FixupReturnValue(
                        interpreter, interpreter.Binder,
                        interpreter.CultureInfo, type, objectFlags,
                        GetInvokeOptions(aliasAll), objectName, interpName,
                        returnValue, create, dispose, alias, aliasAll,
                        toString, ref result);
                }
            }
            else
            {
                MethodInfoList methodInfoList = new MethodInfoList();

                if (index != Index.Invalid)
                    methodInfoList.Add(methodInfo[methodIndexList[index]]);
                else
                    foreach (int methodIndex in methodIndexList)
                        methodInfoList.Add(methodInfo[methodIndex]);

                code = MarshalOps.FixupReturnValue(
                    interpreter, interpreter.Binder, interpreter.CultureInfo,
                    type, objectFlags, GetInvokeOptions(aliasAll), objectName,
                    interpName, methodInfoList, create, dispose, alias,
                    aliasAll, toString, ref result);
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region Object Disposal Support Methods
        public static ReturnCode TryDisposeObject(
            object @object,  /* in */
            ref Result error /* out */
            )
        {
            return TryDisposeObject(@object, DefaultDispose, ref error);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode TryDisposeObject(
            object @object,  /* in */
            bool dispose,    /* in */
            ref Result error /* out */
            )
        {
            Exception exception = null;

            return TryDisposeObject(
                @object, ref dispose, ref error, ref exception);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode TryDisposeObject(
            object @object,   /* in */
            ref bool dispose, /* out */
            ref Result error  /* out */
            )
        {
            Exception exception = null;

            return TryDisposeObject(
                @object, ref dispose, ref error, ref exception);
        }

        ///////////////////////////////////////////////////////////////////////

        public static ReturnCode TryDisposeObject(
            object @object,         /* in */
            ref bool dispose,       /* out */
            ref Result error,       /* out */
            ref Exception exception /* out */
            )
        {
            if (@object != null)
            {
                try
                {
                    IDisposable disposable = @object as IDisposable;

                    if (disposable != null)
                    {
                        try
                        {
                            disposable.Dispose(); /* throw */
                            dispose = true; /* disposed */

                            return ReturnCode.Ok; /* success */
                        }
                        catch (Exception e)
                        {
                            //
                            // NOTE: The object threw an exception during its
                            //       disposal.  This is technically allowed;
                            //       however, it is typically discouraged.
                            //       Save the information and report it to our
                            //       caller.
                            //
                            error = e;
                            exception = e;

                            return ReturnCode.Error; /* failure */
                        }
                        finally
                        {
                            disposable = null;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Object does not implement IDisposable.
                        //
                        dispose = false; /* not disposable */

                        return ReturnCode.Ok; /* success */
                    }
                }
                finally
                {
                    @object = null;
                }
            }
            else
            {
                //
                // NOTE: Object is not valid.
                //
                dispose = false; /* invalid object */

                return ReturnCode.Ok;
            }
        }
        #endregion
    }
}
