/*
 * Interpreter.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections.Generic;

#if CAS_POLICY
using System.Configuration.Assemblies;
#endif

#if DATA
using System.Data;
#endif

using System.Globalization;
using System.IO;

#if NETWORK
using System.Net.Sockets;
#endif

using System.Reflection;
using System.Resources;

#if CAS_POLICY
using System.Security;
#endif

using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Security.Permissions;
using System.Security.Policy;
using System.Text;
using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Private;

#if NATIVE && TCL
using Eagle._Components.Private.Tcl;
#endif

using Eagle._Components.Public.Delegates;
using Eagle._Constants;
using Eagle._Containers.Private;

#if NATIVE && TCL
using Eagle._Containers.Private.Tcl;
#endif

using Eagle._Containers.Public;
using Eagle._Encodings;
using Eagle._Interfaces.Private;

#if NATIVE && TCL
using Eagle._Interfaces.Private.Tcl;
#endif

using Eagle._Interfaces.Public;

#if !CONSOLE
using ConsoleColor = Eagle._Components.Public.ConsoleColor;
#endif

#if DEBUGGER
using _Private = Eagle._Components.Private;
#endif

using _Public = Eagle._Components.Public;

namespace Eagle._Components.Public
{
    [ObjectId("b98042f9-d66a-46db-81e9-a2f44c1f03fd")]
    public sealed class Interpreter :
#if ISOLATED_INTERPRETERS || ISOLATED_PLUGINS
        ScriptMarshalByRefObject,
#endif
#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
        ICacheManager,
#endif
        IDisposable, IDebugManager, IEngineManager,
        IEntityManager, IErrorManager,
#if HISTORY
        IHistoryManager,
#endif
        IInteractiveManager, IInterpreter,
#if NOTIFY || NOTIFY_OBJECT
        INotifyManager,
#endif
        IPackageManager,
        IPluginManager, IPropertyManager, IResolveManager,
        IResourceManager, IScriptManager, ISecurityManager,
        ISynchronize,
#if NATIVE && TCL
        ITclManager,
#endif
        IVariableManager
    {
        #region Global Read-Only Data (shared by all Interpreters)
        #region Static Data Lock -- Private
        private static readonly object staticSyncRoot = new object();
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Master "Safe" Variable / Array Element Name Lists -- Private
        private static StringList safeVariableNames;
        private static StringList safeTclPlatformElementNames;
        private static StringList safeEaglePlatformElementNames;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Host Integration -- Private
        internal static readonly bool DefaultQuiet = false;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System & Default Encodings -- Public
        //
        // WARNING: If any of these encodings are changed things may not work correctly.
        //
        public static readonly Encoding SystemEncoding = new UnicodeEncoding(false, false);
        public static readonly Encoding DefaultEncoding = new UTF8Encoding(false);
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Global Data (shared by all Interpreters) -- Private
        private static int breakCount = 0;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Global Host Integration Data -- Private
        private static NewHostCallback newHostCallback;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Global Tcl/Tk Integration Data -- Private
        #region Zombie Tcl/Tk Object Tracking -- Private
#if NATIVE && TCL
        private static int deadTclThreadId;
        private static ITclApi deadTclApi;
        private static IntPtrDictionary deadTclInterps;
        private static TclBridgeDictionary deadTclBridges;

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if TCL_THREADS
        private static TclThreadDictionary deadTclThreads;
#endif
#endif
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Constants
        #region Timeout Defaults
#if NETWORK
        private static readonly int DefaultThreadStartTimeout = 3000;
        private static readonly int MinimumSocketPollTimeout = 500; /* in milliseconds */
#endif

        private static readonly int DefaultTimeout = _Timeout.Infinite; // never timeout by default.
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region History & Caching
#if PARSE_CACHE || LIST_CACHE || TYPE_CACHE || COM_TYPE_CACHE
        private const int MaximumCacheSize = 8000;                 /* TODO: *PERF* Good default? */

#if PARSE_CACHE || LIST_CACHE
        private const int MaximumCacheItemSize = 8000;
#endif

#if CACHE_DICTIONARY
        private const int MinimumCacheSize = MaximumCacheSize / 2; /* TODO: *PERF* Good default? */
#endif
#endif

#if HISTORY
        private static readonly int DefaultHistoryLimit = 100; /* TODO: Good default? */
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Error Messages
        #region Resource Error Messages (FIXED-STRING)
        //
        // NOTE: These six strings must be compile time constants because
        //       they are used before the cultureInfo and resourceManager
        //       objects are available to resolve runtime string resources.
        //
        private const string CultureInfoError =
            "could not interpret \"{0}\" as a culture name or identifier";

        private const string InvalidCultureInfoError =
            "invalid culture";

        private const string InvalidBaseResourceName =
            "invalid base resource name";

        private const string ResourceManagerError =
            "could not create resource manager \"{0}\"";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Cancellation Error Messages
        private const string demandCancelResultFormat = "user canceled";

        private const string timeoutCancelPromptFormat =
            "possible runaway script ({0} milliseconds), cancel?";

        private const string timeoutCancelResultFormat =
            "runaway script unwound after {0} milliseconds";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Procedure Declaration Error Messages
        //
        // NOTE: These two strings are used by the procedure declaration engine.
        //       For localization, we need to make them into properties and load
        //       them from a resource file.
        //
        internal static readonly string ArgumentNotScalarError =
            "formal parameter \"{0}\" is an array element";

        internal static readonly string ArgumentNotSimpleError =
            "formal parameter \"{0}\" is not a simple name";
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Opaque Handles
#if NATIVE && TCL
        #region Tcl Interpreters
        internal const string tclMasterInterpPrefix = "masterInterp";
        private const string tclSafeInterpPrefix = "safeInterp";
        private const string tclInterpPrefix = "interp";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Threads (Isolated Worker Threads)
#if TCL_THREADS
        private const string tclThreadPrefix = "thread";
#endif
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Loop
        #region Interactive Prompt Defaults
        private const string PrimaryPrompt = "% ";
        private const string ContinuePrompt = ">\t";

        private const string DebugPrefix = "(debug) ";
        private const string QueuePrefix = "^ ";

        private static readonly StringList DefaultPrompts = new StringList(new string[] {
            null, PrimaryPrompt, ContinuePrompt
        });
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Nesting Level Defaults
#if NATIVE && WINDOWS
        //
        // NOTE: The number of nesting levels before we start checking
        //       native stack space.
        //
        // TODO: We really need to adjust these numbers dynamically
        //       depending on the maximum stack size of the thread.
        //
        private static readonly int NoStackLevels = 100;
        private static readonly int NoStackParserLevels = 100;
        private static readonly int NoStackExpressionLevels = 100;
#endif

        private static readonly int DefaultRecursionLimit = 1000;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Context Data
#if THREADING
        private IContextManager contextManager;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Read-Only Data
        //
        // NOTE: The AppDomain-wide unique Id for this interpreter.
        //
        private readonly long id = GlobalState.NextInterpreterId();

        //
        // FIXME: Review and revise usage of locking in the interpreter object,
        //        paying careful attention to any routines whose primary purpose
        //        is to modify collections of data (e.g. commands, plugins, etc).
        //
        internal readonly object syncRoot = new object();
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Context Data (!THREADING, per-interpreter, per-thread)
        #region Private Engine Context Data (!THREADING, per-interpreter, per-thread)
#if !THREADING
        private int levels;           // evaluation nesting level
        private int previousLevels;   // previous evaluation nesting level (for [test*], etc.)
        private int parserLevels;     // parser nesting level
        private int expressionLevels; // expression nesting level
        private int unknownLevels;    // [unknown] nesting level
        private int traceLevels;      // variable trace nesting level

#if NOTIFY || NOTIFY_OBJECT
        private int notifyLevels;     // notification nesting level
#endif

        private int policyLevels;     // policy checking nesting level
        private int testLevels;       // the nesting level of the test(s).

        private bool stackOverflow;   // stack overflow flag, always reset at level 0.

#if PREVIOUS_RESULT
        private Result previousResult;
#endif

        private ReturnCode returnCode; // NOTE: For internal engine use only.

        private int errorLine;    // NOTE: For internal engine use only.
        private string errorCode; // NOTE: For internal engine use only.
        private string errorInfo; // NOTE: For internal engine use only.
        private int errorFrames;  // NOTE: For internal engine use only.

        private IScriptLocation scriptLocation; // used only when set by [info script ?fileName?]
        private StackList<IScriptLocation> scriptLocations;    // used by [info] command directly.

        private int previousProcessId;   // id of the previous Process to be [exec]'d

        private ArraySearchDictionary arraySearches;

#if HISTORY
        private IHistoryFilter historyEngineFilter;
        private ClientDataList _history;
#endif
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Interactive Context Data (!THREADING, per-interpreter, per-thread)
#if !THREADING
        private bool interactive;
        private string interactiveInput;
        private string previousInteractiveInput;
        private int interactiveLoops;

#if HISTORY
        private IHistoryData historyLoadData;
        private IHistoryData historySaveData;

        private IHistoryFilter historyInfoFilter;
        private IHistoryFilter historyLoadFilter;
        private IHistoryFilter historySaveFilter;

        private string historyFileName;
#endif
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Test Context Data (!THREADING, per-interpreter, per-thread)
#if !THREADING
        //
        // NOTE: TclTest related state.
        //
        private int[] testStatistics; // total tests, passed, failed, etc.
        private StringList testConstraints;
        private StringList testSkipped;
        private StringList testFailures;
        private IntDictionary testCounts; // all tests that have been run.
        private StringList testMatch;
        private StringList testSkip;
        private ReturnCodeDictionary testReturnCodeMessages; // used by the test suite.

#if DEBUGGER
        private StringDictionary testBreakpoints;
#endif

        private string testPath;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Variable Context Data (!THREADING, per-interpreter, per-thread)
#if !THREADING
        private CallStack callStack;       // the script call stack.
        private ICallFrame procedureFrame; // current procedure call frame, if any.
        private ICallFrame currentFrame;   // always the current call frame.
        private ITraceInfo traceInfo;      // current trace info, if any.
#endif

        //
        // WARNING: This field is always the global call frame; however, it is
        //          for use by the ContextManager class only.  Please use the
        //          "GlobalFrame" property instead.
        //
        internal ICallFrame globalFrame;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Data (per-interpreter)
        #region Plugin Token Data
        internal long corePluginToken;

#if NOTIFY || NOTIFY_OBJECT
        internal long objectPluginToken;
#endif

#if DEBUG
        internal long testPluginToken;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Cancellation Data
        //
        // NOTE: These cannot be per-thread because we allow script cancellation
        //       to occur from any thread.
        //
        private bool cancel;
        private bool unwind;
        private bool halt;

        private Result cancelResult;
        private Result haltResult;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Threading Data
        private Thread cancelThread;
        internal Thread testGcThread;

        private int threadId; // NOTE: The "primary" (i.e. creation) thread for this interpreter.
        private Thread thread;
        private EventWaitHandle variableEvent;

        private Thread interactiveThread; // NOTE: The current interactive thread for this interpreter, if any.
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Extensibility Data
        private object applicationObject;
        private object policyObject;
        private object resolverObject;
        private object userObject;
        private StringDictionary runtimeOptions;
        private bool throwOnFeatureNotSupported;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Xml Data
#if XML
        private bool validateXml;
        private bool allXml;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Host Integration Data
#if ISOLATED_PLUGINS
        internal IHost coreHost;
#endif

        private IHost host;
        private bool quiet;

#if POLICY_TRACE
        private bool policyTrace;
#endif

        //
        // NOTE: This is not currently used.
        //
        private TextWriter traceTextWriter;

        //
        // NOTE: This is written to by "DebugOps.Complain" for persisting the error information
        //       to a stream (e.g. a file on disk).
        //
        private TextWriter debugTextWriter;

        //
        // NOTE: This is written to by "DebugOps.Complain" for inclusion in the interactive
        //       debugger output.
        //
        private string complaint;

        //
        // NOTE: The system default foreground and background colors for this interpreter.  Right
        //       now, these are only used in certain circumstances (i.e. when GetColors is called
        //       and cannot determine from the interpreter host what color settings to use).
        //
        private ConsoleColor defaultForegroundColor;
        private ConsoleColor defaultBackgroundColor;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Library Initialization Data
        private string preInitializeText; // saved copy of the pre-init script.

        private string libraryPath;       // the location of the "init" script.
        private StringList autoPathList;  // the list of additional dirs to check for packages.

        internal string initializedPath;  // path actually used for the "init" script.
        private StringList trustedPaths;  // paths trusted by "safe" interpreters.

        //
        // NOTE: The managed object types trusted by "safe" interpreters
        //       (for static invoke only).  Instance object types are
        //       trusted on a per-instance basis.
        //
        private ObjectDictionary trustedTypes;

        private bool initialized; // core script library initialized?
        private bool initializedShell; // shell script library initialized?
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl/Tk Integration Data
#if NATIVE && TCL
        internal ITclApi tclApi;

        internal IntPtrDictionary tclInterps; // TODO: per-thread?

#if TCL_THREADS
        internal TclThreadDictionary tclThreads;
#endif

        internal TclBridgeDictionary tclBridges; // TODO: per-thread?
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region CLR Integration Data
        private AppDomain appDomain;
        private IBinder binder;

        private StringDictionary objectTypes;
        private StringDictionary objectNamespaces;
        private TypePairDictionary<string, long> objectInterfaces;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Cache Data
#if PARSE_CACHE
        internal ParseStateDictionary parseStateCache;

        #region Dead Code
#if DEAD_CODE
        internal ExpressionStateDictionary expressionStateCache;
#endif
        #endregion
#endif

#if LIST_CACHE
        internal StringListDictionary stringListCache;
#endif

#if EXECUTE_CACHE
        internal ExecuteCache executeCache;
#endif

#if TYPE_CACHE
        internal TypeDictionary typeCache;
#endif

#if COM_TYPE_CACHE
        internal IntPtrTypeListDictionary comTypeListCache;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Entity Data
#if DATA
        internal DbConnectionDictionary connections;   // used by [sql] command directly.
        internal DbTransactionDictionary transactions; // used by [sql] command directly.
#endif

        internal PolicyWrapperDictionary policies;
        internal TraceWrapperDictionary traces;
        internal ChannelDictionary channels;
        internal CallFrameDictionary scopes; // used by [scope] command directly.
        internal ObjectWrapperDictionary objects;

#if APPDOMAINS
        internal AppDomainDictionary appDomains;
#endif

        internal List<IResolve> resolvers;

        internal PluginWrapperDictionary plugins;
        internal FunctionWrapperDictionary functions;
        internal OperatorWrapperDictionary operators;
        internal AliasWrapperDictionary aliases;
        internal CommandWrapperDictionary commands;
        internal CommandWrapperDictionary hiddenCommands;
        internal CallbackDictionary callbacks;

#if CALLBACK_QUEUE
        internal CallbackQueue callbackQueue;
#endif

        internal ProcedureWrapperDictionary procedures;
        internal ProcedureWrapperDictionary hiddenProcedures;
        internal ExecuteWrapperDictionary executes;
        internal ExecuteWrapperDictionary hiddenExecutes;
        // internal LambdaWrapperDictionary lambdas;
        internal PackageIndexDictionary packageIndexes;
        internal PackageWrapperDictionary packages;
        internal EncodingDictionary encodings;

#if NATIVE && LIBRARY
        internal ModuleWrapperDictionary modules;
        internal DelegateWrapperDictionary delegates;
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Core Trace Lists
        private TraceList autoPathTraceList;
        private TraceList environmentTraceList;
        private TraceList linkedVariableTraceList;
        private TraceList objectTraceList;
        private TraceList precisionTraceList;
        private TraceList testsTraceList;
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Slave Interpreter Data
        internal Interpreter masterInterpreter;
        internal string slaveName;
        internal InterpreterDictionary slaveInterpreters;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Resource Data
        private CultureInfo cultureInfo;
        internal ResourceManager resourceManager;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Debugger Data
#if DEBUGGER && !THREADING
        private IDebugger debugger;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: This is used by the Debugger class when it needs to break into
        //       the interactive loop.
        //
        private InteractiveLoopCallback interactiveLoopCallback;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Expression Engine Data
        internal Random random;
        internal RandomNumberGenerator cryptoRandom;

        private int precision; // COMPAT: Tcl.
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Event Data
        private IEventManager eventManager;
        private int waitCount;
        private int waitSpinCount;

#if NETWORK
        private int serverSockets;
#endif

        private int sleepTime;
        private int timeout;
        private int threadStackSize;

        private string backgroundError; // NOTE: Background error handler.
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Finalization Data
        private int disposeCount;
        private bool disposed;
        private bool deleted;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Engine Data
        private long nextId;
        private bool readOnly;

        private PolicyDecision commandDecision;
        private PolicyDecision scriptDecision;
        private PolicyDecision fileDecision;
        private PolicyDecision streamDecision;

        private long commandCount; // COMPAT: Tcl.
        private int recursionLimit; // COMPAT: Tcl.

        private bool exit;
        private ExitCode exitCode;

#if HISTORY
        private bool history;
        private int historyLimit;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Flags Data
        internal CreateFlags createFlags;
        private ScriptFlags scriptFlags;

#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
        private CacheFlags cacheFlags;
#endif

        private EngineFlags engineFlags;
        private SubstitutionFlags substitutionFlags;
        private ExpressionFlags expressionFlags;
        private HeaderFlags headerFlags;
        private InterpreterFlags interpreterFlags;
        private EngineFlags interactiveEngineFlags;
        private SubstitutionFlags interactiveSubstitutionFlags;
        private EventFlags interactiveEventFlags;
        private ExpressionFlags interactiveExpressionFlags;
        private EventFlags afterEventFlags;
        private EventFlags engineEventFlags;
        private EventFlags queueEventFlags;
        private EventFlags serviceEventFlags;
        private EventFlags updateEventFlags;
        private EventFlags waitEventFlags;
        private PluginFlags pluginFlags;

#if NOTIFY || NOTIFY_OBJECT
        private NotifyType notifyTypes;
        private NotifyFlags notifyFlags;
        private bool notify;
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Unknown Handler Data
        private string unknown; // NOTE: Unknown command handler.
        private string packageUnknown; // NOTE: Unknown package handler.
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICacheManager Members
#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
        public int ClearCaches(
            CacheFlags cacheFlags
            )
        {
            // CheckDisposed();

            int result = 0;

            lock (syncRoot)
            {
                #region StringList Cache
#if LIST_CACHE
                if ((stringListCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.StringList, true))
                {
                    result += stringListCache.Count;
                    stringListCache.Clear();
                }
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

#if PARSE_CACHE
                #region ParseState Cache
                if ((parseStateCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.ParseState, true))
                {
                    result += parseStateCache.Count;
                    parseStateCache.Clear();
                }
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region ExpressionState Cache
                #region Dead Code
#if DEAD_CODE
                if ((expressionStateCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.ExpressionState, true))
                {
                    result += expressionStateCache.Count;
                    expressionStateCache.Clear();
                }
#endif
                #endregion
                #endregion
#endif

                ///////////////////////////////////////////////////////////////////////////////////////

                #region IExecute Cache
#if EXECUTE_CACHE
                if ((executeCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.IExecute, true))
                {
                    result += executeCache.Count;
                    executeCache.Clear();
                }
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Type Cache
#if TYPE_CACHE
                if ((typeCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.Type, true))
                {
                    result += typeCache.Count;
                    typeCache.Clear();
                }
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region ComTypeList Cache
#if COM_TYPE_CACHE
                if ((comTypeListCache != null) &&
                    FlagOps.HasFlags(cacheFlags, CacheFlags.ComTypeList, true))
                {
                    result += comTypeListCache.Count;
                    comTypeListCache.Clear();
                }
#endif
                #endregion
            }

            return result;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDebugManager Members
        public bool Debug
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return FlagOps.HasFlags(createFlags, CreateFlags.Debug, true);
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    if (value)
                        createFlags |= CreateFlags.Debug;
                    else
                        createFlags &= ~CreateFlags.Debug;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUGGER
        public bool SingleStep
        {
            get
            {
                CheckDisposed();

#if !THREADING
                lock (syncRoot)
#endif
                {
                    IDebugger debugger = Debugger;

                    return (debugger != null) ?
                        debugger.SingleStep : false;
                }
            }
            set
            {
                CheckDisposed();

#if !THREADING
                lock (syncRoot)
#endif
                {
                    IDebugger debugger = Debugger;

                    if (debugger != null)
                        debugger.SingleStep = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsDebuggerAvailable()
        {
            CheckDisposed();

#if !THREADING
            lock (syncRoot)
#endif
            {
                return (Debugger != null);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsDebuggerActive()
        {
            CheckDisposed();

            return PrivateIsDebuggerActive();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public InteractiveLoopCallback InteractiveLoopCallback
        {
            get
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.InteractiveLoopCallback;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return interactiveLoopCallback;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.InteractiveLoopCallback = value;
#else
                lock (syncRoot)
                {
                    interactiveLoopCallback = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public TextWriter TraceTextWriter /* WARNING: NOT USED */
        {
            get { CheckDisposed(); lock (syncRoot) { return traceTextWriter; } }
            set { CheckDisposed(); lock (syncRoot) { traceTextWriter = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public TextWriter DebugTextWriter
        {
            get { CheckDisposed(); lock (syncRoot) { return debugTextWriter; } }
            set { CheckDisposed(); lock (syncRoot) { debugTextWriter = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
#if DEBUGGER
        private bool PrivateIsDebuggerActive()
        {
#if !THREADING
            lock (syncRoot)
#endif
            {
                IDebugger debugger = Debugger;

                return ((debugger != null) && (debugger.Active > 0));
            }
        }
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IEngineManager Members
        #region Script Cancellation
        public ReturnCode CancelEvaluate(
            bool unwind,
            Result result,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.CancelEvaluate(
                this, unwind, result, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CancelAnyEvaluate(
            bool unwind,
            bool strict,
            Result result,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code;
            Result localError = null;

            //
            // NOTE: Cancel any outstanding script in the interpreter.
            //
            code = Engine.CancelEvaluate(this, unwind, result, ref localError);

            if (code != ReturnCode.Ok)
            {
                if (strict)
                {
                    error = localError;
                    return code;
                }
                else
                {
                    DebugOps.Complain(this, code, localError);

                    code = ReturnCode.Ok;
                }
            }

            //
            // NOTE: Attempt to cancel any outstanding Tcl scripts in the Tcl
            //       interpreters for this interpreter.
            //
#if NATIVE && TCL
            ITclApi tclApi = null;
            IntPtrList interps = null;

            lock (syncRoot)
            {
                //
                // NOTE: Get the Tcl API object from the interpreter.
                //
                tclApi = this.tclApi;

                //
                // NOTE: Get the list of Tcl interpreters from the interpreter.
                //
                if (HasTclInterpreters())
                    interps = new IntPtrList(tclInterps.Values);
            }

            //
            // NOTE: If the interpreter has a Tcl API object and a list of
            //       interpreters, try to cancel them all now.  It must also
            //       support script cancellation (via TIP 285).
            //
            if ((tclApi != null) && (tclApi.CancelEval != null) && (interps != null))
            {
                IClientData clientData = null;

                foreach (IntPtr interp in interps)
                {
                    code = TclWrapper.CancelEvaluate(tclApi, interp, unwind,
                        result, ref clientData, ref localError);

                    if (code != ReturnCode.Ok)
                    {
                        if (strict)
                        {
                            error = localError;
                            return code;
                        }
                        else
                        {
                            DebugOps.Complain(this, code, localError);

                            code = ReturnCode.Ok;
                        }
                    }
                }
            }
#endif

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool Cancel
        {
            get
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    return cancel;
                //}
            }
            set
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    cancel = value;

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Interpreter, NotifyFlags.Canceled,
                        cancel, this,
                        null, null, null);
#endif
                //}
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool Unwind
        {
            get
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    return unwind;
                //}
            }
            set
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    unwind = value;

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Interpreter, NotifyFlags.Unwound,
                        unwind, this,
                        null, null, null);
#endif
                //}
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool Halt
        {
            get
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    return halt;
                //}
            }
            set
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    halt = value;

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Interpreter, NotifyFlags.Halted,
                        halt, this,
                        null, null, null);
#endif
                //}
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal Result CancelResult
        {
            get
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    return cancelResult;
                //}
            }
            set
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    cancelResult = value;
                //}
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal Result HaltResult
        {
            get
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    return haltResult;
                //}
            }
            set
            {
                // CheckDisposed();

                //lock (syncRoot)
                //{
                    haltResult = value;
                //}
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Evaluation
        public ReturnCode EvaluateScript(
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateScript(
            string text,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, ref result,
                ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateScript(
            IScript script,
            ref Result result
            )
        {
            CheckDisposed();

            int errorLine = 0;

            return EvaluateScript(script, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateScript(
            IScript script,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, script, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalScript(
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalScript(
            string text,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateFile(
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            int errorLine = 0;

            return EvaluateFile(
                fileName, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateFile(
            string fileName,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(
                this, fileName, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, ref result,
                ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateFile(
            Encoding encoding,
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            int errorLine = 0;

            return EvaluateFile(
                encoding, fileName, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateFile(
            Encoding encoding,
            string fileName,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(
                this, encoding, fileName, engineFlags,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalFile(
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            int errorLine = 0;

            return EvaluateGlobalFile(
                fileName, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalFile(
            string fileName,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(
                this, fileName, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalFile(
            Encoding encoding,
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            int errorLine = 0;

            return EvaluateGlobalFile(
                encoding, fileName, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalFile(
            Encoding encoding,
            string fileName,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(this, encoding, fileName,
                engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateStream(
                this, name, textReader, startIndex, characters,
                engineFlags, substitutionFlags, engineEventFlags,
                expressionFlags, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            ref Result result,
            ref int errorLine
            )
        {
            CheckDisposed();

            return Engine.EvaluateStream(
                this, name, textReader, startIndex, characters,
                engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode EvaluateScript(
            string text,
            IScriptLocation location,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            PushScriptLocation(location);

            try
            {
#endif
                return EvaluateScript(text, ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(location);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateScript( /* CORE COMMAND USE ONLY */
            Argument argument,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            PushScriptLocation((IScriptLocation)argument);

            try
            {
#endif
                return EvaluateScript((string)argument, ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(argument);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateScript( /* FOR [catch] USE ONLY */
            Argument argument,
            EngineFlags engineFlags,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            PushScriptLocation((IScriptLocation)argument);

            try
            {
#endif
                return Engine.EvaluateScript(
                    this, argument, engineFlags, substitutionFlags,
                    engineEventFlags, expressionFlags, ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(argument);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateScript(
            ArgumentList arguments,
            int startIndex,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

            int errorLine = 0;

            return EvaluateScript(arguments, startIndex, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateScript(
            ArgumentList arguments,
            int startIndex,
            ref Result result,
            ref int errorLine
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            IScriptLocation location = null;

            if (RuntimeOps.GetScriptLocation(arguments, startIndex,
                    ref location, ref result) != ReturnCode.Ok)
            {
                return ReturnCode.Error;
            }

            PushScriptLocation(location);

            try
            {
#endif
                return EvaluateScript(
                    ListOps.Concat(arguments, startIndex), ref result,
                    ref errorLine);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(location);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluatePromptScript(
            string text,
            ref Result result,
            ref int errorLine
            )
        {
            return Engine.EvaluateScript(
                this, text, engineFlags | EngineFlags.ResetCancel,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result, ref errorLine);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Expression Evaluation
        public ReturnCode EvaluateExpression(
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.EvaluateExpression(
                this, text, engineFlags,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode EvaluateExpression(
            Argument argument,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            PushScriptLocation((IScriptLocation)argument);

            try
            {
#endif
                return EvaluateExpression((string)argument, ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(argument);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateExpression(
            ArgumentList arguments,
            int startIndex,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            IScriptLocation location = null;

            if (RuntimeOps.GetScriptLocation(arguments, startIndex,
                    ref location, ref result) != ReturnCode.Ok)
            {
                return ReturnCode.Error;
            }

            PushScriptLocation(location);

            try
            {
#endif
                return EvaluateExpression(
                    ListOps.Concat(arguments, startIndex), ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(location);
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode EvaluateExpressionWithErrorInfo(
            Argument argument,
            string errorInfo,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

#if DEBUGGER && BREAKPOINTS
            PushScriptLocation((IScriptLocation)argument);

            try
            {
#endif
                return Engine.EvaluateExpressionWithErrorInfo(
                    this, argument, engineFlags, substitutionFlags,
                    engineEventFlags, expressionFlags, errorInfo,
                    ref result);
#if DEBUGGER && BREAKPOINTS
            }
            finally
            {
                PopScriptLocation(argument);
            }
#endif
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Substitution Processing
        public ReturnCode SubstituteString(
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteString(text, substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteString(
            string text,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteString(
                this, text, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalString(
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteGlobalString(text, substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalString(
            string text,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteString(
                this, text, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteFile(
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteFile(
                fileName, substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteFile(
            string fileName,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteFile(
                this, fileName, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalFile(
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteGlobalFile(
                fileName, substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalFile(
            string fileName,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteFile(
                this, fileName, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteStream(
                name, textReader, startIndex, characters,
                substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteStream(
                this, name, textReader, startIndex, characters,
                engineFlags, substitutionFlags, engineEventFlags,
                expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteGlobalStream(
                name, textReader, startIndex, characters,
                substitutionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalStream(
            string name,
            TextReader textReader,
            int startIndex,
            int characters,
            SubstitutionFlags substitutionFlags,
            ref Result result
            )
        {
            CheckDisposed();

            return Engine.SubstituteStream(
                this, name, textReader, startIndex, characters,
                engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Asynchronous Script Evaluation
        public ReturnCode EvaluateScript(
            string text,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags, substitutionFlags, engineEventFlags,
                expressionFlags, callback, clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateFile(
            string fileName,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(
                this, fileName, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, callback,
                clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalScript(
            string text,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.EvaluateScript(
                this, text, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                callback, clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateGlobalFile(
            string fileName,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.EvaluateFile(
                this, fileName, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                callback, clientData, ref error);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Asynchronous Substitution Processing
        public ReturnCode SubstituteString(
            string text,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return SubstituteString(
                text, substitutionFlags, callback, clientData,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteString(
            string text,
            SubstitutionFlags substitutionFlags,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.SubstituteString(
                this, text, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, callback,
                clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalString(
            string text,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return SubstituteGlobalString(
                text, substitutionFlags, callback, clientData,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalString(
            string text,
            SubstitutionFlags substitutionFlags,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.SubstituteString(
                this, text, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                callback, clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteFile(
            string fileName,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return SubstituteFile(
                fileName, substitutionFlags,
                callback, clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteFile(
            string fileName,
            SubstitutionFlags substitutionFlags,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.SubstituteFile(
                this, fileName, engineFlags, substitutionFlags,
                engineEventFlags, expressionFlags, callback,
                clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalFile(
            string fileName,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return SubstituteGlobalFile(
                fileName, substitutionFlags,
                callback, clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteGlobalFile(
            string fileName,
            SubstitutionFlags substitutionFlags,
            AsynchronousCallback callback, /* NOTE: May be null for "fire-and-forget" type scripts. */
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            return Engine.SubstituteFile(
                this, fileName, engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, engineEventFlags, expressionFlags,
                callback, clientData, ref error);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Command Execution
        public ReturnCode Invoke(
            string name,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            CheckDisposed();

            EngineFlags engineFlags = GetResolveEngineFlags(true);
            IExecute execute = null;

            if (GetIExecuteViaResolvers(engineFlags, name, arguments, true,
                    true, ref execute, ref result) == ReturnCode.Ok)
            {
                return Engine.Execute(
                    name, execute, this, clientData, arguments,
                    engineFlags, substitutionFlags, engineEventFlags,
                    expressionFlags, ref result);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal long CommandCount
        {
            get
            {
                lock (syncRoot)
                {
                    return commandCount;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void IncrementCommandCount()
        {
            Interlocked.Increment(ref commandCount);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode Execute(
            string name,
            IExecute execute,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

            return Engine.Execute(
                name, execute, this, clientData, arguments, engineFlags,
                substitutionFlags, engineEventFlags, expressionFlags,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Hidden IProcedure, IExecute, ICommand
        internal ReturnCode ExecuteHidden(
            string name,
            IExecute execute,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            // CheckDisposed(); /* NON-PUBLIC, SANITY */

            return Engine.Execute(
                name, execute, this, clientData, arguments,
                engineFlags | EngineFlags.InvokeHidden,
                substitutionFlags, engineEventFlags,
                expressionFlags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode MoveExposedAndHiddenCommands(
            ref Result error
            )
        {
            /* CheckDisposed(); */

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if ((procedures != null) && (hiddenProcedures != null))
                    {
                        ProcedureWrapperDictionary localProcedures;

                        ///////////////////////////////////////////////////////////////////////////////

                        localProcedures = new ProcedureWrapperDictionary(procedures);

                        foreach (KeyValuePair<string, _Wrappers.Procedure> pair in localProcedures)
                        {
                            if ((pair.Value != null) && EntityOps.IsHidden(pair.Value))
                            {
                                hiddenProcedures.Add(pair.Key, pair.Value);
                                procedures.Remove(pair.Key);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(pair.Key, false);
#endif
                            }
                        }

                        ///////////////////////////////////////////////////////////////////////////////

                        localProcedures = new ProcedureWrapperDictionary(hiddenProcedures);

                        foreach (KeyValuePair<string, _Wrappers.Procedure> pair in localProcedures)
                        {
                            if ((pair.Value != null) && !EntityOps.IsHidden(pair.Value))
                            {
                                procedures.Add(pair.Key, pair.Value);
                                hiddenProcedures.Remove(pair.Key);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(pair.Key, false);
#endif
                            }
                        }
                    }

                    if ((commands != null) && (hiddenCommands != null))
                    {
                        CommandWrapperDictionary localCommands;

                        ///////////////////////////////////////////////////////////////////////////////

                        localCommands = new CommandWrapperDictionary(commands);

                        foreach (KeyValuePair<string, _Wrappers.Command> pair in localCommands)
                        {
                            if ((pair.Value != null) && EntityOps.IsHidden(pair.Value))
                            {
                                hiddenCommands.Add(pair.Key, pair.Value);
                                commands.Remove(pair.Key);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(pair.Key, false);
#endif
                            }
                        }

                        ///////////////////////////////////////////////////////////////////////////////

                        localCommands = new CommandWrapperDictionary(hiddenCommands);

                        foreach (KeyValuePair<string, _Wrappers.Command> pair in localCommands)
                        {
                            if ((pair.Value != null) && !EntityOps.IsHidden(pair.Value))
                            {
                                commands.Add(pair.Key, pair.Value);
                                hiddenCommands.Remove(pair.Key);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(pair.Key, false);
#endif
                            }
                        }
                    }

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ExposeCommand(
            string name,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    IExecute execute = null;

                    if (GetIExecuteViaResolvers(
                            GetResolveEngineFlags(true) | EngineFlags.UseHidden, name,
                            null, true, true, ref execute) == ReturnCode.Ok)
                    {
                        if (execute is _Wrappers._Execute)
                        {
                            if (HasIExecutes(ref error) && HasHiddenIExecutes(ref error))
                            {
                                _Wrappers._Execute executeWrapper = (_Wrappers._Execute)execute;

                                //
                                // TODO: Modify this when full [namespace] support has been added.
                                //
                                string newName = StringOps.TrimLeadingColons(name);

                                executes.Add(newName, executeWrapper);
                                hiddenExecutes.Remove(newName);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(newName, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (execute is _Wrappers.Procedure)
                        {
                            if (HasProcedures(ref error) && HasHiddenProcedures(ref error))
                            {
                                _Wrappers.Procedure procedureWrapper = (_Wrappers.Procedure)execute;

                                procedures.Add(procedureWrapper.Name, procedureWrapper);
                                hiddenProcedures.Remove(procedureWrapper.Name);

                                procedureWrapper.Flags &= ~ProcedureFlags.Hidden;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(procedureWrapper.Name, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (execute is _Wrappers.Command)
                        {
                            if (HasCommands(ref error) && HasHiddenCommands(ref error))
                            {
                                _Wrappers.Command commandWrapper = (_Wrappers.Command)execute;

                                commands.Add(commandWrapper.Name, commandWrapper);
                                hiddenCommands.Remove(commandWrapper.Name);

                                commandWrapper.Flags &= ~CommandFlags.Hidden;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(commandWrapper.Name, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "unknown execution type for \"{0}\"",
                                name);
                        }
                    }
                    else
                    {
                        error = String.Format(
                            "unknown hidden command \"{0}\"",
                            name);
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode HideCommand(
            string name,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    IExecute execute = null;

                    if (GetIExecuteViaResolvers(
                            GetResolveEngineFlags(true), name, null,
                            true, true, ref execute) == ReturnCode.Ok)
                    {
                        if (execute is _Wrappers._Execute)
                        {
                            if (HasIExecutes(ref error) && HasHiddenIExecutes(ref error))
                            {
                                _Wrappers._Execute executeWrapper = (_Wrappers._Execute)execute;

                                //
                                // TODO: Modify this when full [namespace] support has been added.
                                //
                                string newName = StringOps.TrimLeadingColons(name);

                                hiddenExecutes.Add(newName, executeWrapper);
                                executes.Remove(newName);

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(newName, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (execute is _Wrappers.Procedure)
                        {
                            if (HasProcedures(ref error) && HasHiddenProcedures(ref error))
                            {
                                _Wrappers.Procedure procedureWrapper = (_Wrappers.Procedure)execute;

                                hiddenProcedures.Add(procedureWrapper.Name, procedureWrapper);
                                procedures.Remove(procedureWrapper.Name);

                                procedureWrapper.Flags |= ProcedureFlags.Hidden;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(procedureWrapper.Name, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (execute is _Wrappers.Command)
                        {
                            if (HasCommands(ref error) && HasHiddenCommands(ref error))
                            {
                                _Wrappers.Command commandWrapper = (_Wrappers.Command)execute;

                                hiddenCommands.Add(commandWrapper.Name, commandWrapper);
                                commands.Remove(commandWrapper.Name);

                                commandWrapper.Flags |= CommandFlags.Hidden;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.Remove(commandWrapper.Name, false);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "unknown execution type for \"{0}\"",
                                name);
                        }
                    }
                    else
                    {
                        error = String.Format(
                            "unknown command \"{0}\"",
                            name);
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IEntityManager Members
        #region Entity Checking
        public bool HasAliases(ref Result error)
        {
            CheckDisposed();

            bool result = (aliases != null);

            if (!result)
                error = "aliases not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasCallbacks(ref Result error)
        {
            CheckDisposed();

            bool result = (callbacks != null);

            if (!result)
                error = "callbacks not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasChannels(ref Result error)
        {
            CheckDisposed();

            bool result = (channels != null);

            if (!result)
                error = "channels not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasCommands(ref Result error)
        {
            CheckDisposed();

            bool result = (commands != null);

            if (!result)
                error = "commands not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasFunctions(ref Result error)
        {
            CheckDisposed();

            bool result = (functions != null);

            if (!result)
                error = "functions not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasIExecutes(ref Result error)
        {
            CheckDisposed();

            bool result = (executes != null);

            if (!result)
                error = "executes not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasObjects(ref Result error)
        {
            CheckDisposed();

            bool result = (objects != null);

            if (!result)
                error = "objects not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasOperators(ref Result error)
        {
            CheckDisposed();

            bool result = (operators != null);

            if (!result)
                error = "operators not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasPackageIndexes(ref Result error)
        {
            CheckDisposed();

            bool result = (packageIndexes != null);

            if (!result)
                error = "package indexes not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasPackages(ref Result error)
        {
            CheckDisposed();

            bool result = (packages != null);

            if (!result)
                error = "packages not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasPlugins(ref Result error)
        {
            CheckDisposed();

            bool result = (plugins != null);

            if (!result)
                error = "plugins not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasPolicies(ref Result error)
        {
            CheckDisposed();

            bool result = (policies != null);

            if (!result)
                error = "policies not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasProcedures(ref Result error)
        {
            CheckDisposed();

            bool result = (procedures != null);

            if (!result)
                error = "procedures not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasScopes(ref Result error)
        {
            CheckDisposed();

            bool result = (scopes != null);

            if (!result)
                error = "scopes not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasTraces(ref Result error)
        {
            CheckDisposed();

            bool result = (traces != null);

            if (!result)
                error = "traces not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if CALLBACK_QUEUE
        public bool HasCallbackQueue(ref Result error)
        {
            CheckDisposed();

            bool result = (callbackQueue != null);

            if (!result)
                error = "callback queue not available";

            return result;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DATA
        public bool HasDbConnections(ref Result error)
        {
            CheckDisposed();

            bool result = (connections != null);

            if (!result)
                error = "database connections not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasDbTransactions(ref Result error)
        {
            CheckDisposed();

            bool result = (transactions != null);

            if (!result)
                error = "database transactions not available";

            return result;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && LIBRARY
        public bool HasDelegates(ref Result error)
        {
            CheckDisposed();

            bool result = (delegates != null);

            if (!result)
                error = "delegates not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasModules(ref Result error)
        {
            CheckDisposed();

            bool result = (modules != null);

            if (!result)
                error = "modules not available";

            return result;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Entity Detection
        public ReturnCode DoesAliasExist(string name)
        {
            CheckDisposed();

            IAlias alias = null;
            Result error = null;

            return GetAlias(name, false, true, ref alias, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesCallbackExist(string name)
        {
            CheckDisposed();

            ICallback callback = null;
            Result error = null;

            return GetCallback(name, false, true, ref callback, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesChannelExist(string name)
        {
            CheckDisposed();

            Result error = null;
            Channel channel = GetChannel(name, ref error);

            if (channel != null)
                return ReturnCode.Ok;
            else
                return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesCommandExist(string name)
        {
            CheckDisposed();

            ICommand command = null;
            Result error = null;

            return GetCommand(name, false, true, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesFunctionExist(string name)
        {
            CheckDisposed();

            IFunction function = null;
            Result error = null;

            return GetFunction(name, false, true, ref function, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesIExecuteExist(string name)
        {
            CheckDisposed();

            IExecute execute = null;
            Result error = null;

            return GetIExecute(name, false, true, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesObjectExist(string name)
        {
            CheckDisposed();

            IObject @object = null;
            Result error = null;

            return GetObject(name, false, true, ref @object, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesOperatorExist(string name)
        {
            CheckDisposed();

            IOperator @operator = null;
            Result error = null;

            return GetOperator(name, false, true, ref @operator, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesPackageExist(string name)
        {
            CheckDisposed();

            IPackage package = null;
            Result error = null;

            return GetPackage(name, false, true, ref package, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesPluginExist(string name)
        {
            CheckDisposed();

            IPlugin plugin = null;
            Result error = null;

            return GetPlugin(name, false, true, ref plugin, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesPolicyExist(string name)
        {
            CheckDisposed();

            IPolicy policy = null;
            Result error = null;

            return GetPolicy(name, false, true, ref policy, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesProcedureExist(string name)
        {
            CheckDisposed();

            IProcedure procedure = null;
            Result error = null;

            return GetProcedure(name, false, true, ref procedure, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesScopeExist(string name)
        {
            CheckDisposed();

            ICallFrame frame = null;
            Result error = null;

            return GetScope(name, ref frame, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesTraceExist(string name)
        {
            CheckDisposed();

            ITrace trace = null;
            Result error = null;

            return GetTrace(name, false, true, ref trace, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && LIBRARY
        public ReturnCode DoesDelegateExist(string name)
        {
            CheckDisposed();

            IDelegate @delegate = null;
            Result error = null;

            return GetDelegate(name, false, true, ref @delegate, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesModuleExist(string name)
        {
            CheckDisposed();

            IModule module = null;
            Result error = null;

            return GetModule(name, false, true, ref module, ref error);
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode DoesAliasExist(long token)
        {
            string name = null;
            ICommand command = null;
            Result error = null;

            if (GetCommand(token, false, true, ref name, ref command, ref error) == ReturnCode.Ok)
                if (FlagOps.HasFlags(command.Flags, CommandFlags.Alias, true))
                    return ReturnCode.Ok;

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DoesCommandExist(long token)
        {
            string name = null;
            ICommand command = null;
            Result error = null;

            return GetCommand(token, false, true, ref name, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DoesFunctionExist(long token)
        {
            string name = null;
            IFunction function = null;
            Result error = null;

            return GetFunction(token, false, true, ref name, ref function, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DoesPolicyExist(long token)
        {
            string name = null;
            IPolicy policy = null;
            Result error = null;

            return GetPolicy(token, false, true, ref name, ref policy, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DoesTraceExist(long token)
        {
            string name = null;
            ITrace trace = null;
            Result error = null;

            return GetTrace(token, false, true, ref name, ref trace, ref error);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Entity Management
        #region IIdentifier
        public ReturnCode GetIdentifier(
            IdentifierKind kind,
            string name,
            bool wrapper,
            bool validate,
            ref IIdentifier identifier,
            ref Result error
            )
        {
            CheckDisposed();

            switch (kind)
            {
                case IdentifierKind.Command:
                    {
                        ICommand command = null;

                        if (GetCommand(name, wrapper, validate, ref command, ref error) == ReturnCode.Ok)
                        {
                            identifier = command;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.HiddenCommand:
                    {
                        ICommand command = null;

                        if (GetHiddenCommand(name, wrapper, validate, ref command, ref error) == ReturnCode.Ok)
                        {
                            identifier = command;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Procedure:
                    {
                        IProcedure procedure = null;

                        if (GetProcedure(name, wrapper, validate, ref procedure, ref error) == ReturnCode.Ok)
                        {
                            identifier = procedure;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.HiddenProcedure:
                    {
                        IProcedure procedure = null;

                        if (GetHiddenProcedure(name, wrapper, validate, ref procedure, ref error) == ReturnCode.Ok)
                        {
                            identifier = procedure;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.IExecute: /* NOTE: IExecute does not derive from IIdentifier. */
                case IdentifierKind.HiddenIExecute:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                case IdentifierKind.Lambda:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                case IdentifierKind.Operator:
                    {
                        IOperator @operator = null;

                        if (GetOperator(name, wrapper, validate, ref @operator, ref error) == ReturnCode.Ok)
                        {
                            identifier = @operator;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Function:
                    {
                        IFunction function = null;

                        if (GetFunction(name, wrapper, validate, ref function, ref error) == ReturnCode.Ok)
                        {
                            identifier = function;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Variable:
                    {
                        VariableFlags flags = VariableFlags.NoElement;
                        IVariable variable = null;

                        if (GetVariableViaResolversWithSplit(
                                name, ref flags, ref variable, ref error) == ReturnCode.Ok)
                        {
                            identifier = variable;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.CallFrame:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                case IdentifierKind.Package:
                    {
                        IPackage package = null;

                        if (GetPackage(name, wrapper, validate, ref package, ref error) == ReturnCode.Ok)
                        {
                            identifier = package;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Plugin:
                    {
                        IPlugin plugin = null;

                        if (GetPlugin(name, wrapper, validate, ref plugin, ref error) == ReturnCode.Ok)
                        {
                            identifier = plugin;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Object:
                    {
                        IObject @object = null;

                        if (GetObject(name, wrapper, validate, ref @object, ref error) == ReturnCode.Ok)
                        {
                            identifier = @object;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.ObjectType:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                case IdentifierKind.NativeModule:
                    {
#if NATIVE && LIBRARY
                        IModule module = null;

                        if (GetModule(name, wrapper, validate, ref module, ref error) == ReturnCode.Ok)
                        {
                            identifier = module;
                            return ReturnCode.Ok;
                        }
#else
                        error = "not implemented";
#endif
                        break;
                    }
                case IdentifierKind.NativeDelegate:
                    {
#if NATIVE && LIBRARY
                        IDelegate @delegate = null;

                        if (GetDelegate(name, wrapper, validate, ref @delegate, ref error) == ReturnCode.Ok)
                        {
                            identifier = @delegate;
                            return ReturnCode.Ok;
                        }
#else
                        error = "not implemented";
#endif
                        break;
                    }
                case IdentifierKind.Host:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                case IdentifierKind.Alias:
                    {
                        IAlias alias = null;

                        if (GetAlias(name, wrapper, validate, ref alias, ref error) == ReturnCode.Ok)
                        {
                            identifier = alias;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Callback:
                    {
                        ICallback callback = null;

                        if (GetCallback(name, wrapper, validate, ref callback, ref error) == ReturnCode.Ok)
                        {
                            identifier = callback;
                            return ReturnCode.Ok;
                        }
                        break;
                    }
                case IdentifierKind.Resolve:
                    {
                        error = "identifier kind not supported";
                        break;
                    }
                default:
                    {
                        error = "unknown identifier kind";
                        break;
                    }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICallback
        public ReturnCode GetCallback(
            string name,
            bool wrapper, /* IGNORED */
            bool validate,
            ref long token,
            ref ICallback callback,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCallbacks(ref error))
                {
                    if ((callbacks.Count > 0) &&
                        (name != null) &&
                        callbacks.ContainsKey(name))
                    {
                        code = PrivateGetCallback(name, wrapper, validate, ref token, ref callback, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid callbacks name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddCallback(
            ICallback callback,
            IClientData clientData,
            ref long token, /* IGNORED */
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (callback == null)
                {
                    result = "invalid callback";
                    return ReturnCode.Error;
                }

                if (callback.Name == null)
                {
                    result = "invalid callback name";
                    return ReturnCode.Error;
                }

                if (!HasCallbacks(ref result))
                    return ReturnCode.Error;

                if (DoesCallbackExist(callback.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": callback already exists",
                        callback.Name);

                    return ReturnCode.Error;
                }

                callbacks.Add(callback.Name, callback);

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Callback, NotifyFlags.Added,
                    callback, this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveCallback(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasCallbacks(ref result))
                    {
                        long token = 0;
                        ICallback callback = null;

                        if (GetCallback(name, true, true, ref token,
                                ref callback, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(callback))
                            {
                                callbacks.Remove(name);

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Callback, NotifyFlags.Removed,
                                    new ObjectPair(callback, token), this,
                                    clientData, null, null, ref result);
#endif

                                return ReturnCode.Ok;
                            }
                            else
                            {
                                result = "callback is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetCallback(
            string name,
            bool wrapper, /* IGNORED */
            bool validate,
            ref long token, /* IGNORED */
            ref ICallback callback,
            ref Result error
            )
        {
            if (HasCallbacks(ref error))
            {
                callback = callbacks[name];

                if (validate && (callback == null))
                    error = "invalid callback";
                else
                    return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // TODO: Make sure this method needs to be public.
        //       This is marked as internal for now.
        //
        internal ReturnCode GetCallback(
            string name,
            bool wrapper, /* IGNORED */
            bool validate,
            ref ICallback callback
            )
        {
            // CheckDisposed();

            Result error = null;

            return GetCallback(name, wrapper, validate, ref callback, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetCallback(
            string name,
            bool wrapper, /* IGNORED */
            bool validate,
            ref ICallback callback,
            ref Result error
            )
        {
            long token = 0;

            return GetCallback(name, wrapper, validate, ref token, ref callback, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddCallback(
            ICallback callback,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddCallback(callback, clientData, ref token, ref result);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IExecute
        public ReturnCode GetIExecute(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IExecute execute,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasIExecutes(ref error))
                {
                    if ((executes.Count > 0) &&
                        executes.ContainsKey(token))
                    {
                        code = PrivateGetIExecute(token, wrapper, validate, ref name, ref execute, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid execute token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasIExecutes(ref error))
                {
                    if ((executes.Count > 0) &&
                        (name != null) &&
                        executes.ContainsKey(name))
                    {
                        code = PrivateGetIExecute(name, wrapper, validate, ref token, ref execute, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid execute name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ListIExecutes(
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasIExecutes(ref localError))
                {
                    code = executes.ToList(
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddIExecute(
            string name,
            IExecute execute,
            IClientData clientData, /* NOT USED */
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (execute == null)
                {
                    result = "invalid execute";
                    return ReturnCode.Error;
                }

                if (name == null)
                {
                    result = "invalid execute name";
                    return ReturnCode.Error;
                }

                if (!HasIExecutes(ref result))
                    return ReturnCode.Error;

                //
                // WARNING: Currently, if somebody adds an IExecute that shadows a
                //          command or procedure (by name), it will take priority.
                //          This behavior could be prevented here by changing this
                //          check to be DoesAnyExecuteExist instead of DoesExecuteExist;
                //          however, I am not convinced that task should be the
                //          responsibility of this method (i.e. why should a method
                //          that deals with IExecute objects need to know about command
                //          or procedure objects?).  The spirit of this check is basically
                //          "can we physically add the object in question?".
                //
                if (DoesIExecuteExist(name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": execute already exists",
                        name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers._Execute executeWrapper = new _Wrappers._Execute(id, execute);

                executes.Add(name, executeWrapper);
                token = executeWrapper.Token;

#if EXECUTE_CACHE
                if (executeCache != null)
                    executeCache.AddOrUpdate(name, executeWrapper, false);
#endif

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.IExecute, NotifyFlags.Added,
                    new ObjectPair(execute, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RenameIExecute(
            string oldName,
            string newName,
            bool delete,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasIExecutes(ref result))
                    {
                        if (!delete || !String.IsNullOrEmpty(newName))
                        {
                            IExecute oldExecute = null;

                            if (GetIExecute(oldName, true, true, ref oldExecute, ref result) == ReturnCode.Ok)
                            {
                                _Wrappers._Execute oldWrapper = oldExecute as _Wrappers._Execute;

                                if (oldWrapper != null)
                                {
                                    IAlias oldAlias = oldWrapper.execute as IAlias;

                                    if (PreventAliasLoop(oldAlias, newName, ref result) == ReturnCode.Ok)
                                    {
                                        if (IsUniqueCommandName(newName, ref result) == ReturnCode.Ok)
                                        {
                                            /* IGNORED */
                                            executes.Rename(oldName, newName);

#if EXECUTE_CACHE
                                            if (executeCache != null)
                                                executeCache.Rename(oldName, newName, oldWrapper, false);
#endif

#if NOTIFY
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.IExecute, NotifyFlags.Renamed,
                                                new ObjectTriplet(oldExecute, oldName, newName), this,
                                                null, null, null, ref result);
#endif

                                            result = String.Empty;
                                            return ReturnCode.Ok;
                                        }
                                    }
                                }
                                else
                                {
                                    result = String.Format(
                                        "can't rename \"{0}\": invalid execute wrapper",
                                        oldName);
                                }
                            }
                            else
                            {
                                result = String.Format(
                                    "can't rename \"{0}\": execute doesn't exist",
                                    oldName);
                            }
                        }
                        else
                        {
                            return RemoveIExecute(oldName, null, ref result);
                        }
                    }
                }
                else
                {
                    result = String.Format(
                        "can't rename \"{0}\": interpreter is read-only",
                        oldName);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveIExecute(
            long token,
            IClientData clientData, /* NOT USED */
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasIExecutes(ref result))
                    {
                        string name = null;
                        IExecute execute = null;

                        if (GetIExecute(token, true, true, ref name,
                                ref execute, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveIExecute(
                                name, token, execute, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveIExecute(
            string name,
            IClientData clientData /* NOT USED */,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasIExecutes(ref result))
                    {
                        long token = 0;
                        IExecute execute = null;

                        if (GetIExecute(name, true, true, ref token,
                                ref execute, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveIExecute(
                                name, token, execute, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchIExecute(
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                bool useHidden = Engine.HasUseHidden(flags);

                ExecuteWrapperDictionary localExecutes =
                    useHidden ? hiddenExecutes : executes;

                if (localExecutes != null)
                {
                    //
                    // NOTE: *WARNING* Empty command/procedure names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (localExecutes.Count > 0)
                        {
                            string exactName = null;
                            StringList list = new StringList();

                            if (localExecutes.ContainsKey(name))
                            {
                                //
                                // NOTE: Exact execute match.
                                //
                                exactName = name;

                                //
                                // NOTE: Store the exact execute match in the result
                                //       list.
                                //
                                list.Add(name);
                            }
                            else
                            {
                                foreach (KeyValuePair<string, _Wrappers._Execute> pair in localExecutes)
                                {
                                    if (String.Compare(
                                            pair.Key, 0, name, 0, name.Length,
                                            StringOps.SystemStringComparisonType) == 0)
                                    {
                                        //
                                        // NOTE: Was the key valid (this should always succeed).
                                        //
                                        if (pair.Key != null)
                                        {
                                            //
                                            // NOTE: It was a match; however, was it an exact match?
                                            //
                                            if (pair.Key.Length == name.Length)
                                                //
                                                // NOTE: Preserve match, it may differ in case.
                                                //
                                                exactName = pair.Key;

                                            //
                                            // NOTE: Was it an exact match or did we match at least one
                                            //       character in a partial match?
                                            //
                                            if ((pair.Key.Length == name.Length) || (name.Length > 0))
                                                //
                                                // NOTE: Store the exact or partial match in the results
                                                //       dictionary.
                                                //
                                                list.Add(pair.Key);
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: If there was an exact match, just use it.
                            //
                            if (exactName != null)
                            {
                                //
                                // NOTE: No need to change the caller's variable,
                                //       it already contains the correct value.
                                //
                                if (PrivateGetIExecute(
                                        exactName, wrapper, validate, ref token,
                                        ref execute, ref error) == ReturnCode.Ok)
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenIExecute : NotifyType.IExecute,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, exactName, execute, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count == 1)
                            {
                                //
                                // NOTE: Normal case, exactly one command matched.
                                //
                                if (PrivateGetIExecute(
                                        list[0], wrapper, validate, ref token,
                                        ref execute, ref error) == ReturnCode.Ok)
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenIExecute : NotifyType.IExecute,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, list[0], execute, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count > 1)
                            {
                                error = RuntimeOps.BadValue(
                                    "ambiguous", "execute", name,
                                    list, null, null);

                                ambiguous = true;
                            }
                            else
                            {
                                error = RuntimeOps.BadValue(
                                    null, "execute", name,
                                    localExecutes.Keys, null, null);
                            }
                        }
                        else
                        {
                            error = "no executes defined";
                        }
                    }
                    else
                    {
                        error = "invalid execute name";
                    }
                }
                else
                {
                    error = String.Format(
                        "{0}executes not available",
                        useHidden ? "hidden " : String.Empty);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetIExecute(
            long token,
            bool wrapper,
            bool validate,
            ref string name, /* NOT USED */
            ref IExecute execute,
            ref Result error
            )
        {
            if (HasIExecutes(ref error))
            {
                if (wrapper)
                {
                    execute = executes[token];

                    if (validate && (execute == null))
                        error = "invalid execute";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (executes[token] != null)
                    {
                        execute = executes[token].execute;

                        if (validate && (execute == null))
                            error = "invalid execute";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid execute wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            if (HasIExecutes(ref error))
            {
                if (wrapper)
                {
                    execute = executes[name];
                    token = EntityOps.GetToken(execute as IWrapperData);

                    if (validate && (execute == null))
                        error = "invalid execute";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (executes[name] != null)
                    {
                        execute = executes[name].execute;

                        //
                        // WARNING: Cannot grab the token from the IExecute here
                        //          because that interface does not inherit from
                        //          IWrapperData.
                        //
                        token = EntityOps.GetToken(executes[name] as IWrapperData);

                        if (validate && (execute == null))
                            error = "invalid execute";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid execute wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref IExecute execute
            )
        {
            Result error = null;

            return GetIExecute(name, wrapper, validate, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref IExecute execute,
            ref Result error
            )
        {
            long token = 0;

            return GetIExecute(name, wrapper, validate, ref token, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveIExecute(
            string name,
            long token,
            IExecute execute,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid execute name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
#if NATIVE && TCL
                if ((execute == null) ||
                    (DisposeTclBridges(execute, false, ref result) == ReturnCode.Ok))
#endif
                {
                    executes.Remove(name);

#if EXECUTE_CACHE
                    if (executeCache != null)
                        executeCache.Remove(name, false);
#endif

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.IExecute, NotifyFlags.Removed,
                        new ObjectTriplet(execute, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

#if NATIVE && TCL
                return ReturnCode.Error;
#endif
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating execute: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IObject
        public ReturnCode GetObject(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IObject @object,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasObjects(ref error))
                {
                    if ((objects.Count > 0) &&
                        objects.ContainsKey(token))
                    {
                        if (wrapper)
                        {
                            @object = objects[token];
                            name = EntityOps.GetName(@object as IIdentifier);

                            if (!validate || (@object != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (objects[token] != null)
                            {
                                @object = objects[token].@object;
                                name = EntityOps.GetName(@object as IIdentifier);

                                if (!validate || (@object != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid object token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetObject(
            string name,
            bool wrapper,
            bool validate,
            ref IObject @object
            )
        {
            CheckDisposed();

            Result error = null;

            return GetObject(name, wrapper, validate, ref @object, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetObject(
            string name,
            bool wrapper,
            bool validate,
            ref IObject @object,
            ref Result error
            )
        {
            CheckDisposed();

            long token = 0;

            return GetObject(name, wrapper, validate, ref token, ref @object, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetObject(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IObject @object,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasObjects(ref error))
                {
                    if ((objects.Count > 0) &&
                        (name != null) &&
                        objects.ContainsKey(name))
                    {
                        if (wrapper)
                        {
                            @object = objects[name];
                            token = EntityOps.GetToken(@object as IWrapperData);

                            if (!validate || (@object != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (objects[name] != null)
                            {
                                @object = objects[name].@object;
                                token = EntityOps.GetToken(@object as IWrapperData);

                                if (!validate || (@object != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid object name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetObject(
            object value,
            bool wrapper,
            bool validate,
            ref string name,
            ref long token,
            ref IObject @object,
            ref Result error
            )
        {
            CheckDisposed();

            return GetObject(
                value, ObjectFlags.None, ObjectFlags.None, false, false,
                wrapper, validate, ref name, ref token, ref @object,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddObject(
            string name,
            Type type,
            ObjectFlags flags,
            IClientData clientData,
            int referenceCount,
            string interpName,
            object value,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            return AddObject(new _Objects.Default(new ObjectData(
                name, null, null, clientData, type, flags, referenceCount,
                interpName, 0), value, null), null, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddObject(
            IObject @object,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (@object == null)
                {
                    result = "invalid object";
                    return ReturnCode.Error;
                }

                if (@object.Name == null)
                {
                    result = "invalid object name";
                    return ReturnCode.Error;
                }

                if (!HasObjects(ref result))
                    return ReturnCode.Error;

                if (DoesObjectExist(@object.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": object already exists",
                        @object.Name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers._Object objectWrapper = new _Wrappers._Object(id, @object);

                objects.Add(@object.Name, objectWrapper);
                token = objectWrapper.Token;
                @object.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Object, NotifyFlags.Added,
                    new ObjectPair(@object, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveObject(
            long token,
            IClientData clientData,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjects(ref result))
                    {
                        string name = null;
                        IObject @object = null;

                        if (GetObject(token, false, true, ref name,
                                ref @object, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveObject(
                                name, token, @object, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveObject(
            string name,
            IClientData clientData,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjects(ref result))
                    {
                        long token = 0;
                        IObject @object = null;

                        if (GetObject(name, false, true, ref token,
                                ref @object, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveObject(
                                name, token, @object, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode GetObject(
            object value,
            ObjectFlags hasFlags,
            ObjectFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            bool wrapper,
            bool validate,
            ref string name
            )
        {
            IObject @object = null;

            return GetObject(
                value, hasFlags, notHasFlags, hasAll, notHasAll,
                wrapper, validate, ref name, ref @object);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetObject(
            object value,
            ObjectFlags hasFlags,
            ObjectFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            bool wrapper,
            bool validate,
            ref string name,
            ref IObject @object
            )
        {
            Result error = null;

            return GetObject(
                value, hasFlags, notHasFlags, hasAll, notHasAll,
                wrapper, validate, ref name, ref @object, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetObject(
            object value,
            ObjectFlags hasFlags,
            ObjectFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            bool wrapper,
            bool validate,
            ref string name,
            ref IObject @object,
            ref Result error
            )
        {
            long token = 0;

            return GetObject(
                value, hasFlags, notHasFlags, hasAll, notHasAll, wrapper,
                validate, ref name, ref token, ref @object, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetObject( /* FOR USE BY [sql] ONLY */
            object value,
            bool wrapper,
            bool validate
            )
        {
            string name = null;

            return GetObject(value, wrapper, validate, ref name);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetObject( /* FOR USE BY MarshalOps.WrapHandle ONLY */
            object value,
            bool wrapper,
            bool validate,
            ref string name
            )
        {
            long token = 0;
            IObject @object = null;
            Result error = null;

            return GetObject(
                value, ObjectFlags.None, ObjectFlags.None, false, false,
                wrapper, validate, ref name, ref token, ref @object,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetObject(
            object value,
            ObjectFlags hasFlags,
            ObjectFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            bool wrapper,
            bool validate,
            ref string name,
            ref long token,
            ref IObject @object,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasObjects(ref error))
                {
                    if (objects.Count > 0)
                    {
                        foreach (KeyValuePair<string, _Wrappers._Object> pair in objects)
                        {
                            if (pair.Value != null)
                            {
                                if (Object.ReferenceEquals(pair.Value.Value, value))
                                {
                                    if (((hasFlags == ObjectFlags.None) ||
                                            FlagOps.HasFlags(pair.Value.Flags, hasFlags, hasAll)) &&
                                        ((notHasFlags == ObjectFlags.None) ||
                                            !FlagOps.HasFlags(pair.Value.Flags, notHasFlags, notHasAll)))
                                    {
                                        name = pair.Key;

                                        if (wrapper)
                                            @object = pair.Value;
                                        else
                                            @object = pair.Value.@object;

                                        token = pair.Value.Token;

                                        if (!validate || (@object != null))
                                            code = ReturnCode.Ok;

                                        break;
                                    }
                                }
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                    {
                        //
                        // HACK: Prevent ToString method from throwing any
                        //       exceptions here (e.g. proxy).
                        //
                        try
                        {
                            if (!RuntimeOps.IsTransparentProxy(value))
                                error = String.Format(
                                    "invalid object value \"{0}\"",
                                    value);
                            else
                                error = "invalid object value";
                        }
                        catch (Exception e)
                        {
                            error = String.Format(
                                "invalid object value: {0}",
                                e);
                        }
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddObject(
            string name,
            Type type,
            ObjectFlags flags,
            IClientData clientData,
            int referenceCount,
            string interpName,
            object value,
            ref Result result
            )
        {
            long token = 0;

            return AddObject(
                name, type, flags, clientData, referenceCount, interpName,
                value, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveObject(
            string name,
            long token,
            IObject @object,
            IClientData clientData,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid object name";
                return ReturnCode.Error;
            }

            if (dispose)
            {
                if (@object != null)
                {
                    if (!FlagOps.HasFlags(@object.Flags, ObjectFlags.Locked, true))
                    {
#if NATIVE && TCL
                        //
                        // NOTE: Remove the Tcl command for this object, if it exists.
                        //
                        string bridgeName = FormatOps.TclBridgeName(
                            @object.InterpName, @object.Name);

                        if ((DoesTclBridgeExist(bridgeName) != ReturnCode.Ok) ||
                            (RemoveTclBridge(@object.InterpName, @object.Name,
                                    null, ref result) == ReturnCode.Ok))
#endif
                        {
                            if (@object.Value != null)
                            {
                                if (!Object.ReferenceEquals(@object.Value, this) &&
                                    !IsMasterOrSlaveInterpreter(@object.Value) &&
                                    !FlagOps.HasFlags(@object.Flags, ObjectFlags.NoDispose, true))
                                {
                                    GlobalState.PushActiveInterpreter(this);

                                    try
                                    {
                                        if (RuntimeOps.TryDisposeObject(@object.Value,
                                                ref dispose, ref result) == ReturnCode.Ok)
                                        {
#if NOTIFY
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Object, NotifyFlags.Disposed,
                                                new ObjectTriplet(@object.Value, name, token), this,
                                                clientData, null, null, ref result);
#endif

                                            @object.Value = null;
                                            @object = null;

                                            if (synchronous)
                                                CommonOps.CollectGarbage(null, null); /* throw */
                                        }
                                        else
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    catch (Exception e)
                                    {
                                        result = String.Format(
                                            "caught exception while disposing object: {0}",
                                            e);

                                        return ReturnCode.Error;
                                    }
                                    finally
                                    {
                                        /* IGNORED */
                                        GlobalState.PopActiveInterpreter();
                                    }
                                }
                                else
                                {
                                    //
                                    // NOTE: We did not dispose the object because
                                    //       it was prohibited.
                                    //
                                    dispose = false;
                                }
                            }
                            else
                            {
                                //
                                // NOTE: We did not dispose the object because
                                //       it was invalid.
                                //
                                dispose = false;
                            }
                        }
#if NATIVE && TCL
                        else
                        {
                            //
                            // NOTE: Failed to remove the Tcl bridge.
                            //
                            return ReturnCode.Error;
                        }
#endif
                    }
                    else
                    {
                        result = "cannot dispose or remove a locked object";

                        return ReturnCode.Error;
                    }
                }
                else
                {
                    //
                    // NOTE: We did not dispose the object because the wrapper
                    //       was invalid.
                    //
                    dispose = false;
                }
            }

            //
            // NOTE: Remove the command alias for this object, if it exists.
            //       We cannot use the token here because it is the object
            //       token, not the command (alias) token.
            //
            if ((DoesAliasExist(name) == ReturnCode.Ok) &&
                (RemoveAliasAndCommand(name, clientData, ref result) != ReturnCode.Ok))
            {
                return ReturnCode.Error;
            }

            objects.Remove(name);

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Object, NotifyFlags.Removed,
                new ObjectTriplet(@object, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode RemoveObject(
            long token,
            IClientData clientData,
            bool synchronous,
            ref Result result
            )
        {
            bool dispose = true;

            return RemoveObject(token, clientData, synchronous, ref dispose, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveObject(
            string name,
            IClientData clientData,
            bool synchronous,
            ref Result result
            )
        {
            bool dispose = true;

            return RemoveObject(name, clientData, synchronous, ref dispose, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveObjects(
            IClientData clientData,
            bool stopOnError,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjects(ref result))
                    {
                        StringList keys = new StringList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the objects collection in the containing interpreter.
                        //
                        foreach (KeyValuePair<string, _Wrappers._Object> pair in objects)
                            if (pair.Value != null)
                                if (pair.Value.ReferenceCount == 0)
                                    if (!FlagOps.HasFlags(pair.Value.Flags, ObjectFlags.Locked, true))
                                        keys.Add(pair.Key);

                        int disposed = 0;

                        //
                        // NOTE: Cleanup, PASS #2, dispose and remove all the objects we found
                        //       from PASS #1 (above).
                        //
                        foreach (string key in keys)
                        {
                            bool localDispose = dispose;

                            if ((RemoveObject(key, clientData, synchronous,
                                    ref localDispose, ref result) != ReturnCode.Ok) && stopOnError)
                            {
                                return ReturnCode.Error;
                            }

                            //
                            // NOTE: At least one object has been disposed.
                            //
                            if (localDispose)
                                disposed++;
                        }

                        //
                        // NOTE: Inform the caller if any of the objects were disposed.
                        //
                        dispose = (disposed > 0);

                        //
                        // NOTE: Return the number of objects we disposed and removed.
                        //
                        result = StringList.MakeList(
                            "disposed", disposed, "removed", keys.Count);

                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPackage
        public ReturnCode GetPackage(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IPackage package,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPackages(ref error))
                {
                    if ((packages.Count > 0) &&
                        packages.ContainsKey(token))
                    {
                        if (wrapper)
                        {
                            package = packages[token];
                            name = EntityOps.GetName(package as IIdentifier);

                            if (!validate || (package != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (packages[token] != null)
                            {
                                package = packages[token].package;
                                name = EntityOps.GetName(package as IIdentifier);

                                if (!validate || (package != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid package token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetPackage(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IPackage package,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPackages(ref error))
                {
                    if ((packages.Count > 0) &&
                        (name != null) &&
                        packages.ContainsKey(name))
                    {
                        if (wrapper)
                        {
                            package = packages[name];
                            token = EntityOps.GetToken(package as IWrapperData);

                            if (!validate || (package != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (packages[name] != null)
                            {
                                package = packages[name].package;
                                token = EntityOps.GetToken(package as IWrapperData);

                                if (!validate || (package != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid package name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddPackage(
            IPackage package,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (package == null)
                {
                    result = "invalid package";
                    return ReturnCode.Error;
                }

                if (package.Name == null)
                {
                    result = "invalid package name";
                    return ReturnCode.Error;
                }

                if (!HasPackages(ref result))
                    return ReturnCode.Error;

                if (DoesPackageExist(package.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": package already exists",
                        package.Name);

                    return ReturnCode.Error;
                }

                ReturnCode code = ReturnCode.Ok;

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    try
                    {
                        code = package.Initialize(this,
                            (clientData != null) ? clientData : package.ClientData,
                            ref result);
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while initializing package: {0}",
                            e);

                        code = ReturnCode.Error;
                    }
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        ReturnCode terminateCode;
                        Result terminateResult = null;

                        try
                        {
                            terminateCode = package.Terminate(this,
                                (clientData != null) ? clientData : package.ClientData,
                                ref terminateResult);
                        }
                        catch (Exception e)
                        {
                            terminateResult = String.Format(
                                "caught exception while terminating package: {0}",
                                e);

                            terminateCode = ReturnCode.Error;
                        }

                        if (terminateCode != ReturnCode.Ok)
                            DebugOps.Complain(this, terminateCode, terminateResult);
                    }
                }

                if (code != ReturnCode.Ok)
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                    return code;
                }

                long id = NextId();
                _Wrappers.Package packageWrapper = new _Wrappers.Package(id, package);

                packages.Add(package.Name, packageWrapper);
                token = packageWrapper.Token;
                package.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Package, NotifyFlags.Added,
                    new ObjectPair(package, token), this,
                    clientData, null, null, ref result);
#endif

                /* IGNORED */
                GlobalState.PopActiveInterpreter();
                return code;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePackage(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPackages(ref result))
                    {
                        string name = null;
                        IPackage package = null;

                        if (GetPackage(token, true, true, ref name,
                                ref package, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemovePackage(
                                name, token, package, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePackage(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPackages(ref result))
                    {
                        long token = 0;
                        IPackage package = null;

                        if (GetPackage(name, true, true, ref token,
                                ref package, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemovePackage(
                                name, token, package, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode GetPackage(
            string name,
            bool wrapper,
            bool validate,
            ref IPackage package
            )
        {
            Result error = null;

            return GetPackage(name, wrapper, validate, ref package, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetPackage(
            string name,
            bool wrapper,
            bool validate,
            ref IPackage package,
            ref Result error
            )
        {
            long token = 0;

            return GetPackage(name, wrapper, validate, ref token, ref package, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddPackage(
            IPackage package,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddPackage(package, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemovePackage(
            string name,
            long token,
            IPackage package,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid package name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
                if ((package == null) ||
                    (package.Terminate(this, (clientData != null) ?
                            clientData : package.ClientData, ref result) == ReturnCode.Ok))
                {
                    packages.Remove(name);

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Package, NotifyFlags.Removed,
                        new ObjectTriplet(package, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

                return ReturnCode.Error;
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating package: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPlugin
        public ReturnCode GetPlugin(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IPlugin plugin,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPlugins(ref error))
                {
                    if ((plugins.Count > 0) &&
                        plugins.ContainsKey(token))
                    {
                        if (wrapper)
                        {
                            plugin = plugins[token];
                            name = EntityOps.GetName(plugin as IIdentifier);

                            if (!validate || (plugin != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (plugins[token] != null)
                            {
                                plugin = plugins[token].plugin;
                                name = EntityOps.GetName(plugin as IIdentifier);

                                if (!validate || (plugin != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid plugin token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetPlugin(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IPlugin plugin,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPlugins(ref error))
                {
                    if ((plugins.Count > 0) &&
                        (name != null) &&
                        plugins.ContainsKey(name))
                    {
                        if (wrapper)
                        {
                            plugin = plugins[name];
                            token = EntityOps.GetToken(plugin as IWrapperData);

                            if (!validate || (plugin != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (plugins[name] != null)
                            {
                                plugin = plugins[name].plugin;
                                token = EntityOps.GetToken(plugin as IWrapperData);

                                if (!validate || (plugin != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid plugin name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetPlugin(
            Assembly assembly,
            bool wrapper,
            bool validate,
            ref string name,
            ref long token,
            ref IPlugin plugin,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPlugins(ref error))
                {
                    if (plugins.Count > 0)
                    {
                        foreach (KeyValuePair<string, _Wrappers.Plugin> pair in plugins)
                        {
                            if (pair.Value != null)
                            {
                                if (Object.ReferenceEquals(pair.Value.Assembly, assembly))
                                {
                                    name = pair.Key;

                                    if (wrapper)
                                        plugin = pair.Value;
                                    else
                                        plugin = pair.Value.plugin;

                                    token = pair.Value.Token;

                                    if (!validate || (plugin != null))
                                        code = ReturnCode.Ok;

                                    break;
                                }
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid plugin assembly \"{0}\"",
                            FormatOps.AssemblyName(assembly, 0));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchPlugin(
            string name,
            bool wrapper,
            bool validate,
            ref IPlugin plugin,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPlugins(ref error))
                {
                    if (plugins.Count > 0)
                    {
                        foreach (KeyValuePair<string, _Wrappers.Plugin> pair in plugins)
                        {
                            if (pair.Value != null)
                            {
                                if (StringOps.Match(
                                        MatchMode.Glob, pair.Key, name, false))
                                {
                                    if (wrapper)
                                        plugin = pair.Value;
                                    else
                                        plugin = pair.Value.plugin;

                                    if (!validate || (plugin != null))
                                        code = ReturnCode.Ok;

                                    break;
                                }
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid plugin name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddPlugin(
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (plugin == null)
                {
                    result = "invalid plugin";
                    return ReturnCode.Error;
                }

                if (plugin.Name == null)
                {
                    result = "invalid plugin name";
                    return ReturnCode.Error;
                }

                if (!HasPlugins(ref result))
                    return ReturnCode.Error;

                if (DoesPluginExist(plugin.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": plugin already exists",
                        plugin.Name);

                    return ReturnCode.Error;
                }

                ReturnCode code = ReturnCode.Ok;

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    try
                    {
                        code = plugin.Initialize(this,
                            (clientData != null) ? clientData : plugin.ClientData,
                            ref result);
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while initializing plugin: {0}",
                            e);

                        code = ReturnCode.Error;
                    }
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        ReturnCode terminateCode;
                        Result terminateResult = null;

                        try
                        {
                            terminateCode = plugin.Terminate(this,
                                (clientData != null) ? clientData : plugin.ClientData,
                                ref terminateResult);
                        }
                        catch (Exception e)
                        {
                            terminateResult = String.Format(
                                "caught exception while terminating plugin: {0}",
                                e);

                            terminateCode = ReturnCode.Error;
                        }

                        if (terminateCode != ReturnCode.Ok)
                            DebugOps.Complain(this, terminateCode, terminateResult);
                    }
                }

                if (code != ReturnCode.Ok)
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                    return code;
                }

                long id = NextId();
                _Wrappers.Plugin pluginWrapper = new _Wrappers.Plugin(id, plugin);

                plugins.Add(plugin.Name, pluginWrapper);
                token = pluginWrapper.Token;
                plugin.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Plugin, NotifyFlags.Added,
                    new ObjectPair(plugin, token), this,
                    clientData, null, null, ref result);
#endif

                /* IGNORED */
                GlobalState.PopActiveInterpreter();
                return code;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePlugin(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPlugins(ref result))
                    {
                        string name = null;
                        IPlugin plugin = null;

                        if (GetPlugin(token, true, true, ref name,
                                ref plugin, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemovePlugin(
                                name, token, plugin, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePlugin(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPlugins(ref result))
                    {
                        long token = 0;
                        IPlugin plugin = null;

                        if (GetPlugin(name, true, true, ref token,
                                ref plugin, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemovePlugin(
                                name, token, plugin, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode GetPlugin(
            long token,
            bool wrapper,
            bool validate,
            ref IPlugin plugin,
            ref Result error
            )
        {
            string name = null;

            return GetPlugin(token, wrapper, validate, ref name, ref plugin, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetPlugin(
            string name,
            bool wrapper,
            bool validate,
            ref IPlugin plugin,
            ref Result error
            )
        {
            long token = 0;

            return GetPlugin(name, wrapper, validate, ref token, ref plugin, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemovePlugin(
            string name,
            long token,
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid plugin name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
                if ((plugin == null) ||
                    (plugin.Terminate(this, (clientData != null) ?
                            clientData : plugin.ClientData, ref result) == ReturnCode.Ok))
                {
                    plugins.Remove(name);

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Plugin, NotifyFlags.Removed,
                        new ObjectTriplet(plugin, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

                return ReturnCode.Error;
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating plugin: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICommand
        public ReturnCode GetCommand(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ICommand command,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCommands(ref error))
                {
                    if ((commands.Count > 0) &&
                        commands.ContainsKey(token))
                    {
                        code = PrivateGetCommand(token, wrapper, validate, ref name, ref command, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid command token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetCommand(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCommands(ref error))
                {
                    if ((commands.Count > 0) &&
                        (name != null) &&
                        commands.ContainsKey(name))
                    {
                        code = PrivateGetCommand(name, wrapper, validate, ref token, ref command, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid command name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ListCommands(
            CommandFlags hasFlags,
            CommandFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasCommands(ref localError))
                {
                    code = commands.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddExecuteCallback(
            string name,
            ExecuteCallback callback,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            //
            // NOTE: Synthesize a command out of the clear blue sky and
            //       populate it with the callback supplied by the user.  We do
            //       not bother checking if the supplied callback is valid
            //       because we do not actually care.  The worst case is that
            //       the engine will call the Execute method of the default
            //       command created here, which does nothing.
            //
            ICommand command = new _Commands.Default(
                new CommandData(name, null, null, clientData,
                    typeof(_Commands.Default).FullName,
                    CommandFlags.None, null, 0));

            command.Callback = callback;

            return AddCommand(command, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddCommand(
            ICommand command,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (command == null)
                {
                    result = "invalid command";
                    return ReturnCode.Error;
                }

                if (command.Name == null)
                {
                    result = "invalid command name";
                    return ReturnCode.Error;
                }

                if (!HasCommands(ref result))
                    return ReturnCode.Error;

                //
                // WARNING: Currently, if somebody adds an ICommand that shadows an
                //          IExecute or procedure (by name), it will basically be
                //          ignored in favor of the IExecute or procedure, which
                //          have higher priority.  This behavior could be prevented
                //          here by changing this check to be DoesAnyExecuteExist
                //          instead of DoesCommandExist; however, I am not convinced
                //          that task should be the responsibility of this method
                //          (i.e. why should a method that deals with commands need
                //          to know about IExecute or procedure objects?).  The spirit
                //          of this check is basically "can we physically add the
                //          object in question?".
                //
                if (DoesCommandExist(command.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": command already exists",
                        command.Name);

                    return ReturnCode.Error;
                }

                ReturnCode code = ReturnCode.Ok;

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    try
                    {
                        code = command.Initialize(this,
                            (clientData != null) ? clientData : command.ClientData,
                            ref result);
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while initializing command: {0}",
                            e);

                        code = ReturnCode.Error;
                    }
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        ReturnCode terminateCode;
                        Result terminateResult = null;

                        try
                        {
                            terminateCode = command.Terminate(this,
                                (clientData != null) ? clientData : command.ClientData,
                                ref terminateResult);
                        }
                        catch (Exception e)
                        {
                            terminateResult = String.Format(
                                "caught exception while terminating command: {0}",
                                e);

                            terminateCode = ReturnCode.Error;
                        }

                        if (terminateCode != ReturnCode.Ok)
                            DebugOps.Complain(this, terminateCode, terminateResult);
                    }
                }

                if (code != ReturnCode.Ok)
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                    return code;
                }

                long id = NextId();
                _Wrappers.Command commandWrapper = new _Wrappers.Command(id, command);
                long newToken = commandWrapper.Token;

                //
                // NOTE: Are we allowed to assist in managing the command
                //       token via the associated plugin?
                //
                if (!EntityOps.IsNoToken(command))
                {
                    //
                    // NOTE: Store the command token so that we can remove
                    //       them later without using their names (which
                    //       can change).
                    //
                    IPlugin plugin = command.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.CommandTokens;

                        if ((tokens != null) && !tokens.Contains(newToken))
                        {
                            tokens.Add(newToken);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the command tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.CommandTokens = tokens;
#endif
                        }
                    }
                }

                command.Token = newToken;
                commands.Add(command.Name, commandWrapper);
                token = newToken;

#if EXECUTE_CACHE
                if (executeCache != null)
                    executeCache.AddOrUpdate(command.Name, commandWrapper, false);
#endif

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Command, NotifyFlags.Added,
                    new ObjectPair(command, token), this,
                    clientData, null, null, ref result);
#endif

                /* IGNORED */
                GlobalState.PopActiveInterpreter();
                return code;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RenameCommand(
            string oldName,
            string newName,
            bool delete,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasCommands(ref result))
                    {
                        if (!delete || !String.IsNullOrEmpty(newName))
                        {
                            ICommand oldCommand = null;

                            if (GetCommand(oldName, true, true, ref oldCommand, ref result) == ReturnCode.Ok)
                            {
                                if (!EntityOps.IsReadOnly(oldCommand))
                                {
                                    _Wrappers.Command oldWrapper = oldCommand as _Wrappers.Command;

                                    if (oldWrapper != null)
                                    {
                                        IAlias oldAlias = oldWrapper.command as IAlias;

                                        if (PreventAliasLoop(oldAlias, newName, ref result) == ReturnCode.Ok)
                                        {
                                            if (IsUniqueCommandName(newName, ref result) == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: New name, same token.
                                                //
                                                oldWrapper.Name = newName;

                                                /* IGNORED */
                                                commands.Rename(oldName, newName);

#if EXECUTE_CACHE
                                                if (executeCache != null)
                                                    executeCache.Rename(oldName, newName, oldWrapper, false);
#endif

#if NOTIFY
                                                /* IGNORED */
                                                CheckNotifications(
                                                    NotifyType.Command, NotifyFlags.Renamed,
                                                    new ObjectTriplet(oldCommand, oldName, newName), this,
                                                    null, null, null, ref result);
#endif

                                                result = String.Empty;
                                                return ReturnCode.Ok;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        result = String.Format(
                                            "can't rename \"{0}\": invalid command wrapper",
                                            oldName);
                                    }
                                }
                                else
                                {
                                    result = String.Format(
                                        "can't rename \"{0}\": command is read-only",
                                        oldName);
                                }
                            }
                            else
                            {
                                result = String.Format(
                                    "can't rename \"{0}\": command doesn't exist",
                                    oldName);
                            }
                        }
                        else
                        {
                            return RemoveCommand(oldName, null, ref result);
                        }
                    }
                }
                else
                {
                    result = String.Format(
                        "can't rename \"{0}\": interpreter is read-only",
                        oldName);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveCommand(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasCommands(ref result))
                    {
                        string name = null;
                        ICommand command = null;

                        if (GetCommand(token, true, true, ref name,
                                ref command, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(command))
                            {
                                return PrivateRemoveCommand(
                                    name, token, command, clientData, ref result);
                            }
                            else
                            {
                                result = "command is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveCommand(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasCommands(ref result))
                    {
                        long token = 0;
                        ICommand command = null;

                        if (GetCommand(name, true, true, ref token,
                                ref command, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(command))
                            {
                                return PrivateRemoveCommand(
                                    name, token, command, clientData, ref result);
                            }
                            else
                            {
                                result = "command is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchCommand(
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                bool useHidden = Engine.HasUseHidden(flags);

                CommandWrapperDictionary localCommands =
                    useHidden ? hiddenCommands : commands;

                if (localCommands != null)
                {
                    //
                    // NOTE: *WARNING* Empty command/procedure names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (localCommands.Count > 0)
                        {
                            string exactName = null;
                            StringList list = new StringList();

                            if (localCommands.ContainsKey(name) &&
                                (Engine.HasMatchHidden(flags) ||
                                !EntityOps.IsHidden(localCommands[name])))
                            {
                                //
                                // NOTE: Exact command match.
                                //
                                exactName = name;

                                //
                                // NOTE: Store the exact command match in the result
                                //       list.
                                //
                                list.Add(name);
                            }
                            else
                            {
                                foreach (KeyValuePair<string, _Wrappers.Command> pair in localCommands)
                                {
                                    if (String.Compare(
                                            pair.Key, 0, name, 0, name.Length,
                                            StringOps.SystemStringComparisonType) == 0)
                                    {
                                        //
                                        // NOTE: Is the command visible or are we allowed to
                                        //       find hidden command?
                                        //
                                        if (Engine.HasMatchHidden(flags) || !EntityOps.IsHidden(pair.Value))
                                        {
                                            //
                                            // NOTE: Was the key valid (this should always succeed).
                                            //
                                            if (pair.Key != null)
                                            {
                                                //
                                                // NOTE: It was a match; however, was it an exact match?
                                                //
                                                if (pair.Key.Length == name.Length)
                                                    //
                                                    // NOTE: Preserve match, it may differ in case.
                                                    //
                                                    exactName = pair.Key;

                                                //
                                                // NOTE: Was it an exact match or did we match at least one
                                                //       character in a partial match?
                                                //
                                                if ((pair.Key.Length == name.Length) || (name.Length > 0))
                                                    //
                                                    // NOTE: Store the exact or partial match in the results
                                                    //       dictionary.
                                                    //
                                                    list.Add(pair.Key);
                                            }
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: If there was an exact match, just use it.
                            //
                            if (exactName != null)
                            {
                                //
                                // NOTE: No need to change the caller's variable,
                                //       it already contains the correct value.
                                //
                                if ((useHidden && PrivateGetHiddenCommand(
                                        exactName, wrapper, validate, ref token,
                                        ref command, ref error) == ReturnCode.Ok) ||
                                    (!useHidden && PrivateGetCommand(
                                        exactName, wrapper, validate, ref token,
                                        ref command, ref error) == ReturnCode.Ok))
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenCommand : NotifyType.Command,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, exactName, command, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count == 1)
                            {
                                //
                                // NOTE: Normal case, exactly one command matched.
                                //
                                if ((useHidden && PrivateGetHiddenCommand(
                                        list[0], wrapper, validate, ref token,
                                        ref command, ref error) == ReturnCode.Ok) ||
                                    (!useHidden && PrivateGetCommand(
                                        list[0], wrapper, validate, ref token,
                                        ref command, ref error) == ReturnCode.Ok))
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenCommand : NotifyType.Command,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, list[0], command, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count > 1)
                            {
                                error = RuntimeOps.BadValue(
                                    "ambiguous", "command", name,
                                    list, null, null);

                                ambiguous = true;
                            }
                            else
                            {
                                error = RuntimeOps.BadValue(
                                    null, "command", name,
                                    localCommands.Keys, null, null);
                            }
                        }
                        else
                        {
                            error = "no commands defined";
                        }
                    }
                    else
                    {
                        error = "invalid command name";
                    }
                }
                else
                {
                    error = String.Format(
                        "{0}commands not available",
                        useHidden ? "hidden " : String.Empty);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetCommand(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ICommand command,
            ref Result error
            )
        {
            if (HasCommands(ref error))
            {
                if (wrapper)
                {
                    command = commands[token];
                    name = EntityOps.GetName(command as IIdentifier);

                    if (validate && (command == null))
                        error = "invalid command";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (commands[token] != null)
                    {
                        command = commands[token].command;
                        name = EntityOps.GetName(command as IIdentifier);

                        if (validate && (command == null))
                            error = "invalid command";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid command wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetCommand(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            if (HasCommands(ref error))
            {
                if (wrapper)
                {
                    command = commands[name];
                    token = EntityOps.GetToken(command as IWrapperData);

                    if (validate && (command == null))
                        error = "invalid command";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (commands[name] != null)
                    {
                        command = commands[name].command;
                        token = EntityOps.GetToken(command as IWrapperData);

                        if (validate && (command == null))
                            error = "invalid command";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid command wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetCommand(
            long token,
            bool wrapper,
            bool validate,
            ref ICommand command
            )
        {
            Result error = null;

            return GetCommand(token, wrapper, validate, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetCommand(
            long token,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            string name = null;

            return GetCommand(token, wrapper, validate, ref name, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetCommand(
            string name,
            bool wrapper,
            bool validate,
            ref ICommand command
            )
        {
            Result error = null;

            return GetCommand(name, wrapper, validate, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetCommand(
            string name,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            long token = 0;

            return GetCommand(name, wrapper, validate, ref token, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetCommand(
            Type type,
            bool wrapper,
            bool validate,
            ref ICommand command
            )
        {
            Result error = null;

            return GetCommand(type, wrapper, validate, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetCommand(
            Type type,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            string name = null;

            return GetCommand(type, wrapper, validate, ref name, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetCommand(
            Type type,
            bool wrapper,
            bool validate,
            ref string name,
            ref ICommand command,
            ref Result error
            )
        {
            long token = 0;

            return GetCommand(type, wrapper, validate, ref name, ref token, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetCommand(
            Type type,
            bool wrapper,
            bool validate,
            ref string name,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCommands(ref error))
                {
                    if ((commands.Count > 0) && (type != null))
                    {
                        foreach (KeyValuePair<string, _Wrappers.Command> pair in commands)
                        {
                            if (pair.Value != null)
                            {
                                if (type.IsInstanceOfType(pair.Value.command))
                                {
                                    name = pair.Key;

                                    if (wrapper)
                                        command = pair.Value;
                                    else
                                        command = pair.Value.command;

                                    token = pair.Value.Token;

                                    if (!validate || (command != null))
                                        code = ReturnCode.Ok;

                                    break;
                                }
                            }
                        }

                        foreach (KeyValuePair<string, _Wrappers.Command> pair in hiddenCommands)
                        {
                            if (pair.Value != null)
                            {
                                if (type.IsInstanceOfType(pair.Value.command))
                                {
                                    name = pair.Key;

                                    if (wrapper)
                                        command = pair.Value;
                                    else
                                        command = pair.Value.command;

                                    token = pair.Value.Token;

                                    if (!validate || (command != null))
                                        code = ReturnCode.Ok;

                                    break;
                                }
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid command type \"{0}\"",
                            FormatOps.TypeName(type));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode IsUniqueCommandName(
            string newName,
            ref Result error
            )
        {
            IExecute execute = null;

            if (GetIExecute(newName, true, false, ref execute, ref error) == ReturnCode.Ok)
            {
                error = String.Format(
                    "can't rename to \"{0}\": execute already exists",
                    newName);

                return ReturnCode.Error;
            }

            IProcedure procedure = null;

            if (GetProcedure(newName, true, false, ref procedure, ref error) == ReturnCode.Ok)
            {
                error = String.Format(
                    "can't rename to \"{0}\": procedure already exists",
                    newName);

                return ReturnCode.Error;
            }

            ICommand command = null;

            if (GetCommand(newName, true, false, ref command, ref error) == ReturnCode.Ok)
            {
                error = String.Format(
                    "can't rename to \"{0}\": command already exists",
                    newName);

                return ReturnCode.Error;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ListHiddenCommands(
            CommandFlags hasFlags,
            CommandFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            // CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasHiddenCommands(ref localError))
                {
                    code = hiddenCommands.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddCommand(
            ICommand command,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddCommand(command, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveCommand(
            string name,
            long token,
            ICommand command,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid command name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
                if (
#if NATIVE && TCL
                    ((command == null) ||
                    (DisposeTclBridges(command, false, ref result) == ReturnCode.Ok)) &&
#endif
                    ((command == null) ||
                    (command.Terminate(this, (clientData != null) ?
                            clientData : command.ClientData, ref result) == ReturnCode.Ok)))
                {
                    /* IGNORED */
                    commands.Remove(name);

#if EXECUTE_CACHE
                    if (executeCache != null)
                        executeCache.Remove(name, false);
#endif

                    if (command != null)
                    {
                        //
                        // NOTE: Are we allowed to assist in managing the command
                        //       token via the associated plugin?
                        //
                        bool noToken = EntityOps.IsNoToken(command);

                        if (!noToken)
                        {
                            //
                            // NOTE: If this command is being managed by a plugin,
                            //       remove the associated token from said plugin.
                            //
                            IPlugin plugin = command.Plugin;

                            if (plugin != null)
                            {
                                LongList tokens = plugin.CommandTokens;

                                if ((tokens != null) && tokens.Contains(token))
                                {
                                    tokens.Remove(token);

#if ISOLATED_PLUGINS
                                    //
                                    // HACK: Force the command tokens for the
                                    //       isolated plugin to be updated.
                                    //
                                    if (RuntimeOps.IsPluginIsolated(plugin))
                                        plugin.CommandTokens = tokens;
#endif
                                }
                            }
                        }
                    }

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Command, NotifyFlags.Removed,
                        new ObjectTriplet(command, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

                return ReturnCode.Error;
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating command: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPolicy
        public ReturnCode GetPolicy(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IPolicy policy,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPolicies(ref error))
                {
                    if ((policies.Count > 0) &&
                        policies.ContainsKey(token))
                    {
                        code = PrivateGetPolicy(token, wrapper, validate, ref name, ref policy, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid policy token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetPolicy(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IPolicy policy,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasPolicies(ref error))
                {
                    if ((policies.Count > 0) &&
                        (name != null) &&
                        policies.ContainsKey(name))
                    {
                        code = PrivateGetPolicy(name, wrapper, validate, ref token, ref policy, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid policy name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddPolicy(
            ExecuteCallback callback,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            IPolicy policy = RuntimeOps.NewCorePolicy(
                callback, clientData, plugin, ref result);

            if (policy != null)
            {
                if (AddPolicy(
                        policy, clientData, ref token,
                        ref result) == ReturnCode.Ok)
                {
                    result = policy.Name;
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddPolicy(
            IPolicy policy,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (policy == null)
                {
                    result = "invalid policy";
                    return ReturnCode.Error;
                }

                if (policy.Name == null)
                {
                    result = "invalid policy name";
                    return ReturnCode.Error;
                }

                if (!HasPolicies(ref result))
                    return ReturnCode.Error;

                if (DoesPolicyExist(policy.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": policy already exists",
                        policy.Name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers.Policy policyWrapper = new _Wrappers.Policy(id, policy);
                long newToken = policyWrapper.Token;

                //
                // NOTE: Are we allowed to assist in managing the policy
                //       token via the associated plugin?
                //
                if (!EntityOps.IsNoToken(policy))
                {
                    //
                    // NOTE: Store the policy token so that we can remove
                    //       them later without using their names (which
                    //       can change).
                    //
                    IPlugin plugin = policy.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.PolicyTokens;

                        if ((tokens != null) && !tokens.Contains(newToken))
                        {
                            tokens.Add(newToken);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the policy tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.PolicyTokens = tokens;
#endif
                        }
                    }
                }

                policy.Token = newToken;
                policies.Add(policy.Name, policyWrapper);
                token = newToken;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Policy, NotifyFlags.Added,
                    new ObjectPair(policy, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddScriptPolicy(
            IScriptPolicy scriptPolicy,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            if (scriptPolicy != null)
                return AddPolicy(scriptPolicy.Execute, plugin, clientData, ref token, ref result);
            else
                result = "invalid script policy";

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePolicy(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPolicies(ref result))
                    {
                        string name = null;
                        IPolicy policy = null;

                        if (GetPolicy(token, true, true, ref name,
                                ref policy, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(policy))
                            {
                                return PrivateRemovePolicy(
                                    name, token, policy, clientData, ref result);
                            }
                            else
                            {
                                result = "policy is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePolicy(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasPolicies(ref result))
                    {
                        long token = 0;
                        IPolicy policy = null;

                        if (GetPolicy(name, true, true, ref token,
                                ref policy, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(policy))
                            {
                                return PrivateRemovePolicy(
                                    name, token, policy, clientData, ref result);
                            }
                            else
                            {
                                result = "policy is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetPolicy(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IPolicy policy,
            ref Result error
            )
        {
            if (HasPolicies(ref error))
            {
                if (wrapper)
                {
                    policy = policies[token];
                    name = EntityOps.GetName(policy as IIdentifier);

                    if (validate && (policy == null))
                        error = "invalid policy";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (policies[token] != null)
                    {
                        policy = policies[token].policy;
                        name = EntityOps.GetName(policy as IIdentifier);

                        if (validate && (policy == null))
                            error = "invalid policy";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid policy wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetPolicy(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IPolicy policy,
            ref Result error
            )
        {
            if (HasPolicies(ref error))
            {
                if (wrapper)
                {
                    policy = policies[name];
                    token = EntityOps.GetToken(policy as IWrapperData);

                    if (validate && (policy == null))
                        error = "invalid policy";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (policies[name] != null)
                    {
                        policy = policies[name].policy;
                        token = EntityOps.GetToken(policy as IWrapperData);

                        if (validate && (policy == null))
                            error = "invalid policy";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid policy wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetPolicy(
            string name,
            bool wrapper,
            bool validate,
            ref IPolicy policy,
            ref Result error
            )
        {
            long token = 0;

            return GetPolicy(name, wrapper, validate, ref token, ref policy, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckCommandPolicies(
            PolicyFlags flags,
            IExecute execute,
            ArgumentList arguments,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            return CheckPolicies(
                MethodFlags.CommandPolicy, flags, execute, arguments, null,
                null, null, null, null, null, clientData, ref decision,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckScriptPolicies(
            PolicyFlags flags,
            IScript script,
            Encoding encoding,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            string hashAlgorithm = null; /* NOTE: Means "use default". */
            byte[] hashValue = null;

            if (!FlagOps.HasFlags(flags, PolicyFlags.NoHash, true))
            {
                Result hashError = null;

                hashValue = RuntimeOps.HashScript(
                    hashAlgorithm, script, encoding, ref hashError);

#if DEBUG && VERBOSE
                if (hashValue == null)
                    DebugOps.Complain(this, ReturnCode.Error, hashError);
#endif
            }

            return CheckPolicies(
                MethodFlags.ScriptPolicy, flags, null, null, script, null,
                null, encoding, null, null, clientData, ref decision,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckBeforeFilePolicies(
            PolicyFlags flags,
            string fileName,
            Encoding encoding,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            string hashAlgorithm = null; /* NOTE: Means "use default". */
            byte[] hashValue = null;

            if (!FlagOps.HasFlags(flags, PolicyFlags.NoHash, true))
            {
                Result hashError = null;

                hashValue = RuntimeOps.HashFile(
                    hashAlgorithm, fileName, encoding, ref hashError);

#if DEBUG && VERBOSE
                if (hashValue == null)
                    DebugOps.Complain(this, ReturnCode.Error, hashError);
#endif
            }

            return CheckPolicies(
                MethodFlags.FilePolicy, flags, null, null, null, fileName,
                null, encoding, hashValue, hashAlgorithm, clientData,
                ref decision, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckAfterFilePolicies(
            PolicyFlags flags,
            string fileName,
            string text,
            Encoding encoding,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            string hashAlgorithm = null; /* NOTE: Means "use default". */
            byte[] hashValue = null;

            if (!FlagOps.HasFlags(flags, PolicyFlags.NoHash, true))
            {
                Result hashError = null;

                hashValue = RuntimeOps.HashString(
                    hashAlgorithm, text, encoding, ref hashError);

#if DEBUG && VERBOSE
                if (hashValue == null)
                    DebugOps.Complain(this, ReturnCode.Error, hashError);
#endif
            }

            return CheckPolicies(
                MethodFlags.FilePolicy, flags, null, null, null, fileName,
                text, encoding, hashValue, hashAlgorithm, clientData,
                ref decision, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckBeforeStreamPolicies(
            PolicyFlags flags,
            string name,
            Encoding encoding,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            return CheckPolicies(
                MethodFlags.StreamPolicy, flags, null, null, null, name, null,
                encoding, null, null, clientData, ref decision, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckAfterStreamPolicies(
            PolicyFlags flags,
            string name,
            string text,
            Encoding encoding,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            string hashAlgorithm = null; /* NOTE: Means "use default". */
            byte[] hashValue = null;

            if (!FlagOps.HasFlags(flags, PolicyFlags.NoHash, true))
            {
                Result hashError = null;

                hashValue = RuntimeOps.HashString(
                    hashAlgorithm, text, encoding, ref hashError);

#if DEBUG && VERBOSE
                if (hashValue == null)
                    DebugOps.Complain(this, ReturnCode.Error, hashError);
#endif
            }

            return CheckPolicies(
                MethodFlags.StreamPolicy, flags, null, null, null, name, text,
                encoding, hashValue, hashAlgorithm, clientData, ref decision,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode CheckPolicies(
            MethodFlags methodFlags,
            PolicyFlags policyFlags,
            IExecute execute,
            ArgumentList arguments,
            IScript script,
            string fileName,
            string text,
            Encoding encoding,
            byte[] hashValue,
            string hashAlgorithm,
            IClientData clientData,
            ref PolicyDecision decision,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(interpreterFlags, InterpreterFlags.NoPolicies, true) &&
                (PolicyLevels == 0))
            {
                //
                // NOTE: Does this interpreter support policies?
                //
                if (policies != null)
                {
                    //
                    // NOTE: Prevent endless policy recursion.
                    //
                    interpreterFlags |= InterpreterFlags.NoPolicies;

                    //
                    // NOTE: Create the policy data object with the current policy
                    //       related information (including the original decision,
                    //       which is typically "unknown").
                    //
                    IPolicyContext policyContext = PolicyContext.Create(
                        policyFlags, execute, arguments, script, fileName,
                        text, encoding, hashValue, hashAlgorithm, clientData,
                        /* plugin */ null, decision);

                    //
                    // NOTE: When going through the generic callback delegates, we
                    //       need to wrap our object in a clientData.
                    //
                    IClientData policyClientData = new ClientData(policyContext);

                    //
                    // NOTE: Process each policy (as long as they all continue
                    //       to succeed).
                    //
                    foreach (KeyValuePair<string, _Wrappers.Policy> pair in policies)
                    {
                        if (pair.Value != null)
                        {
                            //
                            // NOTE: Do not invoke the policy if it has been disabled.
                            //
                            if (!EntityOps.IsDisabled(pair.Value))
                            {
                                //
                                // NOTE: Only execute the policies which match the specified
                                //       flags (i.e. command or file policies).
                                //
                                // FIXME: *PERF* Make sure this does not impact policy
                                //        performance too seriously.
                                //
                                if (EntityOps.HasFlags(pair.Value, methodFlags, true))
                                {
                                    //
                                    // NOTE: Since policies can basically do anything they want,
                                    //       we wrap them in a try block to prevent exceptions
                                    //       from escaping.
                                    //
                                    EnterPolicyLevel();

                                    try
                                    {
                                        //
                                        // HACK: Since we need the plugin to vary with each
                                        //       policy called, do that now (i.e. to provide
                                        //       more context for isolated policy support).
                                        //
                                        policyContext.Plugin = pair.Value.Plugin;

                                        code = pair.Value.Execute(
                                            this, policyClientData, arguments, ref result);
                                    }
                                    catch (Exception e)
                                    {
                                        //
                                        // NOTE: Translate exceptions to a failure return.
                                        //
                                        result = String.Format(
                                            "caught exception while checking policies: {0}",
                                            e);

                                        code = ReturnCode.Error;
                                    }
                                    finally
                                    {
                                        ExitPolicyLevel();
                                    }

                                    //
                                    // NOTE: Check for exception results specially because we
                                    //       treat "Break" different from other return codes.
                                    //
                                    if (code == ReturnCode.Break)
                                    {
                                        //
                                        // NOTE: Success; however, skip processing further
                                        //       traces for this policy operation.
                                        //
                                        code = ReturnCode.Ok;
                                        break;
                                    }
                                    else if (code != ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: Some type of failure (or exception), stop
                                        //       processing for this policy operation.
                                        //
                                        break;
                                    }
                                }
                            }
                        }
                    }

                    //
                    // NOTE: If all the policies succeeded and our local "pass" status
                    //       is valid, copy it to the variable provided by the caller.
                    //
                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Each policy gets one "vote" about whether or not to
                        //       allow the execution of the command.  One or more of
                        //       the current policies must agree for the execution of
                        //       the command to be allowed.  There must always be at
                        //       least one policy that "votes" to allow execution of
                        //       the command before it will be allowed (i.e. the
                        //       default policy denies everything).
                        //
                        decision = policyContext.Decision;

                        //
                        // NOTE: If there is no result so far, use the reason given
                        //       for the policy decision, if any.  We do not care if
                        //       the reason is null at this point.
                        //
                        if (result == null)
                            result = policyContext.Reason;
                    }

#if POLICY_TRACE
                    if (policyTrace)
                    {
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: CheckPolicies: methodFlags = {1}, " +
                            "policyFlags = {2}, execute = {3}, " +
                            "arguments = {4}, script = {5}, " +
                            "fileName = {6}, text = {7}, encoding = {8}, " +
                            "hashValue = {9}, hashAlgorithm = {10}, " +
                            "clientData = {11}, decision = {12}, " +
                            "code = {13}, result = {14}{15}",
                            GlobalState.GetCurrentThreadId(),
                            FormatOps.WrapOrNull(methodFlags),
                            FormatOps.WrapOrNull(policyFlags),
                            FormatOps.WrapOrNull(execute),
                            FormatOps.WrapOrNull(true, true, arguments),
                            FormatOps.WrapOrNull(true, true, script),
                            FormatOps.WrapOrNull(fileName),
                            FormatOps.WrapOrNull(true, true, text),
                            FormatOps.WrapOrNull(encoding),
                            ArrayOps.ToHexadecimalString(hashValue),
                            FormatOps.WrapOrNull(hashAlgorithm),
                            FormatOps.WrapOrNull(clientData),
                            decision, code, FormatOps.WrapOrNull(result),
                            Environment.NewLine), typeof(Interpreter).Name);
                    }
#endif

                    //
                    // NOTE: Remove policy prevention flag.
                    //
                    interpreterFlags &= ~InterpreterFlags.NoPolicies;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddPolicy(
            ExecuteCallback callback,
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddPolicy(callback, plugin, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddScriptPolicy(
            PolicyFlags flags,
            Type commandType,
            long commandToken,
            Interpreter policyInterpreter,
            string text,
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddScriptPolicy(
                ScriptPolicy.Create(flags, commandType, commandToken, policyInterpreter, text),
                plugin, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemovePolicy(
            string name,
            long token,
            IPolicy policy,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid policy name";
                return ReturnCode.Error;
            }

            policies.Remove(name);

            if (policy != null)
            {
                //
                // NOTE: Are we allowed to assist in managing the policy
                //       token via the associated plugin?
                //
                bool noToken = EntityOps.IsNoToken(policy);

                if (!noToken)
                {
                    //
                    // NOTE: If this policy is being managed by a plugin,
                    //       remove the associated token from said plugin.
                    //
                    IPlugin plugin = policy.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.PolicyTokens;

                        if ((tokens != null) && tokens.Contains(token))
                        {
                            tokens.Remove(token);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the policy tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.PolicyTokens = tokens;
#endif
                        }
                    }
                }
            }

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Policy, NotifyFlags.Removed,
                new ObjectTriplet(policy, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ITrace
        public ReturnCode GetTrace(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ITrace trace,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasTraces(ref error))
                {
                    if ((traces.Count > 0) &&
                        traces.ContainsKey(token))
                    {
                        code = PrivateGetTrace(token, wrapper, validate, ref name, ref trace, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid trace token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetTrace(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ITrace trace,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasTraces(ref error))
                {
                    if ((traces.Count > 0) &&
                        (name != null) &&
                        traces.ContainsKey(name))
                    {
                        code = PrivateGetTrace(name, wrapper, validate, ref token, ref trace, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid trace name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddTrace(
            TraceCallback callback,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            //
            // NOTE: All traces added via this method overload are global
            //       (i.e. interpreter-wide).
            //
            ITrace trace = RuntimeOps.NewCoreTrace(
                callback, clientData, TraceFlags.Global, plugin, ref result);

            if (traces != null)
            {
                if (AddTrace(trace, clientData, ref token, ref result) == ReturnCode.Ok)
                {
                    result = trace.Name;
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddTrace(
            ITrace trace,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (trace == null)
                {
                    result = "invalid trace";
                    return ReturnCode.Error;
                }

                if (trace.Name == null)
                {
                    result = "invalid trace name";
                    return ReturnCode.Error;
                }

                if (!HasTraces(ref result))
                    return ReturnCode.Error;

                if (DoesTraceExist(trace.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": trace already exists",
                        trace.Name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers.Trace traceWrapper = new _Wrappers.Trace(id, trace);
                long newToken = traceWrapper.Token;

                //
                // NOTE: Are we allowed to assist in managing the trace
                //       token via the associated plugin?
                //
                if (!EntityOps.IsNoToken(trace))
                {
                    //
                    // NOTE: Store the trace token so that we can remove
                    //       them later without using their names (which
                    //       can change).
                    //
                    IPlugin plugin = trace.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.TraceTokens;

                        if ((tokens != null) && !tokens.Contains(newToken))
                        {
                            tokens.Add(newToken);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the trace tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.TraceTokens = tokens;
#endif
                        }
                    }
                }

                trace.Token = newToken;
                traces.Add(trace.Name, traceWrapper);
                token = newToken;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Trace, NotifyFlags.Added,
                    new ObjectPair(trace, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveTrace(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTraces(ref result))
                    {
                        string name = null;
                        ITrace trace = null;

                        if (GetTrace(token, true, true, ref name,
                                ref trace, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(trace))
                            {
                                return PrivateRemoveTrace(
                                    name, token, trace, clientData, ref result);
                            }
                            else
                            {
                                result = "trace is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveTrace(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTraces(ref result))
                    {
                        long token = 0;
                        ITrace trace = null;

                        if (GetTrace(name, true, true, ref token,
                                ref trace, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(trace))
                            {
                                return PrivateRemoveTrace(
                                    name, token, trace, clientData, ref result);
                            }
                            else
                            {
                                result = "trace is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetTrace(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ITrace trace,
            ref Result error
            )
        {
            if (HasTraces(ref error))
            {
                if (wrapper)
                {
                    trace = traces[token];
                    name = EntityOps.GetName(trace as IIdentifier);

                    if (validate && (trace == null))
                        error = "invalid trace";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (traces[token] != null)
                    {
                        trace = traces[token].trace;
                        name = EntityOps.GetName(trace as IIdentifier);

                        if (validate && (trace == null))
                            error = "invalid trace";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid trace wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetTrace(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ITrace trace,
            ref Result error
            )
        {
            if (HasTraces(ref error))
            {
                if (wrapper)
                {
                    trace = traces[name];
                    token = EntityOps.GetToken(trace as IWrapperData);

                    if (validate && (trace == null))
                        error = "invalid trace";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (traces[name] != null)
                    {
                        trace = traces[name].trace;
                        token = EntityOps.GetToken(trace as IWrapperData);

                        if (validate && (trace == null))
                            error = "invalid trace";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid trace wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetTrace(
            string name,
            bool wrapper,
            bool validate,
            ref ITrace trace,
            ref Result error
            )
        {
            long token = 0;

            return GetTrace(name, wrapper, validate, ref token, ref trace, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal TraceList GetTraces(
            TraceList traces
            )
        {
            TraceList result = null;

            if ((this.traces != null) || (traces != null))
            {
                result = new TraceList();

                if (traces != null)
                    result.AddRange(traces);

                if (this.traces != null)
                    result.AddRange(this.traces);
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddTrace(
            ITrace trace,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddTrace(trace, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveTrace(
            string name,
            long token,
            ITrace trace,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid trace name";
                return ReturnCode.Error;
            }

            traces.Remove(name);

            if (trace != null)
            {
                //
                // NOTE: Are we allowed to assist in managing the trace
                //       token via the associated plugin?
                //
                bool noToken = EntityOps.IsNoToken(trace);

                if (!noToken)
                {
                    //
                    // NOTE: If this trace is being managed by a plugin,
                    //       remove the associated token from said plugin.
                    //
                    IPlugin plugin = trace.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.TraceTokens;

                        if ((tokens != null) && tokens.Contains(token))
                        {
                            tokens.Remove(token);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the trace tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.TraceTokens = tokens;
#endif
                        }
                    }
                }
            }

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Trace, NotifyFlags.Removed,
                new ObjectTriplet(trace, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IAlias
        public ReturnCode GetAlias(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IAlias alias,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasAliases(ref error))
                {
                    if ((aliases.Count > 0) &&
                        aliases.ContainsKey(token))
                    {
                        code = PrivateGetAlias(token, wrapper, validate, ref name, ref alias, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid alias token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetAlias(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IAlias alias,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasAliases(ref error))
                {
                    if ((aliases.Count > 0) &&
                        (name != null) &&
                        aliases.ContainsKey(name))
                    {
                        code = PrivateGetAlias(name, wrapper, validate, ref token, ref alias, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid alias name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddAlias(
            string name,
            CommandFlags flags,
            IClientData clientData,
            Interpreter targetInterpreter,
            IExecute target,
            ArgumentList arguments,
            OptionDictionary options,
            int startIndex,
            ref long token,
            ref IAlias alias,
            ref Result result
            )
        {
            CheckDisposed();

            //
            // NOTE: Create the new alias with a unique name token (which may
            //       be the same as the name).
            //
            IAlias newAlias = RuntimeOps.NewAlias(
                name, flags, clientData, GetAliasNameToken(name),
                targetInterpreter, target, arguments, options, startIndex);

            //
            // NOTE: For now, rely entirely upon the logic in AddCommand.
            //       Eventually, we may want to perform some specialized
            //       argument validation here, especially if safe interpreters
            //       are going to be implemented.
            //
            ReturnCode code = AddAliasAndCommand(
                newAlias, null, ref token, ref result);

            if (code == ReturnCode.Ok)
            {
                alias = newAlias;
                result = newAlias.NameToken;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddAlias(
            IAlias alias,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (alias == null)
                {
                    result = "invalid alias";
                    return ReturnCode.Error;
                }

                if (alias.NameToken == null)
                {
                    result = "invalid alias name";
                    return ReturnCode.Error;
                }

                if (!HasAliases(ref result))
                    return ReturnCode.Error;

                if (DoesAliasExist(alias.NameToken) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": alias already exists",
                        alias.NameToken);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers.Alias aliasWrapper = new _Wrappers.Alias(id, alias);

                aliases.Add(alias.NameToken, aliasWrapper);
                token = aliasWrapper.Token;
                alias.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Alias, NotifyFlags.Added,
                    new ObjectPair(alias, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetAlias(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IAlias alias,
            ref Result error
            )
        {
            if (HasAliases(ref error))
            {
                if (wrapper)
                {
                    alias = aliases[token];
                    name = EntityOps.GetName(alias as IIdentifier);

                    if (validate && (alias == null))
                        error = "invalid alias";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (aliases[token] != null)
                    {
                        alias = aliases[token].alias;
                        name = EntityOps.GetName(alias as IIdentifier);

                        if (validate && (alias == null))
                            error = "invalid alias";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid alias wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetAlias(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IAlias alias,
            ref Result error
            )
        {
            if (HasAliases(ref error))
            {
                if (wrapper)
                {
                    alias = aliases[name];
                    token = EntityOps.GetToken(alias as IWrapperData);

                    if (validate && (alias == null))
                        error = "invalid alias";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (aliases[name] != null)
                    {
                        alias = aliases[name].alias;
                        token = EntityOps.GetToken(alias as IWrapperData);

                        if (validate && (alias == null))
                            error = "invalid alias";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid alias wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetAlias(
            string name,
            bool wrapper,
            bool validate,
            ref IAlias alias
            )
        {
            Result error = null;

            return GetAlias(name, wrapper, validate, ref alias, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAlias(
            string name,
            bool wrapper,
            bool validate,
            ref IAlias alias,
            ref Result error
            )
        {
            long token = 0;

            return GetAlias(name, wrapper, validate, ref token, ref alias, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAliasArguments(
            IAlias alias,
            ArgumentList arguments,
            ref string targetName
            )
        {
            ArgumentList targetArguments = null;

            return GetAliasArguments(
                alias, arguments, ref targetName, ref targetArguments);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetAliasArguments(
            IAlias alias,
            ArgumentList arguments,
            ref string targetName,
            ref ArgumentList targetArguments
            )
        {
            Result error = null;

            return GetAliasArguments(
                alias, arguments, ref targetName, ref targetArguments, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAliasArguments(
            IAlias alias,
            ArgumentList arguments,
            ref string targetName,
            ref ArgumentList targetArguments,
            ref Result error
            )
        {
            ReturnCode code;

            if (alias != null)
            {
                ICommand command = null;

                code = RuntimeOps.GetAliasCommand(alias, true, true, ref command, ref error);

                if (code == ReturnCode.Ok)
                {
                    bool mergeArguments = FlagOps.HasFlags(command.Flags, CommandFlags.MergeArguments, true);
                    bool skipTargetName = FlagOps.HasFlags(command.Flags, CommandFlags.SkipTargetName, true);
                    bool skipSourceName = FlagOps.HasFlags(command.Flags, CommandFlags.SkipSourceName, true);

                    if (mergeArguments)
                    {
                        //
                        // NOTE: Merged argument handling, carefully merge the option and
                        //       non-option arguments from the first and second lists.  The
                        //       start index for the second list here is hard-coded to one
                        //       because we know that the command name itself is the first
                        //       argument in the second list.
                        //
                        code = MergeArguments(
                            alias.Options, alias.Arguments, arguments, alias.StartIndex, 1,
                            skipTargetName, skipSourceName, ref targetArguments, ref error);
                    }
                    else
                    {
                        targetArguments = null;

                        //
                        // NOTE: Standard argument handling, simply append the arguments from
                        //       the first and second lists, in that order, to the target
                        //       argument list.
                        //
                        if ((alias.Arguments != null) || (arguments != null))
                            targetArguments = new ArgumentList();

                        if (alias.Arguments != null)
                            targetArguments.AddRange(ArgumentList.GetRange(
                                alias.Arguments, skipTargetName ? 1 : 0));

                        if (arguments != null)
                            targetArguments.AddRange(ArgumentList.GetRange(
                                arguments, skipSourceName ? 1 : 0));

                        code = ReturnCode.Ok;
                    }

                    //
                    // NOTE: If necessary, modify the first argument to be the alias
                    //       name.  Afterwards, always provide the value of the first
                    //       argument to the caller.
                    //
                    targetName = null;

                    if ((targetArguments != null) && (targetArguments.Count > 0))
                        targetName = targetArguments[0];
                }
            }
            else
            {
                error = "invalid alias";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAliasTarget(
            IAlias alias,
            string targetName,
            bool wrapper,
            bool validate,
            bool strict,
            ref IExecute target
            )
        {
            Result error = null;

            return GetAliasTarget(alias, targetName, wrapper, validate, strict, ref target, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAliasTarget(
            IAlias alias,
            string targetName,
            bool wrapper,
            bool validate,
            bool strict,
            ref IExecute target,
            ref Result error
            )
        {
            ReturnCode code;

            //
            // NOTE: First, try for the pre-resolved target.
            //
            target = alias.Target;

            //
            // NOTE: Otherwise, is this a late-bound alias?
            //
            if (target != null)
            {
                //
                // NOTE: Success, just return it.
                //
                code = ReturnCode.Ok;
            }
            else
            {
                Result localError = null;

                //
                // NOTE: Not found; lookup the target by name.
                //
                code = GetIExecuteViaResolvers(
                    GetResolveEngineFlags(true), targetName, null,
                    wrapper, validate, ref target, ref localError);

                //
                // NOTE: During the alias creation process, the target
                //       does not have to be defined.  However, during
                //       the execution process, it does.
                //
                if (code != ReturnCode.Ok)
                {
                    if (!strict)
                        code = ReturnCode.Ok;
                    else
                        error = localError;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetObjectAliasTarget(
            ref IExecute target,
            ref Result error
            )
        {
            ICommand command = null;

            if (GetCommand(typeof(_Commands.Object), true, true, ref command, ref error) == ReturnCode.Ok)
            {
                target = command;

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal string GetAliasNameToken(string name)
        {
            //
            // NOTE: Initially, the name token is the same as the name;
            //       however, this may change later if the command is
            //       renamed.
            //
            string result = name;

            if (!String.IsNullOrEmpty(result))
                while (DoesAliasExist(result) == ReturnCode.Ok)
                    /*
                     * NOTE: This comment stolen from "generic/tclInterp.c".
                     *
                     * The alias name cannot be used as unique token, it is already taken.
                     * We can produce a unique token by prepending "::" repeatedly. This
                     * algorithm is a stop-gap to try to maintain the command name as
                     * token for most use cases, fearful of possible backwards compat
                     * problems. A better algorithm would produce unique tokens that need
                     * not be related to the command name.
                     *
                     * ATTENTION: the tests in interp.test and possibly safe.test depend
                     * on the precise definition of these tokens.
                     *
                     */
                    result = TclVars.NamespaceSeparator + result;

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private IAlias FollowAlias(
            IAlias alias,
            IAlias nextAlias,
            string newName
            )
        {
            IAlias result = null;

            if (nextAlias.Target != null)
            {
                result = nextAlias.Target as IAlias;
            }
            else
            {
                string targetName = null;
                ArgumentList targetArguments = null;

                if (GetAliasArguments(nextAlias, null, ref targetName,
                        ref targetArguments) == ReturnCode.Ok)
                {
                    if (targetName != null)
                    {
                        if (GetAlias(targetName, true, true,
                                ref result) != ReturnCode.Ok)
                        {
                            //
                            // NOTE: If the alias is being renamed, make sure
                            //       we are using the new name (i.e. in that
                            //       case the caller(s) is/are responsible for
                            //       passing it to us).
                            //
                            if (String.Compare(targetName,
                                (newName != null) ? newName : alias.Name,
                                StringOps.SystemStringComparisonType) == 0)
                            {
                                result = alias;
                            }
                        }
                    }
                }
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PreventAliasLoop(
            IAlias alias,
            string newName,
            ref Result error
            )
        {
            IAlias nextAlias = alias;

            while (nextAlias != null)
            {
                //
                // NOTE: Proceed to the next alias in the chain.
                //
                nextAlias = FollowAlias(alias, nextAlias, newName);

                //
                // NOTE: Check to see if we have found a loop.
                //
                if (Object.ReferenceEquals(nextAlias, alias))
                {
                    error = String.Format(
                        "cannot define or rename alias \"{0}\": would create a loop",
                        (newName != null) ? newName : alias.Name);

                    return ReturnCode.Error;
                }
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddAlias(
            string name,
            CommandFlags flags,
            IClientData clientData,
            Interpreter targetInterpreter,
            IExecute target,
            ArgumentList arguments,
            OptionDictionary options,
            int startIndex,
            ref Result result
            )
        {
            long token = 0;

            return AddAlias(name, flags, clientData, targetInterpreter, target, arguments,
                options, startIndex, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddAlias(
            string name,
            CommandFlags flags,
            IClientData clientData,
            Interpreter targetInterpreter,
            IExecute target,
            ArgumentList arguments,
            OptionDictionary options,
            int startIndex,
            ref long token,
            ref Result result
            )
        {
            IAlias alias = null;

            return AddAlias(name, flags, clientData, targetInterpreter, target, arguments,
                options, startIndex, ref token, ref alias, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddAlias(
            string name,
            CommandFlags flags,
            IClientData clientData,
            Interpreter targetInterpreter,
            IExecute target,
            ArgumentList arguments,
            OptionDictionary options,
            int startIndex,
            ref IAlias alias,
            ref Result result
            )
        {
            long token = 0;

            return AddAlias(name, flags, clientData, targetInterpreter, target, arguments,
                options, startIndex, ref token, ref alias, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveAlias(
            string name,
            long token,
            IAlias alias,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid alias name";
                return ReturnCode.Error;
            }

            aliases.Remove(name);

            if (alias != null)
            {
                //
                // NOTE: Prevent the command from trying to cleanup
                //       the alias since it has now been deleted.
                //
                alias.NameToken = null;
            }

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Alias, NotifyFlags.Removed,
                new ObjectTriplet(alias, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode RemoveAlias(
            long token,
            IClientData clientData,
            ref Result result
            ) /* NOT USED */
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasAliases(ref result))
                    {
                        string name = null;
                        IAlias alias = null;

                        if (GetAlias(token, true, true, ref name,
                                ref alias, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveAlias(
                                name, token, alias, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveAlias(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            /* CheckDisposed(); */

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasAliases(ref result))
                    {
                        long token = 0;
                        IAlias alias = null;

                        if (GetAlias(name, true, true, ref token,
                                ref alias, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveAlias(
                                name, token, alias, clientData, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectAlias(
            string name,
            OptionDictionary options,
            bool all,
            ref IAlias alias,
            ref Result error
            )
        {
            IExecute target = null;

            if (GetObjectAliasTarget(ref target, ref error) == ReturnCode.Ok)
            {
                ArgumentList arguments = null;

                if (RuntimeOps.GetObjectAliasArguments(all, ref arguments, ref error) == ReturnCode.Ok)
                {
                    Result result = null;

                    if (AddAlias(name, CommandFlags.MergeArguments, ClientData.Empty, this,
                            target, arguments, options, 2 /* [object invoke] */,
                            ref alias, ref result) == ReturnCode.Ok)
                    {
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        //
                        // NOTE: Transfer error message to caller.
                        //
                        error = result;
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddAliasAndCommand(
            IAlias alias,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            if (PreventAliasLoop(alias, null, ref result) == ReturnCode.Ok)
                //
                // NOTE: If the provided alias is not also a command, this cast will
                //       silently fail; however, that is fine because AddCommand will
                //       return an appropriate error message to the caller.
                //
                if (AddCommand(alias as ICommand, clientData, ref result) == ReturnCode.Ok)
                    if (AddAlias(alias, clientData, ref token, ref result) == ReturnCode.Ok)
                        return ReturnCode.Ok;

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveAliasAndCommand(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            /* CheckDisposed(); */

            IAlias alias = null;

            if (GetAlias(name, true, true, ref alias, ref result) == ReturnCode.Ok)
                if (RemoveAlias(alias.NameToken, clientData, ref result) == ReturnCode.Ok)
                    if (RemoveCommand(alias.Name, clientData, ref result) == ReturnCode.Ok)
                        return ReturnCode.Ok;

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IFunction
        public ReturnCode GetFunction(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IFunction function,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasFunctions(ref error))
                {
                    if ((functions.Count > 0) &&
                        functions.ContainsKey(token))
                    {
                        if (wrapper)
                        {
                            function = functions[token];
                            name = EntityOps.GetName(function as IIdentifier);

                            if (!validate || (function != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (functions[token] != null)
                            {
                                function = functions[token].function;
                                name = EntityOps.GetName(function as IIdentifier);

                                if (!validate || (function != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid function token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetFunction(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IFunction function,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasFunctions(ref error))
                {
                    if ((functions.Count > 0) &&
                        (name != null) &&
                        functions.ContainsKey(name))
                    {
                        if (wrapper)
                        {
                            function = functions[name];
                            token = EntityOps.GetToken(function as IWrapperData);

                            if (!validate || (function != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (functions[name] != null)
                            {
                                function = functions[name].function;
                                token = EntityOps.GetToken(function as IWrapperData);

                                if (!validate || (function != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid function name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ListFunctions(
            FunctionFlags hasFlags,
            FunctionFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasFunctions(ref localError))
                {
                    code = functions.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddFunction(
            Type type,
            string name,
            int arguments,
            TypeList types,
            FunctionFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            if (type == null)
                type = Type.GetType(
                    FormatOps.FunctionTypeName(name), false, true);

            if (type == null)
                type = Type.GetType(
                    FormatOps.FunctionTypeName(Characters.Underscore + name), false, true);

            if (type != null)
            {
                try
                {
                    if (arguments == (int)Arity.Automatic)
                        arguments = AttributeOps.GetArguments(type);

                    object[] args = {
                        new FunctionData(
                            name, null, null, arguments, types,
                            flags, clientData, plugin, 0)
                    };

                    IFunction function = (IFunction)Activator.CreateInstance(type, args);

                    return AddFunction(function, null, ref token, ref result);
                }
                catch (Exception e)
                {
                    result = e;
                }
            }
            else
            {
                result = String.Format(
                    "function \"{0}\" not found",
                    FormatOps.FunctionTypeName(name));
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddFunction(
            IFunction function,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (function == null)
                {
                    result = "invalid function";
                    return ReturnCode.Error;
                }

                if (function.Name == null)
                {
                    result = "invalid function name";
                    return ReturnCode.Error;
                }

                if (!HasFunctions(ref result))
                    return ReturnCode.Error;

                if (DoesFunctionExist(function.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": function already exists",
                        function.Name);

                    return ReturnCode.Error;
                }

                //
                // HACK: Due to how the expression parser works, we cannot allow a
                //       function to be added that has the name of an existing
                //       operator.
                //
                if (DoesOperatorExist(function.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": operator already exists",
                        function.Name);

                    return ReturnCode.Error;
                }

                ReturnCode code = ReturnCode.Ok;

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    try
                    {
                        code = function.Initialize(this,
                            (clientData != null) ? clientData : function.ClientData,
                            ref result);
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while initializing function: {0}",
                            e);

                        code = ReturnCode.Error;
                    }
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        ReturnCode terminateCode;
                        Result terminateResult = null;

                        try
                        {
                            terminateCode = function.Terminate(this,
                                (clientData != null) ? clientData : function.ClientData,
                                ref terminateResult);
                        }
                        catch (Exception e)
                        {
                            terminateResult = String.Format(
                                "caught exception while terminating function: {0}",
                                e);

                            terminateCode = ReturnCode.Error;
                        }

                        if (terminateCode != ReturnCode.Ok)
                            DebugOps.Complain(this, terminateCode, terminateResult);
                    }
                }

                if (code != ReturnCode.Ok)
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                    return code;
                }

                long id = NextId();
                _Wrappers.Function functionWrapper = new _Wrappers.Function(id, function);
                long newToken = functionWrapper.Token;

                //
                // NOTE: Are we allowed to assist in managing the function
                //       token via the associated plugin?
                //
                if (!EntityOps.IsNoToken(function))
                {
                    //
                    // NOTE: Store the function token so that we can remove
                    //       them later without using their names (which
                    //       can change).
                    //
                    IPlugin plugin = function.Plugin;

                    if (plugin != null)
                    {
                        LongList tokens = plugin.FunctionTokens;

                        if ((tokens != null) && !tokens.Contains(newToken))
                        {
                            tokens.Add(newToken);

#if ISOLATED_PLUGINS
                            //
                            // HACK: Force the function tokens for the
                            //       isolated plugin to be updated.
                            //
                            if (RuntimeOps.IsPluginIsolated(plugin))
                                plugin.FunctionTokens = tokens;
#endif
                        }
                    }
                }

                function.Token = newToken;
                functions.Add(function.Name, functionWrapper);
                token = newToken;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Function, NotifyFlags.Added,
                    new ObjectPair(function, token), this,
                    clientData, null, null, ref result);
#endif

                /* IGNORED */
                GlobalState.PopActiveInterpreter();
                return code;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveFunction(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasFunctions(ref result))
                    {
                        string name = null;
                        IFunction function = null;

                        if (GetFunction(token, true, true, ref name,
                                ref function, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(function))
                            {
                                return PrivateRemoveFunction(
                                    name, token, function, clientData, ref result);
                            }
                            else
                            {
                                result = "function is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveFunction(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasFunctions(ref result))
                    {
                        long token = 0;
                        IFunction function = null;

                        if (GetFunction(name, true, true, ref token,
                                ref function, ref result) == ReturnCode.Ok)
                        {
                            if (!EntityOps.IsReadOnly(function))
                            {
                                return PrivateRemoveFunction(
                                    name, token, function, clientData, ref result);
                            }
                            else
                            {
                                result = "function is read-only";
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode GetFunction(
            string name,
            bool wrapper,
            bool validate,
            ref IFunction function,
            ref Result error
            )
        {
            long token = 0;

            return GetFunction(name, wrapper, validate, ref token, ref function, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddFunction(
            Type type,
            string name,
            TypeList types,
            FunctionFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddFunction(
                type, name, types, flags, plugin, clientData,
                ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddFunction(
            Type type,
            string name,
            TypeList types,
            FunctionFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            return AddFunction(
                type, name, (int)Arity.Automatic, types, flags,
                plugin, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveFunction(
            string name,
            long token,
            IFunction function,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid function name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
                if ((function == null) ||
                    (function.Terminate(this, (clientData != null) ?
                            clientData : function.ClientData, ref result) == ReturnCode.Ok))
                {
                    functions.Remove(name);

                    if (function != null)
                    {
                        //
                        // NOTE: Are we allowed to assist in managing the function
                        //       token via the associated plugin?
                        //
                        bool noToken = EntityOps.IsNoToken(function);

                        if (!noToken)
                        {
                            //
                            // NOTE: If this function is being managed by a plugin,
                            //       remove the associated token from said plugin.
                            //
                            IPlugin plugin = function.Plugin;

                            if (plugin != null)
                            {
                                LongList tokens = plugin.FunctionTokens;

                                if ((tokens != null) && tokens.Contains(token))
                                {
                                    tokens.Remove(token);

#if ISOLATED_PLUGINS
                                    //
                                    // HACK: Force the function tokens for the
                                    //       isolated plugin to be updated.
                                    //
                                    if (RuntimeOps.IsPluginIsolated(plugin))
                                        plugin.FunctionTokens = tokens;
#endif
                                }
                            }
                        }
                    }

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Function, NotifyFlags.Removed,
                        new ObjectTriplet(function, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

                return ReturnCode.Error;
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating function: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IProcedure
        public ReturnCode GetProcedure(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IProcedure procedure,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasProcedures(ref error))
                {
                    if ((procedures.Count > 0) &&
                        procedures.ContainsKey(token))
                    {
                        code = PrivateGetProcedure(token, wrapper, validate, ref name, ref procedure, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid procedure token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IProcedure procedure,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasProcedures(ref error))
                {
                    if ((procedures.Count > 0) &&
                        (name != null) &&
                        procedures.ContainsKey(name))
                    {
                        code = PrivateGetProcedure(name, wrapper, validate, ref token, ref procedure, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid procedure name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ListProcedures(
            ProcedureFlags hasFlags,
            ProcedureFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasProcedures(ref localError))
                {
                    code = procedures.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddOrUpdateProcedure(
            string name,
            ProcedureFlags flags,
            StringList arguments,
            string body,
            IScriptLocation location,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            CheckDisposed();

            IProcedure procedure = RuntimeOps.NewCoreProcedure(
                name, null, null, flags,
                new ArgumentList(arguments, ArgumentFlags.NameOnly),
                body, location, clientData);

            return AddOrUpdateProcedure(procedure, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RenameProcedure(
            string oldName,
            string newName,
            bool delete,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasProcedures(ref result))
                    {
                        if (!delete || !String.IsNullOrEmpty(newName))
                        {
                            IProcedure oldProcedure = null;

                            if (GetProcedure(oldName, true, true, ref oldProcedure, ref result) == ReturnCode.Ok)
                            {
                                if (!EntityOps.IsReadOnly(oldProcedure))
                                {
                                    _Wrappers.Procedure oldWrapper = oldProcedure as _Wrappers.Procedure;

                                    if (oldWrapper != null)
                                    {
                                        IAlias oldAlias = oldWrapper.procedure as IAlias;

                                        if (PreventAliasLoop(oldAlias, newName, ref result) == ReturnCode.Ok)
                                        {
                                            if (IsUniqueCommandName(newName, ref result) == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: New name, same token.
                                                //
                                                oldWrapper.Name = newName;

                                                /* IGNORED */
                                                procedures.Rename(oldName, newName);

#if EXECUTE_CACHE
                                                if (executeCache != null)
                                                    executeCache.Rename(oldName, newName, oldWrapper, false);
#endif

#if NOTIFY
                                                /* IGNORED */
                                                CheckNotifications(
                                                    NotifyType.Procedure, NotifyFlags.Renamed,
                                                    new ObjectTriplet(oldProcedure, oldName, newName), this,
                                                    null, null, null, ref result);
#endif

                                                result = String.Empty;
                                                return ReturnCode.Ok;
                                            }
                                        }
                                    }
                                    else
                                    {
                                        result = String.Format(
                                            "can't rename \"{0}\": invalid procedure wrapper",
                                            oldName);
                                    }
                                }
                                else
                                {
                                    result = "procedure is read-only";
                                }
                            }
                            else
                            {
                                result = String.Format(
                                    "can't rename \"{0}\": procedure doesn't exist",
                                    oldName);
                            }
                        }
                        else
                        {
                            return RemoveProcedure(oldName, null, ref result);
                        }
                    }
                }
                else
                {
                    result = String.Format(
                        "can't rename \"{0}\": interpreter is read-only",
                        oldName);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveProcedure(
            long token,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    string name = null;
                    IProcedure procedure = null;

                    if (GetProcedure(token, true, true, ref name,
                            ref procedure, ref result) == ReturnCode.Ok)
                    {
                        if (!EntityOps.IsReadOnly(procedure))
                        {
                            return PrivateRemoveProcedure(
                                name, token, procedure, clientData, ref result);
                        }
                        else
                        {
                            result = "procedure is read-only";
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveProcedure(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    long token = 0;
                    IProcedure procedure = null;

                    if (GetProcedure(name, true, true, ref token,
                            ref procedure, ref result) == ReturnCode.Ok)
                    {
                        if (!EntityOps.IsReadOnly(procedure))
                        {
                            return PrivateRemoveProcedure(
                                name, token, procedure, clientData, ref result);
                        }
                        else
                        {
                            result = "procedure is read-only";
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MatchProcedure(
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref IProcedure procedure,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                bool useHidden = Engine.HasUseHidden(flags);

                ProcedureWrapperDictionary localProcedures =
                    useHidden ? hiddenProcedures : procedures;

                if (localProcedures != null)
                {
                    //
                    // NOTE: *WARNING* Empty command/procedure names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (localProcedures.Count > 0)
                        {
                            string exactName = null;
                            StringList list = new StringList();

                            if (localProcedures.ContainsKey(name) &&
                                (Engine.HasMatchHidden(flags) ||
                                !EntityOps.IsHidden(localProcedures[name])))
                            {
                                //
                                // NOTE: Exact procedure match.
                                //
                                exactName = name;

                                //
                                // NOTE: Store the exact procedure match in the result
                                //       list.
                                //
                                list.Add(name);
                            }
                            else
                            {
                                foreach (KeyValuePair<string, _Wrappers.Procedure> pair in localProcedures)
                                {
                                    if (String.Compare(
                                            pair.Key, 0, name, 0, name.Length,
                                            StringOps.SystemStringComparisonType) == 0)
                                    {
                                        //
                                        // NOTE: Is the procedure visible or are we allowed to
                                        //       find hidden procedures?
                                        //
                                        if (Engine.HasMatchHidden(flags) || !EntityOps.IsHidden(pair.Value))
                                        {
                                            //
                                            // NOTE: Was the key valid (this should always succeed).
                                            //
                                            if (pair.Key != null)
                                            {
                                                //
                                                // NOTE: It was a match; however, was it an exact match?
                                                //
                                                if (pair.Key.Length == name.Length)
                                                    //
                                                    // NOTE: Preserve match, it may differ in case.
                                                    //
                                                    exactName = pair.Key;

                                                //
                                                // NOTE: Was it an exact match or did we match at least one
                                                //       character in a partial match?
                                                //
                                                if ((pair.Key.Length == name.Length) || (name.Length > 0))
                                                    //
                                                    // NOTE: Store the exact or partial match in the results
                                                    //       dictionary.
                                                    //
                                                    list.Add(pair.Key);
                                            }
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: If there was an exact match, just use it.
                            //
                            if (exactName != null)
                            {
                                //
                                // NOTE: No need to change the caller's variable,
                                //       it already contains the correct value.
                                //
                                if ((useHidden && PrivateGetHiddenProcedure(
                                        exactName, wrapper, validate, ref token,
                                        ref procedure, ref error) == ReturnCode.Ok) ||
                                    (!useHidden && PrivateGetProcedure(
                                        exactName, wrapper, validate, ref token,
                                        ref procedure, ref error) == ReturnCode.Ok))
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenProcedure : NotifyType.Procedure,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, exactName, procedure, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count == 1)
                            {
                                //
                                // NOTE: Normal case, exactly one procedure matched.
                                //
                                if ((useHidden && PrivateGetHiddenProcedure(
                                        list[0], wrapper, validate, ref token,
                                        ref procedure, ref error) == ReturnCode.Ok) ||
                                    (!useHidden && PrivateGetProcedure(
                                        list[0], wrapper, validate, ref token,
                                        ref procedure, ref error) == ReturnCode.Ok))
                                {
#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        useHidden ? NotifyType.HiddenProcedure : NotifyType.Procedure,
                                        NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                        new ObjectList(flags, name, list[0], procedure, token), this, null, null,
                                        null, ref error);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            else if (list.Count > 1)
                            {
                                error = RuntimeOps.BadValue(
                                    "ambiguous", "procedure", name,
                                    list, null, null);

                                ambiguous = true;
                            }
                            else
                            {
                                error = RuntimeOps.BadValue(
                                    null, "procedure", name,
                                    localProcedures.Keys, null, null);
                            }
                        }
                        else
                        {
                            error = "no procedures defined";
                        }
                    }
                    else
                    {
                        error = "invalid procedure name";
                    }
                }
                else
                {
                    error = String.Format(
                        "{0}procedures not available",
                        useHidden ? "hidden " : String.Empty);
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private ReturnCode PrivateGetProcedure(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IProcedure procedure,
            ref Result error
            )
        {
            if (HasProcedures(ref error))
            {
                if (wrapper)
                {
                    procedure = procedures[token];
                    name = EntityOps.GetName(procedure as IIdentifier);

                    if (validate && (procedure == null))
                        error = "invalid procedure";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (procedures[token] != null)
                    {
                        procedure = procedures[token].procedure;
                        name = EntityOps.GetName(procedure as IIdentifier);

                        if (validate && (procedure == null))
                            error = "invalid procedure";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid procedure wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IProcedure procedure,
            ref Result error
            )
        {
            if (HasProcedures(ref error))
            {
                if (wrapper)
                {
                    procedure = procedures[name];
                    token = EntityOps.GetToken(procedure as IWrapperData);

                    if (validate && (procedure == null))
                        error = "invalid procedure";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (procedures[name] != null)
                    {
                        procedure = procedures[name].procedure;
                        token = EntityOps.GetToken(procedure as IWrapperData);

                        if (validate && (procedure == null))
                            error = "invalid procedure";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid procedure wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref IProcedure procedure
            )
        {
            Result error = null;

            return GetProcedure(name, wrapper, validate, ref procedure, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref IProcedure procedure,
            ref Result error
            )
        {
            long token = 0;

            return GetProcedure(name, wrapper, validate, ref token, ref procedure, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ListHiddenProcedures(
            ProcedureFlags hasFlags,
            ProcedureFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            // CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasHiddenProcedures(ref localError))
                {
                    code = hiddenProcedures.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddOrUpdateProcedureWithReplace(
            IProcedure procedure,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddOrUpdateProcedureWithReplace(
                procedure, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOrUpdateProcedureWithReplace(
            IProcedure procedure,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            if (procedure != null)
            {
                ReturnCode code = AddOrUpdateProcedure(procedure, ref token, ref result);

                if (code == ReturnCode.Ok)
                {
                    try
                    {
                        string name = procedure.Name;

                        if ((code == ReturnCode.Ok) &&
                            (DoesIExecuteExist(name) == ReturnCode.Ok))
                        {
                            code = RemoveIExecute(name, clientData, ref result);
                        }

                        if ((code == ReturnCode.Ok) &&
                            (DoesCommandExist(name) == ReturnCode.Ok))
                        {
                            code = RemoveCommand(name, clientData, ref result);
                        }
                    }
                    finally
                    {
                        //
                        // BUGFIX: If we fail to remove an ICommand or IExecute,
                        //         we must still remove the procedure we just
                        //         added.
                        //
                        if (code != ReturnCode.Ok)
                        {
                            ReturnCode removeCode;
                            Result removeResult = null;

                            removeCode = RemoveProcedure(token, clientData, ref removeResult);

                            if (removeCode != ReturnCode.Ok)
                                DebugOps.Complain(this, removeCode, removeResult);
                        }
                    }
                }

                return code;
            }
            else
            {
                result = "invalid procedure";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOrUpdateProcedure(
            IProcedure procedure,
            ref long token,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (procedure == null)
                {
                    result = "invalid procedure";
                    return ReturnCode.Error;
                }

                if (procedure.Name == null)
                {
                    result = "invalid procedure name";
                    return ReturnCode.Error;
                }

                if (!HasProcedures(ref result))
                    return ReturnCode.Error;

                long id = NextId();
                _Wrappers.Procedure procedureWrapper = new _Wrappers.Procedure(id, procedure);
                long newToken = procedureWrapper.Token;

                procedure.Token = newToken;

                if (procedures.ContainsKey(procedure.Name))
                {
#if NOTIFY
                    IProcedure oldProcedure = procedures[procedure.Name];
#endif

                    procedures[procedure.Name] = procedureWrapper;

#if EXECUTE_CACHE
                    if (executeCache != null)
                        executeCache.AddOrUpdate(procedure.Name, procedureWrapper, false);
#endif

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Procedure, NotifyFlags.Updated,
                        new ObjectPair(oldProcedure, procedure), this,
                        null, null, null, ref result);
#endif
                }
                else
                {
                    procedures.Add(procedure.Name, procedureWrapper);

#if EXECUTE_CACHE
                    if (executeCache != null)
                        executeCache.AddOrUpdate(procedure.Name, procedureWrapper, false);
#endif

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Procedure, NotifyFlags.Added,
                        new ObjectPair(procedure, newToken), this,
                        null, null, null, ref result);
#endif
                }

                token = newToken;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveProcedure(
            string name,
            long token,
            IProcedure procedure,
            IClientData clientData,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid procedure name";
                return ReturnCode.Error;
            }

            GlobalState.PushActiveInterpreter(this);

            try
            {
#if NATIVE && TCL
                if ((procedure == null) ||
                    (DisposeTclBridges(procedure, false, ref result) == ReturnCode.Ok))
#endif
                {
                    procedures.Remove(name);

#if EXECUTE_CACHE
                    if (executeCache != null)
                        executeCache.Remove(name, false);
#endif

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Procedure, NotifyFlags.Removed,
                        new ObjectTriplet(procedure, name, token), this,
                        clientData, null, null, ref result);
#endif

                    return ReturnCode.Ok;
                }

#if NATIVE && TCL
                return ReturnCode.Error;
#endif
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating procedure: {0}",
                    e);

                return ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Channel
        public bool IsStreamForChannel(
            string name,
            ChannelType channelType,
            Stream stream
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                string channelId = RuntimeOps.ChannelTypeToName(name, channelType);

                if ((channelId == null) || (stream == null))
                    return false;

                if (!HasChannels())
                    return false;

                Channel channel;

                if (!channels.TryGetValue(channelId, out channel))
                    return false;

                if (channel == null)
                    return false;

                ChannelStream channelStream = channel.GetStream();

                if (channelStream == null)
                    return false;

                Stream innerStream = channelStream.GetStream();

                return Object.ReferenceEquals(stream, innerStream);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ListChannels(
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasChannels(ref localError))
                {
                    code = channels.ToList(
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddChannel(
            string name,
            ChannelType channelType,
            Stream stream,
            OptionDictionary options,
            StreamFlags flags,
            StreamTranslation inTranslation,
            StreamTranslation outTranslation,
            Encoding encoding,
            bool appendMode,
            bool autoFlush,
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Ok;

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                string channelId = RuntimeOps.ChannelTypeToName(name, channelType);

                if (channelId == null)
                {
                    error = "invalid channel name";
                    return ReturnCode.Error;
                }

                if (!HasChannels(ref error))
                    return ReturnCode.Error;

                if (channels.ContainsKey(channelId))
                {
                    error = String.Format(
                        "can't add \"{0}\": channel already exists",
                        channelId);

                    return ReturnCode.Error;
                }

                ChannelStream channelStream = null;
                Channel channel = null;

                try
                {
                    channelStream = new ChannelStream(
                        stream, options, flags, inTranslation,
                        outTranslation);

                    channel = new Channel(
                        channelStream, encoding, appendMode,
                        autoFlush, clientData);

                    channels.Add(name, channel);

                    code = ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    code = ReturnCode.Error;
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        if (channelStream != null)
                        {
                            channelStream.PreventClose = false;
                            channelStream.Close();
                            channelStream = null;
                        }

                        if (channel != null)
                        {
                            channel.Close();
                            channel = null;
                        }
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveChannel(
            string name,
            ChannelType channelType,
            bool flush,
            bool close,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                string channelId = RuntimeOps.ChannelTypeToName(name, channelType);

                if (channelId == null)
                {
                    error = "invalid channel name";
                    return ReturnCode.Error;
                }

                if (!HasChannels(ref error))
                    return ReturnCode.Error;

                Channel channel;

                if (channels.TryGetValue(channelId, out channel))
                {
                    if (flush && (channel != null))
                    {
                        try
                        {
                            //
                            // NOTE: Attempt to flush the channel.
                            //
                            channel.Flush(); /* throw */
                        }
                        catch (Exception e)
                        {
                            if (strict)
                            {
                                error = e;
                                return ReturnCode.Error;
                            }
                        }
                    }

                    if (close && (channel != null))
                    {
                        try
                        {
                            //
                            // NOTE: Attempt to close the channel.
                            //
                            channel.Close(); /* throw */
                        }
                        catch (Exception e)
                        {
                            if (strict)
                            {
                                error = e;
                                return ReturnCode.Error;
                            }
                        }
                    }

                    channels.Remove(channelId);

                    return ReturnCode.Ok;
                }
                else
                {
                    error = String.Format(
                        "can not find channel named \"{0}\"",
                        channelId);

                    return ReturnCode.Error;
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICallFrame (Scope)
        public ReturnCode GetScope(
            string name,
            ref ICallFrame frame,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasScopes(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty scope names are technically allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    //       That being said, we actually block them for practical
                    //       reasons in the [scope] command itself.
                    //
                    if (name != null)
                    {
                        if (scopes.TryGetValue(name, out frame))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "scope \"{0}\" not found",
                                name);
                    }
                    else
                    {
                        error = "invalid scope name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool HasScopes()
        {
            Result error = null;

            return HasScopes(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetScope(
            string name,
            ref ICallFrame frame
            )
        {
            Result error = null;

            return GetScope(name, ref frame, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ListScopes(
            string pattern,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasScopes())
                    result = scopes.ToString(pattern, false);
                else
                    result = String.Empty;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddScope(
            ICallFrame frame,
            IClientData clientData,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (frame == null)
                {
                    result = "invalid scope";
                    return ReturnCode.Error;
                }

                if (frame.Name == null)
                {
                    result = "invalid scope name";
                    return ReturnCode.Error;
                }

                if (!HasScopes(ref result))
                    return ReturnCode.Error;

                if (DoesScopeExist(frame.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": scope already exists",
                        frame.Name);

                    return ReturnCode.Error;
                }

                scopes.Add(frame.Name, frame);
                result = String.Empty;
            }

#if NOTIFY
            //
            // NOTE: This must be done outside of the lock statement (above),
            //       because the code called may have arbitrary side-effects.
            //
            /* IGNORED */
            CheckNotifications(
                NotifyType.CallFrame, NotifyFlags.Added,
                frame, this, clientData, null, null,
                ref result);
#endif

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveScope(
            string name,
            IClientData clientData,
            ref Result result
            )
        {
            ICallFrame frame = null;

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (String.IsNullOrEmpty(name))
                {
                    result = "invalid scope name";
                    return ReturnCode.Error;
                }

                if (!HasScopes(ref result))
                    return ReturnCode.Error;

                if (GetScopeCallFrame(name, true, true, ref frame) == ReturnCode.Ok)
                {
                    //
                    // NOTE: We destroyed the current scope call frame (after
                    //       popping it).
                    //
                    result = String.Empty;
                    return ReturnCode.Ok;
                }
                else
                {
                    //
                    // NOTE: Make sure the scope exists and return an error if
                    //       it does not.
                    //
                    if (GetScope(name, ref frame, ref result) == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Remove the scope from the interpreter.
                        //
                        scopes.Remove(name);
                        result = String.Empty;
                    }
                    else
                    {
                        return ReturnCode.Error;
                    }
                }
            }

#if NOTIFY || NOTIFY_OBJECT
            //
            // NOTE: This must be done outside of the lock statement (above),
            //       because the code called may have arbitrary side-effects.
            //
            /* IGNORED */
            CheckNotifications(
                NotifyType.CallFrame, NotifyFlags.Deleted,
                frame, this, clientData, null, null,
                ref result);
#endif

            return ReturnCode.Ok;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region CommandCallback (CallbackQueue)
#if CALLBACK_QUEUE
        public ReturnCode ClearCallbackQueue(
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCallbackQueue(ref error))
                {
                    callbackQueue.Clear();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EnqueueCallback(
            ICallback callback,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCallbackQueue(ref error))
                {
                    CommandCallback commandCallback =
                        callback as CommandCallback;

                    if (commandCallback != null)
                    {
                        callbackQueue.Enqueue(commandCallback);

                        code = ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid command callback";
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DequeueCallback(
            ref ICallback callback,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasCallbackQueue(ref error))
                {
                    if (callbackQueue.Count > 0)
                    {
                        callback = callbackQueue.Dequeue();

                        code = ReturnCode.Ok;
                    }
                    else
                    {
                        error = "callback queue is empty";
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode DequeueCallback(
            ref CommandCallback callback
            )
        {
            ICallback localCallback = null;
            Result error = null;

            if (DequeueCallback(ref localCallback, ref error) == ReturnCode.Ok)
            {
                if (localCallback is CommandCallback)
                {
                    callback = (CommandCallback)localCallback;

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid command callback";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ExecuteCallbackQueue(
            ref Result result
            )
        {
            return Engine.ExecuteCallbackQueue(
                this, engineFlags, substitutionFlags, engineEventFlags,
                expressionFlags, ref result);
        }
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDbConnection / IDbTransaction
#if DATA
        public ReturnCode GetDbConnection(
            string name,
            ref IDbConnection connection,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasDbConnections(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty database connection names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (connections.TryGetValue(name, out connection))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "database connection \"{0}\" not found",
                                name);
                    }
                    else
                    {
                        error = "invalid database connection name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetDbTransaction(
            string name,
            ref IDbTransaction transaction,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasDbTransactions(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty database transaction names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (transactions.TryGetValue(name, out transaction))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "database transaction \"{0}\" not found",
                                name);
                    }
                    else
                    {
                        error = "invalid database transaction name";
                    }
                }
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        #region Hidden ICommand
        private bool HasHiddenCommands(ref Result error)
        {
            /* CheckDisposed(); */

            bool result = (hiddenCommands != null);

            if (!result)
                error = "hidden commands not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenCommand(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ICommand command,
            ref Result error
            )
        {
            if (HasHiddenCommands(ref error))
            {
                if (wrapper)
                {
                    command = hiddenCommands[token];
                    name = EntityOps.GetName(command as IIdentifier);

                    if (validate && (command == null))
                        error = "invalid command";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenCommands[token] != null)
                    {
                        command = hiddenCommands[token].command;
                        name = EntityOps.GetName(command as IIdentifier);

                        if (validate && (command == null))
                            error = "invalid command";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid command wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenCommand(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            if (HasHiddenCommands(ref error))
            {
                if (wrapper)
                {
                    command = hiddenCommands[name];
                    token = EntityOps.GetToken(command as IWrapperData);

                    if (validate && (command == null))
                        error = "invalid command";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenCommands[name] != null)
                    {
                        command = hiddenCommands[name].command;
                        token = EntityOps.GetToken(command as IWrapperData);

                        if (validate && (command == null))
                            error = "invalid command";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid command wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            long token,
            bool wrapper,
            bool validate,
            ref ICommand command
            )
        {
            Result error = null;

            return GetHiddenCommand(token, wrapper, validate, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            long token,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            string name = null;

            return GetHiddenCommand(token, wrapper, validate, ref name, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref ICommand command,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenCommands(ref error))
                {
                    if ((hiddenCommands.Count > 0) &&
                        hiddenCommands.ContainsKey(token))
                    {
                        code = PrivateGetHiddenCommand(token, wrapper, validate, ref name, ref command, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden command token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            string name,
            bool wrapper,
            bool validate,
            ref ICommand command
            )
        {
            Result error = null;

            return GetHiddenCommand(name, wrapper, validate, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            string name,
            bool wrapper,
            bool validate,
            ref ICommand command,
            ref Result error
            )
        {
            long token = 0;

            return GetHiddenCommand(name, wrapper, validate, ref token, ref command, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenCommand(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref ICommand command,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenCommands(ref error))
                {
                    if ((hiddenCommands.Count > 0) &&
                        (name != null) &&
                        hiddenCommands.ContainsKey(name))
                    {
                        code = PrivateGetHiddenCommand(name, wrapper, validate, ref token, ref command, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden command name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Hidden IProcedure
        private bool HasHiddenProcedures(ref Result error)
        {
            /* CheckDisposed(); */

            bool result = (hiddenProcedures != null);

            if (!result)
                error = "hidden procedures not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenProcedure(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IProcedure procedure,
            ref Result error
            )
        {
            if (HasHiddenProcedures(ref error))
            {
                if (wrapper)
                {
                    procedure = hiddenProcedures[token];
                    name = EntityOps.GetName(procedure as IIdentifier);

                    if (validate && (procedure == null))
                        error = "invalid procedure";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenProcedures[token] != null)
                    {
                        procedure = hiddenProcedures[token].procedure;
                        name = EntityOps.GetName(procedure as IIdentifier);

                        if (validate && (procedure == null))
                            error = "invalid procedure";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid procedure wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IProcedure procedure,
            ref Result error
            )
        {
            if (HasHiddenProcedures(ref error))
            {
                if (wrapper)
                {
                    procedure = hiddenProcedures[name];
                    token = EntityOps.GetToken(procedure as IWrapperData);

                    if (validate && (procedure == null))
                        error = "invalid procedure";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenProcedures[name] != null)
                    {
                        procedure = hiddenProcedures[name].procedure;
                        token = EntityOps.GetToken(procedure as IWrapperData);

                        if (validate && (procedure == null))
                            error = "invalid procedure";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid procedure wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenProcedure(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IProcedure procedure,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenProcedures(ref error))
                {
                    if ((hiddenProcedures.Count > 0) &&
                        hiddenProcedures.ContainsKey(token))
                    {
                        code = PrivateGetHiddenProcedure(token, wrapper, validate, ref name, ref procedure, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden procedure token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref IProcedure procedure
            )
        {
            Result error = null;

            return GetHiddenProcedure(name, wrapper, validate, ref procedure, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref IProcedure procedure,
            ref Result error
            )
        {
            long token = 0;

            return GetHiddenProcedure(name, wrapper, validate, ref token, ref procedure, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenProcedure(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IProcedure procedure,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenProcedures(ref error))
                {
                    if ((hiddenProcedures.Count > 0) &&
                        (name != null) &&
                        hiddenProcedures.ContainsKey(name))
                    {
                        code = PrivateGetHiddenProcedure(name, wrapper, validate, ref token, ref procedure, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden procedure name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Hidden IExecute
        private bool HasHiddenIExecutes(ref Result error)
        {
            /* CheckDisposed(); */

            bool result = (hiddenExecutes != null);

            if (!result)
                error = "hidden executes not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenIExecute(
            long token,
            bool wrapper,
            bool validate,
            ref string name, /* NOT USED */
            ref IExecute execute,
            ref Result error
            )
        {
            if (HasHiddenIExecutes(ref error))
            {
                if (wrapper)
                {
                    execute = hiddenExecutes[token];

                    if (validate && (execute == null))
                        error = "invalid execute";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenExecutes[token] != null)
                    {
                        execute = hiddenExecutes[token].execute;

                        if (validate && (execute == null))
                            error = "invalid execute";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid execute wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetHiddenIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            if (HasHiddenIExecutes(ref error))
            {
                if (wrapper)
                {
                    execute = hiddenExecutes[name];
                    token = EntityOps.GetToken(execute as IWrapperData);

                    if (validate && (execute == null))
                        error = "invalid execute";
                    else
                        return ReturnCode.Ok;
                }
                else
                {
                    if (hiddenExecutes[name] != null)
                    {
                        execute = hiddenExecutes[name].execute;

                        //
                        // WARNING: Cannot grab the token from the IExecute here
                        //          because that interface does not inherit from
                        //          IWrapperData.
                        //
                        token = EntityOps.GetToken(hiddenExecutes[name] as IWrapperData);

                        if (validate && (execute == null))
                            error = "invalid execute";
                        else
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "invalid execute wrapper";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenIExecute(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IExecute execute,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenIExecutes(ref error))
                {
                    if ((hiddenExecutes.Count > 0) &&
                        hiddenExecutes.ContainsKey(token))
                    {
                        code = PrivateGetHiddenIExecute(token, wrapper, validate, ref name, ref execute, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden execute token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref IExecute execute
            )
        {
            Result error = null;

            return GetHiddenIExecute(name, wrapper, validate, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref IExecute execute,
            ref Result error
            )
        {
            long token = 0;

            return GetHiddenIExecute(name, wrapper, validate, ref token, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetHiddenIExecute(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasHiddenIExecutes(ref error))
                {
                    if ((hiddenExecutes.Count > 0) &&
                        (name != null) &&
                        hiddenExecutes.ContainsKey(name))
                    {
                        code = PrivateGetHiddenIExecute(name, wrapper, validate, ref token, ref execute, ref error);
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid hidden execute name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IOperator
        private ReturnCode GetOperator(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IOperator @operator,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasOperators(ref error))
                {
                    if (operators.Count > 0)
                    {
                        foreach (KeyValuePair<string, _Wrappers.Operator> pair in operators)
                        {
                            if (pair.Value != null)
                            {
                                if (pair.Value.Token == token)
                                {
                                    name = pair.Key;

                                    if (wrapper)
                                        @operator = pair.Value;
                                    else
                                        @operator = pair.Value.@operator;

                                    if (!validate || (@operator != null))
                                        code = ReturnCode.Ok;

                                    break;
                                }
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid operator token \"{0}\"",
                            token);
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetOperator(
            string name,
            bool wrapper,
            bool validate,
            ref IOperator @operator
            )
        {
            Result error = null;

            return GetOperator(name, wrapper, validate, ref @operator, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetOperator(
            string name,
            bool wrapper,
            bool validate,
            ref IOperator @operator,
            ref Result error
            )
        {
            long token = 0;

            return GetOperator(name, wrapper, validate, ref token, ref @operator, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetOperator(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IOperator @operator,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                if (HasOperators(ref error))
                {
                    if ((operators.Count > 0) &&
                        (name != null) &&
                        operators.ContainsKey(name))
                    {
                        if (wrapper)
                        {
                            @operator = operators[name];
                            token = EntityOps.GetToken(@operator as IWrapperData);

                            if (!validate || (@operator != null))
                                code = ReturnCode.Ok;
                        }
                        else
                        {
                            if (operators[name] != null)
                            {
                                @operator = operators[name].@operator;
                                token = EntityOps.GetToken(@operator as IWrapperData);

                                if (!validate || (@operator != null))
                                    code = ReturnCode.Ok;
                            }
                        }
                    }

                    if (code != ReturnCode.Ok)
                        error = String.Format(
                            "invalid operator name \"{0}\"",
                            FormatOps.DisplayName(name));
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ListOperators(
            OperatorFlags hasFlags,
            OperatorFlags notHasFlags,
            bool hasAll,
            bool notHasAll,
            string pattern,
            bool noCase,
            bool strict,
            ref StringList list,
            ref Result error
            )
        {
            // CheckDisposed();

            ReturnCode code = ReturnCode.Error;

            lock (syncRoot)
            {
                Result localError = null;

                if (HasOperators(ref localError))
                {
                    code = operators.ToList(
                        hasFlags, notHasFlags, hasAll, notHasAll,
                        pattern, noCase, ref list, ref error);
                }
                else if (strict)
                {
                    error = localError;
                }
                else
                {
                    if (list == null)
                        list = new StringList();

                    code = ReturnCode.Ok;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOperator(
            Type type,
            string name,
            Lexeme lexeme,
            TypeList types,
            OperatorFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddOperator(
                type, name, lexeme, types, flags, plugin,
                clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOperator(
            Type type,
            string name,
            Lexeme lexeme,
            TypeList types,
            OperatorFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            return AddOperator(
                type, name, lexeme, (int)Arity.Automatic, types,
                flags, plugin, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOperator(
            Type type,
            string name,
            Lexeme lexeme,
            int operands,
            TypeList types,
            OperatorFlags flags,
            IPlugin plugin,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            // CheckDisposed(); /* NOTE: This API may be public someday. */

            if (type == null)
                type = Type.GetType(
                    FormatOps.OperatorTypeName(name), false, true);

            if (type == null)
                type = Type.GetType(
                    FormatOps.OperatorTypeName(Characters.Underscore + name), false, true);

            if (type != null)
            {
                try
                {
                    if (operands == (int)Arity.Automatic)
                        operands = AttributeOps.GetOperands(type);

                    object[] args = {
                        new OperatorData(
                            name, null, null, lexeme, operands,
                            types, flags, clientData, plugin, 0)
                    };

                    IOperator @operator = (IOperator)Activator.CreateInstance(type, args);

                    return AddOperator(@operator, null, ref token, ref result);
                }
                catch (Exception e)
                {
                    result = e;
                }
            }
            else
            {
                result = String.Format(
                    "operator \"{0}\" not found",
                    FormatOps.OperatorTypeName(name));
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddOperator(
            IOperator @operator,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (@operator == null)
                {
                    result = "invalid operator";
                    return ReturnCode.Error;
                }

                if (@operator.Name == null)
                {
                    result = "invalid operator name";
                    return ReturnCode.Error;
                }

                if (!HasOperators(ref result))
                    return ReturnCode.Error;

                if (DoesOperatorExist(@operator.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": operator already exists",
                        @operator.Name);

                    return ReturnCode.Error;
                }

                //
                // HACK: Due to how the expression parser works, we cannot allow an
                //       operator to be added that has the name of an existing
                //       function.
                //
                if (DoesFunctionExist(@operator.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": function already exists",
                        @operator.Name);

                    return ReturnCode.Error;
                }

                ReturnCode code = ReturnCode.Ok;

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    try
                    {
                        code = @operator.Initialize(this,
                            (clientData != null) ? clientData : @operator.ClientData,
                            ref result);
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while initializing operator: {0}",
                            e);

                        code = ReturnCode.Error;
                    }
                }
                finally
                {
                    if (code != ReturnCode.Ok)
                    {
                        ReturnCode terminateCode;
                        Result terminateResult = null;

                        try
                        {
                            terminateCode = @operator.Terminate(this,
                                (clientData != null) ? clientData : @operator.ClientData,
                                ref terminateResult);
                        }
                        catch (Exception e)
                        {
                            terminateResult = String.Format(
                                "caught exception while terminating operator: {0}",
                                e);

                            terminateCode = ReturnCode.Error;
                        }

                        if (terminateCode != ReturnCode.Ok)
                            DebugOps.Complain(this, terminateCode, terminateResult);
                    }
                }

                if (code != ReturnCode.Ok)
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                    return code;
                }

                long id = NextId();
                _Wrappers.Operator operatorWrapper = new _Wrappers.Operator(id, @operator);

                operators.Add(@operator.Name, operatorWrapper);
                token = operatorWrapper.Token;
                @operator.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Operator, NotifyFlags.Added,
                    new ObjectPair(@operator, token), this,
                    clientData, null, null, ref result);
#endif

                /* IGNORED */
                GlobalState.PopActiveInterpreter();
                return code;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IModule
#if NATIVE && LIBRARY
        internal ReturnCode GetModule(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IModule module,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasModules(ref error))
                {
                    _Wrappers._Module moduleWrapper = null;

                    if (modules.TryGetValue(token, out moduleWrapper))
                    {
                        name = EntityOps.GetName(moduleWrapper);

                        if (wrapper)
                            module = moduleWrapper;
                        else
                            module = (moduleWrapper != null) ?
                                moduleWrapper.module : null;

                        if (!validate || (module != null))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "invalid module token \"{0}\"",
                                token);
                    }
                    else
                    {
                        error = String.Format(
                            "module \"{0}\" not found",
                            token);
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetModule(
            string name,
            bool wrapper,
            bool validate,
            ref IModule module,
            ref Result error
            )
        {
            long token = 0;

            return GetModule(name, wrapper, validate, ref token, ref module, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetModule(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IModule module,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasModules(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty module names are technically allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        _Wrappers._Module moduleWrapper = null;

                        if (modules.TryGetValue(name, out moduleWrapper))
                        {
                            token = EntityOps.GetToken(moduleWrapper);

                            if (wrapper)
                                module = moduleWrapper;
                            else
                                module = (moduleWrapper != null) ?
                                    moduleWrapper.module : null;

                            if (!validate || (module != null))
                                return ReturnCode.Ok;
                            else
                                error = String.Format(
                                    "invalid module name \"{0}\"",
                                    FormatOps.DisplayName(name));
                        }
                        else
                        {
                            error = String.Format(
                                "module \"{0}\" not found",
                                name);
                        }
                    }
                    else
                    {
                        error = "invalid module name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddModule(
            IModule module,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddModule(module, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddModule(
            IModule module,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (module == null)
                {
                    result = "invalid module";
                    return ReturnCode.Error;
                }

                if (module.Name == null)
                {
                    result = "invalid module name";
                    return ReturnCode.Error;
                }

                if (!HasModules(ref result))
                    return ReturnCode.Error;

                if (DoesModuleExist(module.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": module already exists",
                        module.Name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers._Module moduleWrapper = new _Wrappers._Module(id, module);

                modules.Add(module.Name, moduleWrapper);
                token = moduleWrapper.Token;
                module.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Module, NotifyFlags.Added,
                    new ObjectPair(module, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveModule(
            string name,
            long token,
            IModule module,
            IClientData clientData,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            if (dispose)
            {
                if (module != null)
                {
                    if (!FlagOps.HasFlags(module.Flags, ModuleFlags.Locked, true))
                    {
                        GlobalState.PushActiveInterpreter(this);

                        try
                        {
                            if (RuntimeOps.TryDisposeObject(module,
                                    ref dispose, ref result) == ReturnCode.Ok)
                            {
                                module = null;

                                if (synchronous)
                                    CommonOps.CollectGarbage(null, null); /* throw */
                            }
                            else
                            {
                                return ReturnCode.Error;
                            }
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while disposing module: {0}",
                                e);

                            return ReturnCode.Error;
                        }
                        finally
                        {
                            /* IGNORED */
                            GlobalState.PopActiveInterpreter();
                        }
                    }
                    else
                    {
                        result = "cannot dispose or remove a locked module";

                        return ReturnCode.Error;
                    }
                }
            }

            modules.Remove(name);

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Module, NotifyFlags.Removed,
                new ObjectTriplet(module, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveModule(
            string name,
            IClientData clientData /* NOT USED */,
            bool synchronous,
            ref Result result
            )
        {
            bool dispose = true;

            return RemoveModule(name, clientData, synchronous, ref dispose, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveModule(
            string name,
            IClientData clientData /* NOT USED */,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasModules(ref result))
                    {
                        long token = 0;
                        IModule module = null;

                        if (GetModule(name, false, true, ref token,
                                ref module, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveModule(
                                name, token, module, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveModule(
            long token,
            IClientData clientData, /* NOT USED */
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasModules(ref result))
                    {
                        string name = null;
                        IModule module = null;

                        if (GetModule(token, false, true, ref name,
                                ref module, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveModule(
                                name, token, module, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDelegate
#if NATIVE && LIBRARY
        internal ReturnCode GetDelegate(
            long token,
            bool wrapper,
            bool validate,
            ref string name,
            ref IDelegate @delegate,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasDelegates(ref error))
                {
                    _Wrappers.Delegate delegateWrapper = null;

                    if (delegates.TryGetValue(token, out delegateWrapper))
                    {
                        name = EntityOps.GetName(delegateWrapper);

                        if (wrapper)
                            @delegate = delegateWrapper;
                        else
                            @delegate = (delegateWrapper != null) ?
                                delegateWrapper.@delegate : null;

                        if (!validate || (@delegate != null))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "invalid delegate token \"{0}\"",
                                token);
                    }
                    else
                    {
                        error = String.Format(
                            "delegate \"{0}\" not found",
                            token);
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetDelegate(
            string name,
            bool wrapper,
            bool validate,
            ref IDelegate @delegate,
            ref Result error
            )
        {
            long token = 0;

            return GetDelegate(name, wrapper, validate, ref token, ref @delegate, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetDelegate(
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IDelegate @delegate,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasDelegates(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty delegate names are technically allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        _Wrappers.Delegate delegateWrapper = null;

                        if (delegates.TryGetValue(name, out delegateWrapper))
                        {
                            token = EntityOps.GetToken(delegateWrapper);

                            if (wrapper)
                                @delegate = delegateWrapper;
                            else
                                @delegate = (delegateWrapper != null) ?
                                    delegateWrapper.@delegate : null;

                            if (!validate || (@delegate != null))
                                return ReturnCode.Ok;
                            else
                                error = String.Format(
                                    "invalid delegate name \"{0}\"",
                                    FormatOps.DisplayName(name));
                        }
                        else
                        {
                            error = String.Format(
                                "delegate \"{0}\" not found",
                                name);
                        }
                    }
                    else
                    {
                        error = "invalid delegate name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddDelegate(
            IDelegate @delegate,
            IClientData clientData,
            ref Result result
            )
        {
            long token = 0;

            return AddDelegate(@delegate, clientData, ref token, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddDelegate(
            IDelegate @delegate,
            IClientData clientData,
            ref long token,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (@delegate == null)
                {
                    result = "invalid delegate";
                    return ReturnCode.Error;
                }

                if (@delegate.Name == null)
                {
                    result = "invalid delegate name";
                    return ReturnCode.Error;
                }

                if (!HasDelegates(ref result))
                    return ReturnCode.Error;

                if (DoesDelegateExist(@delegate.Name) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": delegate already exists",
                        @delegate.Name);

                    return ReturnCode.Error;
                }

                long id = NextId();
                _Wrappers.Delegate delegateWrapper = new _Wrappers.Delegate(id, @delegate);

                delegates.Add(@delegate.Name, delegateWrapper);
                token = delegateWrapper.Token;
                @delegate.Token = token;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Delegate, NotifyFlags.Added,
                    new ObjectPair(@delegate, token), this,
                    clientData, null, null, ref result);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateRemoveDelegate(
            string name,
            long token,
            IDelegate @delegate,
            IClientData clientData,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            if (name == null)
            {
                result = "invalid delegate name";
                return ReturnCode.Error;
            }

            if (dispose)
            {
                if (@delegate != null)
                {
                    GlobalState.PushActiveInterpreter(this);

                    try
                    {
                        if (RuntimeOps.TryDisposeObject(@delegate,
                                ref dispose, ref result) == ReturnCode.Ok)
                        {
                            @delegate = null;

                            if (synchronous)
                                CommonOps.CollectGarbage(null, null); /* throw */
                        }
                        else
                        {
                            return ReturnCode.Error;
                        }
                    }
                    catch (Exception e)
                    {
                        result = String.Format(
                            "caught exception while disposing delegate: {0}",
                            e);

                        return ReturnCode.Error;
                    }
                    finally
                    {
                        /* IGNORED */
                        GlobalState.PopActiveInterpreter();
                    }
                }
            }

            delegates.Remove(name);

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Delegate, NotifyFlags.Removed,
                new ObjectTriplet(@delegate, name, token), this,
                clientData, null, null, ref result);
#endif

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveDelegate(
            string name,
            IClientData clientData /* NOT USED */,
            bool synchronous,
            ref Result result
            )
        {
            bool dispose = true;

            return RemoveDelegate(name, clientData, synchronous, ref dispose, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveDelegate(
            string name,
            IClientData clientData /* NOT USED */,
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasDelegates(ref result))
                    {
                        long token = 0;
                        IDelegate @delegate = null;

                        if (GetDelegate(name, false, true, ref token,
                                ref @delegate, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveDelegate(
                                name, token, @delegate, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveDelegate(
            long token,
            IClientData clientData, /* NOT USED */
            bool synchronous,
            ref bool dispose,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasDelegates(ref result))
                    {
                        string name = null;
                        IDelegate @delegate = null;

                        if (GetDelegate(token, false, true, ref name,
                                ref @delegate, ref result) == ReturnCode.Ok)
                        {
                            return PrivateRemoveDelegate(
                                name, token, @delegate, clientData, synchronous,
                                ref dispose, ref result);
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Encoding
        internal EncodingDictionary Encodings /* NOTE: For use by default host only. */
        {
            get
            {
                lock (syncRoot)
                {
                    return encodings;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static ReturnCode GetEncoding(
            Interpreter interpreter,
            string name,
            ref Encoding encoding
            )
        {
            Result error = null;

            return GetEncoding(interpreter, name, ref encoding, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetEncoding(
            Interpreter interpreter,
            string name,
            ref Encoding encoding,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                if (name != null)
                {
                    try
                    {
                        EncodingDictionary encodings = interpreter.Encodings;

                        if ((encodings == null) ||
                            (!encodings.TryGetValue(name, out encoding)))
                        {
                            //
                            // NOTE: Either our encodings are not available or we failed to
                            //       find the one they specified in our collection.  Try to
                            //       fallback to the .NET Framework encoding matching their
                            //       specified name.
                            //
                            encoding = Encoding.GetEncoding(name);
                        }

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                else
                {
                    encoding = null;

                    return ReturnCode.Ok;
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Channel
        internal bool HasChannels()
        {
            Result error = null;

            return HasChannels(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private Channel GetChannel(string name)
        {
            Result error = null;

            return GetChannel(name, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal Channel GetChannel(
            string name,
            ref Result error
            )
        {
            Channel result = null;

            if (name != null)
            {
                lock (syncRoot)
                {
                    if (channels != null)
                    {
                        if (channels.TryGetValue(name, out result))
                        {
                            if (result == null)
                                error = "invalid channel";
                        }
                        else
                        {
                            error = String.Format(
                                "can not find channel named \"{0}\"",
                                name);
                        }
                    }
                    else
                    {
                        error = "no channels available";
                    }
                }
            }
            else
            {
                error = "invalid channel name";
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetChannelEncoding(
            string name,
            ref Encoding encoding
            )
        {
            Channel channel = GetChannel(name);

            if (channel != null)
            {
                encoding = channel.GetEncoding();

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal Stream GetChannelStream(string name)
        {
            Stream result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.GetStream();

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Reader & Writer
        internal BinaryReader GetChannelBinaryReader(string name)
        {
            BinaryReader result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.GetBinaryReader();

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal BinaryWriter GetChannelBinaryWriter(string name)
        {
            BinaryWriter result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.GetBinaryWriter();

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal TextReader GetChannelTextReader(string name)
        {
            TextReader result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.GetStreamReader();

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal TextWriter GetChannelTextWriter(string name)
        {
            TextWriter result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.GetStreamWriter();

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Virtual Output
        internal StringBuilder GetChannelVirtualOutput(string name)
        {
            StringBuilder result = null;
            Channel channel = GetChannel(name);

            if (channel != null)
                result = channel.VirtualOutput;

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetChannelVirtualOutput(
            string name,
            ref Result result
            )
        {
            Channel channel = GetChannel(name, ref result);

            if (channel != null)
            {
                if (channel.VirtualOutput != null)
                {
                    result = channel.VirtualOutput;

                    return ReturnCode.Ok;
                }

                result = String.Format(
                    "virtual output is not enabled for channel named \"{0}\"",
                    name);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetChannelVirtualOutput(
            string name,
            bool enable,
            ref Result result
            )
        {
            Channel channel = GetChannel(name, ref result);

            if (channel != null)
            {
                if (enable)
                    channel.VirtualOutput = StringOps.NewStringBuilder();
                else
                    channel.VirtualOutput = null;

                result = String.Format(
                    "virtual output {0} for channel \"{1}\"",
                    ConversionOps.ToEnabled(enable), name);

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Adding Channels
        internal ReturnCode AddFileOrSocketChannel(
            string name,
            Stream stream,
            OptionDictionary options,
            StreamFlags flags,
            bool appendMode,
            bool autoFlush,
            IClientData clientData,
            ref Result error
            )
        {
            return AddChannel(
                name, ChannelType.None, stream, options, flags, StreamTranslation.auto,
                StreamTranslation.auto, Channel.DefaultEncoding, appendMode, autoFlush,
                clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NETWORK
        private ReturnCode AddTcpListenerChannel(
            string name,
            ChannelType channelType,
            TcpListener listener,
            SocketClientData clientData,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                string channelId = RuntimeOps.ChannelTypeToName(name, channelType);

                if (channelId == null)
                {
                    error = "invalid channel name";
                    return ReturnCode.Error;
                }

                if (!HasChannels(ref error))
                    return ReturnCode.Error;

                if (channels.ContainsKey(channelId))
                {
                    error = String.Format(
                        "can't add \"{0}\": channel already exists",
                        channelId);

                    return ReturnCode.Error;
                }

                OptionDictionary options = (clientData != null) ?
                    clientData.Options : null;

                channels.Add(channelId, new Channel(listener, options,
                    StreamFlags.ListenSocket, clientData));
            }

            return ReturnCode.Ok;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Standard Channels
        internal ReturnCode AddStandardChannels(
            IStreamHost streamHost,
            string name, /* NOTE: Cannot be non-null when adding multiple channels. */
            ChannelType channelType,
            bool replace,
            bool strict,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (channels != null)
                {
                    if (streamHost != null)
                    {
                        HostFlags hostFlags = streamHost.GetFlags();

                        try
                        {
                            string channelId;
                            bool contains;

                            if (FlagOps.HasFlags(channelType, ChannelType.Input, true))
                            {
                                channelId = (name != null) ? name : StandardChannel.Input;
                                contains = channels.ContainsKey(channelId);

                                if (!contains || replace)
                                {
                                    //
                                    // HACK: We have no clean way to determine if the stdin stream
                                    //       has additional data to read; therefore, we need to
                                    //       allow any end-of-line character to terminate an input
                                    //       line or we can run into problems on wherever the
                                    //       end-of-line sequence differs.
                                    //
                                    StreamFlags streamFlags = PlatformOps.IsWindowsOperatingSystem() ?
                                        StreamFlags.None : StreamFlags.UseAnyEndOfLineChar;

                                    Channel channel = new Channel(
                                        streamHost.DefaultIn, null, streamFlags, StreamTranslation.auto,
                                        StreamTranslation.auto, streamHost.InputEncoding, false,
                                        false, null);

                                    if (contains)
                                    {
                                        if (channels[channelId] != null)
                                            channels[channelId].Dispose();

                                        channels[channelId] = channel;
                                    }
                                    else
                                    {
                                        channels.Add(channelId, channel);
                                    }
                                }
                                else if (strict)
                                {
                                    error = String.Format(
                                        "can't add \"{0}\": channel already exists",
                                        channelId);

                                    return ReturnCode.Error;
                                }
                            }

                            if (FlagOps.HasFlags(channelType, ChannelType.Output, true))
                            {
                                channelId = (name != null) ? name : StandardChannel.Output;
                                contains = channels.ContainsKey(channelId);

                                if (!contains || replace)
                                {
                                    Channel channel = new Channel(
                                        streamHost.DefaultOut, null, StreamFlags.None, StreamTranslation.auto,
                                        StreamTranslation.auto, streamHost.OutputEncoding, false,
                                        FlagOps.HasFlags(hostFlags, HostFlags.AutoFlushOutput, true), null);

                                    if (contains)
                                    {
                                        if (channels[channelId] != null)
                                            channels[channelId].Dispose();

                                        channels[channelId] = channel;
                                    }
                                    else
                                    {
                                        channels.Add(channelId, channel);
                                    }
                                }
                                else if (strict)
                                {
                                    error = String.Format(
                                        "can't add \"{0}\": channel already exists",
                                        channelId);

                                    return ReturnCode.Error;
                                }
                            }

                            if (FlagOps.HasFlags(channelType, ChannelType.Error, true))
                            {
                                channelId = (name != null) ? name : StandardChannel.Error;
                                contains = channels.ContainsKey(channelId);

                                if (!contains || replace)
                                {
                                    Channel channel = new Channel(
                                        streamHost.DefaultError, null, StreamFlags.None, StreamTranslation.auto,
                                        StreamTranslation.auto, streamHost.ErrorEncoding, false,
                                        FlagOps.HasFlags(hostFlags, HostFlags.AutoFlushError, true), null);

                                    if (contains)
                                    {
                                        if (channels[channelId] != null)
                                            channels[channelId].Dispose();

                                        channels[channelId] = channel;
                                    }
                                    else
                                    {
                                        channels.Add(channelId, channel);
                                    }
                                }
                                else if (strict)
                                {
                                    error = String.Format(
                                        "can't add \"{0}\": channel already exists",
                                        channelId);

                                    return ReturnCode.Error;
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        catch (Exception e)
                        {
                            error = e;
                        }
                    }
                    else
                    {
                        error = "interpreter host not available";
                    }
                }
                else
                {
                    error = "no channels available";
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region AppDomain
#if APPDOMAINS
        private ReturnCode AddAppDomain(
            string friendlyName,
            string packagePath,
#if CAS_POLICY
            Evidence evidence,
#endif
            IClientData clientData,
            bool useBasePath,
            ref AppDomain appDomain,
            ref Result error
            )
        {
            //
            // NOTE: *WARNING* Empty application domain names are allowed,
            //       please do not change this to "!String.IsNullOrEmpty".
            //
            if (friendlyName != null)
            {
                lock (syncRoot)
                {
                    if (appDomains != null)
                    {
                        if (RuntimeOps.CreateAppDomain(
                            friendlyName, packagePath,
#if CAS_POLICY
                            evidence,
#endif
                            clientData, useBasePath, ref appDomain,
                            ref error) == ReturnCode.Ok)
                        {
                            appDomains.Add(friendlyName, appDomain);

#if NOTIFY
                            /* IGNORED */
                            CheckNotifications(
                                NotifyType.AppDomain, NotifyFlags.Added,
                                new ObjectPair(friendlyName, appDomain), this,
                                clientData, null, null, ref error);
#endif

                            return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        error = "application domains not available";
                    }
                }
            }
            else
            {
                error = "invalid friendly name";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode RemoveAppDomain(
            string friendlyName,
            IClientData clientData,
            ref Result error
            )
        {
            //
            // NOTE: *WARNING* Empty application domain names are allowed,
            //       please do not change this to "!String.IsNullOrEmpty".
            //
            if (friendlyName != null)
            {
                lock (syncRoot)
                {
                    if (appDomains != null)
                    {
                        if (appDomains.ContainsKey(friendlyName))
                        {
                            AppDomain appDomain = appDomains[friendlyName];

                            if (RuntimeOps.UnloadAppDomain(appDomain,
                                    clientData, ref error) == ReturnCode.Ok)
                            {
                                appDomains.Remove(friendlyName);

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.AppDomain, NotifyFlags.Removed,
                                    new ObjectPair(friendlyName, appDomain), this,
                                    clientData, null, null, ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "invalid application domain name \"{0}\"",
                                FormatOps.DisplayName(friendlyName));
                        }
                    }
                    else
                    {
                        error = "application domains not available";
                    }
                }
            }
            else
            {
                error = "invalid friendly name";
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region CLR Integration
        //
        // NOTE: For System.Type short-name to long-name lookups (i.e. TypeDefs).
        //
        internal StringDictionary ObjectTypes
        {
            get
            {
                lock (syncRoot)
                {
                    return objectTypes;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: For System.Type objects, not related to Tcl namespaces.
        //
        internal StringDictionary ObjectNamespaces
        {
            get
            {
                lock (syncRoot)
                {
                    return objectNamespaces;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: For COM interop assemblies, we keep track of interfaces.
        //
        internal TypePairDictionary<string, long> ObjectInterfaces
        {
            get
            {
                lock (syncRoot)
                {
                    return objectInterfaces;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool HasObjectTypes(ref Result error)
        {
            bool result = (objectTypes != null);

            if (!result)
                error = "object types not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool HasObjectNamespaces(ref Result error)
        {
            bool result = (objectNamespaces != null);

            if (!result)
                error = "object namespaces not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool HasObjectInterfaces(ref Result error)
        {
            bool result = (objectInterfaces != null);

            if (!result)
                error = "interfaces not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectTypes(
            StringList list,
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectTypes(ref error))
                    {
                        if (list != null)
                        {
                            if ((list.Count % 2) == 0)
                            {
                                for (int index = 0; index < list.Count; index += 2)
                                {
                                    string key = list[index];

                                    if (key != null) // empty string allowed.
                                    {
                                        string value = list[index + 1];

                                        if ((mode == MatchMode.None) ||
                                            StringOps.Match(mode, key, pattern, noCase))
                                        {
                                            if (objectTypes.ContainsKey(key))
                                                objectTypes[key] = value;
                                            else
                                                objectTypes.Add(key, value);
                                        }
                                    }
                                }

                                return ReturnCode.Ok;
                            }
                            else
                            {
                                error = "type list must have an even number of elements";
                            }
                        }
                        else
                        {
                            error = "invalid type list";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveObjectTypes(
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectTypes(ref error))
                    {
                        StringList keys = new StringList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the types dictionary in the containing interpreter.
                        //
                        foreach (KeyValuePair<string, string> pair in objectTypes)
                        {
                            string type = pair.Key;

                            if ((mode == MatchMode.None) ||
                                StringOps.Match(mode, type, pattern, noCase))
                            {
                                keys.Add(type);
                            }
                        }

                        //
                        // NOTE: Cleanup, PASS #2, and remove all the types we found from
                        //       PASS #1 (above).
                        //
                        foreach (string key in keys)
                            objectTypes.Remove(key);

                        //
                        // NOTE: Ok, we were successful.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectNamespaces(
            StringDictionary dictionary,
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectNamespaces(ref error))
                    {
                        if (dictionary != null)
                        {
                            foreach (KeyValuePair<string, string> pair in dictionary)
                            {
                                string key = pair.Key;

                                if ((mode == MatchMode.None) ||
                                    StringOps.Match(mode, key, pattern, noCase))
                                {
                                    string value = pair.Value;

                                    if (!objectNamespaces.ContainsKey(key))
                                        objectNamespaces.Add(key, value);
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "invalid namespace list";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectNamespaces(
            Assembly assembly,
            bool nonPublic,
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectNamespaces(ref error))
                    {
                        if (assembly != null)
                        {
                            TypeList types = new TypeList(assembly.GetTypes());

                            foreach (Type type in types)
                            {
                                if ((type != null) &&
                                    (nonPublic || type.IsPublic) &&
                                    !String.IsNullOrEmpty(type.Namespace))
                                {
                                    if ((mode == MatchMode.None) ||
                                        StringOps.Match(mode, type.Namespace, pattern, noCase))
                                    {
                                        if (!objectNamespaces.ContainsKey(type.Namespace))
                                        {
                                            objectNamespaces.Add(type.Namespace,
                                                AssemblyOps.GetFullName(type));
                                        }
                                    }
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "invalid assembly";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveObjectNamespaces(
            MatchMode mode,
            string pattern,
            bool noCase,
            bool values,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectNamespaces(ref error))
                    {
                        StringList keys = new StringList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the namespaces collection in the containing interpreter.
                        //
                        foreach (KeyValuePair<string, string> pair in objectNamespaces)
                        {
                            string @namespace = pair.Key;

                            if (values)
                            {
                                string assemblyName = pair.Value;

                                if ((mode == MatchMode.None) ||
                                    StringOps.Match(mode, assemblyName, pattern, noCase))
                                {
                                    keys.Add(@namespace);
                                }
                            }
                            else
                            {
                                if ((mode == MatchMode.None) ||
                                    StringOps.Match(mode, @namespace, pattern, noCase))
                                {
                                    keys.Add(@namespace);
                                }
                            }
                        }

                        //
                        // NOTE: Cleanup, PASS #2, and remove all the namespaces we found from
                        //       PASS #1 (above).
                        //
                        foreach (string key in keys)
                            objectNamespaces.Remove(key);

                        //
                        // NOTE: Ok, we were successful.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectInterfaces(
            TypeList types,
            bool nonPublic,
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectInterfaces(ref error))
                    {
                        if (types != null)
                        {
                            foreach (Type type in types)
                            {
                                if ((type != null) &&
                                    type.IsInterface &&
                                    (nonPublic || type.IsPublic))
                                {
                                    if ((mode == MatchMode.None) ||
                                        StringOps.Match(mode, type.FullName, pattern, noCase))
                                    {
                                        if (!objectInterfaces.ContainsKey(type))
                                        {
                                            objectInterfaces.Add(type,
                                                new AnyPair<string, long>(
                                                    AssemblyOps.GetFullName(type),
                                                    NextId()));
                                        }
                                    }
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "invalid interface list";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddObjectInterfaces(
            Assembly assembly,
            bool nonPublic,
            MatchMode mode,
            string pattern,
            bool noCase,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectInterfaces(ref error))
                    {
                        if (assembly != null)
                        {
                            TypeList types = new TypeList(assembly.GetTypes());

                            return AddObjectInterfaces(
                                types, nonPublic, mode, pattern, noCase, ref error);
                        }
                        else
                        {
                            error = "invalid assembly";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode RemoveObjectInterfaces(
            MatchMode mode,
            string pattern,
            bool noCase,
            bool values,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasObjectInterfaces(ref error))
                    {
                        TypeList keys = new TypeList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the interfaces collection in the containing interpreter.
                        //
                        foreach (KeyValuePair<Type, IAnyPair<string, long>> pair in objectInterfaces)
                        {
                            Type @interface = pair.Key;

                            if (@interface != null)
                            {
                                if (values)
                                {
                                    IAnyPair<string, long> anyPair = pair.Value;

                                    if (anyPair != null)
                                    {
                                        string assemblyName = anyPair.X;

                                        if ((mode == MatchMode.None) ||
                                            StringOps.Match(mode, assemblyName, pattern, noCase))
                                        {
                                            keys.Add(@interface);
                                        }
                                    }
                                }
                                else
                                {
                                    if ((mode == MatchMode.None) ||
                                        StringOps.Match(mode, @interface.FullName, pattern, noCase))
                                    {
                                        keys.Add(@interface);
                                    }
                                }
                            }
                        }

                        //
                        // NOTE: Cleanup, PASS #2, and remove all the interfaces we found from
                        //       PASS #1 (above).
                        //
                        foreach (Type key in keys)
                            objectInterfaces.Remove(key);

                        //
                        // NOTE: Ok, we were successful.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region INamespace
        public static bool IsPendingCleanup(
            Interpreter interpreter
            )
        {
            return (interpreter != null) ?
                IsPendingCleanup(interpreter.InterpreterFlags) : false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static bool IsPendingCleanup(
            InterpreterFlags interpreterFlags
            )
        {
            return FlagOps.HasFlags(
                interpreterFlags, InterpreterFlags.Cleanup, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void NotPendingCleanup()
        {
            //
            // NOTE: Remove the pending cleanup flag.
            //
            interpreterFlags &= ~InterpreterFlags.Cleanup;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DeleteNamespace(
            string name,
            bool force,
            ref Result result
            ) /* NOTE: Currently, only deletes the "global" namespace. */
        {
            ReturnCode code = ReturnCode.Ok;

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (RuntimeOps.IsGlobalNamespace(name))
                    {
                        if (force || (Levels == 0))
                        {
                            //
                            // NOTE: Clear all the IExecutes now.
                            //
                            if (code == ReturnCode.Ok)
                            {
                                if (executes != null)
                                {
                                    //
                                    // BUGBUG: IExecute objects have no way to belong
                                    //         to a namespace; therefore, we assume
                                    //         they are all global.
                                    //
                                    StringList keys = new StringList(executes.Keys);

                                    foreach (string key in keys)
                                        executes.Remove(key);

#if EXECUTE_CACHE
                                    //
                                    // BUGFIX: Everything may have just gone away,
                                    //         clear the cache now.
                                    //
                                    if (executeCache != null)
                                        executeCache.Clear();
#endif
                                }
                            }

                            //
                            // NOTE: Terminate and remove all procedures now.
                            //
                            if (code == ReturnCode.Ok)
                            {
                                if (procedures != null)
                                {
                                    StringList keys = new StringList(procedures.Keys);

                                    foreach (string key in keys)
                                    {
                                        IProcedure procedure = procedures[key];

                                        if (procedure != null)
                                        {
                                            //
                                            // BUGFIX: Skip deleting read-only procedures.
                                            //
                                            if (!EntityOps.IsReadOnly(procedure))
                                                procedures.Remove(key);
                                        }
                                        else
                                        {
                                            procedures.Remove(key);
                                        }
                                    }

#if EXECUTE_CACHE
                                    //
                                    // BUGFIX: Everything may have just gone away,
                                    //         clear the cache now.
                                    //
                                    if (executeCache != null)
                                        executeCache.Clear();
#endif
                                }
                            }

                            //
                            // NOTE: Terminate and remove all commands now.
                            //
                            if (code == ReturnCode.Ok)
                            {
                                if (commands != null)
                                {
                                    StringList keys = new StringList(commands.Keys);

                                    foreach (string key in keys)
                                    {
                                        ICommand command = commands[key];

                                        if (command != null)
                                        {
                                            //
                                            // BUGFIX: Skip deleting read-only commands.
                                            //
                                            if (!EntityOps.IsReadOnly(command))
                                            {
                                                GlobalState.PushActiveInterpreter(this);

                                                try
                                                {
                                                    code = command.Terminate(this, null, ref result);
                                                }
                                                catch (Exception e)
                                                {
                                                    result = String.Format(
                                                        "caught exception while terminating command: {0}",
                                                        e);

                                                    code = ReturnCode.Error;
                                                }
                                                finally
                                                {
                                                    /* IGNORED */
                                                    GlobalState.PopActiveInterpreter();
                                                }

                                                if (code == ReturnCode.Ok)
                                                {
                                                    //
                                                    // NOTE: Command successfully terminated, remove it.
                                                    //
                                                    commands.Remove(key);
                                                }
                                                else
                                                {
                                                    //
                                                    // NOTE: Failure, halt deletion process.
                                                    //
                                                    break;
                                                }
                                            }
                                        }
                                        else
                                        {
                                            commands.Remove(key);
                                        }
                                    }

#if EXECUTE_CACHE
                                    //
                                    // BUGFIX: Everything may have just gone away,
                                    //         clear the cache now.
                                    //
                                    if (executeCache != null)
                                        executeCache.Clear();
#endif
                                }
                            }

                            //
                            // NOTE: Finally, delete all global variables now by marking
                            //       them as deleted.
                            //
                            if (code == ReturnCode.Ok)
                            {
                                lock (syncRoot)
                                {
                                    if ((GlobalFrame != null) && (GlobalFrame.Variables != null))
                                    {
                                        foreach (KeyValuePair<string, IVariable> pair in GlobalFrame.Variables)
                                        {
                                            if (pair.Value != null)
                                            {
                                                //
                                                // BUGFIX: Skip deleting read-only (and system) variables.
                                                //
                                                if (!EntityOps.IsReadOnly(pair.Value) && !EntityOps.IsSystem(pair.Value))
                                                {
                                                    //
                                                    // NOTE: The variable is now dead.
                                                    //
                                                    EntityOps.SetUndefined(pair.Value, true);

                                                    //
                                                    // NOTE: Remove any pending array searches that may exist
                                                    //       for this variable.
                                                    //
                                                    CleanupArraySearches(pair.Value);

                                                    //
                                                    // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                                    //         modifications have been completed.
                                                    //
                                                    EntityOps.SetDirty(pair.Value, true);
                                                }
                                            }
                                        }

                                        CallFrameOps.Cleanup(CurrentFrame, GlobalFrame, false);
                                    }
                                }
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Mark the global namespace as requiring cleanup after
                            //       the current evaluation stack unwinds.
                            //
                            interpreterFlags |= InterpreterFlags.Cleanup;
                        }
                    }
                    else
                    {
                        result = String.Format(
                            "unknown namespace \"{0}\" in namespace delete command",
                            FormatOps.DisplayName(name));

                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                    code = ReturnCode.Error;
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Slave Interpreter
        public bool HasSlaveInterpreters(ref Result error)
        {
            CheckDisposed();

            bool result = (slaveInterpreters != null);

            if (!result)
                error = "slave interpreters not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesSlaveInterpreterExist(string path)
        {
            CheckDisposed();

            return DoesSlaveInterpreterExist(path, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetSlaveInterpreter(
            string path,
            bool nested,
            bool validate,
            bool create,
            ref Interpreter interpreter,
            ref string name,
            ref Result error
            )
        {
            CheckDisposed();

            if (nested)
                return GetNestedSlaveInterpreter(path, validate, create,
                    ref interpreter, ref name, ref error);
            else
                return GetSlaveInterpreter(path, validate, ref interpreter,
                    ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddSlaveInterpreter(
            string name,
            Interpreter interpreter,
            IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (interpreter == null)
                {
                    error = "invalid interpreter";
                    return ReturnCode.Error;
                }

                if (name == null)
                {
                    error = "invalid interpreter name";
                    return ReturnCode.Error;
                }

                if (!HasSlaveInterpreters(ref error))
                    return ReturnCode.Error;

                if (Object.ReferenceEquals(interpreter, this))
                {
                    error = String.Format(
                        "can't add \"{0}\": same as current interpreter",
                        name);

                    return ReturnCode.Error;
                }

                if (DoesSlaveInterpreterExist(name, false) == ReturnCode.Ok)
                {
                    error = String.Format(
                        "can't add \"{0}\": slave interpreter already exists",
                        name);

                    return ReturnCode.Error;
                }

                slaveInterpreters.Add(name, interpreter);

#if NOTIFY && NOTIFY_GLOBAL
                if (this.GlobalNotify)
                    /* IGNORED */
                    CheckNotifications(
                        null, false,
                        NotifyType.Interpreter, NotifyFlags.Added,
                        //
                        // BUGBUG: In order to use this class for notification
                        //         parameters, it really should probably be
                        //         made public.
                        //
                        new ObjectList(name), interpreter,
                        clientData, null, null, ref error);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveSlaveInterpreter(
            string name,
            IClientData clientData,
            bool synchronous,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasSlaveInterpreters(ref error))
                    {
                        Interpreter interpreter = null;

                        if (GetSlaveInterpreter(name, true, ref interpreter, ref error) == ReturnCode.Ok)
                        {
                            if (interpreter != null)
                            {
                                if (!Object.ReferenceEquals(interpreter, this))
                                {
                                    GlobalState.PushActiveInterpreter(this);

                                    try
                                    {
                                        if (RuntimeOps.TryDisposeObject(interpreter, ref error) == ReturnCode.Ok)
                                        {
                                            interpreter = null;

                                            if (synchronous)
                                                CommonOps.CollectGarbage(null, null); /* throw */
                                        }
                                        else
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    catch (Exception e)
                                    {
                                        error = String.Format(
                                            "caught exception while disposing slave interpreter: {0}",
                                            e);

                                        return ReturnCode.Error;
                                    }
                                    finally
                                    {
                                        /* IGNORED */
                                        GlobalState.PopActiveInterpreter();
                                    }
                                }
                            }

                            slaveInterpreters.Remove(name);

#if NOTIFY && NOTIFY_GLOBAL
                            if (this.GlobalNotify)
                                /* IGNORED */
                                CheckNotifications(
                                    null, false,
                                    NotifyType.Interpreter, NotifyFlags.Removed,
                                    //
                                    // BUGBUG: In order to use this class for notification
                                    //         parameters, it really should probably be
                                    //         made public.
                                    //
                                    new ObjectList(name), interpreter,
                                    clientData, null, null, ref error);
#endif

                            return ReturnCode.Ok;
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private bool IsMasterInterpreter(
            Interpreter interpreter
            )
        {
            lock (syncRoot)
            {
                return (interpreter != null) &&
                    Object.ReferenceEquals(interpreter, masterInterpreter);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool IsSlaveInterpreter(
            Interpreter interpreter
            )
        {
            lock (syncRoot)
            {
                return (interpreter != null) && (slaveInterpreters != null) &&
                    slaveInterpreters.ContainsValue(interpreter);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool IsMasterOrSlaveInterpreter(
            object value
            )
        {
            Interpreter interpreter = value as Interpreter;

            return (interpreter != null) ?
                IsMasterInterpreter(interpreter) || IsSlaveInterpreter(interpreter) : false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DoesSlaveInterpreterExist(
            string path,
            bool nested
            )
        {
            string name = null;

            return DoesSlaveInterpreterExist(path, nested, ref name);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DoesSlaveInterpreterExist(
            string path,
            bool nested,
            ref string name
            )
        {
            Interpreter interpreter = null;
            Result error = null;

            return GetSlaveInterpreter(path, nested, true, false,
                ref interpreter, ref name, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetInterpreterPath(
            Interpreter interpreter,
            ref Result result
            )
        {
            if (interpreter == null)
            {
                result = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (Object.ReferenceEquals(interpreter, this))
            {
                result = String.Empty;
                return ReturnCode.Ok;
            }

            Interpreter otherInterpreter = interpreter;
            StringList list = new StringList();

            while (!Object.ReferenceEquals(otherInterpreter, this))
            {
                string slaveName = otherInterpreter.slaveName;

                if (slaveName == null)
                    slaveName = otherInterpreter.ToString();

                //
                // NOTE: Keep track of the names going up the chain.
                //
                list.Add(slaveName);

                //
                // NOTE: Traverse to our master interpreter.
                //
                otherInterpreter = otherInterpreter.masterInterpreter;

                //
                // NOTE: Make sure the interpreter is still valid
                //       because we still haven't found this
                //       interpreter yet.
                //
                if (otherInterpreter == null)
                {
                    //
                    // NOTE: This should not really happen.  If so,
                    //       it is an error.
                    //
                    result = "invalid master interpreter";
                    return ReturnCode.Error;
                }
            }

            list.Reverse();
            result = list;

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetSlaveInterpreter(
            string path,
            bool nested,
            bool validate,
            bool create,
            ref Interpreter interpreter,
            ref Result error
            )
        {
            string name = null;

            return GetSlaveInterpreter(path, nested, validate, create,
                ref interpreter, ref name, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetSlaveInterpreter(
            string name,
            bool validate,
            ref Interpreter interpreter,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasSlaveInterpreters(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty interpreter names are technically allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (name.Length > 0)
                        {
                            if (slaveInterpreters.ContainsKey(name))
                            {
                                interpreter = slaveInterpreters[name];

                                if (!validate || (interpreter != null))
                                    return ReturnCode.Ok;
                                else
                                    error = String.Format(
                                        "invalid interpreter name \"{0}\"",
                                        FormatOps.DisplayName(name));
                            }
                            else
                            {
                                error = String.Format(
                                    "could not find interpreter \"{0}\"",
                                    FormatOps.DisplayName(name));
                            }
                        }
                        else
                        {
                            //
                            // NOTE: The empty string always represents the current
                            //       interpreter.
                            //
                            interpreter = this;

                            return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        error = "invalid interpreter name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetNestedSlaveInterpreter(
            string path,
            bool validate,
            bool create,
            ref Interpreter interpreter, /* NOTE: Only set upon success. */
            ref Result error             /* NOTE: Only set upon failure. */
            )
        {
            string name = null;

            return GetNestedSlaveInterpreter(path, validate, create,
                ref interpreter, ref name, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetNestedSlaveInterpreter(
            string path,                 /* in */
            bool validate,               /* in */
            bool create,                 /* in */
            ref Interpreter interpreter, /* out */
            ref string name,             /* out */
            ref Result error             /* out */
            )
        {
            StringList list = null;

            if (Parser.SplitList(
                    this, path, 0, Length.Invalid, false,
                    true, ref list, ref error) == ReturnCode.Ok)
            {
                Interpreter otherInterpreter = this;

                if (list.Count > 0)
                {
                    for (int index = 0; index < list.Count - 1; index++)
                    {
                        if (otherInterpreter.GetSlaveInterpreter(list[index], true,
                                ref otherInterpreter, ref error) != ReturnCode.Ok)
                        {
                            return ReturnCode.Error;
                        }
                    }

                    //
                    // NOTE: This will be the name of the new interpreter
                    //       to be created in the target interpreter if the
                    //       "create" flag is true; otherwise, it will simply
                    //       be the name of the innertmost nested interpreter.
                    //
                    name = list[list.Count - 1];

                    if (!create)
                        //
                        // NOTE: Return the last interpreter in the "chain";
                        //       this one will be validated (i.e. checked for
                        //       null) only if the caller has requested it.
                        //
                        return otherInterpreter.GetSlaveInterpreter(
                            name, validate, ref interpreter, ref error);
                }

                interpreter = otherInterpreter;
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CreateSlaveInterpreter(
            string path,
            IClientData clientData,
            CreateFlags createFlags,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;
            Interpreter interpreter = null;
            string name = null;

            if (path != null)
                code = GetNestedSlaveInterpreter(
                    path, true, true, ref interpreter, ref name, ref result);
            else
                interpreter = this;

            if (code == ReturnCode.Ok)
            {
                Interpreter otherInterpreter = Create(null,
                    createFlags & ~CreateFlags.ThrowOnError, ref result);

                try
                {
                    if (otherInterpreter != null)
                    {
                        string otherId = otherInterpreter.ToString();

                        if (name == null)
                            name = otherId;

                        code = interpreter.AddSlaveInterpreter(
                            name, otherInterpreter, clientData, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Set the master interpreter for the newly
                            //       created interpreter.
                            //
                            otherInterpreter.masterInterpreter = interpreter;

                            //
                            // NOTE: We need a way to find this new interpreter
                            //       by the arbitrary name assigned to it by the
                            //       caller, not just the automatic Id; therefore,
                            //       store that now.
                            //
                            if (String.Compare(name, otherId,
                                    StringOps.SystemStringComparisonType) != 0)
                            {
                                otherInterpreter.slaveName = name;
                            }

                            //
                            // NOTE: Return the interpreter path if it is valid;
                            //       otherwise, return the generated interpreter
                            //       name.
                            //
                            result = (path != null) ? path : name;
                        }
                    }
                    else
                    {
                        code = ReturnCode.Error;
                    }
                }
                catch (Exception e)
                {
                    result = e;
                    code = ReturnCode.Error;
                }
                finally
                {
                    if ((code != ReturnCode.Ok) &&
                        (otherInterpreter != null))
                    {
                        otherInterpreter.Dispose();
                        otherInterpreter = null;
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DeleteSlaveInterpreter(
            string path,
            IClientData clientData,
            bool synchronous,
            ref Result error
            )
        {
            Interpreter interpreter = null;
            string name = null;

            if (GetNestedSlaveInterpreter(path, true, true, ref interpreter,
                    ref name, ref error) == ReturnCode.Ok)
            {
                return interpreter.RemoveSlaveInterpreter(name, clientData,
                    synchronous, ref error);
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IErrorManager Members
        public int ErrorLine
        {
            get
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ErrorLine;
                else
                    return 0;
#else
                lock (syncRoot)
                {
                    return errorLine;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ErrorLine = value;
#else
                lock (syncRoot)
                {
                    errorLine = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CopyErrorInformation(
            VariableFlags flags,
            ref Result result
            )
        {
            CheckDisposed();

            return CopyErrorInformation(flags, false, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CopyErrorInformation(
            VariableFlags flags,
            bool strict,
            ref Result result
            )
        {
            CheckDisposed();

            ReturnCode code;
            Result errorCode = null;
            Result errorInfo = null;
            ResultList errors = null;

            code = CopyErrorInformation(
                flags, strict, ref errorCode, ref errorInfo,
                ref errors);

            if (code == ReturnCode.Ok)
            {
                if (result == null)
                    result = String.Empty;

                result.ErrorCode = errorCode;
                result.ErrorInfo = errorInfo;
            }
            else
            {
                if (errors != null)
                    result = errors.ToString();
                else
                    result = "could not copy error information";
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CopyErrorInformation(
            VariableFlags flags,
            bool strict,
            ref Result errorCode,
            ref Result errorInfo
            )
        {
            CheckDisposed();

            ResultList errors = null;

            return CopyErrorInformation(flags, strict,
                ref errorCode, ref errorInfo, ref errors);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CopyErrorInformation(
            VariableFlags flags,
            bool strict,
            ref Result errorCode,
            ref Result errorInfo,
            ref ResultList errors
            )
        {
            CheckDisposed();

            //
            // NOTE: For now, only allow flags related to disabling watches,
            //       traces, and notifications, etc.
            //
            flags &= VariableFlags.Fast;

            //
            // NOTE: Keep track of any errors encountered while querying the
            //       error related variables and report them to the caller,
            //       even in non-strict mode (i.e. when we would return Ok).
            //
            int errorCount = 0;
            Result error = null;

            if (GetVariableValue(
                    Engine.ErrorVariableFlags | flags,
                    TclVars.ErrorCode, ref errorCode,
                    ref error) != ReturnCode.Ok)
            {
                if (errors == null)
                    errors = new ResultList();

                errors.Add(error); errorCount++;
            }

            if (GetVariableValue(
                    Engine.ErrorVariableFlags | flags,
                    TclVars.ErrorInfo, ref errorInfo,
                    ref error) != ReturnCode.Ok)
            {
                if (errors == null)
                    errors = new ResultList();

                errors.Add(error); errorCount++;
            }

            //
            // NOTE: In strict mode, return Error if there were any errors
            //       querying the variables; otherwise, return Ok.
            //
            return strict ?
                (errorCount > 0 ? ReturnCode.Error : ReturnCode.Ok) :
                ReturnCode.Ok;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IHistoryManager Members
#if HISTORY
        public bool History
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return history;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    history = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int HistoryLimit
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return historyLimit;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    historyLimit = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasHistory(ref Result error)
        {
            CheckDisposed();

            bool result = (_History != null);

            if (!result)
                error = "history not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryData HistoryLoadData
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistoryLoadData;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historyLoadData;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistoryLoadData = value;
#else
                lock (syncRoot)
                {
                    historyLoadData = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryData HistorySaveData
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistorySaveData;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historySaveData;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistorySaveData = value;
#else
                lock (syncRoot)
                {
                    historySaveData = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryFilter HistoryEngineFilter
        {
            get
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.HistoryEngineFilter;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historyEngineFilter;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.HistoryEngineFilter = value;
#else
                lock (syncRoot)
                {
                    historyEngineFilter = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryFilter HistoryInfoFilter
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistoryInfoFilter;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historyInfoFilter;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistoryInfoFilter = value;
#else
                lock (syncRoot)
                {
                    historyInfoFilter = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryFilter HistoryLoadFilter
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistoryLoadFilter;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historyLoadFilter;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistoryLoadFilter = value;
#else
                lock (syncRoot)
                {
                    historyLoadFilter = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHistoryFilter HistorySaveFilter
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistorySaveFilter;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historySaveFilter;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistorySaveFilter = value;
#else
                lock (syncRoot)
                {
                    historySaveFilter = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string HistoryFileName
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.HistoryFileName;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return historyFileName;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.HistoryFileName = value;
#else
                lock (syncRoot)
                {
                    historyFileName = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ClearHistory(
            IHistoryFilter historyFilter, /* NOT USED */
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (!HasHistory(ref error))
                    return ReturnCode.Error;

                //
                // NOTE: Clear the history of executed commands.
                //
                _History.Clear();

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.History, NotifyFlags.Cleared,
                    null, this,
                    null, null, null, ref error);
#endif
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddHistory(
            ArgumentList arguments,
            IHistoryData historyData,
            IHistoryFilter historyFilter, /* NOT USED */
            ref Result error
            )
        {
            CheckDisposed();

            int levels;
            HistoryFlags flags;

            if (historyData != null)
            {
                levels = historyData.Levels;
                flags = historyData.Flags;
            }
            else
            {
                levels = Level.Invalid;
                flags = HistoryFlags.None;
            }

            return AddHistory(arguments, levels, flags, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode LoadHistory(
            Encoding encoding,
            string fileName,
            IHistoryData historyData,
            IHistoryFilter historyFilter, /* NOT USED */
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (String.IsNullOrEmpty(fileName))
                {
                    error = "invalid file name";
                    return ReturnCode.Error;
                }

                if (!File.Exists(fileName))
                {
                    error = String.Format(
                        "couldn't read file \"{0}\": no such file or directory",
                        fileName);

                    return ReturnCode.Error;
                }

                if (!HasHistory(ref error))
                    return ReturnCode.Error;

                try
                {
                    if (encoding == null)
                        encoding = DefaultEncoding;

                    using (StreamReader streamReader = new StreamReader(
                            new FileStream(fileName, FileMode.Open,
                                FileAccess.Read), encoding))
                    {
                        int levels;
                        HistoryFlags flags;

                        if (historyData != null)
                        {
                            levels = historyData.Levels;
                            flags = historyData.Flags;
                        }
                        else
                        {
                            levels = Level.Invalid;
                            flags = HistoryFlags.None;
                        }

                        while (!streamReader.EndOfStream)
                        {
                            StringList list = null;
                            Result localError = null;

                            if (Parser.SplitList(this, streamReader.ReadLine(),
                                    0, Length.Invalid, false, true, ref list,
                                    ref localError) == ReturnCode.Ok)
                            {
                                IClientData clientData = new HistoryClientData(
                                    null, new ArgumentList(list, ArgumentFlags.None),
                                    levels, flags & HistoryFlags.InstanceMask);

                                _History.AddLast(clientData);
                            }
                            else if (strict)
                            {
                                error = localError;
                                return ReturnCode.Error;
                            }
                        }
                    }

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SaveHistory(
            Encoding encoding,
            string fileName,
            IHistoryData historyData, /* NOT USED */
            IHistoryFilter historyFilter,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (String.IsNullOrEmpty(fileName))
                {
                    error = "invalid file name";
                    return ReturnCode.Error;
                }

                if (!HasHistory(ref error))
                    return ReturnCode.Error;

                try
                {
                    if (encoding == null)
                        encoding = DefaultEncoding;

                    using (StreamWriter streamWriter = new StreamWriter(
                            new FileStream(fileName, FileMode.Create,
                                FileAccess.Write), encoding))
                    {
                        foreach (IClientData clientData in _History)
                        {
                            HistoryClientData historyClientData =
                                clientData as HistoryClientData;

                            if (historyClientData != null)
                            {
                                if (RuntimeOps.MatchHistoryData(
                                        historyClientData, historyFilter))
                                {
                                    ArgumentList arguments =
                                        historyClientData.Arguments;

                                    if (arguments != null)
                                        streamWriter.WriteLine(
                                            arguments.ToString());
                                }
                            }
                            else if (strict)
                            {
                                error = "invalid history data";
                                return ReturnCode.Error;
                            }
                        }
                    }

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool CanAddHistory()
        {
            // CheckDisposed();

            // lock (syncRoot) // BUGBUG: *PERF* Used by engine.
            {
                return history &&
#if THREADING
                    (_History != null);
#else
                    (_history != null);
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool HasHistory() // NOTE: For _Hosts.Default use only.
        {
#if THREADING
            return (_History != null);
#else
            return (_history != null);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddHistory(
            ArgumentList arguments,
            int levels,
            HistoryFlags flags
            )
        {
            Result error = null;

            return AddHistory(arguments, levels, flags, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddHistory(
            ArgumentList arguments,
            int levels,
            HistoryFlags flags,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (arguments == null)
                {
                    error = "invalid arguments";
                    return ReturnCode.Error;
                }

                if (!HasHistory(ref error))
                    return ReturnCode.Error;

                //
                // NOTE: Add the arguments to the history of executed commands.
                //
                IClientData clientData =
                    new HistoryClientData(null, arguments, levels, flags);

                _History.AddFirst(clientData);

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.History, NotifyFlags.Added,
                    clientData, this,
                    null, null, null, ref error);
#endif

                //
                // NOTE: Prevent the number of history items from exceeding the
                //       limit (unless the limit is -1).
                //
                int limit = historyLimit;

                if (limit >= 0)
                {
                    //
                    // NOTE: If the limit is zero, use the default limit.
                    //
                    if (limit == 0)
                        limit = DefaultHistoryLimit;

                    while (_History.Count > limit)
                    {
                        //
                        // NOTE: Remove old argument lists in FIFO style.
                        //
                        LinkedListNode<IClientData> last = _History.Last;

                        if (last == null)
                            break;

                        clientData = last.Value;
                        _History.RemoveLast();

#if NOTIFY
                        /* IGNORED */
                        CheckNotifications(
                            NotifyType.History, NotifyFlags.Removed,
                            clientData, this,
                            null, null, null, ref error);
#endif
                    }
                }
            }

            return ReturnCode.Ok;
        }
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInteractiveManager Members
        public bool Interactive // using interactive shell?
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.Interactive;
                else
                    return false;
#else
                lock (syncRoot)
                {
                    return interactive;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.Interactive = value;
#else
                lock (syncRoot)
                {
                    interactive = value;
                }
#endif

                if (IsPrimaryThread())
                {
                    //
                    // NOTE: Do not create the variable if we have been prohibited
                    //       from doing so.
                    //
                    VariableFlags flags = VariableFlags.GlobalOnly | VariableFlags.NoReady;

                    if (FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                        flags |= VariableFlags.NoCreate;

                    //
                    // NOTE: Update script-level informational variable.
                    //
                    SetVariableValue(flags, TclVars.Interactive, value.ToString(), null);
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string InteractiveInput
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.InteractiveInput;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return interactiveInput;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.InteractiveInput = value;
#else
                lock (syncRoot)
                {
                    interactiveInput = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string PreviousInteractiveInput
        {
            get
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.PreviousInteractiveInput;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return previousInteractiveInput;
                }
#endif
            }
            set
            {
                CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.PreviousInteractiveInput = value;
#else
                lock (syncRoot)
                {
                    previousInteractiveInput = value;
                }
#endif
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInterpreter Members
        #region Identity & Affinity
        public long Id
        {
            get { CheckDisposed(); return id; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int ThreadId
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return threadId;
                }
            }
            internal set
            {
                CheckDisposed(); /* NON-PUBLIC, SANITY */

                lock (syncRoot)
                {
                    threadId = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public Thread Thread
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return thread;
                }
            }
            internal set
            {
                CheckDisposed(); /* NON-PUBLIC, SANITY */

                lock (syncRoot)
                {
                    thread = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public Thread InteractiveThread
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return interactiveThread;
                }
            }
            internal set
            {
                CheckDisposed(); /* NON-PUBLIC, SANITY */

                lock (syncRoot)
                {
                    interactiveThread = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public EventWaitHandle VariableEvent
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return variableEvent;
                }
            }
            internal set
            {
                CheckDisposed(); /* NON-PUBLIC, SANITY */

                lock (syncRoot)
                {
                    variableEvent = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public AppDomain GetAppDomain()
        {
            CheckDisposed();

            return appDomain;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public StrongName GetStrongName()
        {
            CheckDisposed();

            return AssemblyOps.GetStrongName(GlobalState.GetAssembly());
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public Hash GetHash()
        {
            CheckDisposed();

            return AssemblyOps.GetHash(GlobalState.GetAssembly());
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public X509Certificate GetCertificate()
        {
            CheckDisposed();

            return AssemblyOps.GetCertificate(GlobalState.GetAssembly());
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public void DemandStrongName()
        {
            CheckDisposed();

            StrongName strongName = GetStrongName();

            new StrongNameIdentityPermission(
                (strongName != null) ? strongName.PublicKey : null,
                null, null).Demand();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public void DemandCertificate()
        {
            CheckDisposed();

            new PublisherIdentityPermission(GetCertificate()).Demand();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public long NextId() /* THREAD-SAFE */
        {
            CheckDisposed();

            long result;

            //
            // NOTE: This is our cheap per-interpreter unique Id
            //       generator for the various script visible
            //       identifiers (such as channel names, etc).
            //
            result = Interlocked.Increment(ref nextId);

            if (result < 0)
                DebugOps.Complain(this, ReturnCode.Error,
                    "next identifier is negative");

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool IsPrimaryThread()
        {
            return IsPrimaryThread(GlobalState.GetCurrentThreadId());
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool IsPrimaryThread(int threadId)
        {
            return (threadId == this.threadId);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void SignalVariableEvent()
        {
            ReturnCode code;
            Result error = null;

            code = SignalVariableEvent(ref error);

            if (code != ReturnCode.Ok)
                DebugOps.Complain(this, code, error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SignalVariableEvent(
            ref Result error
            )
        {
            EventWaitHandle variableEvent;

            lock (syncRoot)
            {
                variableEvent = this.variableEvent;
            }

            if (variableEvent != null)
            {
                if (ThreadOps.SetEvent(variableEvent))
                {
                    return ReturnCode.Ok;
                }
                else
                {
                    error = "failed to signal variable event";
                }
            }
            else
            {
                error = "cannot signal variable event";
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Host & Script Environment
        #region Global New Host Callback
        public static NewHostCallback NewHostCallback
        {
            get
            {
                lock (staticSyncRoot)
                {
                    return newHostCallback;
                }
            }
            set
            {
                lock (staticSyncRoot)
                {
                    newHostCallback = value;
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ScriptFlags ScriptFlags
        {
            get { CheckDisposed(); lock (syncRoot) { return scriptFlags; } }
            set { CheckDisposed(); lock (syncRoot) { scriptFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IHost Host
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return host;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    host = value;

                    //
                    // BUGFIX: We may have just been handed a host that we have
                    //         never seen before.  Make sure it is setup properly.
                    //
                    ReturnCode code;
                    Result error = null;

                    if (host != null)
                    {
                        code = PreSetupHost(host, false, ref error);

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, error);

                        code = SetupHostChannels(this, host, ref error);

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, error);
                    }

#if DEBUGGER
                    code = SetupDebuggerHost(this, host, ref error);

                    if (code != ReturnCode.Ok)
                        DebugOps.Complain(this, code, error);
#endif
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IBinder Binder
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return binder;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    binder = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public CultureInfo CultureInfo
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return cultureInfo;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    cultureInfo = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: Be relatively quiet (i.e. suppress complaints)?  This is
        //       exposed as a public property; however, it should only be
        //       used in circumstances where it is extremely undesirable
        //       to have any extraneous output sent to the interpreter
        //       host (e.g. the console) because it can prevent important
        //       warnings and other information from being displayed.  An
        //       example is that MSBuild (v3.5) appears to interpret output
        //       being sent to the standard error channel as an indication
        //       that an "Exec" task has failed, even if the exit code
        //       of the command is zero.
        //
        public bool Quiet
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return quiet;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    quiet = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if POLICY_TRACE
        public bool PolicyTrace
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return policyTrace;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    policyTrace = value;
                }
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string BackgroundError
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return backgroundError;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    backgroundError = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string Unknown
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return unknown;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    unknown = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string PackageUnknown
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return packageUnknown;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    packageUnknown = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
#if ISOLATED_PLUGINS
        internal IHost CoreHost
        {
            get
            {
                // CheckDisposed();

                lock (syncRoot)
                {
                    return coreHost;
                }
            }
            set
            {
                // CheckDisposed();

                lock (syncRoot)
                {
                    coreHost = value;
                }
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool ShouldBeQuiet()
        {
            // CheckDisposed();

            return ShouldBeQuiet(DebugOps.GetDefaultQuiet(DefaultQuiet));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: This method uses more-or-less the same general idea as the
        //       SafeGetQuiet method of the DebugOps class; however, this one
        //       avoids causing an exception to be thrown (and caught) by
        //       accessing the quiet field directly instead of via the property.
        //
        internal bool ShouldBeQuiet(
            bool @default
            )
        {
            // CheckDisposed();

            if (quiet) /* FIELD */
                return true;

            //
            // HOOK: Allow the test suite (and others components) to override
            //       the quietness setting via the environment.
            //
            if (CommonOps.GetEnvironmentVariable(EnvVars.Quiet) != null)
                return true;

            return @default;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool HasNoBackgroundError()
        {
            return FlagOps.HasFlags(
                interpreterFlags, InterpreterFlags.NoBackgroundError, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void SetNoBackgroundError(
            bool noBackgroundError
            )
        {
            if (noBackgroundError)
                interpreterFlags |= InterpreterFlags.NoBackgroundError;
            else
                interpreterFlags &= ~InterpreterFlags.NoBackgroundError;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Xml Data Handling
#if XML
        public bool ValidateXml
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return validateXml;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    validateXml = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool AllXml
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return allXml;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    allXml = value;
                }
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Initialization & Termination
        public bool Exit
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return exit;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    exit = value;

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Interpreter, NotifyFlags.Exit,
                        exit, this,
                        null, null, null);
#endif
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ExitCode ExitCode
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return exitCode;
                }

            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    exitCode = value;
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Event Queue Management
        public int SleepTime
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return sleepTime;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    sleepTime = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public IEventManager EventManager
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return eventManager;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public EventFlags ServiceEventFlags
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return serviceEventFlags;
                }
            }
            private set
            {
                // CheckDisposed();

                lock (syncRoot)
                {
                    serviceEventFlags = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode QueueScript(
            DateTime dateTime,
            string text,
            ref Result error
            )
        {
            CheckDisposed();

            IEvent @event = null;

            return QueueScript(dateTime, text, ref @event, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode QueueScript(
            DateTime dateTime,
            string text,
            ref IEvent @event,
            ref Result error
            )
        {
            CheckDisposed();

            //
            // BUGFIX: Prevent deadlock caused by holding the interpreter lock
            //         while calling the QueueScript method, which acquires and
            //         holds the event manager lock.  The other thread holds
            //         the event manager lock via the DequeueAnyReadyEvent
            //         method and then ends up trying to acquire the
            //         interpreter lock while attempting to get the value of
            //         the NotifyFlags property via the CheckNotifications
            //         method.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            if (eventManager != null)
            {
                return eventManager.QueueScript(
                    dateTime, text, engineFlags, substitutionFlags,
                    queueEventFlags, expressionFlags,
                    EventPriority.QueueScript, ref @event, ref error);
            }
            else
            {
                error = "event manager not available";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Pending Wait Count
        public static bool IsPendingWait(
            Interpreter interpreter
            )
        {
            return (interpreter != null) ? interpreter.IsPendingWait() : false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal int WaitCount
        {
            get { return waitCount; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int WaitSpinCount
        {
            get { return waitSpinCount; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal IScript CreateAfterScript(
            string name,
            string group,
            string description,
            string type,
            string text,
            DateTime timeStamp,
            EngineMode engineMode,
            ScriptFlags scriptFlags,
            IClientData clientData,
            bool idle
            )
        {
            return Script.Create(
                name, group, description, type, text, timeStamp,
                engineMode, scriptFlags,
                engineFlags | EngineFlags.EvaluateGlobal,
                substitutionFlags, afterEventFlags |
                    (idle ? EventFlags.Idle : EventFlags.None) |
                    EventFlags.Interpreter, expressionFlags,
                clientData);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags CombineEngineEventFlags(
            EventFlags eventFlags
            )
        {
            return FlagOps.HasFlags(eventFlags, EventFlags.Interpreter, true) ?
                engineEventFlags | (eventFlags & ~EventFlags.DequeueMask) :
                eventFlags;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void ClearEvents()
        {
            //
            // BUGFIX: Prevent deadlock caused by holding the interpreter lock
            //         while calling the ClearEvents method, which acquires and
            //         holds the event manager lock.  The other thread holds
            //         the event manager lock via the DequeueAnyReadyEvent
            //         method and then ends up trying to acquire the
            //         interpreter lock while attempting to get the value of
            //         the NotifyFlags property via the CheckNotifications
            //         method.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            if (eventManager != null)
            {
                ReturnCode code;
                Result error = null;

                code = eventManager.ClearEvents(ref error);

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode QueueScript( /* ScriptThread USE ONLY */
            DateTime dateTime,
            string text,
            EventFlags eventFlags,
            ref Result error
            )
        {
            // CheckDisposed();

            //
            // BUGFIX: Prevent deadlock caused by holding the interpreter lock
            //         while calling the QueueScript method, which acquires and
            //         holds the event manager lock.  The other thread holds
            //         the event manager lock via the DequeueAnyReadyEvent
            //         method and then ends up trying to acquire the
            //         interpreter lock while attempting to get the value of
            //         the NotifyFlags property via the CheckNotifications
            //         method.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            if (eventManager != null)
            {
                return eventManager.QueueScript(
                    dateTime, text, engineFlags, substitutionFlags,
                    eventFlags, expressionFlags, EventPriority.QueueScript,
                    ref error);
            }
            else
            {
                error = "event manager not available";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int GetSleepTime()
        {
            //
            // BUGFIX: Prevent possible future deadlock caused by holding the
            //         interpreter lock while calling the GetSleepTime method,
            //         which may at some point acquire and hold the event
            //         manager lock.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            return (eventManager != null) && !eventManager.Disposed ?
                eventManager.GetSleepTime() :
                _Public.EventManager.DefaultSleepTime;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int GetMinimumSleepTime()
        {
            //
            // BUGFIX: Prevent possible future deadlock caused by holding the
            //         interpreter lock while calling the GetMinimumSleepTime
            //         method, which may at some point acquire and hold the
            //         event manager lock.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            //
            // BUGFIX: Since this method is called from Dispose, do not attempt
            //         to access the event manager if it too has already been
            //         disposed.
            //
            return (eventManager != null) && !eventManager.Disposed ?
                eventManager.GetMinimumSleepTime() :
                _Public.EventManager.MinimumSleepTime;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Pending Wait Count
        private bool IsPendingWait()
        {
            return (waitCount > 0);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterWait()
        {
            return Interlocked.Increment(ref waitCount);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitWait()
        {
            return Interlocked.Decrement(ref waitCount);
        }
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Argument & Option Handling
        public ReturnCode GetOptions(
            OptionDictionary options,
            ArgumentList arguments,
            int listCount /* for MustBeIndex, the count of the list tied to the option(s) */,
            int startIndex,
            int stopIndex,
            bool strict, /* raise an error if any unknown option is encountered (instead of stopping). */
            ref int nextIndex /* upon success, location of the first non-option argument. */,
            ref Result error
            )
        {
            CheckDisposed();

            int endIndex = Index.Invalid;

            return GetOptions(
                options, arguments, listCount, startIndex, stopIndex,
                strict ? OptionBehaviorFlags.Strict : OptionBehaviorFlags.Default,
                true, false, false, ref nextIndex, ref endIndex, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetArguments(
            StringList arguments,
            ref Result error
            )
        {
            CheckDisposed();

            return SetArguments(arguments, true, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetArguments(
            StringList arguments,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            ReturnCode code;

            if (arguments != null)
            {
                code = SetVariableValue(
                    VariableFlags.GlobalOnly, TclVars.ShellArgumentCount,
                    arguments.Count.ToString(), null, ref error);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.ShellArguments,
                        arguments.ToString(), null, ref error);
            }
            else if (!strict)
            {
                code = SetVariableValue(
                    VariableFlags.GlobalOnly, TclVars.ShellArgumentCount,
                    0.ToString(), null, ref error);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.ShellArguments,
                        null, null, ref error);
            }
            else
            {
                error = "invalid arguments";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        //
        // NOTE: This is the "Universal Option Parser".
        //
        internal ReturnCode GetOptions(
            OptionDictionary options,
            ArgumentList arguments,
            int listCount /* for MustBeIndex, the count of the list tied to the option(s) */,
            int startIndex,
            int stopIndex,
            OptionBehaviorFlags behaviorFlags, /* do we want to stop, error, or ignore unknown options? */
            bool noCase, /* case-insensitive option names? */
            bool noValue, /* do not try to interpret strings as values of any kind. */
            bool noSet, /* scan mode, do not set any values (no changes to options dictionary). */
            ref int nextIndex, /* upon success, location of the first non-option argument. */
            ref int endIndex, /* upon success, location of EndOfOptions if found, otherwise unchanged. */
            ref Result error
            )
        {
            ReturnCode code;

            if (options != null)
            {
                if (arguments != null)
                {
                    if (ListOps.CheckStartAndStopIndex(0, arguments.Count - 1,
                            ref startIndex, ref stopIndex, ref error))
                    {
                        code = ReturnCode.Ok;

                        for (int argumentIndex = startIndex; argumentIndex <= stopIndex; argumentIndex++)
                        {
                            string current = arguments[argumentIndex];
                            string next = ((argumentIndex + 1) <= stopIndex) ?
                                arguments[argumentIndex + 1] : null;

                            if (Option.IsEndOfOptions(current))
                            {
                                if (options.Has(Option.EndOfOptions))
                                {
                                    if ((argumentIndex + 1) < arguments.Count)
                                        nextIndex = argumentIndex + 1;
                                    else
                                        nextIndex = Index.Invalid;

                                    endIndex = argumentIndex;
                                    break;
                                }
                                else if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.ErrorOnEndOfOptions, true))
                                {
                                    //
                                    // NOTE: This is rarely used because if they do not
                                    //       explicitly allow the end-of-options marker
                                    //       they normally also want it to be considered
                                    //       the first "non-option" argument.
                                    //
                                    error = "unexpected end-of-options marker";
                                    code = ReturnCode.Error;
                                    break;
                                }
                                else if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.StopOnEndOfOptions, true))
                                {
                                    nextIndex = argumentIndex;
                                    break;
                                }
                                else if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.IgnoreOnEndOfOptions, true))
                                {
                                    //
                                    // NOTE: This is not a valid option and we want to
                                    //       simply ignore it.  We may also want to skip
                                    //       the next argument.
                                    //
                                    if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.SkipOnEndOfOptions, true))
                                        argumentIndex++;
                                }
                            }
                            else if (Option.LooksLikeOption(current))
                            {
                                IOption option = null;
                                Result localError = null;

                                if (options.TryResolve(current, true, noCase,
                                        ref option, ref localError) == ReturnCode.Ok)
                                {
                                    if (option != null)
                                    {
                                        if (option.IsIgnored(options))
                                        {
                                            if (option.MustHaveValue(options))
                                            {
                                                if (next != null)
                                                {
                                                    argumentIndex++; // skip option value...
                                                }
                                                else
                                                {
                                                    error = String.Format(
                                                        "\"{0}\" option must be followed by {1}",
                                                        current, option.ToString(option.Flags));

                                                    code = ReturnCode.Error;
                                                    break;
                                                }
                                            }
                                        }
                                        else if (!option.CanBePresent(options, ref error))
                                        {
                                            code = ReturnCode.Error;
                                            break;
                                        }
                                        else if (option.HasFlags(OptionFlags.EndOfOptions, true))
                                        {
                                            if ((argumentIndex + 1) < arguments.Count)
                                                nextIndex = argumentIndex + 1;
                                            else
                                                nextIndex = Index.Invalid;

                                            endIndex = argumentIndex;
                                            break;
                                        }
                                        else if (option.IsUnsafe(options) && IsSafe())
                                        {
                                            error = String.Format(
                                                "permission denied: safe interpreter cannot use option \"{0}\"",
                                                option.Name);

                                            code = ReturnCode.Error;
                                        }
                                        else if (option.MustHaveValue(options))
                                        {
                                            if (next != null)
                                            {
                                                //
                                                // NOTE: Initially, this value of this option is null.  The block
                                                //       below has the implicit guarantee that either the option
                                                //       value will be set to something non-null OR the return
                                                //       code will be set to something non-Ok.
                                                //
                                                Variant variant = null;

                                                //
                                                // NOTE: Are we allowed to interpret the string as the type of
                                                //       value required for this option?  Otherwise, we will not
                                                //       attempt any conversion and if we are in non-scan mode
                                                //       the final option value will be set as a string.  We
                                                //       assume that the caller knows this.
                                                //
                                                if (!noValue)
                                                {
                                                    //
                                                    // NOTE: Does this option require strict value recognition
                                                    //       semantics?
                                                    //
                                                    bool strictOption = option.IsStrict(options);

                                                    //
                                                    // NOTE: Does this option not care about case?
                                                    //
                                                    bool noCaseOption = option.IsNoCase(options);

                                                    //
                                                    // NOTE: Ok, now figure out what kind of value this option
                                                    //       requires and try to parse it from the option value
                                                    //       string.
                                                    //
                                                    if (option.HasFlags(OptionFlags.MustBeTypeList, true))
                                                    {
                                                        TypeList typeList = null;
                                                        ResultList errors = null;

                                                        code = Value.GetTypeList(
                                                            this, next, appDomain, strictOption, false,
                                                            noCase || noCaseOption, cultureInfo, ref typeList,
                                                            ref errors);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(typeList);
                                                        else
                                                            error = errors;
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeReturnCodeList, true))
                                                    {
                                                        ReturnCodeList returnCodeList = null;

                                                        code = Value.GetReturnCodeList(this, next, cultureInfo,
                                                            ref returnCodeList, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(returnCodeList);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeVersion, true))
                                                    {
                                                        Version version = null;

                                                        code = Value.GetVersion(next, cultureInfo,
                                                            ref version, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(version);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeAlias, true))
                                                    {
                                                        IAlias alias = null;

                                                        code = GetAlias(next, true, true, ref alias, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(alias);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeOption, true))
                                                    {
                                                        IOption optionValue = Option.FromString(
                                                            this, next, appDomain, strictOption, false,
                                                            noCase || noCaseOption, cultureInfo, ref error);

                                                        if (optionValue != null)
                                                            variant = new Variant(optionValue);
                                                        else
                                                            code = ReturnCode.Error;
                                                    }
#if NATIVE && TCL
                                                    else if (option.HasFlags(OptionFlags.MustBeTclInterpreter, true))
                                                    {
                                                        IntPtr interp = IntPtr.Zero;

#if TCL_THREADS
                                                        code = GetTclInterpreterOrThread(next, ref interp, ref error);
#else
                                                        code = GetTclInterpreter(next, ref interp, ref error);
#endif

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(next); /* interpName */
                                                    }
#endif
                                                    else if (option.HasFlags(OptionFlags.MustBeType, true))
                                                    {
                                                        Type type = null;
                                                        ResultList errors = null;

                                                        code = Value.GetType(
                                                            this, next, appDomain, strictOption, false,
                                                            noCase || noCaseOption, cultureInfo, ref type,
                                                            ref errors);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(type);
                                                        else
                                                            error = errors;
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeInterpreter, true))
                                                    {
                                                        Interpreter interpreter = null;

                                                        code = Value.GetInterpreter(
                                                            this, next, InterpreterType.Default, ref interpreter,
                                                            ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(interpreter);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeObject, true))
                                                    {
                                                        IObject @object = null;

                                                        code = GetObject(next, true, true, ref @object, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(@object);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeValue, true))
                                                    {
                                                        object value = null;

                                                        code = Value.GetValue(next, ValueFlags.Any, true, cultureInfo,
                                                            ref value, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(value);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeList, true))
                                                    {
                                                        StringList list = null;

                                                        //
                                                        // TODO: *PERF* We cannot have this call to SplitList perform any
                                                        //       caching because we do not know exactly what the resulting
                                                        //       list will be used for.
                                                        //
                                                        code = Parser.SplitList(
                                                            this, next, 0, Length.Invalid,
                                                            false, false, ref list, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(list);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeDateTime, true))
                                                    {
                                                        DateTime dateTime = DateTime.MinValue;

                                                        code = Value.GetDateTime2(next, ValueFlags.AnyDateTime,
                                                            Value.DefaultDateTimeKind, strictOption, cultureInfo,
                                                            ref dateTime, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(dateTime);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeTimeSpan, true))
                                                    {
                                                        TimeSpan timeSpan = TimeSpan.Zero;

                                                        code = Value.GetTimeSpan2(next, ValueFlags.AnyTimeSpan,
                                                            strictOption, cultureInfo, ref timeSpan, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(timeSpan);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeGuid, true))
                                                    {
                                                        Guid guid = Guid.Empty;

                                                        code = Value.GetGuid(next, cultureInfo, ref guid, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(guid);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeEnum, true))
                                                    {
                                                        if (option.Type != null)
                                                        {
                                                            if (option.Type.IsEnum)
                                                            {
                                                                object enumValue;

                                                                //
                                                                // NOTE: Reset the local error result here because we check it
                                                                //       below (i.e. it will not be set by TryParseEnum).
                                                                //
                                                                localError = null;

                                                                if (EnumOps.IsFlagsEnum(option.Type))
                                                                    enumValue = EnumOps.TryParseFlagsEnum(
                                                                        this, option.Type, (option.Value != null) ?
                                                                            option.Value.ToString() : null, next,
                                                                        cultureInfo, noCase || noCaseOption,
                                                                        ref localError);
                                                                else
                                                                    enumValue = EnumOps.TryParseEnum(
                                                                        option.Type, next, noCase || noCaseOption);

                                                                //
                                                                // NOTE: Did we end up with a valid enum value for this type?
                                                                //
                                                                if (enumValue != null)
                                                                {
                                                                    //
                                                                    // NOTE: It is now guaranteed to be the correct Enum type because
                                                                    //       TryParseEnum uses our System.Type object directly and
                                                                    //       cannot return non-null unless Enum.Parse succeeds.
                                                                    //
                                                                    variant = new Variant(enumValue);
                                                                }
                                                                else
                                                                {
                                                                    //
                                                                    // NOTE: Retain any locally generated error information (from
                                                                    //       TryParseFlagsEnum) because it provides more specific
                                                                    //       feedback than we can at this point.
                                                                    //
                                                                    if (!String.IsNullOrEmpty(localError))
                                                                        error = localError;
                                                                    else
                                                                        error = RuntimeOps.BadValue(
                                                                            null, String.Format("\"{0}\" value", option.Type.FullName),
                                                                            next, Enum.GetNames(option.Type), null, null);

                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                error = String.Format(
                                                                    "option \"{0}\" base type is not an enum",
                                                                    option.Name);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            error = String.Format(
                                                                "option \"{0}\" has an invalid base type",
                                                                option.Name);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeReturnCode, true))
                                                    {
                                                        ReturnCode returnCode = ReturnCode.Ok;

                                                        code = Value.GetReturnCode2(next, ValueFlags.AnyReturnCode,
                                                            cultureInfo, ref returnCode, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(returnCode);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeMatchMode, true))
                                                    {
                                                        MatchMode matchMode = MatchMode.None;

                                                        code = Value.GetMatchMode2(next, ValueFlags.AnyMatchMode,
                                                            cultureInfo, ref matchMode, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(matchMode);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeLevel, true))
                                                    {
                                                        //
                                                        // NOTE: Not currently supported.
                                                        //
                                                        error = "options of type \"level\" are not currently supported";
                                                        code = ReturnCode.Error;
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeIndex, true))
                                                    {
                                                        int index = Index.Invalid;

                                                        code = Value.GetIndex(next, listCount, cultureInfo,
                                                            ref index, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(index);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeWideInteger, true))
                                                    {
                                                        long longValue = 0;

                                                        code = Value.GetWideInteger2(next, ValueFlags.AnyWideInteger,
                                                            cultureInfo, ref longValue, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(longValue);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeInteger, true))
                                                    {
                                                        int intValue = 0;

                                                        code = Value.GetInteger2(next, ValueFlags.AnyInteger,
                                                            cultureInfo, ref intValue, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(intValue);
                                                    }
                                                    else if (option.HasFlags(OptionFlags.MustBeBoolean, true))
                                                    {
                                                        bool boolValue = false;

                                                        code = Value.GetBoolean2(next, ValueFlags.AnyBoolean,
                                                            strictOption, cultureInfo, ref boolValue, ref error);

                                                        if (code == ReturnCode.Ok)
                                                            variant = new Variant(boolValue);
                                                    }
                                                    else
                                                    {
                                                        //
                                                        // NOTE: No strongly-typed value flags were set for this
                                                        //       option, just use the string verbatim.
                                                        //
                                                        variant = new Variant(next); // String
                                                    }
                                                }
                                                else
                                                {
                                                    //
                                                    // NOTE: We are forbidden from interpreting the string as a
                                                    //       strongly-typed value, just use the string verbatim.
                                                    //
                                                    variant = new Variant(next); // String
                                                }

                                                //
                                                // NOTE: If we succeeded, set the option value; otherwise, bail.
                                                //
                                                if (code == ReturnCode.Ok)
                                                {
                                                    if (!noSet)
                                                        option.SetPresent(options, true, argumentIndex, variant);

                                                    argumentIndex++; // skip option value...
                                                }
                                                else
                                                {
                                                    break;
                                                }
                                            }
                                            else
                                            {
                                                error = String.Format(
                                                    "\"{0}\" option must be followed by {1}",
                                                    current, option.ToString(option.Flags));

                                                code = ReturnCode.Error;
                                                break;
                                            }
                                        }
                                        else if (!noSet)
                                        {
                                            option.SetPresent(options, true, argumentIndex, null);
                                        }
                                    }
                                }
                                else
                                {
                                    if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.ErrorOnUnknown, true))
                                    {
                                        //
                                        // NOTE: This is not a valid option and we want to
                                        //       treat this as an error.
                                        //
                                        error = localError;
                                        code = ReturnCode.Error;
                                        break;
                                    }
                                    else if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.StopOnUnknown, true))
                                    {
                                        //
                                        // NOTE: This is not a valid option and we want to
                                        //       stop now because we want to assume it is
                                        //       the first non-option argument.
                                        //
                                        nextIndex = argumentIndex;
                                        break;
                                    }
                                    else if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.IgnoreOnUnknown, true))
                                    {
                                        //
                                        // NOTE: This is not a valid option and we want to
                                        //       simply ignore it.  We may also want to skip
                                        //       the next argument.
                                        //
                                        if (FlagOps.HasFlags(behaviorFlags, OptionBehaviorFlags.SkipOnUnknown, true))
                                            argumentIndex++;
                                    }
                                }
                            }
                            else
                            {
                                nextIndex = argumentIndex;
                                break;
                            }
                        }
                    }
                    else
                    {
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    error = "invalid arguments";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                error = "invalid options";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode MergeArguments(
            OptionDictionary options,
            ArgumentList arguments1,
            ArgumentList arguments2,
            int startIndex1,
            int startIndex2,
            bool skipName1,
            bool skipName2,
            ref ArgumentList arguments,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Ok;

            //
            // NOTE: The purpose of this method is to properly merge two
            //       arguments lists, including any options, so that the new
            //       combined argument list is constructed in a well-defined
            //       way and can be processed properly by the target command.
            //
            if ((arguments1 != null) || (arguments2 != null))
                arguments = new ArgumentList();
            else
                arguments = null;

            //
            // NOTE: *PHASE 1* Figure out where the options are in the first
            //       argument list (if there are any at all).
            //
            int nextIndex1 = Index.Invalid;
            int endIndex1 = Index.Invalid;

            if (code == ReturnCode.Ok)
            {
                if (startIndex1 < 0)
                    startIndex1 = 0;

                if ((options != null) && (arguments1 != null))
                {
                    if (arguments1.Count > startIndex1)
                    {
                        //
                        // NOTE: Execute the universal option parser in "scan only"
                        //       mode.  We do not want to waste any time processing
                        //       the option values here just to throw them away, we
                        //       just need to know where they end so that we can build
                        //       the final argument list for the target command.
                        //
                        code = GetOptions(
                            options, arguments1, 0, startIndex1, Index.Invalid,
                            OptionBehaviorFlags.Default, true, true, true,
                            ref nextIndex1, ref endIndex1, ref error);
                    }
                }
                else if ((options == null) && (arguments1 != null))
                {
                    //
                    // BUGFIX: We must set the next index even if there are no
                    //         options because the default value means that
                    //         there are "no non-option arguments left".
                    //
                    nextIndex1 = startIndex1;
                }
            }

            //
            // NOTE: *PHASE 2* Figure out where the options are in the second
            //       argument list (if there are any at all).
            //
            int nextIndex2 = Index.Invalid;
            int endIndex2 = Index.Invalid;

            if (code == ReturnCode.Ok)
            {
                if (startIndex2 < 0)
                    startIndex2 = 0;

                if ((options != null) && (arguments2 != null))
                {
                    if (arguments2.Count > startIndex2)
                    {
                        //
                        // NOTE: Execute the universal option parser in "scan only"
                        //       mode.  We do not want to waste any time processing
                        //       the option values here just to throw them away, we
                        //       just need to know where they end so that we can build
                        //       the final argument list for the target command.
                        //
                        code = GetOptions(
                            options, arguments2, 0, startIndex2, Index.Invalid,
                            OptionBehaviorFlags.Default, true, true, true,
                            ref nextIndex2, ref endIndex2, ref error);
                    }
                }
                else if ((options == null) && (arguments2 != null))
                {
                    //
                    // BUGFIX: We must set the next index even if there are no
                    //         options because the default value means that
                    //         there are "no non-option arguments left".
                    //
                    nextIndex2 = startIndex2;
                }
            }

            //
            // NOTE: *PHASE 3* With the information gathered above, perform the
            //       merge of the two argument lists into one, retaining the
            //       options chosen and any trailing end-of-options marker we
            //       find (one maximum).
            //
            if (code == ReturnCode.Ok)
            {
                //
                // NOTE: Process the arguments from the first list, if any.
                //
                if (arguments1 != null)
                {
                    //
                    // NOTE: Figure out where we need to stop adding option
                    //       arguments from the first list.
                    //
                    int stopIndex1;

                    if (endIndex1 != Index.Invalid) /* Exclude EndOfOptions? */
                        stopIndex1 = endIndex1 - ((arguments2 != null) ? 1 : 0);
                    else if ((nextIndex1 != Index.Invalid) && (nextIndex1 > 0))
                        stopIndex1 = nextIndex1 - 1; /* Before First Non-Option */
                    else
                        stopIndex1 = arguments1.Count - 1; /* All */

                    //
                    // NOTE: First, we want all the arguments from the first list
                    //       except the trailing end-of-options marker, if any.
                    //
                    for (int argumentIndex1 = (skipName1 ? 1 : 0);
                            argumentIndex1 <= stopIndex1;
                            argumentIndex1++)
                        arguments.Add(arguments1[argumentIndex1]);
                }

                //
                // NOTE: Process the arguments from the second list, if any.
                //
                if (arguments2 != null)
                {
                    //
                    // NOTE: Figure out where we need to stop adding option
                    //       arguments from the second list.
                    //
                    int stopIndex2;

                    if (endIndex2 != Index.Invalid)
                        stopIndex2 = endIndex2; /* Include EndOfOptions */
                    else if ((nextIndex2 != Index.Invalid) && (nextIndex2 > 0))
                        stopIndex2 = nextIndex2 - 1; /* Before First Non-Option */
                    else
                        stopIndex2 = Index.Invalid; /* None */

                    //
                    // NOTE: Next, we want all the option arguments from the second
                    //       list, including the trailing end-of-options marker, if
                    //       any.  If there is no trailing end-of-options marker in
                    //       the second list and there was in the first list, we
                    //       "move" the one from the first list to the end of the
                    //       options arguments for the second list.
                    //
                    if (stopIndex2 != Index.Invalid)
                        for (int argumentIndex2 = startIndex2;
                                argumentIndex2 <= stopIndex2; argumentIndex2++)
                            arguments.Add(arguments2[argumentIndex2]);

                    //
                    // NOTE: Next, if the first list had a trailing end-of-options
                    //       marker (which we skipped) and the second list does not,
                    //       add it to the resulting argument list now.
                    //
                    if ((endIndex1 != Index.Invalid) && (endIndex2 == Index.Invalid))
                        arguments.Add(Option.EndOfOptions);

                    //
                    // NOTE: Next, we want all the [non-option] arguments from the
                    //       second list starting at the beginning of the list and
                    //       ending before the first option argument.
                    //
                    for (int argumentIndex2 = (skipName2 ? 1 : 0);
                            (argumentIndex2 < arguments2.Count) &&
                            (argumentIndex2 < startIndex2); argumentIndex2++)
                        arguments.Add(arguments2[argumentIndex2]);

                    //
                    // NOTE: Finally, we want all the [non-option] arguments from the
                    //       second list starting after the last option argument until
                    //       the end of the list.
                    //
                    if (stopIndex2 != Index.Invalid)
                        for (int argumentIndex2 = stopIndex2 + 1;
                                argumentIndex2 < arguments2.Count; argumentIndex2++)
                            arguments.Add(arguments2[argumentIndex2]);
                }
            }

            return code;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Expression Precision
        public int Precision /* throw */
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return precision;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    precision = value;
                }

                //
                // NOTE: Do not create the variable if we have been prohibited
                //       from doing so.
                //
                VariableFlags flags = VariableFlags.GlobalOnly | VariableFlags.NoReady;

                if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                    FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                {
                    flags |= VariableFlags.NoCreate;
                }

                //
                // NOTE: Update script-level informational variable.
                //
                SetVariableValue(flags, TclVars.PrecisionName,
                    value.ToString(), precisionTraceList);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal static decimal FixIntermediatePrecision(decimal value)
        {
            return value; /* NOP */
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static double FixIntermediatePrecision(double value)
        {
            return value; /* NOP */
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal decimal FixFinalPrecision(decimal value)
        {
            if (precision > 0)
                return Math.Round(value, precision);
            else if (precision < 0)
                return Math.Round(value, 0);
            else
                return value;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal double FixFinalPrecision(double value)
        {
            if (precision > 0)
                return Math.Round(value, precision);
            else if (precision < 0)
                return Math.Round(value, 0);
            else
                return value;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Execution Limits
        //
        // NOTE: If this value is true, no objects (commands, functions, procecures,
        //       etc) may be added, renamed, or removed and no variables may be added,
        //       set, or unset.
        //
        public bool ReadOnly
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return readOnly;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    readOnly = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int RecursionLimit
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return recursionLimit;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    recursionLimit = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int Timeout
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return timeout;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    timeout = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int ThreadStackSize
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return threadStackSize;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    threadStackSize = value;
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Notifications
#if NOTIFY || NOTIFY_OBJECT
        public ReturnCode FireNotification(
            IScriptEventArgs eventArgs,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            CheckDisposed();

            return FireNotification(this, eventArgs, clientData, arguments, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private bool CanFireNotifications()
        {
            return (NotifyLevels == 0) &&
                !FlagOps.HasFlags(NotifyFlags, NotifyFlags.NoNotify, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool CheckNotifyTypesAndFlags(
            NotifyType type,         /* in */
            NotifyFlags flags        /* in */
            )
        {
            //
            // NOTE: Using the properties instead of the fields here takes
            //       into account the per-thread types and flags as well
            //       because the properties return a bitwise OR of the
            //       global types and flags with the per-thread values.
            //
            if (FlagOps.HasFlags(NotifyTypes, type, true) &&
                FlagOps.HasFlags(NotifyFlags, flags, true))
            {
                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode FireNotification(
            Interpreter interpreter,
            IScriptEventArgs eventArgs,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                if (interpreter.CanFireNotifications())
                {
                    PluginWrapperDictionary plugins;

                    lock (interpreter.SyncRoot)
                    {
                        //
                        // NOTE: Snapshot semantics.
                        //
                        plugins = interpreter.plugins;

                        if (plugins != null)
                            plugins = new PluginWrapperDictionary(plugins);
                    }

                    if (plugins != null)
                    {
                        //
                        // NOTE: Set the initial result to success.
                        //
                        code = ReturnCode.Ok;

                        //
                        // NOTE: Prevent endless event notification recursion.
                        //
                        interpreter.NotifyFlags |= NotifyFlags.NoNotify;

                        //
                        // NOTE: Process each trace (as long as they all continue
                        //       to succeed).
                        //
                        foreach (KeyValuePair<string, _Wrappers.Plugin> pair in plugins)
                        {
                            IPlugin plugin = pair.Value;

                            if (plugin != null)
                            {
                                INotify notify = plugin as INotify;

                                if (notify != null)
                                {
                                    try
                                    {
                                        //
                                        // NOTE: Query the plugin to find out which notification
                                        //       types it has elected to receive.
                                        //
                                        NotifyType types = notify.GetTypes(interpreter);

                                        //
                                        // NOTE: Query the plugin to find out which notification
                                        //       sub-types it has elected to receive.
                                        //
                                        NotifyFlags flags = notify.GetFlags(interpreter);

                                        //
                                        // NOTE: Query the event arguments to figure out which
                                        //       notification types are currently being sent.
                                        //
                                        NotifyType hasTypes = (eventArgs != null) ?
                                            eventArgs.Types : NotifyType.None;

                                        //
                                        // NOTE: Query the event arguments to figure out which
                                        //       notification sub-types are currently being sent.
                                        //
                                        NotifyFlags hasFlags = (eventArgs != null) ?
                                            eventArgs.Flags : NotifyFlags.None;

                                        //
                                        // NOTE: Does this plugin want to receive the notifications
                                        //       being sent?
                                        //
                                        if (FlagOps.HasFlags(types, hasTypes, true) &&
                                            FlagOps.HasFlags(flags, hasFlags, true))
                                        {
                                            //
                                            // NOTE: Since events can basically do anything they want,
                                            //       we wrap them in a try block to prevent exceptions
                                            //       from escaping.
                                            //
                                            interpreter.EnterNotifyLevel();

                                            try
                                            {
                                                //
                                                // NOTE: Call their notification function.  This could
                                                //       do practially anything; however, in practice it
                                                //       will not being modifying the state of the passed
                                                //       in objects, if any.  No guarantee is made about
                                                //       the ability of plugins to affect changes in the
                                                //       state of the passed in objects OR to alter the
                                                //       course of action to be taken after the
                                                //       notification has been processed.
                                                //
                                                code = notify.Notify(interpreter, eventArgs, clientData,
                                                    arguments, ref result);
                                            }
                                            finally
                                            {
                                                interpreter.ExitNotifyLevel();
                                            }
                                        }
                                    }
                                    catch (Exception e)
                                    {
                                        //
                                        // NOTE: Translate exceptions to a failure return.
                                        //
                                        result = String.Format(
                                            "caught exception while firing notification: {0}",
                                            e);

                                        code = ReturnCode.Error;
                                    }

                                    //
                                    // NOTE: Check for exception results specially because we
                                    //       treat "Break" different from other return codes.
                                    //
                                    if (code == ReturnCode.Break)
                                    {
                                        //
                                        // NOTE: Success; however, skip processing further
                                        //       events.
                                        //
                                        code = ReturnCode.Ok;
                                        break;
                                    }
                                    else if (code != ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: Some type of failure (or exception), stop
                                        //       processing events.
                                        //
                                        break;
                                    }
                                }
                            }
                        }

                        //
                        // NOTE: Remove event notification prevention flag.
                        //
                        interpreter.NotifyFlags &= ~NotifyFlags.NoNotify;
                    }
                    else
                    {
                        result = "plugins not available";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    //
                    // NOTE: Event notifications are disabled or we are
                    //       already processing one on this thread.  Just
                    //       return success.
                    //
                    code = ReturnCode.Ok;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckNotifications(
            NotifyType type,         /* in */
            NotifyFlags flags,       /* in */
            object data,             /* in */
            Interpreter interpreter, /* in */
            IClientData clientData,  /* in */
            ArgumentList arguments,  /* in */
            Exception exception      /* in */
            )
        {
            Result result = null;

            return CheckNotifications(type, flags, data, interpreter, clientData,
                arguments, exception, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckNotifications(
            NotifyType type,         /* in */
            NotifyFlags flags,       /* in */
            object data,             /* in */
            Interpreter interpreter, /* in */
            IClientData clientData,  /* in */
            ArgumentList arguments,  /* in */
            Exception exception,     /* in */
            ref Result result        /* in, out */
            )
        {
            if (!IsPendingDispose(this) &&
                !Engine.HasNoNotify(engineFlags) &&
                CanFireNotifications())
            {
                if (CheckNotifyTypesAndFlags(type, flags))
                {
#if MONO_BUILD
#pragma warning disable 219
#endif
                    ReturnCode eventCode; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

                    Result eventResult = null;

                    try
                    {
                        ScriptException scriptException = null;

                        if (exception != null)
                        {
                            if (exception is ScriptException)
                                scriptException = (ScriptException)exception;
                            else
                                scriptException = new ScriptException(null, exception);
                        }

                        IScriptEventArgs eventArgs = RuntimeOps.GetEventArgs(
                            type, flags, interpreter, new ClientData(data) /* WRAP */,
                            arguments, result, scriptException);

                        IClientData eventClientData = clientData; /* REF */

                        ArgumentList eventArguments = (arguments != null) ?
                            new ArgumentList(arguments) : null; /* COPY */

                        eventResult = Result.Copy(result, true); /* COPY */

                        eventCode = FireNotification(
                            eventArgs, eventClientData, eventArguments, ref eventResult);
                    }
                    catch (Exception e)
                    {
                        eventResult = String.Format(
                            "caught exception while firing notifications: {0}",
                            e);

                        eventCode = ReturnCode.Error;
                    }

#if DEBUG && VERBOSE
                    if (eventCode != ReturnCode.Ok)
                        DebugOps.Complain(interpreter, eventCode, eventResult);
#endif
                }
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY_GLOBAL
        internal static ReturnCode CheckNotifications(
            string pattern,          /* in */
            bool noCase,             /* in */
            NotifyType type,         /* in */
            NotifyFlags flags,       /* in */
            object data,             /* in */
            Interpreter interpreter, /* in */
            IClientData clientData,  /* in */
            ArgumentList arguments,  /* in */
            Exception exception      /* in */
            )
        {
            Result result = null;

            return CheckNotifications(
                pattern, noCase, type, flags, data, interpreter,
                clientData, arguments, exception, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode CheckNotifications(
            string pattern,          /* in */
            bool noCase,             /* in */
            NotifyType type,         /* in */
            NotifyFlags flags,       /* in */
            object data,             /* in */
            Interpreter interpreter, /* in */
            IClientData clientData,  /* in */
            ArgumentList arguments,  /* in */
            Exception exception,     /* in */
            ref Result result        /* in, out */
            )
        {
            InterpreterDictionary interpreters = GlobalState.GetInterpreters();

            if (interpreters != null)
            {
                ReturnCode code = ReturnCode.Ok;

                foreach (KeyValuePair<string, Interpreter> pair in interpreters)
                {
                    //
                    // NOTE: Do the pattern matching first to avoid attempting to
                    //       acquire interpreter locks without a valid reason.
                    //
                    if ((pattern == null) ||
                        Parser.StringMatch(null, pair.Key, 0, pattern, 0, noCase))
                    {
                        Interpreter otherInterpreter = pair.Value;

                        if (otherInterpreter == null)
                            continue;

                        object syncRoot = otherInterpreter.syncRoot; /* FIELD */

                        if (!Monitor.TryEnter(syncRoot))
                            continue;

                        try
                        {
                            //
                            // BUGFIX: Do not send any "global" notifications
                            //         to safe interpreters.  Also, do not
                            //         notify an interpreter if we cannot
                            //         obtain the lock for the interpreter OR
                            //         the "global" notification flag is
                            //         disabled for it.
                            //
                            if (otherInterpreter.IsSafe())
                                continue;

                            code = otherInterpreter.CheckNotifications(
                                type, flags, data, interpreter, clientData,
                                arguments, exception, ref result);

                            if (code != ReturnCode.Ok)
                                break;
                        }
                        finally
                        {
                            Monitor.Exit(syncRoot);
                        }
                    }
                }

                return code;
            }
            else
            {
                result = "no interpreters available";
            }

            return ReturnCode.Error;
        }
#endif
        #endregion
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region INotifyManager Members
#if NOTIFY || NOTIFY_OBJECT
        public NotifyType GlobalNotifyTypes
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return notifyTypes;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    notifyTypes = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public NotifyFlags GlobalNotifyFlags
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return notifyFlags;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    notifyFlags = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool GlobalNotify
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return notify;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    notify = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public NotifyType NotifyTypes
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
#if THREADING
                    IEngineContext context = (contextManager != null) ?
                        contextManager.GetEngineContext(true) : null;

                    if (context != null)
                        return notifyTypes | context.NotifyTypes;
                    else
                        return notifyTypes;
#else
                    return notifyTypes;
#endif
                }
            }
            set
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.NotifyTypes = value;
#else
                lock (syncRoot)
                {
                    notifyTypes = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public NotifyFlags NotifyFlags
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
#if THREADING
                    IEngineContext context = (contextManager != null) ?
                        contextManager.GetEngineContext(true) : null;

                    if (context != null)
                        return notifyFlags | context.NotifyFlags;
                    else
                        return notifyFlags;
#else
                    return notifyFlags;
#endif
                }
            }
            set
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.NotifyFlags = value;
#else
                lock (syncRoot)
                {
                    notifyFlags = value;
                }
#endif
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPackageManager Members
        public ReturnCode PresentPackage(
            string name,
            Version version,
            bool exact,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    IPackage package = null;

                    if ((GetPackage(name, true, true, ref package) == ReturnCode.Ok) &&
                        (package.Loaded != null) &&
                        RuntimeOps.VersionSatisfies(package.Loaded, version, exact))
                    {
                        result = package.Loaded;
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = String.Format(
                            "package \"{0}\" is not present",
                            FormatOps.PackageName(name, version));
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ProvidePackage(
            string name,
            Version version,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    IPackage package = null;

                    if (GetPackage(name, true, true, ref package) == ReturnCode.Ok)
                    {
                        if (version != null)
                        {
                            if (package.Loaded == null)
                            {
                                //
                                // NOTE: This version of the package is now "provided".
                                //
                                package.Loaded = version;

                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                            else if (RuntimeOps.VersionCompare(package.Loaded, version) != 0)
                            {
                                //
                                // NOTE: This is an error, a different version of this
                                //       package has already been provided.
                                //
                                result = String.Format(
                                    "conflicting versions provided for package \"{0}\": " +
                                    "{1} then {2} ", name, package.Loaded, version);
                            }
                            else
                            {
                                //
                                // NOTE: Do nothing, the identical version has already
                                //       been provided.
                                //
                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Do nothing, just return the version of the
                            //       package that has been provided.
                            //
                            if (package.Loaded != null)
                                result = package.Loaded;
                            else
                                result = String.Empty;

                            return ReturnCode.Ok;
                        }
                    }
                    else if (version != null)
                    {
                        //
                        // NOTE: *WARNING* Empty package names are allowed, please
                        //       do not change this to "!String.IsNullOrEmpty".
                        //
                        if (name != null)
                        {
                            //
                            // NOTE: We have never seen this package before.  Create
                            //       a new package and mark it as loaded.
                            //
                            if (AddPackage(RuntimeOps.NewCorePackage(name, null, null,
                                    ClientData.Empty, PackageFlags.None, version,
                                    null), ClientData.Empty, ref result) == ReturnCode.Ok)
                            {
                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            result = "invalid package name";
                        }
                    }
                    else
                    {
                        result = "invalid package version";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RequirePackage(
            string name,
            Version version,
            bool exact,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    IPackage package = null;

                    if (GetPackage(name, true, true, ref package) == ReturnCode.Ok)
                    {
                        ReturnCode code;

                        if (package.Loaded != null)
                        {
                            //
                            // NOTE: Skip trying to load the package, it is already
                            //       loaded.
                            //
                            code = ReturnCode.Ok;
                        }
                        else
                        {
                            //
                            // NOTE: Attempt to load the package using the provided
                            //       script.
                            //
                            Version best = null;

                            try
                            {
                                code = package.Select(
                                    PackagePreference.Default, ref best, ref result);

                                if (code == ReturnCode.Ok)
                                    code = package.Load(this, best, ref result);
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while loading package: {0}",
                                    e);

                                code = ReturnCode.Error;
                            }
                        }

                        //
                        // NOTE: Make sure the package did not fail to load.
                        //
                        if (code == ReturnCode.Ok)
                        {
                            if (package.Loaded != null)
                            {
                                if ((version == null) ||
                                    RuntimeOps.VersionSatisfies(package.Loaded, version, exact))
                                {
                                    result = package.Loaded;
                                }
                                else
                                {
                                    result = String.Format(
                                        "attempt to provide package \"{0}\" failed: " +
                                        "package \"{1}\" provided instead",
                                        FormatOps.PackageName(name, version),
                                        FormatOps.PackageName(name, package.Loaded));

                                    code = ReturnCode.Error;
                                }
                            }
                            else
                            {
                                result = String.Format(
                                    "attempt to provide package \"{0}\" failed: " +
                                    "no version of package \"{1}\" provided",
                                    FormatOps.PackageName(name, version), name);

                                code = ReturnCode.Error;
                            }
                        }

                        return code;
                    }
                    else
                    {
                        result = String.Format(
                            "can't find package \"{0}\"",
                            FormatOps.PackageName(name, version));
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode WithdrawPackage(
            string name,
            Version version,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    IPackage package = null;

                    if (GetPackage(name, true, true, ref package) == ReturnCode.Ok)
                    {
                        if (version != null)
                        {
                            if (package.Loaded != null)
                            {
                                if (RuntimeOps.VersionCompare(package.Loaded, version) == 0)
                                {
                                    //
                                    // NOTE: This version of the package is now "withdrawn".
                                    //
                                    package.Loaded = null;

                                    result = String.Empty;
                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    //
                                    // NOTE: This is an error, a different version of this
                                    //       package has been provided.
                                    //
                                    result = String.Format(
                                        "conflicting versions withdrawn for package \"{0}\": " +
                                        "{1} then {2} ", FormatOps.DisplayName(name),
                                        package.Loaded, version);
                                }
                            }
                            else
                            {
                                //
                                // NOTE: Do nothing, the package has already been
                                //       unprovided.
                                //
                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Do nothing, just return the version of the
                            //       package that has been provided.
                            //
                            if (package.Loaded != null)
                                result = package.Loaded;
                            else
                                result = String.Empty;

                            return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        result = String.Format(
                            "can't find package \"{0}\"",
                            FormatOps.PackageName(name, version));
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal PackageIndexDictionary PackageIndexes /* FOR [package] USE ONLY */
        {
            get { return packageIndexes; }
            set { packageIndexes = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ResetPkgIndexes(
            ref Result error /* NOT USED */
            )
        {
            lock (syncRoot)
            {
                if (packageIndexes != null)
                    packageIndexes.Clear();
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgIndexes(
            string pattern,
            bool noCase,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackageIndexes(ref result))
                {
                    result = packageIndexes.ToString(pattern, noCase);
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgLoaded(
            string pattern,
            bool noCase,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    //
                    // NOTE: Return a list of all loaded packages.
                    //
                    StringList list = new StringList();

                    foreach (KeyValuePair<string, _Wrappers.Package> pair in packages)
                        if ((pair.Value != null) && (pair.Value.Loaded != null))
                            list.Add(pair.Key);

                    result = list.ToString(pattern, noCase);
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgNames(
            string pattern,
            bool noCase,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    //
                    // NOTE: Return a list of all registered packages.
                    //
                    result = packages.ToString(pattern, noCase);
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PkgForget(
            string name
            )
        {
            Result result = null;

            return PkgForget(name, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgForget(
            string name,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    if (DoesPackageExist(name) == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Remove the package.
                        //
                        if (RemovePackage(name, null, ref result) == ReturnCode.Ok)
                            return ReturnCode.Ok;
                    }
                    else
                    {
                        result = String.Format(
                            "can't find package \"{0}\"",
                            FormatOps.PackageName(name, null));
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgForget(
            StringList names,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    if (names != null)
                    {
                        foreach (string name in names)
                            /* IGNORED */
                            PkgForget(name);

                        result = String.Empty;
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "invalid package list";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgIfNeeded(
            string name,
            Version version,
            string text,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    if (version != null)
                    {
                        if (text != null)
                        {
                            ReturnCode code;
                            IPackage package = null;

                            if (GetPackage(name, true, true, ref package) != ReturnCode.Ok)
                            {
                                //
                                // NOTE: *WARNING* Empty package names are allowed, please
                                //       do not change this to "!String.IsNullOrEmpty".
                                //
                                if (name != null)
                                {
                                    package = RuntimeOps.NewCorePackage(name, null, null,
                                        ClientData.Empty, PackageFlags.None, null, null);

                                    code = AddPackage(package, ClientData.Empty, ref result);
                                }
                                else
                                {
                                    result = "invalid package name";
                                    code = ReturnCode.Error;
                                }
                            }
                            else
                            {
                                //
                                // NOTE: Package is already present in the collection.
                                //
                                code = ReturnCode.Ok;
                            }

                            //
                            // NOTE: Make sure we added the package OK (or that we did not need to).
                            //
                            if (code == ReturnCode.Ok)
                            {
                                VersionStringDictionary ifNeeded = package.IfNeeded;

                                if (ifNeeded != null)
                                {
                                    if (ifNeeded.ContainsKey(version))
                                        ifNeeded[version] = text;
                                    else
                                        ifNeeded.Add(version, text);

                                    result = String.Empty;
                                }
                                else
                                {
                                    result = String.Format(
                                        "package \"{0}\" ifneeded scripts not available",
                                        name);

                                    code = ReturnCode.Error;
                                }
                            }

                            return code;
                        }
                        else
                        {
                            IPackage package = null;

                            if (GetPackage(name, true, true, ref package) == ReturnCode.Ok)
                            {
                                VersionStringDictionary ifNeeded = package.IfNeeded;

                                if (ifNeeded != null)
                                {
                                    if (ifNeeded.ContainsKey(version))
                                        result = ifNeeded[version];
                                    else
                                        result = String.Empty;

                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    result = String.Format(
                                        "package \"{0}\" ifneeded scripts not available",
                                        name);
                                }
                            }
                            else
                            {
                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                        }
                    }
                    else
                    {
                        result = "invalid package version";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode PkgVersions(
            string name,
            ref Result result
            )
        {
            lock (syncRoot)
            {
                if (HasPackages(ref result))
                {
                    IPackage package = null;

                    if (GetPackage(name, true, true, ref package) == ReturnCode.Ok)
                    {
                        VersionStringDictionary ifNeeded = package.IfNeeded;

                        if (ifNeeded != null)
                            result = ifNeeded.ToString();
                        else
                            result = String.Empty;
                    }
                    else
                    {
                        result = String.Empty;
                    }

                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPluginManager Members
        [Obsolete()]
        public ReturnCode LoadPlugin( /* NOT USED */
            AssemblyName assemblyName,
#if CAS_POLICY
            Evidence evidence,
#endif
            string typeName,
            IClientData clientData,
            PluginFlags flags,
            ref IPlugin plugin,
            ref Result result
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Ok;

            if (assemblyName != null)
            {
#if ISOLATED_PLUGINS
                bool isolated = FlagOps.HasFlags(flags, PluginFlags.Isolated, true);
#endif
                AppDomain appDomain = null;

#if ISOLATED_PLUGINS
                try
#endif
                {
                    code = RuntimeOps.GetOrCreatePluginAppDomain(
                        this,
                        FormatOps.AppDomainFriendlyName(assemblyName, typeName),
                        null,
#if CAS_POLICY
                        evidence,
#endif
                        clientData,
#if ISOLATED_PLUGINS
                        isolated,
#else
                        false,
#endif
                        true, ref appDomain, ref result);

                    if (code == ReturnCode.Ok)
                    {
#if NOTIFY
                        /* IGNORED */
                        CheckNotifications(
                            NotifyType.Plugin, NotifyFlags.PreLoad,
                            new ObjectList(appDomain, assemblyName, typeName, flags, plugin), this,
                            clientData, null, null, ref result);
#endif

                        Assembly assembly = null;
                        string fileName = null;

                        try
                        {
                            //
                            // BUGBUG: This will not work when using isolated AppDomains
                            //         (which is mostly fine because we never load plugins
                            //         via this method overload anyhow).
                            //
                            assembly = appDomain.Load(
                                assemblyName
#if CAS_POLICY
                                , evidence
#endif
                            );

                            fileName = assembly.Location;
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while loading plugin: {0}",
                                e);

                            code = ReturnCode.Error;
                        }

                        if (code == ReturnCode.Ok)
                            code = CreatePlugin(
                                appDomain, assembly, fileName, typeName,
                                clientData, flags, ref plugin, ref result);

#if NOTIFY
                        if (code == ReturnCode.Ok)
                        {
                            /* IGNORED */
                            CheckNotifications(
                                NotifyType.Plugin, NotifyFlags.Load,
                                new ObjectList(appDomain, fileName, typeName, flags, plugin), this,
                                clientData, null, null, ref result);
                        }
#endif
                    }
                }
#if ISOLATED_PLUGINS
                finally
                {
                    if ((code != ReturnCode.Ok) && isolated && (appDomain != null))
                    {
                        ReturnCode unloadCode;
                        Result unloadError = null;

                        unloadCode = RuntimeOps.UnloadAppDomain(
                            appDomain, clientData, ref unloadError);

                        if (unloadCode != ReturnCode.Ok)
                            DebugOps.Complain(this, unloadCode, unloadError);
                    }
                }
#endif
            }
            else
            {
                result = "invalid assembly name";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode LoadPlugin(
            string fileName,
#if CAS_POLICY
            Evidence evidence,
            byte[] hashValue,
            AssemblyHashAlgorithm hashAlgorithm,
#endif
            string typeName,
            IClientData clientData,
            PluginFlags flags,
            ref IPlugin plugin,
            ref Result result
            )
        {
            CheckDisposed();

            ReturnCode code = ReturnCode.Ok;

            if (!String.IsNullOrEmpty(fileName) && File.Exists(fileName))
            {
                if (FlagOps.HasFlags(flags, PluginFlags.VerifiedOnly, true) &&
                    (FlagOps.HasFlags(flags, PluginFlags.SkipVerified, true) ||
                    !RuntimeOps.IsStrongNameVerified(fileName, true)))
                {
                    result = "plugin file is not strong name signed or cannot be verified";
                    return ReturnCode.Error;
                }

                if (FlagOps.HasFlags(flags, PluginFlags.TrustedOnly, true) &&
                    (FlagOps.HasFlags(flags, PluginFlags.SkipTrusted, true) ||
                    !RuntimeOps.IsFileTrusted(fileName)))
                {
                    result = "plugin file is not Authenticode signed or cannot be trusted";
                    return ReturnCode.Error;
                }

#if ISOLATED_PLUGINS
                bool isolated = FlagOps.HasFlags(flags, PluginFlags.Isolated, true);
#endif
                AppDomain appDomain = null;

#if ISOLATED_PLUGINS
                try
#endif
                {
                    code = RuntimeOps.GetOrCreatePluginAppDomain(
                        this,
                        FormatOps.AppDomainFriendlyName(fileName, typeName),
                        Path.GetDirectoryName(fileName),
#if CAS_POLICY
                        evidence,
#endif
                        clientData,
#if ISOLATED_PLUGINS
                        isolated,
#else
                        false,
#endif
                        true, ref appDomain, ref result);

                    if (code == ReturnCode.Ok)
                    {
#if NOTIFY
                        /* IGNORED */
                        CheckNotifications(
                            NotifyType.Plugin, NotifyFlags.PreLoad,
                            new ObjectList(appDomain, fileName, typeName, flags, plugin), this,
                            clientData, null, null, ref result);
#endif

#if ISOLATED_PLUGINS
                        AssemblyName assemblyName = null;
                        Uri uri = null;
#endif
                        Assembly assembly = null;

                        try
                        {
                            PluginLoadHelper loadHelper = new PluginLoadHelper(
                                fileName, typeName,
#if CAS_POLICY
                                isolated ? null : evidence, hashValue, hashAlgorithm,
#endif
                                new PluginPropertyHelper(), true,
                                FlagOps.HasFlags(flags, PluginFlags.Verbose, true));

#if CAS_POLICY
#if !MONO && !MONO_HACKS
                            if (!CommonOps.Runtime.IsMono())
                                appDomain.DoCallBack(loadHelper.Load4);
                            else
#endif
                                appDomain.DoCallBack(loadHelper.Load2);
#else
                            appDomain.DoCallBack(loadHelper.Load1);
#endif

#if ISOLATED_PLUGINS
                            if (isolated)
                            {
                                //
                                // NOTE: Get the plugin property helper for the plugin being
                                //       loaded from the plugin load helper.
                                //
                                PluginPropertyHelper propertyHelper = loadHelper.PropertyHelper;

                                if (propertyHelper != null)
                                {
                                    //
                                    // NOTE: Grab the name of the resolved and loaded assembly
                                    //       for the plugin being loaded from the plugin property
                                    //       helper.
                                    //
                                    assemblyName = propertyHelper.AssemblyName;

                                    //
                                    // NOTE: Grab the full name of the type for this plugin from
                                    //       the plugin property helper.  This is needed to
                                    //       determine exactly which class we need to create for
                                    //       the plugin being loaded in case a plugin assembly
                                    //       contains multiple plugins.
                                    //
                                    typeName = propertyHelper.TypeName;

                                    //
                                    // NOTE: Grab the URI for the plugin being loaded from the
                                    //       plugin property helper.  This is needed to pass to
                                    //       the constructor of the plugin type.
                                    //
                                    uri = propertyHelper.Uri;

                                    //
                                    // NOTE: Grab the extra plugin flags from the plugin property
                                    //       helper and combine them with the ones specified by
                                    //       the caller.
                                    //
                                    flags |= propertyHelper.PluginFlags;
                                }
                                else
                                {
                                    //
                                    // NOTE: We require the plugin property helper to be available
                                    //       and cannot continue to load the plugin without it.
                                    //
                                    result = "invalid plugin property helper";
                                    code = ReturnCode.Error;
                                }
                            }
                            else
#endif
                            {
                                //
                                // NOTE: Grab the loaded assembly directly from the plugin load
                                //       helper.  This is only legal (i.e. it will only work) when
                                //       we load the plugin assembly into the current application
                                //       domain.
                                //
                                assembly = loadHelper.Assembly;
                            }
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while loading plugin: {0}",
                                e);

                            code = ReturnCode.Error;
                        }

                        if ((code == ReturnCode.Ok) &&
#if ISOLATED_PLUGINS
                            !isolated &&
#endif
                            String.IsNullOrEmpty(typeName))
                        {
                            //
                            // NOTE: Introspect the assembly for types that implement the IPlugin
                            //       interface and simply default to loading the first one marked
                            //       as 'Primary' that we find.
                            //
                            code = RuntimeOps.FindPrimaryPlugin(assembly, ref typeName, ref result);
                        }

                        if (code == ReturnCode.Ok)
                        {
#if ISOLATED_PLUGINS
                            if (isolated)
                                code = CreatePlugin(
                                    appDomain, assemblyName, fileName, typeName,
                                    uri, clientData, flags, ref plugin, ref result);
                            else
#endif
                                code = CreatePlugin(
                                    appDomain, assembly, fileName, typeName, clientData,
                                    flags, ref plugin, ref result);

#if NOTIFY
                            if (code == ReturnCode.Ok)
                            {
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Plugin, NotifyFlags.Load,
                                    new ObjectList(appDomain, fileName, typeName, flags, plugin), this,
                                    clientData, null, null, ref result);
                            }
#endif
                        }
                    }
                }
#if ISOLATED_PLUGINS
                finally
                {
                    if ((code != ReturnCode.Ok) && isolated && (appDomain != null))
                    {
                        ReturnCode unloadCode;
                        Result unloadError = null;

                        unloadCode = RuntimeOps.UnloadAppDomain(
                            appDomain, clientData, ref unloadError);

                        if (unloadCode != ReturnCode.Ok)
                            DebugOps.Complain(this, unloadCode, unloadError);
                    }
                }
#endif
            }
            else
            {
                result = "invalid assembly file name";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnloadPlugin(
            IPlugin plugin,
            IClientData clientData,
            PluginFlags flags,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin == null)
            {
                result = "invalid plugin";
                return ReturnCode.Error;
            }

            ReturnCode code;

#if ISOLATED_PLUGINS
            bool isolated = RuntimeOps.IsPluginIsolated(plugin);
            AppDomain appDomain = isolated ? plugin.AppDomain : null;
#endif

            GlobalState.PushActiveInterpreter(this);

            try
            {
#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Plugin, NotifyFlags.PreUnload,
                    new ObjectPair(plugin, flags), this,
                    clientData, null, null, ref result);
#endif

                code = plugin.Terminate(this, (clientData != null) ?
                    clientData : plugin.ClientData, ref result);

                if (code == ReturnCode.Ok)
                {
#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Plugin, NotifyFlags.Unload,
                        new ObjectPair(plugin, flags), this,
                        clientData, null, null, ref result);
#endif
                }
            }
            catch (Exception e)
            {
                result = String.Format(
                    "caught exception while terminating plugin: {0}",
                    e);

                code = ReturnCode.Error;
            }
            finally
            {
                /* IGNORED */
                GlobalState.PopActiveInterpreter();
            }

#if ISOLATED_PLUGINS
            //
            // NOTE: This cannot be done completely without full AppDomain
            //       isolation of loaded plugin Assemblies.  At that point,
            //       we can simply call AppDomain.Unload to unload the plugin.
            //
            if ((code == ReturnCode.Ok) && isolated)
                code = RuntimeOps.UnloadAppDomain(
                    appDomain, clientData, ref result);
#endif

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnloadPlugin(
            long token,
            IClientData clientData,
            PluginFlags flags, /* NOT USED */
            ref Result result
            )
        {
            CheckDisposed();

            ReturnCode code;
#if ISOLATED_PLUGINS
            IPlugin plugin = null;

            code = GetPlugin(token, true, true, ref plugin, ref result);

            if (code == ReturnCode.Ok)
#endif
            {
#if ISOLATED_PLUGINS
                bool isolated = RuntimeOps.IsPluginIsolated(plugin);
                AppDomain appDomain = isolated ? plugin.AppDomain : null;
#endif

                code = RemovePlugin(token, clientData, ref result);

#if ISOLATED_PLUGINS
                //
                // NOTE: This cannot be done completely without full AppDomain
                //       isolation of loaded plugin Assemblies.  At that point,
                //       we can simply call AppDomain.Unload to unload the plugin.
                //
                if ((code == ReturnCode.Ok) && isolated)
                    code = RuntimeOps.UnloadAppDomain(
                        appDomain, clientData, ref result);
#endif
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnloadPlugin(
            string name,
            IClientData clientData,
            PluginFlags flags, /* NOT USED */
            ref Result result
            )
        {
            CheckDisposed();

            ReturnCode code;
#if ISOLATED_PLUGINS
            IPlugin plugin = null;

            code = GetPlugin(name, true, true, ref plugin, ref result);

            if (code == ReturnCode.Ok)
#endif
            {
#if ISOLATED_PLUGINS
                bool isolated = RuntimeOps.IsPluginIsolated(plugin);
                AppDomain appDomain = isolated ? plugin.AppDomain : null;
#endif

                code = RemovePlugin(name, clientData, ref result);

#if ISOLATED_PLUGINS
                //
                // NOTE: This cannot be done completely without full AppDomain
                //       isolation of loaded plugin Assemblies.  At that point,
                //       we can simply call AppDomain.Unload to unload the plugin.
                //
                if ((code == ReturnCode.Ok) && isolated)
                    code = RuntimeOps.UnloadAppDomain(
                        appDomain, clientData, ref result);
#endif
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddCommands(
            IPlugin plugin,
            CommandFlags flags,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                AppDomain appDomain = plugin.AppDomain;

                if (appDomain != null)
                {
                    CommandDataList commands = plugin.Commands;

                    if (commands != null)
                    {
                        //
                        // NOTE: Are we adding commands in command
                        //       restoration mode?
                        //
                        bool restore = FlagOps.HasFlags(
                            flags, CommandFlags.Restore, true);

#if ISOLATED_PLUGINS
                        //
                        // NOTE: Is this plugin supposed to be loaded
                        //       into an isolated application domain?
                        //
                        bool isolated = RuntimeOps.IsPluginIsolated(plugin);
#endif

                        //
                        // NOTE: Is this interpreter marked as "safe"?
                        //
                        bool createSafe = FlagOps.HasFlags(
                            createFlags, CreateFlags.Safe, true);

                        //
                        // NOTE: Are we supposed to hide unsafe commands?
                        //
                        bool hideUnsafe = FlagOps.HasFlags(
                            createFlags, CreateFlags.HideUnsafe, true);

                        //
                        // NOTE: Is this interpreter marked as "standard"?
                        //
                        bool createStandard = FlagOps.HasFlags(
                            createFlags, CreateFlags.Standard, true);

                        //
                        // NOTE: Are we supposed to hide non-standard
                        //       commands?
                        //
                        bool hideNonStandard = FlagOps.HasFlags(
                            createFlags, CreateFlags.HideNonStandard, true);

                        //
                        // NOTE: Process each command in no particular order
                        //       (actually, the type of the command, which
                        //       currently equates to the same thing).
                        //
                        foreach (ICommandData commandData in commands)
                        {
                            //
                            // NOTE: Query the command flags associated with
                            //       this type.
                            //
                            CommandFlags commandFlags = commandData.Flags;

                            //
                            // NOTE: Check if this command is supposed to be
                            //       added to the interpreter.
                            //
                            bool noAdd = FlagOps.HasFlags(
                                commandFlags, CommandFlags.NoAdd, true);

                            //
                            // NOTE: If we are supposed to skip adding this
                            //       command altogether, then proceed no
                            //       further.
                            //
                            if (!noAdd)
                            {
                                string typeName = commandData.TypeName;

                                //
                                // NOTE: Check if this command has been marked as "safe" or
                                //       "unsafe" by its author.
                                //
                                bool safe = FlagOps.HasFlags(commandFlags, CommandFlags.Safe, true) &&
                                    !FlagOps.HasFlags(commandFlags, CommandFlags.Unsafe, true);

                                //
                                // NOTE: If the interpreter is "safe", make sure we only add
                                //       commands that are marked as "safe" and NOT marked as
                                //       "unsafe" unless we are merely hiding commands not
                                //       considered to be "safe".  In that case, we add all
                                //       commands, hidden them as necessary.
                                //
                                if (safe || !createSafe || hideUnsafe)
                                {
                                    //
                                    // NOTE: Check if this command is part of the "Tcl Standard"
                                    //       (i.e. it is largely (or completely) compatible with
                                    //       an identically named command from Tcl/Tk 8.4, 8.5,
                                    //       and/or 8.6.)
                                    //
                                    bool standard = FlagOps.HasFlags(
                                        commandFlags, CommandFlags.Standard, true);

                                    //
                                    // NOTE: If the interpreter is "standard", make sure we only
                                    //       add commands that are marked as "standard" and NOT
                                    //       marked as "non-standard" unless we are merely hiding
                                    //       commands not considered to be "standard".  In that
                                    //       case, we add all commands, hidden them as necessary.
                                    //
                                    if (standard || !createStandard || hideNonStandard)
                                    {
                                        //
                                        // NOTE: Is the command named "Default" or "Core"?  If
                                        //       so, we need to skip over it because they are
                                        //       used as the base implementation for other
                                        //       commands.
                                        //
                                        if ((String.Compare(typeName, typeof(_Commands.Default).FullName,
                                                StringOps.SystemStringComparisonType) != 0) &&
                                            (String.Compare(typeName, typeof(_Commands._Delegate).FullName,
                                                StringOps.SystemStringComparisonType) != 0) &&
                                            (String.Compare(typeName, typeof(_Commands.Core).FullName,
                                                StringOps.SystemStringComparisonType) != 0) &&
                                            (String.Compare(typeName, typeof(_Commands.Alias).FullName,
                                                StringOps.SystemStringComparisonType) != 0))
                                        {
                                            //
                                            // NOTE: Grab the command name and make sure it is
                                            //       valid.  For now, we just ignore invalid
                                            //       command names (i.e. we skip adding the
                                            //       command).
                                            //
                                            string name = commandData.Name;

                                            if (name != null)
                                            {
                                                //
                                                // HACK: All core command names are lowercase;
                                                //       culture is invariant because these are
                                                //       "system" identifiers.
                                                //
                                                name = name.ToLowerInvariant();

                                                //
                                                // HACK: Remove leading underscore from core
                                                //       command names to accommodate the special
                                                //       circumstance where we were using a
                                                //       leading underscore to get around .NET
                                                //       Framework "reserved" type names (e.g.
                                                //       File, String, Object, etc).
                                                //
                                                if ((name.Length > 0) &&
                                                    (name[0] == Characters.Underscore))
                                                {
                                                    name = name.Substring(1);
                                                }

                                                //
                                                // NOTE: Always try to add the command unless we
                                                //       are in "restoration" mode, in that case
                                                //       only try to add the command if it does
                                                //       not already exist.
                                                //
                                                bool add = true;

                                                if (restore &&
                                                    (DoesCommandExist(name) == ReturnCode.Ok))
                                                {
                                                    add = false;
                                                }

                                                //
                                                // NOTE: Do we still want to add this command?
                                                //
                                                if (add)
                                                {
                                                    //
                                                    // NOTE: Start with the flags provided by the
                                                    //       caller and then mask off the command
                                                    //       restoration flag.
                                                    //
                                                    CommandFlags newFlags =
                                                        flags & ~CommandFlags.Restore;

                                                    //
                                                    // NOTE: If this command is not considered
                                                    //       "safe" and the caller instructed us
                                                    //       to hide such commands, do so now.
                                                    //
                                                    if (!safe && hideUnsafe)
                                                        newFlags |= CommandFlags.Hidden;
                                                    //
                                                    // NOTE: Otherwise, if this command is not
                                                    //       considered to be "standard" and the
                                                    //       caller instructed us to hide such
                                                    //       commands, do so now.
                                                    //
                                                    else if (!standard && hideNonStandard)
                                                        newFlags |= CommandFlags.Hidden;

#if ISOLATED_PLUGINS
                                                    //
                                                    // NOTE: If the plugin has been loaded into
                                                    //       an isolated AppDomain, the command
                                                    //       should be marked isolated as well.
                                                    //
                                                    if (isolated)
                                                        newFlags |= CommandFlags.Isolated;
#endif

                                                    //
                                                    // NOTE: Setup the argument array for the
                                                    //       command constructor.  The command
                                                    //       MUST expose a constructor that
                                                    //       accepts this argument type.  It
                                                    //       is also free to expose other
                                                    //       constructors; however, they will
                                                    //       not be used by the script engine.
                                                    //
                                                    object[] args = { new CommandData(name,
                                                        commandData.Group, commandData.Description,
                                                        commandData.ClientData, commandData.TypeName,
                                                        newFlags, commandData.Plugin, commandData.Token)
                                                    };

                                                    //
                                                    // NOTE: Attempt to create an instance of
                                                    //       the command in the application
                                                    //       domain for the plugin.
                                                    //
                                                    ReturnCode code = ReturnCode.Ok;
                                                    ICommand command = null;

#if ISOLATED_PLUGINS
                                                    if (isolated)
                                                    {
                                                        AssemblyName assemblyName = plugin.AssemblyName;

                                                        if (assemblyName != null)
                                                        {
                                                            command = (ICommand)appDomain.CreateInstanceAndUnwrap(
                                                                assemblyName.ToString(), typeName, false,
                                                                RuntimeOps.CreateInstanceBindingFlags, null, args,
                                                                null, null, null);
                                                        }
                                                        else
                                                        {
                                                            result = "plugin has invalid assembly name";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
#endif
                                                    {
                                                        Assembly assembly = plugin.Assembly;

                                                        if (assembly != null)
                                                        {
                                                            command = (ICommand)assembly.CreateInstance(
                                                                typeName, false,
                                                                RuntimeOps.CreateInstanceBindingFlags,
                                                                null, args, null, null);
                                                        }
                                                        else
                                                        {
                                                            result = "plugin has invalid assembly";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }

                                                    if (code == ReturnCode.Ok)
                                                    {
                                                        if (command != null)
                                                        {
                                                            //
                                                            // NOTE: Attempt to assign the the ObjectId
                                                            //       of the command to the Id property,
                                                            //       if necessary.
                                                            //
                                                            // FIXME: Maybe this should be moved into
                                                            //        AddCommand and similar functionality
                                                            //        added to AddFunction, etc?
                                                            //
                                                            if (command.Id.Equals(Guid.Empty))
                                                                command.Id = AttributeOps.GetObjectId(command);

                                                            //
                                                            // NOTE: Add the command to the interpreter.
                                                            //
                                                            long token = 0;

                                                            code = AddCommand(command, clientData, ref token, ref result);

                                                            if (code == ReturnCode.Ok)
                                                                commandData.Token = token;
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "could not create command \"{0}\"",
                                                                typeName);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }

                                                    if (code != ReturnCode.Ok)
                                                        return code;
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }

#if ISOLATED_PLUGINS
                        //
                        // NOTE: Update the command data list for isolated plugins.
                        //
                        if (isolated)
                            plugin.Commands = commands;
#endif

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "plugin has invalid command data list";
                    }
                }
                else
                {
                    result = "plugin has invalid application domain";
                }
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveCommands(
            IPlugin plugin,
            CommandFlags flags,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                LongList localTokens = plugin.CommandTokens;

                if (localTokens != null)
                    localTokens = (LongList)localTokens.Clone();

                if (localTokens != null)
                {
                    foreach (long token in localTokens)
                    {
                        if (DoesCommandExist(token) == ReturnCode.Ok)
                        {
                            ReturnCode code = RemoveCommand(
                                token, clientData, ref result);

                            if (code != ReturnCode.Ok)
                                return code;
                        }
                        else
                        {
                            //
                            // NOTE: We must actually ignore "failures" here because,
                            //       in theory, the command could have been removed
                            //       using the [rename] script command or some other
                            //       mechanism.
                            //
                        }
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveFunctions(
            IPlugin plugin,
            FunctionFlags flags,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                LongList localTokens = plugin.FunctionTokens;

                if (localTokens != null)
                    localTokens = (LongList)localTokens.Clone();

                if (localTokens != null)
                {
                    foreach (long token in localTokens)
                    {
                        if (DoesFunctionExist(token) == ReturnCode.Ok)
                        {
                            ReturnCode code = RemoveFunction(
                                token, clientData, ref result);

                            if (code != ReturnCode.Ok)
                                return code;
                        }
                        else
                        {
                            //
                            // NOTE: We must actually ignore "failures" here because,
                            //       in theory, the function could have been removed
                            //       using some other mechanism.
                            //
                        }
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddPolicies(
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                AppDomain appDomain = plugin.AppDomain;

                if (appDomain != null)
                {
                    PolicyDataList policies = plugin.Policies;

                    if (policies != null)
                    {
#if ISOLATED_PLUGINS
                        //
                        // NOTE: Is this plugin supposed to be loaded
                        //       into an isolated application domain?
                        //
                        bool isolated = RuntimeOps.IsPluginIsolated(plugin);
#endif

                        //
                        // NOTE: Process each policy in no particular order
                        //       (actually, the type and method name of the
                        //       policy, which currently equates to the same
                        //       thing).
                        //
                        foreach (IPolicyData policyData in policies)
                        {
                            //
                            // NOTE: Query the flags associated with this
                            //       policy method.
                            //
                            MethodFlags methodFlags = policyData.MethodFlags;

                            //
                            // NOTE: Check if this policy is supposed to be
                            //       added to the interpreter.
                            //
                            bool noAdd = FlagOps.HasFlags(
                                methodFlags, MethodFlags.NoAdd, true);

                            //
                            // NOTE: If we are supposed to skip adding this
                            //       policy altogether, then proceed no
                            //       further.
                            //
                            if (!noAdd)
                            {
                                string typeName =
                                    typeof(Eagle._Policies.Core).FullName;

                                //
                                // NOTE: Setup the argument array for the
                                //       policy constructor.  The policy
                                //       MUST expose a constructor that
                                //       accepts this argument type.  It
                                //       is also free to expose other
                                //       constructors; however, they will
                                //       not be used by the script engine.
                                //
                                object[] args = { new PolicyData(
                                    policyData.Name, policyData.Group,
                                    policyData.Description, clientData,
                                    policyData.TypeName, policyData.MethodName,
                                    policyData.BindingFlags, policyData.MethodFlags,
                                    policyData.PolicyFlags, policyData.Plugin,
                                    policyData.Token)
                                };

                                //
                                // NOTE: Attempt to create an instance of
                                //       the policy in the application
                                //       domain for the plugin.
                                //
                                ReturnCode code = ReturnCode.Ok;
                                IPolicy policy = null;

#if ISOLATED_PLUGINS
                                if (isolated)
                                {
                                    AssemblyName assemblyName = GlobalState.GetAssemblyName();

                                    if (assemblyName != null)
                                    {
                                        policy = (IPolicy)appDomain.CreateInstanceAndUnwrap(
                                            assemblyName.ToString(), typeName, false,
                                            RuntimeOps.CreateInstanceBindingFlags, null, args,
                                            null, null, null);
                                    }
                                    else
                                    {
                                        //
                                        // HACK: This error message is not quite correct;
                                        //       however, it's close enough for now.
                                        //
                                        result = "plugin has invalid assembly name";
                                        code = ReturnCode.Error;
                                    }
                                }
                                else
#endif
                                {
                                    Assembly assembly = GlobalState.GetAssembly();

                                    if (assembly != null)
                                    {
                                        policy = (IPolicy)assembly.CreateInstance(
                                            typeName, false,
                                            RuntimeOps.CreateInstanceBindingFlags,
                                            null, args, null, null);
                                    }
                                    else
                                    {
                                        //
                                        // HACK: This error message is not quite correct;
                                        //       however, it's close enough for now.
                                        //
                                        result = "plugin has invalid assembly";
                                        code = ReturnCode.Error;
                                    }
                                }

                                if (code == ReturnCode.Ok)
                                {
                                    if (policy != null)
                                    {
                                        //
                                        // NOTE: First, attempt to assign the the ObjectId
                                        //       of the policy to the Id property, if
                                        //       necessary.
                                        //
                                        // FIXME: Maybe this should be moved into AddPolicy
                                        //        and similar functionality added to
                                        //        AddFunction, etc?
                                        //
                                        if (policy.Id.Equals(Guid.Empty))
                                            policy.Id = AttributeOps.GetObjectId(policy);

                                        //
                                        // NOTE: Next, make sure the actual delegate is setup.
                                        //
                                        code = policy.Setup(ref result);

                                        //
                                        // NOTE: Finally, add the policy to the interpreter.
                                        //
                                        if (code == ReturnCode.Ok)
                                        {
                                            long token = 0;

                                            code = AddPolicy(policy, clientData, ref token, ref result);

                                            if (code == ReturnCode.Ok)
                                                policyData.Token = token;
                                        }
                                    }
                                    else
                                    {
                                        result = String.Format(
                                            "could not create policy \"{0}\"",
                                            typeName);

                                        code = ReturnCode.Error;
                                    }
                                }

                                if (code != ReturnCode.Ok)
                                    return code;
                            }
                        }

#if ISOLATED_PLUGINS
                        //
                        // NOTE: Update the policy data list for isolated plugins.
                        //
                        if (isolated)
                            plugin.Policies = policies;
#endif

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        result = "plugin has invalid policy data list";
                    }
                }
                else
                {
                    result = "plugin has invalid application domain";
                }
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemovePolicies(
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                LongList localTokens = (LongList)plugin.PolicyTokens;

                if (localTokens != null)
                    localTokens = (LongList)localTokens.Clone();

                if (localTokens != null)
                {
                    foreach (long token in localTokens)
                    {
                        if (DoesPolicyExist(token) == ReturnCode.Ok)
                        {
                            ReturnCode code = RemovePolicy(
                                token, clientData, ref result);

                            if (code != ReturnCode.Ok)
                                return code;
                        }
                        else
                        {
                            //
                            // NOTE: We should actually ignore "failures" here because,
                            //       in theory, the policy could have been removed using
                            //       some other mechanism.
                            //
                        }
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveTraces(
            IPlugin plugin,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            if (plugin != null)
            {
                LongList localTokens = (LongList)plugin.TraceTokens;

                if (localTokens != null)
                    localTokens = (LongList)localTokens.Clone();

                if (localTokens != null)
                {
                    foreach (long token in localTokens)
                    {
                        if (DoesTraceExist(token) == ReturnCode.Ok)
                        {
                            ReturnCode code = RemoveTrace(
                                token, clientData, ref result);

                            if (code != ReturnCode.Ok)
                                return code;
                        }
                        else
                        {
                            //
                            // NOTE: We should actually ignore "failures" here because,
                            //       in theory, the trace could have been removed using
                            //       some other mechanism.
                            //
                        }
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                result = "invalid plugin";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RestoreCorePlugin(
            bool strict,
            ref Result result
            )
        {
            CheckDisposed();

            IPlugin plugin = null;

            if (corePluginToken != 0)
            {
                if ((GetPlugin(corePluginToken, true, true, ref plugin, ref result) == ReturnCode.Ok) &&
                    (AddCommands(plugin, CommandFlags.Restore, null, ref result) == ReturnCode.Ok) &&
                    (MoveExposedAndHiddenCommands(ref result) == ReturnCode.Ok) &&
                    (!strict || (SetupFunctions(createFlags, ref result) == ReturnCode.Ok)))
                {
                    return ReturnCode.Ok;
                }
            }
            else
            {
                plugin = NewCorePlugin();

                if ((RuntimeOps.PrepareStaticPlugin(plugin, ref result) == ReturnCode.Ok) &&
                    (AddPlugin(plugin, null, ref corePluginToken, ref result) == ReturnCode.Ok) &&
                    (!strict || (SetupFunctions(createFlags, ref result) == ReturnCode.Ok)))
                {
                    return ReturnCode.Ok;
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        private static ReturnCode CreatePlugin(
            AppDomain appDomain,
            AssemblyName assemblyName,
            string fileName,
            string typeName,
            Uri uri,
            IClientData clientData,
            PluginFlags flags,
            ref IPlugin plugin,
            ref Result error
            )
        {
            ReturnCode code;

            if (appDomain != null)
            {
                if (assemblyName != null)
                {
                    if (typeName != null)
                    {
                        if (uri != null)
                        {
                            try
                            {
                                //
                                // NOTE: Setup the arguments for the plugin constructor.
                                //       Unfortunately, constructors cannot be part of an interface;
                                //       however, we require all plugins to expose a constructor with
                                //       argument types compatible with the values we supply here.
                                //       If a plugin does not comply with this requirement then it
                                //       cannot be loaded successfully by the plugin loader and will
                                //       not be available for use by scripts.
                                //
                                // WARNING: Please do not change this because it will prevent any
                                //          plugins from loading that have not been updated to use
                                //          the new argument list.
                                //
                                object[] args = { new PluginData(
                                    FormatOps.PluginName(assemblyName.FullName, typeName),
                                    null, null, clientData, flags, assemblyName.Version, uri,
                                    null, null, assemblyName, fileName, null, null, null, null,
                                    null, null, null, 0)
                                };

                                //
                                // NOTE: Attempt to create an instance of the plugin.  If this throws
                                //       an exception then the plugin probably does not have a
                                //       constructor matching the arguments we are attempting to
                                //       supply; therefore, it is not in compliance with the plugin
                                //       loader interface and cannot be loaded.  Another common cause
                                //       of failures here is that the plugin has a different version
                                //       of the Eagle runtime library specified in its manifest than
                                //       the one currently running.
                                //
                                plugin = (IPlugin)appDomain.CreateInstanceAndUnwrap(
                                    assemblyName.ToString(), typeName, false,
                                    RuntimeOps.CreateInstanceBindingFlags, null, args,
                                    null, null, null);

                                if (plugin != null)
                                {
                                    //
                                    // NOTE: All plugins must have an ObjectId attribute.  Attempt
                                    //       to assign the ObjectId attribute to the Id property now.
                                    //
                                    if (plugin.Id.Equals(Guid.Empty))
                                        plugin.Id = AttributeOps.GetObjectId(plugin);

                                    //
                                    // NOTE: We need the plugin to have a valid assembly within it so
                                    //       that the command and policy loaders will be able to use
                                    //       it.
                                    //
                                    PluginCreateHelper createHelper =
                                        new PluginCreateHelper(plugin, assemblyName);

                                    appDomain.DoCallBack(createHelper.Setup);

                                    code = ReturnCode.Ok;
                                }
                                else
                                {
                                    error = "invalid plugin";
                                    code = ReturnCode.Error;
                                }
                            }
                            catch (Exception e)
                            {
                                error = e;
                                code = ReturnCode.Error;
                            }
                        }
                        else
                        {
                            error = "invalid uri";
                            code = ReturnCode.Error;
                        }
                    }
                    else
                    {
                        error = "invalid type name";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    error = "invalid assembly name";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                error = "invalid application domain";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode CreatePlugin(
            AppDomain appDomain,
            Assembly assembly,
            string fileName,
            string typeName,
            IClientData clientData,
            PluginFlags flags,
            ref IPlugin plugin,
            ref Result error
            )
        {
            ReturnCode code;

            if (appDomain != null)
            {
                if (assembly != null)
                {
                    AssemblyName assemblyName = assembly.GetName();

                    if (assemblyName != null)
                    {
                        if (typeName != null)
                        {
                            try
                            {
                                Type type = assembly.GetType(typeName, true, false);

                                if (type != null)
                                {
                                    //
                                    // NOTE: Get the plugin flags related to the assembly itself and
                                    //       combine them with the ones we already have.
                                    //
                                    flags |= RuntimeOps.GetAssemblyPluginFlags(assembly,
                                        RuntimeOps.GetSkipCheckPluginFlags());

                                    //
                                    // NOTE: Check if the plugin assembly contains a an AssemblyUri
                                    //       attribute.
                                    //
                                    Uri uri = AttributeOps.GetAssemblyUri(assembly);

                                    //
                                    // NOTE: If the plugin assembly does NOT contain an AssemblyUri
                                    //       attribute, use the location.
                                    //
                                    if (uri == null)
                                        uri = new Uri(assembly.Location); /* throw */

                                    //
                                    // NOTE: Setup the arguments for the plugin constructor.
                                    //       Unfortunately, constructors cannot be part of an interface;
                                    //       however, we require all plugins to expose a constructor with
                                    //       argument types compatible with the values we supply here.
                                    //       If a plugin does not comply with this requirement then it
                                    //       cannot be loaded successfully by the plugin loader and will
                                    //       not be available for use by scripts.
                                    //
                                    // WARNING: Please do not change this because it will prevent any
                                    //          plugins from loading that have not been updated to use
                                    //          the new argument list.
                                    //
                                    object[] args = { new PluginData(
                                        FormatOps.PluginName(assemblyName.FullName, typeName),
                                        null, null, clientData, flags, assemblyName.Version, uri,
                                        appDomain, assembly, assemblyName, fileName, null, null,
                                        null, null, null, null, null, 0)
                                    };

                                    //
                                    // NOTE: Attempt to create an instance of the plugin.  If this throws
                                    //       an exception then the plugin probably does not have a
                                    //       constructor matching the arguments we are attempting to
                                    //       supply; therefore, it is not in compliance with the plugin
                                    //       loader interface and cannot be loaded.  Another common cause
                                    //       of failures here is that the plugin has a different version
                                    //       of the Eagle runtime library specified in its manifest than
                                    //       the one currently running.
                                    //
                                    plugin = (IPlugin)assembly.CreateInstance(type.FullName, false,
                                        RuntimeOps.CreateInstanceBindingFlags, null, args, null, null);

                                    if (plugin != null)
                                    {
                                        //
                                        // NOTE: All plugins must have an ObjectId attribute.  Attempt
                                        //       to assign the ObjectId attribute to the Id property now.
                                        //
                                        if (plugin.Id.Equals(Guid.Empty))
                                            plugin.Id = AttributeOps.GetObjectId(plugin);

                                        code = RuntimeOps.PopulatePluginCommands(plugin, ref error);

                                        if (code == ReturnCode.Ok)
                                            code = RuntimeOps.PopulatePluginPolicies(plugin, ref error);
                                    }
                                    else
                                    {
                                        error = "invalid plugin";
                                        code = ReturnCode.Error;
                                    }
                                }
                                else
                                {
                                    error = "type not found";
                                    code = ReturnCode.Error;
                                }
                            }
                            catch (Exception e)
                            {
                                error = e;
                                code = ReturnCode.Error;
                            }
                        }
                        else
                        {
                            error = "invalid type name";
                            code = ReturnCode.Error;
                        }
                    }
                    else
                    {
                        error = "invalid assembly name";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    error = "invalid assembly";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                error = "invalid application domain";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetPluginCommands( /* POLICY HELPER METHOD */
            IPlugin plugin,
            ref List<ICommand> commands,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (plugin != null)
                {
                    if (HasCommands(ref error))
                    {
                        List<ICommand> localCommands = new List<ICommand>();

                        foreach (ICommandData commandData in plugin.Commands)
                        {
                            ICommand command = null;

                            if (GetCommand(commandData.Token, true, true,
                                    ref command, ref error) == ReturnCode.Ok)
                            {
                                localCommands.Add(command);
                            }
                            else
                            {
                                return ReturnCode.Error;
                            }
                        }

                        if (commands == null)
                            commands = new List<ICommand>(localCommands);
                        else
                            commands.AddRange(localCommands);

                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = "invalid plugin";
                }

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal IPlugin GetCorePlugin(
            ref Result error
            )
        {
            if (corePluginToken != 0)
            {
                IPlugin plugin = null;

                if (GetPlugin(corePluginToken, true, true,
                        ref plugin, ref error) == ReturnCode.Ok)
                {
                    return plugin;
                }
            }
            else
            {
                error = "invalid core plugin token";
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private IPlugin NewCorePlugin()
        {
            Assembly assembly = GlobalState.GetAssembly();
            PluginFlags flags = RuntimeOps.GetAssemblyPluginFlags(assembly,
                RuntimeOps.GetSkipCheckPluginFlags());

            return new _Plugins.Core(new PluginData(
                FormatOps.PluginName(GlobalState.GetAssemblyFullName(),
                    typeof(_Plugins.Core).FullName),
                null, null, ClientData.Empty, flags,
                GlobalState.GetAssemblyVersion(), GlobalState.GetAssemblyUri(),
                appDomain, assembly, GlobalState.GetAssemblyName(),
                GlobalState.GetAssemblyLocation(), null, null, null, null, null, null,
                resourceManager, 0));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY || NOTIFY_OBJECT
        private IPlugin NewObjectPlugin()
        {
            Assembly assembly = GlobalState.GetAssembly();
            PluginFlags flags = RuntimeOps.GetAssemblyPluginFlags(assembly,
                RuntimeOps.GetSkipCheckPluginFlags());

            return new _Plugins.Object(new PluginData(
                FormatOps.PluginName(GlobalState.GetAssemblyFullName(),
                    typeof(_Plugins.Object).FullName),
                null, null, ClientData.Empty, flags,
                GlobalState.GetAssemblyVersion(), GlobalState.GetAssemblyUri(),
                appDomain, assembly, GlobalState.GetAssemblyName(),
                GlobalState.GetAssemblyLocation(), null, null, null, null, null, null,
                resourceManager, 0));
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUG
        internal IPlugin GetTestPlugin(
            ref Result error
            )
        {
            if (testPluginToken != 0)
            {
                IPlugin plugin = null;

                if (GetPlugin(testPluginToken, true, true,
                        ref plugin, ref error) == ReturnCode.Ok)
                {
                    return plugin;
                }
            }
            else
            {
                error = "invalid test plugin token";
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private IPlugin NewTestPlugin()
        {
            Assembly assembly = GlobalState.GetAssembly();
            PluginFlags flags = RuntimeOps.GetAssemblyPluginFlags(assembly,
                RuntimeOps.GetSkipCheckPluginFlags());

            return new _Plugins.Test(new PluginData(
                FormatOps.PluginName(GlobalState.GetAssemblyFullName(),
                    typeof(_Plugins.Test).FullName),
                null, null, ClientData.Empty, flags,
                GlobalState.GetAssemblyVersion(), GlobalState.GetAssemblyUri(),
                appDomain, assembly, GlobalState.GetAssemblyName(),
                GlobalState.GetAssemblyLocation(), null, null, null, null, null, null,
                resourceManager, 0));
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region PluginPropertyHelper Class (MarshalByRefObject)
#if SERIALIZATION
        [Serializable()]
#endif
        [ObjectId("0a332e4c-1619-45ce-b624-14099e37cb7a")]
        private sealed class PluginPropertyHelper
#if ISOLATED_INTERPRETERS || ISOLATED_PLUGINS
            : ScriptMarshalByRefObject
#endif
        {
            #region Private Data
            private AssemblyName assemblyName;
            private string fileName;
            private string typeName;
            private Uri uri;
            private PluginFlags pluginFlags;
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Constructors
            public PluginPropertyHelper()
            {
                // do nothing.
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Properties
            public AssemblyName AssemblyName
            {
                get { return assemblyName; }
                set { assemblyName = value; }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            public string FileName
            {
                get { return fileName; }
                set { fileName = value; }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            public string TypeName
            {
                get { return typeName; }
                set { typeName = value; }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            public Uri Uri
            {
                get { return uri; }
                set { uri = value; }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            public PluginFlags PluginFlags
            {
                get { return pluginFlags; }
                set { pluginFlags = value; }
            }
            #endregion
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region PluginCreateHelper Class (Serializable)
#if SERIALIZATION
        [Serializable()]
#endif
        [ObjectId("44989f7c-828c-4170-8655-404ee607460e")]
        private sealed class PluginCreateHelper
        {
            #region Private Constants
            private const string InvalidPluginError = "invalid plugin";
            private const string InvalidAssemblyNameError = "invalid assembly name";
            private const string PluginAppDomainAlreadySetError = "plugin application domain already set";
            private const string PluginAssemblyAlreadySetError = "plugin assembly name already set";
            private const string CouldNotPopulatePluginCommandsError = "could not populate plugin commands";
            private const string CouldNotPopulatePluginPoliciesError = "could not populate plugin policies";
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Private Data
            private IPlugin plugin; /* in */
            private AssemblyName assemblyName; /* in */
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Constructors
            public PluginCreateHelper(
                IPlugin plugin,
                AssemblyName assemblyName
                )
            {
                this.plugin = plugin;
                this.assemblyName = assemblyName;
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Methods
            public void Setup()
            {
                if (plugin == null)
                    throw new InvalidOperationException(InvalidPluginError);

                if (assemblyName == null)
                    throw new InvalidOperationException(InvalidAssemblyNameError);

                if (plugin.AppDomain != null)
                    throw new InvalidOperationException(PluginAppDomainAlreadySetError);

                if (plugin.Assembly != null)
                    throw new InvalidOperationException(PluginAssemblyAlreadySetError);

                //
                // NOTE: Setup the application domain and assembly for the
                //       plugin.
                //
                AppDomain appDomain = AppDomain.CurrentDomain;

                plugin.AppDomain = appDomain;
                plugin.Assembly = AssemblyOps.FindInAppDomain(
                    appDomain, assemblyName);

                //
                // NOTE: Get the command and policy type metadata for this
                //       plugin now because this cannot be done from the
                //       primary application domain.
                //
                ReturnCode code;
                Result error = null;

                code = RuntimeOps.PopulatePluginCommands(plugin, ref error);

                if (code != ReturnCode.Ok)
                    throw new TypeLoadException(
                        CouldNotPopulatePluginCommandsError,
                        new ScriptException(code, error));

                code = RuntimeOps.PopulatePluginPolicies(plugin, ref error);

                if (code != ReturnCode.Ok)
                    throw new TypeLoadException(
                        CouldNotPopulatePluginPoliciesError,
                        new ScriptException(code, error));
            }
            #endregion
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region PluginLoadHelper Class (Serializable)
#if SERIALIZATION
        [Serializable()]
#endif
        [ObjectId("532f533d-6b21-45d8-8139-54c7a4be4dbd")]
        private sealed class PluginLoadHelper
        {
            #region Private Constants
            private const string AssemblyNotLoadedError = "assembly not loaded";
            private const string AssemblyAlreadyLoadedError = "assembly already loaded";
            private const string AssemblyNameAlreadySetError = "assembly name already set";
            private const string TypeAlreadySetError = "type already set";
            private const string PrimaryPluginTypeNotFoundError = "primary plugin type not found";
            private const string UriAlreadySetError = "uri already set";
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Private Data
            private string fileName; /* in */
            private string typeName; /* in, out */

#if CAS_POLICY
#if SERIALIZATION
            [NonSerialized()]
#endif
            private Evidence evidence; /* in */

            private byte[] hashValue; /* in */
            private AssemblyHashAlgorithm hashAlgorithm; /* in */
#endif

            private PluginPropertyHelper propertyHelper; /* in, out */
            private bool setup; /* in */
            private bool verbose; /* in */

#if SERIALIZATION
            [NonSerialized()]
#endif
            private Assembly assembly; /* out */

#if SERIALIZATION
            [NonSerialized()]
#endif
            private Type type; /* out */

            private AssemblyName assemblyName; /* out */
            private Uri uri; /* out */
            private PluginFlags pluginFlags; /* out */
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Constructors
            public PluginLoadHelper(
                string fileName,
                string typeName,
#if CAS_POLICY
                Evidence evidence,
                byte[] hashValue,
                AssemblyHashAlgorithm hashAlgorithm,
#endif
                PluginPropertyHelper propertyHelper,
                bool setup,
                bool verbose
                )
            {
                this.fileName = fileName;
                this.typeName = typeName;

#if CAS_POLICY
                this.evidence = evidence;
                this.hashValue = hashValue;
                this.hashAlgorithm = hashAlgorithm;
#endif

                this.propertyHelper = propertyHelper;
                this.setup = setup;
                this.verbose = verbose;
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Private Methods
            private void Setup() /* throw */
            {
                if (verbose)
                    DebugOps.DumpAppDomain(AppDomain.CurrentDomain);

                SetupAssemblyName(); /* throw */
                SetupTypeName(); /* throw */
                SetupUri(); /* throw */
                SetupAssemblyPluginFlags(); /* throw */

                //
                // NOTE: If the property helper was provided, use it to send our
                //       output data back to the other application domain.
                //
                if (propertyHelper != null)
                {
                    propertyHelper.AssemblyName = assemblyName;
                    propertyHelper.FileName = fileName;
                    propertyHelper.TypeName = typeName;
                    propertyHelper.Uri = uri;
                    propertyHelper.PluginFlags = pluginFlags;
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            private void SetupAssemblyName()
            {
                if (assembly == null)
                    throw new InvalidOperationException(AssemblyNotLoadedError);

                if (assemblyName != null)
                    throw new InvalidOperationException(AssemblyNameAlreadySetError);

                //
                // NOTE: Get the full name of the assembly and store it for
                //       later use by the plugin loader.
                //
                assemblyName = assembly.GetName();
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            private void SetupAssemblyPluginFlags()
            {
                if (assembly == null)
                    throw new InvalidOperationException(AssemblyNotLoadedError);

                pluginFlags |= RuntimeOps.GetAssemblyPluginFlags(assembly,
                    RuntimeOps.GetSkipCheckPluginFlags());
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            private void SetupTypeName()
            {
                if (assembly == null)
                    throw new InvalidOperationException(AssemblyNotLoadedError);

                if (type != null)
                    throw new InvalidOperationException(TypeAlreadySetError);

                //
                // NOTE: Do they want a specific plugin type or simply the
                //       default one?
                //
                if (!String.IsNullOrEmpty(typeName))
                {
                    type = assembly.GetType(typeName, true, false); /* throw */

                    //
                    // NOTE: Did we find the type?  For now, this should already be
                    //       true because the above code should throw an exception
                    //       if the type is not found.
                    //
                    if (type != null)
                        typeName = type.FullName;
                }
                else
                {
                    ReturnCode code;
                    Result error = null;

                    code = RuntimeOps.FindPrimaryPlugin(assembly, ref typeName, ref error);

                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Attempt to get the type from the assembly now.
                        //
                        type = assembly.GetType(typeName, true, false); /* throw */
                    }
                    else
                    {
                        throw new TypeLoadException(
                            PrimaryPluginTypeNotFoundError,
                            new ScriptException(code, error));
                    }
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            private void SetupUri()
            {
                if (assembly == null)
                    throw new InvalidOperationException(AssemblyNotLoadedError);

                if (uri != null)
                    throw new InvalidOperationException(UriAlreadySetError);

                //
                // NOTE: Check if the plugin assembly contains a an AssemblyUri
                //       attribute.
                //
                uri = AttributeOps.GetAssemblyUri(assembly);

                //
                // NOTE: If the plugin assembly does NOT contain an AssemblyUri
                //       attribute, use the location.
                //
                if (uri == null)
                    uri = new Uri(assembly.Location); /* throw */
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Properties
            public Assembly Assembly
            {
                get { return assembly; }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            public PluginPropertyHelper PropertyHelper
            {
                get { return propertyHelper; }
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Public Methods
#if !CAS_POLICY
            public void Load1()
            {
                if (assembly != null)
                    throw new InvalidOperationException(AssemblyAlreadyLoadedError);

                //
                // NOTE: Load the assembly into this AppDomain now.
                //
                assembly = Assembly.LoadFrom(fileName); /* throw */

                //
                // NOTE: Perform the post-load setup steps?
                //
                if (setup)
                    Setup(); /* throw */
            }
#endif

            ///////////////////////////////////////////////////////////////////////////////////////////

#if CAS_POLICY
            public void Load2()
            {
                if (assembly != null)
                    throw new InvalidOperationException(AssemblyAlreadyLoadedError);

                //
                // NOTE: Load the assembly into this AppDomain using the
                //       provided evidence now.
                //
                assembly = Assembly.LoadFrom(fileName, evidence); /* throw */

                //
                // NOTE: Perform the post-load setup steps?
                //
                if (setup)
                    Setup(); /* throw */
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

#if !MONO && !MONO_HACKS
            public void Load4()
            {
                if (assembly != null)
                    throw new InvalidOperationException(AssemblyAlreadyLoadedError);

                //
                // NOTE: Load the assembly into this AppDomain using the
                //       provided evidence and hash parameters now.
                //
                assembly = Assembly.LoadFrom(
                    fileName, evidence, hashValue, hashAlgorithm); /* throw */

                //
                // NOTE: Perform the post-load setup steps?
                //
                if (setup)
                    Setup(); /* throw */
            }
#endif
#endif
            #endregion
        }
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IPropertyManager Members
        public object ApplicationObject
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return applicationObject;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    applicationObject = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public object PolicyObject
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return policyObject;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    policyObject = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public object ResolverObject
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return resolverObject;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    resolverObject = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public object UserObject
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return userObject;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    userObject = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public StringDictionary RuntimeOptions
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return runtimeOptions;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    runtimeOptions = value;
                }

                //
                // NOTE: Do not create the variable if we have been prohibited
                //       from doing so.
                //
                VariableFlags flags = VariableFlags.GlobalOnly | VariableFlags.NoReady;

                if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                    FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                {
                    flags |= VariableFlags.NoCreate;
                }

                //
                // NOTE: Update script-level informational variable.
                //
                SetVariableValue2(flags, Vars.Platform.Name, Vars.Platform.RuntimeOptions,
                    value.ToString(), null);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool HasRuntimeOption(
            string name
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!String.IsNullOrEmpty(name))
                    if ((runtimeOptions != null) && runtimeOptions.ContainsKey(name))
                        return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool ClearRuntimeOptions()
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (runtimeOptions != null)
                    {
                        runtimeOptions.Clear();

                        //
                        // NOTE: Do not create the variable if we have been prohibited
                        //       from doing so.
                        //
                        VariableFlags flags =
                            VariableFlags.GlobalOnly | VariableFlags.NoReady;

                        if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                            FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                        {
                            flags |= VariableFlags.NoCreate;
                        }

                        //
                        // NOTE: Update script-level informational variable.
                        //
                        /* IGNORED */
                        SetVariableValue2(flags, Vars.Platform.Name,
                            Vars.Platform.RuntimeOptions,
                            runtimeOptions.ToString(), null);

                        return true;
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool AddRuntimeOption(
            string name
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (!String.IsNullOrEmpty(name))
                    {
                        if ((runtimeOptions != null) &&
                            !runtimeOptions.ContainsKey(name))
                        {
                            runtimeOptions.Add(name, null);

                            //
                            // NOTE: Do not create the variable if we have been prohibited
                            //       from doing so.
                            //
                            VariableFlags flags =
                                VariableFlags.GlobalOnly | VariableFlags.NoReady;

                            if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                                FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                            {
                                flags |= VariableFlags.NoCreate;
                            }

                            //
                            // NOTE: Update script-level informational variable.
                            //
                            /* IGNORED */
                            SetVariableValue2(flags, Vars.Platform.Name,
                                Vars.Platform.RuntimeOptions,
                                runtimeOptions.ToString(), null);

                            return true;
                        }
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool RemoveRuntimeOption(
            string name
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (!String.IsNullOrEmpty(name))
                    {
                        if ((runtimeOptions != null) &&
                            runtimeOptions.ContainsKey(name))
                        {
                            runtimeOptions.Remove(name);

                            //
                            // NOTE: Do not create the variable if we have been prohibited
                            //       from doing so.
                            //
                            VariableFlags flags =
                                VariableFlags.GlobalOnly | VariableFlags.NoReady;

                            if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                                FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                            {
                                flags |= VariableFlags.NoCreate;
                            }

                            //
                            // NOTE: Update script-level informational variable.
                            //
                            /* IGNORED */
                            SetVariableValue2(flags, Vars.Platform.Name,
                                Vars.Platform.RuntimeOptions,
                                runtimeOptions.ToString(), null);

                            return true;
                        }
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool ChangeRuntimeOption(
            string name
            )
        {
            CheckDisposed();

            if (!String.IsNullOrEmpty(name))
            {
                switch (name[0])
                {
                    case Characters.PlusSign:
                        {
                            if (name.Length > 1)
                                return AddRuntimeOption(name.Substring(1));

                            break;
                        }
                    case Characters.MinusSign:
                        {
                            if (name.Length > 1)
                                return RemoveRuntimeOption(name.Substring(1));

                            break;
                        }
                    case Characters.EqualSign:
                        {
                            if (name.Length > 1)
                                return ClearRuntimeOptions() &&
                                    AddRuntimeOption(name.Substring(1));

                            break;
                        }
                    default:
                        {
                            if (Parser.IsIdentifier(name[0]))
                                return AddRuntimeOption(name);

                            break;
                        }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool ThrowOnFeatureNotSupported
        {
            get
            {
                lock (syncRoot)
                {
                    return throwOnFeatureNotSupported;
                }
            }
            set
            {
                lock (syncRoot)
                {
                    throwOnFeatureNotSupported = value;
                }
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IResolveManager Members
        #region Call Frame Management (WARNING: EXPERIMENTAL)
        //
        // WARNING: *EXPERIMENTAL* This API may change until the
        //          resolver subsystem is completed.
        //
        public ReturnCode GetCallFrame( /* EXPERIMENTAL */
            bool absolute,              /* in */
            int level,                  /* in */
            CallFrameFlags hasFlags,    /* in */
            CallFrameFlags notHasFlags, /* in */
            bool hasAll,                /* in */
            bool notHasAll,             /* in */
            ref ICallFrame frame,       /* in, out */
            ref Result error            /* out */
            )
        {
            CheckDisposed();

#if !THREADING
            lock (syncRoot) /* TRANSACTIONAL */
#endif
            {
                ICallFrame currentFrame =
                    (frame != null) ? frame : CurrentFrame;

                return CallFrameOps.GetOrFind(
                    CallStack, GlobalFrame, currentFrame, absolute, level,
                    hasFlags, notHasFlags, hasAll, notHasAll, ref frame,
                    ref error);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Command & Variable Resolution (WARNING: EXPERIMENTAL)
        //
        // WARNING: *EXPERIMENTAL* This API may change until the
        //          resolver subsystem is completed.
        //
        public ReturnCode GetIExecuteViaResolvers( /* EXPERIMENTAL */
            ICallFrame frame,
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasResolvers(ref error))
                {
                    ResultList errors = null;

                    foreach (IResolve resolve in resolvers)
                    {
                        if (resolve != null)
                        {
                            Result localError = null;

                            try
                            {
                                ReturnCode code = resolve.GetIExecute(frame,
                                    flags, name, arguments, wrapper, validate,
                                    ref ambiguous, ref token, ref execute,
                                    ref localError);

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Success, return the entity that
                                    //       this resolver found.
                                    //
                                    return ReturnCode.Ok;
                                }
                                else if (code == ReturnCode.Return)
                                {
                                    //
                                    // NOTE: Failure, return the error that
                                    //       this resolver returned verbatim.
                                    //
                                    error = localError;

                                    return ReturnCode.Error;
                                }
                                else if (code == ReturnCode.Break)
                                {
                                    //
                                    // NOTE: Failure, skip all other resolvers
                                    //       for this entity and return all the
                                    //       previously collected errors.
                                    //
                                    break;
                                }
                                else if (code == ReturnCode.Continue)
                                {
                                    //
                                    // NOTE: Failure, keep trying; however, do
                                    //       not add the error message from this
                                    //       resolver.
                                    //
                                    continue;
                                }
                            }
                            catch (Exception e)
                            {
                                localError = e;
                            }

                            if (errors == null)
                                errors = new ResultList();

                            if (localError != null)
                                errors.Add(localError);
                        }
                    }

                    if (errors != null)
                        error = errors;
                    else
                        error = "no resolvers defined";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // WARNING: *EXPERIMENTAL* This API may change until the
        //          resolver subsystem is completed.
        //
        public ReturnCode GetVariableViaResolvers( /* EXPERIMENTAL */
            ICallFrame frame,
            string varName,
            string varIndex,
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasResolvers(ref error))
                {
                    bool resolveNull = FlagOps.HasFlags(
                        flags, VariableFlags.ResolveNull, true);

                    ResultList errors = null;

                    foreach (IResolve resolve in resolvers)
                    {
                        if (resolve != null)
                        {
                            Result localError = null;

                            try
                            {
                                ReturnCode code = resolve.GetVariable(
                                    frame, varName, varIndex, ref flags,
                                    ref variable, ref localError);

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Success, return the entity that
                                    //       this resolver found.  If the
                                    //       entity is now null and that is
                                    //       disallowed via the flags, fail.
                                    //
                                    if (!resolveNull && (variable == null))
                                    {
                                        error = String.Format(
                                            "can't resolve {0}: variable is invalid",
                                            FormatOps.ErrorVariableName(varName, varIndex));

                                        return ReturnCode.Error;
                                    }

                                    return ReturnCode.Ok;
                                }
                                else if (code == ReturnCode.Return)
                                {
                                    //
                                    // NOTE: Failure, return the error that
                                    //       this resolver returned verbatim.
                                    //
                                    error = localError;

                                    return ReturnCode.Error;
                                }
                                else if (code == ReturnCode.Break)
                                {
                                    //
                                    // NOTE: Failure, skip all other resolvers
                                    //       for this entity and return all the
                                    //       previously collected errors.
                                    //
                                    break;
                                }
                                else if (code == ReturnCode.Continue)
                                {
                                    //
                                    // NOTE: Failure, keep trying; however, do
                                    //       not add the error message from this
                                    //       resolver.
                                    //
                                    continue;
                                }
                            }
                            catch (Exception e)
                            {
                                localError = e;
                            }

                            if (errors == null)
                                errors = new ResultList();

                            if (localError != null)
                                errors.Add(localError);
                        }
                    }

                    if (errors != null)
                        error = errors;
                    else
                        error = "no resolvers defined";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal EngineFlags GetResolveEngineFlags(
            bool exact
            )
        {
            return Engine.GetResolveFlags(engineFlags, exact);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateGetAnyIExecute(
            ICallFrame frame, /* NOT USED */
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            bool useHidden = Engine.HasUseHidden(flags);

            //
            // TODO: Modify this when full [namespace] support has been added.
            //
            string newName = StringOps.TrimLeadingColons(name);

            if (Engine.HasUseIExecutes(flags) &&
                 ((useHidden && (GetHiddenIExecute(
                    newName, wrapper, validate, ref execute) == ReturnCode.Ok)) ||
                 (!useHidden && (GetIExecute(
                    newName, wrapper, validate, ref execute) == ReturnCode.Ok))))
            {
                return ReturnCode.Ok;
            }
            else
            {
                IProcedure procedure = null;

                if (Engine.HasUseProcedures(flags) &&
                    ((useHidden && (GetHiddenProcedure(
                        newName, wrapper, validate, ref procedure) == ReturnCode.Ok)) ||
                    (!useHidden && (GetProcedure(
                        newName, wrapper, validate, ref procedure) == ReturnCode.Ok))))
                {
                    execute = procedure;

                    return ReturnCode.Ok;
                }
                else
                {
                    ICommand command = null;

                    if (Engine.HasUseCommands(flags) &&
                        ((useHidden && (GetHiddenCommand(
                            newName, wrapper, validate, ref command) == ReturnCode.Ok)) ||
                        (!useHidden && (GetCommand(
                            newName, wrapper, validate, ref command) == ReturnCode.Ok))))
                    {
                        execute = command;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = String.Format(
                            "invalid {0}command name \"{1}\"",
                            useHidden ? "hidden " : String.Empty,
                            FormatOps.DisplayName(newName));
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateMatchAnyIExecute(
            ICallFrame frame, /* NOT USED */
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref IExecute execute,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                //
                // NOTE: *WARNING* Empty command/procedure names are allowed,
                //       please do not change this to "!String.IsNullOrEmpty".
                //
                if (name != null)
                {
                    //
                    // TODO: Modify this when full [namespace] support has been added.
                    //
                    string newName = StringOps.TrimLeadingColons(name);

#if EXECUTE_CACHE
                    if (executeCache != null)
                    {
                        if (executeCache.TryGet(newName, validate, ref execute))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref executeCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return ReturnCode.Ok;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref executeCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
#endif

                    string exactName = null;
                    ExecuteDictionary matches = new ExecuteDictionary();
                    bool useHidden = Engine.HasUseHidden(flags);

                    if (exactName == null)
                    {
                        if (Engine.HasUseIExecutes(flags))
                        {
                            ExecuteWrapperDictionary localExecutes = useHidden ?
                                hiddenExecutes : executes;

                            if ((localExecutes != null) && (localExecutes.Count > 0))
                            {
                                if (localExecutes.ContainsKey(newName))
                                {
                                    //
                                    // NOTE: Exact execute match.
                                    //
                                    exactName = newName;

                                    //
                                    // NOTE: Store the exact execute match in the results
                                    //       dictionary.
                                    //
                                    matches.Add(newName, localExecutes[newName]);
                                }
                                else
                                {
                                    foreach (KeyValuePair<string, _Wrappers._Execute> pair in localExecutes)
                                    {
                                        if (String.Compare(
                                                pair.Key, 0, newName, 0, newName.Length,
                                                StringOps.SystemStringComparisonType) == 0)
                                        {
                                            //
                                            // NOTE: Was the key valid (this should always succeed).
                                            //
                                            if (pair.Key != null)
                                            {
                                                //
                                                // NOTE: It was a match; however, was it an exact match?
                                                //
                                                if (pair.Key.Length == newName.Length)
                                                    //
                                                    // NOTE: Preserve match, it may differ in case.
                                                    //
                                                    exactName = pair.Key;

                                                //
                                                // NOTE: Was it an exact match or did we match at least one
                                                //       character in a partial match?
                                                //
                                                if ((pair.Key.Length == newName.Length) || (newName.Length > 0))
                                                    //
                                                    // NOTE: Store the exact or partial match in the results
                                                    //       dictionary.
                                                    //
                                                    matches.Add(pair.Key, pair.Value);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if (exactName == null)
                    {
                        if (Engine.HasUseProcedures(flags))
                        {
                            ProcedureWrapperDictionary localProcedures = useHidden ?
                                hiddenProcedures : procedures;

                            if ((localProcedures != null) && (localProcedures.Count > 0))
                            {
                                if (localProcedures.ContainsKey(newName) &&
                                    (Engine.HasMatchHidden(flags) ||
                                    !EntityOps.IsHidden(localProcedures[newName])))
                                {
                                    //
                                    // NOTE: Exact procedure match.
                                    //
                                    exactName = newName;

                                    //
                                    // NOTE: Store the exact procedure match in the results
                                    //       dictionary.
                                    //
                                    matches.Add(newName, localProcedures[newName]);
                                }
                                else
                                {
                                    foreach (KeyValuePair<string, _Wrappers.Procedure> pair in localProcedures)
                                    {
                                        if (String.Compare(
                                                pair.Key, 0, newName, 0, newName.Length,
                                                StringOps.SystemStringComparisonType) == 0)
                                        {
                                            //
                                            // NOTE: Is the procedure visible or are we allowed to
                                            //       find hidden procedures?
                                            //
                                            if (Engine.HasMatchHidden(flags) || !EntityOps.IsHidden(pair.Value))
                                            {
                                                //
                                                // NOTE: Was the key valid (this should always succeed).
                                                //
                                                if (pair.Key != null)
                                                {
                                                    //
                                                    // NOTE: It was a match; however, was it an exact match?
                                                    //
                                                    if (pair.Key.Length == newName.Length)
                                                        //
                                                        // NOTE: Preserve match, it may differ in case.
                                                        //
                                                        exactName = pair.Key;

                                                    //
                                                    // NOTE: Was it an exact match or did we match at least one
                                                    //       character in a partial match?
                                                    //
                                                    if ((pair.Key.Length == newName.Length) || (newName.Length > 0))
                                                        //
                                                        // NOTE: Store the exact or partial match in the results
                                                        //       dictionary.
                                                        //
                                                        matches.Add(pair.Key, pair.Value);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if (exactName == null)
                    {
                        if (Engine.HasUseCommands(flags))
                        {
                            CommandWrapperDictionary localCommands = useHidden ?
                                hiddenCommands : commands;

                            if ((localCommands != null) && (localCommands.Count > 0))
                            {
                                if (localCommands.ContainsKey(newName) &&
                                    (Engine.HasMatchHidden(flags) ||
                                    !EntityOps.IsHidden(localCommands[newName])))
                                {
                                    //
                                    // NOTE: Exact command match.
                                    //
                                    exactName = newName;

                                    //
                                    // NOTE: Store the exact command match in the results
                                    //       dictionary.
                                    //
                                    matches.Add(newName, localCommands[newName]);
                                }
                                else
                                {
                                    foreach (KeyValuePair<string, _Wrappers.Command> pair in localCommands)
                                    {
                                        if (String.Compare(
                                                pair.Key, 0, newName, 0, newName.Length,
                                                StringOps.SystemStringComparisonType) == 0)
                                        {
                                            //
                                            // NOTE: Is the command visible or are we allowed to
                                            //       find hidden command?
                                            //
                                            if (Engine.HasMatchHidden(flags) || !EntityOps.IsHidden(pair.Value))
                                            {
                                                //
                                                // NOTE: Was the key valid (this should always succeed).
                                                //
                                                if (pair.Key != null)
                                                {
                                                    //
                                                    // NOTE: It was a match; however, was it an exact match?
                                                    //
                                                    if (pair.Key.Length == newName.Length)
                                                        //
                                                        // NOTE: Preserve match, it may differ in case.
                                                        //
                                                        exactName = pair.Key;

                                                    //
                                                    // NOTE: Was it an exact match or did we match at least one
                                                    //       character in a partial match?
                                                    //
                                                    if ((pair.Key.Length == newName.Length) || (newName.Length > 0))
                                                        //
                                                        // NOTE: Store the exact or partial match in the results
                                                        //       dictionary.
                                                        //
                                                        matches.Add(pair.Key, pair.Value);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }

                    //
                    // NOTE: If there was an exact match, just use it.
                    //
                    if (exactName != null)
                    {
                        //
                        // NOTE: Exact match, return the matching command or
                        //       procedure.
                        //
                        if (matches[exactName] is _Wrappers._Execute)
                        {
                            if ((useHidden && PrivateGetHiddenIExecute(
                                    exactName, wrapper, validate, ref token,
                                    ref execute, ref error) == ReturnCode.Ok) ||
                                (!useHidden && PrivateGetIExecute(
                                    exactName, wrapper, validate, ref token,
                                    ref execute, ref error) == ReturnCode.Ok))
                            {
#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(exactName, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenIExecute : NotifyType.IExecute),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, exactName, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (matches[exactName] is _Wrappers.Procedure)
                        {
                            IProcedure procedure = null;

                            if ((useHidden && PrivateGetHiddenProcedure(
                                    exactName, wrapper, validate, ref token,
                                    ref procedure, ref error) == ReturnCode.Ok) ||
                                (!useHidden && PrivateGetProcedure(
                                    exactName, wrapper, validate, ref token,
                                    ref procedure, ref error) == ReturnCode.Ok))
                            {
                                execute = procedure;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(exactName, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenProcedure : NotifyType.Procedure),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, exactName, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (matches[exactName] is _Wrappers.Command)
                        {
                            ICommand command = null;

                            if ((useHidden && PrivateGetHiddenCommand(
                                    exactName, wrapper, validate, ref token,
                                    ref command, ref error) == ReturnCode.Ok) ||
                                (!useHidden && PrivateGetCommand(
                                    exactName, wrapper, validate, ref token,
                                    ref command, ref error) == ReturnCode.Ok))
                            {
                                execute = command;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(exactName, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenCommand : NotifyType.Command),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, exactName, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "object \"{0}\" is not an execute, procedure, or command wrapper",
                                exactName);
                        }
                    }
                    else if (matches.Count == 1)
                    {
                        string key = new StringList(matches.Keys)[0];

                        //
                        // NOTE: Normal case, exactly one command matched.
                        //
                        if (matches[key] is _Wrappers._Execute)
                        {
                            if ((useHidden && (PrivateGetHiddenIExecute(
                                    key, wrapper, validate, ref token,
                                    ref execute, ref error) == ReturnCode.Ok)) ||
                                (!useHidden && (PrivateGetIExecute(
                                    key, wrapper, validate, ref token,
                                    ref execute, ref error) == ReturnCode.Ok)))
                            {
#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(key, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenIExecute : NotifyType.IExecute),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, key, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (matches[key] is _Wrappers.Procedure)
                        {
                            IProcedure procedure = null;

                            if ((useHidden && (PrivateGetHiddenProcedure(
                                    key, wrapper, validate, ref token,
                                    ref procedure, ref error) == ReturnCode.Ok)) ||
                                (!useHidden && (PrivateGetProcedure(
                                    key, wrapper, validate, ref token,
                                    ref procedure, ref error) == ReturnCode.Ok)))
                            {
                                execute = procedure;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(key, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenProcedure : NotifyType.Procedure),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, key, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else if (matches[key] is _Wrappers.Command)
                        {
                            ICommand command = null;

                            if ((useHidden && (PrivateGetHiddenCommand(
                                    key, wrapper, validate, ref token,
                                    ref command, ref error) == ReturnCode.Ok)) ||
                                (!useHidden && (PrivateGetCommand(
                                    key, wrapper, validate, ref token,
                                    ref command, ref error) == ReturnCode.Ok)))
                            {
                                execute = command;

#if EXECUTE_CACHE
                                if (executeCache != null)
                                    executeCache.AddOrUpdate(key, execute, false);
#endif

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Resolver | (useHidden ? NotifyType.HiddenCommand : NotifyType.Command),
                                    NotifyFlags.Matched | (useHidden ? NotifyFlags.Hidden : NotifyFlags.None),
                                    new ObjectList(flags, name, key, execute, token), this, null, null, null,
                                    ref error);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = String.Format(
                                "object \"{0}\" is not an execute, procedure, or command wrapper",
                                key);
                        }
                    }
                    else if (matches.Count > 1)
                    {
                        error = RuntimeOps.BadValue("ambiguous",
                            useHidden ? "hidden command" : "command",
                            newName, matches.Keys, null, null);

                        ambiguous = true;
                    }
                    else
                    {
                        error = String.Format(
                            "invalid {0}command name \"{1}\"",
                            useHidden ? "hidden " : String.Empty,
                            newName);
                    }
                }
                else
                {
                    error = "invalid command name";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetAnyIExecute(
            ICallFrame frame,
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref long token,
            ref IExecute execute,
            ref Result error
            ) /* USED BY CORE RESOLVER */
        {
            /* CheckDisposed(); */

            //
            // NOTE: When searching for [non-hidden] commands, fallback to searching
            //       the hidden commands also to give the policies, if any, a chance
            //       to allow or deny their execution; however, only do this if the
            //       appropriate flag (get-hidden) is set.
            //
            Result localError = null;

            if (PrivateGetAnyIExecute(
                    frame, flags, name, wrapper, validate, ref token,
                    ref execute, ref localError) == ReturnCode.Ok)
            {
                return ReturnCode.Ok;
            }
            else if (RuntimeOps.ShouldResolveHidden(flags, false))
            {
                Result localError2 = null; /* VALUE NOT USED */

                if (PrivateGetAnyIExecute(
                        frame, flags | EngineFlags.UseHidden, name, wrapper, validate,
                        ref token, ref execute, ref localError2) == ReturnCode.Ok)
                {
                    return ReturnCode.Ok;
                }
            }

            error = localError;
            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode MatchAnyIExecute(
            ICallFrame frame,
            EngineFlags flags,
            string name,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref long token,
            ref IExecute execute,
            ref Result error
            ) /* USED BY CORE RESOLVER */
        {
            /* CheckDisposed(); */

            //
            // NOTE: When searching for [non-hidden] commands, fallback to searching
            //       the hidden commands also to give the policies, if any, a chance
            //       to allow or deny their execution; however, only do this if the
            //       appropriate flag (match-hidden) is set.
            //
            Result localError = null;

            if (PrivateMatchAnyIExecute(
                    frame, flags, name, wrapper, validate, ref ambiguous,
                    ref token, ref execute, ref localError) == ReturnCode.Ok)
            {
                return ReturnCode.Ok;
            }
            else if (RuntimeOps.ShouldResolveHidden(flags, true))
            {
                Result localError2 = null; /* VALUE NOT USED */

                if (PrivateMatchAnyIExecute(
                        frame, flags | EngineFlags.UseHidden, name, wrapper,
                        validate, ref ambiguous, ref token, ref execute,
                        ref localError2) == ReturnCode.Ok)
                {
                    return ReturnCode.Ok;
                }
            }

            error = localError;
            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetIExecuteViaResolvers(
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref IExecute execute
            )
        {
            return GetIExecuteViaResolvers(
                null, flags, name, arguments, wrapper, validate, ref execute);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetIExecuteViaResolvers(
            ICallFrame frame,
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref IExecute execute
            )
        {
            Result error = null;

            return GetIExecuteViaResolvers(
                frame, flags, name, arguments, wrapper, validate, ref execute,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetIExecuteViaResolvers(
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref IExecute execute,
            ref Result error
            )
        {
            return GetIExecuteViaResolvers(
                null, flags, name, arguments, wrapper, validate, ref execute,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetIExecuteViaResolvers(
            ICallFrame frame,
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref IExecute execute,
            ref Result error
            )
        {
            bool ambiguous = false;

            return GetIExecuteViaResolvers(
                frame, flags, name, arguments, wrapper, validate, ref ambiguous,
                ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetIExecuteViaResolvers(
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref IExecute execute,
            ref Result error
            )
        {
            return GetIExecuteViaResolvers(
                null, flags, name, arguments, wrapper, validate, ref ambiguous,
                ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetIExecuteViaResolvers(
            ICallFrame frame,
            EngineFlags flags,
            string name,
            ArgumentList arguments,
            bool wrapper,
            bool validate,
            ref bool ambiguous,
            ref IExecute execute,
            ref Result error
            )
        {
            long token = 0;

            return GetIExecuteViaResolvers(
                frame, flags, name, arguments, wrapper, validate, ref ambiguous,
                ref token, ref execute, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableViaResolversWithSplit(
            ICallFrame frame,
            string name,
            ref IVariable variable
            )
        {
            VariableFlags flags = VariableFlags.None;
            Result error = null;

            return GetVariableViaResolversWithSplit(
                frame, name, null, ref flags, ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableViaResolversWithSplit(
            string name,
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            )
        {
            return GetVariableViaResolversWithSplit(
                null, name, null, ref flags, ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetVariableViaResolversWithSplit(
            ICallFrame frame,
            string name,
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            )
        {
            return GetVariableViaResolversWithSplit(
                frame, name, null, ref flags, ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetVariableViaResolversWithSplit(
            ICallFrame frame,
            string name,
            string index,
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            )
        {
            //
            // NOTE: Reset the error related flags because we have not
            //       actually searched for the variable yet.
            //
            flags &= ~VariableFlags.ErrorMask;

            //
            // BUGFIX: Prevent any custom resolvers from returning a null
            //         variable along with a successful return code.
            //
            flags &= ~VariableFlags.ResolveNull;

            //
            // BUGFIX: Block all attempts to use an array element reference
            //         in the variable name if an explicit array element
            //         index has been explicitly specified by the caller.
            //
            if (index != null)
                flags |= VariableFlags.NoElement;

            //
            // NOTE: Perform a split of the name (which may refer to an array
            //       element) into the base variable name and the array index,
            //       if any.
            //
            string varName = null;
            string varIndex = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return GetVariableViaResolvers(
                    frame, varName, varIndex, ref flags,
                    ref variable, ref error);
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Resolver Management (WARNING: EXPERIMENTAL)
        public bool HasResolvers(ref Result error)
        {
            CheckDisposed();

            bool result = (resolvers != null);

            if (!result)
                error = "resolvers not available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ResetResolvers(
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    error = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (!HasResolvers(ref error))
                    return ReturnCode.Error;

                IResolve resolve = RuntimeOps.NewCoreResolver(this);

                resolvers.Clear();
                resolvers.Add(resolve);
                resolve.Token = (long)Priority.Highest;

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Resolve, NotifyFlags.Reset,
                    null, this,
                    null, null, null, ref error);
#endif

                return ReturnCode.Ok;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetResolver(
            Priority priority,
            ref IResolve resolve,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!HasResolvers(ref error))
                    return ReturnCode.Error;

                if (priority == Priority.Lowest)
                    priority = (Priority)resolvers.Count - 1;
                else if (priority == Priority.None)
                    priority = Priority.Highest;

                if (((int)priority < 0) ||
                    ((int)priority > resolvers.Count))
                {
                    error = "invalid priority";
                    return ReturnCode.Error;
                }

                resolve = resolvers[(int)priority];
                return ReturnCode.Ok;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddResolver(
            IResolve resolve,
            IClientData clientData,
            Priority priority,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (resolve == null)
                {
                    result = "invalid resolve";
                    return ReturnCode.Error;
                }

                if (!HasResolvers(ref result))
                    return ReturnCode.Error;

                if (priority == Priority.Lowest)
                {
                    resolvers.Add(resolve);
                    resolve.Token = resolvers.Count - 1;
                }
                else
                {
                    if (priority == Priority.None)
                        priority = Priority.Highest;

                    if (((int)priority < 0) ||
                        ((int)priority > resolvers.Count))
                    {
                        result = "invalid priority";
                        return ReturnCode.Error;
                    }

                    resolvers.Insert((int)priority, resolve);
                    resolve.Token = (long)priority;
                }

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Resolve, NotifyFlags.Added,
                    new ObjectPair(resolve, priority), this,
                    clientData, null, null, ref result);
#endif

                return ReturnCode.Ok;
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IResourceManager Members
        public string GetString(
            string name
            )
        {
            CheckDisposed();

            return GetString(null, name);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string GetString(
            IPlugin plugin,
            string name
            )
        {
            CheckDisposed();

            //
            // NOTE: Use the culture configured for this interpreter.
            //
            return GetString(plugin, name, cultureInfo);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string GetString(
            IPlugin plugin,
            string name,
            ref Result error
            )
        {
            CheckDisposed();

            //
            // NOTE: Use the culture configured for this interpreter.
            //
            return GetString(plugin, name, cultureInfo, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string GetString(
            IPlugin plugin,
            string name,
            CultureInfo cultureInfo
            )
        {
            CheckDisposed();

            Result error = null;

            return GetString(plugin, name, cultureInfo, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public string GetString(
            IPlugin plugin,
            string name,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            CheckDisposed();

            return RuntimeOps.GetString(
                plugin, resourceManager, name, cultureInfo, ref error);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IScriptManager Members
        public string LibraryPath
        {
            get { CheckDisposed(); lock (syncRoot) { return libraryPath; } }
            set { CheckDisposed(); lock (syncRoot) { libraryPath = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public StringList AutoPathList
        {
            get { CheckDisposed(); lock (syncRoot) { return autoPathList; } }
            set { CheckDisposed(); lock (syncRoot) { autoPathList = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode PreInitialize(
            ref Result error /* out */
            )
        {
            CheckDisposed();

            return EvaluateScript(preInitializeText, ref error); /* EXEMPT */
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode Initialize(
            ref Result error /* out */
            )
        {
            CheckDisposed();

            return PrivateInitialize(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode InitializeShell(
            ref Result error /* out */
            )
        {
            CheckDisposed();

            return PrivateInitializeShell(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetScript(
            string name,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            CheckDisposed();

            if (FlagOps.HasFlags(flags, ScriptFlags.Library, true))
                return RuntimeOps.GetLibraryScript(this, host, name, ref flags, ref result);
            else
                return RuntimeOps.GetHostScript(this, host, name, ref flags, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal string GetScript(
            string name,
            ref ScriptFlags flags
            )
        {
            Result result = null;

            if (GetScript(name, ref flags, ref result) != ReturnCode.Ok)
                result = null; /* NOTE: Discard error message. */

            return result;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ISecurityManager Members
        public StringList TrustedPaths
        {
            get { CheckDisposed(); return trustedPaths; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ObjectDictionary TrustedTypes
        {
            get { CheckDisposed(); return trustedTypes; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public PolicyDecision CommandDecision
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return commandDecision;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    commandDecision = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public PolicyDecision ScriptDecision
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return scriptDecision;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    scriptDecision = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public PolicyDecision FileDecision
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return fileDecision;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    fileDecision = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public PolicyDecision StreamDecision
        {
            get
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    return streamDecision;
                }
            }
            set
            {
                CheckDisposed();

                lock (syncRoot)
                {
                    streamDecision = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsSafe()
        {
            CheckDisposed();

            return FlagOps.HasFlags(createFlags, CreateFlags.Safe, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MakeSafe(
            bool safe,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (commands != null)
                {
                    foreach (KeyValuePair<string, _Wrappers.Command> pair in commands)
                    {
                        if (pair.Value != null)
                        {
                            CommandFlags flags = pair.Value.Flags;

                            if (!FlagOps.HasFlags(flags, CommandFlags.Safe, true) ||
                                FlagOps.HasFlags(flags, CommandFlags.Unsafe, true))
                            {
                                if (safe)
                                    flags |= CommandFlags.Hidden;
                                else
                                    flags &= ~CommandFlags.Hidden;

                                pair.Value.Flags = flags;
                            }
                        }
                    }
                }

                if (procedures != null)
                {
                    foreach (KeyValuePair<string, _Wrappers.Procedure> pair in procedures)
                    {
                        if (pair.Value != null)
                        {
                            ProcedureFlags flags = pair.Value.Flags;

                            if (!FlagOps.HasFlags(flags, ProcedureFlags.Safe, true) ||
                                FlagOps.HasFlags(flags, ProcedureFlags.Unsafe, true))
                            {
                                if (safe)
                                    flags |= ProcedureFlags.Hidden;
                                else
                                    flags &= ~ProcedureFlags.Hidden;

                                pair.Value.Flags = flags;
                            }
                        }
                    }
                }

                //
                // NOTE: Move all the newly hidden and exposed commands to their proper
                //       internal collections.
                //
                if (MoveExposedAndHiddenCommands(ref error) == ReturnCode.Ok)
                {
                    //
                    // BUGBUG: Removal of the "unsafe" variables and array elements
                    //         from the global call frame is currently a one-way
                    //         street, just like in Tcl.  We make no attempt to
                    //         restore the deleted variables if/when the interpreter
                    //         is made "unsafe" again.
                    //
                    if (safe)
                    {
                        SetupSafeVariableNames();

                        ICallFrame frame = GlobalFrame;

                        if (frame != null)
                        {
                            VariableDictionary variables = frame.Variables;

                            if (variables != null)
                            {
                                foreach (KeyValuePair<string, IVariable> pair in variables)
                                {
                                    if (pair.Value != null)
                                    {
                                        VariableFlags flags = pair.Value.Flags;

                                        if ((!FlagOps.HasFlags(flags, VariableFlags.Safe, true) ||
                                            FlagOps.HasFlags(flags, VariableFlags.Unsafe, true)) &&
                                            !safeVariableNames.Contains(pair.Key))
                                        {
                                            //
                                            // NOTE: Mark this "unsafe" variable for removal.
                                            //
                                            EntityOps.SetUndefined(pair.Value, true);
                                        }
                                        else if ((String.Compare(pair.Key, TclVars.Platform.Name,
                                                StringOps.SystemStringComparisonType) == 0))
                                        {
                                            //
                                            // NOTE: Prepare to remove the "unsafe" array
                                            //       elements from the "tcl_platform" array.
                                            //
                                            ObjectDictionary arrayValue = pair.Value.ArrayValue;

                                            if (arrayValue != null)
                                            {
                                                //
                                                // NOTE: This must be done in two phases because
                                                //       we cannot modify the dictionary while we
                                                //       enumerate it.
                                                //
                                                StringList keys = new StringList();

                                                foreach (KeyValuePair<string, object> pair2 in arrayValue)
                                                    if (!safeTclPlatformElementNames.Contains(pair2.Key))
                                                        keys.Add(pair2.Key);

                                                //
                                                // NOTE: Purge the "unsafe" array elements.
                                                //
                                                foreach (string key in keys)
                                                    arrayValue.Remove(key);
                                            }
                                        }
                                        else if ((String.Compare(pair.Key, Vars.Platform.Name,
                                                StringOps.SystemStringComparisonType) == 0))
                                        {
                                            //
                                            // NOTE: Prepare to remove the "unsafe" array
                                            //       elements from the "eagle_platform" array.
                                            //
                                            ObjectDictionary arrayValue = pair.Value.ArrayValue;

                                            if (arrayValue != null)
                                            {
                                                //
                                                // NOTE: This must be done in two phases because
                                                //       we cannot modify the dictionary while we
                                                //       enumerate it.
                                                //
                                                StringList keys = new StringList();

                                                foreach (KeyValuePair<string, object> pair2 in arrayValue)
                                                    if (!safeEaglePlatformElementNames.Contains(pair2.Key))
                                                        keys.Add(pair2.Key);

                                                //
                                                // NOTE: Purge the "unsafe" array elements.
                                                //
                                                foreach (string key in keys)
                                                    arrayValue.Remove(key);
                                            }
                                        }
                                    }
                                }

                                //
                                // NOTE: Purge all the undefined (i.e. deleted)
                                //       variables from the [global] call frame.
                                //
                                CallFrameOps.Cleanup(frame, frame, false);
                            }
                        }
                    }

                    if (safe)
                        createFlags |= CreateFlags.SafeAndHideUnsafe;
                    else
                        createFlags &= ~CreateFlags.SafeAndHideUnsafe;
                }
                else
                {
                    return ReturnCode.Error;
                }
            }

            //
            // NOTE: *WARNING* Please do not move this inside of the lock
            //       statement.
            //
            if (safe)
            {
                //
                // NOTE: Finally, attempt to evaluate the safe interpreter
                //       initialization script (i.e. "safe.eagle").
                //
                ScriptFlags scriptFlags = ScriptFlags.CoreLibraryRequiredFile;

                ReturnCode code = EvaluateScript(this, FileName.Safe,
                    ref scriptFlags, ref error);

                if (code != ReturnCode.Ok)
                    return code;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MarkTrusted(ref Result error)
        {
            CheckDisposed();

            lock (syncRoot)
            {
                createFlags &= ~CreateFlags.Safe;
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsStandard()
        {
            CheckDisposed();

            return FlagOps.HasFlags(createFlags, CreateFlags.Standard, true);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode MakeStandard(
            bool standard,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (commands != null)
                {
                    foreach (KeyValuePair<string, _Wrappers.Command> pair in commands)
                    {
                        if (pair.Value != null)
                        {
                            CommandFlags flags = pair.Value.Flags;

                            if (!FlagOps.HasFlags(flags, CommandFlags.Standard, true))
                            {
                                if (standard)
                                    flags |= CommandFlags.Hidden;
                                else
                                    flags &= ~CommandFlags.Hidden;

                                pair.Value.Flags = flags;
                            }
                        }
                    }
                }

                if (standard)
                    createFlags |= CreateFlags.StandardAndHideNonStandard;
                else
                    createFlags &= ~CreateFlags.StandardAndHideNonStandard;
            }

            return ReturnCode.Ok;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ISynchronize Members
        public object SyncRoot
        {
            //
            // TODO: Rethink this strategy.  Should calls for the interpreter
            //       lock from inside the interpreter methods be able to throw
            //       an exception?  If so, why?  If not, why not?
            //
            get { CheckDisposed(); return syncRoot; }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ITclManager Members
#if NATIVE && TCL
        #region Tcl Loader Support
        public bool HasTcl(ref Result error)
        {
            CheckDisposed();

            bool result = (tclApi != null);

            if (!result)
                error = "no Tcl available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode LoadTcl(
            FindFlags flags,
            string path,
            Version minimumRequired,
            Version maximumRequired,
            Version unknown,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclInterpreters(ref result))
                    {
                        if (IsPrimaryThread())
                        {
                            GlobalState.PushActiveInterpreter(this);

                            try
                            {
#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Tcl, NotifyFlags.PreLoad,
                                    //
                                    // BUGBUG: In order to use this class for notification
                                    //         parameters, it really should probably be
                                    //         made public.
                                    //
                                    new ObjectList(flags, path, minimumRequired, maximumRequired, unknown), this,
                                    null, null, null, ref result);
#endif

                                IntPtr interp = IntPtr.Zero;

                                if (TclWrapper.Load(this, flags, null, path, minimumRequired,
                                        maximumRequired, unknown, clientData, ref tclApi,
                                        ref interp, ref result) == ReturnCode.Ok)
                                {
                                    string name = FormatOps.Id(tclMasterInterpPrefix, null, NextId());

                                    //
                                    // NOTE: Even though the master Tcl interpreter is not really intended
                                    //       for general purpose scripting use, add it anyhow so that we
                                    //       can keep track of it and delete it later.  This Tcl interpreter
                                    //       is always considered to be "fully trusted" and has the full set
                                    //       of Tcl commands available to it.
                                    //
                                    tclInterps.Add(name, interp);
                                    result = name;

#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        NotifyType.Tcl, NotifyFlags.Load,
                                        //
                                        // BUGBUG: In order to use this class for notification
                                        //         parameters, it really should probably be
                                        //         made public.
                                        //
                                        new ObjectList(flags, path, minimumRequired, maximumRequired, unknown), this,
                                        null, null, null, ref result);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                            finally
                            {
                                /* IGNORED */
                                GlobalState.PopActiveInterpreter();
                            }
                        }
                        else
                        {
                            result = "cannot load Tcl, wrong thread";
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnloadTcl(
            UnloadFlags flags, /* in */
            ref Result error   /* out */
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclInterpreters(ref error))
                    {
                        if (IsPrimaryThread())
                        {
                            GlobalState.PushActiveInterpreter(this);

                            try
                            {
#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Tcl, NotifyFlags.PreUnload,
                                    flags, this,
                                    null, null, null, ref error);
#endif

                                //
                                // NOTE: First, attempt to cleanup and/or dispose
                                //       any Tcl related objects.
                                //
                                if (DisposeTcl(false, false, false, ref error) == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Since the Tcl master interpreter should
                                    //       already be deleted now, just [implicitly]
                                    //       pass an invalid Tcl interpreter handle to
                                    //       Unload here, which will cause it to skip
                                    //       attempting to delete the Tcl master
                                    //       interpreter.  After this call it may no
                                    //       longer be safe to use the Tcl API because
                                    //       we may be holding the last reference to it
                                    //       and this call will then cause it to be
                                    //       completely unloaded from memory.
                                    //
                                    if (TclWrapper.Unload(this, flags, ref tclApi,
                                            ref error) == ReturnCode.Ok)
                                    {
#if NOTIFY
                                        /* IGNORED */
                                        CheckNotifications(
                                            NotifyType.Tcl, NotifyFlags.Unload,
                                            flags, this,
                                            null, null, null, ref error);
#endif

                                        return ReturnCode.Ok;
                                    }
                                }
                            }
                            finally
                            {
                                /* IGNORED */
                                GlobalState.PopActiveInterpreter();
                            }
                        }
                        else
                        {
                            error = "cannot unload Tcl, wrong thread";
                        }
                    }
                }
                else
                {
                    error = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if TCL_WRAPPER
        public ITclApi TclApi
        {
            get
            {
                CheckDisposed();

                return tclApi;
            }
            set
            {
                CheckDisposed();

                tclApi = value;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Interpreter Support
        public bool HasTclInterpreters(ref Result error)
        {
            CheckDisposed();

            bool result = (tclInterps != null);

            if (!result)
                error = "no Tcl interpreters available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsTclInterpreterReady(
            string name,
            ref Result error
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            if (GetTclInterpreter(name, ref interp, ref error) == ReturnCode.Ok)
                return TclWrapper.IsReady(tclApi, interp, ref error);

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesTclInterpreterExist(string name)
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;
            Result error = null;

            return GetTclInterpreter(name, ref interp, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetTclInterpreter(
            string name,
            ref IntPtr interp,
            ref Result error
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (HasTclInterpreters(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty Tcl interpreter names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (tclInterps.TryGetValue(name, out interp))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "invalid Tcl interpreter \"{0}\"",
                                name);
                    }
                    else
                    {
                        error = "invalid Tcl interpreter name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool IsTclInterpreterActive(
            string name
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            if (GetTclInterpreter(name, ref interp) == ReturnCode.Ok)
                return TclWrapper.GetInterpActive(tclApi, interp);

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public int GetTclInterpreterErrorLine(
            string name
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            if (GetTclInterpreter(name, ref interp) == ReturnCode.Ok)
                return TclWrapper.GetErrorLine(tclApi, interp);

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CreateTclInterpreter(
            bool initialize,
            bool memory,
            bool safe,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclInterpreters(ref result))
                    {
                        GlobalState.PushActiveInterpreter(this);

                        try
                        {
                            IntPtr interp = IntPtr.Zero;

                            if (TclWrapper.CreateInterpreter(tclApi, initialize, memory, safe,
                                    ref interp, ref result) == ReturnCode.Ok)
                            {
                                string name =
                                    FormatOps.Id(safe ? tclSafeInterpPrefix : tclInterpPrefix, null, NextId());

                                tclInterps.Add(name, interp);
                                result = name;

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Tcl | NotifyType.Interpreter, NotifyFlags.Added,
                                    //
                                    // BUGBUG: In order to use this class for notification
                                    //         parameters, it really should probably be
                                    //         made public.
                                    //
                                    new ObjectList(initialize, safe, name, interp), this,
                                    null, null, null, ref result);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        finally
                        {
                            /* IGNORED */
                            GlobalState.PopActiveInterpreter();
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DeleteTclInterpreter(
            string name,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclInterpreters(ref result))
                    {
                        GlobalState.PushActiveInterpreter(this);

                        try
                        {
                            IntPtr interp = IntPtr.Zero;

                            if (GetTclInterpreter(name, ref interp, ref result) == ReturnCode.Ok)
                            {
                                if (
#if NATIVE_PACKAGE
                                    !NativePackage.ShouldDeleteTclInterpreter(name, interp) ||
#endif
                                    TclWrapper.DeleteInterpreter(
                                        tclApi, false, ref interp, ref result) == ReturnCode.Ok)
                                {
                                    tclInterps.Remove(name);
                                    result = String.Empty;

#if NOTIFY
                                    /* IGNORED */
                                    CheckNotifications(
                                        NotifyType.Tcl | NotifyType.Interpreter, NotifyFlags.Removed,
                                        new ObjectPair(name, interp), this,
                                        null, null, null, ref result);
#endif

                                    return ReturnCode.Ok;
                                }
                            }
                        }
                        finally
                        {
                            /* IGNORED */
                            GlobalState.PopActiveInterpreter();
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool HasTclInterpreters()
        {
            Result error = null;
            return HasTclInterpreters(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclInterpreter(
            string name,
            ref IntPtr interp
            )
        {
            Result error = null;

            return GetTclInterpreter(name, ref interp, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetTclMasterInterpreter(
            ref string name,
            ref IntPtr interp,
            ref Result error
            )
        {
            /* CheckDisposed(); */

            lock (syncRoot)
            {
                if (HasTclInterpreters(ref error))
                {
                    //
                    // HACK: Use the convention that we always prefix the
                    //       master Tcl interpreter with a specially formatted
                    //       name in order to find it (we cannot simply use the
                    //       Tcl interpreter at index zero because the Tcl
                    //       interpreters are maintained in a dictionary).
                    //
                    string pattern = FormatOps.Id(tclMasterInterpPrefix, null,
                        Characters.Asterisk.ToString());

                    //
                    // NOTE: Search all the Tcl interpreters in this interpreter
                    //       for the master Tcl interp stopping as soon as we
                    //       find a valid one (there should only be one valid
                    //       master Tcl interpreter per interpreter at any given
                    //       time).
                    //
                    string key = null;
                    IntPtr value = IntPtr.Zero;

                    foreach (KeyValuePair<string, IntPtr> pair in tclInterps)
                    {
                        //
                        // NOTE: Hard-coded match mode is OK here.
                        //
                        if (StringOps.Match(
                                MatchMode.Glob, pair.Key, pattern, false) &&
                            (pair.Value != IntPtr.Zero))
                        {
                            key = pair.Key;
                            value = pair.Value;
                            break;
                        }
                    }

                    //
                    // NOTE: Did we find the handle for the master Tcl
                    //       interpreter?
                    //
                    if (key != null)
                    {
                        name = key;
                        interp = value;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "Tcl master interpreter is not available";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: Used by the Execute method of the _Commands.Tcl class to help
        //       implement the [tcl master] sub-command.
        //
        internal ReturnCode GetAnyTclMasterInterpreter(
            ref string name,
            ref IntPtr interp,
            ref Result error
            )
        {
            if ((GetTclMasterInterpreter(
                ref name, ref interp, ref error) == ReturnCode.Ok)
#if NATIVE_PACKAGE
                || (NativePackage.GetMasterTclInterpreter(
                    ref name, ref interp, ref error) == ReturnCode.Ok)
#endif
                )
            {
                return ReturnCode.Ok;
            }
            else
            {
                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsTclInterpreterReady(
            string name
            )
        {
            Result error = null;

            return IsTclInterpreterReady(name, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool FindTclInterpreterThreadId(
            IntPtr interp,
            ref int threadId
            )
        {
            IntPtrList interps = null;
#if TCL_THREADS
            TclThreadList threads = null;
#endif

            //
            // BUGBUG: This is likely to be a serious performance bottleneck
            //         since this method must be called for every Tcl API
            //         function call that involves a Tcl interpreter or a Tcl
            //         object.
            //
            lock (syncRoot)
            {
                //
                // NOTE: Get the list of Tcl interps from the
                //       interpreter.
                //
                if (HasTclInterpreters())
                    interps = new IntPtrList(tclInterps.Values);

#if TCL_THREADS
                //
                // NOTE: Get the list of Tcl threads from the
                //       interpreter.
                //
                if (HasTclThreads())
                    threads = new TclThreadList(tclThreads.Values);
#endif
            }

            //
            // NOTE: First, check the Tcl interps that belong to the
            //       primary thread for this interpreter.
            //
            if (interps != null)
            {
                foreach (IntPtr localInterp in interps)
                {
                    if (localInterp == interp)
                    {
                        //
                        // NOTE: Return the primary thread Id.
                        //
                        threadId = this.threadId;

                        //
                        // NOTE: Success, we found the Tcl interpreter.
                        //
                        return true;
                    }
                }
            }

#if TCL_THREADS
            //
            // NOTE: Next, check the Tcl threads that belong to this
            //       interpreter.  Each one of these threads has an
            //       isolated Tcl interpreter.
            //
            if (threads != null)
            {
                foreach (TclThread thread in threads)
                {
                    if ((thread != null) && (thread.Interp == interp))
                    {
                        //
                        // NOTE: Return the specified thread Id for this Tcl
                        //       interpreter.
                        //
                        threadId = thread.ThreadId;

                        //
                        // NOTE: Success, we found the Tcl interpreter.
                        //
                        return true;
                    }
                }
            }
#endif

#if TCL_WRAPPER
            //
            // NOTE: In "Tcl wrapper" mode, we assume that any Tcl
            //       interpreter we cannot otherwise locate was
            //       created on the main thread of the Eagle
            //       interpreter.
            //
            if (FlagOps.HasFlags(createFlags, CreateFlags.TclWrapper, true))
            {
                //
                // NOTE: Return the primary thread Id.
                //
                threadId = this.threadId;

                return true;
            }
#endif

            //
            // NOTE: Failure, we did not find the Tcl interpreter.
            //
            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclAliasTarget(
            ref IExecute target,
            ref Result error
            )
        {
            ICommand command = null;

            if (GetCommand(typeof(_Commands.Tcl), true, true, ref command, ref error) == ReturnCode.Ok)
            {
                target = command;

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddTclAlias(string name, ref Result error)
        {
            IExecute target = null;

            if (GetTclAliasTarget(ref target, ref error) == ReturnCode.Ok)
            {
                ArgumentList arguments = null;

                if (RuntimeOps.GetTclAliasArguments(ref arguments, ref error) == ReturnCode.Ok)
                {
                    Result result = null;

                    if (AddAlias(name, CommandFlags.None, ClientData.Empty, this, target,
                            arguments, null, 2 /* tcl eval */, ref result) == ReturnCode.Ok)
                    {
                        return ReturnCode.Ok;
                    }
                    else
                    {
                        //
                        // NOTE: Transfer error message to caller.
                        //
                        error = result;
                    }
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Thread Support
#if TCL_THREADS
        public bool HasTclThreads(ref Result error)
        {
            CheckDisposed();

            bool result = (tclThreads != null);

            if (!result)
                error = "no Tcl threads available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesTclThreadExist(string name)
        {
            CheckDisposed();

            TclThread thread = null;
            Result error = null;

            return GetTclThread(name, ref thread, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if TCL_WRAPPER
        public
#else
        internal
#endif
        ReturnCode GetTclThread(
            string name,
            ref TclThread thread,
            ref Result error
            )
        {
#if TCL_WRAPPER
            CheckDisposed();
#endif

            lock (syncRoot)
            {
                if (HasTclThreads(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty Tcl thread names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (tclThreads.TryGetValue(name, out thread))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "invalid Tcl thread \"{0}\"",
                                name);
                    }
                    else
                    {
                        error = "invalid Tcl thread name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CreateTclThread(
            int timeout,
            bool generic,
            bool debug,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclThreads(ref result))
                    {
                        GlobalState.PushActiveInterpreter(this);

                        try
                        {
                            string name =
                                FormatOps.Id(tclThreadPrefix, null, NextId());

                            TclThread thread = TclThread.Create(
                                this, ClientData.Empty, timeout, name, generic,
                                debug, ref result);

                            if (thread != null)
                            {
                                tclThreads.Add(name, thread);
                                result = name;

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Tcl | NotifyType.Thread, NotifyFlags.Added,
                                    new ObjectPair(name, thread), this,
                                    null, null, null, ref result);
#endif

                                return ReturnCode.Ok;
                            }
                        }
                        finally
                        {
                            /* IGNORED */
                            GlobalState.PopActiveInterpreter();
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DeleteTclThread(
            string name,
            bool strict,
            ref Result result
            )
        {
            CheckDisposed();

            TclThread thread = null;

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclThreads(ref result))
                    {
                        if ((GetTclThread(name, ref thread, ref result) == ReturnCode.Ok) &&
                            (thread == null))
                        {
                            result = "invalid Tcl thread";
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            //
            // NOTE: If we managed to find the thread above then attempt
            //       to shut it down now.  We do this without holding a
            //       lock on the interpreter to avoid a deadlock with the
            //       other thread when it attempts to use the Tcl API
            //       object associated with this interpreter.
            //
            if (thread != null)
            {
                GlobalState.PushActiveInterpreter(this);

                try
                {
                    if (thread.Shutdown(true, false, strict, ref result) == ReturnCode.Ok)
                    {
                        lock (syncRoot)
                        {
                            tclThreads.Remove(name);
                        }

                        result = String.Empty;

#if NOTIFY
                        /* IGNORED */
                        CheckNotifications(
                            NotifyType.Tcl | NotifyType.Thread, NotifyFlags.Removed,
                            new ObjectPair(name, thread), this,
                            null, null, null, ref result);
#endif

                        return ReturnCode.Ok;
                    }
                }
                finally
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool HasTclThreads()
        {
            Result error = null;
            return HasTclThreads(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclThread(
            string name,
            ref TclThread thread
            )
        {
            Result error = null;

            return GetTclThread(name, ref thread, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // WARNING: The returned Tcl interpreter may only be used in conjunction with the Tcl
        //          C API on the thread that it was created on.  This restriction is NOT
        //          enforced by this function as we are merely returning the pointer to it.
        //
        private ReturnCode GetTclInterpreterOrThread(
            string name,
            ref IntPtr interp,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasTclInterpreters(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty Tcl interpreter and thread names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (tclInterps.TryGetValue(name, out interp))
                        {
                            return ReturnCode.Ok;
                        }
                        else
                        {
                            if (HasTclThreads(ref error))
                            {
                                TclThread thread;

                                if (tclThreads.TryGetValue(name, out thread))
                                {
                                    interp = thread.Interp;
                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    error = String.Format(
                                        "invalid Tcl interpreter or thread \"{0}\"",
                                        name);
                                }
                            }
                        }
                    }
                    else
                    {
                        error = "invalid Tcl interpreter or thread name";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclInterpreters(
            string pattern,
            bool noCase,
            bool strict,
            ref IntPtrDictionary interps,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                Result localError = null;

                if (HasTclInterpreters(ref localError))
                {
                    if (interps == null)
                        interps = new IntPtrDictionary();

                    foreach (KeyValuePair<string, IntPtr> pair in tclInterps)
                    {
                        if ((pattern == null) ||
                            StringOps.Match(
                                MatchMode.Glob, pair.Key, pattern, noCase))
                        {
                            interps.Add(pair.Key, pair.Value);
                        }
                    }

                    return ReturnCode.Ok;
                }
                else if (strict)
                {
                    error = localError;
                    return ReturnCode.Error;
                }
                else
                {
                    return ReturnCode.Ok;
                }
            }
        }
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Thread Event Support
#if TCL_THREADS
        public ReturnCode QueueTclThreadEvent(
            string threadName,
            EventType type,
            EventFlags flags,
            object data,
            bool synchronous,
            ref Result result
            )
        {
            CheckDisposed();

            TclThread thread = null;

            if (GetTclThread(threadName, ref thread, ref result) == ReturnCode.Ok)
            {
                if (thread != null)
                {
                    if (thread.QueueEvent(type, flags, data, synchronous,
                            ref result) == ReturnCode.Ok)
                    {
#if NOTIFY
                        /* IGNORED */
                        CheckNotifications(
                            NotifyType.Tcl | NotifyType.Thread | NotifyType.Event, NotifyFlags.Queued,
                            new ObjectList(threadName, thread, type, flags, data, synchronous),
                            this, null, null, null, ref result);
#endif

                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    result = "invalid Tcl thread";
                }
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Script Support
        public ReturnCode CancelTclEvaluate(
            string interpName,
            bool unwind,
            Result result,
            ref Result error
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return CancelTclEvaluate(interpName, unwind, result, ref clientData, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode CancelTclEvaluate(
            string interpName,
            bool unwind,
            Result result,
            ref IClientData clientData,
            ref Result error
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            ReturnCode code = GetTclInterpreter(interpName, ref interp, ref result);

            if (code == ReturnCode.Ok)
                code = TclWrapper.CancelEvaluate(tclApi, interp, unwind, result, ref clientData, ref error);

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.Tcl | NotifyType.Script, NotifyFlags.Canceled,
                new ObjectTriplet(unwind, result, code), this,
                null, null, null, ref error);
#endif

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclExpression(
            string interpName,
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclExpression(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclExpression(
            string interpName,
            string text,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return EvaluateTclExpression(interpName, text, exceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclExpression(
            string interpName,
            string text,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclExpression(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclExpression(
            string interpName,
            string text,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            ReturnCode code = GetTclInterpreter(interpName, ref interp, ref result);

            if (code == ReturnCode.Ok)
            {
                string name = StringList.MakeList("tcl expr", interpName);

                ICallFrame frame = NewLinkedCallFrame(
                    name, null, CallFrameFlags.Expression |
                    CallFrameFlags.Tcl | CallFrameFlags.Tracking);

                PushCallFrame(frame);

                code = TclWrapper.EvaluateExpression(
                    tclApi, interp, text, exceptions, ref clientData, ref result);

                if (code == ReturnCode.Error)
                {
                    Result errorInfo = null;

                    if (TclWrapper.GetVariable(tclApi, interp, Tcl_VarFlags.TCL_GLOBAL_ONLY,
                            TclVars.ErrorInfo, ref errorInfo, ref errorInfo) == ReturnCode.Ok)
                    {
                        Engine.AddErrorInformation(this, null,
                            String.Format("{0}{1}", HasErrorInProgress()
                                ? Environment.NewLine : String.Empty, errorInfo));
                    }

                    Engine.AddErrorInformation(this, result,
                        String.Format("{0}    (\"tcl expr\" body line {1})",
                            Environment.NewLine, TclWrapper.GetErrorLine(tclApi, interp)));
                }

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Tcl | NotifyType.Expression, NotifyFlags.Evaluated,
                    new ObjectTriplet(interpName, text, code), this,
                    null, null, null, ref result);
#endif

                //
                // NOTE: Pop the original call frame that we pushed above and
                //       any intervening scope call frames that may be leftover
                //       (i.e. they were not explicitly closed).
                //
                PopCallFrame(CallFrameFlags.Scope, true);
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text, Tcl_EvalFlags.TCL_EVAL_NONE,
                exceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text, Tcl_EvalFlags.TCL_EVAL_NONE,
                exceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            Tcl_EvalFlags flags,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text, flags,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            Tcl_EvalFlags flags,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return EvaluateTclScript(interpName, text, flags, exceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            Tcl_EvalFlags flags,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclScript(interpName, text, flags,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclScript(
            string interpName,
            string text,
            Tcl_EvalFlags flags,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            ReturnCode code = GetTclInterpreter(interpName, ref interp, ref result);

            if (code == ReturnCode.Ok)
            {
                string name = StringList.MakeList("tcl eval", interpName);

                ICallFrame frame = NewLinkedCallFrame(
                    name, null, CallFrameFlags.Evaluate |
                    CallFrameFlags.Tcl | CallFrameFlags.Tracking);

                PushCallFrame(frame);

                code = TclWrapper.EvaluateScript(
                    tclApi, interp, text, flags, exceptions, ref clientData, ref result);

                if (code == ReturnCode.Error)
                {
                    Result errorInfo = null;

                    if (TclWrapper.GetVariable(tclApi, interp, Tcl_VarFlags.TCL_GLOBAL_ONLY,
                            TclVars.ErrorInfo, ref errorInfo, ref errorInfo) == ReturnCode.Ok)
                    {
                        Engine.AddErrorInformation(this, null,
                            String.Format("{0}{1}", HasErrorInProgress()
                                ? Environment.NewLine : String.Empty, errorInfo));
                    }

                    Engine.AddErrorInformation(this, result,
                        String.Format("{0}    (\"tcl eval\" body line {1})",
                            Environment.NewLine, TclWrapper.GetErrorLine(tclApi, interp)));
                }

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Tcl | NotifyType.Script, NotifyFlags.Evaluated,
                    new ObjectTriplet(interpName, text, flags), this,
                    null, null, null, ref result);
#endif

                //
                // NOTE: Pop the original call frame that we pushed above and
                //       any intervening scope call frames that may be leftover
                //       (i.e. they were not explicitly closed).
                //
                PopCallFrame(CallFrameFlags.Scope, true);
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclFile(
            string interpName,
            string fileName,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclFile(interpName, fileName,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclFile(
            string interpName,
            string fileName,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return EvaluateTclFile(interpName, fileName, exceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclFile(
            string interpName,
            string fileName,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return EvaluateTclFile(interpName, fileName,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode EvaluateTclFile(
            string interpName,
            string fileName,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            ReturnCode code = GetTclInterpreter(interpName, ref interp, ref result);

            if (code == ReturnCode.Ok)
            {
                string name = StringList.MakeList("tcl source", interpName, fileName);

                ICallFrame frame = NewLinkedCallFrame(
                    name, null, CallFrameFlags.Source |
                    CallFrameFlags.Tcl | CallFrameFlags.Tracking);

                PushCallFrame(frame);

                code = TclWrapper.EvaluateFile(
                    tclApi, interp, fileName, exceptions, ref clientData, ref result);

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Tcl | NotifyType.File, NotifyFlags.Evaluated,
                    new ObjectPair(interpName, fileName), this,
                    null, null, null, ref result);
#endif

                //
                // NOTE: Pop the original call frame that we pushed above and
                //       any intervening scope call frames that may be leftover
                //       (i.e. they were not explicitly closed).
                //
                PopCallFrame(CallFrameFlags.Scope, true);
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteTclString(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return SubstituteTclString(interpName, text, exceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteTclString(interpName, text,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteTclString(interpName, text, Tcl_SubstFlags.TCL_SUBST_ALL, exceptions,
                ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            Tcl_SubstFlags flags,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return SubstituteTclString(interpName, text, flags, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            Tcl_SubstFlags flags,
            bool exceptions,
            ref Result result
            )
        {
            CheckDisposed();

            IClientData clientData = null;

            return SubstituteTclString(interpName, text, flags, exceptions, ref clientData,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            Tcl_SubstFlags flags,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            return SubstituteTclString(interpName, text, flags,
                _Components.Private.Tcl.TclApi.DefaultExceptions, ref clientData, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SubstituteTclString(
            string interpName,
            string text,
            Tcl_SubstFlags flags,
            bool exceptions,
            ref IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            IntPtr interp = IntPtr.Zero;

            ReturnCode code = GetTclInterpreter(interpName, ref interp, ref result);

            if (code == ReturnCode.Ok)
            {
                string name = StringList.MakeList("tcl subst", interpName);

                ICallFrame frame = NewLinkedCallFrame(
                    name, null, CallFrameFlags.Substitute |
                    CallFrameFlags.Tcl | CallFrameFlags.Tracking);

                PushCallFrame(frame);

                code = TclWrapper.SubstituteString(
                    tclApi, interp, text, flags, exceptions, ref clientData, ref result);

                if (code == ReturnCode.Error)
                {
                    Result errorInfo = null;

                    if (TclWrapper.GetVariable(tclApi, interp, Tcl_VarFlags.TCL_GLOBAL_ONLY,
                            TclVars.ErrorInfo, ref errorInfo, ref errorInfo) == ReturnCode.Ok)
                    {
                        Engine.AddErrorInformation(this, null,
                            String.Format("{0}{1}", HasErrorInProgress()
                                ? Environment.NewLine : String.Empty, errorInfo));
                    }

                    Engine.AddErrorInformation(this, result,
                        String.Format("{0}    (\"tcl subst\" body line {1})",
                            Environment.NewLine, TclWrapper.GetErrorLine(tclApi, interp)));
                }

#if NOTIFY
                /* IGNORED */
                CheckNotifications(
                    NotifyType.Tcl | NotifyType.String, NotifyFlags.Substituted,
                    new ObjectTriplet(interpName, text, code), this,
                    null, null, null, ref result);
#endif

                //
                // NOTE: Pop the original call frame that we pushed above and
                //       any intervening scope call frames that may be leftover
                //       (i.e. they were not explicitly closed).
                //
                PopCallFrame(CallFrameFlags.Scope, true);
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tcl Command Support
        public bool HasTclBridges(ref Result error)
        {
            CheckDisposed();

            bool result = (tclBridges != null);

            if (!result)
                error = "no Tcl bridges available";

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode DoesTclBridgeExist(string name)
        {
            CheckDisposed();

            TclBridge tclBridge = null;
            Result error = null;

            return GetTclBridge(name, ref tclBridge, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode AddTclBridge(
            IExecute execute,
            string interpName,
            string commandName,
            IClientData clientData,
            bool forceDelete,
            bool noComplain,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (readOnly)
                {
                    result = "interpreter is read-only";
                    return ReturnCode.Error;
                }

                if (execute == null)
                {
                    result = "invalid execute";
                    return ReturnCode.Error;
                }

                if (!HasTclInterpreters(ref result))
                    return ReturnCode.Error;

                if (!HasTclBridges(ref result))
                    return ReturnCode.Error;

                IntPtr interp = IntPtr.Zero;

                //
                // NOTE: *THREADING* Get the Tcl interpreter even if it is in an isolated thread.
                //       If the Tcl interpreter is from an isolated thread then this method may
                //       only be called from that thread.
                //
#if TCL_THREADS
                if (GetTclInterpreterOrThread(interpName, ref interp, ref result) != ReturnCode.Ok)
#else
                if (GetTclInterpreter(interpName, ref interp, ref result) != ReturnCode.Ok)
#endif
                    return ReturnCode.Error;

                string bridgeName =
                    FormatOps.TclBridgeName(interpName, commandName);

                if (bridgeName == null)
                {
                    result = "invalid bridge name";
                    return ReturnCode.Error;
                }

                if (DoesTclBridgeExist(bridgeName) == ReturnCode.Ok)
                {
                    result = String.Format(
                        "can't add \"{0}\": command already exists in Tcl interpreter \"{1}\"",
                        commandName, interpName);

                    return ReturnCode.Error;
                }

                GlobalState.PushActiveInterpreter(this);

                try
                {
                    TclBridge tclBridge = TclBridge.Create(this, execute,
                        (clientData != null) ? clientData : RuntimeOps.GetClientData(execute),
                        interp, commandName, forceDelete, noComplain, ref result);

                    if (tclBridge == null)
                        return ReturnCode.Error;

                    tclBridges.Add(bridgeName, tclBridge);

#if NOTIFY
                    /* IGNORED */
                    CheckNotifications(
                        NotifyType.Tcl | NotifyType.Command, NotifyFlags.Added,
                        //
                        // BUGBUG: In order to use this class for notification
                        //         parameters, it really should probably be
                        //         made public.
                        //
                        new ObjectList(execute, interpName, commandName, bridgeName,
                            forceDelete, tclBridge), this,
                        clientData, null, null, ref result);
#endif

                    result = bridgeName;
                }
                finally
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                }
            }

            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode RemoveTclBridge(
            string interpName,
            string commandName,
            IClientData clientData,
            ref Result result
            )
        {
            CheckDisposed();

            lock (syncRoot)
            {
                if (!readOnly)
                {
                    if (HasTclBridges(ref result))
                    {
                        string bridgeName =
                            FormatOps.TclBridgeName(interpName, commandName);

                        TclBridge tclBridge = null;

                        if (GetTclBridge(bridgeName, ref tclBridge, ref result) == ReturnCode.Ok)
                        {
                            GlobalState.PushActiveInterpreter(this);

                            try
                            {
                                if (tclBridge != null)
                                {
                                    tclBridge.Dispose(); /* throw */
                                    tclBridge = null;
                                }

                                tclBridges.Remove(bridgeName);

#if NOTIFY
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Tcl | NotifyType.Command, NotifyFlags.Removed,
                                    //
                                    // BUGBUG: In order to use this class for notification
                                    //         parameters, it really should probably be
                                    //         made public.
                                    //
                                    new ObjectList(interpName, commandName, bridgeName, tclBridge), this,
                                    clientData, null, null, ref result);
#endif

                                result = String.Empty;
                                return ReturnCode.Ok;
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while disposing Tcl bridge: {0}",
                                    e);

                                return ReturnCode.Error;
                            }
                            finally
                            {
                                /* IGNORED */
                                GlobalState.PopActiveInterpreter();
                            }
                        }
                    }
                }
                else
                {
                    result = "interpreter is read-only";
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal bool HasTclBridges() // NOTE: This overload is for use by the TclBridge object only.
        {
            Result error = null;
            return HasTclBridges(ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclBridge(
            string name,
            ref TclBridge tclBridge
            )
        {
            Result error = null;

            return GetTclBridge(name, ref tclBridge, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTclBridge(
            string name,
            ref TclBridge tclBridge,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasTclBridges(ref error))
                {
                    //
                    // NOTE: *WARNING* Empty Tcl command names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (name != null)
                    {
                        if (tclBridges.TryGetValue(name, out tclBridge))
                            return ReturnCode.Ok;
                        else
                            error = String.Format(
                                "invalid Tcl command \"{0}\"",
                                name);
                    }
                    else
                    {
                        error = "invalid Tcl command name";
                    }
                }
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IVariableManager Members
        #region Variable Checking
        public ReturnCode DoesVariableExist(VariableFlags flags, string name)
        {
            CheckDisposed();

            bool badRequest = false;
            Result error = null;

            return DoesVariableExist(flags, null, name, ref badRequest, ref error);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Variable Management (Single)
        public ReturnCode AddVariable(
            VariableFlags flags,
            string name,
            TraceList traces,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            IVariable variable = null;

            return AddVariable2(flags, name, traces, strict, ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode GetVariableValue(
            VariableFlags flags,
            string name,
            ref Result value,
            ref Result error
            )
        {
            CheckDisposed();

            string varName = null;
            string varIndex = null;
            IVariable variable = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return GetVariableValue2(
                    flags, varName, varIndex, ref variable,
                    ref value, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode ResetVariable(
            VariableFlags flags,
            string name,
            ref Result error
            )
        {
            CheckDisposed();

            string varName = null;
            string varIndex = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return ResetVariable2(flags, null, varName, varIndex, null, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetVariableLink(
            VariableFlags flags,
            string name,
            MemberInfo memberInfo,
            object @object,
            ref Result error
            )
        {
            CheckDisposed();

            string varName = null;
            string varIndex = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                IVariable variable = null;

                return SetVariableLink2(
                    flags, null, varName, varIndex, memberInfo,
                    @object, ref variable, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetVariableValue(
            VariableFlags flags,
            string name,
            string value,
            TraceList traces,
            ref Result error
            )
        {
            CheckDisposed();

            IVariable variable = null;

            return SetVariableValue(
                flags, name, value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnsetVariable(
            VariableFlags flags,
            string name,
            ref Result error
            )
        {
            CheckDisposed();

            string varName = null;
            string varIndex = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return UnsetVariable2(flags, varName, varIndex, null, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode WaitVariable(
            VariableFlags flags,    /* in */
            string name,            /* in */
            int limit,              /* in */
            EventWaitHandle @event, /* in, optional */
            bool noCancel,          /* in */
            bool stopOnError,       /* in */
            bool errorOnEmpty,      /* in */
            bool userInterface,     /* in */
            ref Result error        /* out */
            ) /* THREAD-SAFE */
        {
            CheckDisposed();

            ReturnCode code;

            if (EnterWait() > 0)
            {
                //
                // NOTE: *WARNING* Empty variable names are allowed, please do not change
                //       these to "!String.IsNullOrEmpty".
                //
                if (name != null)
                {
                    //
                    // NOTE: *WARNING* Sanity check the user-interface flag here.  We MUST
                    //       pump messages if this is an STA thread.
                    //
                    if (!userInterface && RuntimeOps.IsStaThread())
                    {
                        userInterface = true;

                        if (FlagOps.HasFlags(flags, VariableFlags.WaitTrace, true))
                        {
                            //
                            // HACK: Uses the interpreter "id" field directly, to avoid
                            //       disposal exception issues.
                            //
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: WaitVariable: detected STA thread, user-interface " +
                                "flag forcibly enabled, interpreter = {1}, flags = {2}, " +
                                "name = {3}, limit = {4}, noCancel = {5}, stopOnError = {6}, " +
                                "errorOnEmpty = {7}, userInterface = {8}, waitCount = {9}, " +
                                "waitSpinCount = {10}, error = {11}{12}",
                                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(id),
                                FormatOps.WrapOrNull(flags), FormatOps.WrapOrNull(name),
                                limit, noCancel, stopOnError, errorOnEmpty, userInterface,
                                waitCount, waitSpinCount, FormatOps.WrapOrNull(true, true, error),
                                Environment.NewLine), typeof(Interpreter).Name);
                        }
                    }

                    //
                    // NOTE: Has the existence of the variable been checked?
                    //
                    bool once = false;

                    //
                    // NOTE: NOTE: Does the variable currently exist?
                    //
                    bool exists = false;

                    //
                    // NOTE: Has the variable been added, modified, or unset since we
                    //       started waiting?
                    //
                    bool dirty = false;

                    //
                    // NOTE: Create the array of event handles to wait on.  The first event
                    //       is signaled when any variable is set or unset in the interpreter.
                    //       The second event is signaled when the queue managed by the event
                    //       manager is empty.  The third [and final] event is signaled when
                    //       an event is queued to the queue managed by the event manager.
                    //
                    EventWaitHandle[] events = GetEventWaitHandles(@event, true);

                    //
                    // NOTE: Keep track of the iterations of main loop of this method (for
                    //       diagnostic purposes only).
                    //
                    int localWaitSpinCount = 0;

                    //
                    // NOTE: Keep waiting for the variable until one of the following
                    //       conditions is true:
                    //
                    //       1. The interpreter is no longer ready (i.e. canceled, deleted,
                    //          halted, exited, etc).
                    //
                    //       2. The variable is created, deleted, or changed.
                    //
                    //       3. Any error is raised while processing events.
                    //
                    while (((code = Ready(this, false, false, noCancel, ref error)) == ReturnCode.Ok) &&
                           ((code = IsVariableNameDirty(this, name, flags, ref once, ref exists,
                                ref dirty, ref error)) == ReturnCode.Ok) && !dirty)
                    {
                        //
                        // NOTE: We need a local result because we do not want to change
                        //       the caller's result based on random async events that
                        //       get processed while waiting for their variable to become
                        //       "signaled".  However, we will change the caller's result
                        //       if an error is encountered.
                        //
                        Result localResult = null;

                        //
                        // NOTE: Attempt to process all pending events stopping if an error
                        //       is encountered.  This grabs and uses the current wait event
                        //       flags without locking.
                        //
                        code = RuntimeOps.DoOneEvent(
                            this, waitEventFlags, EventPriority.WaitVariable, limit,
                            stopOnError, errorOnEmpty, userInterface, ref localResult);

                        //
                        // NOTE: If we encountered an error processing events, break out
                        //       of the loop and return the error code and result to the
                        //       caller.
                        //
                        if (code != ReturnCode.Ok)
                        {
                            error = localResult;
                            break;
                        }

                        //
                        // NOTE: Get the configured sleep time for this interpreter [or
                        //       the minimum possible sleep time].
                        //
                        int milliseconds = GetMinimumSleepTime();

                        //
                        // NOTE: We always yield to other running threads.  This also
                        //       gives them an opportunity to cancel any script being
                        //       evaluated on this thread and/or update the variable we
                        //       are waiting for.
                        //
                        if (events != null)
                        {
#if NATIVE && WINDOWS
                            /* IGNORED */
                            WindowOps.WaitForMultipleHandles(
                                events, milliseconds, userInterface);
#else
                            /* IGNORED */
                            ThreadOps.WaitAnyEvent(events, milliseconds);
#endif
                        }
                        else
                        {
                            RuntimeOps.Sleep(this, milliseconds);
                        }

                        //
                        // NOTE: Keep track of how many times we actually waited for an
                        //       event.  This does not currently distinguish between a
                        //       timeout and one of the event handles being signaled.
                        //
                        localWaitSpinCount++; /* THREAD-SAFE */
                        Interlocked.Increment(ref waitSpinCount);
                    }

                    ClearVariableNameWait(this, name, flags);

                    //
                    // NOTE: Show diagnostic information about this wait operation?
                    //
                    if (FlagOps.HasFlags(flags, VariableFlags.WaitTrace, true))
                    {
                        //
                        // HACK: Uses the interpreter "id" field directly, to avoid disposal
                        //       exception issues.
                        //
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: WaitVariable: interpreter = {1}, code = {2}, flags = {3}, " +
                            "name = {4}, limit = {5}, noCancel = {6}, stopOnError = {7}, " +
                            "errorOnEmpty = {8}, once = {9}, exists = {10}, dirty = {11}, " +
                            "waitCount = {12}, localWaitSpinCount = {13}, waitSpinCount = {14}, " +
                            "error = {15}{16}", GlobalState.GetCurrentThreadId(),
                            FormatOps.WrapOrNull(id), code, FormatOps.WrapOrNull(flags),
                            FormatOps.WrapOrNull(name), limit, noCancel, stopOnError,
                            errorOnEmpty, once, exists, dirty, waitCount, localWaitSpinCount,
                            waitSpinCount, FormatOps.WrapOrNull(true, true, error),
                            Environment.NewLine), typeof(Interpreter).Name);
                    }
                }
                else
                {
                    error = "invalid variable name";
                    code = ReturnCode.Error;
                }

                /* IGNORED */
                ExitWait();
            }
            else
            {
                error = "wait subsystem locked";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal ReturnCode GetVariable(
            ICallFrame frame,
            string varName,
            string varIndex, /* NOT USED */
            ref VariableFlags flags,
            ref IVariable variable,
            ref Result error
            ) /* USED BY CORE RESOLVER */
        {
            /* CheckDisposed(); */

            //
            // NOTE: Reset the "NotFound" flag because we have not actually
            //       searched for the variable yet.
            //
            flags &= ~VariableFlags.NotFound;

            //
            // NOTE: Make sure the supplied variable name is valid.
            //
            if (varName != null)
            {
                lock (syncRoot) /* TRANSACTIONAL */
                {
                    //
                    // NOTE: Get the applicable call frame based on the provided
                    //       variable name and flags.
                    //
                    ICallFrame variableFrame = GetVariableFrame( /* HOT PATH (TSD) */
                        frame, ref varName, ref flags);

                    if (variableFrame != null)
                    {
                        if (variableFrame.Variables != null)
                        {
                            if (variableFrame.Variables.Count > 0)
                            {
                                //
                                // NOTE: Now, try to get the variable in the desired call frame.
                                //
                                IVariable localVariable = null;

                                if (variableFrame.Variables.TryGetValue(varName, out localVariable))
                                {
                                    //
                                    // NOTE: Now, see if they require that this be an undefined
                                    //       variable (or not).
                                    //
                                    bool wantDefined =
                                        FlagOps.HasFlags(flags, VariableFlags.Defined, true);
                                    bool wantUndefined =
                                        FlagOps.HasFlags(flags, VariableFlags.Undefined, true);
                                    bool isUndefined = EntityOps.IsUndefined(localVariable);

                                    if ((!wantDefined && !wantUndefined) ||
                                        (wantDefined && !isUndefined) ||
                                        (wantUndefined && isUndefined))
                                    {
                                        //
                                        // NOTE: Do they want to allow virtual variables to be
                                        //       returned?  If not, raise an error now.
                                        //
                                        bool wantVirtual =
                                            !FlagOps.HasFlags(flags, VariableFlags.NonVirtual, true);
                                        bool isVirtual = EntityOps.IsVirtual(localVariable);

                                        if (wantVirtual || !isVirtual)
                                        {
                                            //
                                            // NOTE: Do they want us to verify that the variable does
                                            //       NOT contain a link index (i.e. is it an upvar link
                                            //       to an array element)?
                                            //
                                            bool noLinkIndex =
                                                FlagOps.HasFlags(flags, VariableFlags.NoLinkIndex, true);
                                            bool hasLinkIndex = (localVariable != null) &&
                                                (localVariable.LinkIndex != null);

                                            if (!noLinkIndex || !hasLinkIndex)
                                            {
                                                //
                                                // NOTE: Now, see if they require that this be an array
                                                //       variable (or not).
                                                //
                                                bool wantArray =
                                                    !FlagOps.HasFlags(flags, VariableFlags.NoGetArray, true) &&
                                                    FlagOps.HasFlags(flags, VariableFlags.Array, true);
                                                bool wantNoArray =
                                                    !FlagOps.HasFlags(flags, VariableFlags.NoGetArray, true) &&
                                                    FlagOps.HasFlags(flags, VariableFlags.NoArray, true);
                                                bool isArray = EntityOps.IsArray2(localVariable);

                                                if ((!wantArray && !wantNoArray) ||
                                                    (wantArray && isArray) ||
                                                    (wantNoArray && !isArray))
                                                {
                                                    variable = localVariable;

                                                    return ReturnCode.Ok;
                                                }
                                                else
                                                {
                                                    error = String.Format(
                                                        "can't get {0}: variable {1} array",
                                                        FormatOps.ErrorVariableName(varName, null),
                                                        isArray ? "is" : "isn't");
                                                }
                                            }
                                            else
                                            {
                                                //
                                                // NOTE: The variable has a link index (to an array
                                                //       element) and according to the caller, it
                                                //       should not.
                                                //
                                                flags |= VariableFlags.HasLinkIndex;

                                                error = String.Format(
                                                    "can't get {0}: variable is array element link",
                                                    FormatOps.ErrorVariableName(varName, null));
                                            }
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: The variable is virtual and according to the
                                            //       caller, it should not be.
                                            //
                                            flags |= VariableFlags.WasVirtual;

                                            error = String.Format(
                                                "can't get {0}: variable is virtual",
                                                FormatOps.ErrorVariableName(varName, null));
                                        }
                                    }
                                    else
                                    {
                                        //
                                        // BUGFIX: If the variable is undefined, we must set
                                        //         the "NotFound" flag because an undefined
                                        //         variable should be treated just like one
                                        //         that is physically missing from the call
                                        //         frame.
                                        //
                                        if (isUndefined)
                                            flags |= VariableFlags.NotFound;

                                        //
                                        // HACK: This is on the "hot path" for WaitVariable on
                                        //       variables that do not yet exist; therefore, do
                                        //       not use String.Format here.
                                        //
                                        // BUGBUG: Just how slow is String.Format and why?
                                        //
                                        // error = String.Format(
                                        //     "can't get {0}: variable {1} defined",
                                        //     FormatOps.ErrorVariableName(varName, null),
                                        //     isUndefined ? "isn't" : "is");
                                        //
                                        error = isUndefined ?
                                            "variable isn't defined" : "variable is defined";
                                    }
                                }
                                else
                                {
                                    //
                                    // NOTE: The variable does not exist in the call frame,
                                    //       set the "NotFound" flag.
                                    //
                                    flags |= VariableFlags.NotFound;

                                    //
                                    // HACK: This is on the "hot path" for WaitVariable on
                                    //       variables that do not yet exist; therefore, do
                                    //       not use String.Format here.
                                    //
                                    // BUGBUG: Just how slow is String.Format and why?
                                    //
                                    // error = String.Format(
                                    //     "variable {0} not found in call frame \"{1}\"",
                                    //     FormatOps.ErrorVariableName(varName, null),
                                    //     variableFrame);
                                    //
                                    error = "variable not found in call frame";
                                }
                            }
                            else
                            {
                                //
                                // NOTE: The variable does not exist in the call frame (i.e.
                                //       there are no variables in the call frame), set the
                                //       "NotFound" flag.
                                //
                                flags |= VariableFlags.NotFound;

                                error = "no variables in call frame";
                            }
                        }
                        else
                        {
                            error = "call frame does not support variables";
                        }
                    }
                    else
                    {
                        error = "invalid call frame";
                    }
                }
            }
            else
            {
                error = "invalid variable name";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DoesVariableExist(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            ref bool badRequest,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                string varName = null;
                string varIndex = null;

                if (RuntimeOps.SplitVariableName(
                        this, flags, name, ref varName,
                        ref varIndex, ref error) == ReturnCode.Ok)
                {
                    IVariable variable = null;

                    if (GetVariableViaResolversWithSplit(
                            frame, varName, ref flags, ref variable,
                            ref error) == ReturnCode.Ok)
                    {
                        //
                        // NOTE: For linked variables, we may be linking to an array
                        //       element.
                        //
                        string linkIndex = null;

                        if (EntityOps.IsLink(variable))
                        {
                            //
                            // NOTE: Get the array element index of the link, if any.
                            //
                            linkIndex = variable.LinkIndex;

                            //
                            // NOTE: Follow the linked variable.
                            //
                            variable = EntityOps.FollowLinks(variable);
                        }

                        //
                        // NOTE: If they specified the NoElement flag, then make sure
                        //       the linked variable does not refer to an array element.
                        //
                        if ((linkIndex == null) ||
                            !FlagOps.HasFlags(flags, VariableFlags.NoElement, true))
                        {
                            badRequest = false; // var name searched OK and found OK.

                            //
                            // BUGFIX: Null and undefined variables do not "exist".
                            //
                            if ((variable != null) &&
                                !EntityOps.IsUndefined(variable))
                            {
                                //
                                // NOTE: Are they requesting an array, directly or
                                //       indirectly (i.e. indirectly if the specified
                                //       variable name or variable link contains an
                                //       element index and they did not specify the
                                //       NoElement flag)?
                                //
                                if (FlagOps.HasFlags(flags, VariableFlags.Array, true) ||
                                    (varIndex != null) || (linkIndex != null))
                                {
                                    //
                                    // NOTE: Is the variable an array?
                                    //
                                    if (EntityOps.IsArray2(variable) &&
                                        ((varIndex == null) || (linkIndex == null)))
                                    {
                                        if ((varIndex != null) || (linkIndex != null))
                                        {
                                            //
                                            // NOTE: If there was no element index specified,
                                            //       use the one from the variable link.
                                            //
                                            if (varIndex == null)
                                                varIndex = linkIndex;

                                            //
                                            // HACK: Handle the global "env" array specially.
                                            //       We must do this because our global "env"
                                            //       array has no backing storage (unlike Tcl's)
                                            //       and we do not have a trace operation for
                                            //       "exists".
                                            //
                                            ObjectDictionary arrayValue = variable.ArrayValue;

                                            if ((IsEnvironmentVariable(variable) &&
                                                    CommonOps.DoesEnvironmentVariableExist(varIndex)) ||
                                                ((arrayValue != null) && arrayValue.ContainsKey(varIndex)))
                                            {
                                                return ReturnCode.Ok;
                                            }
                                            else if ((IsTestsVariable(variable) &&
                                                    DoesTestsVariableExist(varIndex)) ||
                                                ((arrayValue != null) && arrayValue.ContainsKey(varIndex)))
                                            {
                                                return ReturnCode.Ok;
                                            }
                                            else
                                            {
                                                error = String.Format(
                                                    "can't verify {0}: no such element in array",
                                                    FormatOps.ErrorVariableName(
                                                        variable, linkIndex, varName, varIndex));
                                            }
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: The variable is an array and we did not want to
                                            //       validate any actual element.
                                            //
                                            return ReturnCode.Ok;
                                        }
                                    }
                                    else
                                    {
                                        error = String.Format(
                                            "can't verify {0}: variable isn't array",
                                            FormatOps.ErrorVariableName(
                                                variable, linkIndex, varName, varIndex));
                                    }
                                }
                                else if (FlagOps.HasFlags(flags, VariableFlags.NoArray, true))
                                {
                                    if (EntityOps.IsArray2(variable))
                                    {
                                        //
                                        // NOTE: They requested verification of a scalar variable and
                                        //       the variable is actually an array, this is an error.
                                        //
                                        error = String.Format(
                                            "can't verify {0}: variable is array",
                                            FormatOps.ErrorVariableName(
                                                variable, linkIndex, varName, varIndex));
                                    }
                                    else
                                    {
                                        if (varIndex != null)
                                        {
                                            //
                                            // NOTE: This is somewhat confusing.  They requested
                                            //       verification of a scalar variable; however,
                                            //       they supplied an array element index for a
                                            //       non-array variable?
                                            //
                                            error = String.Format(
                                                "can't verify {0}: variable isn't array",
                                                FormatOps.ErrorVariableName(
                                                    variable, linkIndex, varName, varIndex));
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: They requested verification of a scalar variable
                                            //       and [correctly] did not supply an element index,
                                            //       everything is ok.
                                            //
                                            return ReturnCode.Ok;
                                        }
                                    }
                                }
                                else
                                {
                                    //
                                    // NOTE: No special flags were set and they did not specify an
                                    //       element index.
                                    //
                                    return ReturnCode.Ok;
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't verify {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, varName, varIndex));
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Bad request, they wanted to validate that the
                            //       variable does not refer to an array element
                            //       and the linked variable does.  Indicate this
                            //       special condition to the caller.
                            //
                            badRequest = true; // var name NOT searched OK.

                            //
                            // NOTE: Provide an informative error message.
                            //
                            error = "name refers to an element in an array";
                        }
                    }
                    else
                    {
                        //
                        // NOTE: If we did not actually search for the variable
                        //       (due to the internal calls to parse the variable
                        //       name failing, etc), we do not want to pretend that
                        //       the variable was simply not found.
                        //
                        if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                            badRequest = false; // var name searched OK and NOT found.
                        else
                            badRequest = true; // var name NOT searched OK.
                    }
                }
                else
                {
                    //
                    // NOTE: Bad variable name could not be parsed.
                    //       Indicate this special condition to the
                    //       caller.
                    //
                    badRequest = true; // var name NOT searched OK.
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PostProcessVariableValue(
            VariableFlags flags,
            string name,
            string index,
            IVariable variable,
            object oldValue,
            ref Result value,
            ref Result error
            )
        {
            //
            // NOTE: Do we need to post-process this variable value at all
            //       prior to returning it?
            //
            if (EntityOps.IsSubstitute(variable) ||
                EntityOps.IsEvaluate(variable))
            {
                Result result = null;

                //
                // NOTE: Do we need to post-process the variable value by
                //       processing substitutions contained within it?
                //
                if (EntityOps.IsSubstitute(variable))
                {
                    if (SubstituteString(
                            RuntimeOps.GetStringFromObject(oldValue),
                            ref result) != ReturnCode.Ok)
                    {
                        error = result;
                        return ReturnCode.Error;
                    }

                    oldValue = result;
                }

                //
                // NOTE: Do we need to post-process the variable value by
                //       evaluating it as a script?
                //
                if (EntityOps.IsEvaluate(variable))
                {
                    if (EvaluateScript(
                            RuntimeOps.GetStringFromObject(oldValue),
                            ref result) != ReturnCode.Ok) /* EXEMPT */
                    {
                        error = result;
                        return ReturnCode.Error;
                    }

                    oldValue = result;
                }
            }

            //
            // NOTE: If post-processing was skipped or we were
            //       successful, signal success and return.
            //
            value = RuntimeOps.GetStringFromObject(oldValue);
            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode FireTraces( // NOTE: For use by [array] command only.
            BreakpointType breakpointType,
            VariableFlags flags,
            ICallFrame frame,
            string varName,
            string varIndex,
            object newValue,
            ObjectDictionary oldValues,
            ObjectDictionary newValues,
            IVariable variable,
            ref Result result
            )
        {
            if (!FlagOps.HasFlags(flags, VariableFlags.NonTrace, false) &&
                !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
            {
                object oldValue = EntityOps.GetOldValue(
                    flags, variable, varIndex,
                    RuntimeOps.GetDefaultValue(breakpointType));

                ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(
                    this, null, breakpointType, frame, variable,
                    varName, varIndex, flags, oldValue, newValue,
                    oldValues, newValues, null, false, true,
                    ReturnCode.Ok);

                return RuntimeOps.FireTraces(
                    variable, breakpointType, this, traceInfo,
                    ref result);
            }
            else
            {
                return ReturnCode.Ok;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableValue(
            VariableFlags flags,
            string name,
            ref Result value
            )
        {
            string varName = null;
            string varIndex = null;
            IVariable variable = null;
            Result error = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return GetVariableValue2(
                    flags, varName, varIndex, ref variable,
                    ref value, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableValue2(
            VariableFlags flags,
            string name,
            string index,
            ref Result value,
            ref Result error
            )
        {
            IVariable variable = null;

            return GetVariableValue2(flags, name, index, ref variable, ref value, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableValue2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            ref Result value,
            ref Result error
            )
        {
            IVariable variable = null;

            return GetVariableValue2(flags, frame, name, index, ref variable, ref value, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableValue2(
            VariableFlags flags,
            string name,
            string index,
            ref IVariable variable,
            ref Result value,
            ref Result error
            )
        {
            return GetVariableValue2(flags, null, name, index, ref variable, ref value, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetVariableValue2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            ref IVariable variable,
            ref Result value,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                bool haveVariable = (variable != null);

                if ((haveVariable ||
                    (GetVariableViaResolversWithSplit(frame, name, index, ref flags,
                        ref variable, ref error) == ReturnCode.Ok)) &&
                    !EntityOps.IsUndefined(variable, "read", name, index, ref error))
                {
                    //
                    // COMPAT: Tcl, if the index was specified, make sure the name
                    //         itself does not contain an array element reference.
                    //
                    if (haveVariable || (index == null))
                    {
                        VariableFlags splitFlags = flags;
                        string splitIndex = null;

                        if (index != null)
                            splitFlags |= VariableFlags.NoElement;

                        if (RuntimeOps.SplitVariableName(
                                this, splitFlags, name, ref name,
                                ref splitIndex, ref error) == ReturnCode.Ok)
                        {
                            if ((index == null) && (splitIndex != null))
                                index = splitIndex;
                        }
                        else
                        {
                            return ReturnCode.Error;
                        }
                    }

                    //
                    // NOTE: Make sure we have a variable.
                    //
                    if (variable == null)
                    {
                        error = String.Format(
                            "can't read {0}: variable is invalid",
                            FormatOps.ErrorVariableName(name, index));

                        return ReturnCode.Error;
                    }

                    //
                    // NOTE: For linked variables, we may be linking to an array
                    //       element.
                    //
                    string linkIndex = null;

                    //
                    // NOTE: Save original call frame for traces.
                    //
                    ICallFrame savedFrame = variable.Frame;

                    //
                    // NOTE: Is the variable really a link to another variable?
                    //
                    if (EntityOps.IsLink(variable))
                    {
                        //
                        // NOTE: Get the array element index of the link, if any.
                        //
                        linkIndex = variable.LinkIndex;

                        //
                        // NOTE: Follow the linked variable.
                        //
                        variable = EntityOps.FollowLinks(variable);
                    }

                    if (!EntityOps.IsUndefined(variable))
                    {
                        if ((index != null) || (linkIndex != null))
                        {
                            if (EntityOps.IsArray2(variable) &&
                                ((index == null) || (linkIndex == null)))
                            {
                                //
                                // NOTE: If there was no element index specified, use
                                //       the one from the variable link.
                                //
                                if (index == null)
                                    index = linkIndex;

                                if (!EntityOps.IsWriteOnly(variable))
                                {
                                    ObjectDictionary arrayValue = variable.ArrayValue;

                                    if ((arrayValue != null) && arrayValue.ContainsKey(index))
                                    {
                                        object oldValue = EntityOps.GetOldValue(
                                            flags, variable, index, RuntimeOps.GetDefaultValue(
                                                BreakpointType.BeforeVariableGet));

                                        ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                            null, BreakpointType.BeforeVariableGet, savedFrame,
                                            variable, name, index, flags, oldValue, null, null,
                                            null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                            CanHitWatches() && EntityOps.IsBreakOnGet(variable))
                                        {
                                            //
                                            // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                            //       get access for this variable.
                                            //
                                            if (CheckWatches(
                                                    ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                    traceInfo, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
#endif

                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                            !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                        {
                                            try
                                            {
                                                if (RuntimeOps.FireTraces(
                                                        variable, traceInfo.BreakpointType, this, traceInfo,
                                                        ref value, ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                error = String.Format(
                                                    "caught exception while firing variable traces: {0}",
                                                    e);

                                                return ReturnCode.Error;
                                            }
                                        }

#if NOTIFY
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                            !EntityOps.IsNoNotify(variable))
                                        {
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Variable, NotifyFlags.Trace,
                                                new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                this, null, null, null, ref error);
                                        }
#endif

                                        //
                                        // NOTE: If the trace(s) did not cancel the operation, proceed.
                                        //
                                        if (!traceInfo.Cancel)
                                        {
                                            //
                                            // NOTE: Check for post-processing of the variable value.
                                            //
                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonPostProcess, false) &&
                                                traceInfo.PostProcess &&
                                                !FlagOps.HasFlags(interpreterFlags, InterpreterFlags.NoPostProcess, true))
                                            {
                                                if (PostProcessVariableValue(
                                                        traceInfo.Flags, name, traceInfo.Index, variable,
                                                        traceInfo.OldValue, ref value, ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            else
                                            {
                                                value = RuntimeOps.GetStringFromObject(traceInfo.OldValue);
                                            }
                                        }

                                        return traceInfo.ReturnCode;
                                    }
                                    else
                                    {
                                        ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                            null, BreakpointType.BeforeVariableGet, savedFrame,
                                            variable, name, index, flags, null, null, null,
                                            null, null, false, true, ReturnCode.Error);

#if DEBUGGER
                                        if (CanHitWatches() &&
                                            EntityOps.IsBreakOnGet(variable))
                                        {
                                            //
                                            // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                            //       get access for this variable.
                                            //
                                            if (CheckWatches(
                                                    ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                    traceInfo, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
#endif

                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                            !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                        {
                                            try
                                            {
                                                if (RuntimeOps.FireTraces(
                                                        variable, traceInfo.BreakpointType, this, traceInfo,
                                                        ref value, ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                error = String.Format(
                                                    "caught exception while firing variable traces: {0}",
                                                    e);

                                                return ReturnCode.Error;
                                            }
                                        }

#if NOTIFY
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                            !EntityOps.IsNoNotify(variable))
                                        {
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Variable, NotifyFlags.Trace,
                                                new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                this, null, null, null, ref error);
                                        }
#endif

                                        //
                                        // NOTE: If the trace(s) did not cancel the operation, proceed
                                        //       with producing an appropriate error message.  Otherwise,
                                        //       if a trace callback canceled the operation with a
                                        //       successful return code (i.e. not the original error
                                        //       return code) then post-process the variable value (if
                                        //       required) and return.
                                        //
                                        if (traceInfo.Cancel)
                                        {
                                            if (traceInfo.ReturnCode == ReturnCode.Ok)
                                            {
                                                //
                                                // NOTE: Check for post-processing of the variable value.
                                                //
                                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonPostProcess, false) &&
                                                    traceInfo.PostProcess &&
                                                    !FlagOps.HasFlags(interpreterFlags, InterpreterFlags.NoPostProcess, true))
                                                {
                                                    if (PostProcessVariableValue(
                                                            traceInfo.Flags, name, traceInfo.Index, variable,
                                                            traceInfo.OldValue, ref value, ref error) != ReturnCode.Ok)
                                                    {
                                                        return ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    value = RuntimeOps.GetStringFromObject(traceInfo.OldValue);
                                                }
                                            }
                                        }
                                        else
                                        {
                                            error = String.Format(
                                                "can't read {0}: no such element in array",
                                                FormatOps.ErrorVariableName(
                                                    variable, linkIndex, name, traceInfo.Index));
                                        }

                                        return traceInfo.ReturnCode;
                                    }
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't read {0}: variable is write-only",
                                        FormatOps.ErrorVariableName(
                                            variable, linkIndex, name, index));
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't read {0}: variable isn't array",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, index));
                            }
                        }
                        else
                        {
                            if (!EntityOps.IsArray2(variable))
                            {
                                if (!EntityOps.IsWriteOnly(variable))
                                {
                                    object oldValue = (variable.Value != null) ?
                                        variable.Value : RuntimeOps.GetDefaultValue(
                                            BreakpointType.BeforeVariableGet);

                                    ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                        null, BreakpointType.BeforeVariableGet, savedFrame,
                                        variable, name, index, flags, oldValue, null, null,
                                        null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                        CanHitWatches() && EntityOps.IsBreakOnGet(variable))
                                    {
                                        //
                                        // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                        //       get access for this variable.
                                        //
                                        if (CheckWatches(
                                                ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                traceInfo, ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
#endif

                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                        !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                    {
                                        try
                                        {
                                            if (RuntimeOps.FireTraces(
                                                    variable, traceInfo.BreakpointType, this, traceInfo,
                                                    ref value, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            error = String.Format(
                                                "caught exception while firing variable traces: {0}",
                                                e);

                                            return ReturnCode.Error;
                                        }
                                    }

#if NOTIFY
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                        !EntityOps.IsNoNotify(variable))
                                    {
                                        /* IGNORED */
                                        CheckNotifications(
                                            NotifyType.Variable, NotifyFlags.Trace,
                                            new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                            this, null, null, null, ref error);
                                    }
#endif

                                    //
                                    // NOTE: If the trace(s) did not cancel the operation, proceed.
                                    //
                                    if (!traceInfo.Cancel)
                                    {
                                        //
                                        // NOTE: Check for post-processing of the variable value.
                                        //
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonPostProcess, false) &&
                                            traceInfo.PostProcess &&
                                            !FlagOps.HasFlags(interpreterFlags, InterpreterFlags.NoPostProcess, true))
                                        {
                                            if (PostProcessVariableValue(
                                                    traceInfo.Flags, name, traceInfo.Index, variable,
                                                    traceInfo.OldValue, ref value, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
                                        else
                                        {
                                            value = RuntimeOps.GetStringFromObject(traceInfo.OldValue);
                                        }
                                    }

                                    return traceInfo.ReturnCode;
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't read {0}: variable is write-only",
                                        FormatOps.ErrorVariableName(name, null));
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't read {0}: variable is array",
                                    FormatOps.ErrorVariableName(name, null));
                            }
                        }
                    }
                    else
                    {
                        if (!EntityOps.IsWriteOnly(variable))
                        {
                            ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                null, BreakpointType.BeforeVariableGet, savedFrame,
                                variable, name, index, flags, null, null, null,
                                null, null, false, true, ReturnCode.Error);

#if DEBUGGER
                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                CanHitWatches() && EntityOps.IsBreakOnGet(variable))
                            {
                                //
                                // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                //       get access for this variable.
                                //
                                if (CheckWatches(
                                        ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                        traceInfo, ref error) != ReturnCode.Ok)
                                {
                                    return ReturnCode.Error;
                                }
                            }
#endif

                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                            {
                                try
                                {
                                    if (RuntimeOps.FireTraces(
                                            variable, traceInfo.BreakpointType, this, traceInfo,
                                            ref value, ref error) != ReturnCode.Ok)
                                    {
                                        return ReturnCode.Error;
                                    }
                                }
                                catch (Exception e)
                                {
                                    error = String.Format(
                                        "caught exception while firing variable traces: {0}",
                                        e);

                                    return ReturnCode.Error;
                                }
                            }

#if NOTIFY
                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                !EntityOps.IsNoNotify(variable))
                            {
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Variable, NotifyFlags.Trace,
                                    new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                    this, null, null, null, ref error);
                            }
#endif

                            //
                            // NOTE: If the trace(s) did not cancel the operation, proceed
                            //       with producing an appropriate error message.  Otherwise,
                            //       if a trace callback canceled the operation with a
                            //       successful return code (i.e. not the original error
                            //       return code) then post-process the variable value (if
                            //       required) and return.
                            //
                            if (traceInfo.Cancel)
                            {
                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Check for post-processing of the variable value.
                                    //
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonPostProcess, false) &&
                                        traceInfo.PostProcess &&
                                        !FlagOps.HasFlags(interpreterFlags, InterpreterFlags.NoPostProcess, true))
                                    {
                                        if (PostProcessVariableValue(
                                                traceInfo.Flags, name, traceInfo.Index, variable,
                                                traceInfo.OldValue, ref value, ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    else
                                    {
                                        value = RuntimeOps.GetStringFromObject(traceInfo.OldValue);
                                    }
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't read {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, traceInfo.Index));
                            }

                            return traceInfo.ReturnCode;
                        }
                        else
                        {
                            error = String.Format(
                                "can't read {0}: variable is write-only",
                                FormatOps.ErrorVariableName(name, null));
                        }
                    }
                }
                else if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                {
                    error = String.Format(
                        "can't read {0}: no such variable",
                        FormatOps.ErrorVariableName(
                            variable, null, name, index));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddVariable2(
            VariableFlags flags,
            string name,
            TraceList traces,
            bool strict,
            ref IVariable variable,
            ref Result error
            )
        {
            return AddVariable2(flags, null, name, traces, strict, ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode AddVariable2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            TraceList traces,
            bool strict,
            ref IVariable variable,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    bool badRequest = false;
                    Result localError = null;

                    if (DoesVariableExist(flags, frame, name, ref badRequest, ref localError) != ReturnCode.Ok)
                    {
                        if (!badRequest)
                        {
                            ICallFrame variableFrame = GetVariableFrame(frame, flags);

                            //
                            // TODO: Modify this when full [namespace] support has been added.
                            //
                            string newName = StringOps.TrimLeadingColons(name);

                            variable = new Variable(variableFrame, newName,
                                (flags & ~VariableFlags.NonAdd),
                                GetTraces(traces), variableEvent);

                            if (IsGlobalCallFrame(variableFrame))
                                EntityOps.SetGlobal(variable, true);
                            else
                                EntityOps.SetLocal(variable, true);

                            ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                null, BreakpointType.BeforeVariableAdd, variableFrame,
                                variable, name, null, flags, null, null, null, null,
                                null, false, true, ReturnCode.Ok);

#if DEBUGGER
                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                            {
                                //
                                // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                //       set access for this variable.
                                //
                                if (CheckWatches(
                                        ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                        traceInfo, ref error) != ReturnCode.Ok)
                                {
                                    return ReturnCode.Error;
                                }
                            }
#endif

                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                            {
                                try
                                {
                                    if (RuntimeOps.FireTraces(
                                            variable, traceInfo.BreakpointType, this, traceInfo,
                                            ref error) != ReturnCode.Ok)
                                    {
                                        return ReturnCode.Error;
                                    }
                                }
                                catch (Exception e)
                                {
                                    error = String.Format(
                                        "caught exception while firing variable traces: {0}",
                                        e);

                                    return ReturnCode.Error;
                                }
                            }

#if NOTIFY
                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                !EntityOps.IsNoNotify(variable))
                            {
                                /* IGNORED */
                                CheckNotifications(
                                    NotifyType.Variable, NotifyFlags.Trace,
                                    new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                    this, null, null, null, ref error);
                            }
#endif

                            //
                            // NOTE: If the trace(s) did not cancel the operation, proceed.
                            //
                            if (!traceInfo.Cancel)
                            {
                                //
                                // NOTE: All arrays (even virtual ones) must have a valid
                                //       dictionary for their elements.
                                //
                                if (EntityOps.IsArray2(variable))
                                {
                                    variable.Value = null;
                                    variable.ArrayValue = new ObjectDictionary();
                                }

                                //
                                // BUGFIX: If the variable name exists in the call frame at
                                //         this point it is most likely a zombie (i.e.
                                //         undefined); therefore, replace it if necessary.
                                //
                                VariableDictionary variables = variableFrame.Variables;

                                if (variables.ContainsKey(newName))
                                    //
                                    // NOTE: Replace variable in current call frame.
                                    //
                                    variables[newName] = variable;
                                else
                                    //
                                    // NOTE: Add variable in current call frame.
                                    //
                                    variables.Add(newName, variable);

                                //
                                // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                //         modifications have been completed.
                                //
                                EntityOps.SetDirty(variable, true);
                            }

                            return traceInfo.ReturnCode;
                        }
                        else
                        {
                            error = localError;
                        }
                    }
                    else if (strict)
                    {
                        error = String.Format(
                            "can't add {0}: variable already exists",
                            FormatOps.ErrorVariableName(name, null));
                    }
                    else
                    {
                        //
                        // NOTE: Variable already exists (non-strict) mode.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    error = String.Format(
                        "can't add {0}: interpreter is read-only",
                        FormatOps.ErrorVariableName(name, null));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode ResetVariable2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            IVariable variable,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    //
                    // BUGBUG: This may need revision; we do need some way to remove
                    //         "stale" undefined variables from the call frame even
                    //         though the only way to officially create one is via the
                    //         global or upvar commands and those get cleaned up
                    //         automatically upon procedure exit.
                    //
                    bool haveVariable = (variable != null);
                    Result localError = null;

                    if ((haveVariable ||
                        (GetVariableViaResolversWithSplit(frame, name, index, ref flags,
                            ref variable, ref localError) == ReturnCode.Ok)) &&
                        !EntityOps.IsUndefined(variable, "reset", name, index,
                            ref localError))
                    {
                        //
                        // COMPAT: Tcl, if the index was specified, make sure the name
                        //         itself does not contain an array element reference.
                        //
                        if (haveVariable || (index == null))
                        {
                            VariableFlags splitFlags = flags;
                            string splitIndex = null;

                            if (index != null)
                                splitFlags |= VariableFlags.NoElement;

                            if (RuntimeOps.SplitVariableName(
                                    this, splitFlags, name, ref name,
                                    ref splitIndex, ref error) == ReturnCode.Ok)
                            {
                                if ((index == null) && (splitIndex != null))
                                    index = splitIndex;
                            }
                            else
                            {
                                return ReturnCode.Error;
                            }
                        }

                        //
                        // NOTE: Make sure we have a variable.
                        //
                        if (variable == null)
                        {
                            error = String.Format(
                                "can't reset {0}: variable is invalid",
                                FormatOps.ErrorVariableName(name, index));

                            return ReturnCode.Error;
                        }

                        //
                        // NOTE: For linked variables, we may be linking to an array
                        //       element.
                        //
                        string linkIndex = null;

                        //
                        // NOTE: Save original call frame for traces.
                        //
                        ICallFrame savedFrame = variable.Frame;

                        //
                        // NOTE: Is the variable really a link to another variable?
                        //
                        if (EntityOps.IsLink(variable))
                        {
                            //
                            // NOTE: Get the array element index of the link, if any.
                            //
                            linkIndex = variable.LinkIndex;

                            //
                            // NOTE: Follow the linked variable.
                            //
                            variable = EntityOps.FollowLinks(variable);
                        }

                        if ((index != null) || (linkIndex != null))
                        {
                            if (EntityOps.IsArray2(variable) &&
                                    ((index == null) || (linkIndex == null)))
                            {
                                //
                                // NOTE: If there was no element index specified, use
                                //       the one from the variable link.
                                //
                                if (index == null)
                                    index = linkIndex;

                                if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                                {
                                    ObjectDictionary arrayValue = variable.ArrayValue;

                                    if ((arrayValue != null) && arrayValue.ContainsKey(index))
                                    {
                                        object oldValue = EntityOps.GetOldValue(
                                            flags, variable, index, RuntimeOps.GetDefaultValue(
                                                BreakpointType.BeforeVariableSet));

                                        object newValue = EntityOps.GetNewValue(
                                            flags, variable.Name, index, oldValue, null);

                                        ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                            null, BreakpointType.BeforeVariableReset, savedFrame,
                                            variable, name, index, flags, oldValue, newValue, null,
                                            null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                            CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                                        {
                                            //
                                            // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                            //       set access for this variable.
                                            //
                                            if (CheckWatches(
                                                    ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                    traceInfo, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
#endif

                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                            !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                        {
                                            try
                                            {
                                                if (RuntimeOps.FireTraces(
                                                        variable, traceInfo.BreakpointType, this, traceInfo,
                                                        ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                error = String.Format(
                                                    "caught exception while firing variable traces: {0}",
                                                    e);

                                                return ReturnCode.Error;
                                            }
                                        }

#if NOTIFY
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                            !EntityOps.IsNoNotify(variable))
                                        {
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Variable, NotifyFlags.Trace,
                                                new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                this, null, null, null, ref error);
                                        }
#endif

                                        //
                                        // NOTE: If the trace(s) did not cancel the operation, proceed.
                                        //
                                        if (!traceInfo.Cancel)
                                        {
                                            arrayValue = variable.ArrayValue; /* REFRESH */

                                            if (arrayValue != null)
                                            {
                                                if (arrayValue.ContainsKey(index))
                                                    arrayValue[index] = traceInfo.NewValue;
                                                else
                                                    arrayValue.Add(index, traceInfo.NewValue);

                                                //
                                                // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                                //         modifications have been completed.
                                                //
                                                EntityOps.SetDirty(variable, true);
                                            }
                                            else
                                            {
                                                error = String.Format(
                                                    "can't reset {0}: no longer an array",
                                                    FormatOps.ErrorVariableName(
                                                        variable, linkIndex, name, index));

                                                return ReturnCode.Error;
                                            }
                                        }

                                        return traceInfo.ReturnCode;
                                    }
                                    else
                                    {
                                        ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                            null, BreakpointType.BeforeVariableReset, savedFrame,
                                            variable, name, index, flags, null, null, null, null,
                                            null, false, true, ReturnCode.Error);

#if DEBUGGER
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                            CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                                        {
                                            //
                                            // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                            //       unset access for this variable.
                                            //
                                            if (CheckWatches(
                                                    ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                    traceInfo, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
#endif

                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                            !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                        {
                                            try
                                            {
                                                if (RuntimeOps.FireTraces(
                                                        variable, traceInfo.BreakpointType, this, traceInfo,
                                                        ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                error = String.Format(
                                                    "caught exception while firing variable traces: {0}",
                                                    e);

                                                return ReturnCode.Error;
                                            }
                                        }

#if NOTIFY
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                            !EntityOps.IsNoNotify(variable))
                                        {
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Variable, NotifyFlags.Trace,
                                                new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                this, null, null, null, ref error);
                                        }
#endif

                                        //
                                        // NOTE: If the trace(s) did not cancel the operation, proceed.
                                        //
                                        if (!traceInfo.Cancel)
                                        {
                                            error = String.Format(
                                                "can't reset {0}: no such element in array",
                                                FormatOps.ErrorVariableName(
                                                    variable, linkIndex, name, index));
                                        }

                                        return traceInfo.ReturnCode;
                                    }
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't reset {0}: variable is read-only",
                                        FormatOps.ErrorVariableName(
                                            variable, linkIndex, name, index));
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't reset {0}: variable isn't array",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, index));
                            }
                        }
                        else
                        {
                            if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                            {
                                object oldValue = EntityOps.GetOldValue(
                                    flags, variable, index, RuntimeOps.GetDefaultValue(
                                        BreakpointType.BeforeVariableSet));

                                object newValue = EntityOps.GetNewValue(
                                    flags, variable.Name, index, oldValue, null);

                                ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                    null, BreakpointType.BeforeVariableReset, savedFrame,
                                    variable, name, index, flags, oldValue, newValue, null,
                                    null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                    CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                                {
                                    //
                                    // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                    //       set access for this variable.
                                    //
                                    if (CheckWatches(
                                            ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                            traceInfo, ref error) != ReturnCode.Ok)
                                    {
                                        return ReturnCode.Error;
                                    }
                                }
#endif

                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                    !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                {
                                    try
                                    {
                                        if (RuntimeOps.FireTraces(
                                                variable, traceInfo.BreakpointType, this, traceInfo,
                                                ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    catch (Exception e)
                                    {
                                        error = String.Format(
                                            "caught exception while firing variable traces: {0}",
                                            e);

                                        return ReturnCode.Error;
                                    }
                                }

#if NOTIFY
                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                    !EntityOps.IsNoNotify(variable))
                                {
                                    /* IGNORED */
                                    CheckNotifications(
                                        NotifyType.Variable, NotifyFlags.Trace,
                                        new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                        this, null, null, null, ref error);
                                }
#endif

                                //
                                // NOTE: If the trace(s) did not cancel the operation, proceed.
                                //
                                if (!traceInfo.Cancel)
                                {
                                    //
                                    // NOTE: Did the trace(s) change the variable into an array?
                                    //
                                    if (EntityOps.IsArray2(variable))
                                    {
                                        //
                                        // NOTE: Allocate space for the array name/values now.
                                        //
                                        variable.Value = null;
                                        variable.ArrayValue = new ObjectDictionary();
                                    }
                                    //
                                    // NOTE: If the trace(s) changed the variable to be a link, we do
                                    //       not bother doing anything else to it (i.e. we assume it
                                    //       points where they want it to point).
                                    //
                                    else if (!EntityOps.IsLink(variable))
                                    {
                                        //
                                        // NOTE: Set the value to the output of the trace(s).
                                        //
                                        variable.Value = traceInfo.NewValue;
                                    }

                                    //
                                    // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                    //         modifications have been completed.
                                    //
                                    EntityOps.SetDirty(variable, true);
                                }

                                return traceInfo.ReturnCode;
                            }
                            else
                            {
                                error = String.Format(
                                    "can't reset {0}: variable is read-only",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, null));
                            }
                        }
                    }
                    else
                    {
                        if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                        {
                            if (FlagOps.HasFlags(flags, VariableFlags.NoComplain, true))
                            {
                                return ReturnCode.Ok;
                            }
                            else
                            {
                                error = String.Format(
                                    "can't reset {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, null, name, index));
                            }
                        }
                        else
                        {
                            if (FlagOps.HasFlags(flags, VariableFlags.NoComplain, true))
                            {
                                return ReturnCode.Ok;
                            }
                            else
                            {
                                //
                                // NOTE: We did not actually search for the variable,
                                //       inform the caller as to why.
                                //
                                error = localError;
                            }
                        }
                    }
                }
                else
                {
                    error = String.Format(
                        "can't reset {0}: interpreter is read-only",
                        FormatOps.ErrorVariableName(name, index));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetVariableLink2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            MemberInfo memberInfo,
            object @object,
            ref IVariable variable,
            ref Result error
            )
        {
            // CheckDisposed();

            IAnyPair<MemberInfo, object> pair =
                new AnyPair<MemberInfo, object>(memberInfo, @object);

            if (SetVariableValue2(
                    flags, frame, name, index, pair, linkedVariableTraceList,
                    ref variable, ref error) == ReturnCode.Ok)
            {
                if (HasLinkedVariableTraceCallback(variable))
                {
                    return ReturnCode.Ok;
                }
                else
                {
                    error = String.Format(
                        "can't link {0}: could not set variable trace",
                        FormatOps.ErrorVariableName(
                            variable, null, name, index));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetVariableValue(
            VariableFlags flags,
            string name,
            string value,
            TraceList traces
            )
        {
            Result error = null;

            return SetVariableValue(
                flags, name, value, traces, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetVariableValue(
            VariableFlags flags,
            string name,
            string value,
            TraceList traces,
            ref IVariable variable,
            ref Result error
            )
        {
            string varName = null;
            string varIndex = null;

            if (RuntimeOps.SplitVariableName(
                    this, flags, name, ref varName,
                    ref varIndex, ref error) == ReturnCode.Ok)
            {
                return SetVariableValue2(
                    flags, varName, varIndex, value,
                    traces, ref variable, ref error);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetVariableValue2(
            VariableFlags flags,
            string name,
            string index,
            string value,
            TraceList traces
            )
        {
            IVariable variable = null;
            Result error = null;

            return SetVariableValue2(
                flags, name, index, value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetVariableValue2(
            VariableFlags flags,
            string name,
            string index,
            string value,
            TraceList traces,
            ref Result error
            )
        {
            IVariable variable = null;

            return SetVariableValue2(
                flags, name, index, value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetVariableValue2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            string value,
            TraceList traces,
            ref Result error
            )
        {
            IVariable variable = null;

            return SetVariableValue2(
                flags, frame, name, index, value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetVariableValue2(
            VariableFlags flags,
            string name,
            string index,
            string value,
            TraceList traces,
            ref IVariable variable,
            ref Result error
            )
        {
            return SetVariableValue2(
                flags, null, name, index, value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetVariableValue2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            string value,
            TraceList traces,
            ref IVariable variable,
            ref Result error
            )
        {
            return SetVariableValue2(
                flags, frame, name, index, (object)value, traces,
                ref variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetVariableValue2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            object value,
            TraceList traces,
            ref IVariable variable,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    bool haveVariable = (variable != null);
                    Result localError = null;

                    if (haveVariable ||
                        (GetVariableViaResolversWithSplit(frame, name, index, ref flags,
                            ref variable, ref localError) == ReturnCode.Ok))
                    {
                        //
                        // COMPAT: Tcl, if the index was specified, make sure the name
                        //         itself does not contain an array element reference.
                        //
                        if (haveVariable || (index == null))
                        {
                            VariableFlags splitFlags = flags;
                            string splitIndex = null;

                            if (index != null)
                                splitFlags |= VariableFlags.NoElement;

                            if (RuntimeOps.SplitVariableName(
                                    this, splitFlags, name, ref name,
                                    ref splitIndex, ref error) == ReturnCode.Ok)
                            {
                                if ((index == null) && (splitIndex != null))
                                    index = splitIndex;
                            }
                            else
                            {
                                return ReturnCode.Error;
                            }
                        }

                        //
                        // NOTE: Make sure we have a variable.
                        //
                        if (variable == null)
                        {
                            error = String.Format(
                                "can't set {0}: variable is invalid",
                                FormatOps.ErrorVariableName(name, index));

                            return ReturnCode.Error;
                        }

                        //
                        // NOTE: For linked variables, we may be linking to an array
                        //       element.
                        //
                        string linkIndex = null;

                        //
                        // NOTE: Save original call frame for traces.
                        //
                        ICallFrame savedFrame = variable.Frame;

                        //
                        // NOTE: Is the variable really a link to another variable?
                        //
                        // BUGFIX: Also, does the variable at the end of the link exist
                        //         in a defined call frame?
                        //
                        if (EntityOps.HasValidLink(variable, false))
                        {
                            //
                            // NOTE: Get the array element index of the link, if any.
                            //
                            linkIndex = variable.LinkIndex;

                            //
                            // NOTE: Follow the linked variable.
                            //
                            variable = EntityOps.FollowLinks(variable);
                        }

                        if ((index != null) || (linkIndex != null))
                        {
                            //
                            // BUGFIX: (PART 1) If the variable was previously undefined, we
                            //         cannot validate it as an array variable.
                            //
                            // BUGFIX: If the variable is a link at this point the call frame
                            //         for the variable at the link destination is invalid;
                            //         therefore, allow the variable to be re-used as an array.
                            //
                            if ((EntityOps.IsUndefined(variable) || EntityOps.IsLink(variable) ||
                                    EntityOps.IsArray2(variable)) &&
                                ((index == null) || (linkIndex == null)))
                            {
                                //
                                // NOTE: If there was no element index specified, use
                                //       the one from the variable link.
                                //
                                if (index == null)
                                    index = linkIndex;

                                if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                                {
                                    object oldValue = EntityOps.GetOldValue(
                                        flags, variable, index, RuntimeOps.GetDefaultValue(
                                            BreakpointType.BeforeVariableSet));

                                    object newValue = EntityOps.GetNewValue(
                                        flags, variable.Name, index, oldValue, value);

                                    ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                        null, BreakpointType.BeforeVariableSet, savedFrame,
                                        variable, name, index, flags, oldValue, newValue,
                                        null, null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                        CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                                    {
                                        //
                                        // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                        //       set access for this variable.
                                        //
                                        if (CheckWatches(
                                                ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                traceInfo, ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
#endif

                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                        !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                    {
                                        try
                                        {
                                            if (RuntimeOps.FireTraces(
                                                    variable, traceInfo.BreakpointType, this, traceInfo,
                                                    ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            error = String.Format(
                                                "caught exception while firing variable traces: {0}",
                                                e);

                                            return ReturnCode.Error;
                                        }
                                    }

#if NOTIFY
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                        !EntityOps.IsNoNotify(variable))
                                    {
                                        /* IGNORED */
                                        CheckNotifications(
                                            NotifyType.Variable, NotifyFlags.Trace,
                                            new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                            this, null, null, null, ref error);
                                    }
#endif

                                    //
                                    // NOTE: If the trace(s) did not cancel the operation, proceed.
                                    //
                                    if (!traceInfo.Cancel)
                                    {
                                        if (traceInfo.Index != null)
                                        {
                                            //
                                            // BUGFIX: The variable may have been a link to a variable
                                            //         in an undefined call frame.
                                            //
                                            if (EntityOps.IsLink(variable))
                                            {
                                                //
                                                // NOTE: First, unflag this variable as a link and unlink
                                                //       it.
                                                //
                                                EntityOps.SetLink(variable, false);
                                                variable.Link = null;

                                                //
                                                // NOTE: Next, reset the flags to the ones provided by the
                                                //       caller, masking off any invalid ones.
                                                //
                                                variable.Flags = (traceInfo.Flags & ~VariableFlags.NonSet);

                                                //
                                                // NOTE: Next, flag the variable as global or local,
                                                //       depending on the call frame it resides in.
                                                //
                                                ICallFrame variableFrame = GetVariableFrame(frame, traceInfo.Flags);

                                                if (IsGlobalCallFrame(variableFrame))
                                                    EntityOps.SetGlobal(variable, true);
                                                else
                                                    EntityOps.SetLocal(variable, true);

                                                //
                                                // NOTE: Next, reset the traces for this variable to the
                                                //       same ones that would be assigned to brand new
                                                //       variables.
                                                //
                                                variable.Traces = GetTraces(traces);
                                            }

                                            //
                                            // BUGFIX: (PART 2) We need to tag the variable as an array and
                                            //         create the underlying dictionary now if the variable
                                            //         was previously undefined.  Also, make sure the array
                                            //         value dictionary is created (or recreated) before
                                            //         continuing.
                                            //
                                            EntityOps.SetArray(variable, true);
                                            ObjectDictionary arrayValue = variable.ArrayValue;

                                            if (EntityOps.IsUndefined(variable) || (arrayValue == null))
                                            {
                                                variable.Value = null;
                                                variable.ArrayValue = arrayValue = new ObjectDictionary();
                                            }

                                            if (arrayValue.ContainsKey(traceInfo.Index))
                                                arrayValue[traceInfo.Index] = traceInfo.NewValue;
                                            else
                                                arrayValue.Add(traceInfo.Index, traceInfo.NewValue);

                                            //
                                            // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                            //         modifications have been completed.
                                            //
                                            EntityOps.SetUndefined(variable, false);
                                            EntityOps.SetDirty(variable, true);
                                        }
                                        else
                                        {
                                            error = String.Format(
                                                "can't set {0}: variable is array",
                                                FormatOps.ErrorVariableName(name, null));

                                            return ReturnCode.Error;
                                        }
                                    }

                                    return traceInfo.ReturnCode;
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't set {0}: variable is read-only",
                                        FormatOps.ErrorVariableName(
                                            variable, linkIndex, name, index));
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't set {0}: variable isn't array",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, index));
                            }
                        }
                        else
                        {
                            //
                            // BUGFIX: (PART 3) If the variable was previously undefined,
                            //         we cannot validate it as a scalar variable.
                            //
                            if (EntityOps.IsUndefined(variable) || !EntityOps.IsArray2(variable))
                            {
                                if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                                {
                                    object oldValue = (variable.Value != null) ?
                                        variable.Value : RuntimeOps.GetDefaultValue(
                                            BreakpointType.BeforeVariableSet);

                                    object newValue = EntityOps.GetNewValue(
                                        flags, variable.Name, index, oldValue, value);

                                    ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                        null, BreakpointType.BeforeVariableSet, savedFrame,
                                        variable, name, index, flags, oldValue, newValue,
                                        null, null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                        CanHitWatches() && EntityOps.IsBreakOnSet(variable))
                                    {
                                        //
                                        // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                        //       set access for this variable.
                                        //
                                        if (CheckWatches(
                                                ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                traceInfo, ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
#endif

                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                        !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                    {
                                        try
                                        {
                                            if (RuntimeOps.FireTraces(
                                                    variable, traceInfo.BreakpointType, this, traceInfo,
                                                    ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
                                        catch (Exception e)
                                        {
                                            error = String.Format(
                                                "caught exception while firing variable traces: {0}",
                                                e);

                                            return ReturnCode.Error;
                                        }
                                    }

#if NOTIFY
                                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                        !EntityOps.IsNoNotify(variable))
                                    {
                                        /* IGNORED */
                                        CheckNotifications(
                                            NotifyType.Variable, NotifyFlags.Trace,
                                            new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                            this, null, null, null, ref error);
                                    }
#endif

                                    //
                                    // NOTE: If the trace(s) did not cancel the operation, proceed.
                                    //
                                    if (!traceInfo.Cancel)
                                    {
                                        //
                                        // BUGFIX: The variable may have been a link to a variable
                                        //         in an undefined call frame.
                                        //
                                        if (EntityOps.IsLink(variable))
                                        {
                                            //
                                            // NOTE: First, unflag this variable as a link and unlink
                                            //       it.
                                            //
                                            EntityOps.SetLink(variable, false);
                                            variable.Link = null;

                                            //
                                            // NOTE: Next, reset the flags to the ones provided by the
                                            //       caller, masking off any invalid ones.
                                            //
                                            variable.Flags = (traceInfo.Flags & ~VariableFlags.NonSet);

                                            //
                                            // NOTE: Next, flag the variable as global or local,
                                            //       depending on the call frame it resides in.
                                            //
                                            ICallFrame variableFrame = GetVariableFrame(frame, traceInfo.Flags);

                                            if (IsGlobalCallFrame(variableFrame))
                                                EntityOps.SetGlobal(variable, true);
                                            else
                                                EntityOps.SetLocal(variable, true);

                                            //
                                            // NOTE: Next, reset the traces for this variable to the
                                            //       same ones that would be assigned to brand new
                                            //       variables.
                                            //
                                            variable.Traces = GetTraces(traces);
                                        }

                                        //
                                        // BUGFIX: (PART 4) We need to tag the variable as a scalar and
                                        //         remove the underlying dictionary now if the variable
                                        //         was previously undefined or if array value dictionary
                                        //         is actually present.
                                        //
                                        EntityOps.SetArray(variable, false);
                                        ObjectDictionary arrayValue = variable.ArrayValue;

                                        if (EntityOps.IsUndefined(variable) || (arrayValue != null))
                                            variable.ArrayValue = arrayValue = null;

                                        variable.Value = traceInfo.NewValue;

                                        //
                                        // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                        //         modifications have been completed.
                                        //
                                        EntityOps.SetUndefined(variable, false);
                                        EntityOps.SetDirty(variable, true);
                                    }

                                    return traceInfo.ReturnCode;
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't set {0}: variable is read-only",
                                        FormatOps.ErrorVariableName(name, null));
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "can't set {0}: variable is array",
                                    FormatOps.ErrorVariableName(name, null));
                            }
                        }
                    }
                    else
                    {
                        if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                        {
                            //
                            // NOTE: Make sure we are allowed to add new variables (i.e.
                            //       and not just change existing ones).
                            //
                            if (!FlagOps.HasFlags(flags, VariableFlags.NoCreate, true))
                            {
                                ICallFrame variableFrame = GetVariableFrame(frame, flags);

                                //
                                // NOTE: Remove all leading namespace separator characters.
                                //
                                // TODO: Modify this when full [namespace] support has been added.
                                //
                                string newName = StringOps.TrimLeadingColons(name);

                                variable = new Variable(variableFrame, newName,
                                    (flags & ~VariableFlags.NonSet | VariableFlags.Undefined),
                                    GetTraces(traces), variableEvent);

                                if (IsGlobalCallFrame(variableFrame))
                                    EntityOps.SetGlobal(variable, true);
                                else
                                    EntityOps.SetLocal(variable, true);

                                object newValue = EntityOps.GetNewValue(
                                    flags, variable.Name, index, null, value);

                                ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                    null, BreakpointType.BeforeVariableSet, variableFrame,
                                    variable, name, index, flags, null, newValue, null,
                                    null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                if (CanHitWatches() &&
                                    EntityOps.IsBreakOnSet(variable))
                                {
                                    //
                                    // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                    //       set access for this variable.
                                    //
                                    if (CheckWatches(
                                            ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                            traceInfo, ref error) != ReturnCode.Ok)
                                    {
                                        return ReturnCode.Error;
                                    }
                                }
#endif

                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                    !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                {
                                    try
                                    {
                                        if (RuntimeOps.FireTraces(
                                                variable, traceInfo.BreakpointType, this, traceInfo,
                                                ref error) != ReturnCode.Ok)
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    catch (Exception e)
                                    {
                                        error = String.Format(
                                            "caught exception while firing variable traces: {0}",
                                            e);

                                        return ReturnCode.Error;
                                    }
                                }

#if NOTIFY
                                if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                    !EntityOps.IsNoNotify(variable))
                                {
                                    /* IGNORED */
                                    CheckNotifications(
                                        NotifyType.Variable, NotifyFlags.Trace,
                                        new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                        this, null, null, null, ref error);
                                }
#endif

                                //
                                // NOTE: If the trace(s) did not cancel the operation, proceed.
                                //
                                if (!traceInfo.Cancel)
                                {
                                    if ((traceInfo.Index != null) ||
                                        EntityOps.IsArray2(variable))
                                    {
                                        EntityOps.SetArray(variable, true);
                                        ObjectDictionary arrayValue = new ObjectDictionary();

                                        variable.Value = null;
                                        variable.ArrayValue = arrayValue;

                                        if (traceInfo.Index != null)
                                            arrayValue.Add(traceInfo.Index, traceInfo.NewValue);
                                    }
                                    else
                                    {
                                        variable.Value = traceInfo.NewValue;
                                    }

                                    variableFrame.Variables.Add(newName, variable);

                                    //
                                    // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                    //         modifications have been completed.
                                    //
                                    EntityOps.SetUndefined(variable, false);
                                    EntityOps.SetDirty(variable, true);
                                }

                                return traceInfo.ReturnCode;
                            }
                            else
                            {
                                error = String.Format(
                                    "can't set {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, null, name, index));
                            }
                        }
                        else
                        {
                            //
                            // NOTE: We did not actually search for the variable,
                            //       inform the caller as to why.
                            //
                            error = localError;
                        }
                    }
                }
                else
                {
                    error = String.Format(
                        "can't set {0}: interpreter is read-only",
                        FormatOps.ErrorVariableName(name, index));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode UnsetVariable2(
            VariableFlags flags,
            string name,
            string index,
            IVariable variable,
            ref Result error
            )
        {
            return UnsetVariable2(flags, null, name, index, variable, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode UnsetVariable2(
            VariableFlags flags,
            ICallFrame frame,
            string name,
            string index,
            IVariable variable,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (!readOnly)
                {
                    //
                    // BUGBUG: This may need revision; we do need some way to remove
                    //         "stale" undefined variables from the call frame even
                    //         though the only way to officially create one is via the
                    //         global or upvar commands and those get cleaned up
                    //         automatically upon procedure exit.
                    //
                    bool haveVariable = (variable != null);
                    Result localError = null;

                    if ((haveVariable ||
                        (GetVariableViaResolversWithSplit(frame, name, index, ref flags,
                            ref variable, ref localError) == ReturnCode.Ok)) &&
                        !EntityOps.IsUndefined(variable, "unset", name, index,
                            ref localError))
                    {
                        //
                        // COMPAT: Tcl, if the index was specified, make sure the name
                        //         itself does not contain an array element reference.
                        //
                        if (haveVariable || (index == null))
                        {
                            VariableFlags splitFlags = flags;
                            string splitIndex = null;

                            if (index != null)
                                splitFlags |= VariableFlags.NoElement;

                            if (RuntimeOps.SplitVariableName(
                                    this, splitFlags, name, ref name,
                                    ref splitIndex, ref error) == ReturnCode.Ok)
                            {
                                if ((index == null) && (splitIndex != null))
                                    index = splitIndex;
                            }
                            else
                            {
                                return ReturnCode.Error;
                            }
                        }

                        //
                        // NOTE: Make sure we have a variable.
                        //
                        if (variable == null)
                        {
                            error = String.Format(
                                "can't unset {0}: variable is invalid",
                                FormatOps.ErrorVariableName(name, index));

                            return ReturnCode.Error;
                        }

                        //
                        // NOTE: For linked variables, we may be linking to an array
                        //       element.
                        //
                        string linkIndex = null;

                        //
                        // NOTE: Save original call frame for traces.
                        //
                        ICallFrame savedFrame = variable.Frame;

                        //
                        // NOTE: For linked variables, save a reference to the original
                        //       variable link itself.
                        //
                        IVariable savedVariable = null;

                        //
                        // NOTE: Is the variable really a link to another variable?
                        //
                        // BUGFIX: Also, does the variable at the end of the link exist
                        //         in a defined call frame?
                        //
                        if (EntityOps.HasValidLink(variable, false))
                        {
                            //
                            // NOTE: Get the array element index of the link, if any.
                            //
                            linkIndex = variable.LinkIndex;

                            //
                            // NOTE: Save the link itself because we may need to mark it
                            //       as undefined (below).
                            //
                            savedVariable = variable; // TEST: Test this.

                            //
                            // NOTE: Follow the linked variable.
                            //
                            variable = EntityOps.FollowLinks(variable);
                        }

                        if (!EntityOps.IsUndefined(variable))
                        {
                            if ((index != null) || (linkIndex != null))
                            {
                                if (EntityOps.IsArray2(variable) &&
                                        ((index == null) || (linkIndex == null)))
                                {
                                    //
                                    // NOTE: If there was no element index specified, use
                                    //       the one from the variable link.
                                    //
                                    if (index == null)
                                        index = linkIndex;

                                    if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                                    {
                                        ObjectDictionary arrayValue = variable.ArrayValue;

                                        if ((arrayValue != null) && arrayValue.ContainsKey(index))
                                        {
                                            object oldValue = EntityOps.GetOldValue(
                                                flags, variable, index, RuntimeOps.GetDefaultValue(
                                                    BreakpointType.BeforeVariableUnset));

                                            ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                                null, BreakpointType.BeforeVariableUnset, savedFrame,
                                                variable, name, index, flags, oldValue, null, null,
                                                null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                                CanHitWatches() && EntityOps.IsBreakOnUnset(variable))
                                            {
                                                //
                                                // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                                //       unset access for this variable.
                                                //
                                                if (CheckWatches(
                                                        ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                        traceInfo, ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
#endif

                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                                !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                            {
                                                try
                                                {
                                                    if (RuntimeOps.FireTraces(
                                                            variable, traceInfo.BreakpointType, this, traceInfo,
                                                            ref error) != ReturnCode.Ok)
                                                    {
                                                        return ReturnCode.Error;
                                                    }
                                                }
                                                catch (Exception e)
                                                {
                                                    error = String.Format(
                                                        "caught exception while firing variable traces: {0}",
                                                        e);

                                                    return ReturnCode.Error;
                                                }
                                            }

#if NOTIFY
                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                                !EntityOps.IsNoNotify(variable))
                                            {
                                                /* IGNORED */
                                                CheckNotifications(
                                                    NotifyType.Variable, NotifyFlags.Trace,
                                                    new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                    this, null, null, null, ref error);
                                            }
#endif

                                            //
                                            // NOTE: If the trace(s) did not cancel the operation, proceed.
                                            //
                                            if (!traceInfo.Cancel)
                                            {
                                                if (traceInfo.Index != null)
                                                {
                                                    arrayValue = variable.ArrayValue; /* REFRESH */

                                                    if (arrayValue != null)
                                                    {
                                                        arrayValue.Remove(traceInfo.Index);

                                                        //
                                                        // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                                        //         modifications have been completed.
                                                        //
                                                        EntityOps.SetDirty(variable, true);
                                                    }
                                                    else
                                                    {
                                                        error = String.Format(
                                                            "can't unset {0}: no longer an array",
                                                            FormatOps.ErrorVariableName(
                                                                variable, linkIndex, name, index));

                                                        return ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if (savedVariable != null)
                                                    {
                                                        //
                                                        // NOTE: Links to dead variable(s) should be dead too.
                                                        //
                                                        EntityOps.SetUndefined(savedVariable, true); // TEST: Test this.
                                                    }

                                                    //
                                                    // NOTE: The variable is now dead.
                                                    //
                                                    EntityOps.SetUndefined(variable, true);

                                                    //
                                                    // NOTE: Remove any pending array searches that may exist for this
                                                    //       variable.
                                                    //
                                                    CleanupArraySearches(variable);

                                                    //
                                                    // NOTE: Check for the variable purge flag.  If the purge flag is set,
                                                    //       we remove all undefined variables from the current call frame.
                                                    //
                                                    if (FlagOps.HasFlags(traceInfo.Flags, VariableFlags.Purge, true))
                                                        //
                                                        // NOTE: Purge all undefined variables in the call frame for the
                                                        //       variable to be unset.  This implies removal of the variable
                                                        //       to be unset because we just set it to be undefined (above).
                                                        //
                                                        CallFrameOps.Cleanup(CurrentFrame, savedFrame, false);
                                                    else if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NoRemove, true))
                                                        //
                                                        // NOTE: Remove the entire array if there is one; otherwise,
                                                        //       just remove this scalar variable.
                                                        //
                                                        savedFrame.Variables.Remove(name);

                                                    //
                                                    // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                                    //         modifications have been completed.
                                                    //
                                                    EntityOps.SetDirty(variable, true);
                                                }
                                            }

                                            return traceInfo.ReturnCode;
                                        }
                                        else
                                        {
                                            ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                                null, BreakpointType.BeforeVariableUnset, savedFrame,
                                                variable, name, index, flags, null, null, null, null,
                                                null, false, true, ReturnCode.Error);

#if DEBUGGER
                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                                CanHitWatches() && EntityOps.IsBreakOnUnset(variable))
                                            {
                                                //
                                                // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                                //       unset access for this variable.
                                                //
                                                if (CheckWatches(
                                                        ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                        traceInfo, ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
#endif

                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                                !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                            {
                                                try
                                                {
                                                    if (RuntimeOps.FireTraces(
                                                            variable, traceInfo.BreakpointType, this, traceInfo,
                                                            ref error) != ReturnCode.Ok)
                                                    {
                                                        return ReturnCode.Error;
                                                    }
                                                }
                                                catch (Exception e)
                                                {
                                                    error = String.Format(
                                                        "caught exception while firing variable traces: {0}",
                                                        e);

                                                    return ReturnCode.Error;
                                                }
                                            }

#if NOTIFY
                                            if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                                !EntityOps.IsNoNotify(variable))
                                            {
                                                /* IGNORED */
                                                CheckNotifications(
                                                    NotifyType.Variable, NotifyFlags.Trace,
                                                    new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                    this, null, null, null, ref error);
                                            }
#endif

                                            //
                                            // NOTE: If the trace(s) did not cancel the operation, proceed.
                                            //
                                            if (!traceInfo.Cancel)
                                            {
                                                error = String.Format(
                                                    "can't unset {0}: no such element in array",
                                                    FormatOps.ErrorVariableName(
                                                        variable, linkIndex, name, index));
                                            }

                                            return traceInfo.ReturnCode;
                                        }
                                    }
                                    else
                                    {
                                        error = String.Format(
                                            "can't unset {0}: variable is read-only",
                                            FormatOps.ErrorVariableName(
                                                variable, linkIndex, name, index));
                                    }
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't unset {0}: variable isn't array",
                                        FormatOps.ErrorVariableName(
                                            variable, linkIndex, name, index));
                                }
                            }
                            else
                            {
                                if (!EntityOps.IsReadOnly(variable) && !EntityOps.IsSystem(variable))
                                {
                                    ICallFrame variableFrame = GetVariableFrame(frame, flags);

                                    if ((variableFrame != null) && (variableFrame.Variables != null))
                                    {
                                        object oldValue = EntityOps.GetOldValue(
                                            flags, variable, index, RuntimeOps.GetDefaultValue(
                                                BreakpointType.BeforeVariableUnset));

                                        ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(this,
                                            null, BreakpointType.BeforeVariableUnset, savedFrame,
                                            variable, name, index, flags, oldValue, null, null,
                                            null, null, false, true, ReturnCode.Ok);

#if DEBUGGER
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonWatch, false) &&
                                            CanHitWatches() && EntityOps.IsBreakOnUnset(variable))
                                        {
                                            //
                                            // NOTE: Debugger support, call Debugger.Trace if we are breaking on
                                            //       unset access for this variable.
                                            //
                                            if (CheckWatches(
                                                    ReturnCode.Ok, traceInfo.BreakpointType, name, null,
                                                    traceInfo, ref error) != ReturnCode.Ok)
                                            {
                                                return ReturnCode.Error;
                                            }
                                        }
#endif

                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonTrace, false) &&
                                            !EntityOps.IsNoTrace(variable) && (TraceLevels == 0))
                                        {
                                            try
                                            {
                                                if (RuntimeOps.FireTraces(
                                                        variable, traceInfo.BreakpointType, this, traceInfo,
                                                        ref error) != ReturnCode.Ok)
                                                {
                                                    return ReturnCode.Error;
                                                }
                                            }
                                            catch (Exception e)
                                            {
                                                error = String.Format(
                                                    "caught exception while firing variable traces: {0}",
                                                    e);

                                                return ReturnCode.Error;
                                            }
                                        }

#if NOTIFY
                                        if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NonNotify, false) &&
                                            !EntityOps.IsNoNotify(variable))
                                        {
                                            /* IGNORED */
                                            CheckNotifications(
                                                NotifyType.Variable, NotifyFlags.Trace,
                                                new ObjectPair(traceInfo.BreakpointType, traceInfo),
                                                this, null, null, null, ref error);
                                        }
#endif

                                        //
                                        // NOTE: If the trace(s) did not cancel the operation, proceed.
                                        //
                                        if (!traceInfo.Cancel)
                                        {
                                            if (savedVariable != null)
                                            {
                                                //
                                                // NOTE: Links to dead variable(s) should be dead too.
                                                //
                                                EntityOps.SetUndefined(savedVariable, true); // TEST: Test this.
                                            }

                                            //
                                            // NOTE: The variable is now dead.
                                            //
                                            EntityOps.SetUndefined(variable, true);

                                            //
                                            // NOTE: Remove any pending array searches that may exist for this
                                            //       variable.
                                            //
                                            CleanupArraySearches(variable);

                                            //
                                            // NOTE: Check for the variable purge flag.  If the purge flag is set,
                                            //       we remove all undefined variables from the current call frame.
                                            //
                                            if (FlagOps.HasFlags(traceInfo.Flags, VariableFlags.Purge, true))
                                            {
                                                //
                                                // NOTE: Purge all undefined variables in the call frame for the
                                                //       variable to be unset.  This implies removal of the variable
                                                //       to be unset because we just set it to be undefined (above).
                                                //
                                                CallFrameOps.Cleanup(CurrentFrame, variableFrame, false);
                                            }
                                            else if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NoRemove, true))
                                            {
                                                //
                                                // NOTE: Remove all leading namespace separator characters.
                                                //
                                                // TODO: Modify this when full [namespace] support has been added.
                                                //
                                                string newName = StringOps.TrimLeadingColons(name);

                                                //
                                                // NOTE: Remove the entire array if there is one; otherwise,
                                                //       just remove this scalar variable.
                                                //
                                                variableFrame.Variables.Remove(newName);
                                            }

                                            //
                                            // BUGFIX: Mark the variable as "dirty" AFTER the actual
                                            //         modifications have been completed.
                                            //
                                            EntityOps.SetDirty(variable, true);
                                        }

                                        return traceInfo.ReturnCode;
                                    }
                                    else
                                    {
                                        error = String.Format(
                                            "can't unset {0}: invalid call frame",
                                            FormatOps.ErrorVariableName(name, null));
                                    }
                                }
                                else
                                {
                                    error = String.Format(
                                        "can't unset {0}: variable is read-only",
                                        FormatOps.ErrorVariableName(name, null));
                                }
                            }
                        }
                        else
                        {
                            //
                            // BUGBUG: This may need revision; we do need some way to remove
                            //         "stale" undefined variables from the call frame even
                            //         though the only way to officially create one is via the
                            //         global or upvar commands and those get cleaned up
                            //         automatically upon procedure exit.
                            //
                            if (FlagOps.HasFlags(flags, VariableFlags.NoComplain, true))
                            {
                                return ReturnCode.Ok;
                            }
                            else
                            {
                                error = String.Format(
                                    "can't unset {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, linkIndex, name, index));
                            }
                        }
                    }
                    else
                    {
                        if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                        {
                            if (FlagOps.HasFlags(flags, VariableFlags.NoComplain, true))
                            {
                                return ReturnCode.Ok;
                            }
                            else
                            {
                                error = String.Format(
                                    "can't unset {0}: no such variable",
                                    FormatOps.ErrorVariableName(
                                        variable, null, name, index));
                            }
                        }
                        else
                        {
                            if (FlagOps.HasFlags(flags, VariableFlags.NoComplain, true))
                            {
                                return ReturnCode.Ok;
                            }
                            else
                            {
                                //
                                // NOTE: We did not actually search for the variable,
                                //       inform the caller as to why.
                                //
                                error = localError;
                            }
                        }
                    }
                }
                else
                {
                    error = String.Format(
                        "can't unset {0}: interpreter is read-only",
                        FormatOps.ErrorVariableName(name, index));
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode ClearVariableNameWait(
            Interpreter interpreter, /* in */
            string name,             /* in */
            VariableFlags flags      /* in */
            )
        {
            Result error = null;

            return ClearVariableNameWait(interpreter, name, flags, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode ClearVariableNameWait(
            Interpreter interpreter, /* in */
            string name,             /* in */
            VariableFlags flags,     /* in */
            ref Result error         /* out */
            ) /* THREAD-SAFE */
        {
            if (interpreter != null)
            {
                if (name != null)
                {
                    //
                    // BUGFIX: If the interpreter is disposed, skip clearing the variable
                    //         wait flags because the variable belongs to the interpreter,
                    //         which is now invalid.
                    //
                    lock (interpreter.syncRoot) /* TRANSACTIONAL */
                    {                           /* FIELD */
                        if (!IsDeletedOrDisposed(interpreter))
                        {
                            IVariable variable = null;

                            //
                            // NOTE: Attempt to search for the variable in the current
                            //       variable call frame.
                            //
                            if (interpreter.GetVariableViaResolversWithSplit(
                                    name, ref flags, ref variable,
                                    ref error) == ReturnCode.Ok)
                            {
                                EntityOps.SetWait(variable, false);

                                return ReturnCode.Ok;
                            }
                        }
                        else
                        {
                            error = "attempt to call eval in disposed interpreter";
                        }
                    }
                }
                else
                {
                    error = "invalid variable name";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode IsVariableNameDirty(
            Interpreter interpreter, /* in */
            string name,             /* in */
            VariableFlags flags,     /* in */
            ref bool once,           /* in, out */
            ref bool exists,         /* in, out */
            ref bool dirty,          /* out */
            ref Result error         /* out */
            ) /* THREAD-SAFE */
        {
            if (interpreter != null)
            {
                //
                // NOTE: *WARNING* Empty variable names are allowed, please do not
                //       change these to "!String.IsNullOrEmpty".
                //
                if (name != null)
                {
                    lock (interpreter.syncRoot) /* TRANSACTIONAL */
                    {                           /* FIELD */
                        //
                        // BUGFIX: Avoid disposal race condition.
                        //
                        if (!IsDeletedOrDisposed(interpreter))
                        {
                            IVariable variable = null;
                            Result localError = null;

                            //
                            // NOTE: Attempt to search for the variable in the current
                            //       variable call frame.
                            //
                            if (interpreter.GetVariableViaResolversWithSplit(
                                    name, ref flags, ref variable,
                                    ref localError) == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Is the variable really a link to another
                                //       variable?
                                //
                                if (FlagOps.HasFlags(flags, VariableFlags.WaitFollowLink, true) &&
                                    EntityOps.IsLink(variable))
                                {
                                    //
                                    // NOTE: Follow the linked variable.
                                    //
                                    variable = EntityOps.FollowLinks(variable);
                                }

                                //
                                // NOTE: Have we seen this variable already?
                                //
                                if (once)
                                {
                                    //
                                    // NOTE: If the dirty flag is set, the variable
                                    //       is considered to be dirty.
                                    //
                                    dirty = EntityOps.IsDirty(variable);
                                }
                                else
                                {
                                    //
                                    // NOTE: Initially, this variable does exist.
                                    //
                                    exists = true;

                                    //
                                    // NOTE: Flag this variable as non-dirty now.
                                    //
                                    EntityOps.SetDirty(variable, false);
                                    EntityOps.SetWait(variable, true);

                                    //
                                    // NOTE: We have now seen this variable at least
                                    //       once.
                                    //
                                    once = true;
                                }

                                //
                                // NOTE: If the variable is dirty and tracing is enabled,
                                //       display debugging information before returning.
                                //
                                if (dirty && FlagOps.HasFlags(
                                        flags, VariableFlags.WaitTrace, true))
                                {
                                    RuntimeOps.DebugTrace(String.Format(
                                        "{0}: IsVariableNameDirty: variable resolved, code = Ok, " +
                                        "interpreter = {1}, name = {2}, flags = {3}, once = {4}, " +
                                        "exists = {5}, dirty = {6}, error = {7}{8}",
                                        GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                                        FormatOps.WrapOrNull(name), FormatOps.WrapOrNull(flags), once,
                                        exists, dirty, FormatOps.WrapOrNull(true, true, error),
                                        Environment.NewLine), typeof(Interpreter).Name);
                                }

                                return ReturnCode.Ok;
                            }
                            else
                            {
                                //
                                // NOTE: Did we actually search for the variable?
                                //
                                if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                                {
                                    //
                                    // NOTE: Have we seen this variable already?
                                    //
                                    if (once)
                                    {
                                        //
                                        // NOTE: If the variable existed previously and
                                        //       it does not exist now then it is dirty.
                                        //
                                        if (exists)
                                            dirty = true;
                                    }
                                    else
                                    {
                                        //
                                        // NOTE: Initially, this variable does not exist.
                                        //
                                        exists = false;

                                        //
                                        // NOTE: We have now seen this variable at least
                                        //       once.
                                        //
                                        once = true;
                                    }

                                    //
                                    // NOTE: If the variable is dirty and tracing is enabled,
                                    //       display debugging information before returning.
                                    //
                                    if (dirty && FlagOps.HasFlags(
                                            flags, VariableFlags.WaitTrace, true))
                                    {
                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: IsVariableNameDirty: variable unresolved, code = Ok, " +
                                            "interpreter = {1}, name = {2}, flags = {3}, once = {4}, " +
                                            "exists = {5}, dirty = {6}, error = {7}{8}",
                                            GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                                            FormatOps.WrapOrNull(name), FormatOps.WrapOrNull(flags),
                                            once, exists, dirty, FormatOps.WrapOrNull(true, true, error),
                                            Environment.NewLine), typeof(Interpreter).Name);
                                    }

                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    //
                                    // NOTE: We could not search for the variable, let the
                                    //       caller know why.
                                    //
                                    error = localError;
                                }
                            }
                        }
                        else
                        {
                            error = "attempt to call eval in disposed interpreter";
                        }
                    }
                }
                else
                {
                    error = "invalid variable name";
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            //
            // NOTE: If the variable is dirty and tracing is enabled,
            //       display debugging information before returning.
            //
            if (dirty && FlagOps.HasFlags(
                    flags, VariableFlags.WaitTrace, true))
            {
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: IsVariableNameDirty: code = Error, interpreter = {1}, name = {2}, " +
                    "flags = {3}, once = {4}, exists = {5}, dirty = {6}, error = {7}{8}",
                    GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                    FormatOps.WrapOrNull(name), FormatOps.WrapOrNull(flags),
                    once, exists, dirty, FormatOps.WrapOrNull(true, true, error),
                    Environment.NewLine), typeof(Interpreter).Name);
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool ShouldWaitVariable()
        {
            //
            // BUGFIX: Prevent possible future deadlock caused by holding the
            //         interpreter lock while calling the Enabled and
            //         EventCount properties method, which may at some point
            //         acquire and hold the event manager lock.
            //
            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            if (eventManager != null)
            {
                bool enabled = eventManager.Enabled;

                return enabled && ((eventManager.TotalEventCount > 0) ||
#if NETWORK
                    (serverSockets > 0)
#else
                    false /* NOTE: No network support. */
#endif
                    );
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private EventWaitHandle[] GetEventWaitHandles(
            EventWaitHandle @event,
            bool reset
            )
        {
            EventWaitHandle variableEvent;

            lock (syncRoot)
            {
                variableEvent = this.variableEvent;
            }

            IEventManager eventManager;

            lock (syncRoot)
            {
                eventManager = this.eventManager;
            }

            EventWaitHandle emptyEvent = null;
            EventWaitHandle enqueueEvent = null;
            EventWaitHandle[] userEvents = null;

            if (eventManager != null)
            {
                emptyEvent = eventManager.EmptyEvent;
                enqueueEvent = eventManager.EnqueueEvent;
                userEvents = eventManager.UserEvents;
            }

            int length = ConversionOps.ToInt(variableEvent != null) +
                ConversionOps.ToInt(emptyEvent != null) +
                ConversionOps.ToInt(enqueueEvent != null) +
                ConversionOps.ToInt(@event != null) +
                ((userEvents != null) ? userEvents.Length : 0);

            EventWaitHandle[] events;

            if (length > 0)
            {
                events = new EventWaitHandle[length];

                int startIndex = 0;

                if ((length > startIndex) && (variableEvent != null))
                    events[startIndex++] = variableEvent;

                if ((length > startIndex) && (emptyEvent != null))
                    events[startIndex++] = emptyEvent;

                if ((length > startIndex) && (enqueueEvent != null))
                    events[startIndex++] = enqueueEvent;

                //
                // NOTE: *HOOK* Allow the immediate caller to inject an event
                //       into the array to wait for.
                //
                if ((length > startIndex) && (@event != null))
                    events[startIndex++] = @event;

                //
                // NOTE: *HOOK* Allow external applications and/or users to
                //       cleanly interact with the core event loop.
                //
                if (length > startIndex)
                    for (int index = startIndex; index < length; index++)
                        events[index] = userEvents[index - startIndex];

                if (reset)
                {
                    //
                    // NOTE: Minimally, we need to reset the master variable
                    //       event to make sure we do not [always] return
                    //       immediately from the wait call inside the loop.
                    //
                    for (int index = 0; index < length; index++)
                        ThreadOps.ResetEvent(events[index]);
                }

                return events;
            }

            return null;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Variable Management (Multiple)
        public ReturnCode GetVariableValues(
            VariableFlags flags,
            StringSortedList variables,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            if (variables != null)
            {
                StringList keys = new StringList(variables.Keys);

                foreach (string name in keys)
                {
                    ReturnCode code;
                    Result value = null;

                    code = GetVariableValue(flags, name, ref value, ref error);

                    //
                    // NOTE: Non-strict mode allows us to unconditionally query a
                    //       group of variables without worrying that some of them
                    //       may not actually exist.
                    //
                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Success, replace contained value with
                        //       queried value.
                        //
                        variables[name] = value;
                    }
                    else if (strict)
                    {
                        //
                        // NOTE: Failure, return error code to caller.
                        //
                        return code;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid variable list";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode SetVariableValues(
            VariableFlags flags,
            TraceList traces,
            StringSortedList variables,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            if (variables != null)
            {
                foreach (KeyValuePair<string, string> pair in variables)
                {
                    ReturnCode code;

                    code = SetVariableValue(flags, pair.Key, pair.Value, traces, ref error);

                    //
                    // NOTE: Non-strict mode allows us to unconditionally set a
                    //       group of variables without worrying that some of them
                    //       may actually be read-only, etc.
                    //
                    if (strict && (code != ReturnCode.Ok))
                    {
                        //
                        // NOTE: Failure, return error code to caller.
                        //
                        return code;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid variable list";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public ReturnCode UnsetVariables(
            VariableFlags flags,
            StringSortedList variables,
            bool strict,
            ref Result error
            )
        {
            CheckDisposed();

            if (variables != null)
            {
                foreach (string name in variables.Keys)
                {
                    ReturnCode code;

                    code = UnsetVariable(flags, name, ref error);

                    //
                    // NOTE: Non-strict mode allows us to unconditionally unset a
                    //       group of variables without worrying that some of them
                    //       may not actually exist.
                    //
                    if (strict && (code != ReturnCode.Ok))
                    {
                        //
                        // NOTE: Failure, return error code to caller.
                        //
                        return code;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid variable list";
            }

            return ReturnCode.Error;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Creation, Pre-Setup, Setup, & Initialization
        #region Creation
        internal CreateFlags FilterCreateFlags(
            bool debug,
            bool safe,
            bool standard
            )
        {
            CreateFlags result = CreateFlags.None;

#if DEBUGGER
            if (debug)
                result |= (createFlags & CreateFlags.DebuggerUse);
#endif

            if (safe)
                result |= (createFlags & CreateFlags.SafeAndHideUnsafe);

            if (standard)
                result |= (createFlags & CreateFlags.StandardAndHideNonStandard);

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static CreateFlags GetStartupCreateFlags( /* throw */
            IEnumerable<string> args,
            CreateFlags createFlags,
            OptionOriginFlags originFlags,
            bool console,
            bool verbose
            )
        {
            StringList argv = (args != null) ? new StringList(args) : null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: GetStartupCreateFlags: entered, argv = {1}, createFlags = {2}, " +
                "originFlags = {3}, console = {4}, verbose = {5}{6}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(true, true, argv),
                FormatOps.WrapOrNull(createFlags), FormatOps.WrapOrNull(originFlags),
                console, verbose, Environment.NewLine), typeof(Interpreter).Name);

            if (FlagOps.HasFlags(originFlags, OptionOriginFlags.Environment, true))
            {
                //
                // NOTE: Check for the debug environment variable.
                //       If it is set (to anything) then set the
                //       debug flag for the interpreter right now.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.Debug, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.Debug);
#endif

                    createFlags |= CreateFlags.Debug;
                }

                //
                // NOTE: Check for the "safe" environment variable.
                //       If it is set (to anything) then enable the
                //       appropriate "safe mode" flags for the
                //       interpreter.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.Safe, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.Safe); /* throw */
#endif

                    createFlags |= CreateFlags.SafeAndHideUnsafe;
                }

                //
                // NOTE: Check for the "standard" environment variable.
                //       If it is set (to anything) then enable the
                //       appropriate "standard mode" flags for the
                //       interpreter.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.Standard, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.Standard); /* throw */
#endif

                    createFlags |= CreateFlags.StandardAndHideNonStandard;
                }

                //
                // NOTE: Check for the no-color environment variable.
                //       If it is set (to anything) then make sure the
                //       console output will not be in color.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoColor, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoColor); /* throw */
#endif

                    createFlags |= CreateFlags.NoColor;
                }

                //
                // NOTE: Check for the no-title environment variable.
                //       If it is set (to anything) then skip setting the
                //       console title.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoTitle, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoTitle); /* throw */
#endif

                    createFlags |= CreateFlags.NoTitle;
                }

                //
                // NOTE: Check for the no-icon environment variable.
                //       If it is set (to anything) then skip setting the
                //       console icon.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoIcon, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoIcon); /* throw */
#endif

                    createFlags |= CreateFlags.NoIcon;
                }

                //
                // NOTE: Check for the no-profile environment variable.
                //       If it is set (to anything) then skip loading the
                //       host profile.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoProfile, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoProfile); /* throw */
#endif

                    createFlags |= CreateFlags.NoProfile;
                }

                //
                // NOTE: Check for the no-cancel environment variable.
                //       If it is set (to anything) then skip setting up the
                //       console script cancellation keypress, etc.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoCancel, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoCancel); /* throw */
#endif

                    createFlags |= CreateFlags.NoCancel;
                }

                //
                // NOTE: Check for the no-initialize environment variable.
                //       If it is set (to anything) then skip initialization
                //       of the interpreter.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoInitialize, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoInitialize);
#endif

                    createFlags &= ~CreateFlags.Initialize;
                }

                //
                // NOTE: Check for the no-throw-on-disposed environment variable.
                //       If it is set (to anything) then we will not throw exceptions
                //       when a disposed object is accessed.
                //
                if (GlobalState.GetEnvironmentVariable(
                        EnvVars.NoThrowOnDisposed, true, false, verbose) != null)
                {
#if CONSOLE
                    if (console && verbose)
                        Console.WriteLine(_Constants.Prompt.NoThrowOnDisposed);
#endif

                    createFlags &= ~CreateFlags.ThrowOnDisposed;
                }

                //
                // NOTE: Does the caller want to allow the CreateFlags to be
                //       overridden directly via the environment?
                //
                if (FlagOps.HasFlags(originFlags, OptionOriginFlags.Override, true))
                {
                    //
                    // NOTE: Check for the CreateFlags environment variable.
                    //       If it is set (and valid) then we will use it to
                    //       modify the creation flags for the interpreter.
                    //
                    string value = GlobalState.GetEnvironmentVariable(
                        EnvVars.CreateFlags, true, false, verbose);

                    if (!String.IsNullOrEmpty(value))
                    {
#if CONSOLE
                        if (console && verbose)
                            Console.WriteLine(String.Format(
                                _Constants.Prompt.CreateFlags, value));
#endif

                        object enumValue = null;
                        Result enumError = null;

                        enumValue = EnumOps.TryParseFlagsEnum(
                            null, typeof(CreateFlags), createFlags.ToString(),
                            value, null, true, ref enumError);

                        if (enumValue is CreateFlags)
                            createFlags = (CreateFlags)enumValue;
                        else
                            DebugOps.Complain(ReturnCode.Error, enumError);
                    }
                }
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: GetStartupCreateFlags: exited, argv = {1}, createFlags = {2}, " +
                "originFlags = {3}, console = {4}, verbose = {5}{6}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(true, true, argv),
                FormatOps.WrapOrNull(createFlags), FormatOps.WrapOrNull(originFlags),
                console, verbose, Environment.NewLine), typeof(Interpreter).Name);

            return createFlags;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode ProcessStartupOptions(
            Interpreter interpreter,
            IEnumerable<string> args,
            CreateFlags createFlags,
            OptionOriginFlags originFlags,
            bool console,
            bool verbose,
            ref Result error
            )
        {
            bool initialize = false;

            return ProcessStartupOptions(interpreter, args, createFlags, originFlags,
                console, verbose, ref initialize, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode ProcessStartupOptions(
            Interpreter interpreter,
            IEnumerable<string> args,
            CreateFlags createFlags,
            OptionOriginFlags originFlags,
            bool console,
            bool verbose,
            ref bool initialize,
            ref Result error
            )
        {
            bool loop = false;

            return ProcessStartupOptions(interpreter, args, createFlags, originFlags,
                console, verbose, ref initialize, ref loop, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode ProcessStartupOptions(
            Interpreter interpreter,
            IEnumerable<string> args,
            CreateFlags createFlags,
            OptionOriginFlags originFlags,
            bool console,
            bool verbose,
            ref bool initialize,
            ref bool loop,
            ref Result error
            )
        {
            StringList argv = (args != null) ? new StringList(args) : null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: ProcessStartupOptions: entered, interpreter = {1}, argv = {2}, " +
                "createFlags = {3}, originFlags = {4}, console = {5}, " +
                "verbose = {6}, initialize = {7}, loop = {8}, error = {9}{10}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                FormatOps.WrapOrNull(true, true, argv), FormatOps.WrapOrNull(createFlags),
                FormatOps.WrapOrNull(originFlags), console, verbose, initialize, loop,
                FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                typeof(Interpreter).Name);

            try
            {
                if (interpreter != null)
                {
                    if (FlagOps.HasFlags(originFlags, OptionOriginFlags.Environment, true))
                    {
                        //
                        // NOTE: Check for the no-trace environment variable.  If it
                        //       is set (to anything) then we clear all the trace
                        //       listeners.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.NoTrace, true, false) != null) /* GLOBAL */
                        {
                            //
                            // NOTE: Attempt to clear the trace listeners now.
                            //
                            DebugOps.ClearTraceListeners( /* GLOBAL */
                                interpreter.Debug, console, verbose);
                        }

                        //
                        // NOTE: Check for the trace environment variable.
                        //       If it is set (to anything) then we add the console
                        //       trace listener so that we can see debugging output.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.Trace, true, false) != null) /* GLOBAL */
                        {
                            //
                            // NOTE: Attempt to setup the trace listeners in verbose
                            //       mode.  This may fail; however, the success and/or
                            //       failure messages will be displayed on the console
                            //       from inside the method.
                            //
                            DebugOps.SetupTraceListeners( /* GLOBAL */
                                true, interpreter.Debug, console, verbose);
                        }

                        //
                        // NOTE: Check for the trace-to-host environment variable.
                        //       If it is set (to anything) then set the trace-to-host
                        //       flag for the interpreter right now.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.TraceToHost, true, false) != null)
                        {
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(_Constants.Prompt.TraceToHost);
#endif

                            interpreter.InterpreterFlags |= InterpreterFlags.TraceToHost;
                        }

                        //
                        // NOTE: Check for the interactive environment variable.
                        //       If it is set (to anything) then set the
                        //       interactive flag for the interpreter right now.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.Interactive, true, false) != null)
                        {
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(_Constants.Prompt.Interactive);
#endif

                            interpreter.Interactive = true;
                        }

                        //
                        // NOTE: Check for the single-step environment variable.
                        //       If it is set (to anything) then set the single
                        //       step flag for the interpreter right now.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.Step, true, false) != null)
                        {
#if DEBUGGER
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(_Constants.Prompt.SingleStep);
#endif

                            interpreter.SingleStep = true;
#else
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(_Constants.Prompt.NoDebugger);
#endif
#endif
                        }

                        //
                        // NOTE: Check for the no-initialize environment variable.
                        //       If it is set (to anything) then skip initialization
                        //       of the interpreter.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.NoInitialize, true, false) != null)
                        {
#if CONSOLE
                            //
                            // NOTE: Do not bother notifying the user that the script
                            //       library initialization of the interpreter will
                            //       be skipped if it is likely that they have already
                            //       been notified (i.e. if the "Initialize" flag has
                            //       already been disabled via GetStartupCreateFlags).
                            //
                            if (console && verbose &&
                                FlagOps.HasFlags(createFlags, CreateFlags.Initialize, true))
                            {
                                Console.WriteLine(_Constants.Prompt.NoInitialize);
                            }
#endif

                            initialize = false;
                        }

                        //
                        // NOTE: Check for the no-loop environment variable.
                        //       If it is set (to anything) then skip entering
                        //       the interactive loop.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.NoLoop, true, false) != null)
                        {
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(_Constants.Prompt.NoLoop);
#endif

                            loop = false;
                        }

                        //
                        // NOTE: Check for the no-throw-on-disposed environment variable.
                        //       If it is set (to anything) then we will not throw exceptions
                        //       when a disposed object is accessed.
                        //
                        if (GlobalState.GetEnvironmentVariable(
                                EnvVars.NoThrowOnDisposed, true, false) != null) /* GLOBAL */
                        {
#if CONSOLE
                            if (console && verbose &&
                                FlagOps.HasFlags(createFlags, CreateFlags.ThrowOnDisposed, true))
                            {
                                Console.WriteLine(_Constants.Prompt.NoThrowOnDisposed);
                            }
#endif

                            Engine.SetThrowOnDisposed(null, false, false); /* GLOBAL */
                        }

                        //
                        // NOTE: Check for the InterpreterFlags environment variable.
                        //       If it is set (and valid) then we will use it to
                        //       modify the instance flags for the interpreter.
                        //
                        string value = GlobalState.GetEnvironmentVariable(
                            EnvVars.InterpreterFlags, true, false);

                        if (!String.IsNullOrEmpty(value))
                        {
#if CONSOLE
                            if (console && verbose)
                                Console.WriteLine(String.Format(
                                    _Constants.Prompt.InterpreterFlags, value));
#endif

                            object enumValue = null;
                            Result enumError = null;

                            enumValue = EnumOps.TryParseFlagsEnum(
                                null, typeof(InterpreterFlags),
                                interpreter.InterpreterFlags.ToString(),
                                value, null, true, ref enumError);

                            if (enumValue is InterpreterFlags)
                                interpreter.InterpreterFlags = (InterpreterFlags)enumValue;
                            else
                                DebugOps.Complain(ReturnCode.Error, enumError);
                        }
                    }

                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: ProcessStartupOptions: exited, interpreter = {1}, argv = {2}, " +
                        "createFlags = {3}, originFlags = {4}, console = {5}, " +
                        "verbose = {6}, initialize = {7}, loop = {8}, error = {9}{10}",
                        GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                        FormatOps.WrapOrNull(true, true, argv), FormatOps.WrapOrNull(createFlags),
                        FormatOps.WrapOrNull(originFlags), console, verbose, initialize, loop,
                        FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                        typeof(Interpreter).Name);

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "invalid interpreter";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

#if CONSOLE
            if (console)
                Console.WriteLine(RuntimeOps.FormatResult(ReturnCode.Error, error));
#endif

            RuntimeOps.DebugTrace(String.Format(
                "{0}: ProcessStartupOptions: exited, interpreter = {1}, argv = {2}, " +
                "createFlags = {3}, originFlags = {4}, console = {5}, " +
                "verbose = {6}, initialize = {7}, loop = {8}, error = {9}{10}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(interpreter),
                FormatOps.WrapOrNull(true, true, argv), FormatOps.WrapOrNull(createFlags),
                FormatOps.WrapOrNull(originFlags), console, verbose, initialize, loop,
                FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                typeof(Interpreter).Name);

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: This is the "simplest" possible method that can
        //       be used to create an interpreter.
        //
        public static Interpreter Create( /* Example.Program */
            IEnumerable<string> args,
            ref Result result
            )
        {
            return Create(args, CreateFlags.Default, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create(
            InterpreterSettings interpreterSettings,
            bool strict,
            ref Result result
            )
        {
            if (interpreterSettings != null)
            {
                return Create(
                    interpreterSettings.Args, interpreterSettings.Culture,
                    interpreterSettings.CreateFlags, interpreterSettings.ScriptFlags,
                    interpreterSettings.AppDomain, interpreterSettings.Host,
                    interpreterSettings.Profile, interpreterSettings.ApplicationObject,
                    interpreterSettings.PolicyObject, interpreterSettings.ResolverObject,
                    interpreterSettings.UserObject, interpreterSettings.Policies,
                    interpreterSettings.Traces, interpreterSettings.Text,
                    interpreterSettings.LibraryPath, interpreterSettings.AutoPathList,
                    ref result);
            }

            if (strict)
            {
                result = "invalid interpreter settings";
                return null;
            }

            return Create(ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create(
            IEnumerable<string> args,
            CreateFlags createFlags,
            IHost host,
            ref Result result
            )
        {
            return Create(
                args, null, createFlags, ScriptFlags.Default, null, host, null,
                null, null, null, null, null, null, null, null, null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create(
            IEnumerable<string> args,
            CreateFlags createFlags,
            ref Result result
            )
        {
            return Create(args, createFlags, (TraceList)null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static Interpreter Create( /* NOTE: [test2] use only. */
            IEnumerable<string> args,
            CreateFlags createFlags,
            ScriptFlags scriptFlags,
            ref Result result
            )
        {
            return Create(
                args, null, createFlags, scriptFlags, null, null, null, null,
                null, null, null, null, null, null, null, null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create(
            ref Result result
            )
        {
            return Create(null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create( /* Featherlight.Components.Private.CommonOps */
            IEnumerable<string> args,
            CreateFlags createFlags,
            string text,
            string libraryPath,
            StringList autoPathList,
            ref Result result
            )
        {
            return Create(
                args, createFlags, null, null, null, null, null, null, text,
                libraryPath, autoPathList, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static Interpreter Create( /* InterpreterHelper.Create */
            IEnumerable<string> args,
            CreateFlags createFlags,
            ScriptFlags scriptFlags,
            string text,
            string libraryPath,
            StringList autoPathList,
            ref Result result
            )
        {
            return Create(
                args, null, createFlags, scriptFlags, null, null, null, null,
                null, null, null, null, null, text, libraryPath, autoPathList,
                ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create( /* Sample.Class1 */
            IEnumerable<string> args,
            CreateFlags createFlags,
            TraceList traces,
            ref Result result
            )
        {
            return Create(
                args, createFlags, null, null, null, null, null, traces, null,
                null, null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create( /* Eagle._Cmdlets.Script */
            IEnumerable<string> args,
            CreateFlags createFlags,
            object applicationObject,
            object policyObject,
            object resolverObject,
            object userObject,
            ExecuteCallbackDictionary policies,
            TraceList traces,
            string text,
            string libraryPath,
            StringList autoPathList,
            ref Result result
            )
        {
            return Create(
                args, null, createFlags, ScriptFlags.Default, null, null, null,
                applicationObject, policyObject, resolverObject, userObject,
                policies, traces, text, libraryPath, autoPathList, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static Interpreter Create( /* NOTE: INTERNAL DEBUGGER USE ONLY. */
            string culture,
            CreateFlags createFlags,
            AppDomain appDomain
            )
        {
            Result result = null;

            return Create(
                null, culture, createFlags, ScriptFlags.Default, appDomain,
                null, null, null, null, null, null, null, null, null, null,
                null, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter Create(
            IEnumerable<string> args,           // original arguments to the host application, where applicable.
            string culture,                     // currently, used only for parsing (error messages are same).
            CreateFlags createFlags,            // special creation flags.
            ScriptFlags scriptFlags,            // the default script flags used by the engine.
            AppDomain appDomain,                // pass null to use the current application domain.
            IHost host,                         // pass null to use the default host.
            string profile,                     // the host profile to load (pass null for none).
            object applicationObject,           // this is reserved for use by the application (pass null if not needed).
            object policyObject,                // this is reserved for use by custom policies (pass null if not needed).
            object resolverObject,              // this is reserved for use by custom resolvers (pass null if not needed).
            object userObject,                  // this is reserved for use by the user (pass null if not needed).
            ExecuteCallbackDictionary policies, // "safe" interpreter policies, if any.
            TraceList traces,                   // interpreter wide variable traces, if any.
            string text,                        // the "pre-init" (actually "post-setup") script, if any.
            string libraryPath,                 // location of the script library (i.e. "init" script).
            StringList autoPathList,            // extra dirs to check for packages.
            ref Result result                   // result of script library initialization or error upon failure.
            )
        {
            StringList argv = (args != null) ? new StringList(args) : null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Create: argv = {1}, culture = {2}, createFlags = {3}, " +
                "scriptFlags = {4}, appDomain = {5}, host = {6}, " +
                "profile = {7}, applicationObject = {8}, policyObject = {9}, " +
                "resolverObject = {10}, userObject = {11}, policies = {12}, " +
                "traces = {13}, text = {14}, libraryPath = {15}, " +
                "autoPathList = {16}, result = {16}{17}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(true, true, argv),
                FormatOps.WrapOrNull(culture), FormatOps.WrapOrNull(createFlags),
                FormatOps.WrapOrNull(scriptFlags), FormatOps.WrapOrNull(appDomain),
                FormatOps.WrapOrNull(host), FormatOps.WrapOrNull(profile),
                FormatOps.WrapOrNull(applicationObject), FormatOps.WrapOrNull(policyObject),
                FormatOps.WrapOrNull(resolverObject), FormatOps.WrapOrNull(userObject),
                FormatOps.WrapOrNull(policies), FormatOps.WrapOrNull(traces),
                FormatOps.WrapOrNull(text), FormatOps.WrapOrNull(libraryPath),
                FormatOps.WrapOrNull(true, true, autoPathList),
                FormatOps.WrapOrNull(true, true, result), Environment.NewLine),
                typeof(Interpreter).Name);

            //
            // NOTE: Pause for them to attach a debugger, if requested.  For embedders, we
            //       really need the ability to break into the debugger here; otherwise,
            //       being able to easily attach a debugger at a predictable point becomes
            //       a problem.
            //
            if (FlagOps.HasFlags(createFlags, CreateFlags.BreakOnCreate, true) ||
                ((GlobalState.GetEnvironmentVariable(EnvVars.Break, true, false) != null) &&
                 (Interlocked.Increment(ref breakCount) == 1)))
            {
#if CONSOLE
                Console.WriteLine(String.Format(
                    _Constants.Prompt.Debugger,
                    RuntimeOps.GetCurrentProcessId()));

                try
                {
                    Console.ReadKey(true); /* throw */
                }
                catch (InvalidOperationException) // Console.ReadKey
                {
                    // do nothing.
                }
#endif

                DebugOps.Break(); /* throw */
            }

            bool throwOnError = FlagOps.HasFlags(createFlags,
                CreateFlags.ThrowOnError, true);

            Interpreter interpreter = null;

            try
            {
                interpreter = new Interpreter();

                lock (interpreter.SyncRoot)
                {
                    GlobalState.PushActiveInterpreter(interpreter);

                    try
                    {
                        ReturnCode code = interpreter.PreSetup(
                            culture, createFlags, scriptFlags, appDomain,
                            host, profile, applicationObject, policyObject,
                            resolverObject, userObject, new StringDictionary(),
                            text, libraryPath, autoPathList, ref result);

                        //*****************************************************

                        if (code == ReturnCode.Ok)
                        {
                            code = interpreter.Setup(
                                createFlags, args, policies, traces, text,
                                ref result);
                        }

                        //*****************************************************
                        // WARNING: No scripts of any kind may be evaluated
                        //          using the interpreter until after this
                        //          point.
                        //*****************************************************

                        if (code == ReturnCode.Ok)
                        {
#if DEBUGGER
                            //
                            // NOTE: This must be prior to the Initialize()
                            //       call below if we want to be able to debug
                            //       the startup script.
                            //
                            // BUGFIX: Use the same host as we were passed,
                            //         never create a new one here.
                            //
                            if (FlagOps.HasFlags(
                                    createFlags, CreateFlags.Debugger, true))
                            {
                                interpreter.Debugger = RuntimeOps.NewDebugger(
                                    !FlagOps.HasFlags(createFlags,
                                        CreateFlags.NoIsolated, true),
                                    culture, createFlags, interpreter.GetAppDomain(),
                                    interpreter.Host);
                            }
#endif

                            bool initialize = true;

                            if (FlagOps.HasFlags(
                                    createFlags, CreateFlags.Startup, true))
                            {
                                code = ProcessStartupOptions(
                                    interpreter, args, createFlags,
                                    OptionOriginFlags.All,
                                    !FlagOps.HasFlags(createFlags,
                                        CreateFlags.NoHost, true) &&
                                    !FlagOps.HasFlags(createFlags,
                                        CreateFlags.NoConsole, true),
                                    true, ref initialize, ref result);
                            }

                            if (code == ReturnCode.Ok)
                            {
                                //
                                // NOTE: If script library initialization was
                                //       requested, do it now.
                                //
                                if (initialize &&
                                    FlagOps.HasFlags(createFlags,
                                        CreateFlags.Initialize, true))
                                {
                                    code = interpreter.PrivateInitialize(
                                        ref result);
                                }
                            }
                        }

                        //*****************************************************

                        if (code == ReturnCode.Ok)
                        {
                            GlobalState.AddInterpreter(interpreter);

#if NOTIFY && NOTIFY_GLOBAL
                            if (interpreter.GlobalNotify)
                                /* IGNORED */
                                CheckNotifications(
                                    null, false,
                                    NotifyType.Interpreter, NotifyFlags.Added,
                                    new ObjectTriplet(culture, createFlags, host),
                                    interpreter, null, null, null, ref result);
#endif
                        }
                        else
                        {
                            interpreter.Dispose(); /* throw */
                            interpreter = null;
                        }

                        //*****************************************************

                        if ((interpreter == null) && throwOnError)
                        {
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: Create: failure (with throw), " +
                                "code = {1}, result = {2}{3}",
                                GlobalState.GetCurrentThreadId(), code,
                                FormatOps.WrapOrNull(true, true, result),
                                Environment.NewLine), typeof(Interpreter).Name);

                            //
                            // NOTE: If we get to this point we have failed to
                            //       create and/or initialize the interpreter
                            //       AND they requested that we throw an
                            //       exception in that case.
                            //
                            throw new ScriptException(code, result, null);
                        }
                    }
                    finally
                    {
                        /* IGNORED */
                        GlobalState.PopActiveInterpreter();
                    }
                }
            }
            catch (Exception e)
            {
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: {1}{2}",
                    GlobalState.GetCurrentThreadId(),
                    e, Environment.NewLine),
                    typeof(Interpreter).Name);

                if (interpreter != null)
                {
                    interpreter.Dispose(); /* throw */
                    interpreter = null;
                }

                if (throwOnError)
                    throw;
                else
                    result = e;
            }

            if (interpreter != null)
            {
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: Create: success, interpreter = {1}, result = {2}{3}",
                    GlobalState.GetCurrentThreadId(),
                    FormatOps.WrapOrNull(interpreter),
                    FormatOps.WrapOrNull(true, true, result),
                    Environment.NewLine), typeof(Interpreter).Name);
            }
            else
            {
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: Create: failure (without throw), interpreter = {1}, " +
                    "result = {2}{3}", GlobalState.GetCurrentThreadId(),
                    FormatOps.WrapOrNull(interpreter),
                    FormatOps.WrapOrNull(true, true, result),
                    Environment.NewLine), typeof(Interpreter).Name);
            }

            return interpreter;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Pre-Setup
        private ReturnCode PreSetupCulture(
            string culture,
            bool specific,
            ref Result error
            )
        {
            //
            // NOTE: Empty string is valid here, we use it to select the current
            //       culture.
            //
            if (culture != null)
            {
                try
                {
                    //
                    // NOTE: Attempt to set the culture based on using the parameter
                    //       "culture" as a name (either neutral or specific).  Empty
                    //       string is valid here and selects the invariant culture.
                    //
                    if (specific)
                        cultureInfo = CultureInfo.CreateSpecificCulture(culture);
                    else
                        cultureInfo = CultureInfo.GetCultureInfo(culture);

                    return ReturnCode.Ok;
                }
                catch
                {
                    //
                    // NOTE: Ok, it was not a valid culture name, try to interpret the
                    //       parameter "culture" as an Id integer.
                    //
                    int cultureId = 0;

                    if (Value.GetInteger2(culture, ValueFlags.AnyInteger, null /* no culture yet! */,
                            ref cultureId, ref error) == ReturnCode.Ok)
                    {
                        try
                        {
                            cultureInfo = CultureInfo.GetCultureInfo(cultureId);

                            return ReturnCode.Ok;
                        }
                        catch (Exception e)
                        {
                            //
                            // NOTE: It parsed as a valid integer; however, the culture
                            //       specified by the Id was not found.
                            //
                            error = FormatOps.ErrorWithException(String.Format(
                                CultureInfoError, culture), e);

                            return ReturnCode.Error;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: It did not parse as a valid integer, fail.
                        //
                        error = String.Format(CultureInfoError, culture);

                        return ReturnCode.Error;
                    }
                }
            }
            else
            {
                //
                // NOTE: Getting this property value is documented to never fail.
                //
                cultureInfo = CultureInfo.InvariantCulture;

                return ReturnCode.Ok;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PreSetupResourceManager(ref Result error)
        {
            if (cultureInfo != null)
            {
                string resourceBaseName = GlobalState.GetResourceBaseName();

                if (resourceBaseName != null)
                {
                    try
                    {
                        //
                        // FIXME: PRI 4: Now that this resource management code
                        //        is in place and working properly, we need to
                        //        migrate all the error messages and other
                        //        static strings to be managed resources.  The
                        //        original intention was to do this right from
                        //        the start; however, time constraints prevented
                        //        that vision from becoming a reality.
                        //
                        resourceManager = new ResourceManager(
                            resourceBaseName, GlobalState.GetAssembly());

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = FormatOps.ErrorWithException(String.Format(
                            ResourceManagerError, resourceBaseName), e);
                    }
                }
                else
                {
                    error = InvalidBaseResourceName;
                }
            }
            else
            {
                error = InvalidCultureInfoError;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PreSetupHost(
            IHost host,
            bool strict,
            ref Result error
            )
        {
            //
            // NOTE: Set the system default values for the "default" foreground
            //       and background colors now.  That way, even if there is no
            //       interpreter host or it decides not to do anything, these
            //       values will still be explicitly initialized.
            //
            defaultForegroundColor = _ConsoleColor.Default;
            defaultBackgroundColor = _ConsoleColor.Default;

            //
            // NOTE: If this IHost call fails, the caller will fail to setup
            //       the interpreter.  This is considered to be legitimate,
            //       even in non-strict mode, because IHost implementations are
            //       free to simply return "Ok" without having to do anything
            //       else (i.e. the default colors have already been set to the
            //       system defaults before this call).  Therefore, an IHost
            //       implementation that returns "Error" here is explicitly
            //       saying "the interpreter setup should fail".  Furthermore,
            //       we attempt to further minimize the risk of failure here
            //       by explicitly requesting the default hard-coded colors
            //       from the host (i.e. both the "foreground" and "background"
            //       boolean arguments are set to "false") because that should
            //       be even less likely to fail (and it _is_ less likely to
            //       fail with the "stock" implementation of this method,
            //       as provided by the _Hosts.Core host).
            //
            return RuntimeOps.GetHostColors(
                host, ColorName.Default, false, false, strict,
                ref defaultForegroundColor, ref defaultBackgroundColor,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PreSetup(
            string culture,
            CreateFlags createFlags,
            ScriptFlags scriptFlags,
            AppDomain appDomain,
            IHost host,
            string profile,
            object applicationObject,
            object policyObject,
            object resolverObject,
            object userObject,
            StringDictionary runtimeOptions,
            string text,
            string libraryPath,
            StringList autoPathList,
            ref Result result
            )
        {
            #region Thread Affinity / AppDomain Identity
            //
            // NOTE: Keep track of the thread that this interpreter was created on.
            //
            this.Thread = Thread.CurrentThread;
            this.ThreadId = GlobalState.GetCurrentThreadId();
            this.VariableEvent = ThreadOps.CreateEvent(false);

            //
            // NOTE: What is the application domain for this interpreter?
            //
            this.appDomain = (appDomain != null) ?
                appDomain : AppDomain.CurrentDomain;
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Host Integration
            //
            // NOTE: Create and/or setup our host environment first because this
            //       will be used to communicate all other initialization failures.
            //
            if (host != null)
            {
                //
                // NOTE: Use the specified hosting environment.
                //
                this.host = host;
            }
            else if (!FlagOps.HasFlags(createFlags, CreateFlags.NoHost, true))
            {
                if (profile == null)
                    profile = GlobalState.GetEnvironmentVariable(
                        EnvVars.Profile, true, false);

#if CONSOLE
                if (!String.IsNullOrEmpty(profile))
                    Console.WriteLine(String.Format(
                        _Constants.Prompt.Profile, profile));
#endif

                //
                // NOTE: *HOOK* Allow the default host to be overridden by
                //       an external application or third-party plugin.
                //
                NewHostCallback callback = NewHostCallback; /* STATIC */

                if (callback != null)
                {
                    this.host = RuntimeOps.NewCustomHost(
                        callback, this, null, profile, createFlags);
                }
                else
                {
#if CONSOLE
                    //
                    // NOTE: Use the default hosting environment (optimized
                    //       for console apps, like the interactive shell).
                    //
                    if (!FlagOps.HasFlags(createFlags, CreateFlags.NoConsole, true))
                    {
                        this.host = RuntimeOps.NewConsoleHost(
                            this, null, profile, createFlags);
                    }
                    else
#endif
                    {
                        this.host = RuntimeOps.NewDiagnosticHost(
                            this, null, profile, createFlags);
                    }
                }
            }
            else
            {
                this.host = null;
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Application Integration
            //
            // NOTE: Set the application and user objects now, in case they are necessary
            //       during the later phases of interpreter initialization (e.g. for use
            //       with a custom security policy callback).
            //
            this.applicationObject = applicationObject;
            this.policyObject = policyObject;
            this.resolverObject = resolverObject;
            this.userObject = userObject;
            this.runtimeOptions = runtimeOptions;
            this.throwOnFeatureNotSupported = RuntimeOps.DefaultThrowOnFeatureNotSupported;
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            //
            // NOTE: We must have a culture and resource manager setup before we do anything
            //       further because they are required to get access to our string resources
            //       (error messages, etc).
            //
            if ((PreSetupCulture(culture, false, ref result) == ReturnCode.Ok) &&
                (PreSetupResourceManager(ref result) == ReturnCode.Ok) &&
                (PreSetupHost(this.host, false, ref result) == ReturnCode.Ok))
            {
                #region Public Properties
                this.ReadOnly = false;
                this.RecursionLimit = DefaultRecursionLimit;
                this.Timeout = DefaultTimeout;
                this.SleepTime = _Public.EventManager.DefaultSleepTime;
                this.ExitCode = RuntimeOps.SuccessExitCode();
                this.Quiet = DebugOps.GetDefaultQuiet(DefaultQuiet);
                this.CommandDecision = PolicyDecision.None;
                this.ScriptDecision = PolicyDecision.None;
                this.FileDecision = PolicyDecision.None;
                this.StreamDecision = PolicyDecision.None;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Script Library Pre-Initialization
                preInitializeText = text;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Script Library Paths
                this.libraryPath = libraryPath;
                this.autoPathList = autoPathList;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Script Library Miscellaneous
                initialized = false;
                initializedShell = false;
                initializedPath = null;
                trustedPaths = new StringList();
                trustedTypes = new ObjectDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Flags
                this.createFlags = createFlags;
                this.scriptFlags = scriptFlags;

                ///////////////////////////////////////////////////////////////////////////////////////

#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
                cacheFlags = CacheFlags.Default;
#endif

                ///////////////////////////////////////////////////////////////////////////////////////

                interpreterFlags = InterpreterFlags.None;
                headerFlags = HeaderFlags.Invalid;
                engineFlags = EngineFlags.None;
                interactiveEngineFlags = EngineFlags.None;
                substitutionFlags = SubstitutionFlags.All;
                expressionFlags = ExpressionFlags.All;
                interactiveSubstitutionFlags = SubstitutionFlags.None;
                afterEventFlags = EventFlags.After;
                engineEventFlags = EventFlags.Engine;
                interactiveEventFlags = EventFlags.None;
                queueEventFlags = EventFlags.Queue;
                serviceEventFlags = EventFlags.Service;
                updateEventFlags = EventFlags.Wait;
                waitEventFlags = EventFlags.Wait;
                pluginFlags = PluginFlags.None;

                ///////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY || NOTIFY_OBJECT
                notifyTypes = NotifyType.None;
                notifyFlags = NotifyFlags.None;

                //
                // NOTE: For now, default the "global" notification setting to
                //       enabled.  This can be overridden prior to any "global"
                //       notifications being fired by changing this setting
                //       from the "init" script or by using the NoGlobalNotify
                //       flag.
                //
                notify = !FlagOps.HasFlags(createFlags, CreateFlags.NoGlobalNotify, true);
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Plugin Tokens
                corePluginToken = 0;

#if NOTIFY || NOTIFY_OBJECT
                objectPluginToken = 0;
#endif

#if DEBUG
                testPluginToken = 0;
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Core Entities
                //
                // NOTE: How many commands have been executed?
                //
                commandCount = 0;

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Packages
                packageIndexes = new PackageIndexDictionary();
                packages = new PackageWrapperDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Plugins
                plugins = new PluginWrapperDictionary();
                commands = new CommandWrapperDictionary();
                hiddenCommands = new CommandWrapperDictionary();
                policies = new PolicyWrapperDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Executable Objects
                procedures = new ProcedureWrapperDictionary();
                hiddenProcedures = new ProcedureWrapperDictionary();
                // lambdas = new LambdaDictionary();
                executes = new ExecuteWrapperDictionary();
                hiddenExecutes = new ExecuteWrapperDictionary();
                operators = new OperatorWrapperDictionary();
                functions = new FunctionWrapperDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Command & Variable Resolvers
                //
                // NOTE: For now, just add the core resolver.
                //
                resolvers = new List<IResolve>(new IResolve[] {
                    RuntimeOps.NewCoreResolver(this)
                });
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Event Manager
                eventManager = new EventManager(this);

                //
                // NOTE: Initialize the count of pending wait calls to zero.
                //
                waitCount = 0;

                //
                // NOTE: How many times have we waited for an event (in WaitVariable)?
                //
                waitSpinCount = 0;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Variable Traces
                traces = new TraceWrapperDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Variable Scopes
                scopes = new CallFrameDictionary();
                #endregion
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region CLR Integration
                binder = new ScriptBinder(this, null, this.Debug);
                objects = new ObjectWrapperDictionary();
                aliases = new AliasWrapperDictionary();
                callbacks = new CallbackDictionary();

                ///////////////////////////////////////////////////////////////////////////////////////

                objectTypes = new StringDictionary();

                objectNamespaces = new StringDictionary(
                    (IDictionary<string, string>)Value.DefaultObjectNamespaces); /* COPY */

                objectInterfaces = new TypePairDictionary<string, long>();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Interpreters
                masterInterpreter = null;
                slaveInterpreters = new InterpreterDictionary();
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Other Entities
#if DATA
                connections = new DbConnectionDictionary();
                transactions = new DbTransactionDictionary();
#endif

#if CALLBACK_QUEUE
                callbackQueue = new CallbackQueue();
#endif

#if APPDOMAINS
                appDomains = new AppDomainDictionary();
#endif

#if LIST_CACHE
                stringListCache = new StringListDictionary();
#endif

#if PARSE_CACHE
                parseStateCache = new ParseStateDictionary();

                #region Dead Code
#if DEAD_CODE
                expressionStateCache = new ExpressionStateDictionary();
#endif
                #endregion
#endif

#if EXECUTE_CACHE
                executeCache = new ExecuteCache();
#endif

#if TYPE_CACHE
                typeCache = new TypeDictionary();
#endif

#if COM_TYPE_CACHE
                comTypeListCache = new IntPtrTypeListDictionary();
#endif

#if NATIVE && TCL
                tclInterps = new IntPtrDictionary();

#if TCL_THREADS
                tclThreads = new TclThreadDictionary();
#endif

                tclBridges = new TclBridgeDictionary();
#endif

#if NATIVE && LIBRARY
                modules = new ModuleWrapperDictionary();
                delegates = new DelegateWrapperDictionary();
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Random Number Generators
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoRandom, true))
                {
                    random = new Random();
                    cryptoRandom = RNGCryptoServiceProvider.Create();
                }
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Object Identifiers
#if RANDOMIZE_ID
                //
                // NOTE: If possible, randomize the starting Id for objects.
                //
                nextId = (random != null) ? Math.Abs(random.Next()) : 0;
#else
                nextId = 0;
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Script Error Handling
                unknown = TclVars.Unknown;
                backgroundError = TclVars.BackgroundError;
                packageUnknown = TclVars.PackageUnknown;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Channels & Encodings
                channels = new ChannelDictionary();

                //
                // NOTE: Initialize the "well-known" encodings.
                //
                encodings = new EncodingDictionary(new _Comparers.Custom(
                    StringOps.SystemNoCaseStringComparisonType));

                encodings.Add(Channel.NullEncodingName, null);
                encodings.Add(Channel.BinaryEncodingName, Channel.BinaryEncoding);
                encodings.Add(Channel.DefaultEncodingName, Channel.DefaultEncoding);
                encodings.Add(Channel.SystemDefaultEncodingName, Encoding.Default);
                encodings.Add(IdentityEncoding.webName, IdentityEncoding.Identity);
                encodings.Add(OneByteEncoding.webName, OneByteEncoding.OneByte);
                encodings.Add(TwoByteEncoding.webName, TwoByteEncoding.TwoByte);
                encodings.Add(TclEncoding.webName, TclEncoding.Tcl);
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Static Lock Held
                lock (staticSyncRoot)
                {
#if NATIVE && WINDOWS
                    #region Native Stack Checking
                    RuntimeOps.InitializeStackChecking();
                    #endregion
#endif

                    ///////////////////////////////////////////////////////////////////////////////////

#if THREADING
                    ContextManager.Initialize();
#endif

                    ///////////////////////////////////////////////////////////////////////////////////

#if NATIVE && TCL
                    #region Transfer Global Tcl/Tk Integration Data
                    //
                    // NOTE: If this thread (i.e. the primary thread for this
                    //       interpreter) is the same as the thread for the
                    //       dead Tcl objects in the global cache, transfer
                    //       them to us now.
                    //
                    if (IsPrimaryThread(deadTclThreadId))
                    {
                        ResultList errors = null;

                        ///////////////////////////////////////////////////////////////////////////////

                        if (deadTclThreadId != 0)
                            deadTclThreadId = 0;

                        ///////////////////////////////////////////////////////////////////////////////

                        if (deadTclApi != null)
                        {
                            tclApi = deadTclApi;
                            deadTclApi = null;

                            if (errors == null)
                                errors = new ResultList();

                            errors.Add("recovered dead Tcl API object");
                        }

                        ///////////////////////////////////////////////////////////////////////////////

                        if (deadTclInterps != null)
                        {
                            if (tclInterps != null)
                                tclInterps.Add(deadTclInterps);

                            if (errors == null)
                                errors = new ResultList();

                            errors.Add(String.Format(
                                "recovered {0} dead Tcl interpreters: {1}",
                                deadTclInterps.Count, deadTclInterps));

                            deadTclInterps.Clear();
                            deadTclInterps = null;
                        }

                        ///////////////////////////////////////////////////////////////////////////////

#if TCL_THREADS
                        if (deadTclThreads != null)
                        {
                            if (tclThreads != null)
                                tclThreads.Add(deadTclThreads);

                            errors.Add(String.Format(
                                "recovered {0} dead Tcl threads: {1}",
                                deadTclThreads.Count, deadTclThreads));

                            deadTclThreads.Clear();
                            deadTclThreads = null;
                        }
#endif

                        ///////////////////////////////////////////////////////////////////////////////

                        if (deadTclBridges != null)
                        {
                            if (tclBridges != null)
                                tclBridges.Add(deadTclBridges);

                            errors.Add(String.Format(
                                "recovered {0} dead Tcl bridges: {1}",
                                deadTclBridges.Count, deadTclBridges));

                            deadTclBridges.Clear();
                            deadTclBridges = null;
                        }

                        ///////////////////////////////////////////////////////////////////////////////

                        if ((errors != null) && (errors.Count > 0))
                        {
                            errors.Insert(0, String.Format(
                                "interpreter \"{0}\" from thread {1}",
                                this, threadId));

                            DebugOps.Complain(this, ReturnCode.Error, errors);
                        }
                    }
                    #endregion

                    ///////////////////////////////////////////////////////////////////////////////////

                    #region Initialize Global Tcl/Tk Integration Data
                    if (deadTclInterps == null)
                        deadTclInterps = new IntPtrDictionary();

#if TCL_THREADS
                    if (deadTclThreads == null)
                        deadTclThreads = new TclThreadDictionary();
#endif

                    if (deadTclBridges == null)
                        deadTclBridges = new TclBridgeDictionary();
                    #endregion
#endif
                }
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Context Manager
#if THREADING
                //
                // NOTE: This must be done AFTER the static Initialize method
                //       of the ContextManager class has been called (above).
                //
                contextManager = new ContextManager(this);
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Call Stack (Non-Threaded Only)
#if !THREADING
                CallStack = new CallStack(this.RecursionLimit);
                GlobalFrame = CallFrameOps.NewGlobal(this, CallStack);

                PushGlobalCallFrame(false); // NOTE: This call frame is never popped.
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Engine Non-Context Members
                Cancel = false;
                Unwind = false;
                Halt = false;

                CancelResult = null;
                HaltResult = null;

                Complaint = null;
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Engine Context Members
#if !THREADING
                Levels = 0;
                ParserLevels = 0;
                ExpressionLevels = 0;
                UnknownLevels = 0;
                TraceLevels = 0;

#if NOTIFY || NOTIFY_OBJECT
                NotifyLevels = 0;
                NotifyTypes = NotifyType.None;
                NotifyFlags = NotifyFlags.None;
#endif

                PolicyLevels = 0;
                TestLevels = 0;

#if DEBUGGER
                Debugger = null;
#endif

#if PREVIOUS_RESULT
                PreviousResult = null;
#endif

                EngineFlags = EngineFlags.None;

                ReturnCode = ReturnCode.Ok;

                ErrorLine = 0;
                ErrorCode = null;
                ErrorInfo = null;

                ScriptLocation = null;
                ScriptLocations = new StackList<IScriptLocation>();

                PreviousProcessId = 0;

                ArraySearches = new ArraySearchDictionary();

#if HISTORY
                HistoryEngineFilter = null;
                _History = new ClientDataList();
#endif
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Interactive Context Members
#if !THREADING
#if HISTORY
                HistoryLoadData = null;
                HistorySaveData = null;

                HistoryInfoFilter = null;
                HistoryLoadFilter = null;
                HistorySaveFilter = null;

                HistoryFileName = null;
#endif
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                #region Test Context Members
#if !THREADING
                //
                // NOTE: Initialize the test related data.
                //
                TestStatistics = new int[(int)TestInformationType.SizeOf];
                TestConstraints = new StringList();
                TestSkipped = new StringList();
                TestFailures = new StringList();
                TestCounts = new IntDictionary();
                TestMatch = new StringList(); /* TODO: Should this default to all (*)? */
                TestSkip = new StringList();
                TestReturnCodeMessages = TestOps.GetReturnCodeMessages();

#if DEBUGGER
                TestBreakpoints = new StringDictionary();
#endif
#endif
                #endregion

                ///////////////////////////////////////////////////////////////////////////////////////

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Setup
#if DEBUGGER
        private static ReturnCode SetupDebuggerHost(
            Interpreter interpreter,
            IHost host,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    //
                    // NOTE: If this interpreter has no debugger then we
                    //       can simply do nothing.
                    //
                    IDebugger debugger = interpreter.Debugger;

                    if (debugger != null)
                    {
                        //
                        // NOTE: Extract the isolated debugger interpreter
                        //       and check its validity.
                        //
                        Interpreter debugInterpreter = debugger.Interpreter;

                        if (debugInterpreter != null)
                        {
                            try
                            {
                                //
                                // NOTE: If the provided host is valid, clone it
                                //       into the isolated debugger interpreter;
                                //       otherwise, just use the provided null
                                //       value (i.e. if the parent interpreter
                                //       does not have a valid host then the
                                //       isolated debugger interpreter should not
                                //       have one either).
                                //
                                IHost newHost = (host != null) ?
                                    host.Clone(debugInterpreter) : null;

                                //
                                // NOTE: Set the host for the isolated debugger
                                //       interpreter (may be null).
                                //
                                debugInterpreter.Host = newHost;

                                //
                                // NOTE: If we set the host for the isolated
                                //       debugger interpreter to a non-null value,
                                //       create the standard channels for it now;
                                //       otherwise, we are totally done.
                                //
                                if (newHost != null)
                                    return SetupHostChannels(
                                        debugInterpreter, newHost, ref error);
                                else
                                    return ReturnCode.Ok;
                            }
                            catch (Exception e)
                            {
                                error = e;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: Ok, the debugger interpreter is not available.
                            //       This is not an error.
                            //
                            return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Ok, the debugger is not available.  This is not
                        //       an error.
                        //
                        return ReturnCode.Ok;
                    }
                }
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetupTraces(
            TraceList traces,
            IClientData clientData,
            ref Result error
            )
        {
            if (traces != null)
            {
                foreach (ITrace trace in traces)
                {
                    Result localResult = null;

                    if (AddTrace(trace, clientData,
                            ref localResult) != ReturnCode.Ok)
                    {
                        error = localResult;
                        return ReturnCode.Error;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid traces";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetupPolicies(
            ExecuteCallbackDictionary policies,
            IPlugin plugin,
            IClientData clientData,
            ref Result error
            )
        {
            if (policies != null)
            {
                foreach (KeyValuePair<string, ExecuteCallback> pair in policies)
                {
                    Result localResult = null;

                    if (AddPolicy(pair.Value, plugin, clientData,
                            ref localResult) != ReturnCode.Ok)
                    {
                        error = localResult;
                        return ReturnCode.Error;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "invalid policies";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static void SetupSafeVariableNames()
        {
            #region Static Lock Held
            lock (staticSyncRoot)
            {
                //
                // WARNING: These lists MUST be kept in sync with the variable
                //          setup code in the Setup and SetupPlatform methods.
                //
                if (safeVariableNames == null) /* ONCE */
                    safeVariableNames = new StringList(
                        Vars.Null, Vars.Platform.Name, TclVars.Interactive,
                        TclVars.PatchLevelName, TclVars.Platform.Name,
                        TclVars.VersionName);

                if (safeTclPlatformElementNames == null) /* ONCE */
                    safeTclPlatformElementNames = new StringList(
                        TclVars.Platform.ByteOrder, TclVars.Platform.CharacterSize,
                        TclVars.Platform.Debug, TclVars.Platform.DirectorySeparator,
                        TclVars.Platform.Engine, TclVars.Platform.PatchLevel,
                        TclVars.Platform.PathSeparator, TclVars.Platform.PlatformName,
                        TclVars.Platform.PointerSize, TclVars.Platform.Threaded,
                        TclVars.Platform.Unicode, TclVars.Platform.Version,
                        TclVars.Platform.WordSize);

                if (safeEaglePlatformElementNames == null) /* ONCE */
                    safeEaglePlatformElementNames = new StringList(
                        Vars.Platform.Configuration, Vars.Platform.Engine,
                        Vars.Platform.PatchLevel, Vars.Platform.Version);
            }
            #endregion
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetupVariables(
            CreateFlags createFlags,
            IEnumerable<string> args,
            bool strict,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
            {
                SetupSafeVariableNames();

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.VersionName,
                        TclVars.VersionValue, null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.PatchLevelName,
                        TclVars.PatchLevelValue, null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.ErrorCode,
                        null, null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.ErrorInfo,
                        null, null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue(
                        VariableFlags.GlobalOnly, TclVars.PrecisionName,
                        TclVars.PrecisionValue.ToString(),
                        precisionTraceList, ref result);

                if (code == ReturnCode.Ok)
                {
                    if (!FlagOps.HasFlags(createFlags, CreateFlags.Safe, true))
                    {
                        if (args != null)
                        {
                            StringList arguments = new StringList(args);

                            if (code == ReturnCode.Ok)
                                code = SetVariableValue(
                                    VariableFlags.GlobalOnly, TclVars.ShellArgumentCount,
                                    arguments.Count.ToString(), null, ref result);

                            if (code == ReturnCode.Ok)
                                code = SetVariableValue(
                                    VariableFlags.GlobalOnly, TclVars.ShellArguments,
                                    arguments.ToString(), null, ref result);
                        }
                        else if (FlagOps.HasFlags(
                                createFlags, CreateFlags.SetArguments, true))
                        {
                            if (code == ReturnCode.Ok)
                                code = SetVariableValue(
                                    VariableFlags.GlobalOnly, TclVars.ShellArgumentCount,
                                    0.ToString(), null, ref result);

                            if (code == ReturnCode.Ok)
                                code = SetVariableValue(
                                    VariableFlags.GlobalOnly, TclVars.ShellArguments,
                                    null, null, ref result);
                        }

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue(
                                VariableFlags.GlobalOnly, TclVars.ShellArgument0,
                                PathOps.GetExecutableName(), null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue(
                                VariableFlags.GlobalOnly, TclVars.RunCommandsFileName,
                                TclVars.RunCommandsFileValue, null, ref result);

                        if ((code == ReturnCode.Ok) && initialized)
                            code = SetVariableValue(
                                VariableFlags.GlobalOnly, TclVars.AutoPath,
                                GlobalState.GetAutoPathList(this, false).ToString(),
                                autoPathTraceList, ref result);

                        if (code == ReturnCode.Ok)
                            code = AddVariable(
                                VariableFlags.GlobalOnly | VariableFlags.Array,
                                TclVars.Environment,
                                environmentTraceList, strict, ref result);

                        if (code == ReturnCode.Ok)
                            code = AddVariable(
                                VariableFlags.GlobalOnly | VariableFlags.Array,
                                Vars.Tests,
                                testsTraceList, strict, ref result);

                        if (code == ReturnCode.Ok)
                            code = AddVariable(
                                VariableFlags.GlobalOnly | VariableFlags.Array,
                                TclVars.AutoIndex, null, strict, ref result);
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetupFunctions(
            CreateFlags createFlags,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(createFlags, CreateFlags.NoFunctions, true))
            {
                IPlugin plugin = GetCorePlugin(ref result);

                if (plugin != null)
                {
                    code = AddFunction(
                        null, "abs", Value.numberTypes, FunctionFlags.None,
                        plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "acos", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "asin", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "atan", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "atan2", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "bool", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "ceil", Value.nonIntegralTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "cos", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "cosh", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "datetime", Value.allTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "decimal", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "double", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "e", null, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "epsilon", null, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "exp", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "flags", Value.stringTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "floor", Value.nonIntegralTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "fmod", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "hypot", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "int", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "list", Value.allTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "log", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "log10", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "log2", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "logx", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "max", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "min", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "pi", null, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "pow", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "rand", null, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "random", null, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "randstr", Value.integerTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "round", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "sign", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "sin", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "sinh", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "sqrt", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "srand", Value.integerTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "tan", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "tanh", Value.floatTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "timespan", Value.allTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "truncate", Value.nonIntegralTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "typeof", Value.allTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddFunction(
                            null, "wide", Value.numberTypes, FunctionFlags.None,
                            plugin, ClientData.Empty, ref result);
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode SetupOperators(
            CreateFlags createFlags,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(createFlags, CreateFlags.NoOperators, true))
            {
                IPlugin plugin = GetCorePlugin(ref result);

                if (plugin != null)
                {
                    if (code == ReturnCode.Ok)
                        //
                        // NOTE: *SPECIAL CASE* This operator is handled directly
                        //       by the expression engine; therefore, we specify
                        //       the default operator here (which does nothing).
                        //
                        code = AddOperator(
                            typeof(_Operators.Default), Operators.Question,
                            Lexeme.Question, null, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseAnd), Operators.BitwiseAnd,
                            Lexeme.BitwiseAnd, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseEqv), Operators.BitwiseEqv,
                            Lexeme.BitwiseEqv, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseImp), Operators.BitwiseImp,
                            Lexeme.BitwiseImp, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseNot), Operators.BitwiseNot,
                            Lexeme.BitwiseNot, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseOr), Operators.BitwiseOr,
                            Lexeme.BitwiseOr, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.BitwiseXor), Operators.BitwiseXor,
                            Lexeme.BitwiseXor, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Divide), Operators.Divide,
                            Lexeme.Divide, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Equal), Operators.Equal,
                            Lexeme.Equal, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Exponent), Operators.Exponent,
                            Lexeme.Exponent, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.GreaterThan), Operators.GreaterThan,
                            Lexeme.GreaterThan, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.GreaterThanOrEqualTo), Operators.GreaterThanOrEqualTo,
                            Lexeme.GreaterThanOrEqualTo, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LeftRotate), Operators.LeftRotate,
                            Lexeme.LeftRotate, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LeftShift), Operators.LeftShift,
                            Lexeme.LeftShift, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LessThan), Operators.LessThan,
                            Lexeme.LessThan, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LessThanOrEqualTo), Operators.LessThanOrEqualTo,
                            Lexeme.LessThanOrEqualTo, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.ListIn), Operators.ListIn,
                            Lexeme.ListIn, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.ListNotIn), Operators.ListNotIn,
                            Lexeme.ListNotIn, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalAnd), Operators.LogicalAnd,
                            Lexeme.LogicalAnd, Value.numberTypes, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalEqv), Operators.LogicalEqv,
                            Lexeme.LogicalEqv, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalImp), Operators.LogicalImp,
                            Lexeme.LogicalImp, Value.numberTypes, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalNot), Operators.LogicalNot,
                            Lexeme.LogicalNot, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalOr), Operators.LogicalOr,
                            Lexeme.LogicalOr, Value.numberTypes, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.LogicalXor), Operators.LogicalXor,
                            Lexeme.LogicalXor, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Minus), Operators.Minus,
                            Lexeme.Minus, Value.numberTypes, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Modulus), Operators.Modulus,
                            Lexeme.Modulus, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Multiply), Operators.Multiply,
                            Lexeme.Multiply, Value.numberTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.NotEqual), Operators.NotEqual,
                            Lexeme.NotEqual, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.Plus), Operators.Plus,
                            Lexeme.Plus, Value.numberTypes, OperatorFlags.Special,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.RightRotate), Operators.RightRotate,
                            Lexeme.RightRotate, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.RightShift), Operators.RightShift,
                            Lexeme.RightShift, Value.integralTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.StringEqual), Operators.StringEqual,
                            Lexeme.StringEqual, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);

                    if (code == ReturnCode.Ok)
                        code = AddOperator(
                            typeof(_Operators.StringNotEqual), Operators.StringNotEqual,
                            Lexeme.StringNotEqual, Value.allTypes, OperatorFlags.None,
                            plugin, ClientData.Empty, ref result);
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode SetupHostChannels(
            Interpreter interpreter,
            IStreamHost streamHost,
            ref Result error
            )
        {
            if (interpreter != null)
            {
                return interpreter.AddStandardChannels(
                    streamHost, null, ChannelType.All, false, false, ref error);
            }
            else
            {
                error = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode Setup(
            CreateFlags createFlags,
            IEnumerable<string> args,
            ExecuteCallbackDictionary policies,
            TraceList traces,
            string text,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Standard Channels
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoHost, true) &&
                    !FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) &&
                    !FlagOps.HasFlags(createFlags, CreateFlags.NoChannels, true))
                {
                    code = AddStandardChannels(
                        host, null, ChannelType.All, false, false, ref result);
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Core Plugins
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoPlugins, true))
                {
                    IPlugin plugin = null;

                    if ((code == ReturnCode.Ok) &&
                        !FlagOps.HasFlags(createFlags, CreateFlags.NoCorePlugin, true))
                    {
                        if (code == ReturnCode.Ok)
                        {
                            plugin = NewCorePlugin();

                            if (plugin != null)
                            {
                                Result localResult = null;

                                code = RuntimeOps.PrepareStaticPlugin(plugin, ref localResult);

                                if (code == ReturnCode.Ok)
                                    code = AddPlugin(plugin, null, ref corePluginToken, ref localResult);

                                if (code != ReturnCode.Ok)
                                    result = localResult;
                            }
                            else
                            {
                                result = "invalid core plugin";
                                code = ReturnCode.Error;
                            }
                        }
                    }

#if NOTIFY || NOTIFY_OBJECT
                    if ((code == ReturnCode.Ok) &&
                        !FlagOps.HasFlags(createFlags, CreateFlags.NoObjectPlugin, true))
                    {
                        plugin = NewObjectPlugin();

                        if (plugin != null)
                        {
                            Result localResult = null;

                            code = RuntimeOps.PrepareStaticPlugin(plugin, ref localResult);

                            if (code == ReturnCode.Ok)
                                code = AddPlugin(plugin, null, ref objectPluginToken, ref localResult);

                            if (code != ReturnCode.Ok)
                                result = localResult;
                        }
                        else
                        {
                            result = "invalid object plugin";
                            code = ReturnCode.Error;
                        }
                    }
#endif

#if DEBUG
                    if ((code == ReturnCode.Ok) &&
                        !FlagOps.HasFlags(createFlags, CreateFlags.NoTestPlugin, true))
                    {
                        plugin = NewTestPlugin();

                        if (plugin != null)
                        {
                            Result localResult = null;

                            code = RuntimeOps.PrepareStaticPlugin(plugin, ref localResult);

                            if (code == ReturnCode.Ok)
                                code = AddPlugin(plugin, null, ref testPluginToken, ref localResult);

                            if (code != ReturnCode.Ok)
                                result = localResult;
                        }
                        else
                        {
                            result = "invalid test plugin";
                            code = ReturnCode.Error;
                        }
                    }
#endif
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Core command policies (for safe interpreters).
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                IPlugin plugin = GetCorePlugin(ref result);

                if (plugin != null)
                {
                    //
                    // NOTE: For "safe" interpreters, setup the interpreter policies now.  Policies
                    //       provided by the caller are added first followed by the policies needed
                    //       for proper operation of various features (e.g. such as script library
                    //       initialization).
                    //
                    if (policies != null)
                        code = SetupPolicies(policies, null, null, ref result);

                    if ((code == ReturnCode.Ok) &&
                        !FlagOps.HasFlags(createFlags, CreateFlags.NoCorePolicies, true) &&
                        (PolicyOps.Policies != null))
                    {
                        code = SetupPolicies(PolicyOps.Policies, plugin, null, ref result);
                    }
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Pre-defined Trace Lists
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                IPlugin plugin = GetCorePlugin(ref result);

                if (plugin != null)
                {
                    autoPathTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.None, plugin,
                        new TraceCallback[] { AutoPathTraceCallback });

                    environmentTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.None, plugin,
                        new TraceCallback[] { EnvironmentTraceCallback });

                    linkedVariableTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.None, plugin,
                        new TraceCallback[] { LinkedVariableTraceCallback });

                    objectTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.Global, plugin,
                        new TraceCallback[] { ObjectTraceCallback });

                    precisionTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.None, plugin,
                        new TraceCallback[] { PrecisionTraceCallback });

                    testsTraceList = new TraceList(
                        ClientData.Empty, TraceFlags.None, plugin,
                        new TraceCallback[] { TestsTraceCallback });

                    //
                    // NOTE: Setup interpreter wide variable traces now.  Traces provided by the
                    //       caller are added first followed by traces needed for proper operation
                    //       of various features (e.g. the core opaque object handle reference
                    //       counting).
                    //
                    if (traces != null)
                        code = SetupTraces(traces, null, ref result);

                    if ((code == ReturnCode.Ok) &&
                        !FlagOps.HasFlags(createFlags, CreateFlags.NoCoreTraces, true) &&
                        (objectTraceList != null))
                    {
                        code = SetupTraces(objectTraceList, null, ref result);
                    }
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Pre-defined Global Variables
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
                code = SetupVariables(createFlags, args, true, ref result);

            if (code == ReturnCode.Ok)
                code = SetupPlatform(createFlags, true, ref result);

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Pre-defined Environment Variables
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
                code = SetupHome(createFlags, ref result);

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Update script-level interactive status.
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                this.Interactive = false;
                this.InteractiveLoops = 0;
                this.Precision = TclVars.PrecisionValue;
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Predefined Objects
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
            {
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoObjects, true))
                {
                    if (code == ReturnCode.Ok)
                        code = AddObject(Vars.Null, null,
                            ObjectFlags.Locked | ObjectFlags.Safe,
                            ClientData.Empty, 0, null, null,
                            ref result);

                    if (code == ReturnCode.Ok)
                        code = SetVariableValue(
                            VariableFlags.System | VariableFlags.GlobalOnly,
                            Vars.Null, Vars.Null, null, ref result);
                }
            }

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Expression Functions
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
                code = SetupFunctions(createFlags, ref result);

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Expression Operators
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if (code == ReturnCode.Ok)
                code = SetupOperators(createFlags, ref result);

            ///////////////////////////////////////////////////////////////////////////////////////////
            //
            // Evaluate the "pre-init" script, if any...
            //
            ///////////////////////////////////////////////////////////////////////////////////////////

            if ((code == ReturnCode.Ok) && (text != null))
                code = EvaluateScript(text, ref result); /* EXEMPT */

            ///////////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY && NOTIFY_GLOBAL
            if ((code == ReturnCode.Ok) && this.GlobalNotify)
                /* IGNORED */
                CheckNotifications(
                    null, false,
                    NotifyType.Interpreter, NotifyFlags.Setup,
                    new ObjectPair(createFlags, args), this,
                    null, null, null, ref result);
#endif

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode SetupHome(
            CreateFlags createFlags,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) &&
                !FlagOps.HasFlags(createFlags, CreateFlags.NoHome, true))
            {
                if (code == ReturnCode.Ok)
                {
                    string home = PathOps.GetHomeDirectory();

                    if (home == null)
                    {
                        //
                        // NOTE: The HOME environment variable was not found.  We need
                        //       to add it.  Start by creating an empty string.
                        //
                        home = String.Empty;

                        //
                        // NOTE: Check for the HOMEDRIVE environment variable and append
                        //       the value, if any, to the HOME value we are going to set.
                        //
                        string homeDrive = GlobalState.GetEnvironmentVariable(
                            EnvVars.HomeDrive, true, false);

                        if (homeDrive != null)
                            home += homeDrive;

                        //
                        // NOTE: Check for the HOMEPATH environment variable and append
                        //       the value, if any, to the HOME value we are going to set.
                        //
                        string homePath = GlobalState.GetEnvironmentVariable(
                            EnvVars.HomePath, true, false);

                        if (homePath != null)
                            home += homePath;

                        //
                        // NOTE: If the newly constructed candidate value for the HOME
                        //       environment variable has a value, set it now.
                        //
                        if (!String.IsNullOrEmpty(home))
                            PathOps.SetHomeDirectory(home);
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode SetupPlatform(
            CreateFlags createFlags,
            bool create,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true) &&
                !FlagOps.HasFlags(createFlags, CreateFlags.NoPlatform, true))
            {
                SetupSafeVariableNames();

                if (code == ReturnCode.Ok)
                {
                    string byteOrder = BitConverter.IsLittleEndian ?
                        TclVars.Platform.LittleEndianValue :
                        TclVars.Platform.BigEndianValue;

                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.ByteOrder, byteOrder, null, ref result);
                }

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.CharacterSize, sizeof(char).ToString() +
                            Characters.MinusSign + sizeof(char).ToString(),
                        null, ref result);

#if DEBUG
                //
                // NOTE: When built with DEBUG, we are always debug-enabled.
                //
                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.Debug, ConversionOps.ToInt(true).ToString(),
                        null, ref result);
#endif

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.DirectorySeparator,
                        Path.DirectorySeparatorChar.ToString(), null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.Engine, Vars.PackageName, null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.PatchLevel, TclVars.PatchLevelValue,
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.PathSeparator, Path.PathSeparator.ToString(),
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.PlatformName, PlatformOps.Platform,
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.PointerSize, IntPtr.Size.ToString(),
                        null, ref result);

                if (code == ReturnCode.Ok) /* we are always threads-enabled */
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.Threaded, ConversionOps.ToInt(true).ToString(),
                        null, ref result);

                if (code == ReturnCode.Ok) /* we are always unicode-enabled */
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.Unicode, ConversionOps.ToInt(true).ToString(),
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.Version, TclVars.VersionValue,
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, TclVars.Platform.Name,
                        TclVars.Platform.WordSize, sizeof(int).ToString(),
                        null, ref result);

                Version version = GlobalState.GetAssemblyVersion();

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, Vars.Platform.Name,
                        Vars.Platform.Version, FormatOps.MajorMinor(version),
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, Vars.Platform.Name,
                        Vars.Platform.PatchLevel, (version != null) ?
                            version.ToString() : null,
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, Vars.Platform.Name,
                        Vars.Platform.Engine, Vars.PackageName,
                        null, ref result);

                if (code == ReturnCode.Ok)
                    code = SetVariableValue2(
                        VariableFlags.GlobalOnly, Vars.Platform.Name,
                        Vars.Platform.Configuration,
                        AttributeOps.GetAssemblyConfiguration(GlobalState.GetAssembly()),
                        null, ref result);

                if (code == ReturnCode.Ok)
                {
                    if (!FlagOps.HasFlags(createFlags, CreateFlags.Safe, true))
                    {
                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.Processors,
                                Environment.ProcessorCount.ToString(),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.Host, Environment.MachineName,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.Machine, PlatformOps.Processor,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.OsName, PlatformOps.OperatingSystem,
                                null, ref result);

                        /* REUSED */
                        version = CommonOps.Runtime.GetOperatingSystemVersion();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.OsVersion, FormatOps.MajorMinor(version),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.OsPatchLevel,
                                (version != null) ? version.ToString() : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.OsServicePack,
                                PlatformOps.GetServicePackString(),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.OsBits,
                                PlatformOps.GetOperatingSystemBits().ToString(),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, TclVars.Platform.Name,
                                TclVars.Platform.User, Environment.UserName,
                                null, ref result);

#if CAS_POLICY
                        if (code == ReturnCode.Ok)
                        {
                            Result permissionSetResult = null;

                            try
                            {
                                Evidence evidence = GlobalState.GetAssemblyEvidence();

                                PermissionSet permissionSet = (evidence != null) ?
                                    SecurityManager.ResolvePolicy(evidence) : null;

                                permissionSetResult = (permissionSet != null) ?
                                    StringOps.NormalizeWhiteSpace(
                                        permissionSet.ToString(), Characters.Space,
                                        true, false) : null;
                            }
                            catch (Exception e)
                            {
                                permissionSetResult = e;
                            }

                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.PermissionSet, permissionSetResult,
                                null, ref result);
                        }
#endif

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.SourceId,
                                AttributeOps.GetAssemblySourceId(GlobalState.GetAssembly()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.SourceTimeStamp,
                                AttributeOps.GetAssemblySourceTimeStamp(GlobalState.GetAssembly()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Administrator,
                                RuntimeOps.IsAdministrator().ToString(),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Culture,
                                FormatOps.CultureName(GlobalState.GetAssemblyCultureInfo(), false),
                                null, ref result);

                        /* REUSED */
                        version = CommonOps.Runtime.GetFrameworkVersion();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.FrameworkVersion,
                                (version != null) ? version.ToString() : null,
                                null, ref result);

                        OperatingSystem operatingSystem =
                            CommonOps.Runtime.GetOperatingSystem();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.OsName, (operatingSystem != null) ?
                                    operatingSystem.ToString() : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Wow64,
                                PlatformOps.Win32onWin64.ToString(),
                                null, ref result);

                        /* REUSED */
                        version = CommonOps.Runtime.GetRuntimeVersion();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.RuntimeVersion,
                                (version != null) ? version.ToString() : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.ImageRuntimeVersion,
                                AssemblyOps.GetImageRuntimeVersion(
                                    GlobalState.GetAssembly()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Epoch, FormatOps.Iso8601DateTime(
                                    TimeOps.UnixEpoch, true),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            Assembly assembly = GlobalState.GetAssembly();

                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.GlobalAssemblyCache, (assembly != null) ?
                                    assembly.GlobalAssemblyCache.ToString() :
                                    false.ToString(),
                                null, ref result);
                        }

                        StringList options = DefineConstants.OptionList;

                        if (options != null)
                        {
                            options = new StringList(options);
                            options.Sort();
                        }

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.CompileOptions,
                                (options != null) ? options.ToString(false) : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.RuntimeOptions, runtimeOptions.ToString(),
                                null, ref result);

                        if (!FlagOps.HasFlags(createFlags, CreateFlags.NoObjectIds, true))
                        {
                            StringPairList objectIds =
                                AttributeOps.GetObjectIds(GlobalState.GetAssembly(), false);

                            if (code == ReturnCode.Ok)
                                code = SetVariableValue2(
                                    VariableFlags.GlobalOnly, Vars.Platform.Name,
                                    Vars.Platform.ObjectIds,
                                    (objectIds != null) ? objectIds.ToString() : null,
                                    null, ref result);
                        }

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.MinimumDate,
                                FormatOps.Iso8601DateTime(DateTime.MinValue, true),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.MaximumDate,
                                FormatOps.Iso8601DateTime(DateTime.MaxValue, true),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.TimeStamp,
                                FormatOps.Iso8601DateTime(
                                    AttributeOps.GetAssemblyDateTime(
                                        GlobalState.GetAssembly()), true),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Release,
                                AttributeOps.GetAssemblyRelease(GlobalState.GetAssembly()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Tag,
                                AttributeOps.GetAssemblyTag(GlobalState.GetAssembly()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Text,
                                AttributeOps.GetAssemblyText(GlobalState.GetAssembly()),
                                null, ref result);

                        Uri uri = GlobalState.GetAssemblyUri();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Uri,
                                (uri != null) ? uri.ToString() : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.PublicKey,
                                AssemblyOps.GetPublicKey(
                                    GlobalState.GetAssemblyName()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.PublicKeyToken,
                                AssemblyOps.GetPublicKeyToken(
                                    GlobalState.GetAssemblyName()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.ModuleVersionId,
                                AssemblyOps.GetModuleVersionId(
                                    GlobalState.GetAssembly()).ToString(),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Certificate, FormatOps.Certificate(
                                    GlobalState.GetAssemblyLocation(), GetCertificate(),
                                    !create || RuntimeOps.ShouldCheckFileTrusted(),
                                    false, false),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.UpdateFileName,
                                Vars.Platform.UpdateFileValue,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.StrongName, FormatOps.StrongName(
                                    GlobalState.GetAssembly(), GetStrongName(),
                                    !create || RuntimeOps.ShouldCheckStrongNameVerified()),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.Hash, FormatOps.Hash(GetHash()),
                                null, ref result);

                        /* REUSED */
                        version = GlobalState.GetEntryAssemblyVersion();

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.ShellVersion,
                                FormatOps.MajorMinor(version),
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.ShellPatchLevel,
                                (version != null) ? version.ToString() : null,
                                null, ref result);

                        if (code == ReturnCode.Ok)
                            code = SetVariableValue2(
                                VariableFlags.GlobalOnly, Vars.Platform.Name,
                                Vars.Platform.RuntimeName,
                                CommonOps.Runtime.GetRuntimeName(),
                                null, ref result);
                    }
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Initialization
        private static ReturnCode EvaluateScript(
            Interpreter interpreter,
            string name,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            string text = null;

            return EvaluateScript(interpreter, name, ref flags, ref text, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode EvaluateScript(
            Interpreter interpreter,
            string name,
            ref ScriptFlags flags,
            ref string text,
            ref Result result
            )
        {
            int errorLine = 0;

            return EvaluateScript(interpreter, name, ref flags, ref text, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: *WARNING* This method is for use with named (i.e.
        //       predefined) script library and host scripts only.
        //       This method may be radically modified and/or
        //       removed in the future.
        //
        private static ReturnCode EvaluateScript(
            Interpreter interpreter,
            string name,
            ref ScriptFlags flags,
            ref string text,
            ref Result result,
            ref int errorLine
            )
        {
            if (interpreter != null)
            {
                try
                {
                    bool library = FlagOps.HasFlags(flags, ScriptFlags.Library, true);
                    Result localResult = null;

                    if ((library && (RuntimeOps.GetLibraryScript(
                            interpreter, interpreter.Host, name,
                            ref flags, ref localResult) == ReturnCode.Ok)) ||
                        (!library && (RuntimeOps.GetHostScript(
                            interpreter, interpreter.Host, name,
                            ref flags, ref localResult) == ReturnCode.Ok)))
                    {
                        text = localResult;

                        if (!String.IsNullOrEmpty(text))
                        {
                            if (FlagOps.HasFlags(flags, ScriptFlags.File, true))
                            {
                                if (PathOps.IsRemoteUri(text) || File.Exists(text))
                                {
                                    return interpreter.EvaluateFile(
                                        text, ref result, ref errorLine);
                                }
                                else
                                {
                                    result = String.Format(
                                        "the provided \"{0}\" script file \"{1}\" is not " +
                                        "a valid remote uri and does not exist locally",
                                        name, text);
                                }
                            }
                            else
                            {
                                return interpreter.EvaluateScript(
                                    text, ref result, ref errorLine);
                            }
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.Required, true))
                        {
                            //
                            // NOTE: The script is required; therefore, this
                            //       is an error.
                            //
                            result = String.Format(
                                "the provided \"{0}\" script is invalid (required)",
                                name);
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.Optional, true))
                        {
                            //
                            // NOTE: Script is explicitly optional, this is OK.
                            //
                            result = String.Empty;
                            return ReturnCode.Ok;
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.ErrorOnEmpty, true))
                        {
                            //
                            // NOTE: The strict flag is set; therefore, this
                            //       is an error.
                            //
                            result = String.Format(
                                "the provided \"{0}\" script is invalid (strict)",
                                name);
                        }
                        else
                        {
                            //
                            // TODO: Script is neither explicitly requierd nor
                            //       optional and the strict flag is not set;
                            //       therefore, just return success.  These
                            //       semantics may need fine-tuning later.
                            //
                            result = String.Empty;
                            return ReturnCode.Ok;
                        }
                    }
                    else if (FlagOps.HasFlags(flags, ScriptFlags.Required, true) ||
                        FlagOps.HasFlags(flags, ScriptFlags.ErrorOnEmpty, true))
                    {
                        //
                        // NOTE: No suitable script could be located to
                        //       satisfy the request.  Return the inner
                        //       error message verbatim.
                        //
                        result = localResult;
                    }
                    else
                    {
                        //
                        // TODO: Script is neither explicitly requierd nor
                        //       optional and the strict flag is not set;
                        //       therefore, just return success.  These
                        //       semantics may need fine-tuning later.
                        //
                        result = String.Empty;
                        return ReturnCode.Ok;
                    }
                }
                catch (Exception e)
                {
                    result = e;
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            //
            // NOTE: If we get to this point, we have failed and
            //       we do not have error line information (i.e.
            //       we have not evaluated anything).
            //
            errorLine = 0;

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode EvaluateStartupScript(
            Interpreter interpreter,
            ref ScriptFlags flags,
            ref Result result
            )
        {
            int errorLine = 0;

            return EvaluateStartupScript(interpreter, ref flags, ref result, ref errorLine);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode EvaluateStartupScript(
            Interpreter interpreter,
            ref ScriptFlags flags,
            ref Result result,
            ref int errorLine
            )
        {
            if (interpreter != null)
            {
                try
                {
                    string name = TclVars.RunCommandsFileName;
                    Result localResult = null;
                    ResultList errors = null;

                    if (RuntimeOps.GetStartupScript(
                            interpreter, interpreter.Host, name, ref flags,
                            ref localResult, ref errors) == ReturnCode.Ok)
                    {
                        string text = localResult;

                        if (!String.IsNullOrEmpty(text))
                        {
                            if (FlagOps.HasFlags(flags, ScriptFlags.File, true))
                            {
                                if (PathOps.IsRemoteUri(text) || File.Exists(text))
                                {
                                    return interpreter.EvaluateFile(
                                        text, ref result, ref errorLine);
                                }
                                else
                                {
                                    result = String.Format(
                                        "the provided \"{0}\" script file \"{1}\" is not " +
                                        "a valid remote uri and does not exist locally",
                                        name, text);
                                }
                            }
                            else
                            {
                                return interpreter.EvaluateScript(
                                    text, ref result, ref errorLine);
                            }
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.Required, true))
                        {
                            //
                            // NOTE: The script is required; therefore, this
                            //       is an error.
                            //
                            result = String.Format(
                                "the provided \"{0}\" script is invalid (required)",
                                name);
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.Optional, true))
                        {
                            //
                            // NOTE: Script is explicitly optional, this is OK.
                            //
                            result = String.Empty;
                            return ReturnCode.Ok;
                        }
                        else if (FlagOps.HasFlags(flags, ScriptFlags.ErrorOnEmpty, true))
                        {
                            //
                            // NOTE: The strict flag is set; therefore, this
                            //       is an error.
                            //
                            result = String.Format(
                                "the provided \"{0}\" script is invalid (strict)",
                                name);
                        }
                        else
                        {
                            //
                            // TODO: Script is neither explicitly requierd nor
                            //       optional and the strict flag is not set;
                            //       therefore, just return success.  These
                            //       semantics may need fine-tuning later.
                            //
                            result = String.Empty;
                            return ReturnCode.Ok;
                        }
                    }
                    else if (FlagOps.HasFlags(flags, ScriptFlags.Required, true) ||
                        FlagOps.HasFlags(flags, ScriptFlags.ErrorOnEmpty, true))
                    {
                        //
                        // NOTE: No suitable script could be located to
                        //       satisfy the request.  Return the inner
                        //       error message verbatim.
                        //
                        result = errors;
                    }
                    else
                    {
                        //
                        // TODO: Script is neither explicitly requierd nor
                        //       optional and the strict flag is not set;
                        //       therefore, just return success.  These
                        //       semantics may need fine-tuning later.
                        //
                        result = String.Empty;
                        return ReturnCode.Ok;
                    }
                }
                catch (Exception e)
                {
                    result = e;
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            //
            // NOTE: If we get to this point, we have failed and
            //       we do not have error line information (i.e.
            //       we have not evaluated anything).
            //
            errorLine = 0;

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateInitialize(
            ref Result error /* out */
            )
        {
            ReturnCode code;

            if (!initialized)
            {
                //
                // NOTE: Debug all the key paths used to initialize the interpreter.
                //       This was created initially to debug compliance with our path
                //       resolution expectations (primarily on Mono).
                //
                if (FlagOps.HasFlags(createFlags, CreateFlags.ShowAutoPath, true))
                    GlobalState.DisplayPaths(this);

                //
                // NOTE: Pretend to be Tcl for the purpose of source compatibility
                //       with package index files.
                //
                Result localResult = null;

                code = ProvidePackage(TclVars.PackageName,
                    TclVars.PackageVersion, ref localResult);

                if (code != ReturnCode.Ok)
                    error = localResult;

                //
                // NOTE: Register the static Eagle package in the interpreter.
                //
                if (code == ReturnCode.Ok)
                {
                    code = ProvidePackage(
                        GlobalState.GetPackageName(false),
                        GlobalState.GetAssemblyVersion(), ref localResult);

                    if (code != ReturnCode.Ok)
                        error = localResult;
                }

                ScriptFlags scriptFlags;

                if (code == ReturnCode.Ok)
                {
                    string text = null;

                    scriptFlags = ScriptFlags.CoreLibraryRequiredFile;

                    code = EvaluateScript(this,
                        FormatOps.ScriptTypeToFileName(
                            ScriptTypes.Initialization, false),
                        ref scriptFlags, ref text, ref localResult);

                    if (code == ReturnCode.Ok)
                    {
                        if (FlagOps.HasFlags(scriptFlags, ScriptFlags.File, true))
                        {
                            //
                            // BUGFIX: Save the directory where we found the
                            //         "init" script.
                            //
                            initializedPath = PathOps.TranslatePath(
                                PathOps.GetDirectoryName(text),
                                PathTranslationType.Default);
                        }
                        else
                        {
                            //
                            // BUGFIX: No file was used to find the "init"
                            //         script; however, we probably pulled
                            //         it from the assembly resources.
                            //
                            Assembly assembly = GlobalState.GetAssembly();

                            initializedPath = (assembly != null) ?
                                PathOps.TranslatePath(assembly.Location,
                                    PathTranslationType.Default) : null;
                        }
                    }
                    else
                    {
                        error = localResult;
                    }
                }

                if (code == ReturnCode.Ok)
                {
                    scriptFlags = ScriptFlags.ApplicationLibraryOptionalFile;

                    code = EvaluateScript(this, FileName.Embedding,
                        ref scriptFlags, ref localResult);

                    if (code != ReturnCode.Ok)
                        error = localResult;
                }

                if (code == ReturnCode.Ok)
                {
                    scriptFlags = ScriptFlags.VendorLibraryOptionalFile;

                    code = EvaluateScript(this, FileName.Vendor,
                        ref scriptFlags, ref localResult);

                    if (code != ReturnCode.Ok)
                        error = localResult;
                }

                if (code == ReturnCode.Ok)
                {
                    if (!FlagOps.HasFlags(createFlags, CreateFlags.Safe, true))
                    {
                        //
                        // NOTE: Trigger the initialization of the package
                        //       "subsystem".  This must be done after the
                        //       library has been initialized; otherwise,
                        //       the pkgIndex scripts will not be able to
                        //       use those facilities.
                        //
                        if (code == ReturnCode.Ok)
                            code = SetVariableValue(
                                VariableFlags.GlobalOnly, TclVars.AutoPath,
                                GlobalState.GetAutoPathList(this, false).ToString(),
                                autoPathTraceList, ref error);

                        //
                        // NOTE: This variable should have been set above.
                        //       It may be null if no physical file was
                        //       used for the "init" script; however, this
                        //       is OK.
                        //
                        if (code == ReturnCode.Ok)
                            code = SetVariableValue(
                                VariableFlags.GlobalOnly, TclVars.Library,
                                initializedPath, null, ref error);
                    }
                }

#if DEBUGGER
                //
                // NOTE: Now, see if we need still need to initialize
                //       isolated interpreter for our debugger (if we
                //       have one).
                //
                if (code == ReturnCode.Ok)
                {
                    IDebugger debugger = Debugger;

                    if (debugger != null)
                    {
                        Interpreter debugInterpreter = debugger.Interpreter;

                        if ((debugInterpreter != null) &&
                            !debugInterpreter.initialized)
                        {
                            /* RECURSION */
                            code = debugInterpreter.PrivateInitialize(
                                ref error);
                        }
                    }
                }
#endif

                //
                // NOTE: Has everything succeeded?
                //
                if (code == ReturnCode.Ok)
                {
                    initialized = true; /* NOTE: It is now done. */

#if NOTIFY && NOTIFY_GLOBAL
                    if (this.GlobalNotify)
                        /* IGNORED */
                        CheckNotifications(
                            null, false,
                            NotifyType.Interpreter, NotifyFlags.Initialized,
                            new ObjectPair(initializedPath), this,
                            null, null, null, ref error);
#endif
                }
            }
            else
            {
                //
                // NOTE: Already initialized, just return.
                //
                code = ReturnCode.Ok;
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: PrivateInitialize: interpreter = {1}, code = {2}, error = {3}{4}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(this), code,
                FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                typeof(Interpreter).Name);

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode PrivateInitializeShell(
            ref Result error /* out */
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (!initializedShell)
            {
                ScriptFlags scriptFlags;
                Result localResult = null;

                if (code == ReturnCode.Ok)
                {
                    scriptFlags = ScriptFlags.Interactive |
                        ScriptFlags.CoreLibraryRequiredFile;

                    code = EvaluateScript(this, FileName.Shell,
                        ref scriptFlags, ref localResult);

                    if (code != ReturnCode.Ok)
                        error = localResult;
                }

                if (code == ReturnCode.Ok)
                {
                    scriptFlags = ScriptFlags.ApplicationOptionalFile |
                        ScriptFlags.UserOptionalFile;

                    code = EvaluateStartupScript(this,
                        ref scriptFlags, ref localResult);

                    if (code != ReturnCode.Ok)
                        error = localResult;
                }

                if (code == ReturnCode.Ok)
                    initializedShell = true;
            }
            else
            {
                //
                // NOTE: Already initialized, just return.
                //
                code = ReturnCode.Ok;
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: PrivateInitializeShell: interpreter = {1}, code = {2}, error = {3}{4}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(this), code,
                FormatOps.WrapOrNull(true, true, error), Environment.NewLine),
                typeof(Interpreter).Name);

            return code;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Debugger Support
#if DEBUGGER
        private bool CanHitWatches()
        {
            return !Engine.HasNoWatch(engineFlags) && !PrivateIsDebuggerActive();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool CanHitBreakpoints()
        {
            return !Engine.HasNoBreakpoint(engineFlags) && !PrivateIsDebuggerActive();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetTestBreakpoints(
            ref Result result
            )
        {
            StringDictionary testBreakpoints = TestBreakpoints;

            if (testBreakpoints == null)
            {
                result = "test breakpoints not available";
                return ReturnCode.Error;
            }

            result = testBreakpoints.ToString(null, false);
            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool HasTestBreakpoint(
            string name
            )
        {
            if (name != null)
            {
                StringDictionary testBreakpoints = TestBreakpoints;

                if (testBreakpoints != null)
                    return testBreakpoints.ContainsKey(name);
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool SetTestBreakpoint(
            string name,
            bool enable,
            ref Result error
            )
        {
            if (name == null)
            {
                error = "invalid test name";
                return false;
            }

            StringDictionary testBreakpoints = TestBreakpoints;

            if (testBreakpoints == null)
            {
                error = "test breakpoints not available";
                return false;
            }

            if (enable == testBreakpoints.ContainsKey(name))
            {
                error = String.Format(
                    "test \"{0}\" breakpoint is already {1}",
                    name, enable ? "set" : "unset");

                return false;
            }

            if (enable)
                testBreakpoints.Add(name, null);
            else
                testBreakpoints.Remove(name);

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CheckBreakpoints(
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            IExecute execute,
            IExecuteArgument executeArgument,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            return Engine.CheckBreakpoints(
                code, breakpointType, breakpointName, token, traceInfo, engineFlags,
                substitutionFlags, engineEventFlags, expressionFlags, execute,
                executeArgument, this, clientData, arguments, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode CheckWatches(
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            return Engine.CheckWatches(
                code, breakpointType, breakpointName, token, traceInfo, engineFlags,
                substitutionFlags, engineEventFlags, expressionFlags, this, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DebuggerBreak(
            IDebugger debugger,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            return _Private.Debugger.Break(
                debugger, this, code, breakpointType, breakpointName, token,
                traceInfo, engineFlags, substitutionFlags, engineEventFlags,
                expressionFlags, headerFlags, clientData, arguments, ref result);
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Shell & Interactive Loop Support
        #region Shell Support
        #region Shell Thread Support
#if SHELL
        private static void ShellThreadStart(
            object obj
            )
        {
            try
            {
                //
                // TODO: Add support for reporting the overall result here.
                //
                /* IGNORED */
                ShellMain(obj as string[]);
            }
            catch
            {
                //
                // TODO: Add logging support.
                //
                // not much we can really do.
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Thread ShellThreadMain(
            string[] args,
            bool start
            )
        {
            try
            {
                Thread shellThread = Engine.CreateThread(
                    ShellThreadStart, 0, true, false);

                if (shellThread != null)
                {
                    shellThread.Name = "shellThread";

                    if (start)
                        shellThread.Start(args);

                    return shellThread;
                }
            }
            catch
            {
                // do nothing.
            }

            return null;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Shell Core Support
#if SHELL
        //
        // NOTE: This method is [also] specifically for use by alternative shells
        //       (i.e. ones that have no console, etc).  This method assumes that
        //       the interpreter has already been created with the desired options
        //       and that the interpreter host has been set to something with
        //       suitable functionality (i.e. minimally, something that fully
        //       implements the IInteractiveHost interface).
        //
        public static ExitCode ShellMainCore(
            Interpreter interpreter,
            ShellArgumentCallback callback,
            IClientData clientData,
            string[] args,
            bool initialize,
            bool loop,
            ref Result result
            )
        {
            IList<string> argv = (args != null) ? new StringList(args) : null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: ShellMainCore: entered, interpreter = {1}, callback = {2}, " +
                "clientData = {3}, argv = {4}, initialize = {5}, loop = {6}, " +
                "result = {7}{8}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull((interpreter != null) ?
                    (object)interpreter.id : null),
                FormatOps.WrapOrNull(FormatOps.DelegateName(callback)),
                FormatOps.WrapOrNull(clientData),
                FormatOps.WrapOrNull(true, true, argv), initialize, loop,
                FormatOps.WrapOrNull(true, true, result), Environment.NewLine),
                typeof(Interpreter).Name);

            ExitCode exitCode;
            bool dispose = false;

            if (interpreter != null)
            {
                //
                // NOTE: We need the return code all over the place in this
                //       method.
                //
                ReturnCode code;

                //
                // NOTE: Fetch the interpreter host now for result reporting
                //       purposes.
                //
                IInteractiveHost interactiveHost = interpreter.Host;

                //
                // NOTE: In debug mode, show the command line
                //       arguments just as we received them.
                //
                if (interpreter.Debug)
                    RuntimeOps.WriteLineHostOrConsole(interactiveHost,
                        String.Format("The command line is: {0}",
                            RuntimeOps.BuildCommandLine(argv, true)));

                //
                // NOTE: If this value is non-zero after a file is
                //       evaluated, we enter the interactive loop
                //       instead of exiting the process.
                //
                bool noExit = false;

                //
                // NOTE: Command line usage information variables.
                //
                bool showOptions = false;
                bool showEnvironment = false;

                //
                // NOTE: This is the encoding to use for reading a script
                //       file.  It can be set using the "-encoding"
                //       command-line option.
                //
                Encoding encoding = null;

            retry: // NOTE: Used to continue processing arguments.
                int argc = (argv != null) ? argv.Count : 0;

                //
                // NOTE: Re-fetch interpreter host in case it has changed
                //       out from under us during argument processing.
                //
                interactiveHost = interpreter.Host;

                if (argc > 0)
                {
                    //
                    // NOTE: Grab the first argument in the array.  The
                    //       array is modified as we process the arguments;
                    //       however, we are guaranteed to have at least
                    //       one by the containing if statement.
                    //
                    string arg = argv[0];

                    //
                    // NOTE: Save the original, unmodified argument for
                    //       later error reporting.
                    //
                    string savedArg = arg;

                option:
                    //
                    // NOTE: This is the number of switch chars in front
                    //       of the current argument.
                    //
                    int count = 0;

                    //
                    // NOTE: This trims any leading switch chars from the
                    //       current argument and sets the count to the
                    //       number of switch chars actually removed.
                    //
                    arg = StringOps.TrimSwitchChars(arg, ref count);

                    if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Arguments))
                    {
                        if (argc >= 2)
                        {
                            //
                            // NOTE: Get the stream reader for the file containing
                            //       the arguments to process.  If the file name is
                            //       "-" or "stdin", we will end up reading arguments
                            //       from the standard input stream.  Currently, this
                            //       is always done via the Console; however, in the
                            //       future it may use the interpreter host.
                            //
                            TextReader textReader = RuntimeOps.GetTextReaderForFile(
                                encoding, argv[1], ref result);

                            if (textReader != null)
                            {
                                //
                                // NOTE: This method pops its own arguments from
                                //       the list.  The literal "2" in the parameter
                                //       list for ReadArgumentsFromFile is the
                                //       number of arguments that need to be popped
                                //       from the front of the list prior to
                                //       appending any arguments found in the
                                //       specified file.
                                //
                                code = RuntimeOps.ReadArgumentsFromFile(
                                    interpreter, textReader, /* POP */ 2,
                                    ref argv, ref result);

                                if (code == ReturnCode.Ok)
                                    goto retry;
                            }
                            else
                            {
                                code = ReturnCode.Error;
                            }

                            if (code != ReturnCode.Ok)
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg, code, result,
                                    ref argv, ref interactiveHost);
                            }

                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <fileName>\"",
                                CommandLineOption.Arguments), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Debug))
                    {
                        //
                        // NOTE: Set the debug flag for the interpreter now.
                        //
                        interpreter.Debug = true;

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Evaluate))
                    {
                        if (initialize)
                            code = interpreter.PrivateInitialize(ref result);
                        else
                            code = ReturnCode.Ok;

                        int errorLine = 0;

                        if (code == ReturnCode.Ok)
                        {
                            for (int argIndex = 1; argIndex < argc; argIndex++)
                            {
                                code = interpreter.EvaluateScript(
                                    argv[argIndex], ref result, ref errorLine);

                                if (code != ReturnCode.Ok)
                                    break;
                            }
                        }

                        if (code != ReturnCode.Ok)
                        {
                            //
                            // BUGFIX: We may have evaluated some code and the host
                            //         may have been changed; grab it again.
                            //
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                errorLine != 0 ? errorLine : interpreter.ErrorLine,
                                true, false, ref argv, ref interactiveHost);
                        }

                        if (interpreter.exitCode == RuntimeOps.SuccessExitCode())
                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        else
                            exitCode = interpreter.exitCode;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Encoding))
                    {
                        if (argc >= 2)
                        {
                            code = GetEncoding(interpreter, argv[1],
                                ref encoding, ref result);

                            if (code == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Ok, remove the first two arguments and
                                //       process the next one.
                                //
                                GenericOps<string>.PopFirstArgument(ref argv);
                                GenericOps<string>.PopFirstArgument(ref argv);

                                goto retry;
                            }
                            else
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg, code, result,
                                    ref argv, ref interactiveHost);
                            }

                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <encodingName>\"",
                                CommandLineOption.Encoding), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.File))
                    {
                        if (argc >= 2)
                        {
                            if (initialize)
                                code = interpreter.PrivateInitialize(ref result);
                            else
                                code = ReturnCode.Ok;

                            if (code == ReturnCode.Ok)
                                code = interpreter.SetArguments(
                                    new StringList(argv, 2), ref result);

                            int errorLine = 0;

                            if (code == ReturnCode.Ok)
                                code = interpreter.EvaluateFile(
                                    encoding, argv[1], ref result, ref errorLine);

                            if (code != ReturnCode.Ok)
                            {
                                //
                                // BUGFIX: We may have evaluated some code and the host
                                //         may have been changed; grab it again.
                                //
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg, code, result,
                                    errorLine != 0 ? errorLine : interpreter.ErrorLine,
                                    true, false, ref argv, ref interactiveHost);
                            }

                            if (interpreter.exitCode == RuntimeOps.SuccessExitCode())
                                exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                            else
                                exitCode = interpreter.exitCode;
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <fileName> [arguments]\"",
                                CommandLineOption.File), ref argv, ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.About))
                    {
                        //
                        // NOTE: Show the version banner and copyright notice
                        //       only.
                        //
                        showOptions = false;
                        showEnvironment = false;

                        //
                        // NOTE: We will use the common help option to
                        //       display the information we want.
                        //
                        arg = Characters.MinusSign + CommandLineOption.Help;

                        //
                        // NOTE: Force a jump to the start of the option
                        //       dispatcher.
                        //
                        goto option;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.CommandHelp))
                    {
                        //
                        // NOTE: Show command line options.
                        //
                        showOptions = true;
                        showEnvironment = false;

                        //
                        // NOTE: We will use the common help option to
                        //       display the information we want.
                        //
                        arg = Characters.MinusSign + CommandLineOption.Help;

                        //
                        // NOTE: Force a jump to the start of the option
                        //       dispatcher.
                        //
                        goto option;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.EnvironmentHelp))
                    {
                        //
                        // NOTE: Show environment variables.
                        //
                        showOptions = false;
                        showEnvironment = true;

                        //
                        // NOTE: We will use the common help option to
                        //       display the information we want.
                        //
                        arg = Characters.MinusSign + CommandLineOption.Help;

                        //
                        // NOTE: Force a jump to the start of the option
                        //       dispatcher.
                        //
                        goto option;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.FullHelp))
                    {
                        //
                        // NOTE: Show command line options and environment
                        //       settings.
                        //
                        showOptions = true;
                        showEnvironment = true;

                        //
                        // NOTE: We will use the common help option to
                        //       display the information we want.
                        //
                        arg = Characters.MinusSign + CommandLineOption.Help;

                        //
                        // NOTE: Force a jump to the start of the option
                        //       dispatcher.
                        //
                        goto option;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Help))
                    {
                        code = HelpOps.WriteUsage(
                            interpreter, null, true, true, showOptions,
                            showEnvironment, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            if ((interactiveHost != null) &&
                                !showOptions && !showEnvironment)
                            {
                                interactiveHost.WriteLine();
                                interactiveHost.WriteLine(String.Format(
                                    "Use the \"-{0}\", \"-{1}\", \"-{2}\", or " +
                                    "\"-{3}\" options for more information.",
                                    CommandLineOption.About,
                                    CommandLineOption.CommandHelp,
                                    CommandLineOption.EnvironmentHelp,
                                    CommandLineOption.FullHelp));
                            }
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Interactive))
                    {
                        //
                        // NOTE: Enable interactive mode for the script debugger.
                        //
                        interpreter.Interactive = true;

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Initialize))
                    {
                        if (initialize)
                            code = interpreter.PrivateInitialize(ref result);
                        else
                            code = ReturnCode.Ok;

                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Ok, remove the first argument and process
                            //       the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            //
                            // BUGFIX: We may have evaluated some code and the host
                            //         may have been changed; grab it again.
                            //
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                interpreter.ErrorLine, true, false,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.NoExit))
                    {
                        //
                        // NOTE: Always enter the interactive loop instead of
                        //       exiting.
                        //
                        noExit = true;

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.NoTrace))
                    {
                        //
                        // NOTE: Disable all tracing to the console, etc.
                        //
                        DebugOps.ClearTraceListeners(
                            interpreter.Debug, true, true);

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Pause))
                    {
                        if (interactiveHost != null)
                        {
                            //
                            // NOTE: Wait for the user to [attach a debugger
                            //       and] press a key.
                            //
                            interactiveHost.WriteLine(String.Format(
                                _Constants.Prompt.Debugger,
                                RuntimeOps.GetCurrentProcessId()));

                            interactiveHost.Pause();

                            //
                            // NOTE: Ok, remove the first argument and process
                            //       the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            RuntimeOps.WriteConsoleOrComplain(
                                ReturnCode.Error, "interpreter host not available");

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.PostInitialize))
                    {
                        if (argc >= 2)
                        {
                            if (interpreter.initialized)
                            {
                                int errorLine = 0;

                                code = interpreter.EvaluateScript(
                                    argv[1], ref result, ref errorLine);

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Ok, remove the first two arguments and
                                    //       process the next one.
                                    //
                                    GenericOps<string>.PopFirstArgument(ref argv);
                                    GenericOps<string>.PopFirstArgument(ref argv);

                                    goto retry;
                                }
                                else
                                {
                                    //
                                    // BUGFIX: We may have evaluated some code and the host
                                    //         may have been changed; grab it again.
                                    //
                                    RuntimeOps.ShellMainCoreError(
                                        interpreter, savedArg, arg, code, result,
                                        errorLine, true, false, ref argv,
                                        ref interactiveHost);
                                }

                                exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                            }
                            else
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg,
                                    "script library is not initialized",
                                    ref argv, ref interactiveHost);

                                exitCode = RuntimeOps.FailureExitCode();
                            }
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <script>\"",
                                CommandLineOption.PostInitialize), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.PreInitialize))
                    {
                        if (argc >= 2)
                        {
                            if (!interpreter.initialized)
                            {
                                int errorLine = 0;

                                code = interpreter.EvaluateScript(
                                    argv[1], ref result, ref errorLine);

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Ok, remove the first two arguments and
                                    //       process the next one.
                                    //
                                    GenericOps<string>.PopFirstArgument(ref argv);
                                    GenericOps<string>.PopFirstArgument(ref argv);

                                    goto retry;
                                }
                                else
                                {
                                    //
                                    // BUGFIX: We may have evaluated some code and the host
                                    //         may have been changed; grab it again.
                                    //
                                    RuntimeOps.ShellMainCoreError(
                                        interpreter, savedArg, arg, code,
                                        result, errorLine, true, false,
                                        ref argv, ref interactiveHost);
                                }

                                exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                            }
                            else
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg,
                                    "script library is already initialized",
                                    ref argv, ref interactiveHost);

                                exitCode = RuntimeOps.FailureExitCode();
                            }
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <script>\"",
                                CommandLineOption.PreInitialize),
                                ref argv, ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Profile))
                    {
                        if (argc >= 2)
                        {
                            _Hosts.Core coreHost = interactiveHost as _Hosts.Core;

                            if (coreHost != null)
                            {
                                //
                                // NOTE: Set the profile to the one specified on the
                                //       command line.
                                //
                                coreHost.Profile = argv[1];

                                //
                                // NOTE: Now, we can grab the dynamically constructed
                                //       host profile file name and use it to reload
                                //       the profile.
                                //
                                string fileName = coreHost.HostProfileFileName;

                                if (_Hosts.Core.LoadHostProfile(interpreter, coreHost,
                                        coreHost.GetType(), fileName, false, false,
                                        ref result))
                                {
                                    //
                                    // NOTE: Ok, remove the first two arguments and
                                    //       process the next one.
                                    //
                                    GenericOps<string>.PopFirstArgument(ref argv);
                                    GenericOps<string>.PopFirstArgument(ref argv);

                                    goto retry;
                                }
                                else
                                {
                                    RuntimeOps.ShellMainCoreError(
                                        interpreter, savedArg, arg, String.Format(
                                        "failed to reload host profile: {0}", result),
                                        ref argv, ref interactiveHost);

                                    exitCode = RuntimeOps.FailureExitCode();
                                }
                            }
                            else
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg,
                                    String.Format(RuntimeOps.HostNoFeatureError,
                                        typeof(_Hosts.Core).Name),
                                    ref argv, ref interactiveHost);

                                exitCode = RuntimeOps.FailureExitCode();
                            }
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <profile>\"",
                                CommandLineOption.Profile), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Reconfigure))
                    {
                        if (argc >= 2)
                        {
#if XML && SERIALIZATION
                            InterpreterSettings interpreterSettings = null;

                            code = InterpreterSettings.LoadFrom(argv[1], true,
                                ref interpreterSettings, ref result);

                            if (code == ReturnCode.Ok)
                            {
                                try
                                {
                                    //
                                    // NOTE: We plan on re-creating the active
                                    //       interpreter (below); therefore, we
                                    //       must dispose the current one first.
                                    //
                                    if (interpreter != null)
                                    {
                                        interpreter.Dispose(); /* throw */
                                        interpreter = null;
                                    }
                                }
                                catch (Exception e)
                                {
                                    result = e;
                                    code = ReturnCode.Error;
                                }
                            }

                            if (code == ReturnCode.Ok)
                            {
                                interpreter = Interpreter.Create(
                                    interpreterSettings, true, ref result);

                                if (interpreter != null)
                                {
                                    //
                                    // NOTE: This method has now re-created the
                                    //       active interpreter; therefore, it
                                    //       must be disposed by this method
                                    //       before exiting.
                                    //
                                    dispose = true;

                                    //
                                    // NOTE: Ok, remove the first two arguments
                                    //       and process the next one.
                                    //
                                    GenericOps<string>.PopFirstArgument(ref argv);
                                    GenericOps<string>.PopFirstArgument(ref argv);

                                    goto retry;
                                }
                                else
                                {
                                    code = ReturnCode.Error;
                                }
                            }
#else
                            result = "not implemented";
                            code = ReturnCode.Error;
#endif

                            if (code != ReturnCode.Ok)
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg, code, result,
                                    ref argv, ref interactiveHost);
                            }

                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <settings>\"",
                                CommandLineOption.Reconfigure), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.RuntimeOption))
                    {
                        if (argc >= 2)
                        {
                            if (interpreter.initialized)
                            {
                                if (interpreter.ChangeRuntimeOption(argv[1]))
                                {
                                    //
                                    // NOTE: Ok, remove the first two arguments and
                                    //       process the next one.
                                    //
                                    GenericOps<string>.PopFirstArgument(ref argv);
                                    GenericOps<string>.PopFirstArgument(ref argv);

                                    goto retry;
                                }
                                else
                                {
                                    //
                                    // BUGFIX: We may have evaluated some code and the host
                                    //         may have been changed; grab it again.
                                    //
                                    RuntimeOps.ShellMainCoreError(
                                        interpreter, savedArg, arg,
                                        "failed to change runtime option",
                                        ref argv, ref interactiveHost);
                                }

                                exitCode = RuntimeOps.FailureExitCode();
                            }
                            else
                            {
                                RuntimeOps.ShellMainCoreError(
                                    interpreter, savedArg, arg,
                                    "script library is not initialized",
                                    ref argv, ref interactiveHost);

                                exitCode = RuntimeOps.FailureExitCode();
                            }
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <optionName>\"",
                                CommandLineOption.RuntimeOption),
                                ref argv, ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Safe))
                    {
                        //
                        // NOTE: Make the interpreter "safe".
                        //
                        code = interpreter.MakeSafe(true, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Ok, remove the first argument and process
                            //       the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Standard))
                    {
                        //
                        // NOTE: Make the interpreter "standard".
                        //
                        code = interpreter.MakeStandard(true, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Ok, remove the first argument and process
                            //       the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Step))
                    {
#if DEBUGGER
                        //
                        // NOTE: Enable single-step mode for the script debugger.
                        //
                        interpreter.SingleStep = true;
#else
                        //
                        // NOTE: Issue a warning because the script debugger is
                        //       not available.
                        //
                        if (interactiveHost != null)
                            interactiveHost.WriteLine(_Constants.Prompt.NoDebugger);
#endif

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        (StringOps.MatchSwitch(arg, CommandLineOption.Test) ||
                        StringOps.MatchSwitch(arg, CommandLineOption.PluginTest)))
                    {
                        if (initialize)
                            code = interpreter.PrivateInitialize(ref result);
                        else
                            code = ReturnCode.Ok;

                        if (code == ReturnCode.Ok)
                            code = interpreter.SetArguments(
                                new StringList(argv, 3), ref result);

                        bool all = false;

                        if ((code == ReturnCode.Ok) && (argc >= 3))
                            code = Value.GetBoolean2(argv[2], ValueFlags.AnyBoolean, false,
                                interpreter.CultureInfo, ref all, ref result);

                        int errorLine = 0;

                        if (code == ReturnCode.Ok)
                            code = TestOps.ShellMain(
                                interpreter, (argc >= 2) ? argv[1] : null,
                                interpreter.EngineFlags, interpreter.SubstitutionFlags,
                                interpreter.EngineEventFlags, interpreter.ExpressionFlags,
                                StringOps.MatchSwitch(arg, CommandLineOption.PluginTest),
                                all, ref result, ref errorLine);

                        if (code != ReturnCode.Ok)
                        {
                            //
                            // BUGFIX: We may have evaluated some code and the host
                            //         may have been changed; grab it again.
                            //
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                errorLine != 0 ? errorLine : interpreter.ErrorLine,
                                true, false, ref argv, ref interactiveHost);
                        }

                        if (interpreter.exitCode == RuntimeOps.SuccessExitCode())
                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        else
                            exitCode = interpreter.exitCode;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.TestDirectory))
                    {
                        if (argc >= 2)
                        {
                            interpreter.TestPath = argv[1];

                            //
                            // NOTE: Ok, remove the first two arguments and
                            //       process the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <directory>\"",
                                CommandLineOption.TestDirectory), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Trace))
                    {
                        //
                        // NOTE: Add our standard trace listeners (if necessary).
                        //       This should have no effect if they have already
                        //       been added (i.e. by the environment variable
                        //       being set).
                        //
                        DebugOps.SetupTraceListeners(
                            true, interpreter.Debug, true, true);

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.TraceToHost))
                    {
                        //
                        // NOTE: Set the trace-to-host interpreter flag now.
                        //
                        interpreter.InterpreterFlags |= InterpreterFlags.TraceToHost;

                        //
                        // NOTE: Ok, remove the first argument and process
                        //       the next one.
                        //
                        GenericOps<string>.PopFirstArgument(ref argv);

                        goto retry;
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.VendorPath))
                    {
                        if (argc >= 2)
                        {
                            GlobalState.SetEnvironmentVariable(EnvVars.VendorPath,
                                argv[1]);

                            //
                            // NOTE: Ok, remove the first two arguments and
                            //       process the next one.
                            //
                            GenericOps<string>.PopFirstArgument(ref argv);
                            GenericOps<string>.PopFirstArgument(ref argv);

                            goto retry;
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, String.Format(
                                "wrong # args: should be \"-{0} <path>\"",
                                CommandLineOption.VendorPath), ref argv,
                                ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else if ((count > 0) &&
                        StringOps.MatchSwitch(arg, CommandLineOption.Version))
                    {
                        code = HelpOps.WriteVersion(interpreter, true, true, ref result);

                        if (code != ReturnCode.Ok)
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                    }
                    else if (callback != null)
                    {
                        code = callback(
                            interpreter, interactiveHost, clientData,
                            count, arg, ref argv, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            //
                            // NOTE: Process the next [standard or custom]
                            //       argument now.
                            //
                            goto retry;
                        }
                        else
                        {
                            //
                            // BUGFIX: We may have evaluated some code and the host
                            //         may have been changed; grab it again.
                            //
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);

                            exitCode = RuntimeOps.FailureExitCode();
                        }
                    }
                    else
                    {
                        code = HelpOps.WriteUsage(interpreter, String.Format(
                            "invalid argument: \"{0}\"", savedArg), true, true,
                            true, false, ref result);

                        if (code != ReturnCode.Ok)
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, savedArg, arg, code, result,
                                ref argv, ref interactiveHost);
                        }

                        exitCode = RuntimeOps.FailureExitCode();
                    }

                    //
                    // NOTE: If requested, force entry into interactive loop.
                    //
                    if (noExit)
                    {
                        //
                        // NOTE: It is necessary to invalidate the command line
                        //       arguments here; otherwise, the interactive loop
                        //       will not be entered.
                        //
                        argv = null;
                        argc = 0;

                        goto retry;
                    }
                }
                else if (loop)
                {
                    if (initialize)
                        code = interpreter.PrivateInitialize(ref result);
                    else
                        code = ReturnCode.Ok;

                    if (code == ReturnCode.Ok)
                    {
                        //
                        // NOTE: Always use the original command line arguments
                        //       (as passed by the caller) here; otherwise, we
                        //       should always end up with a null argument list
                        //       because we cannot get to this block of code
                        //       unless there are no arguments left to process
                        //       (or there never were any to begin with).
                        //
                        code = InteractiveLoop(interpreter, args, ref result);
                    }
                    else
                    {
                        //
                        // BUGFIX: We may have evaluated some code and the host
                        //         may have been changed; grab it again.
                        //
                        RuntimeOps.ShellMainCoreError(
                            interpreter, null, null, code, result,
                            interpreter.ErrorLine, false, false, ref argv,
                            ref interactiveHost);
                    }

                    exitCode = interpreter.ExitCode;
                }
                else
                {
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: ShellMainCore: interactive loop skipped{1}",
                        GlobalState.GetCurrentThreadId(), Environment.NewLine),
                        typeof(Interpreter).Name);

                    exitCode = interpreter.ExitCode;
                }
            }
            else
            {
                result = "invalid interpreter";

                //
                // NOTE: Interpreter is invalid and we have no host available.
                //
                RuntimeOps.WriteConsoleOrComplain(ReturnCode.Error, result);

                exitCode = RuntimeOps.FailureExitCode();
            }

            if (dispose)
            {
                ReturnCode disposeCode;
                Result disposeError = null;

                disposeCode = RuntimeOps.TryDisposeObject(interpreter,
                    ref disposeError);

                if (disposeCode != ReturnCode.Ok)
                    DebugOps.Complain(interpreter, disposeCode, disposeError);
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: ShellMainCore: exited, interpreter = {1}, callback = {2}, " +
                "clientData = {3}, argv = {4}, initialize = {5}, loop = {6}, " +
                "result = {7}, exitCode = {8}{9}", GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull((interpreter != null) ?
                    (object)interpreter.id : null),
                FormatOps.WrapOrNull(FormatOps.DelegateName(callback)),
                FormatOps.WrapOrNull(clientData), FormatOps.WrapOrNull(true, true, argv),
                initialize, loop, FormatOps.WrapOrNull(true, true, result), exitCode,
                Environment.NewLine), typeof(Interpreter).Name);

            return exitCode;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Standard Shell Entry Point
#if SHELL
        public static ExitCode ShellMain(string[] args)
        {
            ExitCode exitCode;

            try
            {
                //
                // NOTE: Start with default shell flags.
                //
                CreateFlags createFlags = CreateFlags.ShellUse;

                //
                // NOTE: Get the effective interpreter creation flags for the
                //       shell from the environment, etc.
                //
                createFlags = GetStartupCreateFlags(args, createFlags,
                    OptionOriginFlags.All, true, true);

                //
                // NOTE: We need a result variable now (in case creation of the
                //       interpreter fails).
                //
                Result result = null;

                //
                // NOTE: Create an interpreter now inside of a using block so
                //       that we can be relatively sure it will be finalized on
                //       this thread.
                //
                using (Interpreter interpreter = Create(args, createFlags, ref result))
                {
                    //
                    // NOTE: Make sure the interpreter was actually created.
                    //       This can, in theory, be a problem if the
                    //       ThrowOnError flag ends up getting removed somehow
                    //       prior to the call to create the interpreter.
                    //
                    if (interpreter != null)
                    {
                        //
                        // NOTE: We need the return code very early on now (during
                        //       startup option processing).
                        //
                        ReturnCode code;

                        //
                        // NOTE: Fetch the interpreter host now for result reporting
                        //       purposes.
                        //
                        IInteractiveHost interactiveHost = interpreter.Host;

                        //
                        // NOTE: By default, initialize the script library for the
                        //       interpreter and enter the interactive loop when
                        //       necessary.
                        //
                        bool initialize = true;
                        bool loop = true;

                        //
                        // NOTE: Process all the remaining startup options (i.e.
                        //       the ones that do not modify the interpreter
                        //       creation flags) now.
                        //
                        code = ProcessStartupOptions(interpreter, args, createFlags,
                            OptionOriginFlags.All, true, true, ref initialize,
                            ref loop, ref result);

                        if (code == ReturnCode.Ok)
                        {
                            exitCode = ShellMainCore(
                                interpreter, null, null, args, initialize, loop,
                                ref result);
                        }
                        else
                        {
                            RuntimeOps.ShellMainCoreError(
                                interpreter, null, null, code, result,
                                ref interactiveHost);

                            exitCode = RuntimeOps.ReturnCodeToExitCode(code, true);
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Creation of the interpreter failed.
                        //
                        RuntimeOps.WriteConsoleOrComplain(ReturnCode.Error, result);

                        exitCode = RuntimeOps.FailureExitCode();
                    }
                }
            }
            catch (Exception e)
            {
                //
                // WARNING: Top-level unhandled exception handler.
                //          We should never get here.  If we do then
                //          something has gone really wrong and should
                //          be complained about loudly.
                //
                string message = String.Format("UNHANDLED EXCEPTION: {0}", e);

                //
                // NOTE: We cannot rely on the Interpreter now (nor the
                //       IHost contained within it); therefore, output
                //       directly to the console.
                //
                RuntimeOps.WriteConsoleOrComplain(ReturnCode.Error, message);

                //
                // NOTE: Output to trace listeners.
                //
                DebugOps.TraceWriteLine(message, null);

                //
                // NOTE: Break into the debugger, if one is attached.
                //
                if (DebugOps.IsAttached())
                {
                    DebugOps.Log(message);
                    DebugOps.Break();
                }

                //
                // NOTE: Re-throw the exception if we have been requested
                //       to do so.
                //
                if (GlobalState.GetEnvironmentVariable(EnvVars.Throw, true,
                        false) != null)
                {
                    throw;
                }

                //
                // TODO: The ability to diagnostic send memory dumps here.
                //
                exitCode = RuntimeOps.ExceptionExitCode();
            }

            ///////////////////////////////////////////////////////////////////////////////////////////

            try
            {
                long[] totalMemory = { 0, 0 };

                CommonOps.GetTotalMemory(true,
                    ref totalMemory[0], ref totalMemory[1]);

                RuntimeOps.DebugTrace(String.Format(
                    "{0}: ShellMain: exited, exitCode = {1}, " +
                    "totalMemory[0] = {2}, totalMemory[1] = {3}{4}",
                    GlobalState.GetCurrentThreadId(), exitCode,
                    totalMemory[0], totalMemory[1], Environment.NewLine),
                    typeof(Interpreter).Name);
            }
            catch (Exception e)
            {
                RuntimeOps.DebugTrace(String.Format(
                    "{0}: {1}{2}",
                    GlobalState.GetCurrentThreadId(),
                    e, Environment.NewLine),
                    typeof(Interpreter).Name);
            }

            return exitCode;
        }
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Loop Support
        #region Interactive Exit Support
#if SHELL
        private static void CheckExit(
            Interpreter interpreter,
            ref bool exit
            )
        {
            if (interpreter != null)
            {
                bool newExit = interpreter.Exit;

                if (newExit && !exit)
                    exit = newExit;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Prompt Support
        internal static string GetDefaultPrompt(
            PromptType type,
            PromptFlags flags
            )
        {
            string result = null;

            if (((int)type >= 0) && ((int)type < DefaultPrompts.Count))
            {
                result = DefaultPrompts[(int)type];

                if ((result != null) && FlagOps.HasFlags(flags, PromptFlags.Queue, true))
                    result = QueuePrefix + result;

                if ((result != null) && FlagOps.HasFlags(flags, PromptFlags.Debug, true))
                    result = DebugPrefix + result;
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if SHELL
        private static void Prompt(
            IInteractiveHost interactiveHost,
            HostFlags hostFlags,
            PromptType type,
            bool debug,
            bool queue
            )
        {
            PromptFlags promptFlags = PromptFlags.None;

            if (debug) promptFlags |= PromptFlags.Debug;
            if (queue) promptFlags |= PromptFlags.Queue;

            //
            // NOTE: We require a host to display a custom prompt
            //       (or even to display a default prompt).
            //
            if (interactiveHost != null)
            {
                ReturnCode code;
                Result error = null;

                //
                // NOTE: Request the host to display some a prompt
                //       appropriate for the specified prompt type
                //       and flags.
                //
                try
                {
                    if (FlagOps.HasFlags(hostFlags, HostFlags.Prompt, true))
                    {
                        code = interactiveHost.Prompt(
                            type, ref promptFlags, ref error); /* throw */
                    }
                    else
                    {
                        error = String.Format(
                            RuntimeOps.HostNoFeatureError, HostFlags.Prompt);

                        code = ReturnCode.Error;
                    }
                }
                catch (Exception e)
                {
                    error = e;
                    code = ReturnCode.Error;
                }

                //
                // NOTE: If the host failed to display some type of
                //       prompt, fallback to an appropriate default
                //       prompt now.
                //
                if ((code != ReturnCode.Ok) &&
                    !FlagOps.HasFlags(promptFlags, PromptFlags.Done, true))
                {
                    //
                    // NOTE: Get the appropriate default prompt for this
                    //       situation.
                    //
                    string prompt = GetDefaultPrompt(type, promptFlags);

                    if (prompt != null)
                        interactiveHost.Write(prompt);
                }
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Input Support
#if SHELL
        private static bool ReadInteractiveInput(
            IInteractiveHost interactiveHost,
            ref string value
            )
        {
            if (interactiveHost == null)
                return false;

            return interactiveHost.ReadLine(ref value);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // WARNING: This method is private, very specialized, subject to change
        //          at any time, and is really only designed to be called from
        //          within the InteractiveLoop method.  Therefore, use this
        //          method at your own risk.
        //
        internal static void GetInteractiveInput(
            Interpreter interpreter,              /* in */
            bool trace,                           /* in */
            bool debug,                           /* in */
            bool queue,                           /* in */
            SubstitutionFlags substitutionFlags,  /* in */
            IClientData clientData,               /* in */
            bool forceCancel,                     /* in */
            bool forceHalt,                       /* in */
            ref IInteractiveHost interactiveHost, /* in, out */
            ref bool exit,                        /* in, out */
            ref bool done,                        /* in, out */
            ref bool previous,                    /* out */
            out bool canceled,                    /* out */
            out string text,                      /* out */
            out bool notReady,                    /* out */
            out Result parseError                 /* out */
            )
        {
            //
            // NOTE: Re-fetch the interpreter host in case it has been
            //       externally modified or closed.  Also, make sure it is
            //       still open.  If either of these checks fail or return
            //       false, bail out now.
            //
            bool isOpen = RuntimeOps.IsHostOpen(interpreter,
                ref interactiveHost);

            //
            // BUGFIX: Verify that the interactive host is still valid and
            //         open.  If not, the rest of this method is basically
            //         useless and we are done.
            //
            if (!isOpen)
            {
                done = true; /* NOTE: Invalid or closed host, we are done. */

                if (trace)
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: GetInteractiveInput: leaving, host " +
                        "is already invalid or closed, exit = {1}, " +
                        "interactiveHost = {2}, isOpen = {3}{4}",
                        GlobalState.GetCurrentThreadId(), exit,
                        (interactiveHost != null), isOpen,
                        Environment.NewLine), typeof(Interpreter).Name);

                //
                // NOTE: Reset the variables used by our caller (the main
                //       interactive loop).
                //
                canceled = false;
                text = null;
                notReady = false;
                parseError = null;

                return;
            }

            ReturnCode resetCode;
            Result resetError = null;

            //
            // NOTE: We must reset the script cancellation flags if we can
            //       (i.e. if no evals are pending), because the IsComplete
            //       method now relies upon the Ready method prior to the
            //       Engine.Evaluate* method having been called (which would
            //       normally cooperatively reset the script cancellation
            //       flags prior to checking them), which checks the script
            //       cancellation flags.
            //
            bool resetCancel = false;

            resetCode = Engine.ResetCancel(
                interpreter, forceCancel, ref resetCancel, ref resetError);

            if (resetCode == ReturnCode.Ok)
            {
                if (trace)
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: GetInteractiveInput: cancel flags {1}.{2}",
                        GlobalState.GetCurrentThreadId(), resetCancel ?
                        "reset" : "not reset", Environment.NewLine),
                        typeof(Interpreter).Name);
            }
            else
            {
                if (trace)
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: GetInteractiveInput: failed to reset cancel " +
                        "flags.{1}", GlobalState.GetCurrentThreadId(),
                        Environment.NewLine), typeof(Interpreter).Name);

                DebugOps.Complain(interpreter, resetCode, resetError);
            }

            //
            // NOTE: We must reset the debugger halt flags if we can (i.e.
            //       if no evals are pending and no other interactive loops
            //       are active), because Parser.IsComplete now relies upon
            //       Interpreter.Ready prior to Engine.Evaluate* having been
            //       called (which would normally cooperatively reset the
            //       debugger halt flags prior to checking them), which checks
            //       the debugger halt flags.
            //
            bool resetHalt = false;

            resetCode = Engine.ResetHalt(
                interpreter, forceHalt, ref resetHalt, ref resetError);

            if (resetCode == ReturnCode.Ok)
            {
                if (trace)
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: GetInteractiveInput: halt flags {1}.{2}",
                        GlobalState.GetCurrentThreadId(), resetHalt ?
                        "reset" : "not reset", Environment.NewLine),
                        typeof(Interpreter).Name);
            }
            else
            {
                if (trace)
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: GetInteractiveInput: failed to reset halt " +
                        "flags.{1}", GlobalState.GetCurrentThreadId(),
                        Environment.NewLine), typeof(Interpreter).Name);

                DebugOps.Complain(interpreter, resetCode, resetError);
            }

            //
            // NOTE: Grab the host flags now as we need them to determine if
            //       the prompt can and should [always] be displayed.
            //
            HostFlags hostFlags = interactiveHost.GetFlags();

            //
            // NOTE: Does the host want to force the prompt to be displayed
            //       even if input has been redirected?  If so, just do it.
            //       Otherwise, check if the host input is NOT currently being
            //       redirected (i.e. we currently have an interactive user)
            //       and then display the prompt prior to requesting any input
            //       from the host.
            //
            if (FlagOps.HasFlags(hostFlags, HostFlags.ForcePrompt, true) ||
                !interactiveHost.IsInputRedirected())
            {
                Prompt(interactiveHost,
                    hostFlags, PromptType.Start, debug, queue);
            }

            //
            // NOTE: Check if the debugger is set to enter the interactive loop
            //       upon script cancellation.
            //
            bool breakOnCancel =
                RuntimeOps.GetDebuggerBreakOnCancel(interpreter, false);

            //
            // NOTE: Reset the variables used by our caller (the main
            //       interactive loop).
            //
            canceled = false;
            text = String.Empty; // NOTE: Cannot be null.
            notReady = breakOnCancel; // NOTE: Bypass readiness?
            parseError = null;

            do
            {
                //
                // NOTE: Re-grab the host flags now as they may have changed
                //       (?) and we need them to determine if the prompt can
                //       and should [always] be displayed.
                //
                hostFlags = interactiveHost.GetFlags();

                //
                // NOTE: Was there a parser error just now when we checked if
                //       the input text was a complete command?
                //
                if (parseError != null)
                {
                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: incomplete input, " +
                            "parseError = {1}{2}",
                            GlobalState.GetCurrentThreadId(),
                            FormatOps.WrapOrNull(parseError),
                            Environment.NewLine), typeof(Interpreter).Name);

                    //
                    // NOTE: Does this host support multiple-line input?
                    //
                    if (FlagOps.HasFlags(
                            hostFlags, HostFlags.MultipleLineInput, true))
                    {
                        if (trace)
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: GetInteractiveInput: multiple-line " +
                                "input is supported, setting canceled " +
                                "flag...{1}", GlobalState.GetCurrentThreadId(),
                                Environment.NewLine), typeof(Interpreter).Name);

                        //
                        // NOTE: For hosts that support multiple-line input,
                        //       we will not accept any incomplete input.
                        //       Therefore, bail out now with an "error" (i.e.
                        //       or at least our closest approximation of one).
                        //
                        canceled = true;

                        break;
                    }

                    //
                    // NOTE: In the event of a parser error (i.e. if the line
                    //       entered does not represent a complete command)
                    //       display the error in the host title.
                    //
                    interactiveHost.Title = String.Format(
                        "incomplete: {0}", parseError);

                    // interactiveHost.WriteLine(String.Format(
                    //     "incomplete: {0}", parseError));

                    parseError = null;
                }

                //
                // NOTE: Re-check if the host input is being redirected (i.e.
                //       we do not have an interactive user) then display the
                //       input continuation prompt if necessary prior to
                //       requesting more input from the host.
                //
                bool redirected = interactiveHost.IsInputRedirected();

                if (!String.IsNullOrEmpty(text) && (!redirected ||
                    FlagOps.HasFlags(hostFlags, HostFlags.ForcePrompt, true)))
                {
                    Prompt(interactiveHost,
                        hostFlags, PromptType.Continue, debug, queue);
                }

                //
                // NOTE: Check and see if the debugger has command input
                //       injected via the [debug icommand] sub-command for us
                //       to use.
                //
                string line = RuntimeOps.GetDebuggerCommand(interpreter, false);
                int length; // NOTE: Length of the current input line.

                //
                // NOTE: Are we processing a command injected via the debugger?
                //
                if (!String.IsNullOrEmpty(line))
                {
                    length = line.Length;

                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: injecting line of input " +
                            "from debugger containing {1} characters...{2}",
                            GlobalState.GetCurrentThreadId(), length,
                            Environment.NewLine), typeof(Interpreter).Name);

                    //
                    // NOTE: Prevent this command from being used as the
                    //       previous interactive input and disable the ability
                    //       to use the "again" command.
                    //
                    previous = false;

                    //
                    // NOTE: Simulate the command having been entered via the
                    //       host.
                    //
                    interactiveHost.WriteLine(line);
                }

                //
                // NOTE: Has the request for input been canceled by the host
                //       (i.e. either a null line was returned and no input
                //       redirection is in use OR the ReadLine method returned
                //       false).
                //
                bool cancel = false;

                //
                // NOTE: Get a line of input from the host, if necessary (i.e.
                //       we do not already have a command injected via the
                //       debugger, etc).
                //
                if (!String.IsNullOrEmpty(line) ||
                    ReadInteractiveInput(interactiveHost, ref line))
                {
                    if (line != null) /* more input? */
                    {
                        length = line.Length;

                        if (trace)
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: GetInteractiveInput: read line of " +
                                "input containing {1} characters.{2}",
                                GlobalState.GetCurrentThreadId(), length,
                                Environment.NewLine), typeof(Interpreter).Name);

                        if (length > 0)
                            text += line + Characters.NewLine;
                        else
                            text += Characters.NewLine;

                        if (interpreter != null)
                            interpreter.InteractiveInput = text;
                    }
                    else if (redirected)
                    {
                        if (trace)
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: GetInteractiveInput: hit end of " +
                                "redirected input, setting exit flag...{1}",
                                GlobalState.GetCurrentThreadId(),
                                Environment.NewLine), typeof(Interpreter).Name);

                        //
                        // NOTE: It looks like console input is being redirected
                        //       and we have run out of it; therefore, bail out
                        //       of the interactive loop now.
                        //
                        exit = true;
                    }
                    else
                    {
                        if (trace)
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: GetInteractiveInput: read null line " +
                                "of input, setting cancel flag...{1}",
                                GlobalState.GetCurrentThreadId(),
                                Environment.NewLine), typeof(Interpreter).Name);

                        //
                        // NOTE: The ReadLine method returned true; however,
                        //       the actual input line is null and input is not
                        //       being redirected.  Therefore, assume the user
                        //       has canceled the input request.
                        //
                        cancel = true;
                    }
                }
                else
                {
                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: failed to read line " +
                            "of input, setting cancel flag...{1}",
                            GlobalState.GetCurrentThreadId(),
                            Environment.NewLine), typeof(Interpreter).Name);

                    //
                    // NOTE: The ReadLine method has returned false; therefore,
                    //       assume the input request was canceled by the user.
                    //
                    cancel = true;
                }

                //
                // NOTE: Re-fetch the interpreter host in case it has been
                //       externally modified or closed.  Also, make sure it is
                //       still open.  If either of these checks fail or return
                //       false, bail out now.
                //
                isOpen = RuntimeOps.IsHostOpen(interpreter,
                    ref interactiveHost);

                //
                // NOTE: Make sure that we exit if the [exit] command was used.
                //       This is needed because another thread may have set the
                //       exit flag for the interpreter while we were waiting
                //       for input.
                //
                CheckExit(interpreter, ref exit);

                //
                // NOTE: Did the interpreter exit while we were reading (or was
                //       the host invalidated, perhaps by Dispose)?
                //
                if (exit || !isOpen)
                {
                    done = true; /* NOTE: Async exit, we are done. */

                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: leaving, exit = {1}, " +
                            "interactiveHost = {2}, isOpen = {3}{4}",
                            GlobalState.GetCurrentThreadId(), exit,
                            (interactiveHost != null), isOpen,
                            Environment.NewLine), typeof(Interpreter).Name);

                    break;
                }
                //
                // NOTE: Was input canceled while we were reading?
                //
                else if (cancel)
                {
                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: leaving, input was " +
                            "canceled...{1}", GlobalState.GetCurrentThreadId(),
                            Environment.NewLine), typeof(Interpreter).Name);

                    interactiveHost.WriteLine();
                    canceled = true; /* skip command processing */

                    break;
                }
                else if (String.IsNullOrEmpty(text) || (text.Trim().Length == 0))
                {
                    if (trace)
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: GetInteractiveInput: leaving, null or empty " +
                            "input...{1}", GlobalState.GetCurrentThreadId(),
                            Environment.NewLine), typeof(Interpreter).Name);

                    break;
                }
            } while (String.IsNullOrEmpty(text) || (text.Trim().Length == 0) ||
                     !Parser.IsComplete(interpreter, null, Parser.StartLine,
                            text, 0, Length.Invalid, substitutionFlags,
                            ref notReady, ref parseError) &&
                     !notReady);

            //
            // NOTE: Make sure that we have not exited and that the host is
            //       still valid and open.
            //
            if (!done)
            {
                interactiveHost.Title = null;

                if (notReady)
                {
#if INTERACTIVE_COMMANDS
                    //
                    // NOTE: If this is an interactive debugger command, fake
                    //       the interpreter being "ready" (at least as far as
                    //       input validation goes) because we need to give the
                    //       user the opportunity to reset the script
                    //       cancellation flag, etc.
                    //
                    if (CheckInteractiveCommand(
                            interpreter, text, null, clientData, true, false))
                    {
                        if (trace)
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: GetInteractiveInput: input looks like " +
                                "an interactive command, interpreter readiness " +
                                "faked.{1}", GlobalState.GetCurrentThreadId(),
                                Environment.NewLine), typeof(Interpreter).Name);

                        notReady = false;
                    }
                    else // NOTE: Dangling else (see "if" statement below).
#endif
                    if (!breakOnCancel)
                    {
                        //
                        // NOTE: Show the user why the interpreter is not ready.
                        //
                        interactiveHost.WriteResultLine(
                            ReturnCode.Error, String.Format(
                            "interpreter not ready: {0}", parseError));
                    }
                }
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Command Support
#if SHELL && INTERACTIVE_COMMANDS
        private static bool CheckInteractiveCommand(
            Interpreter interpreter, /* in */
            string text,             /* in */
            string command,          /* in */
            IClientData clientData,  /* in */
            bool usePrefix,          /* in */
            bool extensions          /* in */
            )
        {
            bool verbose = false; /* NOTE: Mask substitution errors. */
            ArgumentList arguments = null;
            ReturnCode code = ReturnCode.Ok;
            Result result = null;

            return CheckInteractiveCommand(interpreter, text, command, clientData,
                usePrefix, ref verbose, ref arguments, ref code, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool CheckInteractiveCommand(
            Interpreter interpreter,    /* in */
            string text,                /* in */
            string command,             /* in */
            IClientData clientData,     /* in */
            bool usePrefix,             /* in */
            ref bool verbose,           /* in, out */
            ref ArgumentList arguments, /* in, out */
            ref ReturnCode code,        /* out */
            ref Result result           /* out */
            )
        {
            if (!String.IsNullOrEmpty(text))
            {
                ReturnCode localCode = ReturnCode.Ok;
                EngineFlags engineFlags = EngineFlags.None;
                EngineFlags interactiveEngineFlags = EngineFlags.None;

                //
                // NOTE: To do anything with substitution, we require a
                //       valid interpreter context.
                //
                if (interpreter != null)
                {
                    //
                    // NOTE: Grab the current engine flags for the
                    //       interpreter.  Also, grab the engine and
                    //       substitution flags specific to interactive
                    //       commands.
                    //
                    engineFlags = interpreter.EngineFlags;

                    interactiveEngineFlags =
                        interpreter.InteractiveEngineFlags;

                    //
                    // NOTE: Only perform substitution if the text looks
                    //       like it contains an interactive command.
                    //
                    if (RuntimeOps.LooksLikeInteractiveCommand(text))
                    {
                        SubstitutionFlags interactiveSubstitutionFlags =
                            interpreter.InteractiveSubstitutionFlags;

                        EventFlags interactiveEventFlags =
                            interpreter.InteractiveEventFlags;

                        ExpressionFlags interactiveExpressionFlags =
                            interpreter.InteractiveExpressionFlags;

                        //
                        // NOTE: Has substitution been disabled by the interpreter
                        //       or via the interactive substitution flags?
                        //
                        if (!Engine.HasNoSubstitute(engineFlags) &&
                            !Engine.HasNoSubstitute(interactiveEngineFlags) &&
                            (interactiveSubstitutionFlags != SubstitutionFlags.None))
                        {
                            Result localResult = null;

                            //
                            // NOTE: Perform the substitions on the text using
                            //       the current flags.
                            //
                            localCode = Engine.SubstituteString(
                                interpreter, text, interactiveEngineFlags,
                                interactiveSubstitutionFlags, interactiveEventFlags,
                                interactiveExpressionFlags, ref localResult);

                            if (localCode == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Success, replace original command text
                                //       with substituted result.
                                //
                                text = localResult;
                            }
                            else if (verbose)
                            {
                                //
                                // NOTE: Only show this error once per actual
                                //       interactive command entered, not per
                                //       call to this method.
                                //
                                verbose = false;

                                //
                                // NOTE: Complain about substitution failure.
                                //
                                DebugOps.Complain(interpreter, localCode, localResult);
                            }
                        }
                    }
                }

                //
                // NOTE: Make sure the above subsitution step succeeded,
                //       if applicable.
                //
                if (localCode == ReturnCode.Ok)
                {
                    //
                    // NOTE: Are we checking for a specific command?
                    //
                    if (!String.IsNullOrEmpty(command))
                    {
                        //
                        // NOTE: Does the text end in a new line character?
                        //       If so, strip it off now.
                        //
                        if (text[text.Length - 1] == Characters.NewLine)
                            text = text.Substring(0, text.Length - 1);

                        //
                        // NOTE: Does the caller want to prefix the command to
                        //       check for with the interactive command prefix?
                        //       This is used so that the caller does not have
                        //       to hard-code the prefix inline in their calls
                        //       to this method.
                        //
                        if (usePrefix)
                        {
                            string prefix = RuntimeOps.InteractiveCommandPrefix;

                            if (!String.IsNullOrEmpty(prefix) && !command.StartsWith(
                                    prefix, StringOps.SystemNoCaseStringComparisonType))
                            {
                                command = prefix + command;
                            }
                        }

                        //
                        // NOTE: Did the caller supply pre-existing arguments?
                        //       If so, check the command name (i.e. the first
                        //       argument) against the command to check for.
                        //
                        if (arguments != null)
                        {
                            if ((arguments.Count > 0) &&
                                (String.Compare(arguments[0], command,
                                    StringOps.SystemNoCaseStringComparisonType) == 0))
                            {
                                //
                                // NOTE: Attempt to handle the interactive command
                                //       as an external interactive command.  Since
                                //       this method was called with the name of a
                                //       pre-existing command to check against, this
                                //       will be an overridden built-in interactive
                                //       command if it exists.
                                //
                                if (!Engine.HasNoInteractiveCommand(engineFlags) &&
                                    !Engine.HasNoInteractiveCommand(interactiveEngineFlags))
                                {
                                    bool executed = ExecuteInteractiveCommand(
                                        interpreter, arguments, clientData,
                                        ref code, ref result);

#if HISTORY
                                    //
                                    // NOTE: Add the command to the history even if
                                    //       it was not executed.  If the command was
                                    //       just executed, it has NOT been recorded
                                    //       yet (because ExecuteInteractiveCommand
                                    //       does not handle command history).  If
                                    //       the command was NOT executed, we assume
                                    //       it will be by the caller since the
                                    //       command name does match the one they
                                    //       specified.
                                    //
                                    if ((interpreter != null) && interpreter.History)
                                        interpreter.AddHistory(arguments,
                                            interpreter.Levels, HistoryFlags.Interactive);
#endif

                                    return !executed;
                                }

#if HISTORY
                                if ((interpreter != null) && interpreter.History)
                                    interpreter.AddHistory(arguments,
                                        interpreter.Levels, HistoryFlags.Interactive);
#endif

                                return true;
                            }
                        }
                        else
                        {
                            //
                            // NOTE: The interactive command, like all other commands,
                            //       must be a well-formed list.  Split it into its
                            //       elements now.
                            //
                            StringList list = null;

                            if (Parser.SplitList(interpreter, text, 0, Length.Invalid,
                                    false, true, ref list) == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Compare the first element of the list (i.e.
                                //       the command name) against the command to
                                //       check for.
                                //
                                if ((list.Count > 0) &&
                                    (String.Compare(list[0], command,
                                        StringOps.SystemNoCaseStringComparisonType) == 0))
                                {
                                    //
                                    // NOTE: Save the caller some work by giving them
                                    //       all the arguments for the interactive
                                    //       command in a list.
                                    //
                                    arguments = new ArgumentList(list, ArgumentFlags.None);

                                    //
                                    // NOTE: Attempt to handle the interactive command
                                    //       as an external interactive command.  Since
                                    //       this method was called with the name of a
                                    //       pre-existing command to check against, this
                                    //       will be an overridden built-in interactive
                                    //       command if it exists.
                                    //
                                    if (!Engine.HasNoInteractiveCommand(engineFlags) &&
                                        !Engine.HasNoInteractiveCommand(interactiveEngineFlags))
                                    {
                                        bool executed = ExecuteInteractiveCommand(
                                            interpreter, arguments, clientData,
                                            ref code, ref result);

#if HISTORY
                                        //
                                        // NOTE: Add the command to the history even if
                                        //       it was not executed.  If the command was
                                        //       just executed, it has NOT been recorded
                                        //       yet (because ExecuteInteractiveCommand
                                        //       does not handle command history).  If
                                        //       the command was NOT executed, we assume
                                        //       it will be by the caller since the
                                        //       command name does match the one they
                                        //       specified.
                                        //
                                        if ((interpreter != null) && interpreter.History)
                                            interpreter.AddHistory(arguments,
                                                interpreter.Levels, HistoryFlags.Interactive);
#endif

                                        return !executed;
                                    }

#if HISTORY
                                    if ((interpreter != null) && interpreter.History)
                                        interpreter.AddHistory(arguments,
                                            interpreter.Levels, HistoryFlags.Interactive);
#endif

                                    return true;
                                }
                            }
                        }
                    }
                    else if (usePrefix)
                    {
                        //
                        // NOTE: Just return non-zero if this looks like an
                        //       interactive command.
                        //
                        return RuntimeOps.LooksLikeInteractiveCommand(text);
                    }
                    else if (!Engine.HasNoInteractiveCommand(engineFlags) &&
                        !Engine.HasNoInteractiveCommand(interactiveEngineFlags) &&
                        RuntimeOps.LooksLikeInteractiveCommand(text))
                    {
                        //
                        // NOTE: Does the text end in a new line character?
                        //       If so, strip it off now.
                        //
                        if (text[text.Length - 1] == Characters.NewLine)
                            text = text.Substring(0, text.Length - 1);

                        //
                        // NOTE: Did the caller supply pre-existing arguments?
                        //       If so, check the command name (i.e. the first
                        //       argument) against the command to check for.
                        //
                        if (arguments != null)
                        {
                            //
                            // NOTE: Attempt to handle the interactive command
                            //       as an external interactive command.  Since
                            //       this method was not called with the name of
                            //       a pre-existing command to check against,
                            //       this will be an interactive extension
                            //       command if it exists.
                            //
                            bool executed = ExecuteInteractiveCommand(
                                interpreter, arguments, clientData,
                                ref code, ref result);

#if HISTORY
                            if (executed && (interpreter != null) && interpreter.History)
                                interpreter.AddHistory(arguments,
                                    interpreter.Levels, HistoryFlags.Interactive);
#endif

                            return executed;
                        }
                        else
                        {
                            //
                            // NOTE: The interactive command, like all other commands,
                            //       must be a well-formed list.  Split it into its
                            //       elements now.
                            //
                            StringList list = null;

                            if (Parser.SplitList(interpreter, text, 0, Length.Invalid,
                                    false, true, ref list) == ReturnCode.Ok)
                            {
                                //
                                // NOTE: Save the caller some work by giving them
                                //       all the arguments for the interactive
                                //       command in a list.
                                //
                                arguments = new ArgumentList(list, ArgumentFlags.None);

                                //
                                // NOTE: Attempt to handle the interactive command
                                //       as an external interactive command.  Since
                                //       this method was not called with the name of
                                //       a pre-existing command to check against,
                                //       this will be an interactive extension
                                //       command if it exists.
                                //
                                bool executed = ExecuteInteractiveCommand(
                                    interpreter, arguments, clientData,
                                    ref code, ref result);

#if HISTORY
                                if (executed && (interpreter != null) && interpreter.History)
                                    interpreter.AddHistory(arguments,
                                        interpreter.Levels, HistoryFlags.Interactive);
#endif

                                return executed;
                            }
                        }
                    }
                }
            }

            //
            // NOTE: No, it is not the command they are checking for OR
            //       we had some failure during the substitution process
            //       and skipped further checking.
            //
            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool ResolveInteractiveCommand(
            Interpreter interpreter,
            ArgumentList arguments,
            EngineFlags engineFlags,
            ref string name,
            ref IExecute execute
            )
        {
            //
            // NOTE: If the interactive command prefix is invalid (null or
            //       zero length) we cannot resolve or execute any
            //       interactive commands.
            //
            string prefix = RuntimeOps.InteractiveCommandPrefix;

            if (!String.IsNullOrEmpty(prefix))
            {
                if (interpreter != null)
                {
                    if ((arguments != null) && (arguments.Count > 0))
                    {
                        //
                        // NOTE: Extract the name of the interactive command to
                        //       resolve and execute.  It must start with the
                        //       interactive command prefix (e.g. "#") in order
                        //       to be properly resolved and executed.
                        //
                        // TODO: Modify this when full [namespace] support has
                        //       been added.
                        //
                        name = StringOps.TrimLeadingColons(arguments[0]);

                        if ((name != null) && name.StartsWith(
                                prefix, StringOps.SystemNoCaseStringComparisonType))
                        {
                            //
                            // NOTE: If the name of the interactive command consists
                            //       only of the interactive command prefix itself
                            //       then try for an exact match; otherwise, partial
                            //       non-ambiguous prefix matching will be used.  See
                            //       if this interactive command has been officially
                            //       registered with the interpreter (i.e. it is an
                            //       interactive extension command, which may also
                            //       shadow an existing built-in interactive command).
                            //       Not finding the interactive command is not an
                            //       error; therefore, the error message is ignored.
                            //
                            if (name.Length <= prefix.Length)
                                engineFlags |= EngineFlags.ExactMatch;

                            if (interpreter.GetIExecuteViaResolvers(engineFlags, name,
                                    arguments, true, true, ref execute) == ReturnCode.Ok)
                            {
                                return true;
                            }
                        }
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool ExecuteInteractiveCommand(
            Interpreter interpreter,
            ArgumentList arguments,
            IClientData clientData,
            ref ReturnCode code,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                EngineFlags engineFlags = interpreter.EngineFlags;

                EngineFlags interactiveEngineFlags =
                    interpreter.InteractiveEngineFlags;

                SubstitutionFlags interactiveSubstitutionFlags =
                    interpreter.InteractiveSubstitutionFlags;

                EventFlags interactiveEventFlags =
                    interpreter.InteractiveEventFlags;

                ExpressionFlags interactiveExpressionFlags =
                    interpreter.InteractiveExpressionFlags;

                if (!Engine.HasNoInteractiveCommand(engineFlags) &&
                    !Engine.HasNoInteractiveCommand(interactiveEngineFlags))
                {
                    string name = null;
                    IExecute execute = null;

                    if (ResolveInteractiveCommand(interpreter, arguments,
                            interactiveEngineFlags, ref name, ref execute))
                    {
                        ICallFrame frame = interpreter.NewLinkedCallFrame(
                            StringList.MakeList("interactive", name), null,
                            CallFrameFlags.Interactive | CallFrameFlags.Tracking);

                        interpreter.PushCallFrame(frame);

                        //
                        // NOTE: Save the current engine flags and then enable
                        //       the external execution flags.
                        //
                        EngineFlags savedEngineFlags = Engine.AddStackCheckFlags(
                            ref interactiveEngineFlags);

                        //
                        // NOTE: Execute the command using the interactive engine
                        //       and substitution flags with the interactive
                        //       engine flags having been modified to include the
                        //       flags necessary for external command execution
                        //       (i.e. command execution outside of the engine).
                        //
                        code = Engine.Execute(
                            name, execute, interpreter, clientData, arguments,
                            interactiveEngineFlags, interactiveSubstitutionFlags,
                            interactiveEventFlags, interactiveExpressionFlags,
                            ref result);

                        //
                        // NOTE: Restore the saved engine flags, masking off the
                        //       external execution flags as necessary.
                        //
                        Engine.RemoveStackCheckFlags(savedEngineFlags,
                            ref interactiveEngineFlags);

                        //
                        // NOTE: Pop the original call frame that we pushed above
                        //       and any intervening scope call frames that may be
                        //       leftover (i.e. they were not explicitly closed).
                        //
                        interpreter.PopCallFrame(CallFrameFlags.Scope, true);

                        //
                        // NOTE: Yes, we just executed an interactive extension
                        //       command; therefore, prevent the default handling
                        //       of the built-in interactive command, if any.
                        //
                        return true;
                    }
                }
            }

            //
            // NOTE: We did nothing.  Either the interactive (extension?)
            //       command was not found OR we could not search for it
            //       because the caller provided us invalid arguments.
            //       Therefore, continue default processing.  This will
            //       involve executing the built-in behavior for the
            //       interactive command or doing nothing if it is not
            //       found by the built-in interactive command processor.
            //
            return false;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Script Cancellation Support
#if CONSOLE
        internal static void ConsoleCancelEventHandler(
            object sender,
            ConsoleCancelEventArgs e
            ) /* THREAD-SAFE */
        {
            //
            // NOTE: Cancel all currently running scripts in all threads.
            //       Only handle the first event generated by the Ctrl-C.
            //
            if ((e == null) || !e.Cancel)
            {
                //
                // NOTE: Obtain the master list of interpreters.  This requires
                //       holding the master lock; therefore, do it first.
                //
                InterpreterDictionary interpreters = GlobalState.GetInterpreters();

                if (interpreters != null)
                {
                    foreach (KeyValuePair<string, Interpreter> pair in interpreters)
                    {
                        Interpreter interpreter = pair.Value;

                        if (interpreter != null)
                        {
                            ReturnCode code;
                            Result error = null;

                            code = interpreter.CancelAnyEvaluate(true, false,
                                demandCancelResultFormat, ref error);

                            if (code != ReturnCode.Ok)
                                DebugOps.Complain(interpreter, code, error);
                        }
                    }
                }

                //
                // NOTE: Finally, prevent default event handling (which would
                //       terminate the process); however, we cannot prevent
                //       Ctrl-Break from terminating the process (only Ctrl-C).
                //
                if ((e != null) && (e.SpecialKey == ConsoleSpecialKey.ControlC))
                    e.Cancel = true;
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static bool HasCancelThread(
            Interpreter interpreter
            )
        {
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    return (interpreter.cancelThread != null);
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static void CancelThreadStart(
            object obj
            ) /* THREAD-SAFE */
        {
            try
            {
                Interpreter interpreter = obj as Interpreter;

                if (interpreter != null)
                {
                    int timeout;

                    //
                    // HACK: Workaround for Mono disposal race condition.
                    //
                    lock (interpreter.SyncRoot)
                    {
                        timeout = interpreter.timeout; /* private */
                    }

                    if (timeout != _Timeout.Infinite) // never timeout?
                    {
                        ReturnCode code;
                        Exception exception = null;
                        Result error = null;

                        //
                        // NOTE: Give the script some time to execute.  If this
                        //       thread is interrupted during the wait, do not
                        //       attempt to do anything else.
                        //
                        code = RuntimeOps.Sleep(timeout, ref exception, ref error);

                        if (code == ReturnCode.Ok)
                        {
                            if (WindowOps.YesOrNo(
                                    String.Format(timeoutCancelPromptFormat, timeout),
                                    GlobalState.GetPackageName(false), true))
                            {
                                code = interpreter.CancelAnyEvaluate(true, false,
                                    String.Format(timeoutCancelResultFormat, timeout),
                                    ref error);

                                if (code != ReturnCode.Ok)
                                    DebugOps.Complain(interpreter, code, error);
                            }
                        }
                        else if (!(exception is ThreadInterruptedException))
                        {
                            DebugOps.Complain(interpreter, code, error);
                        }
                    }
                }
            }
            catch (ThreadInterruptedException)
            {
                // do nothing.
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static ReturnCode StartCancelThread(
            Interpreter interpreter,
            bool force,
            bool strict,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    //
                    // NOTE: If there is not timeout we do not bother starting the
                    //       dedicated script cancellation thread unless we are
                    //       forced to by the caller.
                    //
                    if (force ||
                        (interpreter.timeout != _Timeout.Infinite))
                    {
                        Thread thread = interpreter.cancelThread;

                        if (thread == null)
                        {
                            try
                            {
                                thread = Engine.CreateThread(
                                    interpreter, CancelThreadStart, 0, false,
                                    false);

                                if (thread != null)
                                {
                                    thread.Name = String.Format(
                                        "cancelThread: {0}", interpreter);

                                    thread.Start(interpreter);

                                    interpreter.cancelThread = thread;
                                    result = "created and started cancel thread";

                                    return ReturnCode.Ok;
                                }
                                else
                                {
                                    result = "could not create cancel thread";
                                }
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "failed to start cancel thread, caught exception \"{0}\"",
                                    e);
                            }
                        }
                        else
                        {
                            result = "cancel thread already started";

                            //
                            // NOTE: This is not really always an error due to nested
                            //       interactive loops.
                            //
                            if (!strict)
                                return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Skipped, we were not forced to start the cancel thread and
                        //       there is no timeout set.
                        //
                        result = "skipped checking cancel thread";

                        return ReturnCode.Ok;
                    }
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static ReturnCode InterruptCancelThread(
            Interpreter interpreter,
            bool strict,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                lock (interpreter.SyncRoot)
                {
                    Thread thread = interpreter.cancelThread;

                    if (thread != null)
                    {
                        try
                        {
                            if (thread.IsAlive)
                            {
                                thread.Interrupt(); /* throw */

                                if (!thread.Join(ThreadOps.DefaultJoinTimeout))
                                    /* BUGBUG: Leaks? */
                                    thread.Abort(); /* throw */

                                interpreter.cancelThread = null;
                                result = "interrupted cancel thread";

                                return ReturnCode.Ok;
                            }
                            else
                            {
                                interpreter.cancelThread = null; /* NOTE: Dead. */
                                result = "cancel thread is not alive";

                                //
                                // NOTE: This is not really always an error due
                                //        to subtle timing issues.
                                //
                                if (!strict)
                                    return ReturnCode.Ok;
                            }
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "failed to interrupt cancel thread, " +
                                "caught exception \"{0}\"",
                                e);
                        }
                    }
                    else
                    {
                        result = "cancel thread already stopped";

                        //
                        // NOTE: This is not really always an error due to
                        //       nested interactive loops.
                        //
                        if (!strict)
                            return ReturnCode.Ok;
                    }
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Loop
#if SHELL
        public static ReturnCode InteractiveLoop(
            Interpreter interpreter,
            IEnumerable<string> args,
            ref Result result
            )
        {
            bool exit = false;

            return InteractiveLoop(interpreter, args, ref exit, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode InteractiveLoop(
            Interpreter interpreter,
            IEnumerable<string> args,
            ref bool exit,
            ref Result result
            )
        {
            return InteractiveLoop(
                interpreter, false, args, ReturnCode.Ok, BreakpointType.None,
                null, null, null, EngineFlags.None, SubstitutionFlags.All,
                EventFlags.Engine, ExpressionFlags.All, HeaderFlags.None, null,
                null, ref exit, ref result);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode InteractiveLoop(
            Interpreter interpreter,
            bool debug,
            IEnumerable<string> args,
            ReturnCode code,
            BreakpointType breakpointType,
            string breakpointName,
            IToken token,
            ITraceInfo traceInfo,
            EngineFlags engineFlags,
            SubstitutionFlags substitutionFlags,
            EventFlags eventFlags,
            ExpressionFlags expressionFlags,
            HeaderFlags headerFlags,
            IClientData clientData,
            ArgumentList arguments,
            ref bool exit,
            ref Result result
            )
        {
            IList<string> argv = (args != null) ? new StringList(args) : null;

            RuntimeOps.DebugTrace(String.Format(
                "{0}: InteractiveLoop: entered, debug = {1}, argv = {2}, code = {3}, " +
                "breakpointType = {4}, breakpointName = {5}, token = {6}, traceInfo = {7}, " +
                "engineFlags = {8}, substitutionFlags = {9}, eventFlags = {10}, " +
                "expressionFlags = {11}, headerFlags = {12}, clientData = {13}, " +
                "arguments = {14}, exit = {15}, result = {16}{17}",
                GlobalState.GetCurrentThreadId(), debug, FormatOps.WrapOrNull(true, true, argv),
                code, breakpointType, FormatOps.WrapOrNull(breakpointName), (token != null),
                (traceInfo != null), FormatOps.WrapOrNull(engineFlags),
                FormatOps.WrapOrNull(substitutionFlags), FormatOps.WrapOrNull(eventFlags),
                FormatOps.WrapOrNull(expressionFlags), FormatOps.WrapOrNull(headerFlags),
                FormatOps.WrapOrNull(clientData), FormatOps.WrapOrNull(true, true, arguments),
                exit, FormatOps.WrapOrNull(true, true, result), Environment.NewLine),
                typeof(Interpreter).Name);

            if (interpreter != null)
            {
                //
                // NOTE: Make sure that we exit if the [exit] command was used.
                //
                CheckExit(interpreter, ref exit);

                if (!exit)
                {
                    //
                    // NOTE: Grab the interactive host from the interpreter.  The
                    //       methods of this interface represent the absolute bare
                    //       minimum functionality we need for the interactive loop
                    //       to function correctly.
                    //
                    IInteractiveHost interactiveHost = interpreter.Host;

                    //
                    // NOTE: Check for the host after checking whether we are exiting
                    //       because the host may have already been torn down and that
                    //       is not an error.
                    //
                    if (interactiveHost != null)
                    {
                        //
                        // NOTE: Make sure we have checked for native stack space at
                        //       least once at this point.
                        //
#if NATIVE && WINDOWS
                        if (RuntimeOps.CheckForStackSpace(interpreter,
                                Engine.GetStackExtraSpace()) == ReturnCode.Ok)
#endif
                        {
                            //
                            // NOTE: Push this interpreter onto the activation stack
                            //       because it is considered "active" while the
                            //       interactive loop is using it.
                            //
                            GlobalState.PushActiveInterpreter(interpreter);

                            ///////////////////////////////////////////////////////////////////////////

                            #region Save / Push Interactive Loop Level
                            //
                            // NOTE: Save the current interactive thread property so
                            //       that we can restore it later, if necessary.
                            //
                            Thread savedInteractiveThread = interpreter.InteractiveThread;

                            //
                            // NOTE: The thread for the interactive user is this one.
                            //
                            interpreter.InteractiveThread = Thread.CurrentThread;

                            //
                            // NOTE: Save the current interactive property so that we can
                            //       restore it later, if necessary.
                            //
                            bool savedInteractive = interpreter.Interactive;

                            //
                            // NOTE: This is an interactive loop and requires user input;
                            //       therefore, we are in interactive mode.
                            //
                            interpreter.Interactive = true;

                            //
                            // NOTE: We are entering an[other] instance of the interactive
                            //       loop; therefore, increase our tracking count.
                            //
                            interpreter.InteractiveLoops++; /* THREAD-SAFE */
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////
                            //
                            // NOTE: These variables are used when halting the interpreter (i.e.
                            //       cancel any scripts in progress and break out of any nested
                            //       interactive loops).
                            //
                            ReturnCode haltCode;
                            bool haltReset = false;
                            Result haltResult = null;

                            ///////////////////////////////////////////////////////////////////////////

                            #region [Maybe] Initialize Interactive (Shell) Script Library
                            //
                            // NOTE: Have we initialized the interactive shell yet?  If not,
                            //       do it now.
                            //
                            if (interpreter.initialized && /* NOTE: We must be initialized. */
                                (interpreter.InteractiveLoops == 1)) /* NOTE: Outermost loop only. */
                            {
                                #region Enter Debugger Active Level
#if DEBUGGER
                                //
                                // NOTE: Get the active debugger from the interpreter itself.
                                //
                                IDebugger debugger = interpreter.Debugger;

                                //
                                // NOTE: If we are debugging and the debugger is available for
                                //       this interpreter, increase one level of nested debugger
                                //       evaluations.  This is critical because it is used by the
                                //       engine to determine if breakpoints should occur in the
                                //       evaluated script and if the halt flag for the interpreter
                                //       should be reset after the script is evaluated (below).
                                //
                                if (debug && (debugger != null))
                                    debugger.SetActive(true);
#endif
                                #endregion

                                ///////////////////////////////////////////////////////////////////////

                                try
                                {
                                    //
                                    // NOTE: Evaluate the user-specific shell startup file and the
                                    //       interactive shell startup file, if any.
                                    //
                                    ReturnCode shellCode;
                                    Result shellResult = null;

                                    shellCode = interpreter.PrivateInitializeShell(ref shellResult);

                                    if (shellCode != ReturnCode.Ok)
                                        DebugOps.Complain(interpreter, shellCode,
                                            String.Format("shell: {0}", shellResult));

                                    //
                                    // BUGFIX: If the shell initialization script somehow set the
                                    //         exit flag for the interpreter, bail out before
                                    //         entering the actual interactive loop and without
                                    //         displaying any debugging related information.
                                    //
                                    CheckExit(interpreter, ref exit);

                                    //
                                    // NOTE: Did the shell initialization script somehow mark the
                                    //       interpreter as exited?
                                    //
                                    if (exit)
                                    {
                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: InteractiveLoop: leaving, shell " +
                                            "initialization script set the exit flag...{1}",
                                            GlobalState.GetCurrentThreadId(),
                                            Environment.NewLine), typeof(Interpreter).Name);

                                        goto skipLoop;
                                    }
                                }
                                finally
                                {
                                    #region Exit Debugger Active Level
#if DEBUGGER
                                    //
                                    // NOTE: Back out one level of nested debugger evaluations.
                                    //
                                    debugger = interpreter.Debugger; /* NOTE: Refresh. */

                                    if (debug && (debugger != null))
                                        debugger.SetActive(false);
#endif
                                    #endregion
                                }
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////

                            #region Write Interactive Header
                            if (debug)
                            {
                                //
                                // NOTE: Attempt to cast the interactive host to an informational
                                //       host.  If this fails, we simply ignore it because it is
                                //       not strictly required for proper functioning of the
                                //       interactive loop.
                                //
                                IInformationHost informationHost = interactiveHost as IInformationHost;
                                bool proxy = RuntimeOps.IsTransparentProxy(informationHost);

                                if (informationHost != null)
                                    informationHost.WriteHeader(
                                        interpreter, code, breakpointType, breakpointName,
                                        !proxy ? token : null, !proxy ? traceInfo : null,
                                        engineFlags, substitutionFlags, eventFlags, expressionFlags,
                                        RuntimeOps.GetDebuggerHeaderFlags(headerFlags, debug, true,
                                            false, debug),
                                        clientData, arguments, result);
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////
                            //
                            // NOTE: Get the header display flags for the interpreter and add
                            //       the necessary flags for our debugging mode.
                            //
                            HeaderFlags localHeaderFlags = RuntimeOps.GetDebuggerHeaderFlags(
                                interpreter.HeaderFlags, debug, false, false, debug);

                            ///////////////////////////////////////////////////////////////////////////

                            #region Save / Push Engine Flags
                            //
                            // NOTE: Initialize the saved engine flags (to nothing).
                            //
                            EngineFlags savedEngineFlags = EngineFlags.None;

                            //
                            // NOTE: Are we being called by the debugger?
                            //
                            if (debug)
                            {
                                //
                                // BUGFIX: Save the current engine flags for this interpreter to
                                //         be restored as this [potentially nested] interactive
                                //         loop is exiting.
                                //
                                savedEngineFlags = interpreter.EngineFlags;

                                //
                                // NOTE: For functions called from inside the interactive loop
                                //       that do not, directly or indirectly, receive engine
                                //       flags as a formal argument, we still need to be sure
                                //       that the engine flags received from our caller are in
                                //       effect for the entire duration of this interactive
                                //       loop; therefore, we combine the engine flags received
                                //       from our caller into the engine flags for the
                                //       interpreter.
                                //
                                interpreter.EngineFlags |= engineFlags;
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////
                            //
                            // NOTE: Get the local engine, substitution, and event flags for
                            //       this interpreter.
                            //
                            EngineFlags localEngineFlags = interpreter.EngineFlags;
                            SubstitutionFlags localSubstitutionFlags = interpreter.SubstitutionFlags;
                            EventFlags localEventFlags = interpreter.EngineEventFlags;
                            ExpressionFlags localExpressionFlags = interpreter.ExpressionFlags;

                            //
                            // NOTE: Add the engine flags for our debugging mode to our local
                            //       engine flags, if necessary.
                            //
                            localEngineFlags |= RuntimeOps.GetDebuggerEngineFlags(debug);

                            //
                            // NOTE: The "forceCancel" and "forceHalt" variables are used as
                            //       the "force" argument to the ResetCancel and ResetHalt
                            //       methods, respectively.  These methods are used at the top
                            //       of each loop iteration to reset the interpreter flags as
                            //       necessary prior to doing anything with the interpreter.
                            //
                            bool forceCancel = false;
                            bool forceHalt = false;

                            //
                            // NOTE: These result variables need to be preserved between loop
                            //       iterations for possible modification of the REAL result(s)
                            //       (primarily in debug mode).
                            //
                            ReturnCode localCode = ReturnCode.Ok;
                            Result localResult = null;
                            int localErrorLine = 0;

#if NATIVE && TCL
                            //
                            // NOTE: This variable controls which script engine is used to
                            //       evaluate all interactively entered scripts (i.e. using
                            //       Eagle or using "tclsh emulation mode").  By default (zero),
                            //       all interactively entered scripts will be evaluated using
                            //       the interactive Eagle interpreter.  When this variable is
                            //       non-zero, all interactively entered scripts will be
                            //       evaluated using the master Tcl interpreter, if available.
                            //
                            bool tclsh = false;
                            string tclInterpName = null;
#endif

                            //
                            // NOTE: Does the GC test thread need to be shutdown upon exiting
                            //       THIS interactive loop?
                            //
                            bool startedGcThread = false;

                            //
                            // NOTE: This variable controls the primary loop (below).  When it
                            //       is true, the loop will terminate (i.e. there is no more
                            //       input available, the user has exited, etc).
                            //
                            bool done = false;

                            while (!done)
                            {
                                //
                                // NOTE: Reset the interactive processing variables used by input
                                //       loop(s) and the output handling code.
                                //
                                bool show = true; /* NOTE: Show local result by default. */
                                bool previous = true; /* NOTE: Update previous interactive input? */

                                //
                                // NOTE: Enable traces in the interactive input processing?
                                //
                                bool trace = FlagOps.HasFlags(
                                    interpreter.interpreterFlags, InterpreterFlags.TraceInput, true);

                                ///////////////////////////////////////////////////////////////////////

                                #region Interactive Loop Input Processing
                                //
                                // NOTE: Declare the variables used for processing all interactive
                                //       input, redirected or otherwise (i.e. from the actual
                                //       interactive user).
                                //
                                bool canceled;
                                string text;
                                bool notReady;
                                Result parseError;

                                //
                                // NOTE: Invoke the method directly responsible for getting a
                                //       complete [logical] line of interactive input.
                                //
                                GetInteractiveInput(
                                    interpreter, trace, debug, false, localSubstitutionFlags,
                                    clientData, forceCancel, forceHalt, ref interactiveHost,
                                    ref exit, ref done, ref previous, out canceled, out text,
                                    out notReady, out parseError);
                                #endregion

                                ///////////////////////////////////////////////////////////////////////

                                #region Interactive Host Begin-Processing Hook Callback
                                //
                                // NOTE: Give the interactive host the opportunity to preview
                                //       and possibly respond to the entered command.
                                //
                                if (!done && (interactiveHost != null))
                                {
                                    ReturnCode beginCode;
                                    Result beginError = null;

                                    beginCode = interactiveHost.BeginProcessing(
                                        interpreter.InteractiveLoops, ref text, ref beginError);

                                    if (beginCode != ReturnCode.Ok)
                                    {
                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: InteractiveLoop: host processing error, " +
                                            "beginCode = {1}, beginError = {2}{3}",
                                            GlobalState.GetCurrentThreadId(), beginCode,
                                            FormatOps.WrapOrNull(true, true, beginError),
                                            Environment.NewLine), typeof(Interpreter).Name);

                                        //
                                        // NOTE: Processing has been halted by the interactive
                                        //       host, bail out of the interactive loop now.
                                        //
                                        done = true;
                                    }
                                }
                                #endregion

                                ///////////////////////////////////////////////////////////////////////
                                //
                                // NOTE: Now, evaluate the completed script entered by the user
                                //       (if they have not canceled).
                                //
                                if (!done && !canceled && !notReady)
                                {
                                    if (!String.IsNullOrEmpty(text) && (text.Trim().Length > 0))
                                    {
#if INTERACTIVE_COMMANDS
                                        //
                                        // NOTE: Reset the interactive command processing variables
                                        //       used by the CheckInteractiveCommand method.
                                        //
                                        bool debugVerbose = true;
                                        ArgumentList debugArguments = null;
#endif

                                        ///////////////////////////////////////////////////////////////

                                        #region Enter Debugger Active Level
#if DEBUGGER
                                        //
                                        // NOTE: Refresh the active debugger from the interpreter itself.
                                        //
                                        IDebugger debugger = interpreter.Debugger;

                                        //
                                        // NOTE: If we are debugging and the debugger is available for
                                        //       this interpreter, increase one level of nested debugger
                                        //       evaluations.  This is critical because it is used by the
                                        //       engine to determine if breakpoints should occur in the
                                        //       evaluated script and if the halt flag for the interpreter
                                        //       should be reset after the script is evaluated (below).
                                        //
                                        if (debug && (debugger != null))
                                            debugger.SetActive(true);
#endif
                                        #endregion

                                        ///////////////////////////////////////////////////////////////

                                        try
                                        {
                                            //
                                            // NOTE: First, check and see if an "interactive" command
                                            //       was entered (e.g. one that starts with "#");
                                            //       otherwise, evaluate the entered input as a script.
                                            //
                                            // TODO: This code is still too messy.  Resolving and
                                            //       executing interactive commands should be a lot
                                            //       cleaner, simpler, and more efficient.
                                            //
                                            ///////////////////////////////////////////////////////////

                                            #region Interactive Commands
#if INTERACTIVE_COMMANDS
                                            #region "Special" Interactive Commands
                                            //
                                            // NOTE: This first interactive command check is designed
                                            //       to allow any overridden interactive commands to
                                            //       be executed properly (e.g. "#show") even if they
                                            //       do not actually represent a built-in interactive
                                            //       command (e.g. "#foo").  This call modifies the
                                            //       local return code and result.  The result of the
                                            //       interactive command execution will be displayed
                                            //       using the normal mechanisms, after this huge
                                            //       "if/else" block (below).
                                            //
                                            if (CheckInteractiveCommand(
                                                    interpreter, text, null, clientData, false,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                //
                                                // NOTE: Do nothing.
                                                //
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "nop", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                #region Interactive Nop Command
                                                //
                                                // NOTE: This interactive command is specifically
                                                //       designed to do nothing.
                                                //
                                                #endregion
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "go", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                //
                                                // NOTE: This must be done inline (in this method)
                                                //       because it uses the C# break statement.
                                                //
                                                #region Interactive Go Command
                                                //
                                                // NOTE: In debug mode, we simply break out of
                                                //       the loop; otherwise, an error message
                                                //       is displayed.
                                                //
                                                if (debug)
                                                {
                                                    localResult = String.Empty;
                                                    localCode = ReturnCode.Ok;

                                                    break;
                                                }
                                                else
                                                {
                                                    localResult = String.Format(
                                                        "cannot \"{0}go\" when not debugging",
                                                        RuntimeOps.InteractiveCommandPrefix);

                                                    localCode = ReturnCode.Error;
                                                }
                                                #endregion
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "run", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                //
                                                // NOTE: This must be done inline (in this method)
                                                //       because it uses the C# break statement.
                                                //
                                                #region Interactive Run Command
#if DEBUGGER
                                                //
                                                // NOTE: In debug mode, we simply disable further
                                                //       stepping and break out of the loop;
                                                //       otherwise, an error message is displayed.
                                                //
                                                if (debug)
                                                {
                                                    IDebugger localDebugger = null;

                                                    if (Engine.CheckDebugger(interpreter, false,
                                                            ref localDebugger, ref localResult))
                                                    {
                                                        //
                                                        // FIXME: Yes, this is somewhat confusing.
                                                        //        Why does the "#run" command call
                                                        //        the IDebugger.Reset method?
                                                        //
                                                        //        From the perspective of the
                                                        //        IDebugger interface itself, the
                                                        //        Reset method clears all the
                                                        //        internal debugging state
                                                        //        (basically resetting it to null
                                                        //        and zero).
                                                        //
                                                        //        However, from the perspective of
                                                        //        the interactive loop, this command
                                                        //        (which is named "#run") is used to
                                                        //        disable all debugging features and
                                                        //        run the script being evaluated at
                                                        //        full speed.
                                                        //
                                                        localCode = localDebugger.Reset(
                                                            ref localResult);

                                                        if (localCode == ReturnCode.Ok)
                                                            break;
                                                    }
                                                    else
                                                    {
                                                        localCode = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    localResult = String.Format(
                                                        "cannot \"{0}run\" when not debugging",
                                                        RuntimeOps.InteractiveCommandPrefix);

                                                    localCode = ReturnCode.Error;
                                                }
#else
                                                localResult = "not implemented";
                                                localCode = ReturnCode.Error;
#endif
                                                #endregion
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "break", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                //
                                                // NOTE: This must be done inline (in this method)
                                                //       because it uses the C# break statement.
                                                //
                                                #region Interactive Break Command
#if DEBUGGER
                                                IDebugger localDebugger = null;

                                                if (Engine.CheckDebugger(interpreter, false,
                                                        ref localDebugger, ref localResult))
                                                {
#if PREVIOUS_RESULT
                                                    //
                                                    // NOTE: At this point, the result of the
                                                    //       previous command may still be
                                                    //       untouched and will be displayed
                                                    //       verbatim upon entry into the
                                                    //       interactive loop.
                                                    //
                                                    localResult = Result.Copy(
                                                        interpreter.PreviousResult, true); /* COPY */
#endif

                                                    //
                                                    // NOTE: Break into the debugger by
                                                    //       starting a nested interactive
                                                    //       loop.
                                                    //
                                                    localCode = _Private.Debugger.Break(
                                                        localDebugger, interpreter, localCode,
                                                        BreakpointType.Break, debugArguments[0],
                                                        token, traceInfo, localEngineFlags,
                                                        localSubstitutionFlags, localEventFlags,
                                                        localExpressionFlags, localHeaderFlags,
                                                        clientData, arguments, ref localResult);

                                                    //
                                                    // FIXME: If there were no other failures in
                                                    //        the interactive loop, perhaps we
                                                    //        should reflect the previous result?
                                                    //        Better logic here may be needed.
                                                    //
                                                    if ((localCode == ReturnCode.Ok) &&
                                                        (localResult != null))
                                                    {
                                                        localCode = localResult.ReturnCode;
                                                    }
                                                    else if (interpreter.InteractiveLoops > 1)
                                                    {
                                                        //
                                                        // BUGFIX: If the interpreter has been
                                                        //         halted then we need to break
                                                        //         out of this loop and any
                                                        //         nested interactive loops
                                                        //         (except the outermost one).
                                                        //
                                                        result = localResult;
                                                        code = localCode;

                                                        break;
                                                    }
                                                }
                                                else
                                                {
                                                    localCode = ReturnCode.Error;
                                                }
#else
                                                localResult = "not implemented";
                                                localCode = ReturnCode.Error;
#endif
                                                #endregion
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "halt", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                //
                                                // NOTE: This must be done inline (in this method)
                                                //        because it uses the C# break statement.
                                                //
                                                #region Interactive Halt Command
                                                //
                                                // NOTE: In debug mode, we simply break out of
                                                //       the loop and return failure to the
                                                //       caller; otherwise, an error message
                                                //       is displayed.
                                                //
                                                if (debug)
                                                {
                                                    //
                                                    // NOTE: Prevent further trips through the
                                                    //       interpreter and the interactive
                                                    //       loop(s).
                                                    //
                                                    localResult = "INTERPRETER HALTED";
                                                    localCode = Engine.HaltEvaluate(
                                                        interpreter, localResult,
                                                        ref localResult);

                                                    if (localCode == ReturnCode.Ok)
                                                    {
                                                        result = localResult; /* TRANSFER */
                                                        code = ReturnCode.Error;

                                                        break;
                                                    }
                                                }
                                                else
                                                {
                                                    localResult = String.Format(
                                                        "cannot \"{0}halt\" when not debugging",
                                                        RuntimeOps.InteractiveCommandPrefix);

                                                    localCode = ReturnCode.Error;
                                                }
                                                #endregion
                                            }
                                            #endregion

                                            ///////////////////////////////////////////////////////////

                                            #region "Normal" Interactive Commands
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "cmd", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult) ||
                                                CheckInteractiveCommand(
                                                    interpreter, text, "cmd.exe", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult) ||
                                                CheckInteractiveCommand(
                                                    interpreter, text, "cmd", clientData, false,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult) ||
                                                CheckInteractiveCommand(
                                                    interpreter, text, "cmd.exe", clientData, false,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.cmd(
                                                    interpreter, debugArguments, localEventFlags,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "tclshrc", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.tclshrc(
                                                    interpreter, interactiveHost as IEngineHost,
                                                    debugArguments, localEventFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "reset", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.reset(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "color", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.color(
                                                    interactiveHost, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "exceptions", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.exceptions(
                                                    interactiveHost, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "testgc", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.testgc(
                                                    interpreter, debugArguments, ref startedGcThread,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "hcancel", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.hcancel(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "hexit", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.hexit(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "check", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.check(
                                                    interpreter, debugArguments, localEngineFlags,
                                                    localSubstitutionFlags, localEventFlags,
                                                    localExpressionFlags, ref localCode,
                                                    ref localResult, ref localErrorLine);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "again", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.again(
                                                    interpreter, ref previous, ref show,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "help", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.help(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "usage", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.usage(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "version", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.version(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "args", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.args(
                                                    argv, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "ainfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.ainfo(
                                                    interpreter, interactiveHost, code,
                                                    breakpointType, breakpointName, arguments,
                                                    result, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "clearq", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.clearq(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "oinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.oinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "vinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.vinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "complaint", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.complaint(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "cuinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.cuinfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "dinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.dinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "testinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.testinfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "toinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.toinfo(
                                                    interpreter, interactiveHost, token,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "tinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.tinfo(
                                                    interpreter, interactiveHost, traceInfo,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "stack", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.stack(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "finfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.finfo(
                                                    interpreter, interactiveHost, engineFlags,
                                                    substitutionFlags, eventFlags, expressionFlags,
                                                    headerFlags, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "lfinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.lfinfo(
                                                    interpreter, interactiveHost, localEngineFlags,
                                                    localSubstitutionFlags, localEventFlags,
                                                    localExpressionFlags, localHeaderFlags,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "frinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.frinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "einfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.einfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "cinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.cinfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "eninfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.eninfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "sinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.sinfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "histfile", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.histfile(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "histinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.histinfo(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "histclear", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.histclear(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "histload", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.histload(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "histsave", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.histsave(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "hinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.hinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "iinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.iinfo(
                                                    interpreter, interactiveHost, result,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "fresc", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.fresc(
                                                    ref forceCancel, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "fresh", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.fresh(
                                                    ref forceHalt, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "resc", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.resc(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "resh", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.resh(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "rehash", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.rehash(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "eval", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.eval(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult, ref localErrorLine);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "subst", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.subst(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "suspend", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.suspend(
                                                    interpreter, debug, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "resume", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.resume(
                                                    interpreter, debug, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "about", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.about(
                                                    interpreter, interactiveHost, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "chans", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.chans(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "dpath", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.dpath(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "cancel", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.cancel(
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "test", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult) ||
                                                CheckInteractiveCommand(
                                                    interpreter, text, "ptest", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.test(
                                                    interpreter, debugArguments, localEngineFlags,
                                                    localSubstitutionFlags, localEventFlags,
                                                    localExpressionFlags, ref localCode,
                                                    ref localResult, ref localErrorLine);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "trustclr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.trustclr(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "trustdir", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.trustdir(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "testdir", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.testdir(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "purge", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.purge(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "restc", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.restc(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "restv", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.restv(
                                                    interpreter, argv, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "vout", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.vout(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "rlimit", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.rlimit(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "ntypes", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.ntypes(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "nflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.nflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "hflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.hflags(
                                                    interpreter, debugArguments, debug,
                                                    ref localHeaderFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "lhflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.lhflags(
                                                    interpreter, debugArguments, debug,
                                                    ref localHeaderFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "cflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.cflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "scflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.scflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "iflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.iflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "pflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.pflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "eflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.eflags(
                                                    interpreter, debugArguments,
                                                    ref localEngineFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "evflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.evflags(
                                                    interpreter, debugArguments,
                                                    ref localEventFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "exflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.exflags(
                                                    interpreter, debugArguments,
                                                    ref localExpressionFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "ieflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.ieflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "ievflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.ievflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "iexflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.iexflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "leflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.leflags(
                                                    interpreter, debugArguments,
                                                    ref localEngineFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "levflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.levflags(
                                                    interpreter, debugArguments, ref localEventFlags,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "lexflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.lexflags(
                                                    interpreter, debugArguments,
                                                    ref localExpressionFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "sflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.sflags(
                                                    interpreter, debugArguments,
                                                    ref localSubstitutionFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "isflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.isflags(
                                                    interpreter, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "lsflags", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.lsflags(
                                                    interpreter, debugArguments,
                                                    ref localSubstitutionFlags, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "step", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.step(
                                                    interpreter, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "style", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.style(
                                                    interactiveHost, debugArguments, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "exit", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.exit(
                                                    ref exit, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "canexit", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.canexit(
                                                    interactiveHost, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "show", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.show(
                                                    interpreter, interactiveHost, debugArguments,
                                                    debug, code, breakpointType, breakpointName,
                                                    token, traceInfo, engineFlags, substitutionFlags,
                                                    eventFlags, expressionFlags, clientData, arguments,
                                                    result, localHeaderFlags, localCode, localResult,
                                                    ref show);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "overr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.overr(
                                                    interpreter, interactiveHost, debugArguments,
                                                    ref show, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "prevr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.prevr(
                                                    interpreter, interactiveHost, ref show,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "nextr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.nextr(
                                                    interpreter, interactiveHost, localCode,
                                                    localResult, ref show);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "fresr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.fresr(
                                                    interactiveHost, ref show, ref localCode,
                                                    ref localResult, ref code, ref result);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "resr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.resr(
                                                    interactiveHost, ref show, ref localCode,
                                                    ref localResult, ref code, ref result);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "clearr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.clearr(
                                                    interactiveHost, ref show, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "nullr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.nullr(
                                                    interactiveHost, ref show, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "copyr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.copyr(
                                                    interactiveHost, code, result, ref show,
                                                    ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "setr", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.setr(
                                                    interactiveHost, localCode, localResult,
                                                    ref show, ref code, ref result);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "mover", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.mover(
                                                    interactiveHost, ref show, ref localCode,
                                                    ref localResult, ref code, ref result);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "lrinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.lrinfo(
                                                    interactiveHost, localCode, localResult,
                                                    localErrorLine, ref show);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "mrinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.mrinfo(
                                                    interpreter, interactiveHost, debugArguments,
                                                    code, result, ref show);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "rinfo", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.rinfo(
                                                    interpreter, interactiveHost, localCode,
                                                    localResult, localErrorLine, code, result,
                                                    ref show);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "tclsh", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.tclsh(
                                                    ref tclsh, ref localCode, ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "tclinterp", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.tclinterp(
                                                    debugArguments, ref tclInterpName, ref localCode,
                                                    ref localResult);
                                            }
                                            else if (CheckInteractiveCommand(
                                                    interpreter, text, "queue", clientData, true,
                                                    ref debugVerbose, ref debugArguments,
                                                    ref localCode, ref localResult))
                                            {
                                                InteractiveOps.Commands.queue(
                                                    interpreter, trace, debug, localEngineFlags,
                                                    localSubstitutionFlags, localEventFlags,
                                                    localExpressionFlags, clientData, forceCancel,
                                                    forceHalt, ref interactiveHost, ref exit,
                                                    ref done, ref previous, ref canceled, ref text,
                                                    ref notReady, ref parseError, ref localCode,
                                                    ref localResult);
                                            }
                                            #endregion

                                            ///////////////////////////////////////////////////////////

                                            else // NOTE: Dangling else (see "if" statement below).
#endif
                                            #endregion

                                            ///////////////////////////////////////////////////////////

                                            #region Interactive Script Evaluation
                                            if (!String.IsNullOrEmpty(text) && (text.Trim().Length > 0))
                                            {
#if NATIVE && TCL
                                                #region Tcl Shell Emulation Mode
                                                //
                                                // NOTE: Are we operating in "tclsh emulation mode"?
                                                //       If so, try to get the active Tcl master
                                                //       interpreter and use it to do the script
                                                //       evaluation; otherwise, fallback on the normal
                                                //       behavior of using the Eagle interpreter.
                                                //
                                                if (tclsh)
                                                {
                                                    localErrorLine = 0;

                                                    ///////////////////////////////////////////////////

                                                    #region Get Native Tcl Interpreter
                                                    string interpName = null;
                                                    IntPtr interp = IntPtr.Zero;

                                                    if (tclInterpName != null)
                                                    {
                                                        interpName = tclInterpName;

                                                        localCode = interpreter.GetTclInterpreter(
                                                            interpName, ref interp, ref localResult);
                                                    }
                                                    else
                                                    {
                                                        localCode = interpreter.GetAnyTclMasterInterpreter(
                                                            ref interpName, ref interp, ref localResult);
                                                    }
                                                    #endregion

                                                    ///////////////////////////////////////////////////

                                                    #region Evaluate Tcl Script
                                                    if (localCode == ReturnCode.Ok)
                                                    {
                                                        localCode = interpreter.EvaluateTclScript(
                                                            interpName, text, ref localResult);

                                                        if (localCode != ReturnCode.Ok)
                                                            localErrorLine = TclWrapper.GetErrorLine(
                                                                interpreter.tclApi, interp);
                                                    }
                                                    #endregion
                                                }
                                                #endregion

                                                else
#endif
                                                {
                                                    #region Start Script Timeout Thread
                                                    //
                                                    // NOTE: Core interactive shell evaluation sequence.
                                                    //       This is somewhat tricky because of various
                                                    //       re-entrancy concerns.  Matters are further
                                                    //       complicated by script cancellation and
                                                    //       timeouts.
                                                    //
                                                    ReturnCode cancelCode;
                                                    Result cancelResult = null;

                                                    //
                                                    // NOTE: Start the runaway script detection /
                                                    //       cancellation thread.
                                                    //
                                                    cancelCode = StartCancelThread(
                                                        interpreter, false, false, ref cancelResult);

                                                    if (cancelCode != ReturnCode.Ok)
                                                        DebugOps.Complain(interpreter, cancelCode,
                                                            cancelResult);
                                                    #endregion

                                                    ///////////////////////////////////////////////////

                                                    #region Evaluate Eagle Script
                                                    //
                                                    // NOTE: Perform the script evaluation using our
                                                    //       local code and result variables (i.e. do
                                                    //       not change the primary "interpreter
                                                    //       result").  It should also be noted that
                                                    //       this evaluation is not isolated in any
                                                    //       other way and will change the interpreter
                                                    //       state.
                                                    //
                                                    localErrorLine = 0;

                                                    localCode = Engine.EvaluateScript(
                                                        interpreter, text, localEngineFlags,
                                                        localSubstitutionFlags, localEventFlags,
                                                        localExpressionFlags, ref localResult,
                                                        ref localErrorLine);
                                                    #endregion

                                                    ///////////////////////////////////////////////////

                                                    #region [Maybe] Reset Halt Flag
                                                    //
                                                    // NOTE: Reset the interpreter halt flag if this is
                                                    //       the outermost interactive loop and the
                                                    //       interpreter evaluation nesting level is
                                                    //       zero.
                                                    //
                                                    haltCode = Engine.ResetHalt(
                                                        interpreter, false, ref haltReset,
                                                        ref haltResult);

                                                    if (haltCode != ReturnCode.Ok)
                                                    {
                                                        DebugOps.Complain(interpreter, haltCode,
                                                            haltResult);
                                                    }
#if PREVIOUS_RESULT
                                                    else if (haltReset)
                                                    {
                                                        //
                                                        // NOTE: Reset the previous result here;
                                                        //       otherwise, the [debug break] command
                                                        //       will pick up the now stale
                                                        //       "INTERPRETER HALTED" result and run
                                                        //       with it otherwise.
                                                        //
                                                        interpreter.PreviousResult = null;
                                                    }
#endif
                                                    #endregion

                                                    ///////////////////////////////////////////////////

                                                    #region Stop Script Timeout Thread
                                                    //
                                                    // NOTE: Stop the runaway script detection /
                                                    //       cancellation thread.
                                                    //
                                                    cancelCode = InterruptCancelThread(
                                                        interpreter, false, ref cancelResult);

                                                    if (cancelCode != ReturnCode.Ok)
                                                        DebugOps.Complain(interpreter, cancelCode,
                                                            cancelResult);
                                                    #endregion

                                                    ///////////////////////////////////////////////////

                                                    #region Check Halt Flag
                                                    //
                                                    // NOTE: If debugging is still halted (i.e. the
                                                    //       ResetHalt above had no effect because this
                                                    //       is not the outermost interactive loop),
                                                    //       then exit this interactive loop.
                                                    //
                                                    if (Engine.IsHalted(
                                                            interpreter, true,
                                                            ref localResult) != ReturnCode.Ok)
                                                    {
                                                        result = localResult; /* TRANSFER */
                                                        code = ReturnCode.Error;
                                                        break;
                                                    }
                                                    #endregion
                                                }
                                            }
                                            else
                                            {
                                                //
                                                // NOTE: We did absolutely nothing, record no previous
                                                //       interactive input and show no result.
                                                //
                                                previous = false;
                                                show = false;
                                            }
                                            #endregion
                                        }
                                        finally
                                        {
                                            #region Exit Debugger Active Level
#if DEBUGGER
                                            //
                                            // NOTE: Back out one level of nested debugger evaluations.
                                            //
                                            debugger = interpreter.Debugger; /* NOTE: Refresh. */

                                            if (debug && (debugger != null))
                                                debugger.SetActive(false);
#endif
                                            #endregion
                                        }

                                        ///////////////////////////////////////////////////////////////

                                        #region Refresh Interpreter Host
                                        //
                                        // BUGFIX: We must reload the local interpreter host variable
                                        //         from the actual property because it may have been
                                        //         changed by an interactive command or script.
                                        //
                                        interactiveHost = interpreter.Host;
                                        #endregion

                                        ///////////////////////////////////////////////////////////////
                                        //
                                        // NOTE: If there is a debugger available, store a copy of the
                                        //       local result in the debugger result.
                                        //
                                        RuntimeOps.SetDebuggerResult(
                                            interpreter, localCode, localResult, false);

                                        ///////////////////////////////////////////////////////////////

                                        #region Refresh Exit Flag & Process "Previous" / "Show" Flags
                                        //
                                        // NOTE: Make sure that we exit if the [exit] command was used.
                                        //
                                        CheckExit(interpreter, ref exit);

                                        //
                                        // NOTE: Check if we are supposed to exit; if not, write the
                                        //       result of the last operation.
                                        //
                                        if (exit)
                                        {
                                            //
                                            // NOTE: Exit from entire interactive loop [and possibly
                                            //       the entire interpreter as well].  If the entire
                                            //       interpreter has been marked as exited all nested
                                            //       interactive loops will still exit cleanly.
                                            //
                                            done = true;
                                        }
                                        else
                                        {
                                            if (previous)
                                                interpreter.PreviousInteractiveInput = text;

                                            if (show && (interactiveHost != null))
                                                interactiveHost.WriteResultLine(
                                                    localCode, localResult, localErrorLine);
                                        }
                                        #endregion
                                    }
                                }

                                ///////////////////////////////////////////////////////////////////////

                                #region Interactive Host End-Processing Hook Callback
                                if (!done && (interactiveHost != null))
                                {
                                    ReturnCode endCode;
                                    Result endError = null;

                                    endCode = interactiveHost.EndProcessing(
                                        interpreter.InteractiveLoops, ref text, ref endError);

                                    if (endCode != ReturnCode.Ok)
                                    {
                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: InteractiveLoop: host processing error, " +
                                            "endCode = {1}, endError = {2}{3}",
                                            GlobalState.GetCurrentThreadId(), endCode,
                                            FormatOps.WrapOrNull(true, true, endError),
                                            Environment.NewLine), typeof(Interpreter).Name);

                                        //
                                        // NOTE: Processing has been halted by the interactive
                                        //       host, bail out of the interactive loop now.
                                        //
                                        done = true;
                                    }
                                }
                                #endregion
                            }

                            ///////////////////////////////////////////////////////////////////////////

                            #region Interactive Host Done-Processing Hook Callback
                            if (interactiveHost != null)
                            {
                                ReturnCode doneCode;
                                Result doneError = null;

                                doneCode = interactiveHost.DoneProcessing(
                                    interpreter.InteractiveLoops, ref doneError);

                                if (doneCode != ReturnCode.Ok)
                                {
                                    RuntimeOps.DebugTrace(String.Format(
                                        "{0}: InteractiveLoop: host processing error, " +
                                        "doneCode = {1}, doneError = {2}{3}",
                                        GlobalState.GetCurrentThreadId(), doneCode,
                                        FormatOps.WrapOrNull(true, true, doneError),
                                        Environment.NewLine), typeof(Interpreter).Name);
                                }
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////

                            #region Restore / Pop Engine Flags
                            if (debug)
                            {
                                //
                                // BUGFIX: Restore engine flags to the state they were in upon
                                //         entry into this interactive loop.
                                //
                                interpreter.EngineFlags = savedEngineFlags;

                                //
                                // NOTE: clear out the previously saved engine flags.  This
                                //       statement serves as a reminder to re-think the
                                //       interpreter engine flag management, including
                                //       possibly adding a saved engine flags member to the
                                //       interpreter state.
                                //
                                savedEngineFlags = EngineFlags.None;
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////

                            #region GC Test Thread Cleanup
                            //
                            // BUGFIX: Always try to stop the garbage collection test thread
                            //         if this interactive loop started it.
                            //
                            if (startedGcThread)
                            {
                                //
                                // NOTE: Acquire and hold the interpreter lock while shutting
                                //       down the garbage collection test thread, if it is
                                //       active.
                                //
                                lock (interpreter.SyncRoot)
                                {
                                    //
                                    // NOTE: Does the interpreter have an active garbage
                                    //       collection test thread?
                                    //
                                    if (TestOps.HasGcThread(interpreter))
                                    {
                                        ReturnCode interruptCode;
                                        Result interruptResult = null;

                                        interruptCode = TestOps.InterruptGcThread(
                                            interpreter, false, ref interruptResult);

                                        if (interruptCode != ReturnCode.Ok)
                                            DebugOps.Complain(interpreter, interruptCode,
                                                interruptResult);
                                    }
                                }
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////

                            #region Write Interactive Footer
                            if (debug && !exit && (interactiveHost != null))
                            {
                                //
                                // NOTE: Attempt to cast the interactive host to an informational
                                //       host.  If this fails, we simply ignore it because it is
                                //       not strictly required for proper functioning of the
                                //       interactive loop.
                                //
                                IInformationHost informationHost =
                                    interactiveHost as IInformationHost;

                                bool proxy = RuntimeOps.IsTransparentProxy(informationHost);

                                if (informationHost != null)
                                    informationHost.WriteFooter(
                                        interpreter, code, breakpointType, breakpointName,
                                        !proxy ? token : null, !proxy ? traceInfo : null,
                                        engineFlags, substitutionFlags, eventFlags, expressionFlags,
                                        RuntimeOps.GetDebuggerHeaderFlags(
                                            headerFlags, debug, true, false, debug),
                                        clientData, arguments, result);
                            }
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////
                            //
                            // NOTE: We can get to this label if the script evaluated
                            //       in the script initialization code block (above)
                            //       calls the [exit] command (or does something else
                            //       that causes the interpreter to be marked as exited)
                            //       for some (any?) reason.
                            //
                        skipLoop:

                            ///////////////////////////////////////////////////////////////////////////

                            #region Restore / Pop Interactive Loop Level
                            //
                            // NOTE: We are exiting this instance of the interactive
                            //       loop; therefore, decrease our tracking count.
                            //
                            interpreter.InteractiveLoops--; /* THREAD-SAFE */

                            //
                            // BUGFIX: Restore the interactive property to its previous
                            //         setting here, do not simply set it to false.
                            //
                            interpreter.Interactive = savedInteractive;

                            //
                            // BUGFIX: Restore the interactive thread property to its
                            //         previous value here, do not simply set it to null.
                            //
                            interpreter.InteractiveThread = savedInteractiveThread;
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////

                            #region [Maybe] Reset Halt Flag
                            //
                            // NOTE: Possibly reset the "halt" state for the interpreter
                            //       (if there are no more nested loops).
                            //
                            haltCode = Engine.ResetHalt(
                                interpreter, false, ref haltReset, ref haltResult);

                            if (haltCode != ReturnCode.Ok)
                            {
                                DebugOps.Complain(interpreter, haltCode, haltResult);
                            }
#if PREVIOUS_RESULT
                            else if (haltReset)
                            {
                                //
                                // NOTE: Reset the previous result here; otherwise, the
                                //       [debug break] command will pick up the now stale
                                //       "INTERPRETER HALTED" result and run with it
                                //       otherwise.
                                //
                                interpreter.PreviousResult = null;
                            }
#endif
                            #endregion

                            ///////////////////////////////////////////////////////////////////////////
                            //
                            // NOTE: Pop this interpreter from the activation stack
                            //       because the interactive loop is now done using it.
                            //
                            /* IGNORED */
                            GlobalState.PopActiveInterpreter();
                        }
#if NATIVE && WINDOWS
                        else
                        {
                            //
                            // NOTE: We are out of native stack space.  Perhaps there
                            //       are too many nested interactive loops?
                            //
                            result = "out of stack space (infinite loop?)";
                            code = ReturnCode.Error;
                        }
#endif
                    }
                    else
                    {
                        //
                        // NOTE: We absolutely require an interpreter hosting
                        //       environment (such as the provided Console Host)
                        //       for the interactive loop to function properly.
                        //
                        result = "interpreter host not available"; /* ERROR */
                        code = ReturnCode.Error;
                    }
                }
            }
            else
            {
                result = "invalid interpreter"; /* ERROR */
                code = ReturnCode.Error;
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: InteractiveLoop: exited, debug = {1}, argv = {2}, code = {3}, " +
                "breakpointType = {4}, breakpointName = {5}, token = {6}, traceInfo = {7}, " +
                "engineFlags = {8}, substitutionFlags = {9}, eventFlags = {10}, " +
                "expressionFlags = {11}, headerFlags = {12}, clientData = {13}, " +
                "arguments = {14}, exit = {15}, result = {16}{17}",
                GlobalState.GetCurrentThreadId(), debug, FormatOps.WrapOrNull(true, true, argv),
                code, breakpointType, FormatOps.WrapOrNull(breakpointName), (token != null),
                (traceInfo != null), FormatOps.WrapOrNull(engineFlags),
                FormatOps.WrapOrNull(substitutionFlags), FormatOps.WrapOrNull(eventFlags),
                FormatOps.WrapOrNull(expressionFlags), FormatOps.WrapOrNull(headerFlags),
                FormatOps.WrapOrNull(clientData), FormatOps.WrapOrNull(true, true, arguments),
                exit, FormatOps.WrapOrNull(true, true, result), Environment.NewLine),
                typeof(Interpreter).Name);

            return code;
        }
#endif
        #endregion
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Engine Context Members
        public int Levels /* WARNING: NOT GUARANTEED TO BE ACCURATE OR USEFUL. */
        {
            get
            {
                CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.Levels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return levels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.Levels = value;
#else
                // lock (syncRoot)
                {
                    levels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int PreviousLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.PreviousLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return previousLevels;
                }
#endif
            }
            set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.PreviousLevels = value;
#else
                // lock (syncRoot)
                {
                    previousLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ParserLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ParserLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return parserLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ParserLevels = value;
#else
                // lock (syncRoot)
                {
                    parserLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExpressionLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ExpressionLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return expressionLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ExpressionLevels = value;
#else
                // lock (syncRoot)
                {
                    expressionLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int UnknownLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.UnknownLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return unknownLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.UnknownLevels = value;
#else
                // lock (syncRoot)
                {
                    unknownLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int TraceLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.TraceLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return traceLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.TraceLevels = value;
#else
                // lock (syncRoot)
                {
                    traceLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUGGER
        internal IDebugger Debugger
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.Debugger;
                else
                    return null;
#else
                // lock (syncRoot)
                {
                    return debugger;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.Debugger = value;
#else
                // lock (syncRoot)
                {
                    debugger = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY || NOTIFY_OBJECT
        internal int NotifyLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.NotifyLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return notifyLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.NotifyLevels = value;
#else
                // lock (syncRoot)
                {
                    notifyLevels = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int PolicyLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.PolicyLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return policyLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.PolicyLevels = value;
#else
                // lock (syncRoot)
                {
                    policyLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int TestLevels
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.TestLevels;
                else
                    return 0;
#else
                // lock (syncRoot)
                {
                    return testLevels;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.TestLevels = value;
#else
                // lock (syncRoot)
                {
                    testLevels = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool StackOverflow
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.StackOverflow;
                else
                    return false;
#else
                lock (syncRoot)
                {
                    return stackOverflow;
                }
#endif
            }
            set
            {
                // CheckDisposed();

                //
                // NOTE: This property "setter" is only supposed to be used during
                //       interpreter initialization.
                //
#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.StackOverflow = value;
#else
                lock (syncRoot)
                {
                    stackOverflow = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if PREVIOUS_RESULT
        internal Result PreviousResult
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.PreviousResult;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return previousResult;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.PreviousResult = value;
#else
                lock (syncRoot)
                {
                    previousResult = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EngineFlags EngineFlags
        {
            get
            {
                // CheckDisposed();

                lock (syncRoot)
                {
#if THREADING
                    IEngineContext context = (contextManager != null) ?
                        contextManager.GetEngineContext(true) : null;

                    if (context != null)
                        return engineFlags | context.EngineFlags;
                    else
                        return engineFlags;
#else
                    return engineFlags;
#endif
                }
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.EngineFlags = value;
#else
                lock (syncRoot)
                {
                    engineFlags = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode ReturnCode
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ReturnCode;
                else
                    return ReturnCode.Ok;
#else
                lock (syncRoot)
                {
                    return returnCode;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ReturnCode = value;
#else
                lock (syncRoot)
                {
                    returnCode = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal string ErrorCode
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ErrorCode;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return errorCode;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ErrorCode = value;
#else
                lock (syncRoot)
                {
                    errorCode = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal string ErrorInfo
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ErrorInfo;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return errorInfo;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ErrorInfo = value;
#else
                lock (syncRoot)
                {
                    errorInfo = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ErrorFrames
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ErrorFrames;
                else
                    return 0;
#else
                lock (syncRoot)
                {
                    return errorFrames;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ErrorFrames = value;
#else
                lock (syncRoot)
                {
                    errorFrames = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal IScriptLocation ScriptLocation
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ScriptLocation;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return scriptLocation;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ScriptLocation = value;
#else
                lock (syncRoot)
                {
                    scriptLocation = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StackList<IScriptLocation> ScriptLocations
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ScriptLocations;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return scriptLocations;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ScriptLocations = value;
#else
                lock (syncRoot)
                {
                    scriptLocations = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int PreviousProcessId
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.PreviousProcessId;
                else
                    return 0;
#else
                lock (syncRoot)
                {
                    return previousProcessId;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.PreviousProcessId = value;
#else
                lock (syncRoot)
                {
                    previousProcessId = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ArraySearchDictionary ArraySearches
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.ArraySearches;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return arraySearches;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.ArraySearches = value;
#else
                lock (syncRoot)
                {
                    arraySearches = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if HISTORY
        internal ClientDataList _History
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    return context.History;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return _history;
                }
#endif
            }
            private set
            {
                // CheckDisposed();

#if THREADING
                IEngineContext context = (contextManager != null) ?
                    contextManager.GetEngineContext(true) : null;

                if (context != null)
                    context.History = value;
#else
                lock (syncRoot)
                {
                    _history = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EngineFlags BeginExternalExecution()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return context.BeginExternalExecution();
            else
                return EngineFlags.None;
#else
            EnterLevel();

            return Engine.AddStackCheckFlags(ref engineFlags);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void EndExternalExecution(
            EngineFlags savedEngineFlags
            )
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                context.EndExternalExecution(savedEngineFlags);
#else
            Engine.RemoveStackCheckFlags(savedEngineFlags, ref engineFlags);

            ExitLevel();
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int BeginNestedExecution()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return context.BeginNestedExecution();
            else
                return 0;
#else
            int savedPreviousLevels;

            lock (syncRoot)
            {
                savedPreviousLevels = previousLevels;
                previousLevels = levels;
            }

            return savedPreviousLevels;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void EndNestedExecution(
            int savedPreviousLevels
            )
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                context.EndNestedExecution(savedPreviousLevels);
#else
            lock (syncRoot)
            {
                previousLevels = savedPreviousLevels;
            }
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Support Methods
        #region Enter / Exit Context Levels
        internal int EnterLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.Levels;
            else
                return 0;
#else
            return Interlocked.Increment(ref levels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.Levels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref levels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterParserLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.ParserLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref parserLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitParserLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.ParserLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref parserLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterExpressionLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.ExpressionLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref expressionLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitExpressionLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.ExpressionLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref expressionLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterUnknownLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.UnknownLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref unknownLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitUnknownLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.UnknownLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref unknownLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterTraceLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.TraceLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref traceLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitTraceLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.TraceLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref traceLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NOTIFY || NOTIFY_OBJECT
        internal int EnterNotifyLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.NotifyLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref notifyLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitNotifyLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.NotifyLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref notifyLevels);
#endif
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterPolicyLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.PolicyLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref policyLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitPolicyLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.PolicyLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref policyLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int EnterTestLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return ++context.TestLevels;
            else
                return 0;
#else
            return Interlocked.Increment(ref testLevels);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal int ExitTestLevel()
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true) : null;

            if (context != null)
                return --context.TestLevels;
            else
                return 0;
#else
            return Interlocked.Decrement(ref testLevels);
#endif
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interpreter Readiness
        private ReturnCode GetReadyLevels(
            ref int levels,
            ref int parserLevels,
            ref int expressionLevels,
            ref Result error
            )
        {
#if THREADING
            IEngineContext context = (contextManager != null) ?
                contextManager.GetEngineContext(true, ref error) : null;

            if (context != null)
            {
                levels = context.Levels;
                parserLevels = context.ParserLevels;
                expressionLevels = context.ExpressionLevels;

                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
#else
            levels = this.levels;
            parserLevels = this.parserLevels;
            expressionLevels = this.expressionLevels;

            return ReturnCode.Ok;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Ready(
            Interpreter interpreter,
            ref Result error
            ) /* THREAD-SAFE */
        {
            return Ready(interpreter, true, false, false, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static ReturnCode Ready(
            Interpreter interpreter,
            bool stack,
            bool force,
            bool noCancel,
            ref Result error
            ) /* THREAD-SAFE */
        {
            if (interpreter == null)
            {
                error = "invalid interpreter";
                return ReturnCode.Error;
            }

            //
            // NOTE: Grab the interpreter flags without locking for speed.
            //
            InterpreterFlags interpreterFlags = interpreter.interpreterFlags; /* EXEMPT */

            //
            // NOTE: Bypass all interpreter readiness checks?
            //
            bool noReady = FlagOps.HasFlags(
                interpreterFlags, InterpreterFlags.NoReady, true);

            if (noReady)
                return ReturnCode.Ok;

            //
            // NOTE: Bypass all interpreter readiness checks except native stack
            //       checking?
            //
            bool stackOnly = FlagOps.HasFlags(
                interpreterFlags, InterpreterFlags.StackOnly, true);

            ReturnCode code;

            //
            // NOTE: Check if the interpreter is in an appropriate state of readiness
            //       to evaluate scripts and/or execute commands.
            //
            // BUGFIX: Do not use the SyncRoot property here (in case we have been
            //         disposed), use the field instead.
            //
            lock (interpreter.syncRoot) /* FIELD */
            {
                //
                // FIXME: We may need to rethink this code.  If the exit flag has been
                //        set in the interpreter, does that mean cleanup code (e.g.
                //        finally blocks, etc) should also be skipped?  Currently, the
                //        [exit] command (and the interpreter exit flag) in Eagle are
                //        more of a "suggestion" than anything else.  This is one place
                //        where we definiately do NOT follow (nor do we want to follow)
                //        the Tcl "standard" default behavior of immediately exiting
                //        the process upon the [exit] command.
                //
                if (!stackOnly)
                {
                    if (interpreter.exit)
                    {
                        error = "attempt to call eval in exited interpreter";
                        return ReturnCode.Error;
                    }

                    if (interpreter.disposed)
                    {
                        error = "attempt to call eval in disposed interpreter";
                        return ReturnCode.Error;
                    }

                    if (interpreter.deleted)
                    {
                        /* COMPAT: Tcl. */
                        error = "attempt to call eval in deleted interpreter";
                        return ReturnCode.Error;
                    }
                }

                //
                // NOTE: Checking native stack space is a relatively expensive
                //       operation; therefore, only perform this check if there
                //       are actual evaluations or expressions active AND it was
                //       explicitly requested by the caller.
                //
                int levels = 0;
                int parserLevels = 0;
                int expressionLevels = 0;

                //
                // NOTE: Fetch the current evaluation, parser, and expression
                //       nesting levels.  If this call fails, bail out now.
                //
                code = interpreter.GetReadyLevels(ref levels, ref parserLevels,
                    ref expressionLevels, ref error);

                if (code != ReturnCode.Ok)
                    return code;

#if NATIVE && WINDOWS
                if (stack &&
                    (force ||
                     (levels > NoStackLevels) ||
                     (parserLevels > NoStackParserLevels) ||
                     (expressionLevels > NoStackExpressionLevels)))
                {
                    //
                    // NOTE: Make sure that we are not running out of native stack
                    //       space.  If we cannot check native stack space, this
                    //       will simply return "success".
                    //
                    code = RuntimeOps.CheckForStackSpace(
                        interpreter, Engine.GetStackExtraSpace());

                    if (code != ReturnCode.Ok)
                    {
                        //
                        // HACK: Set the stack overflow flag for the interpreter so
                        //       the engine can "quickly" unwind the evaluation stack.
                        //
                        interpreter.StackOverflow = true;

                        //
                        // NOTE: Prefix the trace message with the method name here
                        //       because we are not producing a full stack trace.
                        //
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: Ready: out of stack space, stack = {1}, force = {2}, " +
                            "levels = {3}, parserLevels = {4}, expressionLevels = {5}{6}",
                            GlobalState.GetCurrentThreadId(), stack, force, levels, parserLevels,
                            expressionLevels, Environment.NewLine), typeof(Interpreter).Name);

                        /* COMPAT: Tcl. */
                        error = "out of stack space (infinite loop?)";
                        return code;
                    }
                }
#endif

                //
                // NOTE: If we only care about native stack checking, skip checking
                //       the recursion level.
                //
                if (!stackOnly)
                {
                    //
                    // NOTE: Make sure that we have no exceeded any predefined
                    //       recursion limits.  Access the field directly because
                    //       the interpreter lock is already being held.
                    //
                    int recursionLimit = interpreter.recursionLimit;

                    if ((recursionLimit > 0) && (levels > recursionLimit))
                    {
                        //
                        // HACK: Set the stack overflow flag for the interpreter so
                        //       the engine can "quickly" unwind the evaluation stack.
                        //
                        interpreter.StackOverflow = true;

                        /* COMPAT: Tcl. */
                        error = "too many nested evaluations (infinite loop?)";
                        return ReturnCode.Error;
                    }
                }
            }

            //
            // NOTE: If we only care about native stack checking, skip script
            //       cancellation checking.  Also skip it if specifically
            //       requested to do so by our caller.
            //
            if (!stackOnly && !noCancel)
            {
                //
                // NOTE: These may need to be treated as totally separate
                //       checks at some point, depending on where calls to
                //       ResetCancel and ResetHalt are present.
                //
                // BUGFIX: Do *NOT* do this while holding the interpreter
                //         synchronization lock.
                //
                code = Engine.IsCanceled(interpreter, true, ref error);

                if (code != ReturnCode.Ok)
                    return code;

                code = Engine.IsHalted(interpreter, true, ref error);

                if (code != ReturnCode.Ok)
                    return code;
            }

            return ReturnCode.Ok;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Script Location Support
        private bool ShouldPushScriptLocation(
            bool viaSource
            )
        {
            if (FlagOps.HasFlags(interpreterFlags, /* NOTE: Track all? */
                    InterpreterFlags.ScriptLocation, true))
            {
                return true;
            }

            return viaSource;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool ShouldPopScriptLocation(
            bool viaSource
            )
        {
            if (FlagOps.HasFlags(interpreterFlags, /* NOTE: Track all? */
                    InterpreterFlags.ScriptLocation, true))
            {
                return true;
            }

            return viaSource;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void PushScriptLocation(
            string fileName,
            bool viaSource
            )
        {
            PushScriptLocation(fileName,
                Parser.StartLine, Parser.UnknownLine, viaSource);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void PushScriptLocation(
            string fileName,
            int startLine,
            int endLine,
            bool viaSource
            )
        {
            //
            // HACK: Avoid allocating a ScriptLocation object that we will just
            //       end up throwing away.
            //
            if (!ShouldPushScriptLocation(viaSource))
                return;

            PushScriptLocation(new ScriptLocation(
                fileName, startLine, endLine, viaSource));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void PushScriptLocation(
            IScriptLocation location
            )
        {
            bool viaSource = EntityOps.IsViaSource(location);

            if (!ShouldPushScriptLocation(viaSource))
                return;

            //
            // NOTE: This lock is not necessary when we are compiled with
            //       the THREADING option enabled because ScriptLocation
            //       and ScriptLocations are per-thread values in that
            //       case.
            //
#if !THREADING
            lock (syncRoot) /* TRANSACTIONAL */
#endif
            {
                //
                // NOTE: Undo any previous manual override.
                //
                ScriptLocation = null;
                ScriptLocations.Push(location);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void PopScriptLocation(
            IScriptLocation location
            )
        {
            PopScriptLocation(EntityOps.IsViaSource(location));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void PopScriptLocation(
            bool viaSource
            )
        {
            if (!ShouldPopScriptLocation(viaSource))
                return;

            //
            // NOTE: This lock is not necessary when we are compiled with
            //       the THREADING option enabled because ScriptLocations
            //       is a per-thread value in that case.
            //
#if !THREADING
            lock (syncRoot) /* TRANSACTIONAL */
#endif
            {
                ScriptLocations.Pop();
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool HasErrorInProgress()
        {
            //
            // WARNING: Possibly per-thread engine flags used here,
            //          MUST use property, not field.
            //
            return Engine.HasErrorInProgress(EngineFlags);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool CleanupArraySearches(IVariable variable)
        {
            int removed = 0;

            return CleanupArraySearches(variable, ref removed);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool CleanupArraySearches(
            IVariable variable,
            ref int removed
            )
        {
            if (variable != null)
            {
                ArraySearchDictionary arraySearches = ArraySearches;

                if (arraySearches != null)
                {
                    //
                    // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                    //       from the objects collection in the containing interpreter.
                    //
                    StringList keys = new StringList();

                    foreach (KeyValuePair<string, ArraySearch> pair in arraySearches)
                        if (pair.Value != null)
                            if (Object.ReferenceEquals(pair.Value.Variable, variable))
                                keys.Add(pair.Key);

                    //
                    // NOTE: Cleanup, PASS #2, remove all the objects we found from PASS
                    //       #1 (above).
                    //
                    foreach (string key in keys)
                        LogicOps.Y(arraySearches.Remove(key), removed++);

                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int CountThreadInterpreters()
        {
#if THREADING
            try
            {
                if (contextManager != null)
                    return contextManager.GetEngineContextCount();
            }
            catch
            {
                // do nothing.
            }
#else
            InterpreterDictionary interpreters = GlobalState.GetInterpreters();

            if (interpreters != null)
            {
                int count = 0;

                foreach (KeyValuePair<string, Interpreter> pair in interpreters)
                    if ((pair.Value != null) && pair.Value.IsPrimaryThread())
                        count++;

                return count;
            }
#endif

            return 0;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Interactive Context Members
        internal int InteractiveLoops
        {
            get
            {
                // CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    return context.InteractiveLoops;
                else
                    return 0;
#else
                lock (syncRoot)
                {
                    return interactiveLoops;
                }
#endif
            }
            set
            {
                // CheckDisposed();

#if THREADING
                IInteractiveContext context = (contextManager != null) ?
                    contextManager.GetInteractiveContext(true) : null;

                if (context != null)
                    context.InteractiveLoops = value;
#else
                lock (syncRoot)
                {
                    interactiveLoops = value;
                }
#endif

                if (IsPrimaryThread())
                {
                    //
                    // NOTE: Do not create the variable if we have been prohibited
                    //       from doing so.
                    //
                    VariableFlags flags = VariableFlags.GlobalOnly | VariableFlags.NoReady;

                    if (FlagOps.HasFlags(createFlags, CreateFlags.Safe, true) ||
                        FlagOps.HasFlags(createFlags, CreateFlags.NoVariables, true))
                    {
                        flags |= VariableFlags.NoCreate;
                    }

                    //
                    // NOTE: Update script-level informational variable.
                    //
                    SetVariableValue(flags, TclVars.InteractiveLoops, value.ToString(), null);
                }
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Test Context Members
        internal int[] TestStatistics
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Statistics;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testStatistics;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Statistics = value;
#else
                lock (syncRoot)
                {
                    testStatistics = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringList TestConstraints
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Constraints;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testConstraints;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Constraints = value;
#else
                lock (syncRoot)
                {
                    testConstraints = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringList TestSkipped
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Skipped;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testSkipped;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Skipped = value;
#else
                lock (syncRoot)
                {
                    testSkipped = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringList TestFailures
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Failures;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testFailures;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Failures = value;
#else
                lock (syncRoot)
                {
                    testFailures = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal IntDictionary TestCounts
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Counts;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testCounts;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Counts = value;
#else
                lock (syncRoot)
                {
                    testCounts = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringList TestMatch
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Match;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testMatch;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Match = value;
#else
                lock (syncRoot)
                {
                    testMatch = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringList TestSkip
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Skip;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testSkip;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Skip = value;
#else
                lock (syncRoot)
                {
                    testSkip = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCodeDictionary TestReturnCodeMessages
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.ReturnCodeMessages;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testReturnCodeMessages;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.ReturnCodeMessages = value;
#else
                lock (syncRoot)
                {
                    testReturnCodeMessages = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUGGER
        internal StringDictionary TestBreakpoints
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Breakpoints;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testBreakpoints;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Breakpoints = value;
#else
                lock (syncRoot)
                {
                    testBreakpoints = value;
                }
#endif
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal string TestPath
        {
            get
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    return context.Path;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return testPath;
                }
#endif
            }
            set
            {
#if THREADING
                ITestContext context = (contextManager != null) ?
                    contextManager.GetTestContext(true) : null;

                if (context != null)
                    context.Path = value;
#else
                lock (syncRoot)
                {
                    testPath = value;
                }
#endif
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Variable Context Members
        internal CallStack CallStack /* THREAD-SAFE */
        {
            get
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    return context.CallStack;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return callStack;
                }
#endif
            }
            private set
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    context.CallStack = value;
#else
                lock (syncRoot)
                {
                    callStack = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame GlobalFrame /* THREAD-SAFE */
        {
            get
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    return context.GlobalFrame;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return globalFrame;
                }
#endif
            }
            private set
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    context.GlobalFrame = value;
#else
                lock (syncRoot)
                {
                    globalFrame = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame CurrentFrame /* THREAD-SAFE */
        {
            get
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    return context.CurrentFrame;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return currentFrame;
                }
#endif
            }
            private set
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    context.CurrentFrame = value;
#else
                lock (syncRoot)
                {
                    currentFrame = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame ProcedureFrame /* THREAD-SAFE */
        {
            get
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    return context.ProcedureFrame;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return procedureFrame;
                }
#endif
            }
            /* internal */
            set
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    context.ProcedureFrame = value;
#else
                lock (syncRoot)
                {
                    procedureFrame = value;
                }
#endif
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ITraceInfo TraceInfo /* THREAD-SAFE */
        {
            get
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    return context.TraceInfo;
                else
                    return null;
#else
                lock (syncRoot)
                {
                    return traceInfo;
                }
#endif
            }
            set
            {
#if THREADING
                IVariableContext context = (contextManager != null) ?
                    contextManager.GetVariableContext(true) : null;

                if (context != null)
                    context.TraceInfo = value;
#else
                lock (syncRoot)
                {
                    traceInfo = value;
                }
#endif
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICallFrame Management
        internal bool IsCallFrameInCallStack(ICallFrame frame)
        {
            lock (syncRoot)
            {
                if (CallStack != null)
                    return CallStack.Contains(frame);
                else
                    return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsVariableCallFrame(ICallFrame frame)
        {
            return (frame != null) ?
                FlagOps.HasFlags(frame.Flags, CallFrameFlags.Variables, false) &&
                CallFrameOps.IsSame(frame,
                    GetVariableFrame(null, VariableFlags.None)) : false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsGlobalCallFrame(ICallFrame frame)
        {
            return CallFrameOps.IsSame(frame, GlobalFrame);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsCurrentCallFrame(ICallFrame frame)
        {
            return CallFrameOps.IsSame(frame, CurrentFrame);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsProcedureCallFrame(ICallFrame frame)
        {
            return CallFrameOps.IsSame(frame, ProcedureFrame);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame GetProcedureFrame() /* CANNOT RETURN NULL */
        {
            lock (syncRoot)
            {
                ICallFrame frame = ProcedureFrame;

                return (frame != null) ? frame : GlobalFrame;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame GetVariableFrame(
            ICallFrame frame,
            ref string varName
            ) /* CANNOT RETURN NULL */
        {
            VariableFlags flags = VariableFlags.None;

            return GetVariableFrame(frame, ref varName, ref flags);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame GetVariableFrame(
            ICallFrame frame,
            ref string varName,
            ref VariableFlags flags
            ) /* CANNOT RETURN NULL */
        {
            EntityOps.GetFlags(ref varName, ref flags);

            return GetVariableFrame(frame, flags);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame GetVariableFrame(
            ICallFrame frame,
            VariableFlags flags
            ) /* CANNOT RETURN NULL */
        {
            ICallFrame variableFrame;

            if (FlagOps.HasFlags(flags, VariableFlags.GlobalOnly, true))
            {
                variableFrame = GlobalFrame;
            }
            else
            {
                if (frame != null)
                    variableFrame = CallFrameOps.FollowLinks(frame);
                else
                    variableFrame = CallFrameOps.FollowLinks(CurrentFrame);

                if (variableFrame == null)
                    variableFrame = GlobalFrame;
            }

            return variableFrame;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void GetInfoLevelArguments(
            string subCommand,
            ref CallFrameFlags flags,
            ref ICallFrame frame
            ) /* HACK */
        {
            flags |= CallFrameFlags.Invisible;
            frame = CurrentFrame;

            //
            // NOTE: If we are not performing work for the generic
            //       "level" sub-command then always allow invisible
            //       call frames to be seen; otherwise, only allow
            //       invisible call frames to be seen if we are within
            //       a call frame that has the appropriate flag set.
            //
            if (!String.Equals(subCommand, CallFrameOps.InfoLevelSubCommand,
                    StringOps.SystemStringComparisonType) ||
                CallFrameOps.Find(CallStack, frame, false, 0,
                    CallFrameFlags.NoInvisible, CallFrameFlags.None,
                    false, false) == ReturnCode.Ok)
            {
                flags &= ~CallFrameFlags.Invisible;
                frame = CallFrameOps.FollowLinks(frame);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Peek / Push / Pop
        #region Peek
        internal ICallFrame PeekCallFrame()
        {
#if !THREADING
            lock (syncRoot)
#endif
            {
                return CallFrameOps.Peek(CallStack);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Push
        internal void PushCallFrame(ICallFrame newFrame)
        {
            ICallFrame frame = null;

#if !THREADING
            lock (syncRoot)
#endif
            {
                //
                // NOTE: Push the call frame.
                //
                CallFrameOps.Push(CallStack, newFrame, ref frame);

                //
                // NOTE: Set the current call frame to the one we just pushed.
                //
                CurrentFrame = frame;
            }

#if NOTIFY
            /* IGNORED */
            CheckNotifications(
                NotifyType.CallFrame, NotifyFlags.Pushed,
                newFrame, this,
                null, null, null);
#endif
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Pop
        internal ICallFrame PopCallFrame()
        {
            ICallFrame newFrame = null;
            ICallFrame frame = null;

#if !THREADING
            lock (syncRoot)
#endif
            {
                //
                // NOTE: Pop the call frame.
                //
                newFrame = CallFrameOps.Pop(CallStack, ref frame);

                //
                // NOTE: Set the current call frame to the one we just popped.
                //
                CurrentFrame = frame;
            }

#if NOTIFY || NOTIFY_OBJECT
            /* IGNORED */
            CheckNotifications(
                NotifyType.CallFrame, NotifyFlags.Popped,
                newFrame, this,
                null, null, null);
#endif

            return newFrame;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame PopCallFrame(
            CallFrameFlags hasFlags,
            bool all
            )
        {
            ICallFrame newFrame = PopCallFrames(hasFlags, all);

            if (newFrame != null)
                newFrame = PopCallFrame();

            return newFrame;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame PopCallFrames(
            CallFrameFlags hasFlags,
            bool all
            )
        {
            ICallFrame newFrame = null;

#if !THREADING
            lock (syncRoot)
#endif
            {
                bool done = false;

                do
                {
                    newFrame = PeekCallFrame();

                    if ((newFrame != null) &&
                        FlagOps.HasFlags(newFrame.Flags, hasFlags, all))
                    {
                        PopCallFrame();
                    }
                    else
                    {
                        done = true;
                    }
                }
                while (!done);
            }

            return newFrame;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Global (Engine)
        internal void PushGlobalCallFrame(
            bool mark
            )
        {
            if (mark)
            {
                //
                // BUGFIX: When pushing the global call frame from the engine we
                //         must make sure this is treated just like [uplevel #0]
                //         (i.e. the other call frames must be marked as invisible
                //         for [info level], etc to work properly).
                //
                ReturnCode code;
                Result error = null;

                code = CallFrameOps.MarkMatching(
                    CallStack, CurrentFrame, true, 0, CallFrameFlags.Variables,
                    CallFrameFlags.Invisible | CallFrameFlags.Tracking,
                    CallFrameFlags.Invisible, false, false, true, ref error);

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error); // BUGBUG: Panic!

                ICallFrame frame = NewLinkedCallFrame(
                    "global (engine)", null, CallFrameFlags.Uplevel, GlobalFrame);

                PushCallFrame(frame);

                // PushCallFrame(GlobalFrame);
            }
            else
            {
                PushCallFrame(GlobalFrame);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void PopGlobalCallFrame(
            bool mark
            )
        {
            PopCallFrame(CallFrameFlags.Scope, true);

            if (mark)
            {
                ReturnCode code;
                Result error = null;

                code = CallFrameOps.MarkMatching(
                    CallStack, CurrentFrame, true, 0, CallFrameFlags.Variables,
                    CallFrameFlags.Tracking, CallFrameFlags.Invisible, false,
                    false, false, ref error);

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error); // BUGBUG: Panic!
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICallFrame (Scope)
        internal ReturnCode GetScopeCallFrame(
            string name,
            bool pop,
            bool remove,
            ref ICallFrame frame
            )
        {
            Result error = null;

            return GetScopeCallFrame(name, pop, remove, ref frame, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetScopeCallFrame(
            string name,
            bool pop,
            bool remove,
            ref ICallFrame frame,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                if (HasScopes(ref error))
                {
                    ICallFrame newFrame = PeekCallFrame();

                    //
                    // NOTE: We cannot follow linked call frames if we
                    //       intend to pop the innermost scope call frame.
                    //
                    if (!pop)
                        newFrame = CallFrameOps.FollowLinks(newFrame);

                    //
                    // NOTE: Is the current call frame valid?
                    //
                    if (newFrame != null)
                    {
                        //
                        // NOTE: Is this a [scope] created persistent call frame
                        //       (i.e. one that was created by [scope], but not by
                        //       [scope eval]).
                        //
                        if (FlagOps.HasFlags(newFrame.Flags, CallFrameFlags.Scope, true) &&
                            !FlagOps.HasFlags(newFrame.Flags, CallFrameFlags.Evaluate, true))
                        {
                            //
                            // NOTE: Does the provided name, if any, match?
                            //
                            if ((name == null) ||
                                (String.Compare(newFrame.Name, name, StringOps.SystemStringComparisonType) == 0))
                            {
                                //
                                // NOTE: Pop the call frame from the call stack?
                                //
                                if (pop)
                                    //
                                    // NOTE: We know the call stack has not changed here
                                    //       because we are hodling the interpreter lock.
                                    //
                                    frame = PopCallFrame(); // return and remove it.
                                else
                                    frame = newFrame; // just return it.

                                //
                                // NOTE: Remove the call frame from the interpreter?
                                //
                                if (remove)
                                    scopes.Remove(newFrame.Name);

                                return ReturnCode.Ok;
                            }
                            else
                            {
                                error = String.Format(
                                    "scope \"{0}\" is not innermost",
                                    name);
                            }
                        }
                        else
                        {
                            error = "no scopes are open";
                        }
                    }
                    else
                    {
                        error = "invalid call frame";
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CopyProcedureArgumentsToFrame(
            ICallFrame frame,
            bool force,
            ref Result error
            )
        {
            ArgumentList frameArguments = null;

            lock (syncRoot)
            {
                ICallFrame procedureFrame = ProcedureFrame;

                //
                // NOTE: Steal the arguments from the containing procedure;
                //       otherwise, we cannot access them from the [scope].
                //
                if ((procedureFrame != null) &&
                    (procedureFrame.ProcedureArguments != null))
                {
                    frameArguments = new ArgumentList(
                        procedureFrame.ProcedureArguments);
                }
            }

            ReturnCode code = ReturnCode.Ok;

            if (frameArguments != null)
            {
                //
                // NOTE: Create variables for each of the formal arguments
                //       to the enclosing procedure.  Start the counting at
                //       one to skip the procedure name itself.
                //
                for (int argumentIndex = 1;
                        argumentIndex < frameArguments.Count; argumentIndex++)
                {
                    string varName = frameArguments[argumentIndex].Name;

                    if (force || !frame.Variables.ContainsKey(varName))
                    {
                        string varValue = null;

                        if (frameArguments[argumentIndex].Value != null)
                            varValue = frameArguments[argumentIndex].Value.ToString();

                        code = SetVariableValue2(VariableFlags.Argument, frame,
                            varName, null, varValue, null, ref error);

                        if (code != ReturnCode.Ok)
                            break;
                    }
                }

                //
                // NOTE: Finally, update the arguments for the call frame so
                //       that things like [info level] will work properly.
                //
                if (code == ReturnCode.Ok)
                    frame.Arguments = frameArguments;
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ICallFrame (Factory)
        internal ICallFrame NewProcedureCallFrame(
            string name,
            ObjectDictionary tags,
            CallFrameFlags flags,
            IExecute execute,
            ArgumentList arguments
            )
        {
            return new CallFrame(
                NextId(), name, tags, CallStack.Count + 1,
                Levels, flags, execute, arguments);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame NewLinkedCallFrame(
            string name,
            ObjectDictionary tags,
            CallFrameFlags flags
            )
        {
            return NewLinkedCallFrame(name, tags, flags, CurrentFrame);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame NewLinkedCallFrame(
            string name,
            ObjectDictionary tags,
            CallFrameFlags flags,
            ICallFrame frame
            )
        {
            return new CallFrame(
                NextId(), name, tags, CallStack.Count + 1,
                Levels, flags, frame);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ICallFrame NewCallFrame(
            string name,
            ObjectDictionary tags,
            CallFrameFlags flags,
            VariableDictionary variables,
            ArgumentList arguments
            )
        {
            return new CallFrame(
                NextId(), name, tags, CallStack.Count + 1,
                Levels, flags, variables, arguments);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Flags Management
        internal EventFlags AfterEventFlags /* FOR [after] USE ONLY */
        {
            get { return afterEventFlags; }
            set { afterEventFlags = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal CreateFlags CreateFlags
        {
            get { lock (syncRoot) { return createFlags; } }
            set { lock (syncRoot) { createFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags EngineEventFlags
        {
            get { lock (syncRoot) { return engineEventFlags; } }
            set { lock (syncRoot) { engineEventFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal HeaderFlags HeaderFlags
        {
            get { return headerFlags; }
            set { headerFlags = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EngineFlags InteractiveEngineFlags
        {
            get { lock (syncRoot) { return interactiveEngineFlags; } }
            set { lock (syncRoot) { interactiveEngineFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags InteractiveEventFlags
        {
            get { lock (syncRoot) { return interactiveEventFlags; } }
            set { lock (syncRoot) { interactiveEventFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ExpressionFlags InteractiveExpressionFlags
        {
            get { lock (syncRoot) { return interactiveExpressionFlags; } }
            set { lock (syncRoot) { interactiveExpressionFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal SubstitutionFlags InteractiveSubstitutionFlags
        {
            get { lock (syncRoot) { return interactiveSubstitutionFlags; } }
            set { lock (syncRoot) { interactiveSubstitutionFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal InterpreterFlags InterpreterFlags
        {
            get { lock (syncRoot) { return interpreterFlags; } }
            set { lock (syncRoot) { interpreterFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal PluginFlags PluginFlags
        {
            get { lock (syncRoot) { return pluginFlags; } }
            set { lock (syncRoot) { pluginFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags QueueEventFlags
        {
            get { lock (syncRoot) { return queueEventFlags; } }
            private set { lock (syncRoot) { queueEventFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal SubstitutionFlags SubstitutionFlags
        {
            get { lock (syncRoot) { return substitutionFlags; } }
            set { lock (syncRoot) { substitutionFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ExpressionFlags ExpressionFlags
        {
            get { lock (syncRoot) { return expressionFlags; } }
            set { lock (syncRoot) { expressionFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags UpdateEventFlags
        {
            get { lock (syncRoot) { return updateEventFlags; } }
            private set { lock (syncRoot) { updateEventFlags = value; } }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal EventFlags WaitEventFlags
        {
            get { lock (syncRoot) { return waitEventFlags; } }
            private set { lock (syncRoot) { waitEventFlags = value; } }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Cache Management
#if PARSE_CACHE
        #region ExpressionState Cache
        #region Dead Code
#if DEAD_CODE
        internal bool GetCachedExpressionState( /* NOT USED */
            string text,
            ref IExpressionState exprState
            )
        {
            //
            // NOTE: Part of the experimental expression parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ExpressionState, true))
                {
                    if ((text != null) && (expressionStateCache != null))
                    {
                        if (expressionStateCache.TryGetValue(text, out exprState))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref expressionStateCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return true;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref expressionStateCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool AddCachedExpressionState( /* NOT USED */
            IExpressionState exprState
            )
        {
            //
            // NOTE: Part of the experimental expression parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ExpressionState, true))
                {
                    if ((exprState != null) && (exprState.ParseState != null) &&
                        (exprState.ParseState.Text != null) && (expressionStateCache != null) &&
                        !expressionStateCache.ContainsKey(exprState.ParseState.Text))
                    {
                        if (expressionStateCache.Count > MaximumCacheSize)
                        {
#if CACHE_DICTIONARY
                            expressionStateCache.TrimExcess(
                                MinimumCacheSize, MaximumCacheSize);
#else
                            expressionStateCache.Clear();
#endif
                        }

                        if (exprState.ParseState.Text.Length > MaximumCacheItemSize)
                            return false;

                        if ((exprState.ParseState.Tokens != null) &&
                            (exprState.ParseState.Tokens.Count > MaximumCacheItemSize))
                        {
                            return false;
                        }

                        expressionStateCache.Add(exprState.ParseState.Text, exprState);
                        return true;
                    }
                }
            }

            return false;
        }
#endif
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region ParseState Cache
        internal bool GetCachedParseState(
            string text,
            ref IParseState parseState
            )
        {
            //
            // NOTE: Part of the experimental parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ParseState, true))
                {
                    if ((text != null) && (parseStateCache != null))
                    {
                        if (parseStateCache.TryGetValue(text, out parseState))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref parseStateCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return true;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref parseStateCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool AddCachedParseState(
            IParseState parseState
            )
        {
            //
            // NOTE: Part of the experimental parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ParseState, true))
                {
                    if ((parseState != null) && (parseState.Text != null) &&
                        (parseStateCache != null) &&
                        !parseStateCache.ContainsKey(parseState.Text))
                    {
                        if (parseStateCache.Count > MaximumCacheSize)
                        {
#if CACHE_DICTIONARY
                            parseStateCache.TrimExcess(
                                MinimumCacheSize, MaximumCacheSize);
#else
                            parseStateCache.Clear();
#endif
                        }

                        if (parseState.Text.Length > MaximumCacheItemSize)
                            return false;

                        if ((parseState.Tokens != null) &&
                            (parseState.Tokens.Count > MaximumCacheItemSize))
                        {
                            return false;
                        }

                        parseStateCache.Add(parseState.Text, parseState);
                        return true;
                    }
                }
            }

            return false;
        }
        #endregion
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if LIST_CACHE
        #region StringList Cache
        internal bool GetCachedStringList(
            string text,
            ref StringList list
            )
        {
            //
            // NOTE: Part of the experimental parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.StringList, true))
                {
                    if ((text != null) && (stringListCache != null))
                    {
                        if (stringListCache.TryGetValue(text, out list))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref stringListCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return true;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref stringListCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool AddCachedStringList(
            string text,
            StringList list
            )
        {
            //
            // NOTE: Part of the experimental parser caching mechanism.
            //
            //       No speed improvement was observed; however, further analysis
            //       and profiling may be warranted.
            //
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.StringList, true))
                {
                    if ((text != null) && (stringListCache != null) &&
                        !stringListCache.ContainsKey(text))
                    {
                        if (stringListCache.Count >= MaximumCacheSize)
                        {
#if CACHE_DICTIONARY
                            stringListCache.TrimExcess(
                                MinimumCacheSize, MaximumCacheSize);
#else
                            stringListCache.Clear();
#endif
                        }

                        if (text.Length > MaximumCacheItemSize)
                            return false;

                        if (list.Count > MaximumCacheItemSize)
                            return false;

                        stringListCache.Add(text, list);
                        return true;
                    }
                }
            }

            return false;
        }
        #endregion
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if TYPE_CACHE
        #region Type Cache
        internal bool GetCachedType(
            string name,
            ref Type type
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.Type, true))
                {
                    if ((name != null) && (typeCache != null))
                    {
                        if (typeCache.TryGetValue(name, out type))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref typeCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return true;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref typeCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool AddCachedType(
            string name,
            Type type
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.Type, true))
                {
                    if ((name != null) && (typeCache != null) &&
                        !typeCache.ContainsKey(name))
                    {
                        if (typeCache.Count >= MaximumCacheSize)
                        {
#if CACHE_DICTIONARY
                        typeCache.TrimExcess(
                            MinimumCacheSize, MaximumCacheSize);
#else
                            typeCache.Clear();
#endif
                        }

                        typeCache.Add(name, type);
                        return true;
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool RemoveCachedType(
            string name
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.Type, true))
                {
                    if ((name != null) && (typeCache != null) &&
                        typeCache.ContainsKey(name))
                    {
                        return typeCache.Remove(name);
                    }
                }
            }

            return false;
        }
        #endregion
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if COM_TYPE_CACHE
        #region ComTypeList Cache
        internal bool GetCachedComTypeList(
            IntPtr unknown,
            ref TypeList types
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ComTypeList, true))
                {
                    if ((unknown != IntPtr.Zero) && (comTypeListCache != null))
                    {
                        if (comTypeListCache.TryGetValue(unknown, out types))
                        {
#if CACHE_STATISTICS
                            Interlocked.Increment(
                                ref comTypeListCache.cacheCounts[(int)CacheCountType.Hit]);
#endif

                            return true;
                        }
#if CACHE_STATISTICS
                        else
                        {
                            Interlocked.Increment(
                                ref comTypeListCache.cacheCounts[(int)CacheCountType.Miss]);
                        }
#endif
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool AddCachedComTypeList(
            IntPtr unknown,
            TypeList types
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ComTypeList, true))
                {
                    if ((unknown != IntPtr.Zero) && (comTypeListCache != null) &&
                        !comTypeListCache.ContainsKey(unknown))
                    {
                        if (comTypeListCache.Count >= MaximumCacheSize)
                        {
#if CACHE_DICTIONARY
                            comTypeListCache.TrimExcess(
                                MinimumCacheSize, MaximumCacheSize);
#else
                            comTypeListCache.Clear();
#endif
                        }

                        comTypeListCache.Add(unknown, types);
                        return true;
                    }
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool RemoveCachedComTypeList(
            IntPtr unknown
            )
        {
            lock (syncRoot)
            {
                if (FlagOps.HasFlags(cacheFlags, CacheFlags.ComTypeList, true))
                {
                    if ((unknown != IntPtr.Zero) && (comTypeListCache != null) &&
                        comTypeListCache.ContainsKey(unknown))
                    {
                        return comTypeListCache.Remove(unknown);
                    }
                }
            }

            return false;
        }
        #endregion
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Core Trace Callbacks
        #region AutoPath Trace Callback
        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode AutoPathTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            //
            // NOTE: Intercept writes to the variable "auto_path" and figure out if we
            //       need to rescan for package indexes.
            //
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    //
                    // NOTE: Make sure that an array element index was not specified.
                    //
                    //       *WARNING* Empty array element names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (traceInfo.Index == null)
                    {
                        switch (breakpointType)
                        {
                            case BreakpointType.BeforeVariableSet:
                                {
                                    PackageIndexDictionary packageIndexes;

                                    lock (interpreter.SyncRoot)
                                    {
                                        packageIndexes = interpreter.PackageIndexes;

                                        if (packageIndexes != null)
                                            packageIndexes = new PackageIndexDictionary(packageIndexes);
                                    }

                                    //
                                    // BUGFIX: Save and restore the current [shared] traceInfo object [for
                                    //         this thread] because we may evaluate an arbitrary number of
                                    //         scripts during our package index scan and any variable
                                    //         accesses will cause the [shared] traceInfo object for this
                                    //         thread to change out from under us.
                                    //
                                    ITraceInfo oldTraceInfo = traceInfo.Copy();

                                    try
                                    {
                                        //
                                        // NOTE: We do not really want to force a refresh here; however,
                                        //       since we have no easy way of actually knowing which
                                        //       package index file and by logical extension, which
                                        //       directory, provided the "ifneeded" commands for a given
                                        //       package, we must refresh all directories to be accurate.
                                        //
                                        StringList list = null;

                                        if ((Parser.SplitList(interpreter,
                                                RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                0, Length.Invalid, false, true, ref list,
                                                ref result) == ReturnCode.Ok) &&
                                            (RuntimeOps.FindAllPackages(
                                                interpreter, list, true, true, false, false,
                                                ref packageIndexes, ref result) == ReturnCode.Ok))
                                        {
                                            lock (interpreter.SyncRoot)
                                            {
                                                interpreter.PackageIndexes = packageIndexes;
                                            }
                                        }
                                        else
                                        {
                                            return ReturnCode.Error;
                                        }
                                    }
                                    finally
                                    {
                                        traceInfo.Update(oldTraceInfo);
                                    }
                                    break;
                                }
                            case BreakpointType.BeforeVariableUnset:
                                {
                                    //
                                    // NOTE: They want to unset the auto_path variable.  I guess
                                    //       this should be allowed, it is in Tcl.  Also, make
                                    //       sure it is purged from the call frame so that it
                                    //       cannot be magically restored with this trace
                                    //       callback in place.
                                    //
                                    traceInfo.Flags &= ~VariableFlags.NoRemove;

                                    //
                                    // BUGFIX: Yes, also make sure any waiters are cleared.
                                    //
                                    EntityOps.SetDirty(traceInfo.Variable, true);

                                    //
                                    // NOTE: Ok, it appears that Tcl handles this by forgetting
                                    //       all package index information; therefore, we will
                                    //       emulate that behavior.
                                    //
                                    if (interpreter.ResetPkgIndexes(ref result) != ReturnCode.Ok)
                                        return ReturnCode.Error;

                                    break;
                                }
                        }

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        //
                        // NOTE: We expect "auto_path" to always be a scalar list variable.
                        //
                        result = String.Format("can't {0} {1}: variable isn't array",
                            FormatOps.Breakpoint(breakpointType),
                            FormatOps.ErrorVariableName(traceInfo.Name, traceInfo.Index));
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Environment Trace Callback
        #region Support Methods
        private static bool HasEnvironmentTraceCallback(
            IVariable variable
            )
        {
            return (variable != null) &&
                (variable.Traces != null) &&
                variable.Traces.Contains(EnvironmentTraceCallback);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsEnvironmentVariable(IVariable variable)
        {
            return (variable != null) &&
                IsGlobalCallFrame(variable.Frame) &&
                (String.Compare(variable.Name, TclVars.Environment,
                    StringOps.SystemStringComparisonType) == 0) &&
                HasEnvironmentTraceCallback(variable);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode EnvironmentTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            //
            // NOTE: Intercept reads and writes to the variable "env" and redirect
            //       them to get and set environment variables.
            //
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    //
                    // NOTE: *SPECIAL* Ignore the index when we initially add the variable
                    //       since we do not perform any trace actions during add anyhow.
                    //
                    //       *WARNING* Empty array element names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if ((breakpointType == BreakpointType.BeforeVariableAdd) ||
                        (traceInfo.Index != null))
                    {
                        string oldValue = null;

                        bool exists = (traceInfo.Index != null) ?
                            CommonOps.DoesEnvironmentVariableExist(
                                traceInfo.Index, ref oldValue) : false;

                        switch (breakpointType)
                        {
                            case BreakpointType.BeforeVariableGet:
                                {
                                    if (exists)
                                    {
                                        try
                                        {
                                            result = oldValue;

                                            traceInfo.OldValue = oldValue;
                                            traceInfo.Cancel = true;
                                            traceInfo.ReturnCode = ReturnCode.Ok;
                                        }
                                        catch (Exception e)
                                        {
                                            result = e;
                                            traceInfo.ReturnCode = ReturnCode.Error;
                                        }
                                    }
                                    else
                                    {
                                        result = String.Format(
                                            "can't read {0}: no such element in array",
                                            FormatOps.ErrorVariableName(traceInfo.Name, traceInfo.Index));
                                    }
                                    break;
                                }
                            case BreakpointType.BeforeVariableSet:
                                {
                                    try
                                    {
                                        string newValue = RuntimeOps.GetStringFromObject(traceInfo.NewValue);

                                        CommonOps.SetEnvironmentVariable(traceInfo.Index, newValue);

                                        result = newValue;

                                        EntityOps.SetUndefined(traceInfo.Variable, false);
                                        EntityOps.SetDirty(traceInfo.Variable, true);

                                        //
                                        // BUGFIX: Must set the old and new values in case they
                                        //         are opaque object handles.
                                        //
                                        traceInfo.OldValue = oldValue;
                                        traceInfo.NewValue = newValue;
                                        traceInfo.Cancel = true;
                                        traceInfo.ReturnCode = ReturnCode.Ok;
                                    }
                                    catch (Exception e)
                                    {
                                        result = e;
                                        traceInfo.ReturnCode = ReturnCode.Error;
                                    }
                                    break;
                                }
                            case BreakpointType.BeforeVariableUnset:
                                {
                                    if (exists)
                                    {
                                        try
                                        {
                                            CommonOps.SetEnvironmentVariable(traceInfo.Index, null);

                                            result = String.Empty;

                                            EntityOps.SetDirty(traceInfo.Variable, true);

                                            //
                                            // BUGFIX: Must set the old value in case it was an
                                            //         opaque object handle.
                                            //
                                            traceInfo.OldValue = oldValue;
                                            traceInfo.Cancel = true;
                                            traceInfo.ReturnCode = ReturnCode.Ok;
                                        }
                                        catch (Exception e)
                                        {
                                            result = e;
                                            traceInfo.ReturnCode = ReturnCode.Error;
                                        }
                                    }
                                    else if (!FlagOps.HasFlags(
                                            traceInfo.Flags, VariableFlags.NoComplain, true))
                                    {
                                        result = String.Format(
                                            "can't unset {0}: no such element in array",
                                            FormatOps.ErrorVariableName(traceInfo.Name, traceInfo.Index));
                                    }
                                    else
                                    {
                                        //
                                        // BUGFIX: Do not return an error if the "NoComplain"
                                        //         flag has been set by the caller.
                                        //
                                        // NOTE: The default handling sets an error message
                                        //       into the result for this case and we do not
                                        //       want that; therefore, cancel the default
                                        //       action.
                                        //
                                        traceInfo.Cancel = true;
                                        traceInfo.ReturnCode = ReturnCode.Ok;
                                    }
                                    break;
                                }
                        }

                        return traceInfo.ReturnCode;
                    }
                    else if (breakpointType == BreakpointType.BeforeVariableUnset)
                    {
                        //
                        // NOTE: They want to unset the entire env array.  I guess
                        //       this should be allowed, it is in Tcl.  Also, make
                        //       sure it is purged from the call frame so that it
                        //       cannot be magically restored with this trace
                        //       callback in place.
                        //
                        traceInfo.Flags &= ~VariableFlags.NoRemove;

                        //
                        // BUGFIX: Must make the entire array "concrete" now just
                        //         in case there are any opaque object handles
                        //         that need to be picked up by later traces (e.g.
                        //         ObjectTraceCallback).
                        //
                        IVariable variable = traceInfo.Variable;

                        if (EntityOps.IsArray2(variable))
                            variable.ArrayValue = new ObjectDictionary(
                                Environment.GetEnvironmentVariables());

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        //
                        // NOTE: We (this trace procedure) expect the "env" variable
                        //       to always be an array.
                        //
                        result = String.Format(
                            "can't {0} {1}: variable is array",
                            FormatOps.Breakpoint(breakpointType),
                            FormatOps.ErrorVariableName(traceInfo.Name, null));
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Object Trace Callback
        #region Support Methods
        internal ReturnCode AddObjectReference(
            ReturnCode code,
            Result result,
            ref Result error
            )
        {
            bool added = false;

            return AddObjectReference(code, result, ref added, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode AddObjectReference(
            ReturnCode code,
            Result result,
            ref bool added,
            ref Result error
            )
        {
            //
            // NOTE: Declare a variable for the formal interface for the
            //       objects that we track.
            //
            IObject @object = null; /* RE-USED */

            //
            // NOTE: Check to see if the new variable value refers to a
            //       live object handle.
            //
            if ((result != null) &&
                GetObject(RuntimeOps.GetStringFromObject(result),
                    true, true, ref @object) == ReturnCode.Ok)
            {
                //
                // NOTE: Attempt to cast the object to the proper wrapper
                //       type and then invalidate the object.
                //
                _Wrappers._Object newWrapper = @object as _Wrappers._Object;
                @object = null;

                //
                // NOTE: If the new wrapper object is valid, add a reference
                //       to it.
                //
                if (newWrapper != null)
                {
                    //
                    // NOTE: Do not attempt to manage reference counts for locked
                    //       objects.
                    //
                    if (!FlagOps.HasFlags(newWrapper.Flags, ObjectFlags.Locked, true))
                    {
                        newWrapper.ReferenceCount++;

                        //
                        // NOTE: We successfully added a reference to the object.
                        //
                        added = true;
                    }

                    newWrapper = null;
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        internal static ReturnCode ObjectTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    if (!FlagOps.HasFlags(traceInfo.Flags, VariableFlags.NoObject, true))
                    {
                        //
                        // NOTE: Intercept attempted set and unset operations on
                        //       variable values that could represent opaque object
                        //       handles.
                        //
                        if (RuntimeOps.IsWriteValueTrace(breakpointType))
                        {
                            //
                            // NOTE: Declare a variable for the formal interface for
                            //       the objects that we track.
                            //
                            IObject @object = null; /* RE-USED */

                            //
                            // NOTE: This is the list of old variable values to
                            //       check for opaque object handles (i.e. to
                            //       decrease the reference counts on).
                            //
                            StringList oldValues = null;

                            //
                            // BUGFIX: *SPECIAL CASE* If the variable is [already]
                            //         undefined, we must skip adding any old values
                            //         from it; otherwise, once a variable with an
                            //         opaque object handle has been unset, the
                            //         variable can never be used for that same
                            //         object handle again.  This bug can only be
                            //         seen by using exactly the same opaque object
                            //         handle name for both the old and new values.
                            //
                            if ((traceInfo.Variable == null) ||
                                !EntityOps.IsUndefined(traceInfo.Variable))
                            {
                                ObjectDictionary arrayValue = null;

                                //
                                // NOTE: Handle resetting or unsetting an entire array,
                                //       if necessary.
                                //
                                if (RuntimeOps.NeedValueForTrace(breakpointType, true) &&
                                    (traceInfo.Index == null) &&
                                    EntityOps.IsArray(traceInfo.Variable, ref arrayValue))
                                {
                                    //
                                    // HACK: Delve into the array dictionary directly
                                    //       and extract all the keys and values that
                                    //       are about to be reset or unset.
                                    //
                                    RuntimeOps.GatherTraceValues(null, null, arrayValue,
                                        ref oldValues);
                                }
                                else if (traceInfo.OldValue != null)
                                {
                                    //
                                    // NOTE: Add the one and only old value to the
                                    //       list of old values.
                                    //
                                    RuntimeOps.GatherTraceValues(null, traceInfo.OldValue,
                                        null, ref oldValues);
                                }
                            }

                            //
                            // NOTE: Add any extra old values provided with the trace.
                            //       Currently, this is only used by [array set].
                            //
                            RuntimeOps.GatherTraceValues(
                                RuntimeOps.NeedValueForTrace(breakpointType, true) ?
                                    traceInfo.Index : null, null, traceInfo.OldValues,
                                ref oldValues);

                            //
                            // NOTE: This is the master list of new variable values to
                            //       check for opaque object handles (i.e. to increase
                            //       the reference counts on).
                            //
                            StringList newValues = null;

                            //
                            // NOTE: Add the new value to the list of new values.
                            //       Also, add the extra new values (currently only
                            //       used by [array set]) to the master list of new
                            //       values.
                            //
                            RuntimeOps.GatherTraceValues(
                                RuntimeOps.NeedValueForTrace(breakpointType, false) ?
                                    traceInfo.Index : null, traceInfo.NewValue,
                                traceInfo.NewValues, ref newValues);

                            //
                            // NOTE: For each new value (i.e. there are potentially
                            //       multiple when handling an array).
                            //
                            if (newValues != null)
                            {
                                foreach (string newValue in newValues)
                                {
                                    //
                                    // NOTE: Check to see if the old variable value
                                    //       referred to a live object handle.
                                    //
                                    if ((newValue != null) &&
                                        interpreter.GetObject(newValue, true, true,
                                            ref @object) == ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: Attempt to cast the object to the proper
                                        //       wrapper type and then invalidate the
                                        //       object.
                                        //
                                        _Wrappers._Object newWrapper =
                                            @object as _Wrappers._Object;

                                        @object = null;

                                        //
                                        // NOTE: If the new wrapper object is valid, add
                                        //       a reference to it.
                                        //
                                        if (newWrapper != null)
                                        {
                                            //
                                            // NOTE: Do not attempt to manage reference
                                            //       counts for locked objects.
                                            //
                                            if (!FlagOps.HasFlags(newWrapper.Flags,
                                                    ObjectFlags.Locked, true))
                                            {
                                                newWrapper.ReferenceCount++;
                                            }

                                            newWrapper = null;
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: If there are no old values then we are done.
                            //
                            if (oldValues == null)
                                return ReturnCode.Ok;

                            //
                            // NOTE: This return code is only used when removing
                            //       objects (i.e. when they have no more outstanding
                            //       references). If any of the objects to be removed
                            //       return an error, this return code will reflect
                            //       that.
                            //
                            ReturnCode code = ReturnCode.Ok;

                            //
                            // NOTE: For each old value (i.e. there are potentially
                            //       multiple when handling an array).
                            //
                            foreach (string oldValue in oldValues)
                            {
                                //
                                // NOTE: Check to see if the old variable value
                                //       referred to a live object handle.
                                //
                                if ((oldValue != null) &&
                                    interpreter.GetObject(oldValue, true, true,
                                        ref @object) == ReturnCode.Ok)
                                {
                                    //
                                    // NOTE: Attempt to cast the object to the
                                    //       proper wrapper type and then invalidate
                                    //       the object.
                                    //
                                    _Wrappers._Object oldWrapper =
                                        @object as _Wrappers._Object;

                                    @object = null;

                                    //
                                    // NOTE: If the old wrapper object is valid,
                                    //       release a reference from it.
                                    //
                                    if (oldWrapper != null)
                                    {
                                        //
                                        // NOTE: Do not attempt to manage reference
                                        //       counts for locked objects.
                                        //
                                        if (!FlagOps.HasFlags(oldWrapper.Flags,
                                                ObjectFlags.Locked, true))
                                        {
                                            //
                                            // NOTE: If there are no more outstanding
                                            //       references to the underlying object,
                                            //       dipose and remove it now.
                                            //
                                            if (--oldWrapper.ReferenceCount <= 0)
                                            {
                                                ReturnCode removeCode;
                                                Result removeResult = null;

                                                removeCode = interpreter.RemoveObject(
                                                    oldWrapper.Token, null, true,
                                                    ref removeResult);

                                                if (removeCode != ReturnCode.Ok)
                                                {
                                                    //
                                                    // NOTE: Complain loudly if we could
                                                    //       not remove the object because
                                                    //       this indicates an error probably
                                                    //       occurred during the disposal of
                                                    //       the object.
                                                    //
                                                    DebugOps.Complain(
                                                        interpreter, removeCode, removeResult);

                                                    //
                                                    // NOTE: If any of the objects cannot be
                                                    //       removed, the overall result will
                                                    //       be an error (even if some of the
                                                    //       objects are successfully removed).
                                                    //
                                                    result = removeResult;
                                                    code = removeCode;
                                                }
                                            }
                                        }

                                        oldWrapper = null;
                                    }
                                }
                            }

                            return code;
                        }
                        else
                        {
                            //
                            // NOTE: We do not care about this type of variable trace; however,
                            //       we cannot simply return the returnCode contained within the
                            //       traceInfo object because the original result could have been
                            //       an error (e.g. get access to the "env" array) and we do not
                            //       want this trace to be considered a failure.  The traceInfo
                            //       object itself is not being changed; therefore, the original
                            //       returnCode within it is preserved for the caller(s) to use
                            //       later.
                            //
                            return ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Completely skip all processing of this variable trace.
                        //       The caller certifies that none of the values contained
                        //       within the trace require object reference counting.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Precision Trace Callback
        #region Support Methods
        internal int PrivatePrecision
        {
            set { lock (syncRoot) { precision = value; } }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode PrecisionTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            //
            // NOTE: Sync up our copy of the integer value with what the
            //       script-level knows about.
            //
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    //
                    // NOTE: Make sure that an array element index was not specified.
                    //
                    //       *WARNING* Empty array element names are allowed,
                    //       please do not change this to "String.IsNullOrEmpty".
                    //
                    if (traceInfo.Index == null)
                    {
                        if (breakpointType == BreakpointType.BeforeVariableSet)
                        {
                            int precision = 0;

                            traceInfo.ReturnCode = Value.GetInteger2(
                                RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                ValueFlags.AnyInteger, interpreter.CultureInfo,
                                ref precision, ref result);

                            if (traceInfo.ReturnCode == ReturnCode.Ok)
                                interpreter.PrivatePrecision = precision;
                        }
                        else if (breakpointType == BreakpointType.BeforeVariableUnset)
                        {
                            //
                            // NOTE: They want to unset the tcl_precision variable.  I
                            //       guess this should be allowed, it is in Tcl.  Also,
                            //       make sure it is purged from the call frame so that
                            //       it cannot be magically restored with this trace
                            //       callback in place.  For this variable, this behavior
                            //       is not 100% compatible with Tcl; however, it is
                            //       internally consistent with other traced Eagle
                            //       variables.
                            //
                            traceInfo.Flags &= ~VariableFlags.NoRemove;

                            //
                            // BUGFIX: Yes, also make sure any waiters are cleared.
                            //
                            EntityOps.SetDirty(traceInfo.Variable, true);
                        }

                        return traceInfo.ReturnCode;
                    }
                    else
                    {
                        //
                        // NOTE: We expect "tcl_precision" to always be a
                        //       scalar variable.
                        //
                        result = String.Format(
                            "can't {0} {1}: variable isn't array",
                            FormatOps.Breakpoint(breakpointType),
                            FormatOps.ErrorVariableName(traceInfo.Name, traceInfo.Index));
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Tests Trace Callback
        #region Support Methods
        private static bool HasTestsTraceCallback(IVariable variable)
        {
            return (variable != null) &&
                (variable.Traces != null) &&
                variable.Traces.Contains(TestsTraceCallback);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal bool IsTestsVariable(IVariable variable)
        {
            return (variable != null) &&
                IsGlobalCallFrame(variable.Frame) &&
                (String.Compare(variable.Name, Vars.Tests,
                    StringOps.SystemStringComparisonType) == 0) &&
                HasTestsTraceCallback(variable);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool DoesTestsVariableExist(string name)
        {
            object enumValue = EnumOps.TryParseEnum(
                typeof(TestInformationType), name, true);

            if (enumValue is TestInformationType)
            {
                Result result = null;

                if (GetTestInformation(
                        (TestInformationType)enumValue,
                        ref result) == ReturnCode.Ok)
                {
                    return true;
                }
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetTestInformation(
            TestInformationType type,
            ref Result result
            )
        {
            switch (type)
            {
                case TestInformationType.Constraints:
                    {
                        result = TestConstraints;
                        return ReturnCode.Ok;
                    }
#if DEBUGGER
                case TestInformationType.Breakpoints:
                    {
                        result = TestBreakpoints.KeysAndValuesToString(null, false);
                        return ReturnCode.Ok;
                    }
#endif
                case TestInformationType.Counts:
                    {
                        result = TestCounts.KeysAndValuesToString(null, false);
                        return ReturnCode.Ok;
                    }
                case TestInformationType.SkippedNames:
                    {
                        result = TestSkipped;
                        return ReturnCode.Ok;
                    }
                case TestInformationType.FailedNames:
                    {
                        result = TestFailures;
                        return ReturnCode.Ok;
                    }
                case TestInformationType.SkipNames:
                    {
                        result = TestSkip;
                        return ReturnCode.Ok;
                    }
                case TestInformationType.MatchNames:
                    {
                        result = TestMatch;
                        return ReturnCode.Ok;
                    }
                case TestInformationType.Level:
                    {
                        result = TestLevels;
                        return ReturnCode.Ok;
                    }
                case TestInformationType.Total:
                case TestInformationType.Skipped:
                case TestInformationType.Passed:
                case TestInformationType.Failed:
                    {
                        result = TestStatistics[(int)type];
                        return ReturnCode.Ok;
                    }
                default:
                    {
                        result = RuntimeOps.BadValue(
                            null, "test information type", type.ToString(),
                            Enum.GetNames(typeof(TestInformationType)),
                            null, null);

                        break;
                    }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal StringDictionary GetAllTestInformation(
            ref Result error
            )
        {
            StringDictionary tests = new StringDictionary();

            foreach (TestInformationType type in new TestInformationType[] {
                    TestInformationType.Constraints,
#if DEBUGGER
                    TestInformationType.Breakpoints,
#endif
                    TestInformationType.Counts, TestInformationType.SkippedNames,
                    TestInformationType.FailedNames, TestInformationType.SkipNames,
                    TestInformationType.MatchNames, TestInformationType.Level,
                    TestInformationType.Total, TestInformationType.Skipped,
                    TestInformationType.Passed, TestInformationType.Failed })
            {
                Result result = null;

                if (GetTestInformation(type, ref result) != ReturnCode.Ok)
                {
                    error = result;
                    return null;
                }

                tests.Add(type.ToString(), result);
            }

            return tests;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode TestsTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            //
            // NOTE: Intercept reads and writes to the variable "tests" and redirect
            //       them to get and set interpreter state variables.
            //
            if (interpreter != null)
            {
                if (traceInfo != null)
                {
                    //
                    // NOTE: *SPECIAL* Ignore the index when we initially add the variable
                    //       since we do not perform any trace actions during add anyhow.
                    //
                    //       *WARNING* Empty array element names are allowed,
                    //       please do not change this to "!String.IsNullOrEmpty".
                    //
                    if (breakpointType != BreakpointType.BeforeVariableAdd)
                    {
                        if (traceInfo.Index != null)
                        {
                            object enumValue = EnumOps.TryParseEnum(
                                typeof(TestInformationType), traceInfo.Index,
                                true, ref result);

                            if (enumValue is TestInformationType)
                            {
                                switch (breakpointType)
                                {
                                    case BreakpointType.BeforeVariableGet:
                                        {
                                            traceInfo.ReturnCode = interpreter.GetTestInformation(
                                                (TestInformationType)enumValue, ref result);

                                            traceInfo.Cancel = true;
                                            break;
                                        }
                                    case BreakpointType.BeforeVariableSet:
                                        {
                                            if ((TestInformationType)enumValue == TestInformationType.Constraints)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestConstraints = list;
                                            }
#if DEBUGGER
                                            else if ((TestInformationType)enumValue == TestInformationType.Breakpoints)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestBreakpoints = new StringDictionary(list, true, false);
                                            }
#endif
                                            else if ((TestInformationType)enumValue == TestInformationType.Counts)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestCounts = new IntDictionary(
                                                        list, interpreter.CultureInfo);
                                            }
                                            else if ((TestInformationType)enumValue == TestInformationType.SkippedNames)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestSkipped = list;
                                            }
                                            else if ((TestInformationType)enumValue == TestInformationType.FailedNames)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestFailures = list;
                                            }
                                            else if ((TestInformationType)enumValue == TestInformationType.SkipNames)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestSkip = list;
                                            }
                                            else if ((TestInformationType)enumValue == TestInformationType.MatchNames)
                                            {
                                                StringList list = null;

                                                //
                                                // WARNING: Cannot cache list representation here, the list
                                                //          may be modified via the public property in the
                                                //          future.
                                                //
                                                traceInfo.ReturnCode = Parser.SplitList(
                                                    interpreter, RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    0, Length.Invalid, false, false, ref list, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                    interpreter.TestMatch = list;
                                            }
                                            else if ((TestInformationType)enumValue == TestInformationType.Level)
                                            {
                                                int newValue = 0;

                                                traceInfo.ReturnCode = Value.GetInteger2(
                                                    RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    ValueFlags.AnyInteger, interpreter.CultureInfo,
                                                    ref newValue, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                {
                                                    interpreter.TestLevels = newValue;
                                                    result = newValue;
                                                }
                                            }
                                            else
                                            {
                                                int newValue = 0;

                                                traceInfo.ReturnCode = Value.GetInteger2(
                                                    RuntimeOps.GetStringFromObject(traceInfo.NewValue),
                                                    ValueFlags.AnyInteger, interpreter.CultureInfo,
                                                    ref newValue, ref result);

                                                if (traceInfo.ReturnCode == ReturnCode.Ok)
                                                {
                                                    interpreter.TestStatistics[(int)enumValue] = newValue;
                                                    result = newValue;
                                                }
                                            }

                                            if (traceInfo.ReturnCode == ReturnCode.Ok)
                                            {
                                                EntityOps.SetUndefined(traceInfo.Variable, false);
                                                EntityOps.SetDirty(traceInfo.Variable, true);
                                            }

                                            traceInfo.Cancel = true;
                                            break;
                                        }
                                    case BreakpointType.BeforeVariableUnset:
                                        {
                                            result = String.Format(
                                                "cannot unset \"{0}\", it is a linked variable",
                                                traceInfo.Name);

                                            traceInfo.Cancel = true;
                                            traceInfo.ReturnCode = ReturnCode.Error;
                                            break;
                                        }
                                }

                                return traceInfo.ReturnCode;
                            }
                        }
                        else if (breakpointType == BreakpointType.BeforeVariableUnset)
                        {
                            //
                            // NOTE: They want to unset the entire tests array.  I guess
                            //       this should be allowed.  Also, make sure it is purged
                            //       from the call frame so that it cannot be magically
                            //       restored with this trace callback in place.
                            //
                            traceInfo.Flags &= ~VariableFlags.NoRemove;

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            //
                            // NOTE: We (this trace procedure) expect the "tests" variable to always
                            //       be an array.
                            //
                            result = String.Format(
                                "can't {0} {1}: variable is array",
                                FormatOps.Breakpoint(breakpointType),
                                FormatOps.ErrorVariableName(traceInfo.Name, null));
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Skip add operations altogether.
                        //
                        return ReturnCode.Ok;
                    }
                }
                else
                {
                    result = "invalid trace";
                }
            }
            else
            {
                result = "invalid interpreter";
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Linked Variable Trace Callback
        #region Support Methods
        private static bool HasLinkedVariableTraceCallback(
            IVariable variable
            )
        {
            return (variable != null) &&
                (variable.Traces != null) &&
                variable.Traces.Contains(LinkedVariableTraceCallback);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode ChangeType(
            string text,
            Type type,
            ref object value,
            ref Result error
            )
        {
            if (type == typeof(string))
            {
                value = text;
                return ReturnCode.Ok;
            }

            if (binder == null)
            {
                error = "invalid binder";
                return ReturnCode.Error;
            }

            lock (syncRoot)
            {
                try
                {
                    value = binder.ChangeType(text, type, cultureInfo);

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [MethodFlags(MethodFlags.VariableTrace | MethodFlags.NoAdd)]
        private static ReturnCode LinkedVariableTraceCallback(
            BreakpointType breakpointType,
            Interpreter interpreter,
            ITraceInfo traceInfo,
            ref Result result
            )
        {
            if ((breakpointType != BreakpointType.BeforeVariableGet) &&
                (breakpointType != BreakpointType.BeforeVariableSet) &&
                (breakpointType != BreakpointType.BeforeVariableUnset))
            {
                return ReturnCode.Ok;
            }

            if (interpreter == null)
            {
                result = "invalid interpreter";
                return ReturnCode.Error;
            }

            if (traceInfo == null)
            {
                result = "invalid trace";
                return ReturnCode.Error;
            }

            IVariable variable = traceInfo.Variable;

            if (variable == null)
            {
                result = "invalid variable";
                return ReturnCode.Error;
            }

            //
            // NOTE: If the variable link is still being initially setup,
            //       skip all other handling.
            //
            if ((breakpointType == BreakpointType.BeforeVariableSet) &&
                EntityOps.IsUndefined(variable))
            {
                return ReturnCode.Ok;
            }

            //
            // NOTE: *WARNING* Empty array element names are allowed,
            //       please do not change this to "!String.IsNullOrEmpty".
            //
            string name = traceInfo.Name;
            string index = traceInfo.Index;
            ObjectDictionary arrayValue = null;

            //
            // NOTE: Is the variable an array?  If so, also get the
            //       actual backing storage for it.
            //
            bool isArray = EntityOps.IsArray(variable, ref arrayValue);

            //
            // NOTE: If applicable, does the array element exist?
            //
            bool exists = isArray && (index != null) ?
                arrayValue.ContainsKey(index) : false;

            if (breakpointType == BreakpointType.BeforeVariableUnset)
            {
                if (index != null)
                {
                    if (isArray)
                    {
                        //
                        // NOTE: Unset one array element.
                        //
                        if (exists)
                        {
                            //
                            // NOTE: Must grab the "real" old value in case we
                            //       need to adjust the reference count for it.
                            //
                            object oldValue = null;

                            RuntimeOps.GetLinkedVariableMemberValue(
                                interpreter, traceInfo.OldValue, ref oldValue);

                            if (oldValue != null)
                                traceInfo.OldValue = oldValue;

                            traceInfo.ReturnCode = ReturnCode.Ok;
                        }
                        else if (!FlagOps.HasFlags(
                            traceInfo.Flags, VariableFlags.NoComplain, true))
                        {
                            result = String.Format(
                                "can't unset {0}: no such element in array",
                                FormatOps.ErrorVariableName(
                                    variable, null, name, index));

                            traceInfo.ReturnCode = ReturnCode.Error;
                        }
                        else
                        {
                            //
                            // NOTE: The default handling sets an error message
                            //       into the result for this case and we do not
                            //       want that; therefore, cancel the default
                            //       action.
                            //
                            traceInfo.Cancel = true;
                            traceInfo.ReturnCode = ReturnCode.Ok;
                        }
                    }
                    else
                    {
                        //
                        // NOTE: Error, unset index with non-array?
                        //
                        result = String.Format(
                            "can't unset {0}: variable isn't array",
                            FormatOps.ErrorVariableName(
                                variable, null, name, index));

                        traceInfo.ReturnCode = ReturnCode.Error;
                    }
                }
                else
                {
                    //
                    // NOTE: They want to unset the entire array.  Also, make
                    //       sure it is purged from the call frame so that it
                    //       cannot be magically restored with this trace
                    //       callback in place.
                    //
                    traceInfo.Flags &= ~VariableFlags.NoRemove;

                    //
                    // NOTE: Must grab the "real" old value in case we need
                    //       to adjust the reference count for it.
                    //
                    object oldValue = traceInfo.OldValue;

                    if (oldValue != null)
                    {
                        RuntimeOps.GetLinkedVariableMemberValue(
                            interpreter, oldValue, ref oldValue);

                        if (oldValue != null)
                            traceInfo.OldValue = oldValue;
                    }

                    //
                    // BUGFIX: Must make the entire array "concrete" now just
                    //         in case there are any opaque object handles
                    //         that need to be picked up by later traces (e.g.
                    //         ObjectTraceCallback).
                    //
                    if (arrayValue != null)
                    {
                        ObjectDictionary oldValues = null;

                        RuntimeOps.GetLinkedVariableArrayValues(
                            interpreter, arrayValue, ref oldValues);

                        RuntimeOps.AddOldValuesToTraceInfo(
                            traceInfo, oldValues);
                    }

                    //
                    // NOTE: If we get this far, everything went fine.
                    //
                    traceInfo.ReturnCode = ReturnCode.Ok;
                }
            }
            else if (isArray == (index != null))
            {
                ReturnCode code;
                MemberInfo memberInfo = null;
                Type type = null;
                object @object = null;
                object memberValue = null;

                code = RuntimeOps.GetLinkedVariableMemberAndValue(
                    breakpointType, variable, name, index,
                    traceInfo.OldValue, ref memberInfo, ref type,
                    ref @object, ref memberValue, ref result);

                if (code == ReturnCode.Ok)
                {
                    switch (breakpointType)
                    {
                        case BreakpointType.BeforeVariableGet:
                            {
                                if (!isArray || exists)
                                {
                                    code = RuntimeOps.GetLinkedVariableString(
                                        interpreter, memberValue, ref result);

                                    if (code == ReturnCode.Ok)
                                    {
                                        traceInfo.OldValue =
                                            RuntimeOps.GetStringFromObject(result);

                                        traceInfo.Cancel = true;
                                    }

                                    traceInfo.ReturnCode = code;
                                }
                                else
                                {
                                    result = String.Format(
                                        "can't read {0}: no such element in array",
                                        FormatOps.ErrorVariableName(
                                            variable, null, name, index));

                                    traceInfo.ReturnCode = ReturnCode.Error;
                                }
                                break;
                            }
                        case BreakpointType.BeforeVariableSet:
                            {
                                object newValue = traceInfo.NewValue;

                                if (newValue is string)
                                {
                                    //
                                    // HACK: We do not officially support
                                    //       any array types here (yet?);
                                    //       however, for one dimensional
                                    //       string arrays, just use a
                                    //       StringList.
                                    //
                                    if (type == typeof(string[]))
                                    {
                                        StringList list = null;

                                        code = Parser.SplitList(
                                            interpreter, (string)newValue,
                                            0, Length.Invalid, false, true,
                                            ref list, ref result);

                                        if (code == ReturnCode.Ok)
                                            newValue = list.ToArray();
                                    }
                                    else
                                    {
                                        code = interpreter.ChangeType(
                                            (string)newValue, type,
                                            ref newValue, ref result);
                                    }
                                }

                                if (code == ReturnCode.Ok)
                                {
                                    code = RuntimeOps.SetLinkedVariableMemberValue(
                                        breakpointType, variable, name,
                                        index, memberInfo, @object,
                                        newValue, ref result);
                                }

                                if (code == ReturnCode.Ok)
                                {
                                    //
                                    // HACK: This result is not 100% correct;
                                    //       however, we do not really care.
                                    //       Firstly, setting the overall
                                    //       result is entirely optional for
                                    //       trace callbacks that "succeed".
                                    //       Secondly, using FixupReturnValue
                                    //       (or similar) method here would
                                    //       potentially create opaque object
                                    //       handles which would then just be
                                    //       leaked because they would never
                                    //       actually even be seen by the
                                    //       calling script.  At some point,
                                    //       it might be worthwhile to use
                                    //       the FixupReturnValue method here
                                    //       with the value of the toString
                                    //       parameter hard-wired to true.
                                    //
                                    result = RuntimeOps.GetStringFromObject(
                                        newValue);

                                    EntityOps.SetUndefined(variable, false);
                                    EntityOps.SetDirty(variable, true);

                                    traceInfo.OldValue = memberValue;
                                    traceInfo.NewValue = newValue;
                                    traceInfo.Cancel = true;
                                    traceInfo.ReturnCode = ReturnCode.Ok;
                                }
                                else
                                {
                                    traceInfo.ReturnCode = code;
                                }
                                break;
                            }
                    }
                }
                else
                {
                    traceInfo.ReturnCode = code;
                }
            }
            else
            {
                result = String.Format(
                    "can't {0} {1}: variable {2} array",
                    FormatOps.Breakpoint(breakpointType),
                    FormatOps.ErrorVariableName(
                        variable, null, name, index),
                    isArray ? "is" : "isn't");

                traceInfo.ReturnCode = ReturnCode.Error;
            }

            return traceInfo.ReturnCode;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region "Active" Interpreter Support
        public static Interpreter GetAny() /* THREAD-SAFE */
        {
            Interpreter interpreter = GetActive();

            if (interpreter == null)
            {
                ReturnCode code;
                Result error = null;

                code = GlobalState.GetInterpreter(
                    true, ref interpreter, ref error);

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(code, error);
            }

            return interpreter;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static Interpreter GetActive() /* THREAD-SAFE */
        {
            IAnyPair<Interpreter, IClientData> pair = GetActivePair();
            return (pair != null) ? pair.X : null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private
        internal static IAnyPair<Interpreter, IClientData> GetActivePair()
        {
            return GetActivePair(null);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static IAnyPair<Interpreter, IClientData> GetActivePair(
            Type type
            )
        {
            return GlobalState.GetActiveInterpreter(type);
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Server Socket Support
#if NETWORK
        internal int ServerSockets
        {
            get { return serverSockets; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int EnterSocketThread()
        {
            return Interlocked.Increment(ref serverSockets);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int ExitSocketThread()
        {
            return Interlocked.Decrement(ref serverSockets);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode StartServerSocket(
            OptionDictionary options,
            int timeout,
            string address,
            string port,
            bool exclusive,
            string text, /* command */
            ref Result result)
        {
            ReturnCode code = ReturnCode.Error;
            Thread thread = null;

            try
            {
                EventWaitHandle @event = ThreadOps.CreateEvent(false);

                try
                {
                    SocketClientData clientData = new SocketClientData(
                        null, @event, this, options, address, port, exclusive,
                        text);

                    if (timeout == 0)
                        timeout = DefaultThreadStartTimeout;

                    thread = Engine.CreateThread(
                        this, ServerSocketThreadStart, 0, true, false);

                    if (thread != null)
                    {
                        thread.Name = FormatOps.Id(String.Format(
                            "serverSocketThread: {0}", this), port, 0);

                        thread.Start(clientData);

                        //
                        // HACK: For now, avoid ever using the new overload(s)
                        //       of this method; otherwise, Mono crashes.
                        //
                        if (ThreadOps.WaitEvent(@event, timeout))
                        {
                            result = clientData.Result;
                            code = clientData.ReturnCode;
                        }
                        else
                        {
                            result = "timed out starting server thread";
                        }
                    }
                    else
                    {
                        result = "could not create server thread";
                    }
                }
                finally
                {
                    ThreadOps.CloseEvent(ref @event);
                }
            }
            catch (Exception e)
            {
                result = e;
            }
            finally
            {
                //
                // NOTE: If we have failed for any reason and we
                //       created the thread, abort it now.
                //
                if ((code != ReturnCode.Ok) &&
                    (thread != null))
                {
                    try
                    {
                        thread.Abort(); /* BUGBUG: Leaks? */
                        thread = null;
                    }
                    catch (Exception e)
                    {
                        //
                        // NOTE: Nothing we can do here except log the failure.
                        //
                        RuntimeOps.DebugTrace(String.Format(
                            "{0}: {1}{2}",
                            GlobalState.GetCurrentThreadId(),
                            e, Environment.NewLine),
                            typeof(Interpreter).Name);
                    }
                }
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static void ServerSocketThreadStart(object obj)
        {
            try
            {
                SocketClientData clientData = obj as SocketClientData;

                if (clientData != null)
                {
                    Interpreter interpreter = clientData.Interpreter;

                    if (interpreter != null)
                    {
                        interpreter.EnterSocketThread();

                        try
                        {
                            bool setEvent = false;
                            CultureInfo cultureInfo = interpreter.CultureInfo;

                            Result result = null;

                            TcpListener listener = SocketOps.NewTcpListener(
                                clientData.Address, clientData.Port, cultureInfo,
                                ref result);

                            if (listener != null)
                            {
                                try
                                {
                                    try
                                    {
                                        //
                                        // NOTE: Mono does not support this feature on Unix.
                                        //
                                        if (PlatformOps.IsWindowsOperatingSystem())
                                            listener.ExclusiveAddressUse = clientData.Exclusive;
                                    }
                                    catch (Exception e)
                                    {
                                        //
                                        // NOTE: Mono 2.0/2.2 does not support this feature.
                                        //
                                        RuntimeOps.DebugTrace(String.Format(
                                            "{0}: {1}{2}",
                                            GlobalState.GetCurrentThreadId(),
                                            e, Environment.NewLine),
                                            typeof(Interpreter).Name);
                                    }

                                    string channelId = FormatOps.Id(
                                        "listenSocket", null, interpreter.NextId());

                                    //
                                    // NOTE: Add the "listener" channel to the interpreter.
                                    //
                                    clientData.ReturnCode = interpreter.AddTcpListenerChannel(
                                        channelId, ChannelType.None, listener, clientData,
                                        ref result);

                                    if (clientData.ReturnCode == ReturnCode.Ok)
                                        clientData.Result = channelId;
                                    else
                                        clientData.Result = result;

                                    setEvent = ThreadOps.SetEvent(clientData.Event);

                                    if (clientData.ReturnCode == ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: Success so far, start listening...
                                        //
                                        listener.Start();

                                        try
                                        {
                                            //
                                            // NOTE: Poll the listener for incoming connections.
                                            //       For an incoming connection, accept a TcpClient
                                            //       and evaluate the supplied command.
                                            //
                                            while (SocketOps.IsListenerActive(listener))
                                            {
                                                int milliseconds = interpreter.GetSleepTime();

                                                if (milliseconds < MinimumSocketPollTimeout)
                                                    milliseconds = MinimumSocketPollTimeout;

                                                while (true)
                                                {
                                                    //
                                                    // NOTE: Block until the timeout or a connection request
                                                    //       is ready.
                                                    //
                                                    Socket socket = listener.Server;

                                                    if (socket == null)
                                                        break;

                                                    if (!socket.Poll(milliseconds, SelectMode.SelectRead))
                                                        break;

                                                    //
                                                    // NOTE: If the listener is no longer active or the
                                                    //       underlying socket has been cleaned up (i.e. the
                                                    //       other thread called [close] on it), then bail
                                                    //       out now.
                                                    //
                                                    if (!SocketOps.IsListenerActive(listener) ||
                                                        SocketOps.IsCleanedUp(socket))
                                                    {
                                                        break;
                                                    }

                                                    //
                                                    // NOTE: Attempt to accept the connection.
                                                    //
                                                    TcpClient client = listener.AcceptTcpClient();

                                                    //
                                                    // NOTE: Create an Id for the new client channel.
                                                    //
                                                    channelId = FormatOps.Id("serverSocket", null,
                                                        interpreter.NextId());

                                                    //
                                                    // NOTE: Add the new channel for this client to the
                                                    //       interpreter.
                                                    //
                                                    clientData.ReturnCode = interpreter.AddFileOrSocketChannel(
                                                        channelId, client.GetStream(), clientData.Options,
                                                        StreamFlags.ServerSocket, false, false,
                                                        new ClientData(client), ref result);

                                                    clientData.Result = result;

                                                    if (clientData.ReturnCode == ReturnCode.Ok)
                                                    {
                                                        //
                                                        // NOTE: Evaluate the provided script when a
                                                        //       connection is accepted.
                                                        //
                                                        string text = null;

                                                        clientData.ReturnCode = SocketOps.GetServerScript(
                                                            client, channelId, clientData.Text, ref text,
                                                            ref result);

                                                        clientData.Result = result;

                                                        if (clientData.ReturnCode == ReturnCode.Ok)
                                                        {
                                                            clientData.ReturnCode = interpreter.QueueScript(
                                                                TimeOps.GetUtcNow(), text, ref result);

                                                            clientData.Result = result;
                                                        }
                                                    }

                                                    if (clientData.ReturnCode != ReturnCode.Ok)
                                                        _Public.EventManager.HandleBackgroundError(
                                                            interpreter, clientData.ReturnCode,
                                                            clientData.Result);
                                                }
                                            }
                                        }
                                        finally
                                        {
                                            //
                                            // NOTE: Stop listening for incoming connections, we are done.
                                            //       This call is probably pointless because the only real
                                            //       way we can exit the loop is by externally stopping the
                                            //       channel; however, this should be fairly harmless.
                                            //
                                            listener.Stop();
                                        }
                                    }
                                }
                                catch (Exception e)
                                {
                                    clientData.Result = e;
                                    clientData.ReturnCode = ReturnCode.Error;

                                    if (!setEvent)
                                        ThreadOps.SetEvent(clientData.Event);
                                }
                            }
                            else
                            {
                                clientData.Result = result;
                                clientData.ReturnCode = ReturnCode.Error;

                                if (!setEvent)
                                    ThreadOps.SetEvent(clientData.Event);
                            }
                        }
                        finally
                        {
                            interpreter.ExitSocketThread();
                        }
                    }
                    else
                    {
                        clientData.Result = "invalid interpreter";
                        clientData.ReturnCode = ReturnCode.Error;
                        ThreadOps.SetEvent(clientData.Event);
                    }
                }
            }
            catch (ThreadInterruptedException)
            {
                // do nothing.
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Host Integration Support
        internal string Complaint
        {
            get
            {
                // CheckDisposed();

                lock (syncRoot)
                {
                    return complaint;
                }
            }
            set
            {
                // CheckDisposed();

                lock (syncRoot)
                {
                    complaint = value;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode CanExit(
            ExitCode exitCode,
            bool force,
            bool fail,
            string message,
            ref Result error
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (host != null)
            {
                try
                {
                    if (FlagOps.HasFlags(host.GetFlags(), HostFlags.Exit, true))
                    {
                        if (host.Exiting)
                        {
                            error = "interpreter host is already exiting";
                            code = ReturnCode.Error;
                        }
                        else if (!host.CanExit)
                        {
                            error = "exit denied by interpreter host";
                            code = ReturnCode.Error;
                        }
                        else if (force && !host.CanForceExit)
                        {
                            error = "force exit denied by interpreter host";
                            code = ReturnCode.Error;
                        }
                    }
                }
                catch (Exception e)
                {
                    //
                    // TODO: Should this simply allow the exit to take place?
                    //
                    error = e;
                    code = ReturnCode.Error;
                }
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: CanExit: returning, exitCode = {1}, force = {2}, " +
                "fail = {3}, message = {4}, returning code {5}{6}{7}",
                GlobalState.GetCurrentThreadId(), exitCode, force, fail,
                FormatOps.WrapOrNull(true, true, message), code,
                (code == ReturnCode.Ok) ? String.Empty :
                    String.Format(" with error {0}",
                        FormatOps.WrapOrNull(true, true, error)),
                Environment.NewLine), typeof(Interpreter).Name);

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetStream( /* FOR USE BY Engine.OpenScriptStream METHOD */
            string path,
            FileMode mode,
            FileAccess access,
            bool resolve,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            )
        {
            return GetStream(
                path, mode, access, FileShare.Read,
                Channel.DefaultBufferSize, FileOptions.None,
                resolve, true, ref fullPath, ref stream,
                ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode GetStream( /* FOR USE BY [open] COMMAND */
            string path,
            FileMode mode,
            FileAccess access,
            FileShare share,
            int bufferSize,
            FileOptions options,
            bool strict,
            ref Stream stream,
            ref Result error
            )
        {
            string fullPath = null;

            return GetStream(
                path, mode, access, share, bufferSize,
                options, false, strict, ref fullPath,
                ref stream, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode GetStream(
            string path,
            FileMode mode,
            FileAccess access,
            FileShare share,
            int bufferSize,
            FileOptions options,
            bool resolve,
            bool strict,
            ref string fullPath,
            ref Stream stream,
            ref Result error
            )
        {
            return RuntimeOps.GetHostStream(
                this, host, path, mode, access, share,
                bufferSize, options, resolve, strict,
                ref fullPath, ref stream, ref error);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal void GetHostColors(
            IHost host,
            string name,
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor
            )
        {
            ReturnCode code;
            Result error = null;

            code = RuntimeOps.GetHostColors(host, name, true, true, true,
                ref foregroundColor, ref backgroundColor, ref error);

            if (code != ReturnCode.Ok)
            {
#if DEBUG && VERBOSE
                DebugOps.Complain(this, code, error);
#endif

                foregroundColor = defaultForegroundColor;
                backgroundColor = defaultBackgroundColor;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Disposal Helper Methods (Terminate/Dispose)
        #region Tcl Disposal
#if NATIVE && TCL
        private bool DoesTclNeedDispose(
            ref ResultList errors
            )
        {
            lock (syncRoot)
            {
                if (tclApi != null)
                {
                    if (errors == null)
                        errors = new ResultList();

                    errors.Add("may leak Tcl API object");
                }

                if ((tclInterps != null) && (tclInterps.Count > 0))
                {
                    if (errors == null)
                        errors = new ResultList();

                    errors.Add(String.Format(
                        "may leak {0} Tcl interpreters: {1}",
                        tclInterps.Count, tclInterps));
                }

#if TCL_THREADS
                if ((tclThreads != null) && (tclThreads.Count > 0))
                {
                    if (errors == null)
                        errors = new ResultList();

                    errors.Add(String.Format(
                        "may leak {0} Tcl threads: {1}",
                        tclThreads.Count, tclThreads));
                }
#endif

                if ((tclBridges != null) && (tclBridges.Count > 0))
                {
                    if (errors == null)
                        errors = new ResultList();

                    errors.Add(String.Format(
                        "may leak {0} Tcl bridges: {1}",
                        tclBridges.Count, tclBridges));
                }

                if ((errors != null) && (errors.Count > 0))
                    errors.Insert(0, String.Format(
                        "interpreter \"{0}\" from thread {1} is " +
                        "being disposed on [the wrong] thread {2}",
                        this, threadId, GlobalState.GetCurrentThreadId()));
            }

            return ((errors != null) && (errors.Count > 0));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeTcl(
            bool dispose,
            bool disposing,
            bool disconnect
            ) /* WARNING: For use by Dispose(bool) only. */
        {
            ResultList errors = null;

            if (DoesTclNeedDispose(ref errors))
            {
                ReturnCode code;
                Result error = null;

                //
                // NOTE: Attempt to dispose our Tcl related stuff now.
                //
                code = DisposeTcl(dispose, disposing, disconnect, ref error);

                //
                // NOTE: If we fail for any reason (including thread affinity)
                //       report the error message itself and the objects that
                //       may be leaked.  Also, see if we can transfer the now
                //       "dead" Tcl related objects that  require cleanup to
                //       the global (static) cache for later transfer to an
                //       interpreter created on the appropriate thread (which,
                //       at least in theory, has a greater than zero chance of
                //       properly calling Dispose on the correct thread?).
                //
                if (code != ReturnCode.Ok)
                {
                    if (errors == null)
                        errors = new ResultList();

                    errors.Insert(0, error);
                    error = errors;

                    DebugOps.Complain(this, code, error);

                    //
                    // NOTE: Now, check if we can add all the Tcl related
                    //       objects that would otherwise leak to the global
                    //       (static) cache for later disposal on the correct
                    //       thread.
                    //
                    lock (syncRoot)
                    {
                        #region Static Lock Held
                        lock (staticSyncRoot)
                        {
                            if ((deadTclThreadId == 0) ||
                                IsPrimaryThread(deadTclThreadId))
                            {
                                if (deadTclThreadId == 0)
                                    deadTclThreadId = threadId;

                                ///////////////////////////////////////////////////////////////////////

                                if (tclApi != null)
                                {
                                    if (deadTclApi == null)
                                        deadTclApi = tclApi;

                                    tclApi = null;
                                }

                                ///////////////////////////////////////////////////////////////////////

                                if (tclInterps != null)
                                {
                                    if (deadTclInterps != null)
                                        deadTclInterps.Add(tclInterps);

                                    tclInterps.Clear();
                                    tclInterps = null;
                                }

                                ///////////////////////////////////////////////////////////////////////

#if TCL_THREADS
                                if (tclThreads != null)
                                {
                                    if (deadTclThreads != null)
                                        deadTclThreads.Add(tclThreads);

                                    tclThreads.Clear();
                                    tclThreads = null;
                                }
#endif

                                ///////////////////////////////////////////////////////////////////////

                                if (tclBridges != null)
                                {
                                    if (deadTclBridges != null)
                                        deadTclBridges.Add(tclBridges);

                                    tclBridges.Clear();
                                    tclBridges = null;
                                }
                            }
                        }
                        #endregion
                    }
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DisposeTcl(
            bool dispose,
            bool disposing, /* NOT USED */
            bool disconnect,
            ref Result error
            )
        {
            ReturnCode code;

            if (IsPrimaryThread())
            {
                //
                // NOTE: Dispose of all TclBridge objects we own now.
                //
                code = DisposeTclBridges(null, dispose, ref error);

                if (code != ReturnCode.Ok)
                {
                    if (dispose)
                    {
                        //
                        // NOTE: We are being called from dispose, not much we can really
                        //       do here.  Attempt to show the error to the user.
                        //
                        DebugOps.Complain(this, code, error);
                    }
                    else
                    {
                        //
                        // NOTE: We are not being called from Dispose, bail out now with
                        //       the error.
                        //
                        return code;
                    }
                }

                lock (syncRoot)
                {
                    GlobalState.PushActiveInterpreter(this);

                    try
                    {
#if TCL_THREADS
                        //
                        // NOTE: Attempt to shutdown all the Tcl threads we own now.
                        //
                        if (tclThreads != null)
                        {
                            StringList keys = new StringList(tclThreads.Keys);

                            foreach (string key in keys)
                            {
                                TclThread thread = tclThreads[key];

                                if (thread != null)
                                {
                                    //
                                    // NOTE: If we are being called by the exit handler,
                                    //       "force" the shutdown of the thread.
                                    //
                                    code = thread.Shutdown(true, disconnect, false, ref error);

                                    if (code == ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: The Tcl interpreter has been deleted, remove it
                                        //       from our collection.
                                        //
                                        tclThreads.Remove(key);
                                    }
                                    else
                                    {
                                        if (dispose)
                                        {
                                            //
                                            // NOTE: We are being called from dispose, not much we can really
                                            //       do here.  Attempt to show the error to the user.
                                            //
                                            DebugOps.Complain(this, code, error);
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: We are not being called from Dispose, bail out now with
                                            //       the error.
                                            //
                                            return code;
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: Upon dispose, destroy the collection itself so that it cannot
                            //       be used again.
                            //
                            // BUGBUG: This needs to be reworked at some point when Dispose is
                            //         modified to be capable of "failing".
                            //
                            if (dispose)
                            {
                                tclThreads.Clear();
                                tclThreads = null;
                            }
                        }
#endif

                        //
                        // NOTE: Attempt to delete all the Tcl interps we own now.
                        //
                        if (tclInterps != null)
                        {
                            StringList keys = new StringList(tclInterps.Keys);

                            foreach (string key in keys)
                            {
                                IntPtr interp = tclInterps[key];

                                if (interp != IntPtr.Zero)
                                {
                                    //
                                    // BUGFIX: If we are being called by the exit handler,
                                    //         "force" the deletion of the interpreter.  Tcl
                                    //         should not actually free the underlying memory
                                    //         until the stack is unwound.
                                    //
#if NATIVE_PACKAGE
                                    if (NativePackage.ShouldDeleteTclInterpreter(key, interp))
#endif
                                        code = TclWrapper.DeleteInterpreter(tclApi, disconnect,
                                            ref interp, ref error);
#if NATIVE_PACKAGE
                                    else
                                        code = ReturnCode.Ok;
#endif

                                    if (code == ReturnCode.Ok)
                                    {
                                        //
                                        // NOTE: The Tcl interpreter has been deleted, remove it
                                        //       from our collection.
                                        //
                                        tclInterps.Remove(key);
                                    }
                                    else
                                    {
                                        if (dispose)
                                        {
                                            //
                                            // NOTE: We are being called from dispose, not much we can really
                                            //       do here.  Attempt to show the error to the user.
                                            //
                                            DebugOps.Complain(this, code, error);
                                        }
                                        else
                                        {
                                            //
                                            // NOTE: We are not being called from Dispose, bail out now with
                                            //       the error.
                                            //
                                            return code;
                                        }
                                    }
                                }
                            }

                            //
                            // NOTE: Upon dispose, destroy the collection itself so that it cannot
                            //       be used again.
                            //
                            // BUGBUG: This needs to be reworked at some point when Dispose is
                            //         modified to be capable of "failing".
                            //
                            if (dispose)
                            {
                                tclInterps.Clear();
                                tclInterps = null;
                            }
                        }

                        //
                        // NOTE: Only mess with the Tcl API object if we are being called
                        //       by our Dispose or the Dispose of the Tcl API object itself;
                        //       otherwise, this cleanup will be handled externally.
                        //
                        if (disconnect || dispose)
                        {
                            //
                            // NOTE: Terminate the link with the Tcl API now.
                            //
                            // BUGBUG: This needs to be reworked at some point when Dispose is
                            //         modified to be capable of "failing".
                            //
                            if (tclApi != null)
                            {
                                try
                                {
                                    if (dispose
#if NATIVE_PACKAGE
                                        && !NativePackage.IsTclInterpreterActive()
#endif
                                        )
                                    {
                                        //
                                        // BUGFIX: If we are being explicitly disposed, we need
                                        //         to make sure the actual Tcl library gets
                                        //         unloaded (or at least has its internal
                                        //         reference count decreased in the event it is
                                        //         being used by other Eagle interpreters or
                                        //         random external code).
                                        //
                                        code = TclWrapper.Unload(
                                            this, UnloadFlags.Default, ref tclApi, ref error);
                                    }
                                    else
                                    {
                                        //
                                        // NOTE: Just clear out our Tcl API object reference
                                        //       because we are [already] being called by its
                                        //       Dispose method.
                                        //
                                        tclApi = null;
                                    }
                                }
                                catch (Exception e)
                                {
                                    error = e;
                                    code = ReturnCode.Error;
                                }
                            }
                        }
                    }
                    finally
                    {
                        /* IGNORED */
                        GlobalState.PopActiveInterpreter();
                    }
                }
            }
            else
            {
                error = "cannot dispose Tcl, wrong thread";
                code = ReturnCode.Error;
            }

            return code;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal ReturnCode DisposeTclBridges(
            IntPtr interp,
            bool dispose,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                GlobalState.PushActiveInterpreter(this);

                try
                {
                    if (tclBridges != null)
                    {
                        StringList keys = new StringList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the TclBridges collection in the containing interpreter.
                        //
                        foreach (KeyValuePair<string, TclBridge> pair in tclBridges)
                        {
                            if ((interp == IntPtr.Zero) ||
                                ((pair.Value != null) && (pair.Value.interp == interp)))
                            {
                                keys.Add(pair.Key);
                            }
                        }

                        //
                        // NOTE: Cleanup, PASS #2, dispose and remove all the TclBridge objects
                        //       we found from PASS #1 (above).
                        //
                        foreach (string key in keys)
                        {
                            TclBridge tclBridge = tclBridges[key];

                            if (tclBridge != null)
                            {
                                try
                                {
                                    tclBridge.Dispose(); /* throw */
                                    tclBridge = null;
                                }
                                catch (Exception e)
                                {
                                    error = String.Format(
                                        "caught exception while disposing Tcl bridge: {0}",
                                        e);

                                    return ReturnCode.Error;
                                }
                            }

                            tclBridges.Remove(key);
                        }

                        //
                        // NOTE: Upon dispose, destroy the collection itself so that it cannot
                        //       be used again.
                        //
                        if (dispose && (tclBridges.Count == 0))
                            tclBridges = null;
                    }

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;

                    return ReturnCode.Error;
                }
                finally
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ReturnCode DisposeTclBridges(
            IExecute execute,
            bool dispose,
            ref Result error
            )
        {
            lock (syncRoot)
            {
                GlobalState.PushActiveInterpreter(this);

                try
                {
                    if (tclBridges != null)
                    {
                        StringList keys = new StringList();

                        //
                        // NOTE: Cleanup, PASS #1, collect the list of keys we need to remove
                        //       from the TclBridges collection in the containing interpreter.
                        //
                        foreach (KeyValuePair<string, TclBridge> pair in tclBridges)
                        {
                            if ((execute == null) || ((pair.Value != null) &&
                                Object.ReferenceEquals(pair.Value.execute, execute)))
                            {
                                keys.Add(pair.Key);
                            }
                        }

                        //
                        // NOTE: Cleanup, PASS #2, dispose and remove all the TclBridge objects
                        //       we found from PASS #1 (above).
                        //
                        foreach (string key in keys)
                        {
                            TclBridge tclBridge = tclBridges[key];

                            if (tclBridge != null)
                            {
                                try
                                {
                                    tclBridge.Dispose(); /* throw */
                                    tclBridge = null;
                                }
                                catch (Exception e)
                                {
                                    error = String.Format(
                                        "caught exception while disposing Tcl bridge: {0}",
                                        e);

                                    return ReturnCode.Error;
                                }
                            }

                            tclBridges.Remove(key);
                        }

                        //
                        // NOTE: Upon dispose, destroy the collection itself so that it cannot
                        //       be used again.
                        //
                        if (dispose && (tclBridges.Count == 0))
                            tclBridges = null;
                    }

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;

                    return ReturnCode.Error;
                }
                finally
                {
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();
                }
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Callstack Disposal
#if !THREADING
        private void DisposeCallStack(
            ref CallStack callStack
            )
        {
            if (callStack != null)
            {
                ReturnCode code;
                Result error = null;

                try
                {
                    code = RuntimeOps.TryDisposeObject(callStack, ref error);
                }
                catch (Exception e)
                {
                    error = String.Format(
                        "caught exception while disposing call stack: {0}",
                        e);

                    code = ReturnCode.Error;
                }

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error);

                callStack = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeCallFrame(
            ref ICallFrame frame
            )
        {
            if (frame != null)
            {
                ReturnCode code;
                Result error = null;

                try
                {
                    code = RuntimeOps.TryDisposeObject(frame, ref error);
                }
                catch (Exception e)
                {
                    error = String.Format(
                        "caught exception while disposing call frame: {0}",
                        e);

                    code = ReturnCode.Error;
                }

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error);

                frame = null;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Thread Data Disposal
        internal void DisposeThread(
            bool dispose,
            bool disposing /* NOT USED */
            )
        {
#if NATIVE && WINDOWS
            //
            // NOTE: Try to verify that we are the last interpreter for this
            //       thread prior to getting rid of our native stack related
            //       informmation.
            //
            if (CountThreadInterpreters() == 1) // BUGBUG: Maybe <= 1 here?
                RuntimeOps.FinalizeStackChecking();
#endif

            ///////////////////////////////////////////////////////////////////

#if THREADING
            //
            // NOTE: Dispose all per-interpreter data (for this thread only).
            //       Only "threading" enabled data is disposed here.  Normal
            //       per-interpreter data is disposed by Dispose (see below).
            //
            if (contextManager != null)
                contextManager.Free(dispose);
#else
            //
            // NOTE: Dispose all per-interpreter data (for all threads).  We do this
            //       here, rather than in Dispose, in order to centralize the logic
            //       that handles data that is per-thread in "threading" enabled
            //       builds.
            //
            if (dispose)
            {
                lock (syncRoot)
                {
                    #region Engine Context
                    levels = 0;
                    previousLevels = 0;
                    parserLevels = 0;
                    expressionLevels = 0;
                    unknownLevels = 0;
                    traceLevels = 0;

                    ///////////////////////////////////////////////////////////

#if DEBUGGER
                    interactiveLoopCallback = null;
#endif

                    ///////////////////////////////////////////////////////////

#if NOTIFY || NOTIFY_OBJECT
                    notifyLevels = 0;
                    notifyTypes = NotifyType.None;
                    notifyFlags = NotifyFlags.None;
#endif

                    ///////////////////////////////////////////////////////////

                    policyLevels = 0;
                    testLevels = 0;
                    stackOverflow = false;

                    ///////////////////////////////////////////////////////////

#if PREVIOUS_RESULT
                    previousResult = null;
#endif

                    ///////////////////////////////////////////////////////////

                    engineFlags = EngineFlags.None;
                    returnCode = ReturnCode.Ok;
                    errorLine = 0;
                    errorCode = null;
                    errorInfo = null;
                    errorFrames = 0;
                    scriptLocation = null;

                    ///////////////////////////////////////////////////////////

                    if (scriptLocations != null)
                    {
                        scriptLocations.Clear();
                        scriptLocations = null;
                    }

                    ///////////////////////////////////////////////////////////

                    previousProcessId = 0;

                    ///////////////////////////////////////////////////////////

                    if (arraySearches != null)
                    {
                        arraySearches.Clear();
                        arraySearches = null;
                    }

                    ///////////////////////////////////////////////////////////

#if HISTORY
                    historyEngineFilter = null;

                    ///////////////////////////////////////////////////////////

                    if (_history != null)
                    {
                        _history.Clear();
                        _history = null;
                    }
#endif
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Interactive Context
                    interactive = false;
                    interactiveInput = null;
                    previousInteractiveInput = null;
                    interactiveLoops = 0;

                    ///////////////////////////////////////////////////////////

#if HISTORY
                    historyLoadData = null;
                    historySaveData = null;
                    historyInfoFilter = null;
                    historyLoadFilter = null;
                    historySaveFilter = null;
                    historyFileName = null;
#endif
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Test Context
                    testStatistics = null;

                    ///////////////////////////////////////////////////////////

                    if (testConstraints != null)
                    {
                        testConstraints.Clear();
                        testConstraints = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testSkipped != null)
                    {
                        testSkipped.Clear();
                        testSkipped = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testFailures != null)
                    {
                        testFailures.Clear();
                        testFailures = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testCounts != null)
                    {
                        testCounts.Clear();
                        testCounts = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testMatch != null)
                    {
                        testMatch.Clear();
                        testMatch = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testSkip != null)
                    {
                        testSkip.Clear();
                        testSkip = null;
                    }

                    ///////////////////////////////////////////////////////////

                    if (testReturnCodeMessages != null)
                    {
                        testReturnCodeMessages.Clear();
                        testReturnCodeMessages = null;
                    }

                    ///////////////////////////////////////////////////////////

#if DEBUGGER
                    if (testBreakpoints != null)
                    {
                        testBreakpoints.Clear();
                        testBreakpoints = null;
                    }
#endif

                    ///////////////////////////////////////////////////////////

                    testPath = null;
                    #endregion

                    ///////////////////////////////////////////////////////////

                    #region Variable Context
                    //
                    // NOTE: Dispose of the call stack now.
                    //
                    DisposeCallStack(ref callStack);

                    //
                    // NOTE: Dispose of the current, procedure, and global call
                    //       frames now.
                    //
                    DisposeCallFrame(ref currentFrame);
                    DisposeCallFrame(ref procedureFrame);
                    DisposeCallFrame(ref globalFrame);

                    ///////////////////////////////////////////////////////////

                    traceInfo = null;
                    #endregion
                }
            }
#endif
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Plugin, Command, & Package Disposal
        private void DisposePlugins(
            bool all
            )
        {
            if (plugins != null)
            {
                StringList keys = new StringList(plugins.Keys);

                foreach (string key in keys)
                {
                    IPlugin plugin = plugins[key];

                    if (plugin != null)
                    {
                        bool nativeCode = FlagOps.HasFlags(plugin.Flags,
                            PluginFlags.NativeCode, true);

                        if (all || !nativeCode)
                        {
                            string pluginName = plugin.Name;

                            ReturnCode code;
                            Result result = null;

                            try
                            {
                                code = plugin.Terminate(this, null, ref result);

                                if (code == ReturnCode.Ok)
                                    //
                                    // NOTE: Plugin successfully terminated, remove it.
                                    //
                                    plugins.Remove(key);
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while terminating {0} plugin \"{1}\": {2}",
                                    nativeCode ? "native" : "managed", pluginName, e);

                                code = ReturnCode.Error;
                            }

                            if (code != ReturnCode.Ok)
                                DebugOps.Complain(this, code, result);
                        }
                    }
                    else
                    {
                        plugins.Remove(key);
                    }
                }

                if (all || (plugins.Count == 0))
                    plugins = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeCommands(
            bool all
            )
        {
            if (commands != null)
            {
                StringList keys = new StringList(commands.Keys);

                foreach (string key in keys)
                {
                    ICommand command = commands[key];

                    if (command != null)
                    {
                        bool nativeCode = FlagOps.HasFlags(command.Flags, CommandFlags.NativeCode, true);

                        if (all || !nativeCode)
                        {
                            string commandName = command.Name;

                            ReturnCode code;
                            Result result = null;

                            try
                            {
                                code = command.Terminate(this, null, ref result);

                                if (code == ReturnCode.Ok)
                                    //
                                    // NOTE: Command successfully terminated, remove it.
                                    //
                                    commands.Remove(key);
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while terminating {0} command \"{1}\": {2}",
                                    nativeCode ? "native" : "managed", commandName, e);

                                code = ReturnCode.Error;
                            }

                            if (code != ReturnCode.Ok)
                                DebugOps.Complain(this, code, result);
                        }
                    }
                    else
                    {
                        commands.Remove(key);
                    }
                }

                if (all || (commands.Count == 0))
                    commands = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeHiddenCommands(
            bool all
            )
        {
            if (hiddenCommands != null)
            {
                StringList keys = new StringList(hiddenCommands.Keys);

                foreach (string key in keys)
                {
                    ICommand command = hiddenCommands[key];

                    if (command != null)
                    {
                        bool nativeCode = FlagOps.HasFlags(command.Flags, CommandFlags.NativeCode, true);

                        if (all || !nativeCode)
                        {
                            string commandName = command.Name;

                            ReturnCode code;
                            Result result = null;

                            try
                            {
                                code = command.Terminate(this, null, ref result);

                                if (code == ReturnCode.Ok)
                                    //
                                    // NOTE: Command successfully terminated, remove it.
                                    //
                                    hiddenCommands.Remove(key);
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while terminating {0} hidden command \"{1}\": {2}",
                                    nativeCode ? "native" : "managed", commandName, e);

                                code = ReturnCode.Error;
                            }

                            if (code != ReturnCode.Ok)
                                DebugOps.Complain(this, code, result);
                        }
                    }
                    else
                    {
                        hiddenCommands.Remove(key);
                    }
                }

                if (all || (hiddenCommands.Count == 0))
                    hiddenCommands = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposePackages()
        {
            if (packages != null)
            {
                foreach (KeyValuePair<string, _Wrappers.Package> pair in packages)
                {
                    if (pair.Value != null)
                    {
                        string packageName = pair.Value.Name;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = pair.Value.Terminate(this, null, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while terminating package \"{0}\": {1}",
                                packageName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                packages.Clear();
                packages = null;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Expression Entity Disposal
        private void DisposeFunctions()
        {
            if (functions != null)
            {
                foreach (KeyValuePair<string, _Wrappers.Function> pair in functions)
                {
                    if (pair.Value != null)
                    {
                        string functionName = pair.Value.Name;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = pair.Value.Terminate(this, null, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while terminating function \"{0}\": {1}",
                                functionName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                functions.Clear();
                functions = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeOperators()
        {
            if (operators != null)
            {
                foreach (KeyValuePair<string, _Wrappers.Operator> pair in operators)
                {
                    if (pair.Value != null)
                    {
                        string operatorName = pair.Value.Name;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = pair.Value.Terminate(this, null, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while terminating operator \"{0}\": {1}",
                                operatorName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                operators.Clear();
                operators = null;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Managed Entity Disposal
        private void DisposeObjects()
        {
            if (objects != null)
            {
                foreach (KeyValuePair<string, _Wrappers._Object> pair in objects)
                {
                    if (pair.Value != null)
                    {
                        if (pair.Value.Value != null)
                        {
                            //
                            // BUGFIX: Prevent accidental recursion.  Also, do NOT attempt to
                            //         dispose of objects that we do not own (i.e. the NoDispose
                            //         flag is set).
                            //
                            if (!Object.ReferenceEquals(pair.Value.Value, this) &&
                                !IsMasterOrSlaveInterpreter(pair.Value.Value) &&
                                !FlagOps.HasFlags(pair.Value.Flags, ObjectFlags.NoDispose, true))
                            {
                                string objectName = pair.Value.Name;

                                ReturnCode code;
                                Result result = null;

                                try
                                {
                                    //
                                    // NOTE: *SPECIAL CASE* The IObject interface uses the Value
                                    //       property to refer to the actual object, which is what
                                    //       we want to try and dispose of (i.e. and not the wrapper
                                    //       nor the underlying IObject implementation, which are
                                    //       usually just _Wrappers_Object and _Objects.Default,
                                    //       respectively).
                                    //
                                    code = RuntimeOps.TryDisposeObject(pair.Value.Value, ref result);
                                }
                                catch (Exception e)
                                {
                                    result = String.Format(
                                        "caught exception while disposing object \"{0}\": {1}",
                                        objectName, e);

                                    code = ReturnCode.Error;
                                }

                                if (code != ReturnCode.Ok)
                                    DebugOps.Complain(this, code, result);
                            }
                        }
                    }
                }

                objects.Clear();
                objects = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeChannels()
        {
            if (channels != null)
            {
                foreach (KeyValuePair<string, Channel> pair in channels)
                {
                    if (pair.Value != null)
                    {
                        string channelName = pair.Value.ToString();

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            pair.Value.Close(); /* throw */

                            code = ReturnCode.Ok;
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while closing channel \"{0}\": {1}",
                                channelName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                channels.Clear();
                channels = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if APPDOMAINS
        private void DisposeAppDomains()
        {
            if (appDomains != null)
            {
                foreach (KeyValuePair<string, AppDomain> pair in appDomains)
                {
                    if (pair.Value != null)
                    {
                        string appDomainName = pair.Value.FriendlyName;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = RuntimeOps.UnloadAppDomain(
                                pair.Value, null, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while unloading application domain \"{0}\": {1}",
                                appDomainName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                appDomains.Clear();
                appDomains = null;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Database Entity Disposal
#if DATA
        private void DisposeTransactions()
        {
            if (transactions != null)
            {
                foreach (KeyValuePair<string, IDbTransaction> pair in transactions)
                {
                    if (pair.Value != null)
                    {
                        string transactionName = pair.Value.ToString();

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = RuntimeOps.TryDisposeObject(pair.Value, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while disposing transaction \"{0}\": {1}",
                                transactionName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                transactions.Clear();
                transactions = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeConnections()
        {
            if (connections != null)
            {
                foreach (KeyValuePair<string, IDbConnection> pair in connections)
                {
                    if (pair.Value != null)
                    {
                        string connectionName = pair.Value.ToString();

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = RuntimeOps.TryDisposeObject(pair.Value, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while disposing connection \"{0}\": {1}",
                                connectionName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                connections.Clear();
                connections = null;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Native Entity Disposal
#if NATIVE && LIBRARY
        private void DisposeDelegates()
        {
            if (delegates != null)
            {
                foreach (KeyValuePair<string, _Wrappers.Delegate> pair in delegates)
                {
                    if (pair.Value != null)
                    {
                        string delegateName = pair.Value.Name;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = RuntimeOps.TryDisposeObject(pair.Value, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while disposing delegate \"{0}\": {1}",
                                delegateName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                delegates.Clear();
                delegates = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeModules()
        {
            if (modules != null)
            {
                foreach (KeyValuePair<string, _Wrappers._Module> pair in modules)
                {
                    if (pair.Value != null)
                    {
                        string moduleName = pair.Value.Name;

                        ReturnCode code;
                        Result result = null;

                        try
                        {
                            code = RuntimeOps.TryDisposeObject(pair.Value, ref result);
                        }
                        catch (Exception e)
                        {
                            result = String.Format(
                                "caught exception while disposing module \"{0}\": {1}",
                                moduleName, e);

                            code = ReturnCode.Error;
                        }

                        if (code != ReturnCode.Ok)
                            DebugOps.Complain(this, code, result);
                    }
                }

                modules.Clear();
                modules = null;
            }
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Event Manager, Debugger, & Host Disposal
        private void DisposeEventManager()
        {
            if (eventManager != null)
            {
                ReturnCode code;
                Result error = null;

                try
                {
                    code = RuntimeOps.TryDisposeObject(eventManager, ref error);
                }
                catch (Exception e)
                {
                    error = String.Format(
                        "caught exception while disposing event manager: {0}",
                        e);

                    code = ReturnCode.Error;
                }

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error);

                eventManager = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if DEBUGGER && !THREADING
        private void DisposeDebugger()
        {
            if (debugger != null)
            {
                ReturnCode code;
                Result error = null;

                try
                {
                    code = RuntimeOps.TryDisposeObject(debugger, ref error);
                }
                catch (Exception e)
                {
                    error = String.Format(
                        "caught exception while disposing debugger: {0}",
                        e);

                    code = ReturnCode.Error;
                }

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(this, code, error);

                debugger = null;
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void DisposeHost()
        {
            #region Dispose Host
            if (host != null)
            {
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoDisposeHost, true))
                {
                    ReturnCode code;
                    Result error = null;

                    try
                    {
                        code = RuntimeOps.TryDisposeObject(host, ref error);
                    }
                    catch (Exception e)
                    {
                        error = String.Format(
                            "caught exception while disposing host: {0}",
                            e);

                        code = ReturnCode.Error;
                    }

                    if (code != ReturnCode.Ok)
                        DebugOps.Complain(this, code, error);
                }

                host = null;
            }
            #endregion

            ///////////////////////////////////////////////////////////////////////////////////////////

            #region Dispose Core Host (ISOLATED_PLUGINS only)
#if ISOLATED_PLUGINS
            if (coreHost != null)
            {
                if (!FlagOps.HasFlags(createFlags, CreateFlags.NoDisposeHost, true))
                {
                    ReturnCode code;
                    Result error = null;

                    try
                    {
                        code = RuntimeOps.TryDisposeObject(coreHost, ref error);
                    }
                    catch (Exception e)
                    {
                        error = String.Format(
                            "caught exception while disposing core host: {0}",
                            e);

                        code = ReturnCode.Error;
                    }

                    if (code != ReturnCode.Ok)
                        DebugOps.Complain(this, code, error);
                }

                coreHost = null;
            }
#endif
            #endregion
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Slave Interpreter Disposal
        private void DisposeSlaveInterpreters()
        {
            if (slaveInterpreters != null)
            {
                foreach (KeyValuePair<string, Interpreter> pair in slaveInterpreters)
                {
                    if (pair.Value != null)
                    {
                        if (!Object.ReferenceEquals(pair.Value, this))
                        {
                            ReturnCode code;
                            Result result = null;

                            try
                            {
                                code = RuntimeOps.TryDisposeObject(pair.Value, ref result);
                            }
                            catch (Exception e)
                            {
                                result = String.Format(
                                    "caught exception while disposing slave interpreter \"{0}\": {1}",
                                    pair.Value, e);

                                code = ReturnCode.Error;
                            }

                            if (code != ReturnCode.Ok)
                                DebugOps.Complain(this, code, result);
                        }
                    }
                }

                slaveInterpreters.Clear();
                slaveInterpreters = null;
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Disposal Helper Methods (Reset/Clear)
#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
        private void ResetCaches()
        {
#if LIST_CACHE
            if (stringListCache != null)
            {
                stringListCache.Clear();
                stringListCache = null;
            }
#endif

#if PARSE_CACHE
            if (parseStateCache != null)
            {
                parseStateCache.Clear();
                parseStateCache = null;
            }

            #region Dead Code
#if DEAD_CODE
            if (expressionStateCache != null)
            {
                expressionStateCache.Clear();
                expressionStateCache = null;
            }
#endif
            #endregion
#endif

#if EXECUTE_CACHE
            if (executeCache != null)
            {
                executeCache.Clear();
                executeCache = null;
            }
#endif

#if TYPE_CACHE
            if (typeCache != null)
            {
                typeCache.Clear();
                typeCache = null;
            }
#endif

#if COM_TYPE_CACHE
            if (comTypeListCache != null)
            {
                comTypeListCache.Clear();
                comTypeListCache = null;
            }
#endif
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void ResetEarlyPodObjects()
        {
            if (aliases != null)
            {
                aliases.Clear();
                aliases = null;
            }

            if (procedures != null)
            {
                procedures.Clear();
                procedures = null;
            }

            if (hiddenProcedures != null)
            {
                hiddenProcedures.Clear();
                hiddenProcedures = null;
            }

            if (executes != null)
            {
                executes.Clear();
                executes = null;
            }

            if (hiddenExecutes != null)
            {
                hiddenExecutes.Clear();
                hiddenExecutes = null;
            }

            if (objectInterfaces != null)
            {
                objectInterfaces.Clear();
                objectInterfaces = null;
            }

            if (objectNamespaces != null)
            {
                objectNamespaces.Clear();
                objectNamespaces = null;
            }

            if (objectTypes != null)
            {
                objectTypes.Clear();
                objectTypes = null;
            }

#if CALLBACK_QUEUE
            if (callbackQueue != null)
            {
                callbackQueue.Clear();
                callbackQueue = null;
            }
#endif

            if (callbacks != null)
            {
                callbacks.Clear();
                callbacks = null;
            }

            if (scopes != null)
            {
                scopes.Clear();
                scopes = null;
            }

            if (cryptoRandom != null)
                cryptoRandom = null;

            if (random != null)
                random = null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void ResetLatePodObjects()
        {
            if (resolvers != null)
            {
                resolvers.Clear();
                resolvers = null;
            }

            if (runtimeOptions != null)
            {
                runtimeOptions.Clear();
                runtimeOptions = null;
            }

            if (traces != null)
            {
                traces.Clear();
                traces = null;
            }

            if (policies != null)
            {
                policies.Clear();
                policies = null;
            }

            if (packageIndexes != null)
            {
                packageIndexes.Clear();
                packageIndexes = null;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void ResetUserObjects()
        {
            if (userObject != null)
                userObject = null; /* WARNING: Not owner, do NOT dispose. */

            if (resolverObject != null)
                resolverObject = null; /* WARNING: Not owner, do NOT dispose. */

            if (policyObject != null)
                policyObject = null; /* WARNING: Not owner, do NOT dispose. */

            if (applicationObject != null)
                applicationObject = null; /* WARNING: Not owner, do NOT dispose. */
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable "Pattern" Members
        internal bool Deleted
        {
            get
            {
                lock (syncRoot)
                {
                    //
                    // NOTE: This flag is special.  Non-zero indicates that the
                    //       interpreter is currently being disposed; however,
                    //       the disposal process has not yet completed.  This
                    //       flag is checked and enforced by the Ready method
                    //       to prevent scripts from being evaluated during the
                    //       disposal process.
                    //
                    return deleted;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public bool Disposed
        {
            get { return disposed; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool IsPendingDispose(
            Interpreter interpreter
            )
        {
            return (interpreter != null) && (interpreter.disposeCount > 0);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static bool IsDeletedOrDisposed(
            Interpreter interpreter
            )
        {
            return (interpreter != null) && (interpreter.deleted || interpreter.disposed);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        internal static bool IsDisposed(
            Interpreter interpreter
            )
        {
            return (interpreter != null) && interpreter.disposed;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private void CheckDisposed() /* throw */
        {
#if THROW_ON_DISPOSED
            if (disposed && Engine.IsThrowOnDisposed(this, true))
                throw new ObjectDisposedException(typeof(Interpreter).Name);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private /* protected virtual */ void Dispose(bool disposing)
        {
            RuntimeOps.DebugTrace(String.Format(
                "Dispose: called via {0} for interpreter {1} on thread {2} " +
                "(primary thread is {3}), count is {4}, wait count is {5}, " +
                "wait spin count is {6}, already disposed is {7}{8}", disposing ?
                    typeof(IDisposable).Name : "destructor", id,
                GlobalState.GetCurrentThreadId(), threadId, disposeCount,
                waitCount, waitSpinCount, disposed, Environment.NewLine),
                typeof(Interpreter).Name);

            if (!disposed)
            {
                if (Interlocked.Increment(ref disposeCount) == 1)
                {
#if NOTIFY || NOTIFY_OBJECT
                    //
                    // BUGFIX: Prevent notifications from firing and indirectly
                    //         recreating anything we have torn down (especially
                    //         thread specific data).
                    //
                    notifyFlags |= NotifyFlags.NoNotify;
#endif

                    //
                    // NOTE: Allow callbacks to figure out which interpreter is
                    //       terminating them (even if they were not directly
                    //       passed an Interpreter object).
                    //
                    GlobalState.PushActiveInterpreter(this);

                    //
                    // NOTE: Grab the lock [temporarily] to set the deleted
                    //       flag for the interpreter.
                    //
                    lock (syncRoot)
                    {
                        //
                        // NOTE: Mark this interpreter as "deleted" (i.e. "deletion
                        //       pending") so that any "user callbacks" do not attempt
                        //       to get too clever with their shutdown/cleanup code
                        //       (i.e. no script evaluation, etc).
                        //
                        if (!deleted)
                            deleted = true;
                    }

                    //
                    // NOTE: Unblock any and all pending wait calls now.  When they
                    //       awaken, they should realize that the interpreter has been
                    //       deleted [out from under them] and bail out.
                    //
                    SignalVariableEvent();

                    //
                    // NOTE: Used to keep track of how long we waited for the other
                    //       thread(s) to exit their variable wait states.
                    //
                    int elapsedMilliseconds = 0;

                    //
                    // HACK: Now, block until the pending calls to WaitVariable are
                    //       all cleared.  This is potentially dangerous because it
                    //       could block on the GC thread (?).  This loop is designed
                    //       to keep waiting until the wait count is zero OR until
                    //       the maximum timeout is reached (currently about 3000
                    //       milliseconds).  Also, since the deleted flag has already
                    //       been set for the interpreter (above), we should not have
                    //       to worry about waits being started after this point.
                    //
                    // BUGFIX: Calling the GetMinimumSleepTime method here was causing
                    //         an exception because the event manager was already
                    //         disposed via the GC.  In the future, much more care is
                    //         required to prevent this method from accessing anything
                    //         that can throw ObjectDisposedException.
                    //
                    int sleepMilliseconds = GetMinimumSleepTime();

                    while (Interlocked.CompareExchange(ref waitCount, 0, 0) != 0)
                    {
                        //
                        // NOTE: Wait for a while to give other threads a chance to
                        //       unwind out of their variable wait states.
                        //
                        RuntimeOps.Sleep(sleepMilliseconds);

                        //
                        // NOTE: Keep track of the total elapsed wait time.  If this
                        //       value exceeds the total [default] allowed wait time,
                        //       report this situation as an error and bail out.
                        //
                        elapsedMilliseconds += sleepMilliseconds;

                        if (elapsedMilliseconds > ThreadOps.DefaultJoinTimeout)
                        {
                            RuntimeOps.DebugTrace(String.Format(
                                "{0}: Dispose: timeout waiting, interpreter = {1}, " +
                                "sleepMilliseconds = {2}, elapsedMilliseconds = {3}{4}",
                                GlobalState.GetCurrentThreadId(), id, sleepMilliseconds,
                                elapsedMilliseconds, Environment.NewLine),
                                typeof(Interpreter).Name);

                            break;
                        }
                    }

                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: Dispose: done waiting, interpreter = {1}, " +
                        "sleepMilliseconds = {2}, elapsedMilliseconds = {3}{4}",
                        GlobalState.GetCurrentThreadId(), id, sleepMilliseconds,
                        elapsedMilliseconds, Environment.NewLine),
                        typeof(Interpreter).Name);

                    //
                    // NOTE: We obtain and hold the lock on the Interpreter
                    //       SyncRoot for almost the entire duration of the
                    //       cleanup.
                    //
                    lock (syncRoot)
                    {
                        //
                        // NOTE: If the interpreter was read-only, we need to
                        //       unset that before proceeding; otherwise,
                        //       commands and plugins cannot be terminated
                        //       gracefully.
                        //
                        if (readOnly)
                            readOnly = false;

                        //
                        // BUGFIX: Disposal ordering issue.  We do not care about
                        //         this if the host has already been disposed.
                        //
                        try
                        {
                            //
                            // NOTE: Mark host as in "exit mode" for additional
                            //       help in troubleshooting any shutdown issues.
                            //       This does not need to be undone later because
                            //       the host will also be disposed below.
                            //
                            if ((host != null) &&
                                FlagOps.HasFlags(host.GetFlags(), HostFlags.Exit, true))
                            {
                                host.Exiting = true;
                            }
                        }
                        catch
                        {
                            // do nothing.
                        }

                        //
                        // NOTE: Are we disposing (or being called via the finalizer)?
                        //
                        if (disposing)
                        {
                            ////////////////////////////////////
                            // dispose managed resources here...
                            ////////////////////////////////////

                            //
                            // NOTE: Dispose any nested interpreters first.
                            //
                            DisposeSlaveInterpreters();

                            //
                            // NOTE: Attempt to terminuate managed plugins
                            //       gracefully.
                            //
                            DisposePlugins(false);

                            //
                            // NOTE: Attempt to terminuate managed commands
                            //       gracefully.
                            //
                            DisposeCommands(false);
                            DisposeHiddenCommands(false);

                            //
                            // NOTE: Attempt to terminuate user commands,
                            //       functions, and operators gracefully.
                            //
                            DisposeOperators();
                            DisposeFunctions();

#if DATA
                            //
                            // NOTE: Attempt to dispose our database related
                            //       objects, if any.
                            //
                            DisposeTransactions();
                            DisposeConnections();
#endif

                            //
                            // NOTE: Attempt to dispose all script visible
                            //       objects.
                            //
                            DisposeObjects();

                            //
                            // NOTE: Dispose of the open channels, if any.
                            //
                            DisposeChannels();

#if LIST_CACHE || PARSE_CACHE || EXECUTE_CACHE || TYPE_CACHE || COM_TYPE_CACHE
                            //
                            // NOTE: Clear our cached data.
                            //
                            ResetCaches();
#endif

                            //
                            // NOTE: Dispose the event manager, if any.
                            //
                            DisposeEventManager();

                            //
                            // NOTE: Clear all the "plain old data" objects
                            //       associated with this interpreter.
                            //
                            ResetEarlyPodObjects();
                        }

                        //////////////////////////////////////
                        // release unmanaged resources here...
                        //////////////////////////////////////

                        //
                        // NOTE: Attempt to terminuate all the remaining
                        //       (unmanaged) plugins gracefully.
                        //
                        DisposePlugins(true);

                        //
                        // NOTE: Attempt to terminuate all the remaining
                        //       (unmanaged) commands gracefully.
                        //
                        DisposeCommands(true);
                        DisposeHiddenCommands(true);

                        //
                        // NOTE: Dispose the packages now.  We do this after
                        //       the plugins because they may wish to withdraw
                        //       their associated package(s).
                        //
                        DisposePackages();

#if NATIVE && LIBRARY
                        //
                        // NOTE: Attempt to dispose the native delegates, if
                        //       any.
                        //
                        DisposeDelegates();

                        //
                        // NOTE: Attempt to dispose the native modules, if any.
                        //
                        DisposeModules();
#endif

#if APPDOMAINS
                        //
                        // NOTE: Dispose (i.e. unload) AppDomains that we own,
                        //       if any.
                        //
                        DisposeAppDomains();
#endif

                        //
                        // NOTE: Policies, packages, and package indexes, etc
                        //       are simply data, clear them now.  We have to
                        //       wait until this point in case plugins refer
                        //       to any of them while unloading.
                        //
                        ResetLatePodObjects();

#if DEBUGGER && !THREADING
                        //
                        // NOTE: Dispose our debugger, if any, now.
                        //
                        DisposeDebugger();
#endif

                        //
                        // NOTE: Close the variable event.
                        //
                        ThreadOps.CloseEvent(ref variableEvent);

                        //
                        // NOTE: Reset the user-defined and application-defined
                        //       objects.  We do not "dispose" these objects
                        //       because they do not belong to us.
                        //
                        ResetUserObjects();
                    }

#if NATIVE && TCL
                    //
                    // NOTE: Dispose of our [native] Tcl related objects,
                    //       if any, now.  If we are on the wrong thread, the
                    //       orphaned objects will be turned into "zombies" for
                    //       later pickup by the next Interpreter object to be
                    //       created on this thread.  Complaints (i.e. priority
                    //       error messages) will be reported if any errors are
                    //       encountered during this disposal process.
                    //
                    // HACK: Perhaps stronger measures are needed here to help
                    //       prevent resource leaks?  However, the Dispose
                    //       method is a very limited place.  For one thing, it
                    //       has no means of communicating failure to the caller,
                    //       which may be the GC thread (i.e. it is not supposed
                    //       to throw exceptions and it returns no results).
                    //
                    // WARNING: All users of Interpreter objects are STRONGLY
                    //          advised and encouraged to Dispose of their
                    //          Interpreter objects [on the same thread they
                    //          were created on] just as soon as they are
                    //          finished with them; otherwise, all the [native]
                    //          resources used by them will be disposed of
                    //          on a "best-effort" basis.
                    //
                    DisposeTcl(true, disposing, false);
#endif

                    //
                    // NOTE: Dispose all the thread-specific data for the
                    //       interpreter for this thread [only].
                    //
                    DisposeThread(true, disposing);

                    //
                    // NOTE: Dispose the host environment last.  This must now
                    //       be done without holding the lock on the Interpreter
                    //       SyncRoot; otherwise, a deadlock can occur.
                    //
                    //       Deadlock Example:
                    //
                    //       1. This thread (thread #1) holds the lock on the
                    //          Interpreter SyncRoot and wants the implicit lock
                    //          on the System.Console object to remove the
                    //          ConsoleCancelEventHandler.
                    //
                    //       2. The other thread (thread #2) holds the implicit
                    //          lock on the System.Console object while the call
                    //          to CancelEvaluate is pending awaiting the lock
                    //          on the Interpreter SyncRoot.
                    //
                    DisposeHost();

                    //
                    // NOTE: Remove this interpreter from the active interpreter
                    //       stack.
                    //
                    /* IGNORED */
                    GlobalState.PopActiveInterpreter();

                    //
                    // NOTE: Finally, remove this interpreter from the master
                    //       interpreter list.
                    //
                    /* IGNORED */
                    GlobalState.RemoveInterpreter(this);

                    //
                    // NOTE: Grab the global notify setting now because we need it
                    //       after marking the interpreter as disposed.
                    //
                    bool notify = this.GlobalNotify;

                    //
                    // NOTE: This object is now fully disposed.  We have reset
                    //       all of the internal data necessary for this object to
                    //       function correctly and there is currently no supported
                    //       way to "revive" it; therefore, it should never be used
                    //       again after this point.  This constraint is [normally]
                    //       enforced by the CheckDisposed method, which is called
                    //       from every publically accessible property and method
                    //       of this class.
                    //
                    disposed = true;

#if NOTIFY && NOTIFY_GLOBAL
                    //
                    // BUGFIX: Only notify other interpreters if the global notify
                    //         setting is enabled.  This prevents subtle locking
                    //         issues when tearing down interpreters that belong to
                    //         different threads [OR interpreters that are currently
                    //         being used from different threads].
                    //
                    if (notify)
                        /* IGNORED */
                        CheckNotifications(
                            null, false,
                            NotifyType.Interpreter, NotifyFlags.Removed,
                            null, this,
                            null, null, null);
#endif
                }
            }

            RuntimeOps.DebugTrace(String.Format(
                "{0}: Dispose: exited, interpreter = {1}, disposing = {2}{3}",
                GlobalState.GetCurrentThreadId(), id, disposing,
                Environment.NewLine), typeof(Interpreter).Name);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Object Overrides
        public override bool Equals(object obj)
        {
            CheckDisposed();

            return Object.ReferenceEquals(this, obj);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override int GetHashCode()
        {
            CheckDisposed();

            return ConversionOps.ToInt(id);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override string ToString()
        {
            CheckDisposed();

            return id.ToString();
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable Members
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Static Constructor
#if NATIVE && WINDOWS
        static Interpreter()
        {
            SetupOps.CreateMutexes();
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Constructors
        private Interpreter()
        {
            //
            // NOTE: This constructor is private because Interpreter objects MUST
            //       be created via the static Interpreter.Create factory method.
            //
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Destructor
        ~Interpreter()
        {
            Dispose(false);
        }
        #endregion
    }
}
